/* Copyright (c) 2016, 2024, Oracle and/or its affiliates.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License, version 2.0,
   as published by the Free Software Foundation.

   This program is designed to work with certain software (including
   but not limited to OpenSSL) that is licensed under separate terms,
   as designated in a particular file or component or in included license
   documentation.  The authors of MySQL hereby grant you an additional
   permission to link the program and your derivative works with the
   separately licensed software that they have either included with
   the program or referenced in the documentation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License, version 2.0, for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */

/*
  This file is generated, do not edit.
  See file sql/gen_lex_token.cc.
*/
struct lex_token_string
{
  const char *m_token_string;
  int m_token_length;
  bool m_append_space;
  bool m_start_expr;
};
typedef struct lex_token_string lex_token_string;
#ifdef LEX_TOKEN_WITH_DEFINITION
lex_token_string lex_token_array[]=
{
/* PART 1: character tokens. */
/* 000 */  { "\x00", 1, true, false},
/* 001 */  { "\x01", 1, true, false},
/* 002 */  { "\x02", 1, true, false},
/* 003 */  { "\x03", 1, true, false},
/* 004 */  { "\x04", 1, true, false},
/* 005 */  { "\x05", 1, true, false},
/* 006 */  { "\x06", 1, true, false},
/* 007 */  { "\x07", 1, true, false},
/* 008 */  { "\x08", 1, true, false},
/* 009 */  { "\x09", 1, true, false},
/* 010 */  { "\x0a", 1, true, false},
/* 011 */  { "\x0b", 1, true, false},
/* 012 */  { "\x0c", 1, true, false},
/* 013 */  { "\x0d", 1, true, false},
/* 014 */  { "\x0e", 1, true, false},
/* 015 */  { "\x0f", 1, true, false},
/* 016 */  { "\x10", 1, true, false},
/* 017 */  { "\x11", 1, true, false},
/* 018 */  { "\x12", 1, true, false},
/* 019 */  { "\x13", 1, true, false},
/* 020 */  { "\x14", 1, true, false},
/* 021 */  { "\x15", 1, true, false},
/* 022 */  { "\x16", 1, true, false},
/* 023 */  { "\x17", 1, true, false},
/* 024 */  { "\x18", 1, true, false},
/* 025 */  { "\x19", 1, true, false},
/* 026 */  { "\x1a", 1, true, false},
/* 027 */  { "\x1b", 1, true, false},
/* 028 */  { "\x1c", 1, true, false},
/* 029 */  { "\x1d", 1, true, false},
/* 030 */  { "\x1e", 1, true, false},
/* 031 */  { "\x1f", 1, true, false},
/* 032 */  { "\x20", 1, true, false},
/* 033 */  { "\x21", 1, true, false},
/* 034 */  { "\x22", 1, true, false},
/* 035 */  { "\x23", 1, true, false},
/* 036 */  { "\x24", 1, true, false},
/* 037 */  { "\x25", 1, true, true},
/* 038 */  { "\x26", 1, true, true},
/* 039 */  { "\x27", 1, true, false},
/* 040 */  { "\x28", 1, true, true},
/* 041 */  { "\x29", 1, true, false},
/* 042 */  { "\x2a", 1, true, true},
/* 043 */  { "\x2b", 1, true, true},
/* 044 */  { "\x2c", 1, true, true},
/* 045 */  { "\x2d", 1, true, true},
/* 046 */  { "\x2e", 1, true, false},
/* 047 */  { "\x2f", 1, true, true},
/* 048 */  { "\x30", 1, true, false},
/* 049 */  { "\x31", 1, true, false},
/* 050 */  { "\x32", 1, true, false},
/* 051 */  { "\x33", 1, true, false},
/* 052 */  { "\x34", 1, true, false},
/* 053 */  { "\x35", 1, true, false},
/* 054 */  { "\x36", 1, true, false},
/* 055 */  { "\x37", 1, true, false},
/* 056 */  { "\x38", 1, true, false},
/* 057 */  { "\x39", 1, true, false},
/* 058 */  { "\x3a", 1, true, false},
/* 059 */  { "\x3b", 1, true, false},
/* 060 */  { "\x3c", 1, true, false},
/* 061 */  { "\x3d", 1, true, false},
/* 062 */  { "\x3e", 1, true, false},
/* 063 */  { "\x3f", 1, true, false},
/* 064 */  { "\x40", 1, false, false},
/* 065 */  { "\x41", 1, true, false},
/* 066 */  { "\x42", 1, true, false},
/* 067 */  { "\x43", 1, true, false},
/* 068 */  { "\x44", 1, true, false},
/* 069 */  { "\x45", 1, true, false},
/* 070 */  { "\x46", 1, true, false},
/* 071 */  { "\x47", 1, true, false},
/* 072 */  { "\x48", 1, true, false},
/* 073 */  { "\x49", 1, true, false},
/* 074 */  { "\x4a", 1, true, false},
/* 075 */  { "\x4b", 1, true, false},
/* 076 */  { "\x4c", 1, true, false},
/* 077 */  { "\x4d", 1, true, false},
/* 078 */  { "\x4e", 1, true, false},
/* 079 */  { "\x4f", 1, true, false},
/* 080 */  { "\x50", 1, true, false},
/* 081 */  { "\x51", 1, true, false},
/* 082 */  { "\x52", 1, true, false},
/* 083 */  { "\x53", 1, true, false},
/* 084 */  { "\x54", 1, true, false},
/* 085 */  { "\x55", 1, true, false},
/* 086 */  { "\x56", 1, true, false},
/* 087 */  { "\x57", 1, true, false},
/* 088 */  { "\x58", 1, true, false},
/* 089 */  { "\x59", 1, true, false},
/* 090 */  { "\x5a", 1, true, false},
/* 091 */  { "\x5b", 1, true, false},
/* 092 */  { "\x5c", 1, true, false},
/* 093 */  { "\x5d", 1, true, false},
/* 094 */  { "\x5e", 1, true, true},
/* 095 */  { "\x5f", 1, true, false},
/* 096 */  { "\x60", 1, true, false},
/* 097 */  { "\x61", 1, true, false},
/* 098 */  { "\x62", 1, true, false},
/* 099 */  { "\x63", 1, true, false},
/* 100 */  { "\x64", 1, true, false},
/* 101 */  { "\x65", 1, true, false},
/* 102 */  { "\x66", 1, true, false},
/* 103 */  { "\x67", 1, true, false},
/* 104 */  { "\x68", 1, true, false},
/* 105 */  { "\x69", 1, true, false},
/* 106 */  { "\x6a", 1, true, false},
/* 107 */  { "\x6b", 1, true, false},
/* 108 */  { "\x6c", 1, true, false},
/* 109 */  { "\x6d", 1, true, false},
/* 110 */  { "\x6e", 1, true, false},
/* 111 */  { "\x6f", 1, true, false},
/* 112 */  { "\x70", 1, true, false},
/* 113 */  { "\x71", 1, true, false},
/* 114 */  { "\x72", 1, true, false},
/* 115 */  { "\x73", 1, true, false},
/* 116 */  { "\x74", 1, true, false},
/* 117 */  { "\x75", 1, true, false},
/* 118 */  { "\x76", 1, true, false},
/* 119 */  { "\x77", 1, true, false},
/* 120 */  { "\x78", 1, true, false},
/* 121 */  { "\x79", 1, true, false},
/* 122 */  { "\x7a", 1, true, false},
/* 123 */  { "\x7b", 1, true, false},
/* 124 */  { "\x7c", 1, true, true},
/* 125 */  { "\x7d", 1, true, false},
/* 126 */  { "\x7e", 1, true, false},
/* 127 */  { "\x7f", 1, true, false},
/* 128 */  { "\x80", 1, true, false},
/* 129 */  { "\x81", 1, true, false},
/* 130 */  { "\x82", 1, true, false},
/* 131 */  { "\x83", 1, true, false},
/* 132 */  { "\x84", 1, true, false},
/* 133 */  { "\x85", 1, true, false},
/* 134 */  { "\x86", 1, true, false},
/* 135 */  { "\x87", 1, true, false},
/* 136 */  { "\x88", 1, true, false},
/* 137 */  { "\x89", 1, true, false},
/* 138 */  { "\x8a", 1, true, false},
/* 139 */  { "\x8b", 1, true, false},
/* 140 */  { "\x8c", 1, true, false},
/* 141 */  { "\x8d", 1, true, false},
/* 142 */  { "\x8e", 1, true, false},
/* 143 */  { "\x8f", 1, true, false},
/* 144 */  { "\x90", 1, true, false},
/* 145 */  { "\x91", 1, true, false},
/* 146 */  { "\x92", 1, true, false},
/* 147 */  { "\x93", 1, true, false},
/* 148 */  { "\x94", 1, true, false},
/* 149 */  { "\x95", 1, true, false},
/* 150 */  { "\x96", 1, true, false},
/* 151 */  { "\x97", 1, true, false},
/* 152 */  { "\x98", 1, true, false},
/* 153 */  { "\x99", 1, true, false},
/* 154 */  { "\x9a", 1, true, false},
/* 155 */  { "\x9b", 1, true, false},
/* 156 */  { "\x9c", 1, true, false},
/* 157 */  { "\x9d", 1, true, false},
/* 158 */  { "\x9e", 1, true, false},
/* 159 */  { "\x9f", 1, true, false},
/* 160 */  { "\xa0", 1, true, false},
/* 161 */  { "\xa1", 1, true, false},
/* 162 */  { "\xa2", 1, true, false},
/* 163 */  { "\xa3", 1, true, false},
/* 164 */  { "\xa4", 1, true, false},
/* 165 */  { "\xa5", 1, true, false},
/* 166 */  { "\xa6", 1, true, false},
/* 167 */  { "\xa7", 1, true, false},
/* 168 */  { "\xa8", 1, true, false},
/* 169 */  { "\xa9", 1, true, false},
/* 170 */  { "\xaa", 1, true, false},
/* 171 */  { "\xab", 1, true, false},
/* 172 */  { "\xac", 1, true, false},
/* 173 */  { "\xad", 1, true, false},
/* 174 */  { "\xae", 1, true, false},
/* 175 */  { "\xaf", 1, true, false},
/* 176 */  { "\xb0", 1, true, false},
/* 177 */  { "\xb1", 1, true, false},
/* 178 */  { "\xb2", 1, true, false},
/* 179 */  { "\xb3", 1, true, false},
/* 180 */  { "\xb4", 1, true, false},
/* 181 */  { "\xb5", 1, true, false},
/* 182 */  { "\xb6", 1, true, false},
/* 183 */  { "\xb7", 1, true, false},
/* 184 */  { "\xb8", 1, true, false},
/* 185 */  { "\xb9", 1, true, false},
/* 186 */  { "\xba", 1, true, false},
/* 187 */  { "\xbb", 1, true, false},
/* 188 */  { "\xbc", 1, true, false},
/* 189 */  { "\xbd", 1, true, false},
/* 190 */  { "\xbe", 1, true, false},
/* 191 */  { "\xbf", 1, true, false},
/* 192 */  { "\xc0", 1, true, false},
/* 193 */  { "\xc1", 1, true, false},
/* 194 */  { "\xc2", 1, true, false},
/* 195 */  { "\xc3", 1, true, false},
/* 196 */  { "\xc4", 1, true, false},
/* 197 */  { "\xc5", 1, true, false},
/* 198 */  { "\xc6", 1, true, false},
/* 199 */  { "\xc7", 1, true, false},
/* 200 */  { "\xc8", 1, true, false},
/* 201 */  { "\xc9", 1, true, false},
/* 202 */  { "\xca", 1, true, false},
/* 203 */  { "\xcb", 1, true, false},
/* 204 */  { "\xcc", 1, true, false},
/* 205 */  { "\xcd", 1, true, false},
/* 206 */  { "\xce", 1, true, false},
/* 207 */  { "\xcf", 1, true, false},
/* 208 */  { "\xd0", 1, true, false},
/* 209 */  { "\xd1", 1, true, false},
/* 210 */  { "\xd2", 1, true, false},
/* 211 */  { "\xd3", 1, true, false},
/* 212 */  { "\xd4", 1, true, false},
/* 213 */  { "\xd5", 1, true, false},
/* 214 */  { "\xd6", 1, true, false},
/* 215 */  { "\xd7", 1, true, false},
/* 216 */  { "\xd8", 1, true, false},
/* 217 */  { "\xd9", 1, true, false},
/* 218 */  { "\xda", 1, true, false},
/* 219 */  { "\xdb", 1, true, false},
/* 220 */  { "\xdc", 1, true, false},
/* 221 */  { "\xdd", 1, true, false},
/* 222 */  { "\xde", 1, true, false},
/* 223 */  { "\xdf", 1, true, false},
/* 224 */  { "\xe0", 1, true, false},
/* 225 */  { "\xe1", 1, true, false},
/* 226 */  { "\xe2", 1, true, false},
/* 227 */  { "\xe3", 1, true, false},
/* 228 */  { "\xe4", 1, true, false},
/* 229 */  { "\xe5", 1, true, false},
/* 230 */  { "\xe6", 1, true, false},
/* 231 */  { "\xe7", 1, true, false},
/* 232 */  { "\xe8", 1, true, false},
/* 233 */  { "\xe9", 1, true, false},
/* 234 */  { "\xea", 1, true, false},
/* 235 */  { "\xeb", 1, true, false},
/* 236 */  { "\xec", 1, true, false},
/* 237 */  { "\xed", 1, true, false},
/* 238 */  { "\xee", 1, true, false},
/* 239 */  { "\xef", 1, true, false},
/* 240 */  { "\xf0", 1, true, false},
/* 241 */  { "\xf1", 1, true, false},
/* 242 */  { "\xf2", 1, true, false},
/* 243 */  { "\xf3", 1, true, false},
/* 244 */  { "\xf4", 1, true, false},
/* 245 */  { "\xf5", 1, true, false},
/* 246 */  { "\xf6", 1, true, false},
/* 247 */  { "\xf7", 1, true, false},
/* 248 */  { "\xf8", 1, true, false},
/* 249 */  { "\xf9", 1, true, false},
/* 250 */  { "\xfa", 1, true, false},
/* 251 */  { "\xfb", 1, true, false},
/* 252 */  { "\xfc", 1, true, false},
/* 253 */  { "\xfd", 1, true, false},
/* 254 */  { "\xfe", 1, true, false},
/* 255 */  { "\xff", 1, true, false},
/* PART 2: named tokens from sql/sql_yacc.yy (chunk 1). */
/* 256 */  { "(unknown)", 9, true, false},
/* 257 */  { "(unknown)", 9, true, false},
/* 258 */  { "(unknown)", 9, true, false},
/* 259 */  { "ACCESSIBLE", 10, true, false},
/* 260 */  { "ACCOUNT", 7, true, false},
/* 261 */  { "ACTION", 6, true, false},
/* 262 */  { "ADD", 3, true, false},
/* 263 */  { "ADDDATE", 7, true, false},
/* 264 */  { "AFTER", 5, true, false},
/* 265 */  { "AGAINST", 7, true, false},
/* 266 */  { "AGGREGATE", 9, true, false},
/* 267 */  { "ALGORITHM", 9, true, false},
/* 268 */  { "ALL", 3, true, false},
/* 269 */  { "ALTER", 5, true, false},
/* 270 */  { "ALWAYS", 6, true, false},
/* 271 */  { "(unknown)", 9, true, false},
/* 272 */  { "ANALYZE", 7, true, false},
/* 273 */  { "&&", 2, true, true},
/* 274 */  { "AND", 3, true, true},
/* 275 */  { "SOME", 4, true, false},
/* 276 */  { "AS", 2, true, false},
/* 277 */  { "ASC", 3, true, false},
/* 278 */  { "ASCII", 5, true, false},
/* 279 */  { "ASENSITIVE", 10, true, false},
/* 280 */  { "AT", 2, true, true},
/* 281 */  { "AUTOEXTEND_SIZE", 15, true, false},
/* 282 */  { "AUTO_INCREMENT", 14, true, false},
/* 283 */  { "AVG_ROW_LENGTH", 14, true, false},
/* 284 */  { "AVG", 3, true, false},
/* 285 */  { "BACKUP", 6, true, false},
/* 286 */  { "BEFORE", 6, true, false},
/* 287 */  { "BEGIN", 5, true, false},
/* 288 */  { "BETWEEN", 7, true, true},
/* 289 */  { "INT8", 4, true, false},
/* 290 */  { "BINARY", 6, true, false},
/* 291 */  { "BINLOG", 6, true, false},
/* 292 */  { "(bin)", 5, true, false},
/* 293 */  { "BIT_AND", 7, true, false},
/* 294 */  { "BIT_OR", 6, true, false},
/* 295 */  { "BIT", 3, true, false},
/* 296 */  { "BIT_XOR", 7, true, false},
/* 297 */  { "BLOB", 4, true, false},
/* 298 */  { "BLOCK", 5, true, false},
/* 299 */  { "BOOLEAN", 7, true, false},
/* 300 */  { "BOOL", 4, true, false},
/* 301 */  { "BOTH", 4, true, false},
/* 302 */  { "BTREE", 5, true, false},
/* 303 */  { "BY", 2, true, false},
/* 304 */  { "BYTE", 4, true, false},
/* 305 */  { "CACHE", 5, true, false},
/* 306 */  { "CALL", 4, true, false},
/* 307 */  { "CASCADE", 7, true, false},
/* 308 */  { "CASCADED", 8, true, false},
/* 309 */  { "CASE", 4, true, true},
/* 310 */  { "CAST", 4, true, false},
/* 311 */  { "CATALOG_NAME", 12, true, false},
/* 312 */  { "CHAIN", 5, true, false},
/* 313 */  { "CHANGE", 6, true, false},
/* 314 */  { "CHANGED", 7, true, false},
/* 315 */  { "CHANNEL", 7, true, false},
/* 316 */  { "CHARSET", 7, true, false},
/* 317 */  { "CHARACTER", 9, true, false},
/* 318 */  { "CHECKSUM", 8, true, false},
/* 319 */  { "CHECK", 5, true, false},
/* 320 */  { "CIPHER", 6, true, false},
/* 321 */  { "CLASS_ORIGIN", 12, true, false},
/* 322 */  { "CLIENT", 6, true, false},
/* 323 */  { "CLOSE", 5, true, false},
/* 324 */  { "COALESCE", 8, true, false},
/* 325 */  { "CODE", 4, true, false},
/* 326 */  { "COLLATE", 7, true, false},
/* 327 */  { "COLLATION", 9, true, false},
/* 328 */  { "FIELDS", 6, true, false},
/* 329 */  { "COLUMN", 6, true, false},
/* 330 */  { "COLUMN_FORMAT", 13, true, false},
/* 331 */  { "COLUMN_NAME", 11, true, false},
/* 332 */  { "COMMENT", 7, true, false},
/* 333 */  { "COMMITTED", 9, true, false},
/* 334 */  { "COMMIT", 6, true, false},
/* 335 */  { "COMPACT", 7, true, false},
/* 336 */  { "COMPLETION", 10, true, false},
/* 337 */  { "COMPRESSED", 10, true, false},
/* 338 */  { "COMPRESSION", 11, true, false},
/* 339 */  { "ENCRYPTION", 10, true, false},
/* 340 */  { "CONCURRENT", 10, true, false},
/* 341 */  { "CONDITION", 9, true, false},
/* 342 */  { "CONNECTION", 10, true, false},
/* 343 */  { "CONSISTENT", 10, true, false},
/* 344 */  { "CONSTRAINT", 10, true, false},
/* 345 */  { "CONSTRAINT_CATALOG", 18, true, false},
/* 346 */  { "CONSTRAINT_NAME", 15, true, false},
/* 347 */  { "CONSTRAINT_SCHEMA", 17, true, false},
/* 348 */  { "CONTAINS", 8, true, false},
/* 349 */  { "CONTEXT", 7, true, false},
/* 350 */  { "CONTINUE", 8, true, false},
/* 351 */  { "CONVERT", 7, true, false},
/* 352 */  { "COUNT", 5, true, false},
/* 353 */  { "CPU", 3, true, false},
/* 354 */  { "CREATE", 6, true, false},
/* 355 */  { "CROSS", 5, true, false},
/* 356 */  { "CUBE", 4, true, false},
/* 357 */  { "CURDATE", 7, true, false},
/* 358 */  { "CURRENT", 7, true, false},
/* 359 */  { "CURRENT_USER", 12, true, false},
/* 360 */  { "CURSOR", 6, true, false},
/* 361 */  { "CURSOR_NAME", 11, true, false},
/* 362 */  { "CURTIME", 7, true, false},
/* 363 */  { "SCHEMA", 6, true, false},
/* 364 */  { "SCHEMAS", 7, true, false},
/* 365 */  { "DATAFILE", 8, true, false},
/* 366 */  { "DATA", 4, true, false},
/* 367 */  { "DATETIME", 8, true, false},
/* 368 */  { "DATE_ADD", 8, true, false},
/* 369 */  { "DATE_SUB", 8, true, false},
/* 370 */  { "DATE", 4, true, false},
/* 371 */  { "DAY_HOUR", 8, true, false},
/* 372 */  { "DAY_MICROSECOND", 15, true, false},
/* 373 */  { "DAY_MINUTE", 10, true, false},
/* 374 */  { "DAY_SECOND", 10, true, false},
/* 375 */  { "SQL_TSI_DAY", 11, true, false},
/* 376 */  { "DEALLOCATE", 10, true, false},
/* 377 */  { "(decimal)", 9, true, false},
/* 378 */  { "DECIMAL", 7, true, false},
/* 379 */  { "DECLARE", 7, true, false},
/* 380 */  { "DEFAULT", 7, true, true},
/* 381 */  { "DEFAULT_AUTH", 12, true, false},
/* 382 */  { "DEFINER", 7, true, false},
/* 383 */  { "DELAYED", 7, true, false},
/* 384 */  { "DELAY_KEY_WRITE", 15, true, false},
/* 385 */  { "DELETE", 6, true, false},
/* 386 */  { "DESC", 4, true, false},
/* 387 */  { "EXPLAIN", 7, true, false},
/* 388 */  { "(unknown)", 9, true, false},
/* 389 */  { "DETERMINISTIC", 13, true, false},
/* 390 */  { "DIAGNOSTICS", 11, true, false},
/* 391 */  { "DIRECTORY", 9, true, false},
/* 392 */  { "DISABLE", 7, true, false},
/* 393 */  { "DISCARD", 7, true, false},
/* 394 */  { "DISK", 4, true, false},
/* 395 */  { "DISTINCTROW", 11, true, false},
/* 396 */  { "DIV", 3, true, true},
/* 397 */  { "FLOAT8", 6, true, false},
/* 398 */  { "DO", 2, true, false},
/* 399 */  { "DROP", 4, true, false},
/* 400 */  { "DUAL", 4, true, false},
/* 401 */  { "DUMPFILE", 8, true, false},
/* 402 */  { "DUPLICATE", 9, true, false},
/* 403 */  { "DYNAMIC", 7, true, false},
/* 404 */  { "EACH", 4, true, false},
/* 405 */  { "ELSE", 4, true, false},
/* 406 */  { "ELSEIF", 6, true, true},
/* 407 */  { "ENABLE", 6, true, false},
/* 408 */  { "ENCLOSED", 8, true, false},
/* 409 */  { "END", 3, true, false},
/* 410 */  { "ENDS", 4, true, true},
/* 411 */  { "", 0, true, false},
/* 412 */  { "ENGINES", 7, true, false},
/* 413 */  { "ENGINE", 6, true, false},
/* 414 */  { "ENUM", 4, true, false},
/* 415 */  { "=", 1, true, false},
/* 416 */  { "<=>", 3, true, false},
/* 417 */  { "ERROR", 5, true, false},
/* 418 */  { "ERRORS", 6, true, false},
/* 419 */  { "ESCAPED", 7, true, false},
/* 420 */  { "ESCAPE", 6, true, false},
/* 421 */  { "EVENTS", 6, true, false},
/* 422 */  { "EVENT", 5, true, false},
/* 423 */  { "EVERY", 5, true, true},
/* 424 */  { "EXCHANGE", 8, true, false},
/* 425 */  { "EXECUTE", 7, true, false},
/* 426 */  { "EXISTS", 6, true, false},
/* 427 */  { "EXIT", 4, true, false},
/* 428 */  { "EXPANSION", 9, true, false},
/* 429 */  { "EXPIRE", 6, true, false},
/* 430 */  { "EXPORT", 6, true, false},
/* 431 */  { "EXTENDED", 8, true, false},
/* 432 */  { "EXTENT_SIZE", 11, true, false},
/* 433 */  { "EXTRACT", 7, true, false},
/* 434 */  { "FALSE", 5, true, false},
/* 435 */  { "FAST", 4, true, false},
/* 436 */  { "FAULTS", 6, true, false},
/* 437 */  { "FETCH", 5, true, false},
/* 438 */  { "FILE", 4, true, false},
/* 439 */  { "FILE_BLOCK_SIZE", 15, true, false},
/* 440 */  { "FILTER", 6, true, false},
/* 441 */  { "FIRST", 5, true, false},
/* 442 */  { "FIXED", 5, true, false},
/* 443 */  { "(float)", 7, true, false},
/* 444 */  { "FLOAT4", 6, true, false},
/* 445 */  { "FLUSH", 5, true, false},
/* 446 */  { "FOLLOWS", 7, true, false},
/* 447 */  { "FORCE", 5, true, false},
/* 448 */  { "FOREIGN", 7, true, false},
/* 449 */  { "FOR", 3, true, false},
/* 450 */  { "FORMAT", 6, true, false},
/* 451 */  { "FOUND", 5, true, false},
/* 452 */  { "FROM", 4, true, false},
/* 453 */  { "FULL", 4, true, false},
/* 454 */  { "FULLTEXT", 8, true, false},
/* 455 */  { "FUNCTION", 8, true, false},
/* 456 */  { ">=", 2, true, false},
/* 457 */  { "GENERAL", 7, true, false},
/* 458 */  { "GENERATED", 9, true, false},
/* 459 */  { "GROUP_REPLICATION", 17, true, false},
/* 460 */  { "GEOMETRYCOLLECTION", 18, true, false},
/* 461 */  { "GEOMETRY", 8, true, false},
/* 462 */  { "GET_FORMAT", 10, true, false},
/* 463 */  { "GET", 3, true, false},
/* 464 */  { "GLOBAL", 6, true, false},
/* 465 */  { "GRANT", 5, true, false},
/* 466 */  { "GRANTS", 6, true, false},
/* 467 */  { "GROUP", 5, true, false},
/* 468 */  { "GROUP_CONCAT", 12, true, false},
/* 469 */  { ">", 1, true, false},
/* 470 */  { "HANDLER", 7, true, false},
/* 471 */  { "HASH", 4, true, false},
/* 472 */  { "HAVING", 6, true, false},
/* 473 */  { "HELP", 4, true, false},
/* 474 */  { "(hex)", 5, true, false},
/* 475 */  { "HIGH_PRIORITY", 13, true, false},
/* 476 */  { "HOST", 4, true, false},
/* 477 */  { "HOSTS", 5, true, false},
/* 478 */  { "HOUR_MICROSECOND", 16, true, false},
/* 479 */  { "HOUR_MINUTE", 11, true, false},
/* 480 */  { "HOUR_SECOND", 11, true, false},
/* 481 */  { "SQL_TSI_HOUR", 12, true, false},
/* 482 */  { "(id)", 4, true, false},
/* 483 */  { "IDENTIFIED", 10, true, false},
/* 484 */  { "(id_quoted)", 11, true, false},
/* 485 */  { "IF", 2, true, true},
/* 486 */  { "IGNORE", 6, true, false},
/* 487 */  { "IGNORE_SERVER_IDS", 17, true, false},
/* 488 */  { "IMPORT", 6, true, false},
/* 489 */  { "INDEXES", 7, true, false},
/* 490 */  { "INDEX", 5, true, false},
/* 491 */  { "INFILE", 6, true, false},
/* 492 */  { "INITIAL_SIZE", 12, true, false},
/* 493 */  { "INNER", 5, true, false},
/* 494 */  { "INOUT", 5, true, false},
/* 495 */  { "INSENSITIVE", 11, true, false},
/* 496 */  { "INSERT", 6, true, false},
/* 497 */  { "INSERT_METHOD", 13, true, false},
/* 498 */  { "INSTANCE", 8, true, false},
/* 499 */  { "INSTALL", 7, true, false},
/* 500 */  { "INTERVAL", 8, true, true},
/* 501 */  { "INTO", 4, true, false},
/* 502 */  { "INTEGER", 7, true, false},
/* 503 */  { "INVOKER", 7, true, false},
/* 504 */  { "IN", 2, true, false},
/* 505 */  { "IO_AFTER_GTIDS", 14, true, false},
/* 506 */  { "IO_BEFORE_GTIDS", 15, true, false},
/* 507 */  { "IO", 2, true, false},
/* 508 */  { "IPC", 3, true, false},
/* 509 */  { "IS", 2, true, false},
/* 510 */  { "ISOLATION", 9, true, false},
/* 511 */  { "ISSUER", 6, true, false},
/* 512 */  { "ITERATE", 7, true, false},
/* 513 */  { "JOIN", 4, true, false},
/* 514 */  { "->", 2, true, false},
/* 515 */  { "JSON", 4, true, false},
/* 516 */  { "KEYS", 4, true, false},
/* 517 */  { "KEY_BLOCK_SIZE", 14, true, false},
/* 518 */  { "KEY", 3, true, false},
/* 519 */  { "KILL", 4, true, false},
/* 520 */  { "LANGUAGE", 8, true, false},
/* 521 */  { "LAST", 4, true, false},
/* 522 */  { "<=", 2, true, false},
/* 523 */  { "LEADING", 7, true, false},
/* 524 */  { "LEAVES", 6, true, false},
/* 525 */  { "LEAVE", 5, true, false},
/* 526 */  { "LEFT", 4, true, false},
/* 527 */  { "LESS", 4, true, false},
/* 528 */  { "LEVEL", 5, true, false},
/* 529 */  { "(hostname)", 10, true, false},
/* 530 */  { "LIKE", 4, true, true},
/* 531 */  { "LIMIT", 5, true, false},
/* 532 */  { "LINEAR", 6, true, false},
/* 533 */  { "LINES", 5, true, false},
/* 534 */  { "LINESTRING", 10, true, false},
/* 535 */  { "LIST", 4, true, false},
/* 536 */  { "LOAD", 4, true, false},
/* 537 */  { "LOCAL", 5, true, false},
/* 538 */  { "(unknown)", 9, true, false},
/* 539 */  { "LOCKS", 5, true, false},
/* 540 */  { "LOCK", 4, true, false},
/* 541 */  { "LOGFILE", 7, true, false},
/* 542 */  { "LOGS", 4, true, false},
/* 543 */  { "LONGBLOB", 8, true, false},
/* 544 */  { "LONGTEXT", 8, true, false},
/* 545 */  { "(long)", 6, true, false},
/* 546 */  { "LONG", 4, true, false},
/* 547 */  { "LOOP", 4, true, false},
/* 548 */  { "LOW_PRIORITY", 12, true, false},
/* 549 */  { "<", 1, true, false},
/* 550 */  { "(unknown)", 9, true, false},
/* 551 */  { "(unknown)", 9, true, false},
/* 552 */  { "(unknown)", 9, true, false},
/* 553 */  { "(unknown)", 9, true, false},
/* 554 */  { "(unknown)", 9, true, false},
/* 555 */  { "(unknown)", 9, true, false},
/* 556 */  { "(unknown)", 9, true, false},
/* 557 */  { "(unknown)", 9, true, false},
/* 558 */  { "(unknown)", 9, true, false},
/* 559 */  { "(unknown)", 9, true, false},
/* 560 */  { "(unknown)", 9, true, false},
/* 561 */  { "(unknown)", 9, true, false},
/* 562 */  { "(unknown)", 9, true, false},
/* 563 */  { "(unknown)", 9, true, false},
/* 564 */  { "(unknown)", 9, true, false},
/* 565 */  { "(unknown)", 9, true, false},
/* 566 */  { "(unknown)", 9, true, false},
/* 567 */  { "(unknown)", 9, true, false},
/* 568 */  { "(unknown)", 9, true, false},
/* 569 */  { "(unknown)", 9, true, false},
/* 570 */  { "(unknown)", 9, true, false},
/* 571 */  { "MASTER", 6, true, false},
/* 572 */  { "(unknown)", 9, true, false},
/* 573 */  { "(unknown)", 9, true, false},
/* 574 */  { "MATCH", 5, true, false},
/* 575 */  { "MAX_CONNECTIONS_PER_HOUR", 24, true, false},
/* 576 */  { "MAX_QUERIES_PER_HOUR", 20, true, false},
/* 577 */  { "MAX_ROWS", 8, true, false},
/* 578 */  { "MAX_SIZE", 8, true, false},
/* 579 */  { "MAX", 3, true, false},
/* 580 */  { "MAX_UPDATES_PER_HOUR", 20, true, false},
/* 581 */  { "MAX_USER_CONNECTIONS", 20, true, false},
/* 582 */  { "MAXVALUE", 8, true, false},
/* 583 */  { "MEDIUMBLOB", 10, true, false},
/* 584 */  { "MIDDLEINT", 9, true, false},
/* 585 */  { "MEDIUMTEXT", 10, true, false},
/* 586 */  { "MEDIUM", 6, true, false},
/* 587 */  { "MEMORY", 6, true, false},
/* 588 */  { "MERGE", 5, true, false},
/* 589 */  { "MESSAGE_TEXT", 12, true, false},
/* 590 */  { "MICROSECOND", 11, true, false},
/* 591 */  { "MIGRATE", 7, true, false},
/* 592 */  { "MINUTE_MICROSECOND", 18, true, false},
/* 593 */  { "MINUTE_SECOND", 13, true, false},
/* 594 */  { "SQL_TSI_MINUTE", 14, true, false},
/* 595 */  { "MIN_ROWS", 8, true, false},
/* 596 */  { "MIN", 3, true, false},
/* 597 */  { "MODE", 4, true, false},
/* 598 */  { "MODIFIES", 8, true, false},
/* 599 */  { "MODIFY", 6, true, false},
/* 600 */  { "MOD", 3, true, true},
/* 601 */  { "SQL_TSI_MONTH", 13, true, false},
/* 602 */  { "MULTILINESTRING", 15, true, false},
/* 603 */  { "MULTIPOINT", 10, true, false},
/* 604 */  { "MULTIPOLYGON", 12, true, false},
/* 605 */  { "MUTEX", 5, true, false},
/* 606 */  { "MYSQL_ERRNO", 11, true, false},
/* 607 */  { "NAMES", 5, true, false},
/* 608 */  { "NAME", 4, true, false},
/* 609 */  { "NATIONAL", 8, true, false},
/* 610 */  { "NATURAL", 7, true, false},
/* 611 */  { "(nchar)", 7, true, false},
/* 612 */  { "NCHAR", 5, true, false},
/* 613 */  { "NDBCLUSTER", 10, true, false},
/* 614 */  { "!=", 2, true, false},
/* 615 */  { "(unknown)", 9, true, false},
/* 616 */  { "NEVER", 5, true, false},
/* 617 */  { "NEW", 3, true, false},
/* 618 */  { "NEXT", 4, true, false},
/* 619 */  { "NODEGROUP", 9, true, false},
/* 620 */  { "NONE", 4, true, false},
/* 621 */  { "!", 1, true, false},
/* 622 */  { "NOT", 3, true, true},
/* 623 */  { "NOW", 3, true, false},
/* 624 */  { "NO", 2, true, false},
/* 625 */  { "NO_WAIT", 7, true, false},
/* 626 */  { "NO_WRITE_TO_BINLOG", 18, true, false},
/* 627 */  { "NULL", 4, true, false},
/* 628 */  { "(num)", 5, true, false},
/* 629 */  { "NUMBER", 6, true, false},
/* 630 */  { "NUMERIC", 7, true, false},
/* 631 */  { "NVARCHAR", 8, true, false},
/* 632 */  { "OFFSET", 6, true, false},
/* 633 */  { "ON", 2, true, false},
/* 634 */  { "ONE", 3, true, false},
/* 635 */  { "ONLY", 4, true, false},
/* 636 */  { "OPEN", 4, true, false},
/* 637 */  { "OPTIMIZE", 8, true, false},
/* 638 */  { "OPTIMIZER_COSTS", 15, true, false},
/* 639 */  { "OPTIONS", 7, true, false},
/* 640 */  { "OPTION", 6, true, false},
/* 641 */  { "OPTIONALLY", 10, true, false},
/* 642 */  { "||", 2, true, true},
/* 643 */  { "ORDER", 5, true, false},
/* 644 */  { "||", 2, true, false},
/* 645 */  { "OR", 2, true, true},
/* 646 */  { "OUTER", 5, true, false},
/* 647 */  { "OUTFILE", 7, true, false},
/* 648 */  { "OUT", 3, true, false},
/* 649 */  { "OWNER", 5, true, false},
/* 650 */  { "PACK_KEYS", 9, true, false},
/* 651 */  { "PAGE", 4, true, false},
/* 652 */  { "?", 1, true, false},
/* 653 */  { "PARSER", 6, true, false},
/* 654 */  { "(unknown)", 9, true, false},
/* 655 */  { "PARTIAL", 7, true, false},
/* 656 */  { "PARTITION", 9, true, false},
/* 657 */  { "PARTITIONS", 10, true, false},
/* 658 */  { "PARTITIONING", 12, true, false},
/* 659 */  { "PASSWORD", 8, true, false},
/* 660 */  { "PHASE", 5, true, false},
/* 661 */  { "PLUGIN_DIR", 10, true, false},
/* 662 */  { "PLUGIN", 6, true, false},
/* 663 */  { "PLUGINS", 7, true, false},
/* 664 */  { "POINT", 5, true, false},
/* 665 */  { "POLYGON", 7, true, false},
/* 666 */  { "PORT", 4, true, false},
/* 667 */  { "POSITION", 8, true, false},
/* 668 */  { "PRECEDES", 8, true, false},
/* 669 */  { "PRECISION", 9, true, false},
/* 670 */  { "PREPARE", 7, true, false},
/* 671 */  { "PRESERVE", 8, true, false},
/* 672 */  { "PREV", 4, true, false},
/* 673 */  { "PRIMARY", 7, true, false},
/* 674 */  { "PRIVILEGES", 10, true, false},
/* 675 */  { "PROCEDURE", 9, true, false},
/* 676 */  { "PROCESS", 7, true, false},
/* 677 */  { "PROCESSLIST", 11, true, false},
/* 678 */  { "PROFILE", 7, true, false},
/* 679 */  { "PROFILES", 8, true, false},
/* 680 */  { "PROXY", 5, true, false},
/* 681 */  { "PURGE", 5, true, false},
/* 682 */  { "SQL_TSI_QUARTER", 15, true, false},
/* 683 */  { "QUERY", 5, true, false},
/* 684 */  { "QUICK", 5, true, false},
/* 685 */  { "RANGE", 5, true, false},
/* 686 */  { "READS", 5, true, false},
/* 687 */  { "READ_ONLY", 9, true, false},
/* 688 */  { "READ", 4, true, false},
/* 689 */  { "READ_WRITE", 10, true, false},
/* 690 */  { "REAL", 4, true, false},
/* 691 */  { "REBUILD", 7, true, false},
/* 692 */  { "RECOVER", 7, true, false},
/* 693 */  { "(unknown)", 9, true, false},
/* 694 */  { "REDO_BUFFER_SIZE", 16, true, false},
/* 695 */  { "REDUNDANT", 9, true, false},
/* 696 */  { "REFERENCES", 10, true, false},
/* 697 */  { "RLIKE", 5, true, true},
/* 698 */  { "RELAY", 5, true, false},
/* 699 */  { "RELAYLOG", 8, true, false},
/* 700 */  { "RELAY_LOG_FILE", 14, true, false},
/* 701 */  { "RELAY_LOG_POS", 13, true, false},
/* 702 */  { "RELAY_THREAD", 12, true, false},
/* 703 */  { "RELEASE", 7, true, false},
/* 704 */  { "RELOAD", 6, true, false},
/* 705 */  { "REMOVE", 6, true, false},
/* 706 */  { "RENAME", 6, true, false},
/* 707 */  { "REORGANIZE", 10, true, false},
/* 708 */  { "REPAIR", 6, true, false},
/* 709 */  { "REPEATABLE", 10, true, false},
/* 710 */  { "REPEAT", 6, true, false},
/* 711 */  { "REPLACE", 7, true, false},
/* 712 */  { "REPLICATION", 11, true, false},
/* 713 */  { "REPLICATE_DO_DB", 15, true, false},
/* 714 */  { "REPLICATE_IGNORE_DB", 19, true, false},
/* 715 */  { "REPLICATE_DO_TABLE", 18, true, false},
/* 716 */  { "REPLICATE_IGNORE_TABLE", 22, true, false},
/* 717 */  { "REPLICATE_WILD_DO_TABLE", 23, true, false},
/* 718 */  { "REPLICATE_WILD_IGNORE_TABLE", 27, true, false},
/* 719 */  { "REPLICATE_REWRITE_DB", 20, true, false},
/* 720 */  { "REQUIRE", 7, true, false},
/* 721 */  { "RESET", 5, true, false},
/* 722 */  { "RESIGNAL", 8, true, false},
/* 723 */  { "USER_RESOURCES", 14, true, false},
/* 724 */  { "RESTORE", 7, true, false},
/* 725 */  { "RESTRICT", 8, true, false},
/* 726 */  { "RESUME", 6, true, false},
/* 727 */  { "RETURNED_SQLSTATE", 17, true, false},
/* 728 */  { "RETURNS", 7, true, false},
/* 729 */  { "RETURN", 6, true, true},
/* 730 */  { "REVERSE", 7, true, false},
/* 731 */  { "REVOKE", 6, true, false},
/* 732 */  { "RIGHT", 5, true, false},
/* 733 */  { "ROLLBACK", 8, true, false},
/* 734 */  { "ROLLUP", 6, true, false},
/* 735 */  { "ROTATE", 6, true, false},
/* 736 */  { "ROUTINE", 7, true, false},
/* 737 */  { "ROWS", 4, true, false},
/* 738 */  { "ROW_FORMAT", 10, true, false},
/* 739 */  { "ROW", 3, true, false},
/* 740 */  { "ROW_COUNT", 9, true, false},
/* 741 */  { "RTREE", 5, true, false},
/* 742 */  { "SAVEPOINT", 9, true, false},
/* 743 */  { "SCHEDULE", 8, true, false},
/* 744 */  { "SCHEMA_NAME", 11, true, false},
/* 745 */  { "SECOND_MICROSECOND", 18, true, false},
/* 746 */  { "SQL_TSI_SECOND", 14, true, false},
/* 747 */  { "SECURITY", 8, true, false},
/* 748 */  { "SELECT", 6, true, true},
/* 749 */  { "SENSITIVE", 9, true, false},
/* 750 */  { "SEPARATOR", 9, true, false},
/* 751 */  { "SERIALIZABLE", 12, true, false},
/* 752 */  { "SERIAL", 6, true, false},
/* 753 */  { "SESSION", 7, true, false},
/* 754 */  { "SERVER", 6, true, false},
/* 755 */  { "(unknown)", 9, true, false},
/* 756 */  { "SET", 3, true, false},
/* 757 */  { ":=", 2, true, false},
/* 758 */  { "SHARE", 5, true, false},
/* 759 */  { "<<", 2, true, true},
/* 760 */  { ">>", 2, true, true},
/* 761 */  { "SHOW", 4, true, false},
/* 762 */  { "SHUTDOWN", 8, true, false},
/* 763 */  { "SIGNAL", 6, true, false},
/* 764 */  { "SIGNED", 6, true, false},
/* 765 */  { "SIMPLE", 6, true, false},
/* 766 */  { "SLAVE", 5, true, false},
/* 767 */  { "SLOW", 4, true, false},
/* 768 */  { "SMALLINT", 8, true, false},
/* 769 */  { "SNAPSHOT", 8, true, false},
/* 770 */  { "SOCKET", 6, true, false},
/* 771 */  { "SONAME", 6, true, false},
/* 772 */  { "SOUNDS", 6, true, false},
/* 773 */  { "SOURCE", 6, true, false},
/* 774 */  { "SPATIAL", 7, true, false},
/* 775 */  { "SPECIFIC", 8, true, false},
/* 776 */  { "SQLEXCEPTION", 12, true, false},
/* 777 */  { "SQLSTATE", 8, true, false},
/* 778 */  { "SQLWARNING", 10, true, false},
/* 779 */  { "SQL_AFTER_GTIDS", 15, true, false},
/* 780 */  { "SQL_AFTER_MTS_GAPS", 18, true, false},
/* 781 */  { "SQL_BEFORE_GTIDS", 16, true, false},
/* 782 */  { "SQL_BIG_RESULT", 14, true, false},
/* 783 */  { "SQL_BUFFER_RESULT", 17, true, false},
/* 784 */  { "(unknown)", 9, true, false},
/* 785 */  { "SQL_CALC_FOUND_ROWS", 19, true, false},
/* 786 */  { "SQL_NO_CACHE", 12, true, false},
/* 787 */  { "SQL_SMALL_RESULT", 16, true, false},
/* 788 */  { "SQL", 3, true, false},
/* 789 */  { "SQL_THREAD", 10, true, false},
/* 790 */  { "SSL", 3, true, false},
/* 791 */  { "STACKED", 7, true, false},
/* 792 */  { "STARTING", 8, true, false},
/* 793 */  { "STARTS", 6, true, true},
/* 794 */  { "START", 5, true, false},
/* 795 */  { "STATS_AUTO_RECALC", 17, true, false},
/* 796 */  { "STATS_PERSISTENT", 16, true, false},
/* 797 */  { "STATS_SAMPLE_PAGES", 18, true, false},
/* 798 */  { "STATUS", 6, true, false},
/* 799 */  { "STDDEV_SAMP", 11, true, false},
/* 800 */  { "STDDEV_POP", 10, true, false},
/* 801 */  { "STOP", 4, true, false},
/* 802 */  { "STORAGE", 7, true, false},
/* 803 */  { "STORED", 6, true, false},
/* 804 */  { "STRAIGHT_JOIN", 13, true, false},
/* 805 */  { "STRING", 6, true, false},
/* 806 */  { "SUBCLASS_ORIGIN", 15, true, false},
/* 807 */  { "SUBDATE", 7, true, false},
/* 808 */  { "SUBJECT", 7, true, false},
/* 809 */  { "SUBPARTITIONS", 13, true, false},
/* 810 */  { "SUBPARTITION", 12, true, false},
/* 811 */  { "SUBSTRING", 9, true, false},
/* 812 */  { "SUM", 3, true, false},
/* 813 */  { "SUPER", 5, true, false},
/* 814 */  { "SUSPEND", 7, true, false},
/* 815 */  { "SWAPS", 5, true, false},
/* 816 */  { "SWITCHES", 8, true, false},
/* 817 */  { "SYSDATE", 7, true, false},
/* 818 */  { "TABLES", 6, true, false},
/* 819 */  { "TABLESPACE", 10, true, false},
/* 820 */  { "(unknown)", 9, true, false},
/* 821 */  { "TABLE", 5, true, false},
/* 822 */  { "TABLE_CHECKSUM", 14, true, false},
/* 823 */  { "TABLE_NAME", 10, true, false},
/* 824 */  { "TEMPORARY", 9, true, false},
/* 825 */  { "TEMPTABLE", 9, true, false},
/* 826 */  { "TERMINATED", 10, true, false},
/* 827 */  { "(text)", 6, true, false},
/* 828 */  { "TEXT", 4, true, false},
/* 829 */  { "THAN", 4, true, false},
/* 830 */  { "THEN", 4, true, false},
/* 831 */  { "TIMESTAMP", 9, true, false},
/* 832 */  { "TIMESTAMPADD", 12, true, false},
/* 833 */  { "TIMESTAMPDIFF", 13, true, false},
/* 834 */  { "TIME", 4, true, false},
/* 835 */  { "TINYBLOB", 8, true, false},
/* 836 */  { "TINYINT", 7, true, false},
/* 837 */  { "TINYTEXT", 8, true, false},
/* 838 */  { "TO", 2, true, false},
/* 839 */  { "TRAILING", 8, true, false},
/* 840 */  { "TRANSACTION", 11, true, false},
/* 841 */  { "TRIGGERS", 8, true, false},
/* 842 */  { "TRIGGER", 7, true, false},
/* 843 */  { "TRIM", 4, true, false},
/* 844 */  { "TRUE", 4, true, false},
/* 845 */  { "TRUNCATE", 8, true, false},
/* 846 */  { "TYPES", 5, true, false},
/* 847 */  { "TYPE", 4, true, false},
/* 848 */  { "(unknown)", 9, true, false},
/* 849 */  { "(ulonglong)", 11, true, false},
/* 850 */  { "UNCOMMITTED", 11, true, false},
/* 851 */  { "UNDEFINED", 9, true, false},
/* 852 */  { "(_charset)", 10, true, false},
/* 853 */  { "UNDOFILE", 8, true, false},
/* 854 */  { "UNDO_BUFFER_SIZE", 16, true, false},
/* 855 */  { "UNDO", 4, true, false},
/* 856 */  { "UNICODE", 7, true, false},
/* 857 */  { "UNINSTALL", 9, true, false},
/* 858 */  { "UNION", 5, true, false},
/* 859 */  { "UNIQUE", 6, true, false},
/* 860 */  { "UNKNOWN", 7, true, false},
/* 861 */  { "UNLOCK", 6, true, false},
/* 862 */  { "UNSIGNED", 8, true, false},
/* 863 */  { "UNTIL", 5, true, true},
/* 864 */  { "UPDATE", 6, true, false},
/* 865 */  { "UPGRADE", 7, true, false},
/* 866 */  { "USAGE", 5, true, false},
/* 867 */  { "SYSTEM_USER", 11, true, false},
/* 868 */  { "USE_FRM", 7, true, false},
/* 869 */  { "USE", 3, true, false},
/* 870 */  { "USING", 5, true, false},
/* 871 */  { "UTC_DATE", 8, true, false},
/* 872 */  { "UTC_TIMESTAMP", 13, true, false},
/* 873 */  { "UTC_TIME", 8, true, false},
/* 874 */  { "VALIDATION", 10, true, false},
/* 875 */  { "VALUES", 6, true, false},
/* 876 */  { "VALUE", 5, true, false},
/* 877 */  { "VARBINARY", 9, true, false},
/* 878 */  { "VARCHARACTER", 12, true, false},
/* 879 */  { "VARIABLES", 9, true, false},
/* 880 */  { "VAR_POP", 7, true, false},
/* 881 */  { "VARYING", 7, true, false},
/* 882 */  { "VAR_SAMP", 8, true, false},
/* 883 */  { "VIEW", 4, true, false},
/* 884 */  { "VIRTUAL", 7, true, false},
/* 885 */  { "WAIT", 4, true, false},
/* 886 */  { "WARNINGS", 8, true, false},
/* 887 */  { "WEEK", 4, true, false},
/* 888 */  { "WEIGHT_STRING", 13, true, false},
/* 889 */  { "WHEN", 4, true, true},
/* 890 */  { "WHERE", 5, true, false},
/* 891 */  { "WHILE", 5, true, true},
/* 892 */  { "WITH", 4, true, false},
/* 893 */  { "(unknown)", 9, true, false},
/* 894 */  { "WITH ROLLUP", 11, true, false},
/* 895 */  { "WITHOUT", 7, true, false},
/* 896 */  { "WORK", 4, true, false},
/* 897 */  { "WRAPPER", 7, true, false},
/* 898 */  { "WRITE", 5, true, false},
/* 899 */  { "X509", 4, true, false},
/* 900 */  { "XA", 2, true, false},
/* 901 */  { "XID", 3, true, false},
/* 902 */  { "XML", 3, true, false},
/* 903 */  { "XOR", 3, true, true},
/* 904 */  { "YEAR_MONTH", 10, true, false},
/* 905 */  { "YEAR", 4, true, false},
/* 906 */  { "ZEROFILL", 8, true, false},
/* 907 */  { "->>", 3, true, false},
/* 908 */  { "PERSIST", 7, true, false},
/* 909 */  { "ROLE", 4, true, false},
/* 910 */  { "ADMIN", 5, true, false},
/* 911 */  { "INVISIBLE", 9, true, false},
/* 912 */  { "VISIBLE", 7, true, false},
/* 913 */  { "EXCEPT", 6, true, false},
/* 914 */  { "COMPONENT", 9, true, false},
/* 915 */  { "RECURSIVE", 9, true, false},
/* 916 */  { "(unknown)", 9, true, false},
/* 917 */  { "(unknown)", 9, true, false},
/* 918 */  { "(unknown)", 9, true, false},
/* 919 */  { "(unknown)", 9, true, false},
/* 920 */  { "JSON_OBJECTAGG", 14, true, false},
/* 921 */  { "JSON_ARRAYAGG", 13, true, false},
/* 922 */  { "OF", 2, true, false},
/* 923 */  { "SKIP", 4, true, false},
/* 924 */  { "LOCKED", 6, true, false},
/* 925 */  { "NOWAIT", 6, true, false},
/* 926 */  { "GROUPING", 8, true, false},
/* 927 */  { "PERSIST_ONLY", 12, true, false},
/* 928 */  { "HISTOGRAM", 9, true, false},
/* 929 */  { "BUCKETS", 7, true, false},
/* 930 */  { "(unknown)", 9, true, false},
/* 931 */  { "CLONE", 5, true, false},
/* 932 */  { "CUME_DIST", 9, true, false},
/* 933 */  { "DENSE_RANK", 10, true, false},
/* 934 */  { "EXCLUDE", 7, true, false},
/* 935 */  { "FIRST_VALUE", 11, true, false},
/* 936 */  { "FOLLOWING", 9, true, false},
/* 937 */  { "GROUPS", 6, true, false},
/* 938 */  { "LAG", 3, true, false},
/* 939 */  { "LAST_VALUE", 10, true, false},
/* 940 */  { "LEAD", 4, true, false},
/* 941 */  { "NTH_VALUE", 9, true, false},
/* 942 */  { "NTILE", 5, true, false},
/* 943 */  { "NULLS", 5, true, false},
/* 944 */  { "OTHERS", 6, true, false},
/* 945 */  { "OVER", 4, true, false},
/* 946 */  { "PERCENT_RANK", 12, true, false},
/* 947 */  { "PRECEDING", 9, true, false},
/* 948 */  { "RANK", 4, true, false},
/* 949 */  { "RESPECT", 7, true, false},
/* 950 */  { "ROW_NUMBER", 10, true, false},
/* 951 */  { "TIES", 4, true, false},
/* 952 */  { "UNBOUNDED", 9, true, false},
/* 953 */  { "WINDOW", 6, true, false},
/* 954 */  { "EMPTY", 5, true, false},
/* 955 */  { "JSON_TABLE", 10, true, false},
/* 956 */  { "NESTED", 6, true, false},
/* 957 */  { "ORDINALITY", 10, true, false},
/* 958 */  { "PATH", 4, true, false},
/* 959 */  { "HISTORY", 7, true, false},
/* 960 */  { "REUSE", 5, true, false},
/* 961 */  { "SRID", 4, true, false},
/* 962 */  { "THREAD_PRIORITY", 15, true, false},
/* 963 */  { "RESOURCE", 8, true, false},
/* 964 */  { "SYSTEM", 6, true, false},
/* 965 */  { "VCPU", 4, true, false},
/* 966 */  { "(unknown)", 9, true, false},
/* 967 */  { "(unknown)", 9, true, false},
/* 968 */  { "RESTART", 7, true, false},
/* 969 */  { "DEFINITION", 10, true, false},
/* 970 */  { "DESCRIPTION", 11, true, false},
/* 971 */  { "ORGANIZATION", 12, true, false},
/* 972 */  { "REFERENCE", 9, true, false},
/* 973 */  { "ACTIVE", 6, true, false},
/* 974 */  { "INACTIVE", 8, true, false},
/* 975 */  { "LATERAL", 7, true, false},
/* 976 */  { "ARRAY", 5, true, false},
/* 977 */  { "MEMBER", 6, true, false},
/* 978 */  { "OPTIONAL", 8, true, false},
/* 979 */  { "SECONDARY", 9, true, false},
/* 980 */  { "SECONDARY_ENGINE", 16, true, false},
/* 981 */  { "SECONDARY_LOAD", 14, true, false},
/* 982 */  { "SECONDARY_UNLOAD", 16, true, false},
/* 983 */  { "RETAIN", 6, true, false},
/* 984 */  { "OLD", 3, true, false},
/* 985 */  { "ENFORCED", 8, true, false},
/* 986 */  { "OJ", 2, true, false},
/* 987 */  { "NETWORK_NAMESPACE", 17, true, false},
/* 988 */  { "RANDOM", 6, true, false},
/* 989 */  { "(unknown)", 9, true, false},
/* 990 */  { "(unknown)", 9, true, false},
/* 991 */  { "PRIVILEGE_CHECKS_USER", 21, true, false},
/* 992 */  { "(unknown)", 9, true, false},
/* 993 */  { "REQUIRE_ROW_FORMAT", 18, true, false},
/* 994 */  { "PASSWORD_LOCK_TIME", 18, true, false},
/* 995 */  { "FAILED_LOGIN_ATTEMPTS", 21, true, false},
/* 996 */  { "REQUIRE_TABLE_PRIMARY_KEY_CHECK", 31, true, false},
/* 997 */  { "STREAM", 6, true, false},
/* 998 */  { "OFF", 3, true, false},
/* 999 */  { "RETURNING", 9, true, false},
/* PART 3: padding reserved for sql/sql_yacc.yy extensions. */
/* PART 4: named tokens from sql/sql_hints.yy. */
/* 1000 */  { "MAX_EXECUTION_TIME", 18, true, false},
/* 1001 */  { "RESOURCE_GROUP", 14, true, false},
/* 1002 */  { "BKA", 3, true, false},
/* 1003 */  { "BNL", 3, true, false},
/* 1004 */  { "DUPSWEEDOUT", 11, true, false},
/* 1005 */  { "FIRSTMATCH", 10, true, false},
/* 1006 */  { "INTOEXISTS", 10, true, false},
/* 1007 */  { "LOOSESCAN", 9, true, false},
/* 1008 */  { "MATERIALIZATION", 15, true, false},
/* 1009 */  { "NO_BKA", 6, true, false},
/* 1010 */  { "NO_BNL", 6, true, false},
/* 1011 */  { "NO_ICP", 6, true, false},
/* 1012 */  { "NO_MRR", 6, true, false},
/* 1013 */  { "NO_RANGE_OPTIMIZATION", 21, true, false},
/* 1014 */  { "NO_SEMIJOIN", 11, true, false},
/* 1015 */  { "MRR", 3, true, false},
/* 1016 */  { "QB_NAME", 7, true, false},
/* 1017 */  { "SEMIJOIN", 8, true, false},
/* 1018 */  { "SUBQUERY", 8, true, false},
/* 1019 */  { "MERGE", 5, true, false},
/* 1020 */  { "NO_MERGE", 8, true, false},
/* 1021 */  { "JOIN_PREFIX", 11, true, false},
/* 1022 */  { "JOIN_SUFFIX", 11, true, false},
/* 1023 */  { "JOIN_ORDER", 10, true, false},
/* 1024 */  { "JOIN_FIXED_ORDER", 16, true, false},
/* 1025 */  { "INDEX_MERGE", 11, true, false},
/* 1026 */  { "NO_INDEX_MERGE", 14, true, false},
/* 1027 */  { "SET_VAR", 7, true, false},
/* 1028 */  { "SKIP_SCAN", 9, true, false},
/* 1029 */  { "NO_SKIP_SCAN", 12, true, false},
/* 1030 */  { "HASH_JOIN", 9, true, false},
/* 1031 */  { "NO_HASH_JOIN", 12, true, false},
/* 1032 */  { "(unknown)", 9, true, false},
/* 1033 */  { "(unknown)", 9, true, false},
/* 1034 */  { "(unknown)", 9, true, false},
/* 1035 */  { "(unknown)", 9, true, false},
/* 1036 */  { "(unknown)", 9, true, false},
/* 1037 */  { "(unknown)", 9, true, false},
/* 1038 */  { "(unknown)", 9, true, false},
/* 1039 */  { "INDEX", 5, true, false},
/* 1040 */  { "NO_INDEX", 8, true, false},
/* 1041 */  { "JOIN_INDEX", 10, true, false},
/* 1042 */  { "NO_JOIN_INDEX", 13, true, false},
/* 1043 */  { "GROUP_INDEX", 11, true, false},
/* 1044 */  { "NO_GROUP_INDEX", 14, true, false},
/* 1045 */  { "ORDER_INDEX", 11, true, false},
/* 1046 */  { "NO_ORDER_INDEX", 14, true, false},
/* 1047 */  { "DERIVED_CONDITION_PUSHDOWN", 26, true, false},
/* 1048 */  { "NO_DERIVED_CONDITION_PUSHDOWN", 29, true, false},
/* PART 5: padding reserved for sql/sql_hints.yy extensions. */
/* reserved 1049 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1050 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1051 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1052 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1053 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1054 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1055 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1056 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1057 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1058 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1059 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1060 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1061 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1062 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1063 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1064 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1065 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1066 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1067 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1068 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1069 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1070 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1071 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1072 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1073 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1074 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1075 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1076 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1077 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1078 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1079 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1080 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1081 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1082 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1083 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1084 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1085 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1086 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1087 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1088 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1089 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1090 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1091 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1092 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1093 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1094 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1095 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1096 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1097 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1098 for sql/sql_hints.yy */  { "", 0, false, false},
/* reserved 1099 for sql/sql_hints.yy */  { "", 0, false, false},
/* PART 6: Digest special tokens. */
/* 1100 */  { "?", 1, true, false},
/* 1101 */  { "?, ...", 6, true, false},
/* 1102 */  { "(?)", 3, true, false},
/* 1103 */  { "(?) /* , ... */", 15, true, false},
/* 1104 */  { "(...)", 5, true, false},
/* 1105 */  { "(...) /* , ... */", 17, true, false},
/* 1106 */  { "(tok_id)", 8, true, false},
/* 1107 */  { "(tok_id_at)", 11, true, false},
/* 1108 */  { "/*+", 3, true, false},
/* 1109 */  { "*/", 2, true, false},
/* 1110 */  { "IN (...)", 8, true, false},
/* 1111 */  { "UNUSED", 6, true, false},
/* PART 7: padding reserved for digest special tokens. */
/* reserved 1112 for digest specials */  { "", 0, false, false},
/* reserved 1113 for digest specials */  { "", 0, false, false},
/* reserved 1114 for digest specials */  { "", 0, false, false},
/* reserved 1115 for digest specials */  { "", 0, false, false},
/* reserved 1116 for digest specials */  { "", 0, false, false},
/* reserved 1117 for digest specials */  { "", 0, false, false},
/* reserved 1118 for digest specials */  { "", 0, false, false},
/* reserved 1119 for digest specials */  { "", 0, false, false},
/* reserved 1120 for digest specials */  { "", 0, false, false},
/* reserved 1121 for digest specials */  { "", 0, false, false},
/* reserved 1122 for digest specials */  { "", 0, false, false},
/* reserved 1123 for digest specials */  { "", 0, false, false},
/* reserved 1124 for digest specials */  { "", 0, false, false},
/* reserved 1125 for digest specials */  { "", 0, false, false},
/* reserved 1126 for digest specials */  { "", 0, false, false},
/* reserved 1127 for digest specials */  { "", 0, false, false},
/* reserved 1128 for digest specials */  { "", 0, false, false},
/* reserved 1129 for digest specials */  { "", 0, false, false},
/* reserved 1130 for digest specials */  { "", 0, false, false},
/* reserved 1131 for digest specials */  { "", 0, false, false},
/* reserved 1132 for digest specials */  { "", 0, false, false},
/* reserved 1133 for digest specials */  { "", 0, false, false},
/* reserved 1134 for digest specials */  { "", 0, false, false},
/* reserved 1135 for digest specials */  { "", 0, false, false},
/* reserved 1136 for digest specials */  { "", 0, false, false},
/* reserved 1137 for digest specials */  { "", 0, false, false},
/* reserved 1138 for digest specials */  { "", 0, false, false},
/* reserved 1139 for digest specials */  { "", 0, false, false},
/* reserved 1140 for digest specials */  { "", 0, false, false},
/* reserved 1141 for digest specials */  { "", 0, false, false},
/* reserved 1142 for digest specials */  { "", 0, false, false},
/* reserved 1143 for digest specials */  { "", 0, false, false},
/* reserved 1144 for digest specials */  { "", 0, false, false},
/* reserved 1145 for digest specials */  { "", 0, false, false},
/* reserved 1146 for digest specials */  { "", 0, false, false},
/* reserved 1147 for digest specials */  { "", 0, false, false},
/* reserved 1148 for digest specials */  { "", 0, false, false},
/* reserved 1149 for digest specials */  { "", 0, false, false},
/* PART 8: named tokens from sql/sql_yacc.yy (chunk 2). */
/* 1150 */  { "(unknown)", 9, true, false},
/* 1151 */  { "JSON_VALUE", 10, true, false},
/* 1152 */  { "TLS", 3, true, false},
/* 1153 */  { "ATTRIBUTE", 9, true, false},
/* 1154 */  { "ENGINE_ATTRIBUTE", 16, true, false},
/* 1155 */  { "SECONDARY_ENGINE_ATTRIBUTE", 26, true, false},
/* 1156 */  { "SOURCE_CONNECTION_AUTO_FAILOVER", 31, true, false},
/* 1157 */  { "ZONE", 4, true, false},
/* 1158 */  { "(unknown)", 9, true, false},
/* 1159 */  { "REPLICA", 7, true, false},
/* 1160 */  { "REPLICAS", 8, true, false},
/* 1161 */  { "ASSIGN_GTIDS_TO_ANONYMOUS_TRANSACTIONS", 38, true, false},
/* 1162 */  { "GET_SOURCE_PUBLIC_KEY", 21, true, false},
/* 1163 */  { "SOURCE_AUTO_POSITION", 20, true, false},
/* 1164 */  { "SOURCE_BIND", 11, true, false},
/* 1165 */  { "SOURCE_COMPRESSION_ALGORITHMS", 29, true, false},
/* 1166 */  { "SOURCE_CONNECT_RETRY", 20, true, false},
/* 1167 */  { "SOURCE_DELAY", 12, true, false},
/* 1168 */  { "SOURCE_HEARTBEAT_PERIOD", 23, true, false},
/* 1169 */  { "SOURCE_HOST", 11, true, false},
/* 1170 */  { "SOURCE_LOG_FILE", 15, true, false},
/* 1171 */  { "SOURCE_LOG_POS", 14, true, false},
/* 1172 */  { "SOURCE_PASSWORD", 15, true, false},
/* 1173 */  { "SOURCE_PORT", 11, true, false},
/* 1174 */  { "SOURCE_PUBLIC_KEY_PATH", 22, true, false},
/* 1175 */  { "SOURCE_RETRY_COUNT", 18, true, false},
/* 1176 */  { "SOURCE_SSL", 10, true, false},
/* 1177 */  { "SOURCE_SSL_CA", 13, true, false},
/* 1178 */  { "SOURCE_SSL_CAPATH", 17, true, false},
/* 1179 */  { "SOURCE_SSL_CERT", 15, true, false},
/* 1180 */  { "SOURCE_SSL_CIPHER", 17, true, false},
/* 1181 */  { "SOURCE_SSL_CRL", 14, true, false},
/* 1182 */  { "SOURCE_SSL_CRLPATH", 18, true, false},
/* 1183 */  { "SOURCE_SSL_KEY", 14, true, false},
/* 1184 */  { "SOURCE_SSL_VERIFY_SERVER_CERT", 29, true, false},
/* 1185 */  { "SOURCE_TLS_CIPHERSUITES", 23, true, false},
/* 1186 */  { "SOURCE_TLS_VERSION", 18, true, false},
/* 1187 */  { "SOURCE_USER", 11, true, false},
/* 1188 */  { "SOURCE_ZSTD_COMPRESSION_LEVEL", 29, true, false},
/* 1189 */  { "ST_COLLECT", 10, true, false},
/* 1190 */  { "KEYRING", 7, true, false},
/* 1191 */  { "AUTHENTICATION", 14, true, false},
/* 1192 */  { "FACTOR", 6, true, false},
/* 1193 */  { "FINISH", 6, true, false},
/* 1194 */  { "INITIATE", 8, true, false},
/* 1195 */  { "REGISTRATION", 12, true, false},
/* 1196 */  { "UNREGISTER", 10, true, false},
/* 1197 */  { "INITIAL", 7, true, false},
/* 1198 */  { "CHALLENGE_RESPONSE", 18, true, false},
/* 1199 */  { "GTID_ONLY", 9, true, false},
/* 1200 */  { "INTERSECT", 9, true, false},
/* 1201 */  { "BULK", 4, true, false},
/* 1202 */  { "URL", 3, true, false},
/* 1203 */  { "GENERATE", 8, true, false},
/* 1204 */  { "(unknown)", 9, true, false},
/* 1205 */  { "PARSE_TREE", 10, true, false},
/* 1206 */  { "LOG", 3, true, false},
/* 1207 */  { "GTIDS", 5, true, false},
/* 1208 */  { "PARALLEL", 8, true, false},
/* 1209 */  { "S3", 2, true, false},
/* 1210 */  { "QUALIFY", 7, true, false},
/* 1211 */  { "AUTO", 4, true, false},
/* 1212 */  { "MANUAL", 6, true, false},
/* 1213 */  { "BERNOULLI", 9, true, false},
/* 1214 */  { "TABLESAMPLE", 11, true, false},
/* PART 9: End of token list. */
/* DUMMY */ { "", 0, false, false}
};
#endif /* LEX_TOKEN_WITH_DEFINITION */
/* DIGEST specific tokens. */
#define TOK_GENERIC_VALUE 1100
#define TOK_GENERIC_VALUE_LIST 1101
#define TOK_ROW_SINGLE_VALUE 1102
#define TOK_ROW_SINGLE_VALUE_LIST 1103
#define TOK_ROW_MULTIPLE_VALUE 1104
#define TOK_ROW_MULTIPLE_VALUE_LIST 1105
#define TOK_IDENT 1106
#define TOK_IDENT_AT 1107
#define TOK_HINT_COMMENT_OPEN 1108
#define TOK_HINT_COMMENT_CLOSE 1109
#define TOK_IN_GENERIC_VALUE_EXPRESSION 1110
#define TOK_UNUSED 1111
