/* Copyright (c) 2004, 2024, Oracle and/or its affiliates.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License, version 2.0,
   as published by the Free Software Foundation.

   This program is designed to work with certain software (including
   but not limited to OpenSSL) that is licensed under separate terms,
   as designated in a particular file or component or in included license
   documentation.  The authors of MySQL hereby grant you an additional
   permission to link the program and your derivative works with the
   separately licensed software that they have either included with
   the program or referenced in the documentation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License, version 2.0, for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
/*
  Do not edit this file, it is automatically generated from:
  </build/xtrabackup/src/percona-xtrabackup-8.4.0-6/scripts/mysql_system_tables_data.sql>
*/
const char* mysql_system_data[]={
"-- Copyright (c) 2007, 2024, Oracle and/or its affiliates.\n"
  "-- \n"
  "-- This program is free software; you can redistribute it and/or modify\n"
  "-- it under the terms of the GNU General Public License, version 2.0,\n"
  "-- as published by the Free Software Foundation.\n"
  "--\n"
  "-- This program is designed to work with certain software (including\n"
  "-- but not limited to OpenSSL) that is licensed under separate terms,\n"
  "-- as designated in a particular file or component or in included license\n"
  "-- documentation.  The authors of MySQL hereby grant you an additional\n"
  "-- permission to link the program and your derivative works with the\n"
  "-- separately licensed software that they have either included with\n"
  "-- the program or referenced in the documentation.\n"
  "--\n"
  "-- This program is distributed in the hope that it will be useful,\n"
  "-- but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
  "-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
  "-- GNU General Public License, version 2.0, for more details.\n"
  "--\n"
  "-- You should have received a copy of the GNU General Public License\n"
  "-- along with this program; if not, write to the Free Software\n"
  "-- Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA\n"
  "--\n"
  "-- The inital data for system tables of MySQL Server\n"
  "--\n"
  "START TRANSACTION;\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2000, 1, 'Anguilla 1957 / British West Indies Grid', 'EPSG', 2000, 'PROJCS[\"Anguilla 1957 / B"
  "ritish West Indies Grid\",GEOGCS[\"Anguilla 1957\",DATUM[\"Anguilla 1957\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145"
  ",293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6600\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4600\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-62,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2000\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2001, 1, 'Antigua 1943 / British West Indies Grid', 'EPSG', 2001, 'PROJCS[\"Antigua 1943 / Bri"
  "tish West Indies Grid\",GEOGCS[\"Antigua 1943\",DATUM[\"Antigua 1943\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293"
  ".465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-255,-15,71,0,0,0,0],AUTHORITY[\"EPSG\",\"6601\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4601\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-62,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2001\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2002, 1, 'Dominica 1945 / British West Indies Grid', 'EPSG', 2002, 'PROJCS[\"Dominica 1945 / B"
  "ritish West Indies Grid\",GEOGCS[\"Dominica 1945\",DATUM[\"Dominica 1945\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145"
  ",293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[725,685,536,0,0,0,0],AUTHORITY[\"EPSG\",\"6602\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4602\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-6"
  "2,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2002\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2003, 1, 'Grenada 1953 / British West Indies Grid', 'EPSG', 2003, 'PROJCS[\"Grenada 1953 / Bri"
  "tish West Indies Grid\",GEOGCS[\"Grenada 1953\",DATUM[\"Grenada 1953\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293"
  ".465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[72,213.7,93,0,0,0,0],AUTHORITY[\"EPSG\",\"6603\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4603\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-62,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2003\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2004, 1, 'Montserrat 1958 / British West Indies Grid', 'EPSG', 2004, 'PROJCS[\"Montserrat 1958"
  " / British West Indies Grid\",GEOGCS[\"Montserrat 1958\",DATUM[\"Montserrat 1958\",SPHEROID[\"Clarke 1880 (RGS)\",637"
  "8249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[174,359,365,0,0,0,0],AUTHORITY[\"EPSG\",\"6604\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4604\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-62,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "2004\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2005, 1, 'St. Kitts 1955 / British West Indies Grid', 'EPSG', 2005, 'PROJCS[\"St. Kitts 1955 /"
  " British West Indies Grid\",GEOGCS[\"St. Kitts 1955\",DATUM[\"St. Kitts 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249"
  ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[9,183,236,0,0,0,0],AUTHORITY[\"EPSG\",\"6605\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4605\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-62,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2005\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2006, 1, 'St. Lucia 1955 / British West Indies Grid', 'EPSG', 2006, 'PROJCS[\"St. Lucia 1955 /"
  " British West Indies Grid\",GEOGCS[\"St. Lucia 1955\",DATUM[\"St. Lucia 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249"
  ".145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-153,153,307,0,0,0,0],AUTHORITY[\"EPSG\",\"6606\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4606\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-62,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"200"
  "6\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2007, 1, 'St. Vincent 45 / British West Indies Grid', 'EPSG', 2007, 'PROJCS[\"St. Vincent 45 /"
  " British West Indies Grid\",GEOGCS[\"St. Vincent 1945\",DATUM[\"St. Vincent 1945\",SPHEROID[\"Clarke 1880 (RGS)\",637"
  "8249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[195.671,332.517,274.607,0,0,0,0],AUTHORITY[\"EPSG\",\"6607\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4607\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY["
  "\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-62,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY"
  "[\"EPSG\",\"2007\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2009, 1, 'NAD27(CGQ77) / SCoPQ zone 3', 'EPSG', 2009, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 3\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-58.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2009\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2010, 1, 'NAD27(CGQ77) / SCoPQ zone 4', 'EPSG', 2010, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 4\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-61.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2010\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2011, 1, 'NAD27(CGQ77) / SCoPQ zone 5', 'EPSG', 2011, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 5\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-64.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2011\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2012, 1, 'NAD27(CGQ77) / SCoPQ zone 6', 'EPSG', 2012, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 6\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-67.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2012\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2013, 1, 'NAD27(CGQ77) / SCoPQ zone 7', 'EPSG', 2013, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 7\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2013\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2014, 1, 'NAD27(CGQ77) / SCoPQ zone 8', 'EPSG', 2014, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 8\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-73.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2015, 1, 'NAD27(CGQ77) / SCoPQ zone 9', 'EPSG', 2015, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 9\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2015\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2016, 1, 'NAD27(CGQ77) / SCoPQ zone 10', 'EPSG', 2016, 'PROJCS[\"NAD27(CGQ77) / SCoPQ zone 10\""
  ",GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213"
  "8982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-79.5111111111111,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2016\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2017, 1, 'NAD27(76) / MTM zone 8', 'EPSG', 2017, 'PROJCS[\"NAD27(76) / MTM zone 8\",GEOGCS[\"N"
  "AD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4608\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-73.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2017\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2018, 1, 'NAD27(76) / MTM zone 9', 'EPSG', 2018, 'PROJCS[\"NAD27(76) / MTM zone 9\",GEOGCS[\"N"
  "AD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4608\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2018\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2019, 1, 'NAD27(76) / MTM zone 10', 'EPSG', 2019, 'PROJCS[\"NAD27(76) / MTM zone 10\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2019\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2020, 1, 'NAD27(76) / MTM zone 11', 'EPSG', 2020, 'PROJCS[\"NAD27(76) / MTM zone 11\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2020\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2021, 1, 'NAD27(76) / MTM zone 12', 'EPSG', 2021, 'PROJCS[\"NAD27(76) / MTM zone 12\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2021\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2022, 1, 'NAD27(76) / MTM zone 13', 'EPSG', 2022, 'PROJCS[\"NAD27(76) / MTM zone 13\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2022\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2023, 1, 'NAD27(76) / MTM zone 14', 'EPSG', 2023, 'PROJCS[\"NAD27(76) / MTM zone 14\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2023\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2024, 1, 'NAD27(76) / MTM zone 15', 'EPSG', 2024, 'PROJCS[\"NAD27(76) / MTM zone 15\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2024\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2025, 1, 'NAD27(76) / MTM zone 16', 'EPSG', 2025, 'PROJCS[\"NAD27(76) / MTM zone 16\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2025\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2026, 1, 'NAD27(76) / MTM zone 17', 'EPSG', 2026, 'PROJCS[\"NAD27(76) / MTM zone 17\",GEOGCS[\""
  "NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"460"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2026\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2027, 1, 'NAD27(76) / UTM zone 15N', 'EPSG', 2027, 'PROJCS[\"NAD27(76) / UTM zone 15N\",GEOGCS"
  "[\"NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4608\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2027\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2028, 1, 'NAD27(76) / UTM zone 16N', 'EPSG', 2028, 'PROJCS[\"NAD27(76) / UTM zone 16N\",GEOGCS"
  "[\"NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4608\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2029, 1, 'NAD27(76) / UTM zone 17N', 'EPSG', 2029, 'PROJCS[\"NAD27(76) / UTM zone 17N\",GEOGCS"
  "[\"NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4608\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2029\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2030, 1, 'NAD27(76) / UTM zone 18N', 'EPSG', 2030, 'PROJCS[\"NAD27(76) / UTM zone 18N\",GEOGCS"
  "[\"NAD27(76)\",DATUM[\"North American Datum 1927 (1976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4608\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2030\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2031, 1, 'NAD27(CGQ77) / UTM zone 17N', 'EPSG', 2031, 'PROJCS[\"NAD27(CGQ77) / UTM zone 17N\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2031\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2032, 1, 'NAD27(CGQ77) / UTM zone 18N', 'EPSG', 2032, 'PROJCS[\"NAD27(CGQ77) / UTM zone 18N\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2032\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2033, 1, 'NAD27(CGQ77) / UTM zone 19N', 'EPSG', 2033, 'PROJCS[\"NAD27(CGQ77) / UTM zone 19N\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2033\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2034, 1, 'NAD27(CGQ77) / UTM zone 20N', 'EPSG', 2034, 'PROJCS[\"NAD27(CGQ77) / UTM zone 20N\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2034\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2035, 1, 'NAD27(CGQ77) / UTM zone 21N', 'EPSG', 2035, 'PROJCS[\"NAD27(CGQ77) / UTM zone 21N\","
  "GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138"
  "982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4609\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2039, 1, 'Israel 1993 / Israeli TM Grid', 'EPSG', 2039, 'PROJCS[\"Israel 1993 / Israeli TM Gri"
  "d\",GEOGCS[\"Israel 1993\",DATUM[\"Israel 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[-48,55,52,0,0,0,0],AUTHORITY[\"EPSG\",\"6141\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4141\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",31.7343936111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",35.2045169444444,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000067,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",219529.584,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",626907.39,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "2039\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2040, 1, 'Locodjo 1965 / UTM zone 30N', 'EPSG', 2040, 'PROJCS[\"Locodjo 1965 / UTM zone 30N\","
  "GEOGCS[\"Locodjo 1965\",DATUM[\"Locodjo 1965\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"EPSG\",\"6142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4142\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2040\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2041, 1, 'Abidjan 1987 / UTM zone 30N', 'EPSG', 2041, 'PROJCS[\"Abidjan 1987 / UTM zone 30N\","
  "GEOGCS[\"Abidjan 1987\",DATUM[\"Abidjan 1987\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-124.76,53,466.79,0,0,0,0],AUTHORITY[\"EPSG\",\"6143\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4143\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2041\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2042, 1, 'Locodjo 1965 / UTM zone 29N', 'EPSG', 2042, 'PROJCS[\"Locodjo 1965 / UTM zone 29N\","
  "GEOGCS[\"Locodjo 1965\",DATUM[\"Locodjo 1965\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"EPSG\",\"6142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4142\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2042\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2043, 1, 'Abidjan 1987 / UTM zone 29N', 'EPSG', 2043, 'PROJCS[\"Abidjan 1987 / UTM zone 29N\","
  "GEOGCS[\"Abidjan 1987\",DATUM[\"Abidjan 1987\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-124.76,53,466.79,0,0,0,0],AUTHORITY[\"EPSG\",\"6143\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4143\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2043\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2044, 1, 'Hanoi 1972 / Gauss-Kruger zone 18', 'EPSG', 2044, 'PROJCS[\"Hanoi 1972 / Gauss-Kruge"
  "r zone 18\",GEOGCS[\"Hanoi 1972\",DATUM[\"Hanoi 1972\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\","
  "\"7024\"]],TOWGS84[-17.51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\",\"6147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4147\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2044\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2045, 1, 'Hanoi 1972 / Gauss-Kruger zone 19', 'EPSG', 2045, 'PROJCS[\"Hanoi 1972 / Gauss-Kruge"
  "r zone 19\",GEOGCS[\"Hanoi 1972\",DATUM[\"Hanoi 1972\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\","
  "\"7024\"]],TOWGS84[-17.51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\",\"6147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4147\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2045\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2046, 1, 'Hartebeesthoek94 / Lo15', 'EPSG', 2046, 'PROJCS[\"Hartebeesthoek94 / Lo15\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2046\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2047, 1, 'Hartebeesthoek94 / Lo17', 'EPSG', 2047, 'PROJCS[\"Hartebeesthoek94 / Lo17\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2047\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2048, 1, 'Hartebeesthoek94 / Lo19', 'EPSG', 2048, 'PROJCS[\"Hartebeesthoek94 / Lo19\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2048\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2049, 1, 'Hartebeesthoek94 / Lo21', 'EPSG', 2049, 'PROJCS[\"Hartebeesthoek94 / Lo21\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2049\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2050, 1, 'Hartebeesthoek94 / Lo23', 'EPSG', 2050, 'PROJCS[\"Hartebeesthoek94 / Lo23\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2050\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2051, 1, 'Hartebeesthoek94 / Lo25', 'EPSG', 2051, 'PROJCS[\"Hartebeesthoek94 / Lo25\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2051\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2052, 1, 'Hartebeesthoek94 / Lo27', 'EPSG', 2052, 'PROJCS[\"Hartebeesthoek94 / Lo27\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2052\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2053, 1, 'Hartebeesthoek94 / Lo29', 'EPSG', 2053, 'PROJCS[\"Hartebeesthoek94 / Lo29\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",29,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2053\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2054, 1, 'Hartebeesthoek94 / Lo31', 'EPSG', 2054, 'PROJCS[\"Hartebeesthoek94 / Lo31\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2054\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2055, 1, 'Hartebeesthoek94 / Lo33', 'EPSG', 2055, 'PROJCS[\"Hartebeesthoek94 / Lo33\",GEOGCS[\""
  "Hartebeesthoek94\",DATUM[\"Hartebeesthoek94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4148\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"2055\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2056, 1, 'CH1903+ / LV95', 'EPSG', 2056, 'PROJCS[\"CH1903+ / LV95\",GEOGCS[\"CH1903+\",DATUM[\""
  "CH1903+\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[674.374,15.056,405.3"
  "46,0,0,0,0],AUTHORITY[\"EPSG\",\"6150\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4150\"]],PRO"
  "JECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETER[\"Latitude of projection cent"
  "re\",46.9524055555556,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",7.43958333333333,AUT"
  "HORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",90,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle f"
  "rom Rectified to Skew Grid\",90,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",1,AUTHORITY["
  "\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection centre\",2600000,AUTHORITY[\"EPSG\",\"8816\"]],PARAMETER[\"Nort"
  "hing at projection centre\",1200000,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2056\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2057, 1, 'Rassadiran / Nakhl e Taqi', 'EPSG', 2057, 'PROJCS[\"Rassadiran / Nakhl e Taqi\",GEOG"
  "CS[\"Rassadiran\",DATUM[\"Rassadiran\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG"
  "S84[-133.63,-157.5,-158.62,0,0,0,0],AUTHORITY[\"EPSG\",\"6153\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4153\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETER[\"Lat"
  "itude of projection centre\",27.5188288055556,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centr"
  "e\",52.6035391666667,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",0.571661194444444,AUTHORITY["
  "\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",0.571661194444444,AUTHORITY[\"EPSG\",\"8814\"]],P"
  "ARAMETER[\"Scale factor on initial line\",0.999895934,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection"
  " centre\",658377.437,AUTHORITY[\"EPSG\",\"8816\"]],PARAMETER[\"Northing at projection centre\",3044969.194,AUTHORITY["
  "\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPS"
  "G\",\"2057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2058, 1, 'ED50(ED77) / UTM zone 38N', 'EPSG', 2058, 'PROJCS[\"ED50(ED77) / UTM zone 38N\",GEOG"
  "CS[\"ED50(ED77)\",DATUM[\"European Datum 1950(1977)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-110.33,-97.73,-119.85,0.3423,1.1634,0.2715,0.063],AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4154\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2058\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2059, 1, 'ED50(ED77) / UTM zone 39N', 'EPSG', 2059, 'PROJCS[\"ED50(ED77) / UTM zone 39N\",GEOG"
  "CS[\"ED50(ED77)\",DATUM[\"European Datum 1950(1977)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-110.33,-97.73,-119.85,0.3423,1.1634,0.2715,0.063],AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4154\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2059\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2060, 1, 'ED50(ED77) / UTM zone 40N', 'EPSG', 2060, 'PROJCS[\"ED50(ED77) / UTM zone 40N\",GEOG"
  "CS[\"ED50(ED77)\",DATUM[\"European Datum 1950(1977)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-110.33,-97.73,-119.85,0.3423,1.1634,0.2715,0.063],AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4154\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2060\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2061, 1, 'ED50(ED77) / UTM zone 41N', 'EPSG', 2061, 'PROJCS[\"ED50(ED77) / UTM zone 41N\",GEOG"
  "CS[\"ED50(ED77)\",DATUM[\"European Datum 1950(1977)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-110.33,-97.73,-119.85,0.3423,1.1634,0.2715,0.063],AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4154\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2061\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2062, 1, 'Madrid 1870 (Madrid) / Spain', 'EPSG', 2062, 'PROJCS[\"Madrid 1870 (Madrid) / Spain\""
  ",GEOGCS[\"Madrid 1870 (Madrid)\",DATUM[\"Madrid 1870 (Madrid)\",SPHEROID[\"Struve 1860\",6378298.3,294.73,AUTHORITY[\""
  "EPSG\",\"7028\"]],AUTHORITY[\"EPSG\",\"6903\"]],PRIMEM[\"Madrid\",-3.6879388888888895,AUTHORITY[\"EPSG\",\"8905\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4903\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of "
  "natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9988085293,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2062\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2065, 1, 'S-JTSK (Ferro) / Krovak', 'EPSG', 2065, 'PROJCS[\"S-JTSK (Ferro) / Krovak\",GEOGCS[\""
  "S-JTSK (Ferro)\",DATUM[\"System of the Unified Trigonometrical Cadastral Network (Ferro)\",SPHEROID[\"Bessel 1841\",6"
  "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6818\"]],PRIMEM[\"Ferro\",-17.6777777777777"
  "74,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4818\"]],PROJECTION[\"Krovak\",AUTHORITY[\"EPSG\",\"9819\"]],PARAMETER[\"L"
  "atitude of projection centre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of origin\",42.51"
  "11111111111,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone axis\",30.2881397527778,AUTHORITY[\"EPSG\","
  "\"1036\"]],PARAMETER[\"Latitude of pseudo standard parallel\",78.5111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETE"
  "R[\"Scale factor on pseudo standard parallel\",0.9999,AUTHORITY[\"EPSG\",\"8819\"]],PARAMETER[\"False easting\",0,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"EPSG\",\"2065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2066, 1, 'Mount Dillon / Tobago Grid', 'EPSG', 2066, 'PROJCS[\"Mount Dillon / Tobago Grid\",GE"
  "OGCS[\"Mount Dillon\",DATUM[\"Mount Dillon\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.26067636926103,AUTHORITY["
  "\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6157\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4157\""
  "]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",11.252178611"
  "1111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-60.6860088888889,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"False easting\",187500,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",180000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"Clarke\\'s link\",0.201166195164,AUTHORITY[\"EPSG\",\"9039\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2067, 1, 'Naparima 1955 / UTM zone 20N', 'EPSG', 2067, 'PROJCS[\"Naparima 1955 / UTM zone 20N\""
  ",GEOGCS[\"Naparima 1955\",DATUM[\"Naparima 1955\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-0.465,372.095,171.736,0,0,0,0],AUTHORITY[\"EPSG\",\"6158\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4158\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2067\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2068, 1, 'ELD79 / Libya zone 5', 'EPSG', 2068, 'PROJCS[\"ELD79 / Libya zone 5\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2068\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2069, 1, 'ELD79 / Libya zone 6', 'EPSG', 2069, 'PROJCS[\"ELD79 / Libya zone 6\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2069\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2070, 1, 'ELD79 / Libya zone 7', 'EPSG', 2070, 'PROJCS[\"ELD79 / Libya zone 7\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2070\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2071, 1, 'ELD79 / Libya zone 8', 'EPSG', 2071, 'PROJCS[\"ELD79 / Libya zone 8\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2071\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2072, 1, 'ELD79 / Libya zone 9', 'EPSG', 2072, 'PROJCS[\"ELD79 / Libya zone 9\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2072\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2073, 1, 'ELD79 / Libya zone 10', 'EPSG', 2073, 'PROJCS[\"ELD79 / Libya zone 10\",GEOGCS[\"ELD"
  "79\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2073\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2074, 1, 'ELD79 / Libya zone 11', 'EPSG', 2074, 'PROJCS[\"ELD79 / Libya zone 11\",GEOGCS[\"ELD"
  "79\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2074\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2075, 1, 'ELD79 / Libya zone 12', 'EPSG', 2075, 'PROJCS[\"ELD79 / Libya zone 12\",GEOGCS[\"ELD"
  "79\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2075\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2076, 1, 'ELD79 / Libya zone 13', 'EPSG', 2076, 'PROJCS[\"ELD79 / Libya zone 13\",GEOGCS[\"ELD"
  "79\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2076\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2077, 1, 'ELD79 / UTM zone 32N', 'EPSG', 2077, 'PROJCS[\"ELD79 / UTM zone 32N\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2077\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2078, 1, 'ELD79 / UTM zone 33N', 'EPSG', 2078, 'PROJCS[\"ELD79 / UTM zone 33N\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2078\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2079, 1, 'ELD79 / UTM zone 34N', 'EPSG', 2079, 'PROJCS[\"ELD79 / UTM zone 34N\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2079\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2080, 1, 'ELD79 / UTM zone 35N', 'EPSG', 2080, 'PROJCS[\"ELD79 / UTM zone 35N\",GEOGCS[\"ELD79"
  "\",DATUM[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOW"
  "GS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2080\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2081, 1, 'Chos Malal 1914 / Argentina 2', 'EPSG', 2081, 'PROJCS[\"Chos Malal 1914 / Argentina "
  "2\",GEOGCS[\"Chos Malal 1914\",DATUM[\"Chos Malal 1914\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG"
  "\",\"7022\"]],AUTHORITY[\"EPSG\",\"6160\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4160\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2081\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2082, 1, 'Pampa del Castillo / Argentina 2', 'EPSG', 2082, 'PROJCS[\"Pampa del Castillo / Arge"
  "ntina 2\",GEOGCS[\"Pampa del Castillo\",DATUM[\"Pampa del Castillo\",SPHEROID[\"International 1924\",6378388,297,AUTH"
  "ORITY[\"EPSG\",\"7022\"]],TOWGS84[27.5,14,186.4,0,0,0,0],AUTHORITY[\"EPSG\",\"6161\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4161\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2082\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2083, 1, 'Hito XVIII 1963 / Argentina 2', 'EPSG', 2083, 'PROJCS[\"Hito XVIII 1963 / Argentina "
  "2\",GEOGCS[\"Hito XVIII 1963\",DATUM[\"Hito XVIII 1963\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG"
  "\",\"7022\"]],TOWGS84[18.38,192.45,96.82,0.056,-0.142,-0.2,-0.0013],AUTHORITY[\"EPSG\",\"6254\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4254\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2083\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2084, 1, 'Hito XVIII 1963 / UTM zone 19S', 'EPSG', 2084, 'PROJCS[\"Hito XVIII 1963 / UTM zone "
  "19S\",GEOGCS[\"Hito XVIII 1963\",DATUM[\"Hito XVIII 1963\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[18.38,192.45,96.82,0.056,-0.142,-0.2,-0.0013],AUTHORITY[\"EPSG\",\"6254\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4254\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"2084\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2087, 1, 'ELD79 / TM 12 NE', 'EPSG', 2087, 'PROJCS[\"ELD79 / TM 12 NE\",GEOGCS[\"ELD79\",DATUM"
  "[\"European Libyan Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-19"
  "4.513,-63.978,-25.759,-3.4027,3.756,-3.352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4159\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2087\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2088, 1, 'Carthage / TM 11 NE', 'EPSG', 2088, 'PROJCS[\"Carthage / TM 11 NE\",GEOGCS[\"Carthag"
  "e\",DATUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS"
  "84[-263,6,431,0,0,0,0],AUTHORITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degr"
  "ee\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4"
  "223\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2088\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2089, 1, 'Yemen NGN96 / UTM zone 38N', 'EPSG', 2089, 'PROJCS[\"Yemen NGN96 / UTM zone 38N\",GE"
  "OGCS[\"Yemen NGN96\",DATUM[\"Yemen National Geodetic Network 1996\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORI"
  "TY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4163\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2089\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2090, 1, 'Yemen NGN96 / UTM zone 39N', 'EPSG', 2090, 'PROJCS[\"Yemen NGN96 / UTM zone 39N\",GE"
  "OGCS[\"Yemen NGN96\",DATUM[\"Yemen National Geodetic Network 1996\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORI"
  "TY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4163\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2090\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2093, 1, 'Hanoi 1972 / GK 106 NE', 'EPSG', 2093, 'PROJCS[\"Hanoi 1972 / GK 106 NE\",GEOGCS[\"H"
  "anoi 1972\",DATUM[\"Hanoi 1972\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-17"
  ".51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\",\"6147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4147\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",106,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2093\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2094, 1, 'WGS 72BE / TM 106 NE', 'EPSG', 2094, 'PROJCS[\"WGS 72BE / TM 106 NE\",GEOGCS[\"WGS 7"
  "2BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],T"
  "OWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",106,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2094\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2095, 1, 'Bissau / UTM zone 28N', 'EPSG', 2095, 'PROJCS[\"Bissau / UTM zone 28N\",GEOGCS[\"Bis"
  "sau\",DATUM[\"Bissau\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-173,253,27,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6165\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4165\"]],PROJEC"
  "TION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPS"
  "G\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2096, 1, 'Korean 1985 / East Belt', 'EPSG', 2096, 'PROJCS[\"Korean 1985 / East Belt\",GEOGCS[\""
  "Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\""
  "]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2096\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2097, 1, 'Korean 1985 / Central Belt', 'EPSG', 2097, 'PROJCS[\"Korean 1985 / Central Belt\",GE"
  "OGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\","
  "\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",127,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2097\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2098, 1, 'Korean 1985 / West Belt', 'EPSG', 2098, 'PROJCS[\"Korean 1985 / West Belt\",GEOGCS[\""
  "Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\""
  "]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",125,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2098\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2099, 1, 'Qatar 1948 / Qatar Grid', 'EPSG', 2099, 'PROJCS[\"Qatar 1948 / Qatar Grid\",GEOGCS[\""
  "Qatar 1948\",DATUM[\"Qatar 1948\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"E"
  "PSG\",\"6286\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4286\"]],PROJECTION[\"Cassini-Soldner"
  "\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",25.3823611111111,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",50.7613888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting"
  "\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2099\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2100, 1, 'GGRS87 / Greek Grid', 'EPSG', 2100, 'PROJCS[\"GGRS87 / Greek Grid\",GEOGCS[\"GGRS87\""
  ",DATUM[\"Greek Geodetic Reference System 1987\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[-199.87,74.79,246.62,0,0,0,0],AUTHORITY[\"EPSG\",\"6121\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4121\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2100\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2101, 1, 'Lake / Maracaibo Grid M1', 'EPSG', 2101, 'PROJCS[\"Lake / Maracaibo Grid M1\",GEOGCS"
  "[\"Lake\",DATUM[\"Lake\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\""
  ",\"6249\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4249\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",10.1777777777778,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.6056177777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",-52684.972,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2101\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2102, 1, 'Lake / Maracaibo Grid', 'EPSG', 2102, 'PROJCS[\"Lake / Maracaibo Grid\",GEOGCS[\"Lak"
  "e\",DATUM[\"Lake\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"624"
  "9\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4249\"]],PROJECTION[\"Lambert Conic Conformal (1"
  "SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",10.1777777777778,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-71.6056177777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",147315.028,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2102\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2103, 1, 'Lake / Maracaibo Grid M3', 'EPSG', 2103, 'PROJCS[\"Lake / Maracaibo Grid M3\",GEOGCS"
  "[\"Lake\",DATUM[\"Lake\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\""
  ",\"6249\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4249\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",10.1777777777778,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.6056177777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",447315.028,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2103\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2104, 1, 'Lake / Maracaibo La Rosa Grid', 'EPSG', 2104, 'PROJCS[\"Lake / Maracaibo La Rosa Gri"
  "d\",GEOGCS[\"Lake\",DATUM[\"Lake\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORIT"
  "Y[\"EPSG\",\"6249\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4249\"]],PROJECTION[\"Lambert Co"
  "nic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",10.1777777777778,AUTHORI"
  "TY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.6056177777778,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",-17044,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-23139.97,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2104\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2105, 1, 'NZGD2000 / Mount Eden 2000', 'EPSG', 2105, 'PROJCS[\"NZGD2000 / Mount Eden 2000\",GE"
  "OGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",-36.8797222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",174.764166666"
  "667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"21"
  "05\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2106, 1, 'NZGD2000 / Bay of Plenty 2000', 'EPSG', 2106, 'PROJCS[\"NZGD2000 / Bay of Plenty 200"
  "0\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI"
  "TY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-37.7611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",176.466"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\""
  "2106\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2107, 1, 'NZGD2000 / Poverty Bay 2000', 'EPSG', 2107, 'PROJCS[\"NZGD2000 / Poverty Bay 2000\","
  "GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-38.6244444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177.88555555"
  "5556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2107\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2108, 1, 'NZGD2000 / Hawkes Bay 2000', 'EPSG', 2108, 'PROJCS[\"NZGD2000 / Hawkes Bay 2000\",GE"
  "OGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",-39.6508333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",176.673611111"
  "111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2108\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2109, 1, 'NZGD2000 / Taranaki 2000', 'EPSG', 2109, 'PROJCS[\"NZGD2000 / Taranaki 2000\",GEOGCS"
  "[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",-39.1355555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",174.227777777778,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2109\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2110, 1, 'NZGD2000 / Tuhirangi 2000', 'EPSG', 2110, 'PROJCS[\"NZGD2000 / Tuhirangi 2000\",GEOG"
  "CS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS"
  "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",-39.5122222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",175.64,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2110\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2111, 1, 'NZGD2000 / Wanganui 2000', 'EPSG', 2111, 'PROJCS[\"NZGD2000 / Wanganui 2000\",GEOGCS"
  "[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",-40.2419444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",175.488055555556,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2111\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2112, 1, 'NZGD2000 / Wairarapa 2000', 'EPSG', 2112, 'PROJCS[\"NZGD2000 / Wairarapa 2000\",GEOG"
  "CS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS"
  "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",-40.9252777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",175.647222222222"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2112\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2113, 1, 'NZGD2000 / Wellington 2000', 'EPSG', 2113, 'PROJCS[\"NZGD2000 / Wellington 2000\",GE"
  "OGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",-41.3011111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",174.776388888"
  "889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2113\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2114, 1, 'NZGD2000 / Collingwood 2000', 'EPSG', 2114, 'PROJCS[\"NZGD2000 / Collingwood 2000\","
  "GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-40.7147222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",172.67194444"
  "4444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2114\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2115, 1, 'NZGD2000 / Nelson 2000', 'EPSG', 2115, 'PROJCS[\"NZGD2000 / Nelson 2000\",GEOGCS[\"N"
  "ZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-41.2744444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",173.299166666667,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2115\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2116, 1, 'NZGD2000 / Karamea 2000', 'EPSG', 2116, 'PROJCS[\"NZGD2000 / Karamea 2000\",GEOGCS[\""
  "NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-41.2897222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",172.108888888889,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2116\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2117, 1, 'NZGD2000 / Buller 2000', 'EPSG', 2117, 'PROJCS[\"NZGD2000 / Buller 2000\",GEOGCS[\"N"
  "ZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-41.8105555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171.581111111111,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2117\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2118, 1, 'NZGD2000 / Grey 2000', 'EPSG', 2118, 'PROJCS[\"NZGD2000 / Grey 2000\",GEOGCS[\"NZGD2"
  "000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",-42.3336111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171.549722222222,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2118\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2119, 1, 'NZGD2000 / Amuri 2000', 'EPSG', 2119, 'PROJCS[\"NZGD2000 / Amuri 2000\",GEOGCS[\"NZG"
  "D2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",-42.6888888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",173.01,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2120, 1, 'NZGD2000 / Marlborough 2000', 'EPSG', 2120, 'PROJCS[\"NZGD2000 / Marlborough 2000\","
  "GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-41.5444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",173.80194444"
  "4445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2120\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2121, 1, 'NZGD2000 / Hokitika 2000', 'EPSG', 2121, 'PROJCS[\"NZGD2000 / Hokitika 2000\",GEOGCS"
  "[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",-42.8861111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",170.979722222222,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2121\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2122, 1, 'NZGD2000 / Okarito 2000', 'EPSG', 2122, 'PROJCS[\"NZGD2000 / Okarito 2000\",GEOGCS[\""
  "NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-43.11,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",170.260833333333,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2123, 1, 'NZGD2000 / Jacksons Bay 2000', 'EPSG', 2123, 'PROJCS[\"NZGD2000 / Jacksons Bay 2000\""
  ",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-43.9777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",168.606111"
  "111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"212"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2124, 1, 'NZGD2000 / Mount Pleasant 2000', 'EPSG', 2124, 'PROJCS[\"NZGD2000 / Mount Pleasant 2"
  "000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO"
  "RITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",-43.5905555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",172.7"
  "26944444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\","
  "\"2124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2125, 1, 'NZGD2000 / Gawler 2000', 'EPSG', 2125, 'PROJCS[\"NZGD2000 / Gawler 2000\",GEOGCS[\"N"
  "ZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-43.7486111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171.360555555556,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2125\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2126, 1, 'NZGD2000 / Timaru 2000', 'EPSG', 2126, 'PROJCS[\"NZGD2000 / Timaru 2000\",GEOGCS[\"N"
  "ZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-44.4019444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171.057222222222,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2126\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2127, 1, 'NZGD2000 / Lindis Peak 2000', 'EPSG', 2127, 'PROJCS[\"NZGD2000 / Lindis Peak 2000\","
  "GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-44.735,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",169.4675,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2128, 1, 'NZGD2000 / Mount Nicholas 2000', 'EPSG', 2128, 'PROJCS[\"NZGD2000 / Mount Nicholas 2"
  "000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO"
  "RITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",-45.1327777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",168.3"
  "98611111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\","
  "\"2128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2129, 1, 'NZGD2000 / Mount York 2000', 'EPSG', 2129, 'PROJCS[\"NZGD2000 / Mount York 2000\",GE"
  "OGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",-45.5636111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",167.738611111"
  "111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2129\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2130, 1, 'NZGD2000 / Observation Point 2000', 'EPSG', 2130, 'PROJCS[\"NZGD2000 / Observation P"
  "oint 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",-45.8161111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",170.628333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"E"
  "PSG\",\"2130\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2131, 1, 'NZGD2000 / North Taieri 2000', 'EPSG', 2131, 'PROJCS[\"NZGD2000 / North Taieri 2000\""
  ",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-45.8613888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",170.2825,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99996,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2131\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2132, 1, 'NZGD2000 / Bluff 2000', 'EPSG', 2132, 'PROJCS[\"NZGD2000 / Bluff 2000\",GEOGCS[\"NZG"
  "D2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",-46.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",168.342777777778,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2132\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2133, 1, 'NZGD2000 / UTM zone 58S', 'EPSG', 2133, 'PROJCS[\"NZGD2000 / UTM zone 58S\",GEOGCS[\""
  "NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2133\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2134, 1, 'NZGD2000 / UTM zone 59S', 'EPSG', 2134, 'PROJCS[\"NZGD2000 / UTM zone 59S\",GEOGCS[\""
  "NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2134\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2135, 1, 'NZGD2000 / UTM zone 60S', 'EPSG', 2135, 'PROJCS[\"NZGD2000 / UTM zone 60S\",GEOGCS[\""
  "NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2135\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2136, 1, 'Accra / Ghana National Grid', 'EPSG', 2136, 'PROJCS[\"Accra / Ghana National Grid\","
  "GEOGCS[\"Accra\",DATUM[\"Accra\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG\",\"7029\"]],TOWGS84[-199,32,32"
  "2,0,0,0,0],AUTHORITY[\"EPSG\",\"6168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4168\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",4.6666666666666"
  "7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.99975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",900000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Gold Coast foot\",0.30479971"
  "01815088,AUTHORITY[\"EPSG\",\"9094\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2137, 1, 'Accra / TM 1 NW', 'EPSG', 2137, 'PROJCS[\"Accra / TM 1 NW\",GEOGCS[\"Accra\",DATUM[\""
  "Accra\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG\",\"7029\"]],TOWGS84[-199,32,322,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4168\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"2137\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2138, 1, 'NAD27(CGQ77) / Quebec Lambert', 'EPSG', 2138, 'PROJCS[\"NAD27(CGQ77) / Quebec Lamber"
  "t\",GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698"
  "2138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4609\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",44,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-68.5111111111111,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",60,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2138\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2154, 1, 'RGF93 / Lambert-93', 'EPSG', 2154, 'PROJCS[\"RGF93 / Lambert-93\",GEOGCS[\"RGF93\",D"
  "ATUM[\"Reseau Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"417"
  "1\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin"
  "\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",44,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",700000,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",6600000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2154\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2157, 1, 'IRENET95 / Irish Transverse Mercator', 'EPSG', 2157, 'PROJCS[\"IRENET95 / Irish Tran"
  "sverse Mercator\",GEOGCS[\"IRENET95\",DATUM[\"IRENET95\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6173\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4173\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",53.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99982,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",750000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2157\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2158, 1, 'IRENET95 / UTM zone 29N', 'EPSG', 2158, 'PROJCS[\"IRENET95 / UTM zone 29N\",GEOGCS[\""
  "IRENET95\",DATUM[\"IRENET95\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6173\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4173\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2158\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2159, 1, 'Sierra Leone 1924 / New Colony Grid', 'EPSG', 2159, 'PROJCS[\"Sierra Leone 1924 / Ne"
  "w Colony Grid\",GEOGCS[\"Sierra Leone 1924\",DATUM[\"Sierra Leone Colony 1924\",SPHEROID[\"War Office\",6378300,296,A"
  "UTHORITY[\"EPSG\",\"7029\"]],AUTHORITY[\"EPSG\",\"6174\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4174\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",6.66666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-12,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Gold Coast foot"
  "\",0.3047997101815088,AUTHORITY[\"EPSG\",\"9094\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2159\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2160, 1, 'Sierra Leone 1924 / New War Office Grid', 'EPSG', 2160, 'PROJCS[\"Sierra Leone 1924 "
  "/ New War Office Grid\",GEOGCS[\"Sierra Leone 1924\",DATUM[\"Sierra Leone Colony 1924\",SPHEROID[\"War Office\",63783"
  "00,296,AUTHORITY[\"EPSG\",\"7029\"]],AUTHORITY[\"EPSG\",\"6174\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4174\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",6.66666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-12,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Go"
  "ld Coast foot\",0.3047997101815088,AUTHORITY[\"EPSG\",\"9094\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2160\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2161, 1, 'Sierra Leone 1968 / UTM zone 28N', 'EPSG', 2161, 'PROJCS[\"Sierra Leone 1968 / UTM z"
  "one 28N\",GEOGCS[\"Sierra Leone 1968\",DATUM[\"Sierra Leone 1968\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465"
  ",AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-88,4,101,0,0,0,0],AUTHORITY[\"EPSG\",\"6175\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4175\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2161\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2162, 1, 'Sierra Leone 1968 / UTM zone 29N', 'EPSG', 2162, 'PROJCS[\"Sierra Leone 1968 / UTM z"
  "one 29N\",GEOGCS[\"Sierra Leone 1968\",DATUM[\"Sierra Leone 1968\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465"
  ",AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-88,4,101,0,0,0,0],AUTHORITY[\"EPSG\",\"6175\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4175\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2162\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2163, 1, 'US National Atlas Equal Area', 'EPSG', 2163, 'PROJCS[\"US National Atlas Equal Area\""
  ",GEOGCS[\"Unspecified datum based upon the Clarke 1866 Authalic Sphere\",DATUM[\"Not specified (based on Clarke 1866 "
  "Authalic Sphere)\",SPHEROID[\"Clarke 1866 Authalic Sphere\",6370997,0,AUTHORITY[\"EPSG\",\"7052\"]],AUTHORITY[\"EPSG\""
  ",\"6052\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4052\"]],PROJECTION[\"Lambert Azimuthal Eq"
  "ual Area (Spherical)\",AUTHORITY[\"EPSG\",\"1027\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\""
  "8801\"]],PARAMETER[\"Longitude of natural origin\",-100,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2163\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2164, 1, 'Locodjo 1965 / TM 5 NW', 'EPSG', 2164, 'PROJCS[\"Locodjo 1965 / TM 5 NW\",GEOGCS[\"L"
  "ocodjo 1965\",DATUM[\"Locodjo 1965\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]]"
  ",TOWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"EPSG\",\"6142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4142\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2164\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2165, 1, 'Abidjan 1987 / TM 5 NW', 'EPSG', 2165, 'PROJCS[\"Abidjan 1987 / TM 5 NW\",GEOGCS[\"A"
  "bidjan 1987\",DATUM[\"Abidjan 1987\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]]"
  ",TOWGS84[-124.76,53,466.79,0,0,0,0],AUTHORITY[\"EPSG\",\"6143\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4143\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2165\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2169, 1, 'Luxembourg 1930 / Gauss', 'EPSG', 2169, 'PROJCS[\"Luxembourg 1930 / Gauss\",GEOGCS[\""
  "Luxembourg 1930\",DATUM[\"Luxembourg 1930\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]"
  ",TOWGS84[-189.6806,18.3463,-42.7695,-0.33746,-3.09264,2.53861,0.4598],AUTHORITY[\"EPSG\",\"6181\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4181\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",49.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",6.17777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1"
  "00000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]"
  ",AUTHORITY[\"EPSG\",\"2169\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2172, 1, 'Pulkovo 1942(58) / Poland zone II', 'EPSG', 2172, 'PROJCS[\"Pulkovo 1942(58) / Polan"
  "d zone II\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORI"
  "TY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\","
  "\"9809\"]],PARAMETER[\"Latitude of natural origin\",53.0019444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",21.5027777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0"
  ".9998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4603000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",5806000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],"
  "AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2172\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2173, 1, 'Pulkovo 1942(58) / Poland zone III', 'EPSG', 2173, 'PROJCS[\"Pulkovo 1942(58) / Pola"
  "nd zone III\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHO"
  "RITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\""
  ",\"9809\"]],PARAMETER[\"Latitude of natural origin\",53.5833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",17.0083333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\","
  "0.9998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3501000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",5999000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2174, 1, 'Pulkovo 1942(58) / Poland zone IV', 'EPSG', 2174, 'PROJCS[\"Pulkovo 1942(58) / Polan"
  "d zone IV\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORI"
  "TY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\","
  "\"9809\"]],PARAMETER[\"Latitude of natural origin\",51.6708333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",16.6722222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0"
  ".9998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3703000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",5627000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],"
  "AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2174\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2175, 1, 'Pulkovo 1942(58) / Poland zone V', 'EPSG', 2175, 'PROJCS[\"Pulkovo 1942(58) / Poland"
  " zone V\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",18.9583333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999983,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",237000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4700"
  "000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],A"
  "UTHORITY[\"EPSG\",\"2175\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2176, 1, 'ETRS89 / Poland CS2000 zone 5', 'EPSG', 2176, 'PROJCS[\"ETRS89 / Poland CS2000 zone "
  "5\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999923,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"2176\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2177, 1, 'ETRS89 / Poland CS2000 zone 6', 'EPSG', 2177, 'PROJCS[\"ETRS89 / Poland CS2000 zone "
  "6\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999923,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"2177\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2178, 1, 'ETRS89 / Poland CS2000 zone 7', 'EPSG', 2178, 'PROJCS[\"ETRS89 / Poland CS2000 zone "
  "7\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999923,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"2178\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2179, 1, 'ETRS89 / Poland CS2000 zone 8', 'EPSG', 2179, 'PROJCS[\"ETRS89 / Poland CS2000 zone "
  "8\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999923,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"2179\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2180, 1, 'ETRS89 / Poland CS92', 'EPSG', 2180, 'PROJCS[\"ETRS89 / Poland CS92\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9993,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"2180\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2188, 1, 'Azores Occidental 1939 / UTM zone 25N', 'EPSG', 2188, 'PROJCS[\"Azores Occidental 19"
  "39 / UTM zone 25N\",GEOGCS[\"Azores Occidental 1939\",DATUM[\"Azores Occidental Islands 1939\",SPHEROID[\"Internation"
  "al 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-422.651,-172.995,84.02,0,0,0,0],AUTHORITY[\"EPSG\",\"618"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4182\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"2188\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2189, 1, 'Azores Central 1948 / UTM zone 26N', 'EPSG', 2189, 'PROJCS[\"Azores Central 1948 / U"
  "TM zone 26N\",GEOGCS[\"Azores Central 1948\",DATUM[\"Azores Central Islands 1948\",SPHEROID[\"International 1924\",63"
  "78388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-104,167,-38,0,0,0,0],AUTHORITY[\"EPSG\",\"6183\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4183\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2189\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2190, 1, 'Azores Oriental 1940 / UTM zone 26N', 'EPSG', 2190, 'PROJCS[\"Azores Oriental 1940 /"
  " UTM zone 26N\",GEOGCS[\"Azores Oriental 1940\",DATUM[\"Azores Oriental Islands 1940\",SPHEROID[\"International 1924\""
  ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-203,141,53,0,0,0,0],AUTHORITY[\"EPSG\",\"6184\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4184\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"219"
  "0\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2193, 1, 'NZGD2000 / New Zealand Transverse Mercator 2000', 'EPSG', 2193, 'PROJCS[\"NZGD2000 /"
  " New Zealand Transverse Mercator 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",173,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",1600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"2193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2195, 1, 'NAD83(HARN) / UTM zone 2S', 'EPSG', 2195, 'PROJCS[\"NAD83(HARN) / UTM zone 2S\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2196, 1, 'ETRS89 / Kp2000 Jutland', 'EPSG', 2196, 'PROJCS[\"ETRS89 / Kp2000 Jutland\",GEOGCS[\""
  "ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9.5,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2196\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2197, 1, 'ETRS89 / Kp2000 Zealand', 'EPSG', 2197, 'PROJCS[\"ETRS89 / Kp2000 Zealand\",GEOGCS[\""
  "ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2197\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2198, 1, 'ETRS89 / Kp2000 Bornholm', 'EPSG', 2198, 'PROJCS[\"ETRS89 / Kp2000 Bornholm\",GEOGCS"
  "[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",900"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2198\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2200, 1, 'ATS77 / New Brunswick Stereographic (ATS77)', 'EPSG', 2200, 'PROJCS[\"ATS77 / New Br"
  "unswick Stereographic (ATS77)\",GEOGCS[\"ATS77\",DATUM[\"Average Terrestrial System 1977\",SPHEROID[\"Average Terrest"
  "rial System 1977\",6378135,298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4122\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"9809\"]],"
  "PARAMETER[\"Latitude of natural origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-66.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999912,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"2200\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2201, 1, 'REGVEN / UTM zone 18N', 'EPSG', 2201, 'PROJCS[\"REGVEN / UTM zone 18N\",GEOGCS[\"REG"
  "VEN\",DATUM[\"Red Geodesica Venezolana\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"418"
  "9\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2201\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2202, 1, 'REGVEN / UTM zone 19N', 'EPSG', 2202, 'PROJCS[\"REGVEN / UTM zone 19N\",GEOGCS[\"REG"
  "VEN\",DATUM[\"Red Geodesica Venezolana\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"418"
  "9\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2202\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2203, 1, 'REGVEN / UTM zone 20N', 'EPSG', 2203, 'PROJCS[\"REGVEN / UTM zone 20N\",GEOGCS[\"REG"
  "VEN\",DATUM[\"Red Geodesica Venezolana\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"418"
  "9\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2203\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2204, 1, 'NAD27 / Tennessee', 'EPSG', 2204, 'PROJCS[\"NAD27 / Tennessee\",GEOGCS[\"NAD27\",DAT"
  "UM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],T"
  "OWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PA"
  "RAMETER[\"Longitude of false origin\",-86,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",35.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.4166666666667,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"N"
  "orthing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2204\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2205, 1, 'NAD83 / Kentucky North', 'EPSG', 2205, 'PROJCS[\"NAD83 / Kentucky North\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.25,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2205\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2206, 1, 'ED50 / 3-degree Gauss-Kruger zone 9', 'EPSG', 2206, 'PROJCS[\"ED50 / 3-degree Gauss-"
  "Kruger zone 9\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2206\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2207, 1, 'ED50 / 3-degree Gauss-Kruger zone 10', 'EPSG', 2207, 'PROJCS[\"ED50 / 3-degree Gauss"
  "-Kruger zone 10\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"220"
  "7\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2208, 1, 'ED50 / 3-degree Gauss-Kruger zone 11', 'EPSG', 2208, 'PROJCS[\"ED50 / 3-degree Gauss"
  "-Kruger zone 11\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"220"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2209, 1, 'ED50 / 3-degree Gauss-Kruger zone 12', 'EPSG', 2209, 'PROJCS[\"ED50 / 3-degree Gauss"
  "-Kruger zone 12\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"220"
  "9\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2210, 1, 'ED50 / 3-degree Gauss-Kruger zone 13', 'EPSG', 2210, 'PROJCS[\"ED50 / 3-degree Gauss"
  "-Kruger zone 13\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"221"
  "0\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2211, 1, 'ED50 / 3-degree Gauss-Kruger zone 14', 'EPSG', 2211, 'PROJCS[\"ED50 / 3-degree Gauss"
  "-Kruger zone 14\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"221"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2212, 1, 'ED50 / 3-degree Gauss-Kruger zone 15', 'EPSG', 2212, 'PROJCS[\"ED50 / 3-degree Gauss"
  "-Kruger zone 15\",GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"221"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2213, 1, 'ETRS89 / TM 30 NE', 'EPSG', 2213, 'PROJCS[\"ETRS89 / TM 30 NE\",GEOGCS[\"ETRS89\",DA"
  "TUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2213\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2215, 1, 'Manoca 1962 / UTM zone 32N', 'EPSG', 2215, 'PROJCS[\"Manoca 1962 / UTM zone 32N\",GE"
  "OGCS[\"Manoca 1962\",DATUM[\"Manoca 1962\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPS"
  "G\",\"7011\"]],TOWGS84[-70.9,-151.8,-41.4,0,0,0,0],AUTHORITY[\"EPSG\",\"6193\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4193\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2215\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2216, 1, 'Qornoq 1927 / UTM zone 22N', 'EPSG', 2216, 'PROJCS[\"Qornoq 1927 / UTM zone 22N\",GE"
  "OGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2216\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2217, 1, 'Qornoq 1927 / UTM zone 23N', 'EPSG', 2217, 'PROJCS[\"Qornoq 1927 / UTM zone 23N\",GE"
  "OGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2217\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2218, 1, 'Scoresbysund 1952 / Greenland zone 5 east', 'EPSG', 2218, 'PROJCS[\"Scoresbysund 195"
  "2 / Greenland zone 5 east\",GEOGCS[\"Scoresbysund 1952\",DATUM[\"Scoresbysund 1952\",SPHEROID[\"International 1924\","
  "6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[105,326,-102.5,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6195\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4195\"]],PROJECTION[\"Lambert Conic Conformal (West Orientat"
  "ed)\",AUTHORITY[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",70.5111111111111,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS["
  "\"X\",WEST],AUTHORITY[\"EPSG\",\"2218\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2219, 1, 'ATS77 / UTM zone 19N', 'EPSG', 2219, 'PROJCS[\"ATS77 / UTM zone 19N\",GEOGCS[\"ATS77"
  "\",DATUM[\"Average Terrestrial System 1977\",SPHEROID[\"Average Terrestrial System 1977\",6378135,298.257,AUTHORITY[\""
  "EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4122\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2219\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2220, 1, 'ATS77 / UTM zone 20N', 'EPSG', 2220, 'PROJCS[\"ATS77 / UTM zone 20N\",GEOGCS[\"ATS77"
  "\",DATUM[\"Average Terrestrial System 1977\",SPHEROID[\"Average Terrestrial System 1977\",6378135,298.257,AUTHORITY[\""
  "EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4122\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2221, 1, 'Scoresbysund 1952 / Greenland zone 6 east', 'EPSG', 2221, 'PROJCS[\"Scoresbysund 195"
  "2 / Greenland zone 6 east\",GEOGCS[\"Scoresbysund 1952\",DATUM[\"Scoresbysund 1952\",SPHEROID[\"International 1924\","
  "6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[105,326,-102.5,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6195\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4195\"]],PROJECTION[\"Lambert Conic Conformal (West Orientat"
  "ed)\",AUTHORITY[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",67.5111111111111,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-32,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS["
  "\"X\",WEST],AUTHORITY[\"EPSG\",\"2221\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2222, 1, 'NAD83 / Arizona East (ft)', 'EPSG', 2222, 'PROJCS[\"NAD83 / Arizona East (ft)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.177777777778,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.30"
  "48,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2222\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2223, 1, 'NAD83 / Arizona Central (ft)', 'EPSG', 2223, 'PROJCS[\"NAD83 / Arizona Central (ft)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111.927777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot"
  "\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2223\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2224, 1, 'NAD83 / Arizona West (ft)', 'EPSG', 2224, 'PROJCS[\"NAD83 / Arizona West (ft)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.75,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",70"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,"
  "AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2224\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2225, 1, 'NAD83 / California zone 1 (ftUS)', 'EPSG', 2225, 'PROJCS[\"NAD83 / California zone 1"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-122,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2225\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2226, 1, 'NAD83 / California zone 2 (ftUS)', 'EPSG', 2226, 'PROJCS[\"NAD83 / California zone 2"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-122,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.8333333333333,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.667,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2226\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2227, 1, 'NAD83 / California zone 3 (ftUS)', 'EPSG', 2227, 'PROJCS[\"NAD83 / California zone 3"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120."
  "511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2227\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2228, 1, 'NAD83 / California zone 4 (ftUS)', 'EPSG', 2228, 'PROJCS[\"NAD83 / California zone 4"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-119,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.2611111111111,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2228\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2229, 1, 'NAD83 / California zone 5 (ftUS)', 'EPSG', 2229, 'PROJCS[\"NAD83 / California zone 5"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-118,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.4666666666667,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.667,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2229\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2230, 1, 'NAD83 / California zone 6 (ftUS)', 'EPSG', 2230, 'PROJCS[\"NAD83 / California zone 6"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-116."
  "25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.8833333333333,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.66"
  "7,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2230\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2231, 1, 'NAD83 / Colorado North (ftUS)', 'EPSG', 2231, 'PROJCS[\"NAD83 / Colorado North (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.7944444444444,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,A"
  "UTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2231\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2232, 1, 'NAD83 / Colorado Central (ftUS)', 'EPSG', 2232, 'PROJCS[\"NAD83 / Colorado Central ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.5"
  "11111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.75,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"2232\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2233, 1, 'NAD83 / Colorado South (ftUS)', 'EPSG', 2233, 'PROJCS[\"NAD83 / Colorado South (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,A"
  "UTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2233\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2234, 1, 'NAD83 / Connecticut (ftUS)', 'EPSG', 2234, 'PROJCS[\"NAD83 / Connecticut (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-72.75,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.8666666666667,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"2234\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2235, 1, 'NAD83 / Delaware (ftUS)', 'EPSG', 2235, 'PROJCS[\"NAD83 / Delaware (ftUS)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.4166666666667,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",65"
  "6166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey "
  "foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2235"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2236, 1, 'NAD83 / Florida East (ftUS)', 'EPSG', 2236, 'PROJCS[\"NAD83 / Florida East (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2236\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2237, 1, 'NAD83 / Florida West (ftUS)', 'EPSG', 2237, 'PROJCS[\"NAD83 / Florida West (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2237\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2238, 1, 'NAD83 / Florida North (ftUS)', 'EPSG', 2238, 'PROJCS[\"NAD83 / Florida North (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.5111111111111,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7611111111111,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"2238\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2239, 1, 'NAD83 / Georgia East (ftUS)', 'EPSG', 2239, 'PROJCS[\"NAD83 / Georgia East (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.1777777777778,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US "
  "survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2239\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2240, 1, 'NAD83 / Georgia West (ftUS)', 'EPSG', 2240, 'PROJCS[\"NAD83 / Georgia West (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.1777777777778,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US"
  " survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2240\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2241, 1, 'NAD83 / Idaho East (ftUS)', 'EPSG', 2241, 'PROJCS[\"NAD83 / Idaho East (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.177777777778,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"2241\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2242, 1, 'NAD83 / Idaho Central (ftUS)', 'EPSG', 2242, 'PROJCS[\"NAD83 / Idaho Central (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"2242\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2243, 1, 'NAD83 / Idaho West (ftUS)', 'EPSG', 2243, 'PROJCS[\"NAD83 / Idaho West (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.75,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",2624666.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2243\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2246, 1, 'NAD83 / Kentucky North (ftUS)', 'EPSG', 2246, 'PROJCS[\"NAD83 / Kentucky North (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.25,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"2246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2247, 1, 'NAD83 / Kentucky South (ftUS)', 'EPSG', 2247, 'PROJCS[\"NAD83 / Kentucky South (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-85.75,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9333333333333,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.667,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2247\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2248, 1, 'NAD83 / Maryland (ftUS)', 'EPSG', 2248, 'PROJCS[\"NAD83 / Maryland (ftUS)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of "
  "2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312333."
  "333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US su"
  "rvey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2248\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2249, 1, 'NAD83 / Massachusetts Mainland (ftUS)', 'EPSG', 2249, 'PROJCS[\"NAD83 / Massachusett"
  "s Mainland (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-71.5111"
  "111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.6944444444444,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",246"
  "0625,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2250, 1, 'NAD83 / Massachusetts Island (ftUS)', 'EPSG', 2250, 'PROJCS[\"NAD83 / Massachusetts "
  "Island (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMET"
  "ER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-70.51111111"
  "11111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.4944444444444,AUTHORITY[\"EPSG"
  "\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETE"
  "R[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2250\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2251, 1, 'NAD83 / Michigan North (ft)', 'EPSG', 2251, 'PROJCS[\"NAD83 / Michigan North (ft)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-87,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",26246719.16,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"22"
  "51\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2252, 1, 'NAD83 / Michigan Central (ft)', 'EPSG', 2252, 'PROJCS[\"NAD83 / Michigan Central (ft"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3777777"
  "777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.7,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",19685039.37,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG"
  "\",\"2252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2253, 1, 'NAD83 / Michigan South (ft)', 'EPSG', 2253, 'PROJCS[\"NAD83 / Michigan South (ft)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.37777777777"
  "78,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.6777777777778,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",13123359.58,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\","
  "\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2254, 1, 'NAD83 / Mississippi East (ftUS)', 'EPSG', 2254, 'PROJCS[\"NAD83 / Mississippi East ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.8333333333333,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"2254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2255, 1, 'NAD83 / Mississippi West (ftUS)', 'EPSG', 2255, 'PROJCS[\"NAD83 / Mississippi West ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.3333333333333,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"2255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2256, 1, 'NAD83 / Montana (ft)', 'EPSG', 2256, 'PROJCS[\"NAD83 / Montana (ft)\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-109.511111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitu"
  "de of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968503.937,AUTH"
  "ORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048"
  ",AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2257, 1, 'NAD83 / New Mexico East (ftUS)', 'EPSG', 2257, 'PROJCS[\"NAD83 / New Mexico East (ft"
  "US)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-104.333333333333,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",541337.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"2257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2258, 1, 'NAD83 / New Mexico Central (ftUS)', 'EPSG', 2258, 'PROJCS[\"NAD83 / New Mexico Centr"
  "al (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-106.25,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2259, 1, 'NAD83 / New Mexico West (ftUS)', 'EPSG', 2259, 'PROJCS[\"NAD83 / New Mexico West (ft"
  "US)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.833333333333,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",2723091.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"2259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2260, 1, 'NAD83 / New York East (ftUS)', 'EPSG', 2260, 'PROJCS[\"NAD83 / New York East (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111111111111,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"2260\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2261, 1, 'NAD83 / New York Central (ftUS)', 'EPSG', 2261, 'PROJCS[\"NAD83 / New York Central ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5944444444444,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",820208.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"2261\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2262, 1, 'NAD83 / New York West (ftUS)', 'EPSG', 2262, 'PROJCS[\"NAD83 / New York West (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-78.5944444444444,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",1148291.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2262\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2263, 1, 'NAD83 / New York Long Island (ftUS)', 'EPSG', 2263, 'PROJCS[\"NAD83 / New York Long "
  "Island (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMET"
  "ER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.0333333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"2263\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2264, 1, 'NAD83 / North Carolina (ftUS)', 'EPSG', 2264, 'PROJCS[\"NAD83 / North Carolina (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2264\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2265, 1, 'NAD83 / North Dakota North (ft)', 'EPSG', 2265, 'PROJCS[\"NAD83 / North Dakota North"
  " (ft)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.511111111111,AU"
  "THORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"882"
  "3\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2265\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2266, 1, 'NAD83 / North Dakota South (ft)', 'EPSG', 2266, 'PROJCS[\"NAD83 / North Dakota South"
  " (ft)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.5"
  "11111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.4944444444444,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"2266\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2267, 1, 'NAD83 / Oklahoma North (ftUS)', 'EPSG', 2267, 'PROJCS[\"NAD83 / Oklahoma North (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2267\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2268, 1, 'NAD83 / Oklahoma South (ftUS)', 'EPSG', 2268, 'PROJCS[\"NAD83 / Oklahoma South (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"2268\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2269, 1, 'NAD83 / Oregon North (ft)', 'EPSG', 2269, 'PROJCS[\"NAD83 / Oregon North (ft)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",8202099.738,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2269"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2270, 1, 'NAD83 / Oregon South (ft)', 'EPSG', 2270, 'PROJCS[\"NAD83 / Oregon South (ft)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",4921259.843,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2270"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2271, 1, 'NAD83 / Pennsylvania North (ftUS)', 'EPSG', 2271, 'PROJCS[\"NAD83 / Pennsylvania Nor"
  "th (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77"
  ".75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.95,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"2271\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2272, 1, 'NAD83 / Pennsylvania South (ftUS)', 'EPSG', 2272, 'PROJCS[\"NAD83 / Pennsylvania Sou"
  "th (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77"
  ".75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.9777777777778,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"2272\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2273, 1, 'NAD83 / South Carolina (ft)', 'EPSG', 2273, 'PROJCS[\"NAD83 / South Carolina (ft)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2273\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2274, 1, 'NAD83 / Tennessee (ftUS)', 'EPSG', 2274, 'PROJCS[\"NAD83 / Tennessee (ftUS)\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-86,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2274\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2275, 1, 'NAD83 / Texas North (ftUS)', 'EPSG', 2275, 'PROJCS[\"NAD83 / Texas North (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-101.511111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280833.333,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"2275\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2276, 1, 'NAD83 / Texas North Central (ftUS)', 'EPSG', 2276, 'PROJCS[\"NAD83 / Texas North Cen"
  "tral (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER"
  "[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\","
  "-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.9777777777778,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.1444444444444,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2276\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2277, 1, 'NAD83 / Texas Central (ftUS)', 'EPSG', 2277, 'PROJCS[\"NAD83 / Texas Central (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.33333333"
  "3333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",31.8833333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",2296583.333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",9842500,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2277\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2278, 1, 'NAD83 / Texas South Central (ftUS)', 'EPSG', 2278, 'PROJCS[\"NAD83 / Texas South Cen"
  "tral (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER"
  "[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\","
  "-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.2833333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",13123333.333"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2278\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2279, 1, 'NAD83 / Texas South (ftUS)', 'EPSG', 2279, 'PROJCS[\"NAD83 / Texas South (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.511111111111"
  "1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",16404166.667,AUT"
  "HORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2279\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2280, 1, 'NAD83 / Utah North (ft)', 'EPSG', 2280, 'PROJCS[\"NAD83 / Utah North (ft)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280839.895,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"2280\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2281, 1, 'NAD83 / Utah Central (ft)', 'EPSG', 2281, 'PROJCS[\"NAD83 / Utah Central (ft)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fa"
  "lse origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6561679.79,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2281\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2282, 1, 'NAD83 / Utah South (ft)', 'EPSG', 2282, 'PROJCS[\"NAD83 / Utah South (ft)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER"
  "[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false or"
  "igin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",9842519.685,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2282\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2283, 1, 'NAD83 / Virginia North (ftUS)', 'EPSG', 2283, 'PROJCS[\"NAD83 / Virginia North (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.2111111111111,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6561"
  "666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2283\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2284, 1, 'NAD83 / Virginia South (ftUS)', 'EPSG', 2284, 'PROJCS[\"NAD83 / Virginia South (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280"
  "833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2284\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2285, 1, 'NAD83 / Washington North (ftUS)', 'EPSG', 2285, 'PROJCS[\"NAD83 / Washington North ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.833333333333,AU"
  "THORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"882"
  "3\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"2285\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2286, 1, 'NAD83 / Washington South (ftUS)', 'EPSG', 2286, 'PROJCS[\"NAD83 / Washington South ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.5"
  "11111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.3333333333333,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2286\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2287, 1, 'NAD83 / Wisconsin North (ftUS)', 'EPSG', 2287, 'PROJCS[\"NAD83 / Wisconsin North (ft"
  "US)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latit"
  "ude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.7666666666667,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\","
  "\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"2287\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2288, 1, 'NAD83 / Wisconsin Central (ftUS)', 'EPSG', 2288, 'PROJCS[\"NAD83 / Wisconsin Central"
  " (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"2288\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2289, 1, 'NAD83 / Wisconsin South (ftUS)', 'EPSG', 2289, 'PROJCS[\"NAD83 / Wisconsin South (ft"
  "US)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latit"
  "ude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2289\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2290, 1, 'ATS77 / Prince Edward Isl. Stereographic (ATS77)', 'EPSG', 2290, 'PROJCS[\"ATS77 / P"
  "rince Edward Isl. Stereographic (ATS77)\",GEOGCS[\"ATS77\",DATUM[\"Average Terrestrial System 1977\",SPHEROID[\"Avera"
  "ge Terrestrial System 1977\",6378135,298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4122\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\""
  "9809\"]],PARAMETER[\"Latitude of natural origin\",47.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999912,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "400000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2290\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2294, 1, 'ATS77 / MTM Nova Scotia zone 4', 'EPSG', 2294, 'PROJCS[\"ATS77 / MTM Nova Scotia zon"
  "e 4\",GEOGCS[\"ATS77\",DATUM[\"Average Terrestrial System 1977\",SPHEROID[\"Average Terrestrial System 1977\",6378135"
  ",298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4122\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-61.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2294\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2295, 1, 'ATS77 / MTM Nova Scotia zone 5', 'EPSG', 2295, 'PROJCS[\"ATS77 / MTM Nova Scotia zon"
  "e 5\",GEOGCS[\"ATS77\",DATUM[\"Average Terrestrial System 1977\",SPHEROID[\"Average Terrestrial System 1977\",6378135"
  ",298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4122\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-64.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2295\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2296, 1, 'Ammassalik 1958 / Greenland zone 7 east', 'EPSG', 2296, 'PROJCS[\"Ammassalik 1958 / "
  "Greenland zone 7 east\",GEOGCS[\"Ammassalik 1958\",DATUM[\"Ammassalik 1958\",SPHEROID[\"International 1924\",6378388,"
  "297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-45,417,-3.5,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6196\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4196\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTH"
  "ORITY[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",64.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-40,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST"
  "],AUTHORITY[\"EPSG\",\"2296\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2299, 1, 'Qornoq 1927 / Greenland zone 2 west', 'EPSG', 2299, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 2 west\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",79.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-64,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2299\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2301, 1, 'Qornoq 1927 / Greenland zone 3 west', 'EPSG', 2301, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 3 west\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",76.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-64,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2301\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2303, 1, 'Qornoq 1927 / Greenland zone 4 west', 'EPSG', 2303, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 4 west\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",73.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-52,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2303\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2304, 1, 'Qornoq 1927 / Greenland zone 5 west', 'EPSG', 2304, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 5 west\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",70.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-52,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2304\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2305, 1, 'Qornoq 1927 / Greenland zone 6 west', 'EPSG', 2305, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 6 west\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",67.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-52,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2305\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2306, 1, 'Qornoq 1927 / Greenland zone 7 west', 'EPSG', 2306, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 7 west\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",64.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-52,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2306\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2307, 1, 'Qornoq 1927 / Greenland zone 8 east', 'EPSG', 2307, 'PROJCS[\"Qornoq 1927 / Greenlan"
  "d zone 8 east\",GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHORITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORIT"
  "Y[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",61.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-48,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AU"
  "THORITY[\"EPSG\",\"2307\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2308, 1, 'Batavia / TM 109 SE', 'EPSG', 2308, 'PROJCS[\"Batavia / TM 109 SE\",GEOGCS[\"Batavia"
  "\",DATUM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-377.7,67"
  "5.1,-52.2,0,0,0,0],AUTHORITY[\"EPSG\",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4211\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",109,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2308\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2309, 1, 'WGS 84 / TM 116 SE', 'EPSG', 2309, 'PROJCS[\"WGS 84 / TM 116 SE\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",116,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2309\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2310, 1, 'WGS 84 / TM 132 SE', 'EPSG', 2310, 'PROJCS[\"WGS 84 / TM 132 SE\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2310\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2311, 1, 'WGS 84 / TM 6 NE', 'EPSG', 2311, 'PROJCS[\"WGS 84 / TM 6 NE\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",6,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "96,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\","
  "NORTH],AUTHORITY[\"EPSG\",\"2311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2312, 1, 'Garoua / UTM zone 33N', 'EPSG', 2312, 'PROJCS[\"Garoua / UTM zone 33N\",GEOGCS[\"Gar"
  "oua\",DATUM[\"Garoua\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"E"
  "PSG\",\"6197\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4197\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "96,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\","
  "NORTH],AUTHORITY[\"EPSG\",\"2312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2313, 1, 'Kousseri / UTM zone 33N', 'EPSG', 2313, 'PROJCS[\"Kousseri / UTM zone 33N\",GEOGCS[\""
  "Kousseri\",DATUM[\"Kousseri\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHOR"
  "ITY[\"EPSG\",\"6198\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4198\"]],PROJECTION[\"Transver"
  "se Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS"
  "[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2313\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2314, 1, 'Trinidad 1903 / Trinidad Grid (ftCla)', 'EPSG', 2314, 'PROJCS[\"Trinidad 1903 / Trin"
  "idad Grid (ftCla)\",GEOGCS[\"Trinidad 1903\",DATUM[\"Trinidad 1903\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.2"
  "6067636926103,AUTHORITY[\"EPSG\",\"7007\"]],TOWGS84[-61.702,284.488,472.052,0,0,0,0],AUTHORITY[\"EPSG\",\"6302\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4302\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG"
  "\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",10.4416666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-61.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",283800,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",214500,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Clarke\\'s foot\",0.30"
  "47972654,AUTHORITY[\"EPSG\",\"9005\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2314\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2315, 1, 'Campo Inchauspe / UTM zone 19S', 'EPSG', 2315, 'PROJCS[\"Campo Inchauspe / UTM zone "
  "19S\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2315\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2316, 1, 'Campo Inchauspe / UTM zone 20S', 'EPSG', 2316, 'PROJCS[\"Campo Inchauspe / UTM zone "
  "20S\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2317, 1, 'PSAD56 / ICN Regional', 'EPSG', 2317, 'PROJCS[\"PSAD56 / ICN Regional\",GEOGCS[\"PSA"
  "D56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",6,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-66,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",9,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",3,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2317\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2318, 1, 'Ain el Abd / Aramco Lambert', 'EPSG', 2318, 'PROJCS[\"Ain el Abd / Aramco Lambert\","
  "GEOGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022"
  "\"]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",25.08951,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",48,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",17,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",33,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2319, 1, 'ED50 / TM27', 'EPSG', 2319, 'PROJCS[\"ED50 / TM27\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2319\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2320, 1, 'ED50 / TM30', 'EPSG', 2320, 'PROJCS[\"ED50 / TM30\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2320\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2321, 1, 'ED50 / TM33', 'EPSG', 2321, 'PROJCS[\"ED50 / TM33\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2321\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2322, 1, 'ED50 / TM36', 'EPSG', 2322, 'PROJCS[\"ED50 / TM36\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2323, 1, 'ED50 / TM39', 'EPSG', 2323, 'PROJCS[\"ED50 / TM39\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2323\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2324, 1, 'ED50 / TM42', 'EPSG', 2324, 'PROJCS[\"ED50 / TM42\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2325, 1, 'ED50 / TM45', 'EPSG', 2325, 'PROJCS[\"ED50 / TM45\",GEOGCS[\"ED50\",DATUM[\"European"
  " Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41"
  ",2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2325\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2326, 1, 'Hong Kong 1980 Grid System', 'EPSG', 2326, 'PROJCS[\"Hong Kong 1980 Grid System\",GE"
  "OGCS[\"Hong Kong 1980\",DATUM[\"Hong Kong 1980\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"702"
  "2\"]],TOWGS84[-162.619,-276.959,-161.764,0.067753,-2.243649,-1.158827,-1.094246],AUTHORITY[\"EPSG\",\"6611\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4611\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",22.3121333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",114.178555555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",836694.05,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",819069.8,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],"
  "AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2326\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2327, 1, 'Xian 1980 / Gauss-Kruger zone 13', 'EPSG', 2327, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 13\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH"
  "],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2327\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2328, 1, 'Xian 1980 / Gauss-Kruger zone 14', 'EPSG', 2328, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 14\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH"
  "],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2328\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2329, 1, 'Xian 1980 / Gauss-Kruger zone 15', 'EPSG', 2329, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 15\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH"
  "],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2329\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2330, 1, 'Xian 1980 / Gauss-Kruger zone 16', 'EPSG', 2330, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 16\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH"
  "],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2331, 1, 'Xian 1980 / Gauss-Kruger zone 17', 'EPSG', 2331, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 17\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH"
  "],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2332, 1, 'Xian 1980 / Gauss-Kruger zone 18', 'EPSG', 2332, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 18\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORT"
  "H],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2333, 1, 'Xian 1980 / Gauss-Kruger zone 19', 'EPSG', 2333, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 19\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORT"
  "H],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2334, 1, 'Xian 1980 / Gauss-Kruger zone 20', 'EPSG', 2334, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 20\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORT"
  "H],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2334\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2335, 1, 'Xian 1980 / Gauss-Kruger zone 21', 'EPSG', 2335, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 21\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORT"
  "H],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2335\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2336, 1, 'Xian 1980 / Gauss-Kruger zone 22', 'EPSG', 2336, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 22\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORT"
  "H],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2336\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2337, 1, 'Xian 1980 / Gauss-Kruger zone 23', 'EPSG', 2337, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "zone 23\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORT"
  "H],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2338, 1, 'Xian 1980 / Gauss-Kruger CM 75E', 'EPSG', 2338, 'PROJCS[\"Xian 1980 / Gauss-Kruger C"
  "M 75E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],"
  "AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXI"
  "S[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2338\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2339, 1, 'Xian 1980 / Gauss-Kruger CM 81E', 'EPSG', 2339, 'PROJCS[\"Xian 1980 / Gauss-Kruger C"
  "M 81E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],"
  "AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXI"
  "S[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2340, 1, 'Xian 1980 / Gauss-Kruger CM 87E', 'EPSG', 2340, 'PROJCS[\"Xian 1980 / Gauss-Kruger C"
  "M 87E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],"
  "AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXI"
  "S[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2341, 1, 'Xian 1980 / Gauss-Kruger CM 93E', 'EPSG', 2341, 'PROJCS[\"Xian 1980 / Gauss-Kruger C"
  "M 93E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],"
  "AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXI"
  "S[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2341\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2342, 1, 'Xian 1980 / Gauss-Kruger CM 99E', 'EPSG', 2342, 'PROJCS[\"Xian 1980 / Gauss-Kruger C"
  "M 99E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],"
  "AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXI"
  "S[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2342\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2343, 1, 'Xian 1980 / Gauss-Kruger CM 105E', 'EPSG', 2343, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "CM 105E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2344, 1, 'Xian 1980 / Gauss-Kruger CM 111E', 'EPSG', 2344, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "CM 111E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2345, 1, 'Xian 1980 / Gauss-Kruger CM 117E', 'EPSG', 2345, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "CM 117E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2345\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2346, 1, 'Xian 1980 / Gauss-Kruger CM 123E', 'EPSG', 2346, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "CM 123E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2347, 1, 'Xian 1980 / Gauss-Kruger CM 129E', 'EPSG', 2347, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "CM 129E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2347\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2348, 1, 'Xian 1980 / Gauss-Kruger CM 135E', 'EPSG', 2348, 'PROJCS[\"Xian 1980 / Gauss-Kruger "
  "CM 135E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]"
  "],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH]"
  ",AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2349, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 25', 'EPSG', 2349, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 25\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",25500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2349\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2350, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 26', 'EPSG', 2350, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 26\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",26500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2351, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 27', 'EPSG', 2351, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 27\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",27500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2352, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 28', 'EPSG', 2352, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 28\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",28500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2353, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 29', 'EPSG', 2353, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 29\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",29500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2354, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 30', 'EPSG', 2354, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 30\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2355, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 31', 'EPSG', 2355, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 31\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",31500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2356, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 32', 'EPSG', 2356, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 32\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",32500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2357, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 33', 'EPSG', 2357, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 33\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",33500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2357\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2358, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 34', 'EPSG', 2358, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 34\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",34500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2358\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2359, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 35', 'EPSG', 2359, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 35\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",35500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2359\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2360, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 36', 'EPSG', 2360, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 36\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",36500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2360\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2361, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 37', 'EPSG', 2361, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 37\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",37500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2361\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2362, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 38', 'EPSG', 2362, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 38\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",38500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2362\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2363, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 39', 'EPSG', 2363, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 39\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",39500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2363\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2364, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 40', 'EPSG', 2364, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 40\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2364\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2365, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 41', 'EPSG', 2365, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 41\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",41500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2365\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2366, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 42', 'EPSG', 2366, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 42\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",42500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2366\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2367, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 43', 'EPSG', 2367, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 43\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",43500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2367\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2368, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 44', 'EPSG', 2368, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 44\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",44500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2368\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2369, 1, 'Xian 1980 / 3-degree Gauss-Kruger zone 45', 'EPSG', 2369, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger zone 45\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",45500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2369\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2370, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 75E', 'EPSG', 2370, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 75E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2370\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2371, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 78E', 'EPSG', 2371, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 78E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2371\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2372, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 81E', 'EPSG', 2372, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 81E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2372\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2373, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 84E', 'EPSG', 2373, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 84E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2373\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2374, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 87E', 'EPSG', 2374, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 87E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2374\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2375, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 90E', 'EPSG', 2375, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 90E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2375\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2376, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 93E', 'EPSG', 2376, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 93E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2376\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2377, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 96E', 'EPSG', 2377, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 96E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2377\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2378, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 99E', 'EPSG', 2378, 'PROJCS[\"Xian 1980 / 3-deg"
  "ree Gauss-Kruger CM 99E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY[\""
  "EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2378\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2379, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 102E', 'EPSG', 2379, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 102E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2379\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2380, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 105E', 'EPSG', 2380, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 105E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2380\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2381, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 108E', 'EPSG', 2381, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 108E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2381\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2382, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 111E', 'EPSG', 2382, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 111E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2382\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2383, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 114E', 'EPSG', 2383, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 114E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2383\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2384, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 117E', 'EPSG', 2384, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 117E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2384\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2385, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 120E', 'EPSG', 2385, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 120E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2385\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2386, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 123E', 'EPSG', 2386, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 123E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2386\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2387, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 126E', 'EPSG', 2387, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 126E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2387\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2388, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 129E', 'EPSG', 2388, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 129E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2388\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2389, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 132E', 'EPSG', 2389, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 132E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2389\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2390, 1, 'Xian 1980 / 3-degree Gauss-Kruger CM 135E', 'EPSG', 2390, 'PROJCS[\"Xian 1980 / 3-de"
  "gree Gauss-Kruger CM 135E\",GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1975\",6378140,298.257,AUTHORITY["
  "\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4610\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2390\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2391, 1, 'KKJ / Finland zone 1', 'EPSG', 2391, 'PROJCS[\"KKJ / Finland zone 1\",GEOGCS[\"KKJ\""
  ",DATUM[\"Kartastokoordinaattijarjestelma (1966)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.345,-1.376,1.496],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2391\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2392, 1, 'KKJ / Finland zone 2', 'EPSG', 2392, 'PROJCS[\"KKJ / Finland zone 2\",GEOGCS[\"KKJ\""
  ",DATUM[\"Kartastokoordinaattijarjestelma (1966)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.345,-1.376,1.496],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2392\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2393, 1, 'KKJ / Finland Uniform Coordinate System', 'EPSG', 2393, 'PROJCS[\"KKJ / Finland Unif"
  "orm Coordinate System\",GEOGCS[\"KKJ\",DATUM[\"Kartastokoordinaattijarjestelma (1966)\",SPHEROID[\"International 1924"
  "\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.345,-1.376,1.496],AUTHORITY[\"E"
  "PSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"2393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2394, 1, 'KKJ / Finland zone 4', 'EPSG', 2394, 'PROJCS[\"KKJ / Finland zone 4\",GEOGCS[\"KKJ\""
  ",DATUM[\"Kartastokoordinaattijarjestelma (1966)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.345,-1.376,1.496],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2394\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2395, 1, 'South Yemen / Gauss-Kruger zone 8', 'EPSG', 2395, 'PROJCS[\"South Yemen / Gauss-Krug"
  "er zone 8\",GEOGCS[\"South Yemen\",DATUM[\"South Yemen\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],TOWGS84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"6164\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4164\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",8500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2396, 1, 'South Yemen / Gauss-Kruger zone 9', 'EPSG', 2396, 'PROJCS[\"South Yemen / Gauss-Krug"
  "er zone 9\",GEOGCS[\"South Yemen\",DATUM[\"South Yemen\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],TOWGS84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"6164\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4164\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",9500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2396\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2397, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 3', 'EPSG', 2397, 'PROJCS[\"Pulkovo 19"
  "42(83) / 3-degree Gauss-Kruger zone 3\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"617"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY"
  "[\"EPSG\",\"2397\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2398, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 4', 'EPSG', 2398, 'PROJCS[\"Pulkovo 19"
  "42(83) / 3-degree Gauss-Kruger zone 4\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"617"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"2398\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2399, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 5', 'EPSG', 2399, 'PROJCS[\"Pulkovo 19"
  "42(83) / 3-degree Gauss-Kruger zone 5\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"617"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"2399\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2401, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 25', 'EPSG', 2401, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 25\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2402, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 26', 'EPSG', 2402, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 26\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",26500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2403, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 27', 'EPSG', 2403, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 27\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",27500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2404, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 28', 'EPSG', 2404, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 28\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",28500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "4\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2405, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 29', 'EPSG', 2405, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 29\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",29500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "5\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2406, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 30', 'EPSG', 2406, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 30\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",30500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "6\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2407, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 31', 'EPSG', 2407, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 31\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "7\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2408, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 32', 'EPSG', 2408, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 32\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2409, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 33', 'EPSG', 2409, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 33\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",33500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"240"
  "9\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2410, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 34', 'EPSG', 2410, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 34\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",34500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "10\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2411, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 35', 'EPSG', 2411, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 35\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",35500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "11\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2412, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 36', 'EPSG', 2412, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 36\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",36500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "12\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2413, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 37', 'EPSG', 2413, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 37\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",37500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "13\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2414, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 38', 'EPSG', 2414, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 38\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",38500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "14\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2415, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 39', 'EPSG', 2415, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 39\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",39500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "15\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2416, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 40', 'EPSG', 2416, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 40\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",40500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "16\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2417, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 41', 'EPSG', 2417, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 41\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",41500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "17\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2418, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 42', 'EPSG', 2418, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 42\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",42500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "18\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2419, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 43', 'EPSG', 2419, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 43\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",43500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "19\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2420, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 44', 'EPSG', 2420, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 44\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",44500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "20\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2421, 1, 'Beijing 1954 / 3-degree Gauss-Kruger zone 45', 'EPSG', 2421, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger zone 45\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",45500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "21\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2422, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 75E', 'EPSG', 2422, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 75E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2422\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2423, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 78E', 'EPSG', 2423, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 78E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2423\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2424, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 81E', 'EPSG', 2424, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 81E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2424\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2425, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 84E', 'EPSG', 2425, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 84E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2425\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2426, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 87E', 'EPSG', 2426, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 87E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2426\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2427, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 90E', 'EPSG', 2427, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 90E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2427\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2428, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 93E', 'EPSG', 2428, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 93E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2428\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2429, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 96E', 'EPSG', 2429, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 96E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2429\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2430, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 99E', 'EPSG', 2430, 'PROJCS[\"Beijing 1954 /"
  " 3-degree Gauss-Kruger CM 99E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2430\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2431, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 102E', 'EPSG', 2431, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 102E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2431\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2432, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 105E', 'EPSG', 2432, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 105E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2432\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2433, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 108E', 'EPSG', 2433, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 108E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2433\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2434, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 111E', 'EPSG', 2434, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 111E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2434\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2435, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 114E', 'EPSG', 2435, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 114E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2435\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2436, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 117E', 'EPSG', 2436, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 117E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2436\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2437, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 120E', 'EPSG', 2437, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 120E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2437\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2438, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 123E', 'EPSG', 2438, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 123E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2438\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2439, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 126E', 'EPSG', 2439, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 126E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2439\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2440, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 129E', 'EPSG', 2440, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 129E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2440\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2441, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 132E', 'EPSG', 2441, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 132E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2441\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2442, 1, 'Beijing 1954 / 3-degree Gauss-Kruger CM 135E', 'EPSG', 2442, 'PROJCS[\"Beijing 1954 "
  "/ 3-degree Gauss-Kruger CM 135E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2442\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2443, 1, 'JGD2000 / Japan Plane Rectangular CS I', 'EPSG', 2443, 'PROJCS[\"JGD2000 / Japan Pla"
  "ne Rectangular CS I\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129.5,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2443\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2444, 1, 'JGD2000 / Japan Plane Rectangular CS II', 'EPSG', 2444, 'PROJCS[\"JGD2000 / Japan Pl"
  "ane Rectangular CS II\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2444\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2445, 1, 'JGD2000 / Japan Plane Rectangular CS III', 'EPSG', 2445, 'PROJCS[\"JGD2000 / Japan P"
  "lane Rectangular CS III\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132."
  "177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "45\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2446, 1, 'JGD2000 / Japan Plane Rectangular CS IV', 'EPSG', 2446, 'PROJCS[\"JGD2000 / Japan Pl"
  "ane Rectangular CS IV\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",133.5,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2446\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2447, 1, 'JGD2000 / Japan Plane Rectangular CS V', 'EPSG', 2447, 'PROJCS[\"JGD2000 / Japan Pla"
  "ne Rectangular CS V\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",134.3444"
  "44444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2447\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2448, 1, 'JGD2000 / Japan Plane Rectangular CS VI', 'EPSG', 2448, 'PROJCS[\"JGD2000 / Japan Pl"
  "ane Rectangular CS VI\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",136,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2448\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2449, 1, 'JGD2000 / Japan Plane Rectangular CS VII', 'EPSG', 2449, 'PROJCS[\"JGD2000 / Japan P"
  "lane Rectangular CS VII\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",137."
  "177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"24"
  "49\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2450, 1, 'JGD2000 / Japan Plane Rectangular CS VIII', 'EPSG', 2450, 'PROJCS[\"JGD2000 / Japan "
  "Plane Rectangular CS VIII\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13"
  "8.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2450\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2451, 1, 'JGD2000 / Japan Plane Rectangular CS IX', 'EPSG', 2451, 'PROJCS[\"JGD2000 / Japan Pl"
  "ane Rectangular CS IX\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",139.83"
  "3333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2451\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2452, 1, 'JGD2000 / Japan Plane Rectangular CS X', 'EPSG', 2452, 'PROJCS[\"JGD2000 / Japan Pla"
  "ne Rectangular CS X\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",140.8333"
  "33333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2452\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2453, 1, 'JGD2000 / Japan Plane Rectangular CS XI', 'EPSG', 2453, 'PROJCS[\"JGD2000 / Japan Pl"
  "ane Rectangular CS XI\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",140.25"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2453\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2454, 1, 'JGD2000 / Japan Plane Rectangular CS XII', 'EPSG', 2454, 'PROJCS[\"JGD2000 / Japan P"
  "lane Rectangular CS XII\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",142."
  "25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2454\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2455, 1, 'JGD2000 / Japan Plane Rectangular CS XIII', 'EPSG', 2455, 'PROJCS[\"JGD2000 / Japan "
  "Plane Rectangular CS XIII\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",14"
  "4.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2455\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2456, 1, 'JGD2000 / Japan Plane Rectangular CS XIV', 'EPSG', 2456, 'PROJCS[\"JGD2000 / Japan P"
  "lane Rectangular CS XIV\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",142,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2456\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2457, 1, 'JGD2000 / Japan Plane Rectangular CS XV', 'EPSG', 2457, 'PROJCS[\"JGD2000 / Japan Pl"
  "ane Rectangular CS XV\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127.51"
  "1111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2457\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2458, 1, 'JGD2000 / Japan Plane Rectangular CS XVI', 'EPSG', 2458, 'PROJCS[\"JGD2000 / Japan P"
  "lane Rectangular CS XVI\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",124,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2458\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2459, 1, 'JGD2000 / Japan Plane Rectangular CS XVII', 'EPSG', 2459, 'PROJCS[\"JGD2000 / Japan "
  "Plane Rectangular CS XVII\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2459\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2460, 1, 'JGD2000 / Japan Plane Rectangular CS XVIII', 'EPSG', 2460, 'PROJCS[\"JGD2000 / Japan"
  " Plane Rectangular CS XVIII\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",20,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",1"
  "36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2460\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2461, 1, 'JGD2000 / Japan Plane Rectangular CS XIX', 'EPSG', 2461, 'PROJCS[\"JGD2000 / Japan P"
  "lane Rectangular CS XIX\",GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",154,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2461\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2462, 1, 'Albanian 1987 / Gauss-Kruger zone 4', 'EPSG', 2462, 'PROJCS[\"Albanian 1987 / Gauss-"
  "Kruger zone 4\",GEOGCS[\"Albanian 1987\",DATUM[\"Albanian 1987\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[-44.183,-0.58,-38.489,2.3867,2.7072,-3.5196,-8.2703],AUTHORITY[\"EPSG\",\"6191\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4191\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2462\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2463, 1, 'Pulkovo 1995 / Gauss-Kruger CM 21E', 'EPSG', 2463, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 21E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2463\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2464, 1, 'Pulkovo 1995 / Gauss-Kruger CM 27E', 'EPSG', 2464, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 27E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2464\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2465, 1, 'Pulkovo 1995 / Gauss-Kruger CM 33E', 'EPSG', 2465, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 33E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2465\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2466, 1, 'Pulkovo 1995 / Gauss-Kruger CM 39E', 'EPSG', 2466, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 39E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2466\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2467, 1, 'Pulkovo 1995 / Gauss-Kruger CM 45E', 'EPSG', 2467, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 45E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2467\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2468, 1, 'Pulkovo 1995 / Gauss-Kruger CM 51E', 'EPSG', 2468, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 51E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2468\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2469, 1, 'Pulkovo 1995 / Gauss-Kruger CM 57E', 'EPSG', 2469, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 57E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2469\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2470, 1, 'Pulkovo 1995 / Gauss-Kruger CM 63E', 'EPSG', 2470, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 63E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2470\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2471, 1, 'Pulkovo 1995 / Gauss-Kruger CM 69E', 'EPSG', 2471, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 69E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2471\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2472, 1, 'Pulkovo 1995 / Gauss-Kruger CM 75E', 'EPSG', 2472, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 75E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2472\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2473, 1, 'Pulkovo 1995 / Gauss-Kruger CM 81E', 'EPSG', 2473, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 81E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2473\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2474, 1, 'Pulkovo 1995 / Gauss-Kruger CM 87E', 'EPSG', 2474, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 87E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2474\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2475, 1, 'Pulkovo 1995 / Gauss-Kruger CM 93E', 'EPSG', 2475, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 93E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2475\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2476, 1, 'Pulkovo 1995 / Gauss-Kruger CM 99E', 'EPSG', 2476, 'PROJCS[\"Pulkovo 1995 / Gauss-Kr"
  "uger CM 99E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2476\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2477, 1, 'Pulkovo 1995 / Gauss-Kruger CM 105E', 'EPSG', 2477, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 105E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2477\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2478, 1, 'Pulkovo 1995 / Gauss-Kruger CM 111E', 'EPSG', 2478, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 111E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2478\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2479, 1, 'Pulkovo 1995 / Gauss-Kruger CM 117E', 'EPSG', 2479, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 117E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2479\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2480, 1, 'Pulkovo 1995 / Gauss-Kruger CM 123E', 'EPSG', 2480, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 123E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2480\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2481, 1, 'Pulkovo 1995 / Gauss-Kruger CM 129E', 'EPSG', 2481, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 129E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2481\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2482, 1, 'Pulkovo 1995 / Gauss-Kruger CM 135E', 'EPSG', 2482, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 135E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2482\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2483, 1, 'Pulkovo 1995 / Gauss-Kruger CM 141E', 'EPSG', 2483, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 141E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2483\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2484, 1, 'Pulkovo 1995 / Gauss-Kruger CM 147E', 'EPSG', 2484, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 147E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2484\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2485, 1, 'Pulkovo 1995 / Gauss-Kruger CM 153E', 'EPSG', 2485, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 153E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2485\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2486, 1, 'Pulkovo 1995 / Gauss-Kruger CM 159E', 'EPSG', 2486, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 159E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2486\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2487, 1, 'Pulkovo 1995 / Gauss-Kruger CM 165E', 'EPSG', 2487, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 165E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2487\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2488, 1, 'Pulkovo 1995 / Gauss-Kruger CM 171E', 'EPSG', 2488, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 171E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2488\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2489, 1, 'Pulkovo 1995 / Gauss-Kruger CM 177E', 'EPSG', 2489, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 177E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2489\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2490, 1, 'Pulkovo 1995 / Gauss-Kruger CM 177W', 'EPSG', 2490, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 177W\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2490\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2491, 1, 'Pulkovo 1995 / Gauss-Kruger CM 171W', 'EPSG', 2491, 'PROJCS[\"Pulkovo 1995 / Gauss-K"
  "ruger CM 171W\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2491\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2494, 1, 'Pulkovo 1942 / Gauss-Kruger CM 21E', 'EPSG', 2494, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 21E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2494\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2495, 1, 'Pulkovo 1942 / Gauss-Kruger CM 27E', 'EPSG', 2495, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 27E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2495\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2496, 1, 'Pulkovo 1942 / Gauss-Kruger CM 33E', 'EPSG', 2496, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 33E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2496\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2497, 1, 'Pulkovo 1942 / Gauss-Kruger CM 39E', 'EPSG', 2497, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 39E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2497\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2498, 1, 'Pulkovo 1942 / Gauss-Kruger CM 45E', 'EPSG', 2498, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 45E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2498\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2499, 1, 'Pulkovo 1942 / Gauss-Kruger CM 51E', 'EPSG', 2499, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 51E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2499\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2500, 1, 'Pulkovo 1942 / Gauss-Kruger CM 57E', 'EPSG', 2500, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 57E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2500\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2501, 1, 'Pulkovo 1942 / Gauss-Kruger CM 63E', 'EPSG', 2501, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 63E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2501\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2502, 1, 'Pulkovo 1942 / Gauss-Kruger CM 69E', 'EPSG', 2502, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 69E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2502\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2503, 1, 'Pulkovo 1942 / Gauss-Kruger CM 75E', 'EPSG', 2503, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 75E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2503\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2504, 1, 'Pulkovo 1942 / Gauss-Kruger CM 81E', 'EPSG', 2504, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 81E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2504\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2505, 1, 'Pulkovo 1942 / Gauss-Kruger CM 87E', 'EPSG', 2505, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 87E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2505\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2506, 1, 'Pulkovo 1942 / Gauss-Kruger CM 93E', 'EPSG', 2506, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 93E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2506\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2507, 1, 'Pulkovo 1942 / Gauss-Kruger CM 99E', 'EPSG', 2507, 'PROJCS[\"Pulkovo 1942 / Gauss-Kr"
  "uger CM 99E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"E"
  "PSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2507\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2508, 1, 'Pulkovo 1942 / Gauss-Kruger CM 105E', 'EPSG', 2508, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 105E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2508\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2509, 1, 'Pulkovo 1942 / Gauss-Kruger CM 111E', 'EPSG', 2509, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 111E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2509\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2510, 1, 'Pulkovo 1942 / Gauss-Kruger CM 117E', 'EPSG', 2510, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 117E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2510\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2511, 1, 'Pulkovo 1942 / Gauss-Kruger CM 123E', 'EPSG', 2511, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 123E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2511\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2512, 1, 'Pulkovo 1942 / Gauss-Kruger CM 129E', 'EPSG', 2512, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 129E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2512\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2513, 1, 'Pulkovo 1942 / Gauss-Kruger CM 135E', 'EPSG', 2513, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 135E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2513\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2514, 1, 'Pulkovo 1942 / Gauss-Kruger CM 141E', 'EPSG', 2514, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 141E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2514\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2515, 1, 'Pulkovo 1942 / Gauss-Kruger CM 147E', 'EPSG', 2515, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 147E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2515\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2516, 1, 'Pulkovo 1942 / Gauss-Kruger CM 153E', 'EPSG', 2516, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 153E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2516\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2517, 1, 'Pulkovo 1942 / Gauss-Kruger CM 159E', 'EPSG', 2517, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 159E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2517\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2518, 1, 'Pulkovo 1942 / Gauss-Kruger CM 165E', 'EPSG', 2518, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 165E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2518\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2519, 1, 'Pulkovo 1942 / Gauss-Kruger CM 171E', 'EPSG', 2519, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 171E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2520, 1, 'Pulkovo 1942 / Gauss-Kruger CM 177E', 'EPSG', 2520, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 177E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2520\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2521, 1, 'Pulkovo 1942 / Gauss-Kruger CM 177W', 'EPSG', 2521, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 177W\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2522, 1, 'Pulkovo 1942 / Gauss-Kruger CM 171W', 'EPSG', 2522, 'PROJCS[\"Pulkovo 1942 / Gauss-K"
  "ruger CM 171W\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2523, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 7', 'EPSG', 2523, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger zone 7\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2523\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2524, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 8', 'EPSG', 2524, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger zone 8\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2524\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2525, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 9', 'EPSG', 2525, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger zone 9\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2525\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2526, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 10', 'EPSG', 2526, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 10\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2526\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2527, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 11', 'EPSG', 2527, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 11\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2527\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2528, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 12', 'EPSG', 2528, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 12\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2528\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2529, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 13', 'EPSG', 2529, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 13\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2529\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2530, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 14', 'EPSG', 2530, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 14\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2530\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2531, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 15', 'EPSG', 2531, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 15\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2531\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2532, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 16', 'EPSG', 2532, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 16\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",48,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2532\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2533, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 17', 'EPSG', 2533, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 17\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2533\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2534, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 18', 'EPSG', 2534, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 18\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",54,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2534\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2535, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 19', 'EPSG', 2535, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 19\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2535\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2536, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 20', 'EPSG', 2536, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 20\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",60,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2536\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2537, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 21', 'EPSG', 2537, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 21\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2537\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2538, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 22', 'EPSG', 2538, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 22\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",66,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2538\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2539, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 23', 'EPSG', 2539, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 23\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2539\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2540, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 24', 'EPSG', 2540, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 24\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",72,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",24500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2540\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2541, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 25', 'EPSG', 2541, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 25\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2541\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2542, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 26', 'EPSG', 2542, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 26\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",26500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2542\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2543, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 27', 'EPSG', 2543, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 27\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",27500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2543\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2544, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 28', 'EPSG', 2544, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 28\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",28500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2544\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2545, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 29', 'EPSG', 2545, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 29\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",29500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2545\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2546, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 30', 'EPSG', 2546, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 30\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",30500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2546\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2547, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 31', 'EPSG', 2547, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 31\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2547\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2548, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 32', 'EPSG', 2548, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 32\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2548\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2549, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 33', 'EPSG', 2549, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 33\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",33500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2549\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2551, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 34', 'EPSG', 2551, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 34\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",34500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2552, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 35', 'EPSG', 2552, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 35\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",35500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2553, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 36', 'EPSG', 2553, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 36\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",36500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2554, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 37', 'EPSG', 2554, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 37\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",37500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "4\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2555, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 38', 'EPSG', 2555, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 38\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",38500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "5\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2556, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 39', 'EPSG', 2556, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 39\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",39500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "6\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2557, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 40', 'EPSG', 2557, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 40\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",40500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "7\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2558, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 41', 'EPSG', 2558, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 41\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",41500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2559, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 42', 'EPSG', 2559, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 42\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",42500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"255"
  "9\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2560, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 43', 'EPSG', 2560, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 43\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",43500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "0\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2561, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 44', 'EPSG', 2561, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 44\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",44500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2562, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 45', 'EPSG', 2562, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 45\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",45500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2563, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 46', 'EPSG', 2563, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 46\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",138,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",46500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2564, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 47', 'EPSG', 2564, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 47\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",47500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "4\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2565, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 48', 'EPSG', 2565, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 48\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",144,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",48500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "5\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2566, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 49', 'EPSG', 2566, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 49\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",49500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "6\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2567, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 50', 'EPSG', 2567, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 50\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",50500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "7\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2568, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 51', 'EPSG', 2568, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 51\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",51500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2569, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 52', 'EPSG', 2569, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 52\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",156,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",52500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"256"
  "9\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2570, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 53', 'EPSG', 2570, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 53\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",53500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "0\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2571, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 54', 'EPSG', 2571, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 54\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",54500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2572, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 55', 'EPSG', 2572, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 55\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",55500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2573, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 56', 'EPSG', 2573, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 56\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",56500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2574, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 57', 'EPSG', 2574, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 57\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",57500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "4\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2575, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 58', 'EPSG', 2575, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 58\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",58500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "5\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2576, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 59', 'EPSG', 2576, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 59\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",59500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"257"
  "6\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2578, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 61', 'EPSG', 2578, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 61\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",61500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"25"
  "78\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2579, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 62', 'EPSG', 2579, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 62\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",62500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"25"
  "79\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2580, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 63', 'EPSG', 2580, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 63\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",63500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"25"
  "80\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2581, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 64', 'EPSG', 2581, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 64\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",64500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"25"
  "81\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2582, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 21E', 'EPSG', 2582, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 21E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2582\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2583, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 24E', 'EPSG', 2583, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 24E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2583\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2584, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 27E', 'EPSG', 2584, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 27E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2584\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2585, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 30E', 'EPSG', 2585, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 30E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2585\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2586, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 33E', 'EPSG', 2586, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 33E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2586\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2587, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 36E', 'EPSG', 2587, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 36E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2587\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2588, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 39E', 'EPSG', 2588, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 39E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2588\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2589, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 42E', 'EPSG', 2589, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 42E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2589\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2590, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 45E', 'EPSG', 2590, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 45E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2590\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2591, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 48E', 'EPSG', 2591, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 48E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",48,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2591\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2592, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 51E', 'EPSG', 2592, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 51E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2592\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2593, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 54E', 'EPSG', 2593, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 54E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",54,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2593\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2594, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 57E', 'EPSG', 2594, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 57E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2594\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2595, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 60E', 'EPSG', 2595, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 60E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",60,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2595\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2596, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 63E', 'EPSG', 2596, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 63E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2596\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2597, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 66E', 'EPSG', 2597, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 66E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",66,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2597\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2598, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 69E', 'EPSG', 2598, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 69E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2598\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2599, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 72E', 'EPSG', 2599, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 72E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",72,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2599\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2601, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 75E', 'EPSG', 2601, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 75E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2601\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2602, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 78E', 'EPSG', 2602, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 78E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2602\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2603, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 81E', 'EPSG', 2603, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 81E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2603\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2604, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 84E', 'EPSG', 2604, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 84E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2604\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2605, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 87E', 'EPSG', 2605, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 87E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2605\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2606, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 90E', 'EPSG', 2606, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 90E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2606\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2607, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 93E', 'EPSG', 2607, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 93E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2607\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2608, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 96E', 'EPSG', 2608, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 96E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2608\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2609, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 99E', 'EPSG', 2609, 'PROJCS[\"Pulkovo 1942 /"
  " 3-degree Gauss-Kruger CM 99E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2609\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2610, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 102E', 'EPSG', 2610, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 102E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2610\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2611, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 105E', 'EPSG', 2611, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 105E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2611\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2612, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 108E', 'EPSG', 2612, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 108E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2612\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2613, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 111E', 'EPSG', 2613, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 111E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2613\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2614, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 114E', 'EPSG', 2614, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 114E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2614\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2615, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 117E', 'EPSG', 2615, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 117E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2615\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2616, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 120E', 'EPSG', 2616, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 120E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2616\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2617, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 123E', 'EPSG', 2617, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 123E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2617\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2618, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 126E', 'EPSG', 2618, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 126E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2618\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2619, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 129E', 'EPSG', 2619, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 129E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2619\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2620, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 132E', 'EPSG', 2620, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 132E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2620\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2621, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 135E', 'EPSG', 2621, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 135E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2621\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2622, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 138E', 'EPSG', 2622, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 138E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",138,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2622\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2623, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 141E', 'EPSG', 2623, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 141E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2623\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2624, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 144E', 'EPSG', 2624, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 144E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",144,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2624\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2625, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 147E', 'EPSG', 2625, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 147E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2625\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2626, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 150E', 'EPSG', 2626, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 150E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2626\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2627, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 153E', 'EPSG', 2627, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 153E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2627\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2628, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 156E', 'EPSG', 2628, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 156E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",156,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2628\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2629, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 159E', 'EPSG', 2629, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 159E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2629\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2630, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 162E', 'EPSG', 2630, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 162E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2630\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2631, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 165E', 'EPSG', 2631, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 165E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2631\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2632, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 168E', 'EPSG', 2632, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 168E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2632\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2633, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 171E', 'EPSG', 2633, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 171E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2633\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2634, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 174E', 'EPSG', 2634, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 174E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2634\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2635, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 177E', 'EPSG', 2635, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 177E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2635\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2636, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 180E', 'EPSG', 2636, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 180E\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",180,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2636\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2637, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 177W', 'EPSG', 2637, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 177W\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2637\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2638, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 174W', 'EPSG', 2638, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 174W\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2638\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2639, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 171W', 'EPSG', 2639, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 171W\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2639\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2640, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger CM 168W', 'EPSG', 2640, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger CM 168W\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2640\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2641, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 7', 'EPSG', 2641, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger zone 7\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"26"
  "41\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2642, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 8', 'EPSG', 2642, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger zone 8\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"26"
  "42\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2643, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 9', 'EPSG', 2643, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger zone 9\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"26"
  "43\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2644, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 10', 'EPSG', 2644, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 10\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2644\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2645, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 11', 'EPSG', 2645, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 11\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2645\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2646, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 12', 'EPSG', 2646, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 12\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2646\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2647, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 13', 'EPSG', 2647, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 13\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2647\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2648, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 14', 'EPSG', 2648, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 14\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2648\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2649, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 15', 'EPSG', 2649, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 15\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2649\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2650, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 16', 'EPSG', 2650, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 16\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",48,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2650\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2651, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 17', 'EPSG', 2651, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 17\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2651\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2652, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 18', 'EPSG', 2652, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 18\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",54,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2652\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2653, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 19', 'EPSG', 2653, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 19\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2653\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2654, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 20', 'EPSG', 2654, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 20\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",60,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2654\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2655, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 21', 'EPSG', 2655, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 21\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2655\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2656, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 22', 'EPSG', 2656, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 22\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",66,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2656\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2657, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 23', 'EPSG', 2657, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 23\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2657\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2658, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 24', 'EPSG', 2658, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 24\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",72,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",24500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2658\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2659, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 25', 'EPSG', 2659, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 25\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2659\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2660, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 26', 'EPSG', 2660, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 26\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",26500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2660\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2661, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 27', 'EPSG', 2661, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 27\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",27500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2661\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2662, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 28', 'EPSG', 2662, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 28\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",28500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2662\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2663, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 29', 'EPSG', 2663, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 29\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",29500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2663\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2664, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 30', 'EPSG', 2664, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 30\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",30500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2664\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2665, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 31', 'EPSG', 2665, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 31\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2665\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2666, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 32', 'EPSG', 2666, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 32\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2666\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2667, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 33', 'EPSG', 2667, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 33\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",33500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2667\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2668, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 34', 'EPSG', 2668, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 34\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",34500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2668\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2669, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 35', 'EPSG', 2669, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 35\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",35500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2669\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2670, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 36', 'EPSG', 2670, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 36\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",36500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2670\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2671, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 37', 'EPSG', 2671, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 37\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",37500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2671\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2672, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 38', 'EPSG', 2672, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 38\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",38500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2672\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2673, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 39', 'EPSG', 2673, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 39\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",39500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2673\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2674, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 40', 'EPSG', 2674, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 40\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",40500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2674\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2675, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 41', 'EPSG', 2675, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 41\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",41500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2675\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2676, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 42', 'EPSG', 2676, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 42\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",42500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2676\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2677, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 43', 'EPSG', 2677, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 43\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",43500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2677\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2678, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 44', 'EPSG', 2678, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 44\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",44500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2678\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2679, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 45', 'EPSG', 2679, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 45\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",45500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2679\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2680, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 46', 'EPSG', 2680, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 46\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",138,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",46500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2680\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2681, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 47', 'EPSG', 2681, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 47\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",47500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2681\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2682, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 48', 'EPSG', 2682, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 48\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",144,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",48500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2682\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2683, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 49', 'EPSG', 2683, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 49\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",49500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2683\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2684, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 50', 'EPSG', 2684, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 50\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",50500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2684\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2685, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 51', 'EPSG', 2685, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 51\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",51500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2685\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2686, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 52', 'EPSG', 2686, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 52\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",156,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",52500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2686\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2687, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 53', 'EPSG', 2687, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 53\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",53500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2687\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2688, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 54', 'EPSG', 2688, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 54\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",54500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2688\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2689, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 55', 'EPSG', 2689, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 55\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",55500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2689\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2690, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 56', 'EPSG', 2690, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 56\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",56500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2690\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2691, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 57', 'EPSG', 2691, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 57\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",57500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2691\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2692, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 58', 'EPSG', 2692, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 58\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",58500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2693, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 59', 'EPSG', 2693, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 59\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",59500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2693\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2695, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 61', 'EPSG', 2695, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 61\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",61500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2695\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2696, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 62', 'EPSG', 2696, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 62\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",62500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2696\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2697, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 63', 'EPSG', 2697, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 63\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",63500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2697\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2698, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 64', 'EPSG', 2698, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 64\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",64500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2698\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2699, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 21E', 'EPSG', 2699, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 21E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"26"
  "99\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2700, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 24E', 'EPSG', 2700, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 24E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "00\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2701, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 27E', 'EPSG', 2701, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 27E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "01\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2702, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 30E', 'EPSG', 2702, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 30E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "02\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2703, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 33E', 'EPSG', 2703, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 33E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "03\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2704, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 36E', 'EPSG', 2704, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 36E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "04\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2705, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 39E', 'EPSG', 2705, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 39E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "05\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2706, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 42E', 'EPSG', 2706, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 42E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "06\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2707, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 45E', 'EPSG', 2707, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 45E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "07\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2708, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 48E', 'EPSG', 2708, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 48E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",48,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "08\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2709, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 51E', 'EPSG', 2709, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 51E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "09\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2710, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 54E', 'EPSG', 2710, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 54E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",54,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "10\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2711, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 57E', 'EPSG', 2711, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 57E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "11\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2712, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 60E', 'EPSG', 2712, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 60E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",60,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "12\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2713, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 63E', 'EPSG', 2713, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 63E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "13\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2714, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 66E', 'EPSG', 2714, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 66E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",66,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "14\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2715, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 69E', 'EPSG', 2715, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 69E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "15\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2716, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 72E', 'EPSG', 2716, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 72E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",72,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "16\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2717, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 75E', 'EPSG', 2717, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 75E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "17\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2718, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 78E', 'EPSG', 2718, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 78E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "18\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2719, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 81E', 'EPSG', 2719, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 81E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "19\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2720, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 84E', 'EPSG', 2720, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 84E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "20\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2721, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 87E', 'EPSG', 2721, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 87E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "21\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2722, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 90E', 'EPSG', 2722, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 90E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "22\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2723, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 93E', 'EPSG', 2723, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 93E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "23\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2724, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 96E', 'EPSG', 2724, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 96E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "24\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2725, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 99E', 'EPSG', 2725, 'PROJCS[\"Pulkovo 1995 /"
  " 3-degree Gauss-Kruger CM 99E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,2"
  "98.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"27"
  "25\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2726, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 102E', 'EPSG', 2726, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 102E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2726\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2727, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 105E', 'EPSG', 2727, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 105E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2727\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2728, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 108E', 'EPSG', 2728, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 108E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2728\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2729, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 111E', 'EPSG', 2729, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 111E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2729\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2730, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 114E', 'EPSG', 2730, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 114E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2730\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2731, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 117E', 'EPSG', 2731, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 117E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2731\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2732, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 120E', 'EPSG', 2732, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 120E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2732\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2733, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 123E', 'EPSG', 2733, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 123E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2733\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2734, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 126E', 'EPSG', 2734, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 126E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2734\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2735, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 129E', 'EPSG', 2735, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 129E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2735\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2736, 1, 'Tete / UTM zone 36S', 'EPSG', 2736, 'PROJCS[\"Tete / UTM zone 36S\",GEOGCS[\"Tete\","
  "DATUM[\"Tete\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115.064,-8"
  "7.39,-101.716,-0.058,4.001,-2.062,9.366],AUTHORITY[\"EPSG\",\"6127\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4127\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2736\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2737, 1, 'Tete / UTM zone 37S', 'EPSG', 2737, 'PROJCS[\"Tete / UTM zone 37S\",GEOGCS[\"Tete\","
  "DATUM[\"Tete\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115.064,-8"
  "7.39,-101.716,-0.058,4.001,-2.062,9.366],AUTHORITY[\"EPSG\",\"6127\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4127\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2737\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2738, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 132E', 'EPSG', 2738, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 132E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2739, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 135E', 'EPSG', 2739, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 135E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2739\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2740, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 138E', 'EPSG', 2740, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 138E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",138,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2740\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2741, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 141E', 'EPSG', 2741, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 141E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2741\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2742, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 144E', 'EPSG', 2742, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 144E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",144,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2742\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2743, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 147E', 'EPSG', 2743, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 147E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2743\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2744, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 150E', 'EPSG', 2744, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 150E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2744\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2745, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 153E', 'EPSG', 2745, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 153E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2745\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2746, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 156E', 'EPSG', 2746, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 156E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",156,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2746\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2747, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 159E', 'EPSG', 2747, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 159E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2747\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2748, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 162E', 'EPSG', 2748, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 162E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2748\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2749, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 165E', 'EPSG', 2749, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 165E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2749\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2750, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 168E', 'EPSG', 2750, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 168E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2750\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2751, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 171E', 'EPSG', 2751, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 171E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2751\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2752, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 174E', 'EPSG', 2752, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 174E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2752\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2753, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 177E', 'EPSG', 2753, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 177E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2753\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2754, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 180E', 'EPSG', 2754, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 180E\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",180,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"2754\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2755, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 177W', 'EPSG', 2755, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 177W\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2755\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2756, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 174W', 'EPSG', 2756, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 174W\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-174,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2756\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2757, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 171W', 'EPSG', 2757, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 171W\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2757\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2758, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger CM 168W', 'EPSG', 2758, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger CM 168W\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-168,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"2758\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2759, 1, 'NAD83(HARN) / Alabama East', 'EPSG', 2759, 'PROJCS[\"NAD83(HARN) / Alabama East\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",30.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.83333333333"
  "33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99996,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2759\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2760, 1, 'NAD83(HARN) / Alabama West', 'EPSG', 2760, 'PROJCS[\"NAD83(HARN) / Alabama West\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2760\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2761, 1, 'NAD83(HARN) / Arizona East', 'EPSG', 2761, 'PROJCS[\"NAD83(HARN) / Arizona East\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.177777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2761\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2762, 1, 'NAD83(HARN) / Arizona Central', 'EPSG', 2762, 'PROJCS[\"NAD83(HARN) / Arizona Centra"
  "l\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111.927777"
  "777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"276"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2763, 1, 'NAD83(HARN) / Arizona West', 'EPSG', 2763, 'PROJCS[\"NAD83(HARN) / Arizona West\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.75,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2763\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2764, 1, 'NAD83(HARN) / Arkansas North', 'EPSG', 2764, 'PROJCS[\"NAD83(HARN) / Arkansas North\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.2333333333333,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2764\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2765, 1, 'NAD83(HARN) / Arkansas South', 'EPSG', 2765, 'PROJCS[\"NAD83(HARN) / Arkansas South\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.7666666666667,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"2765\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2766, 1, 'NAD83(HARN) / California zone 1', 'EPSG', 2766, 'PROJCS[\"NAD83(HARN) / California z"
  "one 1\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.6777777777778,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2766\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2767, 1, 'NAD83(HARN) / California zone 2', 'EPSG', 2767, 'PROJCS[\"NAD83(HARN) / California z"
  "one 2\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"2767\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2768, 1, 'NAD83(HARN) / California zone 3', 'EPSG', 2768, 'PROJCS[\"NAD83(HARN) / California z"
  "one 3\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.444444"
  "4444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.0777777777778,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2768\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2769, 1, 'NAD83(HARN) / California zone 4', 'EPSG', 2769, 'PROJCS[\"NAD83(HARN) / California z"
  "one 4\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-119,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.2611111111111,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2769\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2770, 1, 'NAD83(HARN) / California zone 5', 'EPSG', 2770, 'PROJCS[\"NAD83(HARN) / California z"
  "one 5\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-118,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.4666666666667,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"2770\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2771, 1, 'NAD83(HARN) / California zone 6', 'EPSG', 2771, 'PROJCS[\"NAD83(HARN) / California z"
  "one 6\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-116.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.8833333333333,AU"
  "THORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTHORITY[\"EPSG\",\"882"
  "4\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"2771\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2772, 1, 'NAD83(HARN) / Colorado North', 'EPSG', 2772, 'PROJCS[\"NAD83(HARN) / Colorado North\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.794444444444"
  "4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",304800.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2772\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2773, 1, 'NAD83(HARN) / Colorado Central', 'EPSG', 2773, 'PROJCS[\"NAD83(HARN) / Colorado Cent"
  "ral\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.75,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",30480"
  "0.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"2773\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2774, 1, 'NAD83(HARN) / Colorado South', 'EPSG', 2774, 'PROJCS[\"NAD83(HARN) / Colorado South\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.444444444444"
  "4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",304800.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2774\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2775, 1, 'NAD83(HARN) / Connecticut', 'EPSG', 2775, 'PROJCS[\"NAD83(HARN) / Connecticut\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-72.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.8666666666667,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",304800.6096,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1524"
  "00.3048,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"2775\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2776, 1, 'NAD83(HARN) / Delaware', 'EPSG', 2776, 'PROJCS[\"NAD83(HARN) / Delaware\",GEOGCS[\"N"
  "AD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.4166666666667,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2776\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2777, 1, 'NAD83(HARN) / Florida East', 'EPSG', 2777, 'PROJCS[\"NAD83(HARN) / Florida East\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2777\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2778, 1, 'NAD83(HARN) / Florida West', 'EPSG', 2778, 'PROJCS[\"NAD83(HARN) / Florida West\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2778\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2779, 1, 'NAD83(HARN) / Florida North', 'EPSG', 2779, 'PROJCS[\"NAD83(HARN) / Florida North\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.511111"
  "1111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7611111111111,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2779\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2780, 1, 'NAD83(HARN) / Georgia East', 'EPSG', 2780, 'PROJCS[\"NAD83(HARN) / Georgia East\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.1777777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2780\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2781, 1, 'NAD83(HARN) / Georgia West', 'EPSG', 2781, 'PROJCS[\"NAD83(HARN) / Georgia West\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.1777777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2781\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2782, 1, 'NAD83(HARN) / Hawaii zone 1', 'EPSG', 2782, 'PROJCS[\"NAD83(HARN) / Hawaii zone 1\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",18.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "55.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2782\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2783, 1, 'NAD83(HARN) / Hawaii zone 2', 'EPSG', 2783, 'PROJCS[\"NAD83(HARN) / Hawaii zone 2\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",20.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "56.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2783\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2784, 1, 'NAD83(HARN) / Hawaii zone 3', 'EPSG', 2784, 'PROJCS[\"NAD83(HARN) / Hawaii zone 3\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",21.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "58,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2784\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2785, 1, 'NAD83(HARN) / Hawaii zone 4', 'EPSG', 2785, 'PROJCS[\"NAD83(HARN) / Hawaii zone 4\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",21.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "59.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2785\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2786, 1, 'NAD83(HARN) / Hawaii zone 5', 'EPSG', 2786, 'PROJCS[\"NAD83(HARN) / Hawaii zone 5\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",21.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "60.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2786\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2787, 1, 'NAD83(HARN) / Idaho East', 'EPSG', 2787, 'PROJCS[\"NAD83(HARN) / Idaho East\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.177"
  "777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2787\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2788, 1, 'NAD83(HARN) / Idaho Central', 'EPSG', 2788, 'PROJCS[\"NAD83(HARN) / Idaho Central\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "14,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"27"
  "88\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2789, 1, 'NAD83(HARN) / Idaho West', 'EPSG', 2789, 'PROJCS[\"NAD83(HARN) / Idaho West\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.75,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2789\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2790, 1, 'NAD83(HARN) / Illinois East', 'EPSG', 2790, 'PROJCS[\"NAD83(HARN) / Illinois East\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "8.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999975,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EP"
  "SG\",\"2790\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2791, 1, 'NAD83(HARN) / Illinois West', 'EPSG', 2791, 'PROJCS[\"NAD83(HARN) / Illinois West\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "0.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2792, 1, 'NAD83(HARN) / Indiana East', 'EPSG', 2792, 'PROJCS[\"NAD83(HARN) / Indiana East\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85"
  ".6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"2792\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2793, 1, 'NAD83(HARN) / Indiana West', 'EPSG', 2793, 'PROJCS[\"NAD83(HARN) / Indiana West\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87"
  ".0944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"2793\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2794, 1, 'NAD83(HARN) / Iowa North', 'EPSG', 2794, 'PROJCS[\"NAD83(HARN) / Iowa North\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.2777777777778,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1"
  "000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"2794\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2795, 1, 'NAD83(HARN) / Iowa South', 'EPSG', 2795, 'PROJCS[\"NAD83(HARN) / Iowa South\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2795\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2796, 1, 'NAD83(HARN) / Kansas North', 'EPSG', 2796, 'PROJCS[\"NAD83(HARN) / Kansas North\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.7944444444444,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2796\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2797, 1, 'NAD83(HARN) / Kansas South', 'EPSG', 2797, 'PROJCS[\"NAD83(HARN) / Kansas South\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.5666666666667,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",400000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"2797\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2798, 1, 'NAD83(HARN) / Kentucky North', 'EPSG', 2798, 'PROJCS[\"NAD83(HARN) / Kentucky North\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-84.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AU"
  "THORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"2798\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2799, 1, 'NAD83(HARN) / Kentucky South', 'EPSG', 2799, 'PROJCS[\"NAD83(HARN) / Kentucky South\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9333333333333,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000"
  "00,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"2799\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2800, 1, 'NAD83(HARN) / Louisiana North', 'EPSG', 2800, 'PROJCS[\"NAD83(HARN) / Louisiana Nort"
  "h\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.6777777777778,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"2800\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2801, 1, 'NAD83(HARN) / Louisiana South', 'EPSG', 2801, 'PROJCS[\"NAD83(HARN) / Louisiana Sout"
  "h\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91"
  ".3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2801\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2802, 1, 'NAD83(HARN) / Maine East', 'EPSG', 2802, 'PROJCS[\"NAD83(HARN) / Maine East\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-68.5111"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2802\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2803, 1, 'NAD83(HARN) / Maine West', 'EPSG', 2803, 'PROJCS[\"NAD83(HARN) / Maine West\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.1777"
  "777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2803\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2804, 1, 'NAD83(HARN) / Maryland', 'EPSG', 2804, 'PROJCS[\"NAD83(HARN) / Maryland\",GEOGCS[\"N"
  "AD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2804\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2805, 1, 'NAD83(HARN) / Massachusetts Mainland', 'EPSG', 2805, 'PROJCS[\"NAD83(HARN) / Massach"
  "usetts Mainland\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG"
  "\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.694444444"
  "4444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.7277777777778,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",750000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"2805\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2806, 1, 'NAD83(HARN) / Massachusetts Island', 'EPSG', 2806, 'PROJCS[\"NAD83(HARN) / Massachus"
  "etts Island\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.4944444444444,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"2806\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2807, 1, 'NAD83(HARN) / Michigan North', 'EPSG', 2807, 'PROJCS[\"NAD83(HARN) / Michigan North\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-87,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.0944444444444,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",8000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"2807\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2808, 1, 'NAD83(HARN) / Michigan Central', 'EPSG', 2808, 'PROJCS[\"NAD83(HARN) / Michigan Cent"
  "ral\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.7,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",6000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"2808\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2809, 1, 'NAD83(HARN) / Michigan South', 'EPSG', 2809, 'PROJCS[\"NAD83(HARN) / Michigan South\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.677777777777"
  "8,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AU"
  "THORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"2809\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2810, 1, 'NAD83(HARN) / Minnesota North', 'EPSG', 2810, 'PROJCS[\"NAD83(HARN) / Minnesota Nort"
  "h\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-93.1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6333333333333,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",10"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"2810\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2811, 1, 'NAD83(HARN) / Minnesota Central', 'EPSG', 2811, 'PROJCS[\"NAD83(HARN) / Minnesota Ce"
  "ntral\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-9"
  "4.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2811\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2812, 1, 'NAD83(HARN) / Minnesota South', 'EPSG', 2812, 'PROJCS[\"NAD83(HARN) / Minnesota Sout"
  "h\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2812\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2813, 1, 'NAD83(HARN) / Mississippi East', 'EPSG', 2813, 'PROJCS[\"NAD83(HARN) / Mississippi E"
  "ast\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.833"
  "3333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2813\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2814, 1, 'NAD83(HARN) / Mississippi West', 'EPSG', 2814, 'PROJCS[\"NAD83(HARN) / Mississippi W"
  "est\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.333"
  "3333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2814\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2815, 1, 'NAD83(HARN) / Missouri East', 'EPSG', 2815, 'PROJCS[\"NAD83(HARN) / Missouri East\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "0.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2815\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2816, 1, 'NAD83(HARN) / Missouri Central', 'EPSG', 2816, 'PROJCS[\"NAD83(HARN) / Missouri Cent"
  "ral\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"2816\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2817, 1, 'NAD83(HARN) / Missouri West', 'EPSG', 2817, 'PROJCS[\"NAD83(HARN) / Missouri West\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",36.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "4.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",850000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2817\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2818, 1, 'NAD83(HARN) / Montana', 'EPSG', 2818, 'PROJCS[\"NAD83(HARN) / Montana\",GEOGCS[\"NAD"
  "83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-109.51"
  "1111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false or"
  "igin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2818\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2819, 1, 'NAD83(HARN) / Nebraska', 'EPSG', 2819, 'PROJCS[\"NAD83(HARN) / Nebraska\",GEOGCS[\"N"
  "AD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",5000"
  "00,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2819\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2820, 1, 'NAD83(HARN) / Nevada East', 'EPSG', 2820, 'PROJCS[\"NAD83(HARN) / Nevada East\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.59444444444"
  "4,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"28"
  "20\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2821, 1, 'NAD83(HARN) / Nevada Central', 'EPSG', 2821, 'PROJCS[\"NAD83(HARN) / Nevada Central\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-116.666666"
  "666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2821\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2822, 1, 'NAD83(HARN) / Nevada West', 'EPSG', 2822, 'PROJCS[\"NAD83(HARN) / Nevada West\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118.59444444444"
  "4,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"28"
  "22\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2823, 1, 'NAD83(HARN) / New Hampshire', 'EPSG', 2823, 'PROJCS[\"NAD83(HARN) / New Hampshire\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-7"
  "1.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2823\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2824, 1, 'NAD83(HARN) / New Jersey', 'EPSG', 2824, 'PROJCS[\"NAD83(HARN) / New Jersey\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2824\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2825, 1, 'NAD83(HARN) / New Mexico East', 'EPSG', 2825, 'PROJCS[\"NAD83(HARN) / New Mexico Eas"
  "t\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-104.333333"
  "333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",165000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2825\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2826, 1, 'NAD83(HARN) / New Mexico Central', 'EPSG', 2826, 'PROJCS[\"NAD83(HARN) / New Mexico "
  "Central\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-106."
  "25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2826\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2827, 1, 'NAD83(HARN) / New Mexico West', 'EPSG', 2827, 'PROJCS[\"NAD83(HARN) / New Mexico Wes"
  "t\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.833333"
  "333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",830000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2827\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2828, 1, 'NAD83(HARN) / New York East', 'EPSG', 2828, 'PROJCS[\"NAD83(HARN) / New York East\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-7"
  "4.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"2828\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2829, 1, 'NAD83(HARN) / New York Central', 'EPSG', 2829, 'PROJCS[\"NAD83(HARN) / New York Cent"
  "ral\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.59444"
  "44444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2829\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2830, 1, 'NAD83(HARN) / New York West', 'EPSG', 2830, 'PROJCS[\"NAD83(HARN) / New York West\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-78.594444444444"
  "4,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",350000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2830\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2831, 1, 'NAD83(HARN) / New York Long Island', 'EPSG', 2831, 'PROJCS[\"NAD83(HARN) / New York "
  "Long Island\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.0333333333333,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"2831\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2832, 1, 'NAD83(HARN) / North Dakota North', 'EPSG', 2832, 'PROJCS[\"NAD83(HARN) / North Dakot"
  "a North\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"2832\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2833, 1, 'NAD83(HARN) / North Dakota South', 'EPSG', 2833, 'PROJCS[\"NAD83(HARN) / North Dakot"
  "a South\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.494"
  "4444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.1944444444444,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"2833\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2834, 1, 'NAD83(HARN) / Ohio North', 'EPSG', 2834, 'PROJCS[\"NAD83(HARN) / Ohio North\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"2834\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2835, 1, 'NAD83(HARN) / Ohio South', 'EPSG', 2835, 'PROJCS[\"NAD83(HARN) / Ohio South\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2835\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2836, 1, 'NAD83(HARN) / Oklahoma North', 'EPSG', 2836, 'PROJCS[\"NAD83(HARN) / Oklahoma North\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2836\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2837, 1, 'NAD83(HARN) / Oklahoma South', 'EPSG', 2837, 'PROJCS[\"NAD83(HARN) / Oklahoma South\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.2333333333333,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2838, 1, 'NAD83(HARN) / Oregon North', 'EPSG', 2838, 'PROJCS[\"NAD83(HARN) / Oregon North\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2838\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2839, 1, 'NAD83(HARN) / Oregon South', 'EPSG', 2839, 'PROJCS[\"NAD83(HARN) / Oregon South\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2840, 1, 'NAD83(HARN) / Rhode Island', 'EPSG', 2840, 'PROJCS[\"NAD83(HARN) / Rhode Island\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999375,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"2840\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2841, 1, 'NAD83(HARN) / South Dakota North', 'EPSG', 2841, 'PROJCS[\"NAD83(HARN) / South Dakot"
  "a North\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.6944444444444,AU"
  "THORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHORITY[\"EPSG\",\"882"
  "4\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin"
  "\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"2841\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2842, 1, 'NAD83(HARN) / South Dakota South', 'EPSG', 2842, 'PROJCS[\"NAD83(HARN) / South Dakot"
  "a South\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.4,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"2842\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2843, 1, 'NAD83(HARN) / Tennessee', 'EPSG', 2843, 'PROJCS[\"NAD83(HARN) / Tennessee\",GEOGCS[\""
  "NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-86,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2843\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2844, 1, 'NAD83(HARN) / Texas North', 'EPSG', 2844, 'PROJCS[\"NAD83(HARN) / Texas North\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-101.51111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2844\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2845, 1, 'NAD83(HARN) / Texas North Central', 'EPSG', 2845, 'PROJCS[\"NAD83(HARN) / Texas Nort"
  "h Central\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.9"
  "777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.1444444444444,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing "
  "at false origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2845\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2846, 1, 'NAD83(HARN) / Texas Central', 'EPSG', 2846, 'PROJCS[\"NAD83(HARN) / Texas Central\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",31.8833333333333,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",3000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2846\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2847, 1, 'NAD83(HARN) / Texas South Central', 'EPSG', 2847, 'PROJCS[\"NAD83(HARN) / Texas Sout"
  "h Central\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.2833333333333,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",4000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"2847\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2848, 1, 'NAD83(HARN) / Texas South', 'EPSG', 2848, 'PROJCS[\"NAD83(HARN) / Texas South\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",5000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"2848\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2849, 1, 'NAD83(HARN) / Utah North', 'EPSG', 2849, 'PROJCS[\"NAD83(HARN) / Utah North\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "1.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",10"
  "00000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"2849\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2850, 1, 'NAD83(HARN) / Utah Central', 'EPSG', 2850, 'PROJCS[\"NAD83(HARN) / Utah Central\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2000000"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"2850\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2851, 1, 'NAD83(HARN) / Utah South', 'EPSG', 2851, 'PROJCS[\"NAD83(HARN) / Utah South\",GEOGCS"
  "[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "1.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2851\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2852, 1, 'NAD83(HARN) / Vermont', 'EPSG', 2852, 'PROJCS[\"NAD83(HARN) / Vermont\",GEOGCS[\"NAD"
  "83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72.511111111"
  "1111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2852\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2853, 1, 'NAD83(HARN) / Virginia North', 'EPSG', 2853, 'PROJCS[\"NAD83(HARN) / Virginia North\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.211111111111"
  "1,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"2853\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2854, 1, 'NAD83(HARN) / Virginia South', 'EPSG', 2854, 'PROJCS[\"NAD83(HARN) / Virginia South\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.977777777777"
  "8,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"2854\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2855, 1, 'NAD83(HARN) / Washington North', 'EPSG', 2855, 'PROJCS[\"NAD83(HARN) / Washington No"
  "rth\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "0.833333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"2855\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2856, 1, 'NAD83(HARN) / Washington South', 'EPSG', 2856, 'PROJCS[\"NAD83(HARN) / Washington So"
  "uth\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.3333333"
  "333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.8333333333333,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fal"
  "se origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2856\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2857, 1, 'NAD83(HARN) / West Virginia North', 'EPSG', 2857, 'PROJCS[\"NAD83(HARN) / West Virgi"
  "nia North\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.2"
  "611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"2857\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2858, 1, 'NAD83(HARN) / West Virginia South', 'EPSG', 2858, 'PROJCS[\"NAD83(HARN) / West Virgi"
  "nia South\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333333,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"2858\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2859, 1, 'NAD83(HARN) / Wisconsin North', 'EPSG', 2859, 'PROJCS[\"NAD83(HARN) / Wisconsin Nort"
  "h\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.7666666666667,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AU"
  "THORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"2859\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2860, 1, 'NAD83(HARN) / Wisconsin Central', 'EPSG', 2860, 'PROJCS[\"NAD83(HARN) / Wisconsin Ce"
  "ntral\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.5111111111111,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"2860\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2861, 1, 'NAD83(HARN) / Wisconsin South', 'EPSG', 2861, 'PROJCS[\"NAD83(HARN) / Wisconsin Sout"
  "h\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2861\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2862, 1, 'NAD83(HARN) / Wyoming East', 'EPSG', 2862, 'PROJCS[\"NAD83(HARN) / Wyoming East\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-10"
  "5.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EP"
  "SG\",\"2862\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2863, 1, 'NAD83(HARN) / Wyoming East Central', 'EPSG', 2863, 'PROJCS[\"NAD83(HARN) / Wyoming E"
  "ast Central\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-107.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2863\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2864, 1, 'NAD83(HARN) / Wyoming West Central', 'EPSG', 2864, 'PROJCS[\"NAD83(HARN) / Wyoming W"
  "est Central\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-108.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"2864\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2865, 1, 'NAD83(HARN) / Wyoming West', 'EPSG', 2865, 'PROJCS[\"NAD83(HARN) / Wyoming West\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-11"
  "0.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"2865\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2866, 1, 'NAD83(HARN) / Puerto Rico and Virgin Is.', 'EPSG', 2866, 'PROJCS[\"NAD83(HARN) / Pue"
  "rto Rico and Virgin Is.\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 198"
  "0\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",17.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-66.4333333333334,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",18.4333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",18.044444444444"
  "4,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2866\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2867, 1, 'NAD83(HARN) / Arizona East (ft)', 'EPSG', 2867, 'PROJCS[\"NAD83(HARN) / Arizona East"
  " (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.17"
  "7777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG"
  "\",\"2867\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2868, 1, 'NAD83(HARN) / Arizona Central (ft)', 'EPSG', 2868, 'PROJCS[\"NAD83(HARN) / Arizona C"
  "entral (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-"
  "111.927777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"2868\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2869, 1, 'NAD83(HARN) / Arizona West (ft)', 'EPSG', 2869, 'PROJCS[\"NAD83(HARN) / Arizona West"
  " (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.75"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2"
  "869\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2870, 1, 'NAD83(HARN) / California zone 1 (ftUS)', 'EPSG', 2870, 'PROJCS[\"NAD83(HARN) / Calif"
  "ornia zone 1 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.6777"
  "777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2870\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2871, 1, 'NAD83(HARN) / California zone 2 (ftUS)', 'EPSG', 2871, 'PROJCS[\"NAD83(HARN) / Calif"
  "ornia zone 2 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.8333"
  "333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northin"
  "g at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2871\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2872, 1, 'NAD83(HARN) / California zone 3 (ftUS)', 'EPSG', 2872, 'PROJCS[\"NAD83(HARN) / Calif"
  "ornia zone 3 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paral"
  "lel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.0777777777778"
  ",AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012192"
  "4,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2872\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2873, 1, 'NAD83(HARN) / California zone 4 (ftUS)', 'EPSG', 2873, 'PROJCS[\"NAD83(HARN) / Calif"
  "ornia zone 4 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-119,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.2611"
  "111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2873\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2874, 1, 'NAD83(HARN) / California zone 5 (ftUS)', 'EPSG', 2874, 'PROJCS[\"NAD83(HARN) / Calif"
  "ornia zone 5 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-118,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.4666"
  "666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northin"
  "g at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2874\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2875, 1, 'NAD83(HARN) / California zone 6 (ftUS)', 'EPSG', 2875, 'PROJCS[\"NAD83(HARN) / Calif"
  "ornia zone 6 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-116.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.8"
  "833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2875\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2876, 1, 'NAD83(HARN) / Colorado North (ftUS)', 'EPSG', 2876, 'PROJCS[\"NAD83(HARN) / Colorado"
  " North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "40.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2876\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2877, 1, 'NAD83(HARN) / Colorado Central (ftUS)', 'EPSG', 2877, 'PROJCS[\"NAD83(HARN) / Colora"
  "do Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",39.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\"882"
  "4\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2877\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2878, 1, 'NAD83(HARN) / Colorado South (ftUS)', 'EPSG', 2878, 'PROJCS[\"NAD83(HARN) / Colorado"
  " South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2878\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2879, 1, 'NAD83(HARN) / Connecticut (ftUS)', 'EPSG', 2879, 'PROJCS[\"NAD83(HARN) / Connecticut"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-72.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.8666666666667,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2879\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2880, 1, 'NAD83(HARN) / Delaware (ftUS)', 'EPSG', 2880, 'PROJCS[\"NAD83(HARN) / Delaware (ftUS"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.4166666"
  "666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2880\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2881, 1, 'NAD83(HARN) / Florida East (ftUS)', 'EPSG', 2881, 'PROJCS[\"NAD83(HARN) / Florida Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2881\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2882, 1, 'NAD83(HARN) / Florida West (ftUS)', 'EPSG', 2882, 'PROJCS[\"NAD83(HARN) / Florida We"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-82,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2882\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2883, 1, 'NAD83(HARN) / Florida North (ftUS)', 'EPSG', 2883, 'PROJCS[\"NAD83(HARN) / Florida N"
  "orth (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-84.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7611111111111,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2883\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2884, 1, 'NAD83(HARN) / Georgia East (ftUS)', 'EPSG', 2884, 'PROJCS[\"NAD83(HARN) / Georgia Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82"
  ".1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2884\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2885, 1, 'NAD83(HARN) / Georgia West (ftUS)', 'EPSG', 2885, 'PROJCS[\"NAD83(HARN) / Georgia We"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84"
  ".1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2885\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2886, 1, 'NAD83(HARN) / Idaho East (ftUS)', 'EPSG', 2886, 'PROJCS[\"NAD83(HARN) / Idaho East ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-112.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2886\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2887, 1, 'NAD83(HARN) / Idaho Central (ftUS)', 'EPSG', 2887, 'PROJCS[\"NAD83(HARN) / Idaho Cen"
  "tral (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2887\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2888, 1, 'NAD83(HARN) / Idaho West (ftUS)', 'EPSG', 2888, 'PROJCS[\"NAD83(HARN) / Idaho West ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-115.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",2624666.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2888\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2891, 1, 'NAD83(HARN) / Kentucky North (ftUS)', 'EPSG', 2891, 'PROJCS[\"NAD83(HARN) / Kentucky"
  " North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-84.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.97777777"
  "77778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2891\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2892, 1, 'NAD83(HARN) / Kentucky South (ftUS)', 'EPSG', 2892, 'PROJCS[\"NAD83(HARN) / Kentucky"
  " South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.93333333"
  "33333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2892\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2893, 1, 'NAD83(HARN) / Maryland (ftUS)', 'EPSG', 2893, 'PROJCS[\"NAD83(HARN) / Maryland (ftUS"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-77,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1312333.333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2893\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2894, 1, 'NAD83(HARN) / Massachusetts Mainland (ftUS)', 'EPSG', 2894, 'PROJCS[\"NAD83(HARN) / "
  "Massachusetts Mainland (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"G"
  "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AU"
  "THORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longi"
  "tude of false origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",42.6944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.7277777777778,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",2460625,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2894\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2895, 1, 'NAD83(HARN) / Massachusetts Island (ftUS)', 'EPSG', 2895, 'PROJCS[\"NAD83(HARN) / Ma"
  "ssachusetts Island (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude"
  " of false origin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41"
  ".4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2895\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2896, 1, 'NAD83(HARN) / Michigan North (ft)', 'EPSG', 2896, 'PROJCS[\"NAD83(HARN) / Michigan N"
  "orth (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-87,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.0944444444444,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",26246719.16,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"2896\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2897, 1, 'NAD83(HARN) / Michigan Central (ft)', 'EPSG', 2897, 'PROJCS[\"NAD83(HARN) / Michigan"
  " Central (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "45.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",19685039.37,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2897\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2898, 1, 'NAD83(HARN) / Michigan South (ft)', 'EPSG', 2898, 'PROJCS[\"NAD83(HARN) / Michigan S"
  "outh (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.6"
  "777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",13123359.58,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"2898\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2899, 1, 'NAD83(HARN) / Mississippi East (ftUS)', 'EPSG', 2899, 'PROJCS[\"NAD83(HARN) / Missis"
  "sippi East (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-88.8333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2899\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2900, 1, 'NAD83(HARN) / Mississippi West (ftUS)', 'EPSG', 2900, 'PROJCS[\"NAD83(HARN) / Missis"
  "sippi West (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-90.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2900\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2901, 1, 'NAD83(HARN) / Montana (ft)', 'EPSG', 2901, 'PROJCS[\"NAD83(HARN) / Montana (ft)\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-109.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"2901\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2902, 1, 'NAD83(HARN) / New Mexico East (ftUS)', 'EPSG', 2902, 'PROJCS[\"NAD83(HARN) / New Mex"
  "ico East (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-104.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",541337.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2902\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2903, 1, 'NAD83(HARN) / New Mexico Central (ftUS)', 'EPSG', 2903, 'PROJCS[\"NAD83(HARN) / New "
  "Mexico Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-106.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2903\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2904, 1, 'NAD83(HARN) / New Mexico West (ftUS)', 'EPSG', 2904, 'PROJCS[\"NAD83(HARN) / New Mex"
  "ico West (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-107.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",2723091.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2904\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2905, 1, 'NAD83(HARN) / New York East (ftUS)', 'EPSG', 2905, 'PROJCS[\"NAD83(HARN) / New York "
  "East (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2905\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2906, 1, 'NAD83(HARN) / New York Central (ftUS)', 'EPSG', 2906, 'PROJCS[\"NAD83(HARN) / New Yo"
  "rk Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-76.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",820208.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2906\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2907, 1, 'NAD83(HARN) / New York West (ftUS)', 'EPSG', 2907, 'PROJCS[\"NAD83(HARN) / New York "
  "West (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-"
  "78.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",1148291.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2907\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2908, 1, 'NAD83(HARN) / New York Long Island (ftUS)', 'EPSG', 2908, 'PROJCS[\"NAD83(HARN) / Ne"
  "w York Long Island (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMET"
  "ER[\"Longitude of false origin\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41"
  ".0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northin"
  "g at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2908\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2909, 1, 'NAD83(HARN) / North Dakota North (ft)', 'EPSG', 2909, 'PROJCS[\"NAD83(HARN) / North "
  "Dakota North (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.74444444"
  "44444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2909\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2910, 1, 'NAD83(HARN) / North Dakota South (ft)', 'EPSG', 2910, 'PROJCS[\"NAD83(HARN) / North "
  "Dakota South (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",47.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.1944444444444,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2910\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2911, 1, 'NAD83(HARN) / Oklahoma North (ftUS)', 'EPSG', 2911, 'PROJCS[\"NAD83(HARN) / Oklahoma"
  " North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2911\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2912, 1, 'NAD83(HARN) / Oklahoma South (ftUS)', 'EPSG', 2912, 'PROJCS[\"NAD83(HARN) / Oklahoma"
  " South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.23333333333"
  "33,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2912\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2913, 1, 'NAD83(HARN) / Oregon North (ft)', 'EPSG', 2913, 'PROJCS[\"NAD83(HARN) / Oregon North"
  " (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",8202099.738,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"2913\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2914, 1, 'NAD83(HARN) / Oregon South (ft)', 'EPSG', 2914, 'PROJCS[\"NAD83(HARN) / Oregon South"
  " (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",4921259.843,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"2914\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2915, 1, 'NAD83(HARN) / Tennessee (ftUS)', 'EPSG', 2915, 'PROJCS[\"NAD83(HARN) / Tennessee (ft"
  "US)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-86,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2915\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2916, 1, 'NAD83(HARN) / Texas North (ftUS)', 'EPSG', 2916, 'PROJCS[\"NAD83(HARN) / Texas North"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-101.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",32808"
  "33.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2916\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2917, 1, 'NAD83(HARN) / Texas North Central (ftUS)', 'EPSG', 2917, 'PROJCS[\"NAD83(HARN) / Tex"
  "as North Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 198"
  "0\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",33.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.144444444444"
  "4,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,A"
  "UTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2917\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2918, 1, 'NAD83(HARN) / Texas Central (ftUS)', 'EPSG', 2918, 'PROJCS[\"NAD83(HARN) / Texas Cen"
  "tral (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",31."
  "8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2296583.333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",9842500,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2918\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2919, 1, 'NAD83(HARN) / Texas South Central (ftUS)', 'EPSG', 2919, 'PROJCS[\"NAD83(HARN) / Tex"
  "as South Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 198"
  "0\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.2833"
  "333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",13123333.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EP"
  "SG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2919\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2920, 1, 'NAD83(HARN) / Texas South (ftUS)', 'EPSG', 2920, 'PROJCS[\"NAD83(HARN) / Texas South"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.833"
  "3333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",16404166.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2920\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2921, 1, 'NAD83(HARN) / Utah North (ft)', 'EPSG', 2921, 'PROJCS[\"NAD83(HARN) / Utah North (ft"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.794444444"
  "4444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",3280839.895,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2921\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2922, 1, 'NAD83(HARN) / Utah Central (ft)', 'EPSG', 2922, 'PROJCS[\"NAD83(HARN) / Utah Central"
  " (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",6561679.79,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2922\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2923, 1, 'NAD83(HARN) / Utah South (ft)', 'EPSG', 2923, 'PROJCS[\"NAD83(HARN) / Utah South (ft"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",9842519.685,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2923\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2924, 1, 'NAD83(HARN) / Virginia North (ftUS)', 'EPSG', 2924, 'PROJCS[\"NAD83(HARN) / Virginia"
  " North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "39.2111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2924\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2925, 1, 'NAD83(HARN) / Virginia South (ftUS)', 'EPSG', 2925, 'PROJCS[\"NAD83(HARN) / Virginia"
  " South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2925\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2926, 1, 'NAD83(HARN) / Washington North (ftUS)', 'EPSG', 2926, 'PROJCS[\"NAD83(HARN) / Washin"
  "gton North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-120.833333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.74444444"
  "44444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2926\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2927, 1, 'NAD83(HARN) / Washington South (ftUS)', 'EPSG', 2927, 'PROJCS[\"NAD83(HARN) / Washin"
  "gton South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",47.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.8333333333333,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2927\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2928, 1, 'NAD83(HARN) / Wisconsin North (ftUS)', 'EPSG', 2928, 'PROJCS[\"NAD83(HARN) / Wiscons"
  "in North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG"
  "\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longi"
  "tude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.766666666"
  "6667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false"
  " origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2928\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2929, 1, 'NAD83(HARN) / Wisconsin Central (ftUS)', 'EPSG', 2929, 'PROJCS[\"NAD83(HARN) / Wisco"
  "nsin Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.51111"
  "11111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORITY[\"E"
  "PSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2929\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2930, 1, 'NAD83(HARN) / Wisconsin South (ftUS)', 'EPSG', 2930, 'PROJCS[\"NAD83(HARN) / Wiscons"
  "in South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG"
  "\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,A"
  "UTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2930\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2931, 1, 'Beduaram / TM 13 NE', 'EPSG', 2931, 'PROJCS[\"Beduaram / TM 13 NE\",GEOGCS[\"Beduara"
  "m\",DATUM[\"Beduaram\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS"
  "84[-106,-87,188,0,0,0,0],AUTHORITY[\"EPSG\",\"6213\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4213\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2931\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2932, 1, 'QND95 / Qatar National Grid', 'EPSG', 2932, 'PROJCS[\"QND95 / Qatar National Grid\","
  "GEOGCS[\"QND95\",DATUM[\"Qatar National Datum 1995\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-119.4248,-303.65872,-11.00061,1.164298,0.174458,1.096259,3.657065],AUTHORITY[\"EPSG\",\"6614\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4614\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.4611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",51.2166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2932\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2933, 1, 'Segara / UTM zone 50S', 'EPSG', 2933, 'PROJCS[\"Segara / UTM zone 50S\",GEOGCS[\"Seg"
  "ara\",DATUM[\"Gunung Segara\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84["
  "-403,684,41,0,0,0,0],AUTHORITY[\"EPSG\",\"6613\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"461"
  "3\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2933\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2935, 1, 'Pulkovo 1942 / CS63 zone A1', 'EPSG', 2935, 'PROJCS[\"Pulkovo 1942 / CS63 zone A1\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.116666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",41.533333333"
  "3333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",1300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2935\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2936, 1, 'Pulkovo 1942 / CS63 zone A2', 'EPSG', 2936, 'PROJCS[\"Pulkovo 1942 / CS63 zone A2\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.116666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",44.533333333"
  "3333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",2300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2936\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2937, 1, 'Pulkovo 1942 / CS63 zone A3', 'EPSG', 2937, 'PROJCS[\"Pulkovo 1942 / CS63 zone A3\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.116666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",47.533333333"
  "3333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",3300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2937\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2938, 1, 'Pulkovo 1942 / CS63 zone A4', 'EPSG', 2938, 'PROJCS[\"Pulkovo 1942 / CS63 zone A4\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.116666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",50.533333333"
  "3333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",4300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2938\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2939, 1, 'Pulkovo 1942 / CS63 zone K2', 'EPSG', 2939, 'PROJCS[\"Pulkovo 1942 / CS63 zone K2\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.133333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",50.766666666"
  "6667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",2300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2939\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2940, 1, 'Pulkovo 1942 / CS63 zone K3', 'EPSG', 2940, 'PROJCS[\"Pulkovo 1942 / CS63 zone K3\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.133333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",53.766666666"
  "6667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",3300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2940\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2941, 1, 'Pulkovo 1942 / CS63 zone K4', 'EPSG', 2941, 'PROJCS[\"Pulkovo 1942 / CS63 zone K4\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.133333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",56.766666666"
  "6667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",4300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"2941\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2942, 1, 'Porto Santo / UTM zone 28N', 'EPSG', 2942, 'PROJCS[\"Porto Santo / UTM zone 28N\",GE"
  "OGCS[\"Porto Santo\",DATUM[\"Porto Santo 1936\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022"
  "\"]],TOWGS84[-499,-249,314,0,0,0,0],AUTHORITY[\"EPSG\",\"6615\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4615\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2942\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2943, 1, 'Selvagem Grande / UTM zone 28N', 'EPSG', 2943, 'PROJCS[\"Selvagem Grande / UTM zone "
  "28N\",GEOGCS[\"Selvagem Grande\",DATUM[\"Selvagem Grande\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-289,-124,60,0,0,0,0],AUTHORITY[\"EPSG\",\"6616\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4616\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2943\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2945, 1, 'NAD83(CSRS) / MTM zone 3', 'EPSG', 2945, 'PROJCS[\"NAD83(CSRS) / MTM zone 3\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-58.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2945\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2946, 1, 'NAD83(CSRS) / MTM zone 4', 'EPSG', 2946, 'PROJCS[\"NAD83(CSRS) / MTM zone 4\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-61.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2946\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2947, 1, 'NAD83(CSRS) / MTM zone 5', 'EPSG', 2947, 'PROJCS[\"NAD83(CSRS) / MTM zone 5\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-64.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2947\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2948, 1, 'NAD83(CSRS) / MTM zone 6', 'EPSG', 2948, 'PROJCS[\"NAD83(CSRS) / MTM zone 6\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-67.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2948\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2949, 1, 'NAD83(CSRS) / MTM zone 7', 'EPSG', 2949, 'PROJCS[\"NAD83(CSRS) / MTM zone 7\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2949\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2950, 1, 'NAD83(CSRS) / MTM zone 8', 'EPSG', 2950, 'PROJCS[\"NAD83(CSRS) / MTM zone 8\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-73.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2950\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2951, 1, 'NAD83(CSRS) / MTM zone 9', 'EPSG', 2951, 'PROJCS[\"NAD83(CSRS) / MTM zone 9\",GEOGCS"
  "[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUT"
  "HORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-76.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2951\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2952, 1, 'NAD83(CSRS) / MTM zone 10', 'EPSG', 2952, 'PROJCS[\"NAD83(CSRS) / MTM zone 10\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fact"
  "or at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"2952\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2953, 1, 'NAD83(CSRS) / New Brunswick Stereographic', 'EPSG', 2953, 'PROJCS[\"NAD83(CSRS) / Ne"
  "w Brunswick Stereographic\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.009650098"
  "96027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4617\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"9809\"]],PARAMETER[\"Latitude of natural or"
  "igin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-66.5111111111111,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999912,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",7500000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"2953\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2954, 1, 'NAD83(CSRS) / Prince Edward Isl. Stereographic (NAD83)', 'EPSG', 2954, 'PROJCS[\"NAD"
  "83(CSRS) / Prince Edward Isl. Stereographic (NAD83)\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference"
  " System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.02"
  "57899075194932,0.00965009896027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"9809\"]],PARAMETE"
  "R[\"Latitude of natural origin\",47.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999912,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY"
  "[\"EPSG\",\"2954\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2955, 1, 'NAD83(CSRS) / UTM zone 11N', 'EPSG', 2955, 'PROJCS[\"NAD83(CSRS) / UTM zone 11N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",E"
  "AST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2955\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2956, 1, 'NAD83(CSRS) / UTM zone 12N', 'EPSG', 2956, 'PROJCS[\"NAD83(CSRS) / UTM zone 12N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",E"
  "AST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2956\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2957, 1, 'NAD83(CSRS) / UTM zone 13N', 'EPSG', 2957, 'PROJCS[\"NAD83(CSRS) / UTM zone 13N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",E"
  "AST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2957\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2958, 1, 'NAD83(CSRS) / UTM zone 17N', 'EPSG', 2958, 'PROJCS[\"NAD83(CSRS) / UTM zone 17N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2958\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2959, 1, 'NAD83(CSRS) / UTM zone 18N', 'EPSG', 2959, 'PROJCS[\"NAD83(CSRS) / UTM zone 18N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2959\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2960, 1, 'NAD83(CSRS) / UTM zone 19N', 'EPSG', 2960, 'PROJCS[\"NAD83(CSRS) / UTM zone 19N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2960\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2961, 1, 'NAD83(CSRS) / UTM zone 20N', 'EPSG', 2961, 'PROJCS[\"NAD83(CSRS) / UTM zone 20N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2961\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2962, 1, 'NAD83(CSRS) / UTM zone 21N', 'EPSG', 2962, 'PROJCS[\"NAD83(CSRS) / UTM zone 21N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2962\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2963, 1, 'Lisbon 1890 (Lisbon) / Portugal Bonne', 'EPSG', 2963, 'PROJCS[\"Lisbon 1890 (Lisbon)"
  " / Portugal Bonne\",GEOGCS[\"Lisbon 1890 (Lisbon)\",DATUM[\"Lisbon 1890 (Lisbon)\",SPHEROID[\"Bessel 1841\",6377397.1"
  "55,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6904\"]],PRIMEM[\"Lisbon\",-9.131906111111112,AUTHO"
  "RITY[\"EPSG\",\"8902\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4904\"]],PROJECTION[\"Bonne (South Orientated)\",AUTHORITY[\"EPSG\",\"9828\"]],PAR"
  "AMETER[\"Latitude of natural origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"P\",SOUTH],AXIS[\""
  "M\",WEST],AUTHORITY[\"EPSG\",\"2963\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2964, 1, 'NAD27 / Alaska Albers', 'EPSG', 2964, 'PROJCS[\"NAD27 / Alaska Albers\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Albers Equal Area\",AUTHOR"
  "ITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",50,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude"
  " of false origin\",-154,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",55,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eastin"
  "g at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"2964\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2965, 1, 'NAD83 / Indiana East (ftUS)', 'EPSG', 2965, 'PROJCS[\"NAD83 / Indiana East (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.6666666666667,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",328083.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",820208.333,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"2965\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2966, 1, 'NAD83 / Indiana West (ftUS)', 'EPSG', 2966, 'PROJCS[\"NAD83 / Indiana West (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.0944444444444,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",820208.333,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"2966\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2967, 1, 'NAD83(HARN) / Indiana East (ftUS)', 'EPSG', 2967, 'PROJCS[\"NAD83(HARN) / Indiana Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-85.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",820208.333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2967\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2968, 1, 'NAD83(HARN) / Indiana West (ftUS)', 'EPSG', 2968, 'PROJCS[\"NAD83(HARN) / Indiana We"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-87.0944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",820208.333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2968\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2969, 1, 'Fort Marigot / UTM zone 20N', 'EPSG', 2969, 'PROJCS[\"Fort Marigot / UTM zone 20N\","
  "GEOGCS[\"Fort Marigot\",DATUM[\"Fort Marigot\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[137,248,-430,0,0,0,0],AUTHORITY[\"EPSG\",\"6621\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4621\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2969\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2970, 1, 'Guadeloupe 1948 / UTM zone 20N', 'EPSG', 2970, 'PROJCS[\"Guadeloupe 1948 / UTM zone "
  "20N\",GEOGCS[\"Guadeloupe 1948\",DATUM[\"Guadeloupe 1948\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-472.29,-5.63,-304.12,0.4362,-0.8374,0.2563,1.8984],AUTHORITY[\"EPSG\",\"6622\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4622\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "2970\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2971, 1, 'CSG67 / UTM zone 22N', 'EPSG', 2971, 'PROJCS[\"CSG67 / UTM zone 22N\",GEOGCS[\"CSG67"
  "\",DATUM[\"Centre Spatial Guyanais 1967\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-186,230,110,0,0,0,0],AUTHORITY[\"EPSG\",\"6623\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4623\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2971\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2972, 1, 'RGFG95 / UTM zone 22N', 'EPSG', 2972, 'PROJCS[\"RGFG95 / UTM zone 22N\",GEOGCS[\"RGF"
  "G95\",DATUM[\"Reseau Geodesique Francais Guyane 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6624\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4624\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2972\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2973, 1, 'Martinique 1938 / UTM zone 20N', 'EPSG', 2973, 'PROJCS[\"Martinique 1938 / UTM zone "
  "20N\",GEOGCS[\"Martinique 1938\",DATUM[\"Martinique 1938\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[126.93,547.94,130.41,-2.7867,5.1612,-0.8584,13.8227],AUTHORITY[\"EPSG\",\"6625\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4625\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "2973\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2975, 1, 'RGR92 / UTM zone 40S', 'EPSG', 2975, 'PROJCS[\"RGR92 / UTM zone 40S\",GEOGCS[\"RGR92"
  "\",DATUM[\"Reseau Geodesique de la Reunion 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6627\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4627\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2975\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2976, 1, 'Tahiti 52 / UTM zone 6S', 'EPSG', 2976, 'PROJCS[\"Tahiti 52 / UTM zone 6S\",GEOGCS[\""
  "Tahiti 52\",DATUM[\"Tahiti 52\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[162"
  ",117,154,0,0,0,0],AUTHORITY[\"EPSG\",\"6628\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4628\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2976\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2977, 1, 'Tahaa 54 / UTM zone 5S', 'EPSG', 2977, 'PROJCS[\"Tahaa 54 / UTM zone 5S\",GEOGCS[\"T"
  "ahaa 54\",DATUM[\"Tahaa 54\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[72.51,"
  "345.411,79.241,-1.5862,-0.8826,-0.5495,1.3653],AUTHORITY[\"EPSG\",\"6629\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4629\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2977\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2978, 1, 'IGN72 Nuku Hiva / UTM zone 7S', 'EPSG', 2978, 'PROJCS[\"IGN72 Nuku Hiva / UTM zone 7"
  "S\",GEOGCS[\"IGN72 Nuku Hiva\",DATUM[\"IGN72 Nuku Hiva\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG"
  "\",\"7022\"]],TOWGS84[84,274,65,0,0,0,0],AUTHORITY[\"EPSG\",\"6630\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4630\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2978\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2980, 1, 'Combani 1950 / UTM zone 38S', 'EPSG', 2980, 'PROJCS[\"Combani 1950 / UTM zone 38S\","
  "GEOGCS[\"Combani 1950\",DATUM[\"Combani 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-382,-59,-262,0,0,0,0],AUTHORITY[\"EPSG\",\"6632\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4632\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2980\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2981, 1, 'IGN56 Lifou / UTM zone 58S', 'EPSG', 2981, 'PROJCS[\"IGN56 Lifou / UTM zone 58S\",GE"
  "OGCS[\"IGN56 Lifou\",DATUM[\"IGN56 Lifou\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[137.092,131.66,91.475,-1.9436,-11.5993,-4.3321,-7.4824],AUTHORITY[\"EPSG\",\"6633\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4633\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2981\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2985, 1, 'Petrels 1972 / Terre Adelie Polar Stereographic', 'EPSG', 2985, 'PROJCS[\"Petrels 19"
  "72 / Terre Adelie Polar Stereographic\",GEOGCS[\"Petrels 1972\",DATUM[\"Petrels 1972\",SPHEROID[\"International 1924\""
  ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[365,194,166,0,0,0,0],AUTHORITY[\"EPSG\",\"6636\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4636\"]],PROJECTION[\"Polar Stereographic (variant C)\",AUTHORITY[\"E"
  "PSG\",\"9830\"]],PARAMETER[\"Latitude of standard parallel\",-67,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude "
  "of origin\",140,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2985\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2986, 1, 'Perroud 1950 / Terre Adelie Polar Stereographic', 'EPSG', 2986, 'PROJCS[\"Perroud 19"
  "50 / Terre Adelie Polar Stereographic\",GEOGCS[\"Perroud 1950\",DATUM[\"Pointe Geologie Perroud 1950\",SPHEROID[\"Int"
  "ernational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[325,154,172,0,0,0,0],AUTHORITY[\"EPSG\",\"6637\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4637\"]],PROJECTION[\"Polar Stereographic (variant C"
  ")\",AUTHORITY[\"EPSG\",\"9830\"]],PARAMETER[\"Latitude of standard parallel\",-67,AUTHORITY[\"EPSG\",\"8832\"]],PARAM"
  "ETER[\"Longitude of origin\",140,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Easting at false origin\",300000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2986\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2987, 1, 'Saint Pierre et Miquelon 1950 / UTM zone 21N', 'EPSG', 2987, 'PROJCS[\"Saint Pierre "
  "et Miquelon 1950 / UTM zone 21N\",GEOGCS[\"Saint Pierre et Miquelon 1950\",DATUM[\"Saint Pierre et Miquelon 1950\",SP"
  "HEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[11.363,424.148,373.13,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"6638\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4638\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",E"
  "AST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2987\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2988, 1, 'MOP78 / UTM zone 1S', 'EPSG', 2988, 'PROJCS[\"MOP78 / UTM zone 1S\",GEOGCS[\"MOP78\""
  ",DATUM[\"MOP78\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[253,-132,-127,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6639\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4639\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2988\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2991, 1, 'NAD83 / Oregon LCC (m)', 'EPSG', 2991, 'PROJCS[\"NAD83 / Oregon LCC (m)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of "
  "2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,A"
  "UTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2992, 1, 'NAD83 / Oregon GIC Lambert (ft)', 'EPSG', 2992, 'PROJCS[\"NAD83 / Oregon GIC Lambert"
  " (ft)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAM"
  "ETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",1312335.958,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"299"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2993, 1, 'NAD83(HARN) / Oregon LCC (m)', 'EPSG', 2993, 'PROJCS[\"NAD83(HARN) / Oregon LCC (m)\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120."
  "511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "2993\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2994, 1, 'NAD83(HARN) / Oregon GIC Lambert (ft)', 'EPSG', 2994, 'PROJCS[\"NAD83(HARN) / Oregon"
  " GIC Lambert (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fa"
  "lse origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",1312335.958,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"2994\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2995, 1, 'IGN53 Mare / UTM zone 58S', 'EPSG', 2995, 'PROJCS[\"IGN53 Mare / UTM zone 58S\",GEOG"
  "CS[\"IGN53 Mare\",DATUM[\"IGN53 Mare\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG"
  "S84[-408.809,366.856,-412.987,1.8842,-0.5308,2.1655,-121.0993],AUTHORITY[\"EPSG\",\"6641\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4641\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2995\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2996, 1, 'ST84 Ile des Pins / UTM zone 58S', 'EPSG', 2996, 'PROJCS[\"ST84 Ile des Pins / UTM z"
  "one 58S\",GEOGCS[\"ST84 Ile des Pins\",DATUM[\"ST84 Ile des Pins\",SPHEROID[\"International 1924\",6378388,297,AUTHOR"
  "ITY[\"EPSG\",\"7022\"]],TOWGS84[-13,-348,292,0,0,0,0],AUTHORITY[\"EPSG\",\"6642\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4642\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2996\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2997, 1, 'ST71 Belep / UTM zone 58S', 'EPSG', 2997, 'PROJCS[\"ST71 Belep / UTM zone 58S\",GEOG"
  "CS[\"ST71 Belep\",DATUM[\"ST71 Belep\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG"
  "S84[-480.26,-438.32,-643.429,16.3119,20.1721,-4.0349,-111.7002],AUTHORITY[\"EPSG\",\"6643\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4643\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2997\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2998, 1, 'NEA74 Noumea / UTM zone 58S', 'EPSG', 2998, 'PROJCS[\"NEA74 Noumea / UTM zone 58S\","
  "GEOGCS[\"NEA74 Noumea\",DATUM[\"NEA74 Noumea\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-10.18,-350.43,291.37,0,0,0,0],AUTHORITY[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4644\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2998\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (2999, 1, 'Grand Comoros / UTM zone 38S', 'EPSG', 2999, 'PROJCS[\"Grand Comoros / UTM zone 38S\""
  ",GEOGCS[\"Grand Comoros\",DATUM[\"Grand Comoros\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-963,510,-359,0,0,0,0],AUTHORITY[\"EPSG\",\"6646\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4646\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"2999\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3000, 1, 'Segara / NEIEZ', 'EPSG', 3000, 'PROJCS[\"Segara / NEIEZ\",GEOGCS[\"Segara\",DATUM[\""
  "Gunung Segara\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-403,684,41,0,"
  "0,0,0],AUTHORITY[\"EPSG\",\"6613\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292"
  "519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4613\"]],PROJECTI"
  "ON[\"Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG"
  "\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",110,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3900000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3000\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3001, 1, 'Batavia / NEIEZ', 'EPSG', 3001, 'PROJCS[\"Batavia / NEIEZ\",GEOGCS[\"Batavia\",DATUM"
  "[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-377.7,675.1,-52."
  "2,0,0,0,0],AUTHORITY[\"EPSG\",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4211\"]],PROJ"
  "ECTION[\"Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",110,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale facto"
  "r at natural origin\",0.997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3900000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3001\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3002, 1, 'Makassar / NEIEZ', 'EPSG', 3002, 'PROJCS[\"Makassar / NEIEZ\",GEOGCS[\"Makassar\",DA"
  "TUM[\"Makassar\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-587.8,519.75"
  ",145.76,0,0,0,0],AUTHORITY[\"EPSG\",\"6257\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4257\"]"
  "],PROJECTION[\"Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",110,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3900000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3002\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3003, 1, 'Monte Mario / Italy zone 1', 'EPSG', 3003, 'PROJCS[\"Monte Mario / Italy zone 1\",GE"
  "OGCS[\"Monte Mario\",DATUM[\"Monte Mario\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[-104.1,-49.1,-9.9,0.971,-2.917,0.714,-11.68],AUTHORITY[\"EPSG\",\"6265\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4265\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3003\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3004, 1, 'Monte Mario / Italy zone 2', 'EPSG', 3004, 'PROJCS[\"Monte Mario / Italy zone 2\",GE"
  "OGCS[\"Monte Mario\",DATUM[\"Monte Mario\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[-104.1,-49.1,-9.9,0.971,-2.917,0.714,-11.68],AUTHORITY[\"EPSG\",\"6265\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4265\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",2520000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3004\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3005, 1, 'NAD83 / BC Albers', 'EPSG', 3005, 'PROJCS[\"NAD83 / BC Albers\",GEOGCS[\"NAD83\",DAT"
  "UM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,"
  "1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],P"
  "ROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-126,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1"
  "st standard parallel\",50,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",58.5111111111"
  "111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3005\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3006, 1, 'SWEREF99 TM', 'EPSG', 3006, 'PROJCS[\"SWEREF99 TM\",GEOGCS[\"SWEREF99\",DATUM[\"SWER"
  "EF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "96,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"3006\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3007, 1, 'SWEREF99 12 00', 'EPSG', 3007, 'PROJCS[\"SWEREF99 12 00\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"3007\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3008, 1, 'SWEREF99 13 30', 'EPSG', 3008, 'PROJCS[\"SWEREF99 13 30\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",13.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\""
  "E\",EAST],AUTHORITY[\"EPSG\",\"3008\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3009, 1, 'SWEREF99 15 00', 'EPSG', 3009, 'PROJCS[\"SWEREF99 15 00\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"3009\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3010, 1, 'SWEREF99 16 30', 'EPSG', 3010, 'PROJCS[\"SWEREF99 16 30\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",16.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\""
  "E\",EAST],AUTHORITY[\"EPSG\",\"3010\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3011, 1, 'SWEREF99 18 00', 'EPSG', 3011, 'PROJCS[\"SWEREF99 18 00\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"3011\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3012, 1, 'SWEREF99 14 15', 'EPSG', 3012, 'PROJCS[\"SWEREF99 14 15\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",14.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\""
  "E\",EAST],AUTHORITY[\"EPSG\",\"3012\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3013, 1, 'SWEREF99 15 45', 'EPSG', 3013, 'PROJCS[\"SWEREF99 15 45\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",15.7611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NO"
  "RTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3013\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3014, 1, 'SWEREF99 17 15', 'EPSG', 3014, 'PROJCS[\"SWEREF99 17 15\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",17.2611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NO"
  "RTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3015, 1, 'SWEREF99 18 45', 'EPSG', 3015, 'PROJCS[\"SWEREF99 18 45\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",18.7611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NO"
  "RTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3015\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3016, 1, 'SWEREF99 20 15', 'EPSG', 3016, 'PROJCS[\"SWEREF99 20 15\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",20.2611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NO"
  "RTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3016\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3017, 1, 'SWEREF99 21 45', 'EPSG', 3017, 'PROJCS[\"SWEREF99 21 45\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",21.7611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NO"
  "RTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3017\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3018, 1, 'SWEREF99 23 15', 'EPSG', 3018, 'PROJCS[\"SWEREF99 23 15\",GEOGCS[\"SWEREF99\",DATUM["
  "\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",23.2611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NO"
  "RTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3018\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3019, 1, 'RT90 7.5 gon V', 'EPSG', 3019, 'PROJCS[\"RT90 7.5 gon V\",GEOGCS[\"RT90\",DATUM[\"Ri"
  "kets koordinatsystem 1990\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[41"
  "4.1,41.3,603.1,-0.855,2.141,-7.023,0],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4124\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11.3082777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3019\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3020, 1, 'RT90 5 gon V', 'EPSG', 3020, 'PROJCS[\"RT90 5 gon V\",GEOGCS[\"RT90\",DATUM[\"Rikets"
  " koordinatsystem 1990\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[414.1,"
  "41.3,603.1,-0.855,2.141,-7.023,0],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4124\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13.5582777777778,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1"
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3020\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3021, 1, 'RT90 2.5 gon V', 'EPSG', 3021, 'PROJCS[\"RT90 2.5 gon V\",GEOGCS[\"RT90\",DATUM[\"Ri"
  "kets koordinatsystem 1990\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[41"
  "4.1,41.3,603.1,-0.855,2.141,-7.023,0],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4124\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15.8082777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3021\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3022, 1, 'RT90 0 gon', 'EPSG', 3022, 'PROJCS[\"RT90 0 gon\",GEOGCS[\"RT90\",DATUM[\"Rikets koo"
  "rdinatsystem 1990\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[414.1,41.3"
  ",603.1,-0.855,2.141,-7.023,0],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4124\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18.0582777777778,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3022\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3023, 1, 'RT90 2.5 gon O', 'EPSG', 3023, 'PROJCS[\"RT90 2.5 gon O\",GEOGCS[\"RT90\",DATUM[\"Ri"
  "kets koordinatsystem 1990\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[41"
  "4.1,41.3,603.1,-0.855,2.141,-7.023,0],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4124\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20.3082777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3023\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3024, 1, 'RT90 5 gon O', 'EPSG', 3024, 'PROJCS[\"RT90 5 gon O\",GEOGCS[\"RT90\",DATUM[\"Rikets"
  " koordinatsystem 1990\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[414.1,"
  "41.3,603.1,-0.855,2.141,-7.023,0],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4124\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22.5582777777778,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1"
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3024\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3025, 1, 'RT38 7.5 gon V', 'EPSG', 3025, 'PROJCS[\"RT38 7.5 gon V\",GEOGCS[\"RT38\",DATUM[\"St"
  "ockholm 1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"63"
  "08\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",11.3082777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\","
  "EAST],AUTHORITY[\"EPSG\",\"3025\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3026, 1, 'RT38 5 gon V', 'EPSG', 3026, 'PROJCS[\"RT38 5 gon V\",GEOGCS[\"RT38\",DATUM[\"Stockh"
  "olm 1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6308\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",13.5582777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3026\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3027, 1, 'RT38 2.5 gon V', 'EPSG', 3027, 'PROJCS[\"RT38 2.5 gon V\",GEOGCS[\"RT38\",DATUM[\"St"
  "ockholm 1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"63"
  "08\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",15.8082777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\","
  "EAST],AUTHORITY[\"EPSG\",\"3027\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3028, 1, 'RT38 0 gon', 'EPSG', 3028, 'PROJCS[\"RT38 0 gon\",GEOGCS[\"RT38\",DATUM[\"Stockholm "
  "1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6308\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",18.0582777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AU"
  "THORITY[\"EPSG\",\"3028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3029, 1, 'RT38 2.5 gon O', 'EPSG', 3029, 'PROJCS[\"RT38 2.5 gon O\",GEOGCS[\"RT38\",DATUM[\"St"
  "ockholm 1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"63"
  "08\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",20.3082777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\","
  "EAST],AUTHORITY[\"EPSG\",\"3029\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3030, 1, 'RT38 5 gon O', 'EPSG', 3030, 'PROJCS[\"RT38 5 gon O\",GEOGCS[\"RT38\",DATUM[\"Stockh"
  "olm 1938\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6308\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",22.5582777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3030\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3031, 1, 'WGS 84 / Antarctic Polar Stereographic', 'EPSG', 3031, 'PROJCS[\"WGS 84 / Antarctic "
  "Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.25722356"
  "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of"
  " standard parallel\",-71,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",0,AUTHORITY[\"EPSG\",\"8833\""
  "]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3031\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3032, 1, 'WGS 84 / Australian Antarctic Polar Stereographic', 'EPSG', 3032, 'PROJCS[\"WGS 84 /"
  " Australian Antarctic Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\""
  ",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PAR"
  "AMETER[\"Latitude of standard parallel\",-71,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",70,AUTHO"
  "RITY[\"EPSG\",\"8833\"]],PARAMETER[\"False easting\",6000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",6000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\","
  "NORTH],AUTHORITY[\"EPSG\",\"3032\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3033, 1, 'WGS 84 / Australian Antarctic Lambert', 'EPSG', 3033, 'PROJCS[\"WGS 84 / Australian "
  "Antarctic Lambert\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",-50,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",70,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",60"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3033\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3034, 1, 'ETRS89-extended / LCC Europe', 'EPSG', 3034, 'PROJCS[\"ETRS89-extended / LCC Europe\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",52,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2800000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3034\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3035, 1, 'ETRS89-extended / LAEA Europe', 'EPSG', 3035, 'PROJCS[\"ETRS89-extended / LAEA Europ"
  "e\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PAR"
  "AMETER[\"Latitude of natural origin\",52,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",4321000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",3210000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\""
  "X\",EAST],AUTHORITY[\"EPSG\",\"3035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3036, 1, 'Moznet / UTM zone 36S', 'EPSG', 3036, 'PROJCS[\"Moznet / UTM zone 36S\",GEOGCS[\"Moz"
  "net\",DATUM[\"Moznet (ITRF94)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6130\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4130\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]]"
  ",AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3036\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3037, 1, 'Moznet / UTM zone 37S', 'EPSG', 3037, 'PROJCS[\"Moznet / UTM zone 37S\",GEOGCS[\"Moz"
  "net\",DATUM[\"Moznet (ITRF94)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6130\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4130\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]]"
  ",AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3037\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3040, 1, 'ETRS89 / UTM zone 28N (N-E)', 'EPSG', 3040, 'PROJCS[\"ETRS89 / UTM zone 28N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3040\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3041, 1, 'ETRS89 / UTM zone 29N (N-E)', 'EPSG', 3041, 'PROJCS[\"ETRS89 / UTM zone 29N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3041\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3042, 1, 'ETRS89 / UTM zone 30N (N-E)', 'EPSG', 3042, 'PROJCS[\"ETRS89 / UTM zone 30N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3042\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3043, 1, 'ETRS89 / UTM zone 31N (N-E)', 'EPSG', 3043, 'PROJCS[\"ETRS89 / UTM zone 31N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3043\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3044, 1, 'ETRS89 / UTM zone 32N (N-E)', 'EPSG', 3044, 'PROJCS[\"ETRS89 / UTM zone 32N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3044\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3045, 1, 'ETRS89 / UTM zone 33N (N-E)', 'EPSG', 3045, 'PROJCS[\"ETRS89 / UTM zone 33N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3045\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3046, 1, 'ETRS89 / UTM zone 34N (N-E)', 'EPSG', 3046, 'PROJCS[\"ETRS89 / UTM zone 34N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3046\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3047, 1, 'ETRS89 / UTM zone 35N (N-E)', 'EPSG', 3047, 'PROJCS[\"ETRS89 / UTM zone 35N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3047\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3048, 1, 'ETRS89 / UTM zone 36N (N-E)', 'EPSG', 3048, 'PROJCS[\"ETRS89 / UTM zone 36N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3048\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3049, 1, 'ETRS89 / UTM zone 37N (N-E)', 'EPSG', 3049, 'PROJCS[\"ETRS89 / UTM zone 37N (N-E)\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3049\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3052, 1, 'Reykjavik 1900 / Lambert 1900', 'EPSG', 3052, 'PROJCS[\"Reykjavik 1900 / Lambert 190"
  "0\",GEOGCS[\"Reykjavik 1900\",DATUM[\"Reykjavik 1900\",SPHEROID[\"Danish 1876\",6377019.27,300,AUTHORITY[\"EPSG\",\"7"
  "051\"]],TOWGS84[-28,199,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6657\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4657\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORITY[\"EPSG\",\"9826\"]],PARAMETER["
  "\"Latitude of natural origin\",65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-19.022125,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"W\",WEST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3052\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3053, 1, 'Hjorsey 1955 / Lambert 1955', 'EPSG', 3053, 'PROJCS[\"Hjorsey 1955 / Lambert 1955\","
  "GEOGCS[\"Hjorsey 1955\",DATUM[\"Hjorsey 1955\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-73,47,-83,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4658\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORITY[\"EPSG\",\"9826\"]],PARAMETER[\"La"
  "titude of natural origin\",65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-18,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"W\",WEST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3053\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3054, 1, 'Hjorsey 1955 / UTM zone 26N', 'EPSG', 3054, 'PROJCS[\"Hjorsey 1955 / UTM zone 26N\","
  "GEOGCS[\"Hjorsey 1955\",DATUM[\"Hjorsey 1955\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-73,47,-83,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4658\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3054\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3055, 1, 'Hjorsey 1955 / UTM zone 27N', 'EPSG', 3055, 'PROJCS[\"Hjorsey 1955 / UTM zone 27N\","
  "GEOGCS[\"Hjorsey 1955\",DATUM[\"Hjorsey 1955\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-73,47,-83,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4658\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3055\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3056, 1, 'Hjorsey 1955 / UTM zone 28N', 'EPSG', 3056, 'PROJCS[\"Hjorsey 1955 / UTM zone 28N\","
  "GEOGCS[\"Hjorsey 1955\",DATUM[\"Hjorsey 1955\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-73,47,-83,0,0,0,0],AUTHORITY[\"EPSG\",\"6658\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4658\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3056\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3057, 1, 'ISN93 / Lambert 1993', 'EPSG', 3057, 'PROJCS[\"ISN93 / Lambert 1993\",GEOGCS[\"ISN93"
  "\",DATUM[\"Islands Net 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6659\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4659\"]],PROJE"
  "CTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",65,AUTHO"
  "RITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-19,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitu"
  "de of 1st standard parallel\",64.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",65."
  "75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3058, 1, 'Helle 1954 / Jan Mayen Grid', 'EPSG', 3058, 'PROJCS[\"Helle 1954 / Jan Mayen Grid\","
  "GEOGCS[\"Helle 1954\",DATUM[\"Helle 1954\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[982.6087,552.753,-540.873,6.68162662527695,-31.6114924086423,-19.8481610048169,16.805],AUTHORITY[\"EPSG\",\"6"
  "660\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4660\"]],PROJECTION[\"Transverse Mercator\",AU"
  "THORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",-8.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-78"
  "00000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST]"
  ",AUTHORITY[\"EPSG\",\"3058\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3059, 1, 'LKS92 / Latvia TM', 'EPSG', 3059, 'PROJCS[\"LKS92 / Latvia TM\",GEOGCS[\"LKS92\",DAT"
  "UM[\"Latvia 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"6661\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4661\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",-6000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3059\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3060, 1, 'IGN72 Grande Terre / UTM zone 58S', 'EPSG', 3060, 'PROJCS[\"IGN72 Grande Terre / UTM"
  " zone 58S\",GEOGCS[\"IGN72 Grande Terre\",DATUM[\"IGN72 Grande Terre\",SPHEROID[\"International 1924\",6378388,297,AU"
  "THORITY[\"EPSG\",\"7022\"]],TOWGS84[97.295,-263.247,310.882,-1.5999,0.8386,3.1409,13.3259],AUTHORITY[\"EPSG\",\"6634\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4662\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100"
  "00000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3060\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3061, 1, 'Porto Santo 1995 / UTM zone 28N', 'EPSG', 3061, 'PROJCS[\"Porto Santo 1995 / UTM zon"
  "e 28N\",GEOGCS[\"Porto Santo 1995\",DATUM[\"Porto Santo 1995\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY["
  "\"EPSG\",\"7022\"]],TOWGS84[-210.502,-66.902,-48.476,2.094,-15.067,-5.817,0.485],AUTHORITY[\"EPSG\",\"6663\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4663\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPS"
  "G\",\"3061\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3062, 1, 'Azores Oriental 1995 / UTM zone 26N', 'EPSG', 3062, 'PROJCS[\"Azores Oriental 1995 /"
  " UTM zone 26N\",GEOGCS[\"Azores Oriental 1995\",DATUM[\"Azores Oriental Islands 1995\",SPHEROID[\"International 1924\""
  ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-208.719,129.685,52.092,-0.195,-0.014,0.327,0.198],AUTHORITY[\"EPS"
  "G\",\"6664\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4664\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999"
  "6,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3062\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3063, 1, 'Azores Central 1995 / UTM zone 26N', 'EPSG', 3063, 'PROJCS[\"Azores Central 1995 / U"
  "TM zone 26N\",GEOGCS[\"Azores Central 1995\",DATUM[\"Azores Central Islands 1995\",SPHEROID[\"International 1924\",63"
  "78388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-103.088,162.481,-28.276,0.167,0.082,0.168,-1.504],AUTHORITY[\"EPSG\""
  ",\"6665\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4665\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3063\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3064, 1, 'IGM95 / UTM zone 32N', 'EPSG', 3064, 'PROJCS[\"IGM95 / UTM zone 32N\",GEOGCS[\"IGM95"
  "\",DATUM[\"Istituto Geografico Militaire 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6670\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4670\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3064\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3065, 1, 'IGM95 / UTM zone 33N', 'EPSG', 3065, 'PROJCS[\"IGM95 / UTM zone 33N\",GEOGCS[\"IGM95"
  "\",DATUM[\"Istituto Geografico Militaire 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6670\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4670\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3066, 1, 'ED50 / Jordan TM', 'EPSG', 3066, 'PROJCS[\"ED50 / Jordan TM\",GEOGCS[\"ED50\",DATUM["
  "\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17"
  ".16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",37,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-3000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3067, 1, 'ETRS89 / TM35FIN(E,N)', 'EPSG', 3067, 'PROJCS[\"ETRS89 / TM35FIN(E,N)\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3067\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3068, 1, 'DHDN / Soldner Berlin', 'EPSG', 3068, 'PROJCS[\"DHDN / Soldner Berlin\",GEOGCS[\"DHD"
  "N\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\""
  "]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitud"
  "e of natural origin\",52.4186482777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13.627"
  "2036666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",10000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH]"
  ",AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"3068\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3069, 1, 'NAD27 / Wisconsin Transverse Mercator', 'EPSG', 3069, 'PROJCS[\"NAD27 / Wisconsin Tr"
  "ansverse Mercator\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.97869"
  "82138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8."
  "1049164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJEC"
  "TION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPS"
  "G\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",-4500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3069\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3070, 1, 'NAD83 / Wisconsin Transverse Mercator', 'EPSG', 3070, 'PROJCS[\"NAD83 / Wisconsin Tr"
  "ansverse Mercator\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",520000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4480000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3070\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3071, 1, 'NAD83(HARN) / Wisconsin Transverse Mercator', 'EPSG', 3071, 'PROJCS[\"NAD83(HARN) / "
  "Wisconsin Transverse Mercator\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"G"
  "RS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",520000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4480000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"3071\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3072, 1, 'NAD83 / Maine CS2000 East', 'EPSG', 3072, 'PROJCS[\"NAD83 / Maine CS2000 East\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",43.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-67.875,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3072\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3074, 1, 'NAD83 / Maine CS2000 West', 'EPSG', 3074, 'PROJCS[\"NAD83 / Maine CS2000 West\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.375,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3074\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3075, 1, 'NAD83(HARN) / Maine CS2000 East', 'EPSG', 3075, 'PROJCS[\"NAD83(HARN) / Maine CS2000"
  " East\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-67.875,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"3075\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3077, 1, 'NAD83(HARN) / Maine CS2000 West', 'EPSG', 3077, 'PROJCS[\"NAD83(HARN) / Maine CS2000"
  " West\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-70.375,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"3077\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3078, 1, 'NAD83 / Michigan Oblique Mercator', 'EPSG', 3078, 'PROJCS[\"NAD83 / Michigan Oblique"
  " Mercator\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAM"
  "ETER[\"Latitude of projection centre\",45.3091666666667,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projec"
  "tion centre\",-86,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",337.25556,AUTHORITY[\"EPSG\",\""
  "8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",337.25556,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale fa"
  "ctor on initial line\",0.9996,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",2546731.496,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",-4354009.816,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3078\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3079, 1, 'NAD83(HARN) / Michigan Oblique Mercator', 'EPSG', 3079, 'PROJCS[\"NAD83(HARN) / Mich"
  "igan Oblique Mercator\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHO"
  "RITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",45.3091666666667,AUTHORITY[\"EPSG\",\"8811\"]],P"
  "ARAMETER[\"Longitude of projection centre\",-86,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",3"
  "37.25556,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",337.25556,AUTHORITY[\"EPSG\",\""
  "8814\"]],PARAMETER[\"Scale factor on initial line\",0.9996,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\","
  "2546731.496,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4354009.816,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3079\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3080, 1, 'NAD27 / Shackleford', 'EPSG', 3080, 'PROJCS[\"NAD27 / Shackleford\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard par"
  "allel\",27.4166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.92777777777"
  "78,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETE"
  "R[\"Northing at false origin\",3000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3080\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3081, 1, 'NAD83 / Texas State Mapping System', 'EPSG', 3081, 'PROJCS[\"NAD83 / Texas State Map"
  "ping System\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER"
  "[\"Latitude of false origin\",31.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\","
  "-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.4166666666667,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.9277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"3081\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3082, 1, 'NAD83 / Texas Centric Lambert Conformal', 'EPSG', 3082, 'PROJCS[\"NAD83 / Texas Cent"
  "ric Lambert Conformal\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.5,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",15"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3082\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3083, 1, 'NAD83 / Texas Centric Albers Equal Area', 'EPSG', 3083, 'PROJCS[\"NAD83 / Texas Cent"
  "ric Albers Equal Area\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\""
  "Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1500000,AUTHOR"
  "ITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3083\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3084, 1, 'NAD83(HARN) / Texas Centric Lambert Conformal', 'EPSG', 3084, 'PROJCS[\"NAD83(HARN) "
  "/ Texas Centric Lambert Conformal\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID"
  "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\""
  ",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.5,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER"
  "[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000000,AUT"
  "HORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"3084\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3085, 1, 'NAD83(HARN) / Texas Centric Albers Equal Area', 'EPSG', 3085, 'PROJCS[\"NAD83(HARN) "
  "/ Texas Centric Albers Equal Area\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID"
  "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY["
  "\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.5,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6000000,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"3085\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3086, 1, 'NAD83 / Florida GDL Albers', 'EPSG', 3086, 'PROJCS[\"NAD83 / Florida GDL Albers\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\""
  ",24,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER"
  "[\"Latitude of 1st standard parallel\",24,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel"
  "\",31.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3086\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3087, 1, 'NAD83(HARN) / Florida GDL Albers', 'EPSG', 3087, 'PROJCS[\"NAD83(HARN) / Florida GDL"
  " Albers\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAME"
  "TER[\"Latitude of false origin\",24,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",24,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",31.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AU"
  "THORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3087\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3088, 1, 'NAD83 / Kentucky Single Zone', 'EPSG', 3088, 'PROJCS[\"NAD83 / Kentucky Single Zone\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-85.75,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"3088\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3089, 1, 'NAD83 / Kentucky Single Zone (ftUS)', 'EPSG', 3089, 'PROJCS[\"NAD83 / Kentucky Singl"
  "e Zone (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMET"
  "ER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.0944444444444,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280833."
  "333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3089\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3090, 1, 'NAD83(HARN) / Kentucky Single Zone', 'EPSG', 3090, 'PROJCS[\"NAD83(HARN) / Kentucky "
  "Single Zone\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.09444444444"
  "44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3090\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3091, 1, 'NAD83(HARN) / Kentucky Single Zone (ftUS)', 'EPSG', 3091, 'PROJCS[\"NAD83(HARN) / Ke"
  "ntucky Single Zone (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMET"
  "ER[\"Longitude of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",37.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3091\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3092, 1, 'Tokyo / UTM zone 51N', 'EPSG', 3092, 'PROJCS[\"Tokyo / UTM zone 51N\",GEOGCS[\"Tokyo"
  "\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,68"
  "7,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3092\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3093, 1, 'Tokyo / UTM zone 52N', 'EPSG', 3093, 'PROJCS[\"Tokyo / UTM zone 52N\",GEOGCS[\"Tokyo"
  "\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,68"
  "7,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3093\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3094, 1, 'Tokyo / UTM zone 53N', 'EPSG', 3094, 'PROJCS[\"Tokyo / UTM zone 53N\",GEOGCS[\"Tokyo"
  "\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,68"
  "7,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3094\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3095, 1, 'Tokyo / UTM zone 54N', 'EPSG', 3095, 'PROJCS[\"Tokyo / UTM zone 54N\",GEOGCS[\"Tokyo"
  "\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,68"
  "7,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3096, 1, 'Tokyo / UTM zone 55N', 'EPSG', 3096, 'PROJCS[\"Tokyo / UTM zone 55N\",GEOGCS[\"Tokyo"
  "\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,68"
  "7,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3096\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3097, 1, 'JGD2000 / UTM zone 51N', 'EPSG', 3097, 'PROJCS[\"JGD2000 / UTM zone 51N\",GEOGCS[\"J"
  "GD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3097\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3098, 1, 'JGD2000 / UTM zone 52N', 'EPSG', 3098, 'PROJCS[\"JGD2000 / UTM zone 52N\",GEOGCS[\"J"
  "GD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3098\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3099, 1, 'JGD2000 / UTM zone 53N', 'EPSG', 3099, 'PROJCS[\"JGD2000 / UTM zone 53N\",GEOGCS[\"J"
  "GD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3099\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3100, 1, 'JGD2000 / UTM zone 54N', 'EPSG', 3100, 'PROJCS[\"JGD2000 / UTM zone 54N\",GEOGCS[\"J"
  "GD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3100\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3101, 1, 'JGD2000 / UTM zone 55N', 'EPSG', 3101, 'PROJCS[\"JGD2000 / UTM zone 55N\",GEOGCS[\"J"
  "GD2000\",DATUM[\"Japanese Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4612\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3101\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3102, 1, 'American Samoa 1962 / American Samoa Lambert', 'EPSG', 3102, 'PROJCS[\"American Samo"
  "a 1962 / American Samoa Lambert\",GEOGCS[\"American Samoa 1962\",DATUM[\"American Samoa 1962\",SPHEROID[\"Clarke 1866"
  "\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115,118,426,0,0,0,0],AUTHORITY[\"EPSG\",\"6169\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4169\"]],PROJECTION[\"Lambert Conic Conformal (1SP)"
  "\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",-14.2666666666667,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-170,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",312234.65,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EP"
  "SG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3102\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3106, 1, 'Gulshan 303 / Bangladesh Transverse Mercator', 'EPSG', 3106, 'PROJCS[\"Gulshan 303 /"
  " Bangladesh Transverse Mercator\",GEOGCS[\"Gulshan 303\",DATUM[\"Gulshan 303\",SPHEROID[\"Everest 1830 (1937 Adjustme"
  "nt)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[283.7,735.9,261.1,0,0,0,0],AUTHORITY[\"EPSG\",\"6682"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4682\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"3106\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3107, 1, 'GDA94 / SA Lambert', 'EPSG', 3107, 'PROJCS[\"GDA94 / SA Lambert\",GEOGCS[\"GDA94\",D"
  "ATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4283\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",-32,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",135,AUTHORITY[\"EPSG\",\"8822\"]],PAR"
  "AMETER[\"Latitude of 1st standard parallel\",-28,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard p"
  "arallel\",-36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3107\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3108, 1, 'ETRS89 / Guernsey Grid', 'EPSG', 3108, 'PROJCS[\"ETRS89 / Guernsey Grid\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",49.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-2.4166666666"
  "6667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999997,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",47000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",50000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "3108\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3109, 1, 'ETRS89 / Jersey Transverse Mercator', 'EPSG', 3109, 'PROJCS[\"ETRS89 / Jersey Transv"
  "erse Mercator\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",49.225,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-2.135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999999,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",70000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPS"
  "G\",\"3109\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3110, 1, 'AGD66 / Vicgrid66', 'EPSG', 3110, 'PROJCS[\"AGD66 / Vicgrid66\",GEOGCS[\"AGD66\",DAT"
  "UM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7"
  "003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",-37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false orig"
  "in\",145,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-36,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",-38,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3110\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3111, 1, 'GDA94 / Vicgrid', 'EPSG', 3111, 'PROJCS[\"GDA94 / Vicgrid\",GEOGCS[\"GDA94\",DATUM[\""
  "Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]"
  "],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-"
  "37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",145,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",-36,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel"
  "\",-38,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3111\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3112, 1, 'GDA94 / Geoscience Australia Lambert', 'EPSG', 3112, 'PROJCS[\"GDA94 / Geoscience Au"
  "stralia Lambert\",GEOGCS[\"GDA94\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",134"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-18,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",-36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3112\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3113, 1, 'GDA94 / BCSG02', 'EPSG', 3113, 'PROJCS[\"GDA94 / BCSG02\",GEOGCS[\"GDA94\",DATUM[\"G"
  "eocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84"
  "[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-28,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3113\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3114, 1, 'MAGNA-SIRGAS / Colombia Far West zone', 'EPSG', 3114, 'PROJCS[\"MAGNA-SIRGAS / Colom"
  "bia Far West zone\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",4.59620041666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-80.0775079166666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\""
  "E\",EAST],AUTHORITY[\"EPSG\",\"3114\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3115, 1, 'MAGNA-SIRGAS / Colombia West zone', 'EPSG', 3115, 'PROJCS[\"MAGNA-SIRGAS / Colombia "
  "West zone\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",4.59620041666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-77.0775079166666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1"
  "000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST"
  "],AUTHORITY[\"EPSG\",\"3115\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3116, 1, 'MAGNA-SIRGAS / Colombia Bogota zone', 'EPSG', 3116, 'PROJCS[\"MAGNA-SIRGAS / Colombi"
  "a Bogota zone\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",4.59620041666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-74.0775079166666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EA"
  "ST],AUTHORITY[\"EPSG\",\"3116\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3117, 1, 'MAGNA-SIRGAS / Colombia East Central zone', 'EPSG', 3117, 'PROJCS[\"MAGNA-SIRGAS / C"
  "olombia East Central zone\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS"
  " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",4.59620041666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",-71.0775079166666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH]"
  ",AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3117\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3118, 1, 'MAGNA-SIRGAS / Colombia East zone', 'EPSG', 3118, 'PROJCS[\"MAGNA-SIRGAS / Colombia "
  "East zone\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",4.59620041666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-68.0775079166666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1"
  "000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST"
  "],AUTHORITY[\"EPSG\",\"3118\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3119, 1, 'Douala 1948 / AEF west', 'EPSG', 3119, 'PROJCS[\"Douala 1948 / AEF west\",GEOGCS[\"D"
  "ouala 1948\",DATUM[\"Douala 1948\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84["
  "-206.1,-174.7,-87.7,0,0,0,0],AUTHORITY[\"EPSG\",\"6192\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4192\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3120, 1, 'Pulkovo 1942(58) / Poland zone I', 'EPSG', 3120, 'PROJCS[\"Pulkovo 1942(58) / Poland"
  " zone I\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\""
  "9809\"]],PARAMETER[\"Latitude of natural origin\",50.625,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",21.0833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9998,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4637000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",5467000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EA"
  "ST],AUTHORITY[\"EPSG\",\"3120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3121, 1, 'PRS92 / Philippines zone 1', 'EPSG', 3121, 'PROJCS[\"PRS92 / Philippines zone 1\",GE"
  "OGCS[\"PRS92\",DATUM[\"Philippine Reference System 1992\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4683\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3122, 1, 'PRS92 / Philippines zone 2', 'EPSG', 3122, 'PROJCS[\"PRS92 / Philippines zone 2\",GE"
  "OGCS[\"PRS92\",DATUM[\"Philippine Reference System 1992\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4683\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3123, 1, 'PRS92 / Philippines zone 3', 'EPSG', 3123, 'PROJCS[\"PRS92 / Philippines zone 3\",GE"
  "OGCS[\"PRS92\",DATUM[\"Philippine Reference System 1992\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4683\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3123\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3124, 1, 'PRS92 / Philippines zone 4', 'EPSG', 3124, 'PROJCS[\"PRS92 / Philippines zone 4\",GE"
  "OGCS[\"PRS92\",DATUM[\"Philippine Reference System 1992\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4683\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3125, 1, 'PRS92 / Philippines zone 5', 'EPSG', 3125, 'PROJCS[\"PRS92 / Philippines zone 5\",GE"
  "OGCS[\"PRS92\",DATUM[\"Philippine Reference System 1992\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHOR"
  "ITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.068,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4683\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3125\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3126, 1, 'ETRS89 / ETRS-GK19FIN', 'EPSG', 3126, 'PROJCS[\"ETRS89 / ETRS-GK19FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3126\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3127, 1, 'ETRS89 / ETRS-GK20FIN', 'EPSG', 3127, 'PROJCS[\"ETRS89 / ETRS-GK20FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3128, 1, 'ETRS89 / ETRS-GK21FIN', 'EPSG', 3128, 'PROJCS[\"ETRS89 / ETRS-GK21FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3129, 1, 'ETRS89 / ETRS-GK22FIN', 'EPSG', 3129, 'PROJCS[\"ETRS89 / ETRS-GK22FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3129\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3130, 1, 'ETRS89 / ETRS-GK23FIN', 'EPSG', 3130, 'PROJCS[\"ETRS89 / ETRS-GK23FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3130\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3131, 1, 'ETRS89 / ETRS-GK24FIN', 'EPSG', 3131, 'PROJCS[\"ETRS89 / ETRS-GK24FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3131\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3132, 1, 'ETRS89 / ETRS-GK25FIN', 'EPSG', 3132, 'PROJCS[\"ETRS89 / ETRS-GK25FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3132\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3133, 1, 'ETRS89 / ETRS-GK26FIN', 'EPSG', 3133, 'PROJCS[\"ETRS89 / ETRS-GK26FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3133\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3134, 1, 'ETRS89 / ETRS-GK27FIN', 'EPSG', 3134, 'PROJCS[\"ETRS89 / ETRS-GK27FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3134\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3135, 1, 'ETRS89 / ETRS-GK28FIN', 'EPSG', 3135, 'PROJCS[\"ETRS89 / ETRS-GK28FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3135\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3136, 1, 'ETRS89 / ETRS-GK29FIN', 'EPSG', 3136, 'PROJCS[\"ETRS89 / ETRS-GK29FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",29,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3137, 1, 'ETRS89 / ETRS-GK30FIN', 'EPSG', 3137, 'PROJCS[\"ETRS89 / ETRS-GK30FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3137\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3138, 1, 'ETRS89 / ETRS-GK31FIN', 'EPSG', 3138, 'PROJCS[\"ETRS89 / ETRS-GK31FIN\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3138\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3139, 1, 'Vanua Levu 1915 / Vanua Levu Grid', 'EPSG', 3139, 'PROJCS[\"Vanua Levu 1915 / Vanua "
  "Levu Grid\",GEOGCS[\"Vanua Levu 1915\",DATUM[\"Vanua Levu 1915\",SPHEROID[\"Clarke 1880 (international foot)\",637830"
  "6.3696,293.46630765562986,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84[51,391,-36,0,0,0,0],AUTHORITY[\"EPSG\",\"6748\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4748\"]],PROJECTION[\"Hyperbolic Cassini-Soldner\",AUTHOR"
  "ITY[\"EPSG\",\"9833\"]],PARAMETER[\"Latitude of natural origin\",-16.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",179.344444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1251"
  "331.8,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1662888.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"link\""
  ",0.201168,AUTHORITY[\"EPSG\",\"9098\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3139\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3140, 1, 'Viti Levu 1912 / Viti Levu Grid', 'EPSG', 3140, 'PROJCS[\"Viti Levu 1912 / Viti Levu"
  " Grid\",GEOGCS[\"Viti Levu 1912\",DATUM[\"Viti Levu 1912\",SPHEROID[\"Clarke 1880 (international foot)\",6378306.3696"
  ",293.46630765562986,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84[98,390,-22,0,0,0,0],AUTHORITY[\"EPSG\",\"6752\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4752\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"980"
  "6\"]],PARAMETER[\"Latitude of natural origin\",-18,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",178,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",544000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",704000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"link\",0.201168,AUTHORITY[\"EPSG\",\"9098\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3140\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3141, 1, 'Fiji 1956 / UTM zone 60S', 'EPSG', 3141, 'PROJCS[\"Fiji 1956 / UTM zone 60S\",GEOGCS"
  "[\"Fiji 1956\",DATUM[\"Fiji 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84["
  "265.025,384.929,-194.046,0,0,0,0],AUTHORITY[\"EPSG\",\"6721\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4721\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3141\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3142, 1, 'Fiji 1956 / UTM zone 1S', 'EPSG', 3142, 'PROJCS[\"Fiji 1956 / UTM zone 1S\",GEOGCS[\""
  "Fiji 1956\",DATUM[\"Fiji 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[265"
  ".025,384.929,-194.046,0,0,0,0],AUTHORITY[\"EPSG\",\"6721\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4721\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3142\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3144, 1, 'FD54 / Faroe Lambert', 'EPSG', 3144, 'PROJCS[\"FD54 / Faroe Lambert\",GEOGCS[\"FD54\""
  ",DATUM[\"Faroe Datum 1954\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPS"
  "G\",\"6741\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4741\"]],PROJECTION[\"Lambert Conic Con"
  "formal (West Orientated)\",AUTHORITY[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",62,AUTHORITY[\"EPSG"
  "\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at "
  "natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",NORTH],AXIS[\"X\",WEST],AUTHORITY[\"EPSG\",\"3144\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3145, 1, 'ETRS89 / Faroe Lambert', 'EPSG', 3145, 'PROJCS[\"ETRS89 / Faroe Lambert\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (West Orientated)\",AUTHORITY[\"EPSG\",\"9826\"]],PARA"
  "METER[\"Latitude of natural origin\",62,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",WEST],AUTHORITY[\"EPSG\",\"3145\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3148, 1, 'Indian 1960 / UTM zone 48N', 'EPSG', 3148, 'PROJCS[\"Indian 1960 / UTM zone 48N\",GE"
  "OGCS[\"Indian 1960\",DATUM[\"Indian 1960\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY"
  "[\"EPSG\",\"7015\"]],TOWGS84[182,915,344,0,0,0,0],AUTHORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4131\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3148\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3149, 1, 'Indian 1960 / UTM zone 49N', 'EPSG', 3149, 'PROJCS[\"Indian 1960 / UTM zone 49N\",GE"
  "OGCS[\"Indian 1960\",DATUM[\"Indian 1960\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY"
  "[\"EPSG\",\"7015\"]],TOWGS84[182,915,344,0,0,0,0],AUTHORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4131\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3149\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3152, 1, 'ST74', 'EPSG', 3152, 'PROJCS[\"ST74\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",18.05779,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999425,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100178.1808,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",-6500614.7836,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORT"
  "H],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"3152\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3153, 1, 'NAD83(CSRS) / BC Albers', 'EPSG', 3153, 'PROJCS[\"NAD83(CSRS) / BC Albers\",GEOGCS[\""
  "NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUTHOR"
  "ITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Albers E"
  "qual Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PAR"
  "AMETER[\"Longitude of false origin\",-126,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",50,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",58.5111111111111,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fal"
  "se origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3153\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3154, 1, 'NAD83(CSRS) / UTM zone 7N', 'EPSG', 3154, 'PROJCS[\"NAD83(CSRS) / UTM zone 7N\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3154\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3155, 1, 'NAD83(CSRS) / UTM zone 8N', 'EPSG', 3155, 'PROJCS[\"NAD83(CSRS) / UTM zone 8N\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3155\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3156, 1, 'NAD83(CSRS) / UTM zone 9N', 'EPSG', 3156, 'PROJCS[\"NAD83(CSRS) / UTM zone 9N\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3156\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3157, 1, 'NAD83(CSRS) / UTM zone 10N', 'EPSG', 3157, 'PROJCS[\"NAD83(CSRS) / UTM zone 10N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",E"
  "AST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3157\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3158, 1, 'NAD83(CSRS) / UTM zone 14N', 'EPSG', 3158, 'PROJCS[\"NAD83(CSRS) / UTM zone 14N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3158\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3159, 1, 'NAD83(CSRS) / UTM zone 15N', 'EPSG', 3159, 'PROJCS[\"NAD83(CSRS) / UTM zone 15N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3159\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3160, 1, 'NAD83(CSRS) / UTM zone 16N', 'EPSG', 3160, 'PROJCS[\"NAD83(CSRS) / UTM zone 16N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3160\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3161, 1, 'NAD83 / Ontario MNR Lambert', 'EPSG', 3161, 'PROJCS[\"NAD83 / Ontario MNR Lambert\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-85,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",44.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",53.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",930000,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",6430000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3161\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3162, 1, 'NAD83(CSRS) / Ontario MNR Lambert', 'EPSG', 3162, 'PROJCS[\"NAD83(CSRS) / Ontario MN"
  "R Lambert\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.01165"
  "99432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],"
  "PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",0,A"
  "UTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-85,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"La"
  "titude of 1st standard parallel\",44.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "53.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",930000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",6430000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3162\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3163, 1, 'RGNC91-93 / Lambert New Caledonia', 'EPSG', 3163, 'PROJCS[\"RGNC91-93 / Lambert New "
  "Caledonia\",GEOGCS[\"RGNC91-93\",DATUM[\"Reseau Geodesique de Nouvelle Caledonie 91-93\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4749\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",-21.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",166,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-20.6777777777778,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",-22.3444444444444,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3"
  "00000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3163\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3164, 1, 'ST87 Ouvea / UTM zone 58S', 'EPSG', 3164, 'PROJCS[\"ST87 Ouvea / UTM zone 58S\",GEOG"
  "CS[\"ST87 Ouvea\",DATUM[\"ST87 Ouvea\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS8"
  "4[-56.263,16.136,-22.856,0,0,0,0],AUTHORITY[\"EPSG\",\"6750\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4750\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3164\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3165, 1, 'NEA74 Noumea / Noumea Lambert', 'EPSG', 3165, 'PROJCS[\"NEA74 Noumea / Noumea Lamber"
  "t\",GEOGCS[\"NEA74 Noumea\",DATUM[\"NEA74 Noumea\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7"
  "022\"]],TOWGS84[-10.18,-350.43,291.37,0,0,0,0],AUTHORITY[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4644\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",-22.26969175,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",166.442"
  "42575,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-22.24469175,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",-22.29469175,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easti"
  "ng at false origin\",0.66,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1.02,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"31"
  "65\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3166, 1, 'NEA74 Noumea / Noumea Lambert 2', 'EPSG', 3166, 'PROJCS[\"NEA74 Noumea / Noumea Lamb"
  "ert 2\",GEOGCS[\"NEA74 Noumea\",DATUM[\"NEA74 Noumea\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-10.18,-350.43,291.37,0,0,0,0],AUTHORITY[\"EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4644\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",-22.2697222222222,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",166.4425,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-22.2447222222222,AUTHORIT"
  "Y[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",-22.2947222222222,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",8.313,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",-2."
  "354,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3166\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3167, 1, 'Kertau (RSO) / RSO Malaya (ch)', 'EPSG', 3167, 'PROJCS[\"Kertau (RSO) / RSO Malaya ("
  "ch)\",GEOGCS[\"Kertau (RSO)\",DATUM[\"Kertau (RSO)\",SPHEROID[\"Everest 1830 (RSO 1969)\",6377295.664,300.8017,AUTHOR"
  "ITY[\"EPSG\",\"7056\"]],AUTHORITY[\"EPSG\",\"6751\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4751\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of pro"
  "jection centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",102.25,AUTHORITY[\"EPSG"
  "\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",323.0257905,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from "
  "Rectified to Skew Grid\",323.130102361111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0."
  "99984,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"British chain (Sears 1922 truncated)\",20.116756,AUTHORITY[\"EPSG\""
  ",\"9301\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3167\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3168, 1, 'Kertau (RSO) / RSO Malaya (m)', 'EPSG', 3168, 'PROJCS[\"Kertau (RSO) / RSO Malaya (m"
  ")\",GEOGCS[\"Kertau (RSO)\",DATUM[\"Kertau (RSO)\",SPHEROID[\"Everest 1830 (RSO 1969)\",6377295.664,300.8017,AUTHORIT"
  "Y[\"EPSG\",\"7056\"]],AUTHORITY[\"EPSG\",\"6751\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "51\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of proje"
  "ction centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",102.25,AUTHORITY[\"EPSG\""
  ",\"8812\"]],PARAMETER[\"Azimuth of initial line\",323.0257905,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Re"
  "ctified to Skew Grid\",323.130102361111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0.99"
  "984,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",804670.24,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"3168\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3169, 1, 'RGNC91-93 / UTM zone 57S', 'EPSG', 3169, 'PROJCS[\"RGNC91-93 / UTM zone 57S\",GEOGCS"
  "[\"RGNC91-93\",DATUM[\"Reseau Geodesique de Nouvelle Caledonie 91-93\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4749\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3169\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3170, 1, 'RGNC91-93 / UTM zone 58S', 'EPSG', 3170, 'PROJCS[\"RGNC91-93 / UTM zone 58S\",GEOGCS"
  "[\"RGNC91-93\",DATUM[\"Reseau Geodesique de Nouvelle Caledonie 91-93\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4749\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3170\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3171, 1, 'RGNC91-93 / UTM zone 59S', 'EPSG', 3171, 'PROJCS[\"RGNC91-93 / UTM zone 59S\",GEOGCS"
  "[\"RGNC91-93\",DATUM[\"Reseau Geodesique de Nouvelle Caledonie 91-93\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4749\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3172, 1, 'IGN53 Mare / UTM zone 59S', 'EPSG', 3172, 'PROJCS[\"IGN53 Mare / UTM zone 59S\",GEOG"
  "CS[\"IGN53 Mare\",DATUM[\"IGN53 Mare\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG"
  "S84[-408.809,366.856,-412.987,1.8842,-0.5308,2.1655,-121.0993],AUTHORITY[\"EPSG\",\"6641\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4641\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3172\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3173, 1, 'fk89 / Faroe Lambert FK89', 'EPSG', 3173, 'PROJCS[\"fk89 / Faroe Lambert FK89\",GEOG"
  "CS[\"fk89\",DATUM[\"fk89\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG"
  "\",\"6753\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4753\"]],PROJECTION[\"Lambert Conic Conf"
  "ormal (West Orientated)\",AUTHORITY[\"EPSG\",\"9826\"]],PARAMETER[\"Latitude of natural origin\",62,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\""
  ",NORTH],AXIS[\"X\",WEST],AUTHORITY[\"EPSG\",\"3173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3174, 1, 'NAD83 / Great Lakes Albers', 'EPSG', 3174, 'PROJCS[\"NAD83 / Great Lakes Albers\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\""
  ",45.568977,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.455955,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.122774,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of "
  "2nd standard parallel\",49.01518,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3174\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3175, 1, 'NAD83 / Great Lakes and St Lawrence Albers', 'EPSG', 3175, 'PROJCS[\"NAD83 / Great L"
  "akes and St Lawrence Albers\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAM"
  "ETER[\"Latitude of false origin\",45.568977,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-83"
  ".248627,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.122774,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",49.01518,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "3175\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3176, 1, 'Indian 1960 / TM 106 NE', 'EPSG', 3176, 'PROJCS[\"Indian 1960 / TM 106 NE\",GEOGCS[\""
  "Indian 1960\",DATUM[\"Indian 1960\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG"
  "\",\"7015\"]],TOWGS84[182,915,344,0,0,0,0],AUTHORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4131\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",106,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3176\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3177, 1, 'LGD2006 / Libya TM', 'EPSG', 3177, 'PROJCS[\"LGD2006 / Libya TM\",GEOGCS[\"LGD2006\""
  ",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS"
  "84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9965,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3177\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3178, 1, 'GR96 / UTM zone 18N', 'EPSG', 3178, 'PROJCS[\"GR96 / UTM zone 18N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3178\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3179, 1, 'GR96 / UTM zone 19N', 'EPSG', 3179, 'PROJCS[\"GR96 / UTM zone 19N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3179\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3180, 1, 'GR96 / UTM zone 20N', 'EPSG', 3180, 'PROJCS[\"GR96 / UTM zone 20N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3180\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3181, 1, 'GR96 / UTM zone 21N', 'EPSG', 3181, 'PROJCS[\"GR96 / UTM zone 21N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3181\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3182, 1, 'GR96 / UTM zone 22N', 'EPSG', 3182, 'PROJCS[\"GR96 / UTM zone 22N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3182\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3183, 1, 'GR96 / UTM zone 23N', 'EPSG', 3183, 'PROJCS[\"GR96 / UTM zone 23N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3183\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3184, 1, 'GR96 / UTM zone 24N', 'EPSG', 3184, 'PROJCS[\"GR96 / UTM zone 24N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3184\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3185, 1, 'GR96 / UTM zone 25N', 'EPSG', 3185, 'PROJCS[\"GR96 / UTM zone 25N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3185\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3186, 1, 'GR96 / UTM zone 26N', 'EPSG', 3186, 'PROJCS[\"GR96 / UTM zone 26N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3186\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3187, 1, 'GR96 / UTM zone 27N', 'EPSG', 3187, 'PROJCS[\"GR96 / UTM zone 27N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3187\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3188, 1, 'GR96 / UTM zone 28N', 'EPSG', 3188, 'PROJCS[\"GR96 / UTM zone 28N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3188\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3189, 1, 'GR96 / UTM zone 29N', 'EPSG', 3189, 'PROJCS[\"GR96 / UTM zone 29N\",GEOGCS[\"GR96\","
  "DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at nat"
  "ural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3189\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3190, 1, 'LGD2006 / Libya TM zone 5', 'EPSG', 3190, 'PROJCS[\"LGD2006 / Libya TM zone 5\",GEOG"
  "CS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3190\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3191, 1, 'LGD2006 / Libya TM zone 6', 'EPSG', 3191, 'PROJCS[\"LGD2006 / Libya TM zone 6\",GEOG"
  "CS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3192, 1, 'LGD2006 / Libya TM zone 7', 'EPSG', 3192, 'PROJCS[\"LGD2006 / Libya TM zone 7\",GEOG"
  "CS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3193, 1, 'LGD2006 / Libya TM zone 8', 'EPSG', 3193, 'PROJCS[\"LGD2006 / Libya TM zone 8\",GEOG"
  "CS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3194, 1, 'LGD2006 / Libya TM zone 9', 'EPSG', 3194, 'PROJCS[\"LGD2006 / Libya TM zone 9\",GEOG"
  "CS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3194\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3195, 1, 'LGD2006 / Libya TM zone 10', 'EPSG', 3195, 'PROJCS[\"LGD2006 / Libya TM zone 10\",GE"
  "OGCS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3196, 1, 'LGD2006 / Libya TM zone 11', 'EPSG', 3196, 'PROJCS[\"LGD2006 / Libya TM zone 11\",GE"
  "OGCS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3196\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3197, 1, 'LGD2006 / Libya TM zone 12', 'EPSG', 3197, 'PROJCS[\"LGD2006 / Libya TM zone 12\",GE"
  "OGCS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3197\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3198, 1, 'LGD2006 / Libya TM zone 13', 'EPSG', 3198, 'PROJCS[\"LGD2006 / Libya TM zone 13\",GE"
  "OGCS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3198\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3199, 1, 'LGD2006 / UTM zone 32N', 'EPSG', 3199, 'PROJCS[\"LGD2006 / UTM zone 32N\",GEOGCS[\"L"
  "GD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3199\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3200, 1, 'FD58 / Iraq zone', 'EPSG', 3200, 'PROJCS[\"FD58 / Iraq zone\",GEOGCS[\"FD58\",DATUM["
  "\"Final Datum 1958\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-244.72"
  ",-162.773,400.75,0,0,0,0],AUTHORITY[\"EPSG\",\"6132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4132\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural"
  " origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9987864078,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1166200,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3200\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3201, 1, 'LGD2006 / UTM zone 33N', 'EPSG', 3201, 'PROJCS[\"LGD2006 / UTM zone 33N\",GEOGCS[\"L"
  "GD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3201\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3202, 1, 'LGD2006 / UTM zone 34N', 'EPSG', 3202, 'PROJCS[\"LGD2006 / UTM zone 34N\",GEOGCS[\"L"
  "GD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3202\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3203, 1, 'LGD2006 / UTM zone 35N', 'EPSG', 3203, 'PROJCS[\"LGD2006 / UTM zone 35N\",GEOGCS[\"L"
  "GD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],AUTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4754\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3203\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3204, 1, 'WGS 84 / SCAR IMW SP19-20', 'EPSG', 3204, 'PROJCS[\"WGS 84 / SCAR IMW SP19-20\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-66,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-60.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-63.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3204\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3205, 1, 'WGS 84 / SCAR IMW SP21-22', 'EPSG', 3205, 'PROJCS[\"WGS 84 / SCAR IMW SP21-22\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-54,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-60.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-63.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3205\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3206, 1, 'WGS 84 / SCAR IMW SP23-24', 'EPSG', 3206, 'PROJCS[\"WGS 84 / SCAR IMW SP23-24\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-42,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-60.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-63.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3206\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3207, 1, 'WGS 84 / SCAR IMW SQ01-02', 'EPSG', 3207, 'PROJCS[\"WGS 84 / SCAR IMW SQ01-02\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-174,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3207\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3208, 1, 'WGS 84 / SCAR IMW SQ19-20', 'EPSG', 3208, 'PROJCS[\"WGS 84 / SCAR IMW SQ19-20\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-66,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3208\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3209, 1, 'WGS 84 / SCAR IMW SQ21-22', 'EPSG', 3209, 'PROJCS[\"WGS 84 / SCAR IMW SQ21-22\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-54,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3209\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3210, 1, 'WGS 84 / SCAR IMW SQ37-38', 'EPSG', 3210, 'PROJCS[\"WGS 84 / SCAR IMW SQ37-38\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",42,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3210\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3211, 1, 'WGS 84 / SCAR IMW SQ39-40', 'EPSG', 3211, 'PROJCS[\"WGS 84 / SCAR IMW SQ39-40\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",54,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3211\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3212, 1, 'WGS 84 / SCAR IMW SQ41-42', 'EPSG', 3212, 'PROJCS[\"WGS 84 / SCAR IMW SQ41-42\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",66,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3212\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3213, 1, 'WGS 84 / SCAR IMW SQ43-44', 'EPSG', 3213, 'PROJCS[\"WGS 84 / SCAR IMW SQ43-44\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",78,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3213\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3214, 1, 'WGS 84 / SCAR IMW SQ45-46', 'EPSG', 3214, 'PROJCS[\"WGS 84 / SCAR IMW SQ45-46\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3214\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3215, 1, 'WGS 84 / SCAR IMW SQ47-48', 'EPSG', 3215, 'PROJCS[\"WGS 84 / SCAR IMW SQ47-48\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",102,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3215\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3216, 1, 'WGS 84 / SCAR IMW SQ49-50', 'EPSG', 3216, 'PROJCS[\"WGS 84 / SCAR IMW SQ49-50\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",114,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3216\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3217, 1, 'WGS 84 / SCAR IMW SQ51-52', 'EPSG', 3217, 'PROJCS[\"WGS 84 / SCAR IMW SQ51-52\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",126,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3217\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3218, 1, 'WGS 84 / SCAR IMW SQ53-54', 'EPSG', 3218, 'PROJCS[\"WGS 84 / SCAR IMW SQ53-54\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",138,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3218\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3219, 1, 'WGS 84 / SCAR IMW SQ55-56', 'EPSG', 3219, 'PROJCS[\"WGS 84 / SCAR IMW SQ55-56\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",150,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3219\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3220, 1, 'WGS 84 / SCAR IMW SQ57-58', 'EPSG', 3220, 'PROJCS[\"WGS 84 / SCAR IMW SQ57-58\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",162,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-64.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-67.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3221, 1, 'WGS 84 / SCAR IMW SR13-14', 'EPSG', 3221, 'PROJCS[\"WGS 84 / SCAR IMW SR13-14\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-102,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3221\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3222, 1, 'WGS 84 / SCAR IMW SR15-16', 'EPSG', 3222, 'PROJCS[\"WGS 84 / SCAR IMW SR15-16\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3222\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3223, 1, 'WGS 84 / SCAR IMW SR17-18', 'EPSG', 3223, 'PROJCS[\"WGS 84 / SCAR IMW SR17-18\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3223\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3224, 1, 'WGS 84 / SCAR IMW SR19-20', 'EPSG', 3224, 'PROJCS[\"WGS 84 / SCAR IMW SR19-20\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-66,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3224\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3225, 1, 'WGS 84 / SCAR IMW SR27-28', 'EPSG', 3225, 'PROJCS[\"WGS 84 / SCAR IMW SR27-28\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-18,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3225\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3226, 1, 'WGS 84 / SCAR IMW SR29-30', 'EPSG', 3226, 'PROJCS[\"WGS 84 / SCAR IMW SR29-30\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-6,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3226\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3227, 1, 'WGS 84 / SCAR IMW SR31-32', 'EPSG', 3227, 'PROJCS[\"WGS 84 / SCAR IMW SR31-32\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",6,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of"
  " 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paralle"
  "l\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3227\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3228, 1, 'WGS 84 / SCAR IMW SR33-34', 'EPSG', 3228, 'PROJCS[\"WGS 84 / SCAR IMW SR33-34\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",18,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3228\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3229, 1, 'WGS 84 / SCAR IMW SR35-36', 'EPSG', 3229, 'PROJCS[\"WGS 84 / SCAR IMW SR35-36\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",30,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3229\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3230, 1, 'WGS 84 / SCAR IMW SR37-38', 'EPSG', 3230, 'PROJCS[\"WGS 84 / SCAR IMW SR37-38\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",42,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3230\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3231, 1, 'WGS 84 / SCAR IMW SR39-40', 'EPSG', 3231, 'PROJCS[\"WGS 84 / SCAR IMW SR39-40\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",54,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3231\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3232, 1, 'WGS 84 / SCAR IMW SR41-42', 'EPSG', 3232, 'PROJCS[\"WGS 84 / SCAR IMW SR41-42\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",66,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3232\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3233, 1, 'WGS 84 / SCAR IMW SR43-44', 'EPSG', 3233, 'PROJCS[\"WGS 84 / SCAR IMW SR43-44\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",78,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3233\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3234, 1, 'WGS 84 / SCAR IMW SR45-46', 'EPSG', 3234, 'PROJCS[\"WGS 84 / SCAR IMW SR45-46\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3234\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3235, 1, 'WGS 84 / SCAR IMW SR47-48', 'EPSG', 3235, 'PROJCS[\"WGS 84 / SCAR IMW SR47-48\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",102,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3235\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3236, 1, 'WGS 84 / SCAR IMW SR49-50', 'EPSG', 3236, 'PROJCS[\"WGS 84 / SCAR IMW SR49-50\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",114,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3236\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3237, 1, 'WGS 84 / SCAR IMW SR51-52', 'EPSG', 3237, 'PROJCS[\"WGS 84 / SCAR IMW SR51-52\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",126,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3237\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3238, 1, 'WGS 84 / SCAR IMW SR53-54', 'EPSG', 3238, 'PROJCS[\"WGS 84 / SCAR IMW SR53-54\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",138,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3238\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3239, 1, 'WGS 84 / SCAR IMW SR55-56', 'EPSG', 3239, 'PROJCS[\"WGS 84 / SCAR IMW SR55-56\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",150,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3239\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3240, 1, 'WGS 84 / SCAR IMW SR57-58', 'EPSG', 3240, 'PROJCS[\"WGS 84 / SCAR IMW SR57-58\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",162,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3240\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3241, 1, 'WGS 84 / SCAR IMW SR59-60', 'EPSG', 3241, 'PROJCS[\"WGS 84 / SCAR IMW SR59-60\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",174,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-68.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-71.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3241\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3242, 1, 'WGS 84 / SCAR IMW SS04-06', 'EPSG', 3242, 'PROJCS[\"WGS 84 / SCAR IMW SS04-06\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-153,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3242\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3243, 1, 'WGS 84 / SCAR IMW SS07-09', 'EPSG', 3243, 'PROJCS[\"WGS 84 / SCAR IMW SS07-09\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-135,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3243\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3244, 1, 'WGS 84 / SCAR IMW SS10-12', 'EPSG', 3244, 'PROJCS[\"WGS 84 / SCAR IMW SS10-12\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-117,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3244\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3245, 1, 'WGS 84 / SCAR IMW SS13-15', 'EPSG', 3245, 'PROJCS[\"WGS 84 / SCAR IMW SS13-15\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3245\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3246, 1, 'WGS 84 / SCAR IMW SS16-18', 'EPSG', 3246, 'PROJCS[\"WGS 84 / SCAR IMW SS16-18\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3247, 1, 'WGS 84 / SCAR IMW SS19-21', 'EPSG', 3247, 'PROJCS[\"WGS 84 / SCAR IMW SS19-21\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-63,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3247\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3248, 1, 'WGS 84 / SCAR IMW SS25-27', 'EPSG', 3248, 'PROJCS[\"WGS 84 / SCAR IMW SS25-27\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-27,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3248\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3249, 1, 'WGS 84 / SCAR IMW SS28-30', 'EPSG', 3249, 'PROJCS[\"WGS 84 / SCAR IMW SS28-30\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-9,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3250, 1, 'WGS 84 / SCAR IMW SS31-33', 'EPSG', 3250, 'PROJCS[\"WGS 84 / SCAR IMW SS31-33\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",9,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of"
  " 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paralle"
  "l\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3250\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3251, 1, 'WGS 84 / SCAR IMW SS34-36', 'EPSG', 3251, 'PROJCS[\"WGS 84 / SCAR IMW SS34-36\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",27,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3251\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3252, 1, 'WGS 84 / SCAR IMW SS37-39', 'EPSG', 3252, 'PROJCS[\"WGS 84 / SCAR IMW SS37-39\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",45,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3253, 1, 'WGS 84 / SCAR IMW SS40-42', 'EPSG', 3253, 'PROJCS[\"WGS 84 / SCAR IMW SS40-42\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",63,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3254, 1, 'WGS 84 / SCAR IMW SS43-45', 'EPSG', 3254, 'PROJCS[\"WGS 84 / SCAR IMW SS43-45\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3255, 1, 'WGS 84 / SCAR IMW SS46-48', 'EPSG', 3255, 'PROJCS[\"WGS 84 / SCAR IMW SS46-48\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3256, 1, 'WGS 84 / SCAR IMW SS49-51', 'EPSG', 3256, 'PROJCS[\"WGS 84 / SCAR IMW SS49-51\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",117,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3257, 1, 'WGS 84 / SCAR IMW SS52-54', 'EPSG', 3257, 'PROJCS[\"WGS 84 / SCAR IMW SS52-54\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",135,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3258, 1, 'WGS 84 / SCAR IMW SS55-57', 'EPSG', 3258, 'PROJCS[\"WGS 84 / SCAR IMW SS55-57\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",153,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3259, 1, 'WGS 84 / SCAR IMW SS58-60', 'EPSG', 3259, 'PROJCS[\"WGS 84 / SCAR IMW SS58-60\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",171,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-72.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3260, 1, 'WGS 84 / SCAR IMW ST01-04', 'EPSG', 3260, 'PROJCS[\"WGS 84 / SCAR IMW ST01-04\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-168,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3260\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3261, 1, 'WGS 84 / SCAR IMW ST05-08', 'EPSG', 3261, 'PROJCS[\"WGS 84 / SCAR IMW ST05-08\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-144,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3261\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3262, 1, 'WGS 84 / SCAR IMW ST09-12', 'EPSG', 3262, 'PROJCS[\"WGS 84 / SCAR IMW ST09-12\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3262\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3263, 1, 'WGS 84 / SCAR IMW ST13-16', 'EPSG', 3263, 'PROJCS[\"WGS 84 / SCAR IMW ST13-16\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-96,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3263\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3264, 1, 'WGS 84 / SCAR IMW ST17-20', 'EPSG', 3264, 'PROJCS[\"WGS 84 / SCAR IMW ST17-20\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-72,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3264\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3265, 1, 'WGS 84 / SCAR IMW ST21-24', 'EPSG', 3265, 'PROJCS[\"WGS 84 / SCAR IMW ST21-24\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-48,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3265\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3266, 1, 'WGS 84 / SCAR IMW ST25-28', 'EPSG', 3266, 'PROJCS[\"WGS 84 / SCAR IMW ST25-28\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-24,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3266\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3267, 1, 'WGS 84 / SCAR IMW ST29-32', 'EPSG', 3267, 'PROJCS[\"WGS 84 / SCAR IMW ST29-32\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",0,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of"
  " 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paralle"
  "l\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3267\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3268, 1, 'WGS 84 / SCAR IMW ST33-36', 'EPSG', 3268, 'PROJCS[\"WGS 84 / SCAR IMW ST33-36\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",24,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3268\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3269, 1, 'WGS 84 / SCAR IMW ST37-40', 'EPSG', 3269, 'PROJCS[\"WGS 84 / SCAR IMW ST37-40\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",48,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3269\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3270, 1, 'WGS 84 / SCAR IMW ST41-44', 'EPSG', 3270, 'PROJCS[\"WGS 84 / SCAR IMW ST41-44\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",72,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3270\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3271, 1, 'WGS 84 / SCAR IMW ST45-48', 'EPSG', 3271, 'PROJCS[\"WGS 84 / SCAR IMW ST45-48\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",96,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3271\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3272, 1, 'WGS 84 / SCAR IMW ST49-52', 'EPSG', 3272, 'PROJCS[\"WGS 84 / SCAR IMW ST49-52\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",120,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3272\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3273, 1, 'WGS 84 / SCAR IMW ST53-56', 'EPSG', 3273, 'PROJCS[\"WGS 84 / SCAR IMW ST53-56\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",144,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3273\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3274, 1, 'WGS 84 / SCAR IMW ST57-60', 'EPSG', 3274, 'PROJCS[\"WGS 84 / SCAR IMW ST57-60\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",168,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3274\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3275, 1, 'WGS 84 / SCAR IMW SU01-05', 'EPSG', 3275, 'PROJCS[\"WGS 84 / SCAR IMW SU01-05\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-165,AUTHORITY[\"EPSG\",\"8833\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3275\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3276, 1, 'WGS 84 / SCAR IMW SU06-10', 'EPSG', 3276, 'PROJCS[\"WGS 84 / SCAR IMW SU06-10\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-135,AUTHORITY[\"EPSG\",\"8833\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3276\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3277, 1, 'WGS 84 / SCAR IMW SU11-15', 'EPSG', 3277, 'PROJCS[\"WGS 84 / SCAR IMW SU11-15\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-105,AUTHORITY[\"EPSG\",\"8833\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3277\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3278, 1, 'WGS 84 / SCAR IMW SU16-20', 'EPSG', 3278, 'PROJCS[\"WGS 84 / SCAR IMW SU16-20\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-75,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3278\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3279, 1, 'WGS 84 / SCAR IMW SU21-25', 'EPSG', 3279, 'PROJCS[\"WGS 84 / SCAR IMW SU21-25\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-45,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3279\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3280, 1, 'WGS 84 / SCAR IMW SU26-30', 'EPSG', 3280, 'PROJCS[\"WGS 84 / SCAR IMW SU26-30\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-15,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3280\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3281, 1, 'WGS 84 / SCAR IMW SU31-35', 'EPSG', 3281, 'PROJCS[\"WGS 84 / SCAR IMW SU31-35\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",15,AUTHORITY[\"EPSG\",\"8833\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3281\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3282, 1, 'WGS 84 / SCAR IMW SU36-40', 'EPSG', 3282, 'PROJCS[\"WGS 84 / SCAR IMW SU36-40\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",45,AUTHORITY[\"EPSG\",\"8833\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3282\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3283, 1, 'WGS 84 / SCAR IMW SU41-45', 'EPSG', 3283, 'PROJCS[\"WGS 84 / SCAR IMW SU41-45\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",75,AUTHORITY[\"EPSG\",\"8833\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3283\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3284, 1, 'WGS 84 / SCAR IMW SU46-50', 'EPSG', 3284, 'PROJCS[\"WGS 84 / SCAR IMW SU46-50\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",105,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3284\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3285, 1, 'WGS 84 / SCAR IMW SU51-55', 'EPSG', 3285, 'PROJCS[\"WGS 84 / SCAR IMW SU51-55\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",135,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3285\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3286, 1, 'WGS 84 / SCAR IMW SU56-60', 'EPSG', 3286, 'PROJCS[\"WGS 84 / SCAR IMW SU56-60\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",165,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3286\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3287, 1, 'WGS 84 / SCAR IMW SV01-10', 'EPSG', 3287, 'PROJCS[\"WGS 84 / SCAR IMW SV01-10\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-150,AUTHORITY[\"EPSG\",\"8833\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3287\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3288, 1, 'WGS 84 / SCAR IMW SV11-20', 'EPSG', 3288, 'PROJCS[\"WGS 84 / SCAR IMW SV11-20\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-90,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3288\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3289, 1, 'WGS 84 / SCAR IMW SV21-30', 'EPSG', 3289, 'PROJCS[\"WGS 84 / SCAR IMW SV21-30\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-30,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3289\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3290, 1, 'WGS 84 / SCAR IMW SV31-40', 'EPSG', 3290, 'PROJCS[\"WGS 84 / SCAR IMW SV31-40\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",30,AUTHORITY[\"EPSG\",\"8833\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3290\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3291, 1, 'WGS 84 / SCAR IMW SV41-50', 'EPSG', 3291, 'PROJCS[\"WGS 84 / SCAR IMW SV41-50\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",90,AUTHORITY[\"EPSG\",\"8833\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3291\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3292, 1, 'WGS 84 / SCAR IMW SV51-60', 'EPSG', 3292, 'PROJCS[\"WGS 84 / SCAR IMW SV51-60\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",150,AUTHORITY[\"EPSG\",\"8833\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3292\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3293, 1, 'WGS 84 / SCAR IMW SW01-60', 'EPSG', 3293, 'PROJCS[\"WGS 84 / SCAR IMW SW01-60\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-80.2"
  "386111111111,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",0,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETE"
  "R[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3293\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3294, 1, 'WGS 84 / USGS Transantarctic Mountains', 'EPSG', 3294, 'PROJCS[\"WGS 84 / USGS Trans"
  "antarctic Mountains\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.25722356"
  "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",-78,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",162,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3294\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3295, 1, 'Guam 1963 / Yap Islands', 'EPSG', 3295, 'PROJCS[\"Guam 1963 / Yap Islands\",GEOGCS[\""
  "Guam 1963\",DATUM[\"Guam 1963\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TO"
  "WGS84[-100,-248,259,0,0,0,0],AUTHORITY[\"EPSG\",\"6675\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4675\"]],PROJECTION[\"Modified Azimuthal Equidistant\",AUTHORITY[\"EPSG\",\"9832\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",9.54670833333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",138.1687444444"
  "44,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",60000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3295\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3296, 1, 'RGPF / UTM zone 5S', 'EPSG', 3296, 'PROJCS[\"RGPF / UTM zone 5S\",GEOGCS[\"RGPF\",DA"
  "TUM[\"Reseau Geodesique de la Polynesie Francaise\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0.072,-0.507,-0.245,0.0183,-0.0003,0.007,-0.0093],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4687\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "53,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "3296\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3297, 1, 'RGPF / UTM zone 6S', 'EPSG', 3297, 'PROJCS[\"RGPF / UTM zone 6S\",GEOGCS[\"RGPF\",DA"
  "TUM[\"Reseau Geodesique de la Polynesie Francaise\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0.072,-0.507,-0.245,0.0183,-0.0003,0.007,-0.0093],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4687\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "47,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "3297\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3298, 1, 'RGPF / UTM zone 7S', 'EPSG', 3298, 'PROJCS[\"RGPF / UTM zone 7S\",GEOGCS[\"RGPF\",DA"
  "TUM[\"Reseau Geodesique de la Polynesie Francaise\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0.072,-0.507,-0.245,0.0183,-0.0003,0.007,-0.0093],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4687\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "41,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "3298\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3299, 1, 'RGPF / UTM zone 8S', 'EPSG', 3299, 'PROJCS[\"RGPF / UTM zone 8S\",GEOGCS[\"RGPF\",DA"
  "TUM[\"Reseau Geodesique de la Polynesie Francaise\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0.072,-0.507,-0.245,0.0183,-0.0003,0.007,-0.0093],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4687\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "35,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "3299\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3300, 1, 'Estonian Coordinate System of 1992', 'EPSG', 3300, 'PROJCS[\"Estonian Coordinate Sys"
  "tem of 1992\",GEOGCS[\"EST92\",DATUM[\"Estonia 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0.055,-0.541,-0.185,0.0183,-0.0003,-0.007,-0.014],AUTHORITY[\"EPSG\",\"6133\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4133\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",57.5175539305556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",24,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",59.333333333333"
  "3,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",58,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6375000"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTH"
  "ORITY[\"EPSG\",\"3300\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3301, 1, 'Estonian Coordinate System of 1997', 'EPSG', 3301, 'PROJCS[\"Estonian Coordinate Sys"
  "tem of 1997\",GEOGCS[\"EST97\",DATUM[\"Estonia 1997\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6180\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4180\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",57.5175539305556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",24,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",59.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",58,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",50"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6375000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3301\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3302, 1, 'IGN63 Hiva Oa / UTM zone 7S', 'EPSG', 3302, 'PROJCS[\"IGN63 Hiva Oa / UTM zone 7S\","
  "GEOGCS[\"IGN63 Hiva Oa\",DATUM[\"IGN63 Hiva Oa\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"702"
  "2\"]],TOWGS84[374.787,-58.914,-1.202,-16.1928,-11.4629,-5.5287,-0.5502],AUTHORITY[\"EPSG\",\"6689\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4689\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3302\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3303, 1, 'Fatu Iva 72 / UTM zone 7S', 'EPSG', 3303, 'PROJCS[\"Fatu Iva 72 / UTM zone 7S\",GEOG"
  "CS[\"Fatu Iva 72\",DATUM[\"Fatu Iva 72\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO"
  "WGS84[347.175,1077.618,2623.677,33.9058,-70.6776,9.4013,186.0647],AUTHORITY[\"EPSG\",\"6688\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4688\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-14"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"33"
  "03\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3304, 1, 'Tahiti 79 / UTM zone 6S', 'EPSG', 3304, 'PROJCS[\"Tahiti 79 / UTM zone 6S\",GEOGCS[\""
  "Tahiti 79\",DATUM[\"Tahiti 79\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[221"
  ".597,152.441,176.523,2.403,1.3893,0.884,11.4648],AUTHORITY[\"EPSG\",\"6690\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4690\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3304\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3305, 1, 'Moorea 87 / UTM zone 6S', 'EPSG', 3305, 'PROJCS[\"Moorea 87 / UTM zone 6S\",GEOGCS[\""
  "Moorea 87\",DATUM[\"Moorea 87\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[218"
  ".769,150.75,176.75,3.5231,2.0037,1.288,10.9817],AUTHORITY[\"EPSG\",\"6691\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4691\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3305\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3306, 1, 'Maupiti 83 / UTM zone 5S', 'EPSG', 3306, 'PROJCS[\"Maupiti 83 / UTM zone 5S\",GEOGCS"
  "[\"Maupiti 83\",DATUM[\"Maupiti 83\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS8"
  "4[217.109,86.452,23.711,0.0183,-0.0003,0.007,-0.0093],AUTHORITY[\"EPSG\",\"6692\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4692\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3306\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3307, 1, 'Nakhl-e Ghanem / UTM zone 39N', 'EPSG', 3307, 'PROJCS[\"Nakhl-e Ghanem / UTM zone 39"
  "N\",GEOGCS[\"Nakhl-e Ghanem\",DATUM[\"Nakhl-e Ghanem\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\""
  "7030\"]],TOWGS84[0,-0.15,0.68,0,0,0,0],AUTHORITY[\"EPSG\",\"6693\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4693\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3307\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3308, 1, 'GDA94 / NSW Lambert', 'EPSG', 3308, 'PROJCS[\"GDA94 / NSW Lambert\",GEOGCS[\"GDA94\""
  ",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4283\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false o"
  "rigin\",-33.25,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",147,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",-30.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd st"
  "andard parallel\",-35.75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",9300000,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3308\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3309, 1, 'NAD27 / California Albers', 'EPSG', 3309, 'PROJCS[\"NAD27 / California Albers\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Albers Equal Area"
  "\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-120,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETE"
  "R[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",-4000000,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"3309\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3310, 1, 'NAD83 / California Albers', 'EPSG', 3310, 'PROJCS[\"NAD83 / California Albers\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\","
  "0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",34,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",40.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",-4000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3310\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3311, 1, 'NAD83(HARN) / California Albers', 'EPSG', 3311, 'PROJCS[\"NAD83(HARN) / California A"
  "lbers\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETE"
  "R[\"Latitude of false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lat"
  "itude of 2nd standard parallel\",40.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",-4000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3312, 1, 'CSG67 / UTM zone 21N', 'EPSG', 3312, 'PROJCS[\"CSG67 / UTM zone 21N\",GEOGCS[\"CSG67"
  "\",DATUM[\"Centre Spatial Guyanais 1967\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-186,230,110,0,0,0,0],AUTHORITY[\"EPSG\",\"6623\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4623\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3313, 1, 'RGFG95 / UTM zone 21N', 'EPSG', 3313, 'PROJCS[\"RGFG95 / UTM zone 21N\",GEOGCS[\"RGF"
  "G95\",DATUM[\"Reseau Geodesique Francais Guyane 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6624\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4624\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3313\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3316, 1, 'Kasai 1953 / Congo TM zone 22', 'EPSG', 3316, 'PROJCS[\"Kasai 1953 / Congo TM zone 2"
  "2\",GEOGCS[\"Kasai 1953\",DATUM[\"Kasai 1953\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],AUTHORITY[\"EPSG\",\"6696\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4696\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3317, 1, 'Kasai 1953 / Congo TM zone 24', 'EPSG', 3317, 'PROJCS[\"Kasai 1953 / Congo TM zone 2"
  "4\",GEOGCS[\"Kasai 1953\",DATUM[\"Kasai 1953\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],AUTHORITY[\"EPSG\",\"6696\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4696\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3317\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3318, 1, 'IGC 1962 / Congo TM zone 12', 'EPSG', 3318, 'PROJCS[\"IGC 1962 / Congo TM zone 12\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3319, 1, 'IGC 1962 / Congo TM zone 14', 'EPSG', 3319, 'PROJCS[\"IGC 1962 / Congo TM zone 14\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",14,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3319\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3320, 1, 'IGC 1962 / Congo TM zone 16', 'EPSG', 3320, 'PROJCS[\"IGC 1962 / Congo TM zone 16\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3320\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3321, 1, 'IGC 1962 / Congo TM zone 18', 'EPSG', 3321, 'PROJCS[\"IGC 1962 / Congo TM zone 18\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3321\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3322, 1, 'IGC 1962 / Congo TM zone 20', 'EPSG', 3322, 'PROJCS[\"IGC 1962 / Congo TM zone 20\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3323, 1, 'IGC 1962 / Congo TM zone 22', 'EPSG', 3323, 'PROJCS[\"IGC 1962 / Congo TM zone 22\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3323\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3324, 1, 'IGC 1962 / Congo TM zone 24', 'EPSG', 3324, 'PROJCS[\"IGC 1962 / Congo TM zone 24\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3325, 1, 'IGC 1962 / Congo TM zone 26', 'EPSG', 3325, 'PROJCS[\"IGC 1962 / Congo TM zone 26\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",26,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3325\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3326, 1, 'IGC 1962 / Congo TM zone 28', 'EPSG', 3326, 'PROJCS[\"IGC 1962 / Congo TM zone 28\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3326\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3327, 1, 'IGC 1962 / Congo TM zone 30', 'EPSG', 3327, 'PROJCS[\"IGC 1962 / Congo TM zone 30\","
  "GEOGCS[\"IGC 1962 6th Parallel South\",DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4697\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3327\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3328, 1, 'Pulkovo 1942(58) / GUGiK-80', 'EPSG', 3328, 'PROJCS[\"Pulkovo 1942(58) / GUGiK-80\","
  "GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\","
  "\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"9809\"]],P"
  "ARAMETER[\"Latitude of natural origin\",52.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",19.1666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999714,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",E"
  "AST],AUTHORITY[\"EPSG\",\"3328\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3329, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 5', 'EPSG', 3329, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 5\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3329\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3330, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 6', 'EPSG', 3330, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 6\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",18,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3331, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 7', 'EPSG', 3331, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 7\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3332, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 8', 'EPSG', 3332, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 8\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3333, 1, 'Pulkovo 1942(58) / Gauss-Kruger zone 3', 'EPSG', 3333, 'PROJCS[\"Pulkovo 1942(58) / "
  "Gauss-Kruger zone 3\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"3333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3334, 1, 'Pulkovo 1942(58) / Gauss-Kruger zone 4', 'EPSG', 3334, 'PROJCS[\"Pulkovo 1942(58) / "
  "Gauss-Kruger zone 4\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"3334\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3335, 1, 'Pulkovo 1942(58) / Gauss-Kruger zone 5', 'EPSG', 3335, 'PROJCS[\"Pulkovo 1942(58) / "
  "Gauss-Kruger zone 5\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"3335\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3336, 1, 'IGN 1962 Kerguelen / UTM zone 42S', 'EPSG', 3336, 'PROJCS[\"IGN 1962 Kerguelen / UTM"
  " zone 42S\",GEOGCS[\"IGN 1962 Kerguelen\",DATUM[\"IGN 1962 Kerguelen\",SPHEROID[\"International 1924\",6378388,297,AU"
  "THORITY[\"EPSG\",\"7022\"]],TOWGS84[145,-187,103,0,0,0,0],AUTHORITY[\"EPSG\",\"6698\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4698\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3336\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3337, 1, 'Le Pouce 1934 / Mauritius Grid', 'EPSG', 3337, 'PROJCS[\"Le Pouce 1934 / Mauritius G"
  "rid\",GEOGCS[\"Le Pouce 1934\",DATUM[\"Le Pouce 1934\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\""
  "EPSG\",\"7012\"]],TOWGS84[-770.1,158.4,-498.2,0,0,0,0],AUTHORITY[\"EPSG\",\"6699\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4699\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",-20.1950694444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",57.5218277777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000"
  "000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3338, 1, 'NAD83 / Alaska Albers', 'EPSG', 3338, 'PROJCS[\"NAD83 / Alaska Albers\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",50,AUTHO"
  "RITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-154,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latit"
  "ude of 1st standard parallel\",55,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",65,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northin"
  "g at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3338\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3339, 1, 'IGCB 1955 / Congo TM zone 12', 'EPSG', 3339, 'PROJCS[\"IGCB 1955 / Congo TM zone 12\""
  ",GEOGCS[\"IGCB 1955\",DATUM[\"Institut Geographique du Congo Belge 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,"
  "293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4701\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3340, 1, 'IGCB 1955 / Congo TM zone 14', 'EPSG', 3340, 'PROJCS[\"IGCB 1955 / Congo TM zone 14\""
  ",GEOGCS[\"IGCB 1955\",DATUM[\"Institut Geographique du Congo Belge 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,"
  "293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4701\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",14,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3341, 1, 'IGCB 1955 / Congo TM zone 16', 'EPSG', 3341, 'PROJCS[\"IGCB 1955 / Congo TM zone 16\""
  ",GEOGCS[\"IGCB 1955\",DATUM[\"Institut Geographique du Congo Belge 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,"
  "293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4701\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",16,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3341\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3342, 1, 'IGCB 1955 / UTM zone 33S', 'EPSG', 3342, 'PROJCS[\"IGCB 1955 / UTM zone 33S\",GEOGCS"
  "[\"IGCB 1955\",DATUM[\"Institut Geographique du Congo Belge 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465"
  ",AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-168.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4701\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"334"
  "2\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3343, 1, 'Mauritania 1999 / UTM zone 28N', 'EPSG', 3343, 'PROJCS[\"Mauritania 1999 / UTM zone "
  "28N\",GEOGCS[\"Mauritania 1999\",DATUM[\"Mauritania 1999\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6702\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4702\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3344, 1, 'Mauritania 1999 / UTM zone 29N', 'EPSG', 3344, 'PROJCS[\"Mauritania 1999 / UTM zone "
  "29N\",GEOGCS[\"Mauritania 1999\",DATUM[\"Mauritania 1999\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6702\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4702\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3345, 1, 'Mauritania 1999 / UTM zone 30N', 'EPSG', 3345, 'PROJCS[\"Mauritania 1999 / UTM zone "
  "30N\",GEOGCS[\"Mauritania 1999\",DATUM[\"Mauritania 1999\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6702\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4702\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3345\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3346, 1, 'LKS94 / Lithuania TM', 'EPSG', 3346, 'PROJCS[\"LKS94 / Lithuania TM\",GEOGCS[\"LKS94"
  "\",DATUM[\"Lithuania 1994 (ETRS89)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6126\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4669\"]"
  "],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",0.9998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\","
  "\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]"
  "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3347, 1, 'NAD83 / Statistics Canada Lambert', 'EPSG', 3347, 'PROJCS[\"NAD83 / Statistics Canad"
  "a Lambert\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",63.390675,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.877777"
  "7777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origi"
  "n\",6200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3347\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3348, 1, 'NAD83(CSRS) / Statistics Canada Lambert', 'EPSG', 3348, 'PROJCS[\"NAD83(CSRS) / Stat"
  "istics Canada Lambert\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027"
  "041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4617\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false o"
  "rigin\",63.390675,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.8777777777778,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitu"
  "de of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6200000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3350, 1, 'Pulkovo 1942 / CS63 zone C0', 'EPSG', 3350, 'PROJCS[\"Pulkovo 1942 / CS63 zone C0\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21.95,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3351, 1, 'Pulkovo 1942 / CS63 zone C1', 'EPSG', 3351, 'PROJCS[\"Pulkovo 1942 / CS63 zone C1\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24.95,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",125"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3352, 1, 'Pulkovo 1942 / CS63 zone C2', 'EPSG', 3352, 'PROJCS[\"Pulkovo 1942 / CS63 zone C2\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27.95,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",225"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3353, 1, 'Mhast (onshore) / UTM zone 32S', 'EPSG', 3353, 'PROJCS[\"Mhast (onshore) / UTM zone "
  "32S\",GEOGCS[\"Mhast (onshore)\",DATUM[\"Mhast (onshore)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6704\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4704\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3354, 1, 'Mhast (offshore) / UTM zone 32S', 'EPSG', 3354, 'PROJCS[\"Mhast (offshore) / UTM zon"
  "e 32S\",GEOGCS[\"Mhast (offshore)\",DATUM[\"Mhast (offshore)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY["
  "\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6705\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4705\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\","
  "\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3355, 1, 'Egypt Gulf of Suez S-650 TL / Red Belt', 'EPSG', 3355, 'PROJCS[\"Egypt Gulf of Suez "
  "S-650 TL / Red Belt\",GEOGCS[\"Egypt Gulf of Suez S-650 TL\",DATUM[\"Egypt Gulf of Suez S-650 TL\",SPHEROID[\"Helmert"
  " 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-146.21,112.63,4.05,0,0,0,0],AUTHORITY[\"EPSG\",\"6706\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4706\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",31,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",615000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",810000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"3355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3358, 1, 'NAD83(HARN) / North Carolina', 'EPSG', 3358, 'PROJCS[\"NAD83(HARN) / North Carolina\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",609601.22,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3358\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3360, 1, 'NAD83(HARN) / South Carolina', 'EPSG', 3360, 'PROJCS[\"NAD83(HARN) / South Carolina\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PAR"
  "AMETER[\"Latitude of false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false ori"
  "gin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.8333333333333,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",609600,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"3360\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3361, 1, 'NAD83(HARN) / South Carolina (ft)', 'EPSG', 3361, 'PROJCS[\"NAD83(HARN) / South Caro"
  "lina (ft)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.8333333333333,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3361\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3362, 1, 'NAD83(HARN) / Pennsylvania North', 'EPSG', 3362, 'PROJCS[\"NAD83(HARN) / Pennsylvani"
  "a North\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.95,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"3362\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3363, 1, 'NAD83(HARN) / Pennsylvania North (ftUS)', 'EPSG', 3363, 'PROJCS[\"NAD83(HARN) / Penn"
  "sylvania North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.95"
  ",AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3363\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3364, 1, 'NAD83(HARN) / Pennsylvania South', 'EPSG', 3364, 'PROJCS[\"NAD83(HARN) / Pennsylvani"
  "a South\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.9777777777778,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3364\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3365, 1, 'NAD83(HARN) / Pennsylvania South (ftUS)', 'EPSG', 3365, 'PROJCS[\"NAD83(HARN) / Penn"
  "sylvania South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.97"
  "77777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTHORITY["
  "\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing "
  "at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"90"
  "03\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3365\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3367, 1, 'IGN Astro 1960 / UTM zone 28N', 'EPSG', 3367, 'PROJCS[\"IGN Astro 1960 / UTM zone 28"
  "N\",GEOGCS[\"IGN Astro 1960\",DATUM[\"IGN Astro 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\""
  "EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4700\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3367\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3368, 1, 'IGN Astro 1960 / UTM zone 29N', 'EPSG', 3368, 'PROJCS[\"IGN Astro 1960 / UTM zone 29"
  "N\",GEOGCS[\"IGN Astro 1960\",DATUM[\"IGN Astro 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\""
  "EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4700\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3368\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3369, 1, 'IGN Astro 1960 / UTM zone 30N', 'EPSG', 3369, 'PROJCS[\"IGN Astro 1960 / UTM zone 30"
  "N\",GEOGCS[\"IGN Astro 1960\",DATUM[\"IGN Astro 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\""
  "EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4700\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3369\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3370, 1, 'NAD27 / UTM zone 59N', 'EPSG', 3370, 'PROJCS[\"NAD27 / UTM zone 59N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"3370\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3371, 1, 'NAD27 / UTM zone 60N', 'EPSG', 3371, 'PROJCS[\"NAD27 / UTM zone 60N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"3371\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3372, 1, 'NAD83 / UTM zone 59N', 'EPSG', 3372, 'PROJCS[\"NAD83 / UTM zone 59N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3372\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3373, 1, 'NAD83 / UTM zone 60N', 'EPSG', 3373, 'PROJCS[\"NAD83 / UTM zone 60N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3373\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3374, 1, 'FD54 / UTM zone 29N', 'EPSG', 3374, 'PROJCS[\"FD54 / UTM zone 29N\",GEOGCS[\"FD54\","
  "DATUM[\"Faroe Datum 1954\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG"
  "\",\"6741\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4741\"]],PROJECTION[\"Transverse Mercato"
  "r\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER["
  "\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"3374\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3375, 1, 'GDM2000 / Peninsula RSO', 'EPSG', 3375, 'PROJCS[\"GDM2000 / Peninsula RSO\",GEOGCS[\""
  "GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJEC"
  "TION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\""
  ",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",102.25,AUTHORITY[\"EPSG\",\"8812\"]],PA"
  "RAMETER[\"Azimuth of initial line\",323.025796466667,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified t"
  "o Skew Grid\",323.130102361111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0.99984,AUTHO"
  "RITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",804671,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"3375\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3376, 1, 'GDM2000 / East Malaysia BRSO', 'EPSG', 3376, 'PROJCS[\"GDM2000 / East Malaysia BRSO\""
  ",GEOGCS[\"GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\""
  "]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projectio"
  "n centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",115,AUTHORITY[\"EPSG\",\"8812"
  "\"]],PARAMETER[\"Azimuth of initial line\",53.31580995,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified"
  " to Skew Grid\",53.1301023611111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0.99984,AUT"
  "HORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"3376\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3377, 1, 'GDM2000 / Johor Grid', 'EPSG', 3377, 'PROJCS[\"GDM2000 / Johor Grid\",GEOGCS[\"GDM20"
  "00\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJECTION[\""
  "Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",2.12167974444445,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",103.427936236111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"False easting\",-14810.562,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8758.32,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3377\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3378, 1, 'GDM2000 / Sembilan and Melaka Grid', 'EPSG', 3378, 'PROJCS[\"GDM2000 / Sembilan and "
  "Melaka Grid\",GEOGCS[\"GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4742\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\""
  ",2.68234763611111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",101.974905041667,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",3673.785,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-"
  "4240.573,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"3378\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3379, 1, 'GDM2000 / Pahang Grid', 'EPSG', 3379, 'PROJCS[\"GDM2000 / Pahang Grid\",GEOGCS[\"GDM"
  "2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJECTIO"
  "N[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",3.76938808888889,AUTHORI"
  "TY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102.368298983333,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"False easting\",-7368.228,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6485.858,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "3379\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3380, 1, 'GDM2000 / Selangor Grid', 'EPSG', 3380, 'PROJCS[\"GDM2000 / Selangor Grid\",GEOGCS[\""
  "GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJEC"
  "TION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",3.68464905,AUTHORITY["
  "\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",101.389107913889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"False easting\",-34836.161,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",56464.049,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"33"
  "80\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3381, 1, 'GDM2000 / Terengganu Grid', 'EPSG', 3381, 'PROJCS[\"GDM2000 / Terengganu Grid\",GEOG"
  "CS[\"GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],P"
  "ROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",4.9762852,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",103.070275625,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"False easting\",19594.245,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3371.895,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"33"
  "81\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3382, 1, 'GDM2000 / Pinang Grid', 'EPSG', 3382, 'PROJCS[\"GDM2000 / Pinang Grid\",GEOGCS[\"GDM"
  "2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJECTIO"
  "N[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",5.42151754166667,AUTHORI"
  "TY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",100.344376963889,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"False easting\",-23.414,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",62.283,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3382\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3383, 1, 'GDM2000 / Kedah and Perlis Grid', 'EPSG', 3383, 'PROJCS[\"GDM2000 / Kedah and Perlis"
  " Grid\",GEOGCS[\"GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4742\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",5.96"
  "467271388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",100.636371111111,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3383\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3384, 1, 'GDM2000 / Perak Grid', 'EPSG', 3384, 'PROJCS[\"GDM2000 / Perak Grid\",GEOGCS[\"GDM20"
  "00\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJECTION[\""
  "Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",4.85906302222222,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",100.815410586111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"False easting\",-1.769,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",133454.779,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3384\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3385, 1, 'GDM2000 / Kelantan Grid', 'EPSG', 3385, 'PROJCS[\"GDM2000 / Kelantan Grid\",GEOGCS[\""
  "GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]],PROJEC"
  "TION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",5.97254365833333,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102.295241669444,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"False easting\",13227.851,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8739.894,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3385\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3386, 1, 'KKJ / Finland zone 0', 'EPSG', 3386, 'PROJCS[\"KKJ / Finland zone 0\",GEOGCS[\"KKJ\""
  ",DATUM[\"Kartastokoordinaattijarjestelma (1966)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.345,-1.376,1.496],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3386\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3387, 1, 'KKJ / Finland zone 5', 'EPSG', 3387, 'PROJCS[\"KKJ / Finland zone 5\",GEOGCS[\"KKJ\""
  ",DATUM[\"Kartastokoordinaattijarjestelma (1966)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.345,-1.376,1.496],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3387\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3388, 1, 'Pulkovo 1942 / Caspian Sea Mercator', 'EPSG', 3388, 'PROJCS[\"Pulkovo 1942 / Caspian"
  " Sea Mercator\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9805\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",42,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Longitude of natural origin\",51,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"none\",NORTH],AXIS[\"none\",EA"
  "ST],AUTHORITY[\"EPSG\",\"3388\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3389, 1, 'Pulkovo 1942 / 3-degree Gauss-Kruger zone 60', 'EPSG', 3389, 'PROJCS[\"Pulkovo 1942 "
  "/ 3-degree Gauss-Kruger zone 60\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",180,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",60500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"338"
  "9\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3390, 1, 'Pulkovo 1995 / 3-degree Gauss-Kruger zone 60', 'EPSG', 3390, 'PROJCS[\"Pulkovo 1995 "
  "/ 3-degree Gauss-Kruger zone 60\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245"
  ",298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",180,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",60500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"3390\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3391, 1, 'Karbala 1979 / UTM zone 37N', 'EPSG', 3391, 'PROJCS[\"Karbala 1979 / UTM zone 37N\","
  "GEOGCS[\"Karbala 1979\",DATUM[\"Karbala 1979\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[70.995,-335.916,262.898,0,0,0,0],AUTHORITY[\"EPSG\",\"6743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4743\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3392, 1, 'Karbala 1979 / UTM zone 38N', 'EPSG', 3392, 'PROJCS[\"Karbala 1979 / UTM zone 38N\","
  "GEOGCS[\"Karbala 1979\",DATUM[\"Karbala 1979\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[70.995,-335.916,262.898,0,0,0,0],AUTHORITY[\"EPSG\",\"6743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4743\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3392\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3393, 1, 'Karbala 1979 / UTM zone 39N', 'EPSG', 3393, 'PROJCS[\"Karbala 1979 / UTM zone 39N\","
  "GEOGCS[\"Karbala 1979\",DATUM[\"Karbala 1979\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[70.995,-335.916,262.898,0,0,0,0],AUTHORITY[\"EPSG\",\"6743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4743\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3394, 1, 'Nahrwan 1934 / Iraq zone', 'EPSG', 3394, 'PROJCS[\"Nahrwan 1934 / Iraq zone\",GEOGCS"
  "[\"Nahrwan 1934\",DATUM[\"Nahrwan 1934\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012"
  "\"]],TOWGS84[-242.2,-144.9,370.3,0,0,0,0],AUTHORITY[\"EPSG\",\"6744\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4744\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Lat"
  "itude of natural origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9987864078,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1166200,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3394\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3395, 1, 'WGS 84 / World Mercator', 'EPSG', 3395, 'PROJCS[\"WGS 84 / World Mercator\",GEOGCS[\""
  "WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"M"
  "ercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"3395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3396, 1, 'PD/83 / 3-degree Gauss-Kruger zone 3', 'EPSG', 3396, 'PROJCS[\"PD/83 / 3-degree Gaus"
  "s-Kruger zone 3\",GEOGCS[\"PD/83\",DATUM[\"Potsdam Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR"
  "ITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4746\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3396\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3397, 1, 'PD/83 / 3-degree Gauss-Kruger zone 4', 'EPSG', 3397, 'PROJCS[\"PD/83 / 3-degree Gaus"
  "s-Kruger zone 4\",GEOGCS[\"PD/83\",DATUM[\"Potsdam Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHOR"
  "ITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4746\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3397\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3398, 1, 'RD/83 / 3-degree Gauss-Kruger zone 4', 'EPSG', 3398, 'PROJCS[\"RD/83 / 3-degree Gaus"
  "s-Kruger zone 4\",GEOGCS[\"RD/83\",DATUM[\"Rauenberg Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTH"
  "ORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4745\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3398\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3399, 1, 'RD/83 / 3-degree Gauss-Kruger zone 5', 'EPSG', 3399, 'PROJCS[\"RD/83 / 3-degree Gaus"
  "s-Kruger zone 5\",GEOGCS[\"RD/83\",DATUM[\"Rauenberg Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTH"
  "ORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4745\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3399\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3400, 1, 'NAD83 / Alberta 10-TM (Forest)', 'EPSG', 3400, 'PROJCS[\"NAD83 / Alberta 10-TM (Fore"
  "st)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9992,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3400\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3401, 1, 'NAD83 / Alberta 10-TM (Resource)', 'EPSG', 3401, 'PROJCS[\"NAD83 / Alberta 10-TM (Re"
  "source)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9992,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3401\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3402, 1, 'NAD83(CSRS) / Alberta 10-TM (Forest)', 'EPSG', 3402, 'PROJCS[\"NAD83(CSRS) / Alberta"
  " 10-TM (Forest)\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0"
  ".0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"461"
  "7\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9992,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3402\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3403, 1, 'NAD83(CSRS) / Alberta 10-TM (Resource)', 'EPSG', 3403, 'PROJCS[\"NAD83(CSRS) / Alber"
  "ta 10-TM (Resource)\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.009650098960270"
  "41,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4617\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9992,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3403\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3404, 1, 'NAD83(HARN) / North Carolina (ftUS)', 'EPSG', 3404, 'PROJCS[\"NAD83(HARN) / North Ca"
  "rolina (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-79,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUTHORIT"
  "Y[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],"
  "PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3404\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3405, 1, 'VN-2000 / UTM zone 48N', 'EPSG', 3405, 'PROJCS[\"VN-2000 / UTM zone 48N\",GEOGCS[\"V"
  "N-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-191."
  "90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"6756\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"3405\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3406, 1, 'VN-2000 / UTM zone 49N', 'EPSG', 3406, 'PROJCS[\"VN-2000 / UTM zone 49N\",GEOGCS[\"V"
  "N-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-191."
  "90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"6756\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"3406\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3407, 1, 'Hong Kong 1963 Grid System', 'EPSG', 3407, 'PROJCS[\"Hong Kong 1963 Grid System\",GE"
  "OGCS[\"Hong Kong 1963\",DATUM[\"Hong Kong 1963\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.26067636926103,AUTHOR"
  "ITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6738\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4738\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",22.312"
  "1333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114.178555555556,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"False easting\",132033.92,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",62565.96"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Clarke\\'s foot\",0.3047972654,AUTHORITY[\"EPSG\",\"9005\"]],AXIS[\"N\",NORTH],A"
  "XIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3407\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3408, 1, 'NSIDC EASE-Grid North', 'EPSG', 3408, 'PROJCS[\"NSIDC EASE-Grid North\",GEOGCS[\"Uns"
  "pecified datum based upon the International 1924 Authalic Sphere\",DATUM[\"Not specified (based on International 1924"
  " Authalic Sphere)\",SPHEROID[\"International 1924 Authalic Sphere\",6371228,0,AUTHORITY[\"EPSG\",\"7057\"]],AUTHORITY"
  "[\"EPSG\",\"6053\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4053\"]],PROJECTION[\"Lambert Azi"
  "muthal Equal Area (Spherical)\",AUTHORITY[\"EPSG\",\"1027\"]],PARAMETER[\"Latitude of natural origin\",90,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting"
  "\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3408\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3409, 1, 'NSIDC EASE-Grid South', 'EPSG', 3409, 'PROJCS[\"NSIDC EASE-Grid South\",GEOGCS[\"Uns"
  "pecified datum based upon the International 1924 Authalic Sphere\",DATUM[\"Not specified (based on International 1924"
  " Authalic Sphere)\",SPHEROID[\"International 1924 Authalic Sphere\",6371228,0,AUTHORITY[\"EPSG\",\"7057\"]],AUTHORITY"
  "[\"EPSG\",\"6053\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4053\"]],PROJECTION[\"Lambert Azi"
  "muthal Equal Area (Spherical)\",AUTHORITY[\"EPSG\",\"1027\"]],PARAMETER[\"Latitude of natural origin\",-90,AUTHORITY["
  "\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easti"
  "ng\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3409\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3410, 1, 'NSIDC EASE-Grid Global', 'EPSG', 3410, 'PROJCS[\"NSIDC EASE-Grid Global\",GEOGCS[\"U"
  "nspecified datum based upon the International 1924 Authalic Sphere\",DATUM[\"Not specified (based on International 19"
  "24 Authalic Sphere)\",SPHEROID[\"International 1924 Authalic Sphere\",6371228,0,AUTHORITY[\"EPSG\",\"7057\"]],AUTHORI"
  "TY[\"EPSG\",\"6053\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4053\"]],PROJECTION[\"Lambert C"
  "ylindrical Equal Area (Spherical)\",AUTHORITY[\"EPSG\",\"9834\"]],PARAMETER[\"Latitude of 1st standard parallel\",30,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"F"
  "alse easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3410\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3411, 1, 'NSIDC Sea Ice Polar Stereographic North', 'EPSG', 3411, 'PROJCS[\"NSIDC Sea Ice Pola"
  "r Stereographic North\",GEOGCS[\"Unspecified datum based upon the Hughes 1980 ellipsoid\",DATUM[\"Not specified (base"
  "d on Hughes 1980 ellipsoid)\",SPHEROID[\"Hughes 1980\",6378273,298.279411123064,AUTHORITY[\"EPSG\",\"7058\"]],AUTHORI"
  "TY[\"EPSG\",\"6054\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4054\"]],PROJECTION[\"Polar Ste"
  "reographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",70,AUTHORITY[\"EPS"
  "G\",\"8832\"]],PARAMETER[\"Longitude of origin\",-45,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"False easting\",0,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3411\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3412, 1, 'NSIDC Sea Ice Polar Stereographic South', 'EPSG', 3412, 'PROJCS[\"NSIDC Sea Ice Pola"
  "r Stereographic South\",GEOGCS[\"Unspecified datum based upon the Hughes 1980 ellipsoid\",DATUM[\"Not specified (base"
  "d on Hughes 1980 ellipsoid)\",SPHEROID[\"Hughes 1980\",6378273,298.279411123064,AUTHORITY[\"EPSG\",\"7058\"]],AUTHORI"
  "TY[\"EPSG\",\"6054\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4054\"]],PROJECTION[\"Polar Ste"
  "reographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",-70,AUTHORITY[\"EP"
  "SG\",\"8832\"]],PARAMETER[\"Longitude of origin\",0,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"False easting\",0,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3412\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3413, 1, 'WGS 84 / NSIDC Sea Ice Polar Stereographic North', 'EPSG', 3413, 'PROJCS[\"WGS 84 / "
  "NSIDC Sea Ice Polar Stereographic North\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\","
  "6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PAR"
  "AMETER[\"Latitude of standard parallel\",70,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",-45,AUTHO"
  "RITY[\"EPSG\",\"8833\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHO"
  "RITY[\"EPSG\",\"3413\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3414, 1, 'SVY21 / Singapore TM', 'EPSG', 3414, 'PROJCS[\"SVY21 / Singapore TM\",GEOGCS[\"SVY21"
  "\",DATUM[\"SVY21\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6757\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4757\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",1.37777777777778,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",103.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",28001.642,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",38744.572,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3414\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3415, 1, 'WGS 72BE / South China Sea Lambert', 'EPSG', 3415, 'PROJCS[\"WGS 72BE / South China "
  "Sea Lambert\",GEOGCS[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUT"
  "HORITY[\"EPSG\",\"7043\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4324\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",21,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",11"
  "4,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",18,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",24,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",50"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3415\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3416, 1, 'ETRS89 / Austria Lambert', 'EPSG', 3416, 'PROJCS[\"ETRS89 / Austria Lambert\",GEOGCS"
  "[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",47.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",13."
  "3444444444444,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3416\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3417, 1, 'NAD83 / Iowa North (ftUS)', 'EPSG', 3417, 'PROJCS[\"NAD83 / Iowa North (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.2777777777778,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280833.3333,AUT"
  "HORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3417\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3418, 1, 'NAD83 / Iowa South (ftUS)', 'EPSG', 3418, 'PROJCS[\"NAD83 / Iowa South (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3418\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3419, 1, 'NAD83 / Kansas North (ftUS)', 'EPSG', 3419, 'PROJCS[\"NAD83 / Kansas North (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3419\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3420, 1, 'NAD83 / Kansas South (ftUS)', 'EPSG', 3420, 'PROJCS[\"NAD83 / Kansas South (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.51111111111"
  "11,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.5666666666667,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1312333.3"
  "333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3420\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3421, 1, 'NAD83 / Nevada East (ftUS)', 'EPSG', 3421, 'PROJCS[\"NAD83 / Nevada East (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.594444444444,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",26246666.6667,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3421\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3422, 1, 'NAD83 / Nevada Central (ftUS)', 'EPSG', 3422, 'PROJCS[\"NAD83 / Nevada Central (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-116.666666666667,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",19685000,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3422\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3423, 1, 'NAD83 / Nevada West (ftUS)', 'EPSG', 3423, 'PROJCS[\"NAD83 / Nevada West (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118.594444444444,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",13123333.3333,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3423\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3424, 1, 'NAD83 / New Jersey (ftUS)', 'EPSG', 3424, 'PROJCS[\"NAD83 / New Jersey (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111111111111,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"3424\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3425, 1, 'NAD83(HARN) / Iowa North (ftUS)', 'EPSG', 3425, 'PROJCS[\"NAD83(HARN) / Iowa North ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.277777"
  "7777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3425\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3426, 1, 'NAD83(HARN) / Iowa South (ftUS)', 'EPSG', 3426, 'PROJCS[\"NAD83(HARN) / Iowa South ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-9"
  "3.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3426\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3427, 1, 'NAD83(HARN) / Kansas North (ftUS)', 'EPSG', 3427, 'PROJCS[\"NAD83(HARN) / Kansas Nor"
  "th (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.7944444444444,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false"
  " origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3427\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3428, 1, 'NAD83(HARN) / Kansas South (ftUS)', 'EPSG', 3428, 'PROJCS[\"NAD83(HARN) / Kansas Sou"
  "th (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.5"
  "666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHOR"
  "ITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3428\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3429, 1, 'NAD83(HARN) / Nevada East (ftUS)', 'EPSG', 3429, 'PROJCS[\"NAD83(HARN) / Nevada East"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "15.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2624666"
  "6.6667,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3429\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3430, 1, 'NAD83(HARN) / Nevada Central (ftUS)', 'EPSG', 3430, 'PROJCS[\"NAD83(HARN) / Nevada C"
  "entral (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-116.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "19685000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3430\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3431, 1, 'NAD83(HARN) / Nevada West (ftUS)', 'EPSG', 3431, 'PROJCS[\"NAD83(HARN) / Nevada West"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "18.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",131233"
  "33.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3431\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3432, 1, 'NAD83(HARN) / New Jersey (ftUS)', 'EPSG', 3432, 'PROJCS[\"NAD83(HARN) / New Jersey ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3432\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3433, 1, 'NAD83 / Arkansas North (ftUS)', 'EPSG', 3433, 'PROJCS[\"NAD83 / Arkansas North (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3433\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3434, 1, 'NAD83 / Arkansas South (ftUS)', 'EPSG', 3434, 'PROJCS[\"NAD83 / Arkansas South (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1312333.3333,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3434\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3435, 1, 'NAD83 / Illinois East (ftUS)', 'EPSG', 3435, 'PROJCS[\"NAD83 / Illinois East (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.3333333333333,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999975,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"3435\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3436, 1, 'NAD83 / Illinois West (ftUS)', 'EPSG', 3436, 'PROJCS[\"NAD83 / Illinois West (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.1777777777778,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",2296583.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3436\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3437, 1, 'NAD83 / New Hampshire (ftUS)', 'EPSG', 3437, 'PROJCS[\"NAD83 / New Hampshire (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.6666666666667,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"3437\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3438, 1, 'NAD83 / Rhode Island (ftUS)', 'EPSG', 3438, 'PROJCS[\"NAD83 / Rhode Island (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.5111111111111,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999375,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3438\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3439, 1, 'PSD93 / UTM zone 39N', 'EPSG', 3439, 'PROJCS[\"PSD93 / UTM zone 39N\",GEOGCS[\"PSD93"
  "\",DATUM[\"PDO Survey Datum 1993\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T"
  "OWGS84[-180.624,-225.516,173.919,-0.81,-1.898,8.336,16.71006],AUTHORITY[\"EPSG\",\"6134\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4134\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3439\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3440, 1, 'PSD93 / UTM zone 40N', 'EPSG', 3440, 'PROJCS[\"PSD93 / UTM zone 40N\",GEOGCS[\"PSD93"
  "\",DATUM[\"PDO Survey Datum 1993\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],T"
  "OWGS84[-180.624,-225.516,173.919,-0.81,-1.898,8.336,16.71006],AUTHORITY[\"EPSG\",\"6134\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4134\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3440\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3441, 1, 'NAD83(HARN) / Arkansas North (ftUS)', 'EPSG', 3441, 'PROJCS[\"NAD83(HARN) / Arkansas"
  " North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.23333333333"
  "33,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3441\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3442, 1, 'NAD83(HARN) / Arkansas South (ftUS)', 'EPSG', 3442, 'PROJCS[\"NAD83(HARN) / Arkansas"
  " South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.76666666666"
  "67,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3442\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3443, 1, 'NAD83(HARN) / Illinois East (ftUS)', 'EPSG', 3443, 'PROJCS[\"NAD83(HARN) / Illinois "
  "East (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-88.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999975,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3443\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3444, 1, 'NAD83(HARN) / Illinois West (ftUS)', 'EPSG', 3444, 'PROJCS[\"NAD83(HARN) / Illinois "
  "West (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-90.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2296583.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3444\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3445, 1, 'NAD83(HARN) / New Hampshire (ftUS)', 'EPSG', 3445, 'PROJCS[\"NAD83(HARN) / New Hamps"
  "hire (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-71.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3445\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3446, 1, 'NAD83(HARN) / Rhode Island (ftUS)', 'EPSG', 3446, 'PROJCS[\"NAD83(HARN) / Rhode Isla"
  "nd (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999375,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3446\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3447, 1, 'ETRS89 / Belgian Lambert 2005', 'EPSG', 3447, 'PROJCS[\"ETRS89 / Belgian Lambert 200"
  "5\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",50.797815,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\","
  "4.35921583333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",51.1666666666667,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",150328,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "166262,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3447\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3448, 1, 'JAD2001 / Jamaica Metric Grid', 'EPSG', 3448, 'PROJCS[\"JAD2001 / Jamaica Metric Gri"
  "d\",GEOGCS[\"JAD2001\",DATUM[\"Jamaica 2001\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4758\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural o"
  "rigin\",18,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-77,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",750000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",650000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3448\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3449, 1, 'JAD2001 / UTM zone 17N', 'EPSG', 3449, 'PROJCS[\"JAD2001 / UTM zone 17N\",GEOGCS[\"J"
  "AD2001\",DATUM[\"Jamaica 2001\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4758\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3449\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3450, 1, 'JAD2001 / UTM zone 18N', 'EPSG', 3450, 'PROJCS[\"JAD2001 / UTM zone 18N\",GEOGCS[\"J"
  "AD2001\",DATUM[\"Jamaica 2001\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4758\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3450\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3451, 1, 'NAD83 / Louisiana North (ftUS)', 'EPSG', 3451, 'PROJCS[\"NAD83 / Louisiana North (ft"
  "US)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latit"
  "ude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92.5111111111111,AU"
  "THORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.6777777777778,AUTHORITY[\"EPSG\",\"882"
  "3\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3451\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3452, 1, 'NAD83 / Louisiana South (ftUS)', 'EPSG', 3452, 'PROJCS[\"NAD83 / Louisiana South (ft"
  "US)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latit"
  "ude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3333333333333,AU"
  "THORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3452\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3453, 1, 'NAD83 / Louisiana Offshore (ftUS)', 'EPSG', 3453, 'PROJCS[\"NAD83 / Louisiana Offsho"
  "re (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",25.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.33333333333"
  "33,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3453\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3455, 1, 'NAD83 / South Dakota South (ftUS)', 'EPSG', 3455, 'PROJCS[\"NAD83 / South Dakota Sou"
  "th (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-10"
  "0.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.4,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3455\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3456, 1, 'NAD83(HARN) / Louisiana North (ftUS)', 'EPSG', 3456, 'PROJCS[\"NAD83(HARN) / Louisia"
  "na North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG"
  "\",\"9802\"]],PARAMETER[\"Latitude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.6777777"
  "777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing "
  "at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"90"
  "03\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3456\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3457, 1, 'NAD83(HARN) / Louisiana South (ftUS)', 'EPSG', 3457, 'PROJCS[\"NAD83(HARN) / Louisia"
  "na South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG"
  "\",\"9802\"]],PARAMETER[\"Latitude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-91.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3457\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3458, 1, 'NAD83(HARN) / South Dakota North (ftUS)', 'EPSG', 3458, 'PROJCS[\"NAD83(HARN) / Sout"
  "h Dakota North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.6944"
  "444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3458\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3459, 1, 'NAD83(HARN) / South Dakota South (ftUS)', 'EPSG', 3459, 'PROJCS[\"NAD83(HARN) / Sout"
  "h Dakota South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paral"
  "lel\",44.4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3459\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3460, 1, 'Fiji 1986 / Fiji Map Grid', 'EPSG', 3460, 'PROJCS[\"Fiji 1986 / Fiji Map Grid\",GEOG"
  "CS[\"Fiji 1986\",DATUM[\"Fiji Geodetic Datum 1986\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],"
  "TOWGS84[0,0,4.5,0,0,0.554,0.2263],AUTHORITY[\"EPSG\",\"6720\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4720\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",-17,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",178.761111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99985,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3460\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3461, 1, 'Dabola 1981 / UTM zone 28N', 'EPSG', 3461, 'PROJCS[\"Dabola 1981 / UTM zone 28N\",GE"
  "OGCS[\"Dabola 1981\",DATUM[\"Dabola 1981\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPS"
  "G\",\"7011\"]],TOWGS84[-83,37,124,0,0,0,0],AUTHORITY[\"EPSG\",\"6155\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4155\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3461\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3462, 1, 'Dabola 1981 / UTM zone 29N', 'EPSG', 3462, 'PROJCS[\"Dabola 1981 / UTM zone 29N\",GE"
  "OGCS[\"Dabola 1981\",DATUM[\"Dabola 1981\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPS"
  "G\",\"7011\"]],TOWGS84[-83,37,124,0,0,0,0],AUTHORITY[\"EPSG\",\"6155\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4155\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3462\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3463, 1, 'NAD83 / Maine CS2000 Central', 'EPSG', 3463, 'PROJCS[\"NAD83 / Maine CS2000 Central\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69.125,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3463\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3464, 1, 'NAD83(HARN) / Maine CS2000 Central', 'EPSG', 3464, 'PROJCS[\"NAD83(HARN) / Maine CS2"
  "000 Central\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-69.125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"3464\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3465, 1, 'NAD83(NSRS2007) / Alabama East', 'EPSG', 3465, 'PROJCS[\"NAD83(NSRS2007) / Alabama E"
  "ast\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",30.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-85.8333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99996,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"3465\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3466, 1, 'NAD83(NSRS2007) / Alabama West', 'EPSG', 3466, 'PROJCS[\"NAD83(NSRS2007) / Alabama W"
  "est\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-87.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"3466\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3467, 1, 'NAD83(NSRS2007) / Alaska Albers', 'EPSG', 3467, 'PROJCS[\"NAD83(NSRS2007) / Alaska A"
  "lbers\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\""
  "]],PARAMETER[\"Latitude of false origin\",50,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-1"
  "54,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",55,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3467\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3468, 1, 'NAD83(NSRS2007) / Alaska zone 1', 'EPSG', 3468, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 1\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY"
  "[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",57,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitu"
  "de of projection centre\",-133.666666666667,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",323.1"
  "30102361111,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",323.130102361111,AUTHORITY["
  "\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0.9999,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False"
  " easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3468\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3469, 1, 'NAD83(NSRS2007) / Alaska zone 2', 'EPSG', 3469, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 2\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-142,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "69\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3470, 1, 'NAD83(NSRS2007) / Alaska zone 3', 'EPSG', 3470, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 3\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-146,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "70\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3471, 1, 'NAD83(NSRS2007) / Alaska zone 4', 'EPSG', 3471, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 4\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "71\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3472, 1, 'NAD83(NSRS2007) / Alaska zone 5', 'EPSG', 3472, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 5\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-154,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "72\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3473, 1, 'NAD83(NSRS2007) / Alaska zone 6', 'EPSG', 3473, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 6\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-158,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "73\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3474, 1, 'NAD83(NSRS2007) / Alaska zone 7', 'EPSG', 3474, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 7\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "74\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3475, 1, 'NAD83(NSRS2007) / Alaska zone 8', 'EPSG', 3475, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 8\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-166,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "75\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3476, 1, 'NAD83(NSRS2007) / Alaska zone 9', 'EPSG', 3476, 'PROJCS[\"NAD83(NSRS2007) / Alaska z"
  "one 9\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-170,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"34"
  "76\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3477, 1, 'NAD83(NSRS2007) / Alaska zone 10', 'EPSG', 3477, 'PROJCS[\"NAD83(NSRS2007) / Alaska "
  "zone 10\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",51,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-176,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",53.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",51.8333333333333,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3477\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3478, 1, 'NAD83(NSRS2007) / Arizona Central', 'EPSG', 3478, 'PROJCS[\"NAD83(NSRS2007) / Arizon"
  "a Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-111.927777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"3478\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3479, 1, 'NAD83(NSRS2007) / Arizona Central (ft)', 'EPSG', 3479, 'PROJCS[\"NAD83(NSRS2007) / A"
  "rizona Central (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-111.927777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3479\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3480, 1, 'NAD83(NSRS2007) / Arizona East', 'EPSG', 3480, 'PROJCS[\"NAD83(NSRS2007) / Arizona E"
  "ast\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-110.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"3480\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3481, 1, 'NAD83(NSRS2007) / Arizona East (ft)', 'EPSG', 3481, 'PROJCS[\"NAD83(NSRS2007) / Ariz"
  "ona East (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-110.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3481\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3482, 1, 'NAD83(NSRS2007) / Arizona West', 'EPSG', 3482, 'PROJCS[\"NAD83(NSRS2007) / Arizona W"
  "est\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-113.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3482\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3483, 1, 'NAD83(NSRS2007) / Arizona West (ft)', 'EPSG', 3483, 'PROJCS[\"NAD83(NSRS2007) / Ariz"
  "ona West (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-113.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"3483\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3484, 1, 'NAD83(NSRS2007) / Arkansas North', 'EPSG', 3484, 'PROJCS[\"NAD83(NSRS2007) / Arkansa"
  "s North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.233333"
  "3333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3484\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3485, 1, 'NAD83(NSRS2007) / Arkansas North (ftUS)', 'EPSG', 3485, 'PROJCS[\"NAD83(NSRS2007) / "
  "Arkansas North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paralle"
  "l\",36.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,A"
  "UTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3485\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3486, 1, 'NAD83(NSRS2007) / Arkansas South', 'EPSG', 3486, 'PROJCS[\"NAD83(NSRS2007) / Arkansa"
  "s South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.766666"
  "6666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",400000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3486\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3487, 1, 'NAD83(NSRS2007) / Arkansas South (ftUS)', 'EPSG', 3487, 'PROJCS[\"NAD83(NSRS2007) / "
  "Arkansas South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paralle"
  "l\",34.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,A"
  "UTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012192"
  "4,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3487\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3488, 1, 'NAD83(NSRS2007) / California Albers', 'EPSG', 3488, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia Albers\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\","
  "\"9822\"]],PARAMETER[\"Latitude of false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false orig"
  "in\",-120,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",40.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",-4000000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3488\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3489, 1, 'NAD83(NSRS2007) / California zone 1', 'EPSG', 3489, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia zone 1\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41"
  ".6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3489\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3490, 1, 'NAD83(NSRS2007) / California zone 1 (ftUS)', 'EPSG', 3490, 'PROJCS[\"NAD83(NSRS2007)"
  " / California zone 1 (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",41.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3490\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3491, 1, 'NAD83(NSRS2007) / California zone 2', 'EPSG', 3491, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia zone 2\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39"
  ".8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northi"
  "ng at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3491\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3492, 1, 'NAD83(NSRS2007) / California zone 2 (ftUS)', 'EPSG', 3492, 'PROJCS[\"NAD83(NSRS2007)"
  " / California zone 2 (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",39.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.33333333"
  "33333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],P"
  "ARAMETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3492\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3493, 1, 'NAD83(NSRS2007) / California zone 3', 'EPSG', 3493, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia zone 3\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.07777777"
  "77778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAM"
  "ETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3493\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3494, 1, 'NAD83(NSRS2007) / California zone 3 (ftUS)', 'EPSG', 3494, 'PROJCS[\"NAD83(NSRS2007)"
  " / California zone 3 (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",37.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0"
  ".30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3494\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3495, 1, 'NAD83(NSRS2007) / California zone 4', 'EPSG', 3495, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia zone 4\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-119,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37"
  ".2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3495\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3496, 1, 'NAD83(NSRS2007) / California zone 4 (ftUS)', 'EPSG', 3496, 'PROJCS[\"NAD83(NSRS2007)"
  " / California zone 4 (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-119,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",37.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3496\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3497, 1, 'NAD83(NSRS2007) / California zone 5', 'EPSG', 3497, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia zone 5\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-118,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35"
  ".4666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northi"
  "ng at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3497\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3498, 1, 'NAD83(NSRS2007) / California zone 5 (ftUS)', 'EPSG', 3498, 'PROJCS[\"NAD83(NSRS2007)"
  " / California zone 5 (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-118,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",35.4666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.03333333"
  "33333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],P"
  "ARAMETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3498\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3499, 1, 'NAD83(NSRS2007) / California zone 6', 'EPSG', 3499, 'PROJCS[\"NAD83(NSRS2007) / Cali"
  "fornia zone 6\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-116.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",33.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3499\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3500, 1, 'NAD83(NSRS2007) / California zone 6 (ftUS)', 'EPSG', 3500, 'PROJCS[\"NAD83(NSRS2007)"
  " / California zone 6 (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-116.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standa"
  "rd parallel\",33.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.79444"
  "44444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]"
  "],PARAMETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3500\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3501, 1, 'NAD83(NSRS2007) / Colorado Central', 'EPSG', 3501, 'PROJCS[\"NAD83(NSRS2007) / Color"
  "ado Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",39.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",304800.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3501\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3502, 1, 'NAD83(NSRS2007) / Colorado Central (ftUS)', 'EPSG', 3502, 'PROJCS[\"NAD83(NSRS2007) "
  "/ Colorado Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",39.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3502\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3503, 1, 'NAD83(NSRS2007) / Colorado North', 'EPSG', 3503, 'PROJCS[\"NAD83(NSRS2007) / Colorad"
  "o North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",40.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",304800.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3503\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3504, 1, 'NAD83(NSRS2007) / Colorado North (ftUS)', 'EPSG', 3504, 'PROJCS[\"NAD83(NSRS2007) / "
  "Colorado North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",40.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7"
  "277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]"
  "],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601"
  "21924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3504\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3505, 1, 'NAD83(NSRS2007) / Colorado South', 'EPSG', 3505, 'PROJCS[\"NAD83(NSRS2007) / Colorad"
  "o South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",304800.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3505\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3506, 1, 'NAD83(NSRS2007) / Colorado South (ftUS)', 'EPSG', 3506, 'PROJCS[\"NAD83(NSRS2007) / "
  "Colorado South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2"
  "333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]"
  "],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601"
  "21924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3506\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3507, 1, 'NAD83(NSRS2007) / Connecticut', 'EPSG', 3507, 'PROJCS[\"NAD83(NSRS2007) / Connecticu"
  "t\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-72.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.86666666666"
  "67,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",304800.6096,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",152400.3048,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3507\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3508, 1, 'NAD83(NSRS2007) / Connecticut (ftUS)', 'EPSG', 3508, 'PROJCS[\"NAD83(NSRS2007) / Con"
  "necticut (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-72.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",41.8666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3508\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3509, 1, 'NAD83(NSRS2007) / Delaware', 'EPSG', 3509, 'PROJCS[\"NAD83(NSRS2007) / Delaware\",GE"
  "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.41"
  "66666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3509\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3510, 1, 'NAD83(NSRS2007) / Delaware (ftUS)', 'EPSG', 3510, 'PROJCS[\"NAD83(NSRS2007) / Delawa"
  "re (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-75.4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3510\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3511, 1, 'NAD83(NSRS2007) / Florida East', 'EPSG', 3511, 'PROJCS[\"NAD83(NSRS2007) / Florida E"
  "ast\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3511\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3512, 1, 'NAD83(NSRS2007) / Florida East (ftUS)', 'EPSG', 3512, 'PROJCS[\"NAD83(NSRS2007) / Fl"
  "orida East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941"
  "177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3512\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3513, 1, 'NAD83(NSRS2007) / Florida GDL Albers', 'EPSG', 3513, 'PROJCS[\"NAD83(NSRS2007) / Flo"
  "rida GDL Albers\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\""
  ",\"9822\"]],PARAMETER[\"Latitude of false origin\",24,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",-84,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",24,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false "
  "origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3513\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3514, 1, 'NAD83(NSRS2007) / Florida North', 'EPSG', 3514, 'PROJCS[\"NAD83(NSRS2007) / Florida "
  "North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-84.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.76111111"
  "11111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3514\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3515, 1, 'NAD83(NSRS2007) / Florida North (ftUS)', 'EPSG', 3515, 'PROJCS[\"NAD83(NSRS2007) / F"
  "lorida North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-84.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",30.7611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"N"
  "orthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3515\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3516, 1, 'NAD83(NSRS2007) / Florida West', 'EPSG', 3516, 'PROJCS[\"NAD83(NSRS2007) / Florida W"
  "est\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-82,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3516\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3517, 1, 'NAD83(NSRS2007) / Florida West (ftUS)', 'EPSG', 3517, 'PROJCS[\"NAD83(NSRS2007) / Fl"
  "orida West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-82,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941"
  "177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3517\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3518, 1, 'NAD83(NSRS2007) / Georgia East', 'EPSG', 3518, 'PROJCS[\"NAD83(NSRS2007) / Georgia E"
  "ast\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-82.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"3518\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3519, 1, 'NAD83(NSRS2007) / Georgia East (ftUS)', 'EPSG', 3519, 'PROJCS[\"NAD83(NSRS2007) / Ge"
  "orgia East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-82.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3520, 1, 'NAD83(NSRS2007) / Georgia West', 'EPSG', 3520, 'PROJCS[\"NAD83(NSRS2007) / Georgia W"
  "est\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-84.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"3520\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3521, 1, 'NAD83(NSRS2007) / Georgia West (ftUS)', 'EPSG', 3521, 'PROJCS[\"NAD83(NSRS2007) / Ge"
  "orgia West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-84.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3522, 1, 'NAD83(NSRS2007) / Idaho Central', 'EPSG', 3522, 'PROJCS[\"NAD83(NSRS2007) / Idaho Ce"
  "ntral\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"3522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3523, 1, 'NAD83(NSRS2007) / Idaho Central (ftUS)', 'EPSG', 3523, 'PROJCS[\"NAD83(NSRS2007) / I"
  "daho Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99994"
  "7368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3524, 1, 'NAD83(NSRS2007) / Idaho East', 'EPSG', 3524, 'PROJCS[\"NAD83(NSRS2007) / Idaho East\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-112.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"3524\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3525, 1, 'NAD83(NSRS2007) / Idaho East (ftUS)', 'EPSG', 3525, 'PROJCS[\"NAD83(NSRS2007) / Idah"
  "o East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-112.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3525\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3526, 1, 'NAD83(NSRS2007) / Idaho West', 'EPSG', 3526, 'PROJCS[\"NAD83(NSRS2007) / Idaho West\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-115.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"3526\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3527, 1, 'NAD83(NSRS2007) / Idaho West (ftUS)', 'EPSG', 3527, 'PROJCS[\"NAD83(NSRS2007) / Idah"
  "o West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-115.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999333"
  "33,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2624666.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3527\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3528, 1, 'NAD83(NSRS2007) / Illinois East', 'EPSG', 3528, 'PROJCS[\"NAD83(NSRS2007) / Illinois"
  " East\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-88.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  "75,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"3528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3529, 1, 'NAD83(NSRS2007) / Illinois East (ftUS)', 'EPSG', 3529, 'PROJCS[\"NAD83(NSRS2007) / I"
  "llinois East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-88.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\","
  "\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3530, 1, 'NAD83(NSRS2007) / Illinois West', 'EPSG', 3530, 'PROJCS[\"NAD83(NSRS2007) / Illinois"
  " West\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-90.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  "41177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3530\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3531, 1, 'NAD83(NSRS2007) / Illinois West (ftUS)', 'EPSG', 3531, 'PROJCS[\"NAD83(NSRS2007) / I"
  "llinois West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-90.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2296583.3333,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3532, 1, 'NAD83(NSRS2007) / Indiana East', 'EPSG', 3532, 'PROJCS[\"NAD83(NSRS2007) / Indiana E"
  "ast\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-85.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966"
  "667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",250000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3532\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3533, 1, 'NAD83(NSRS2007) / Indiana East (ftUS)', 'EPSG', 3533, 'PROJCS[\"NAD83(NSRS2007) / In"
  "diana East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-85.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.333,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",820208.333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3534, 1, 'NAD83(NSRS2007) / Indiana West', 'EPSG', 3534, 'PROJCS[\"NAD83(NSRS2007) / Indiana W"
  "est\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-87.0944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966"
  "667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",250000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3534\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3535, 1, 'NAD83(NSRS2007) / Indiana West (ftUS)', 'EPSG', 3535, 'PROJCS[\"NAD83(NSRS2007) / In"
  "diana West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-87.0944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",820208.333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHOR"
  "ITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3536, 1, 'NAD83(NSRS2007) / Iowa North', 'EPSG', 3536, 'PROJCS[\"NAD83(NSRS2007) / Iowa North\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.2"
  "777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3536\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3537, 1, 'NAD83(NSRS2007) / Iowa North (ftUS)', 'EPSG', 3537, 'PROJCS[\"NAD83(NSRS2007) / Iowa"
  " North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",43.2777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.07777777"
  "77778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAM"
  "ETER[\"Northing at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121"
  "924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3537\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3538, 1, 'NAD83(NSRS2007) / Iowa South', 'EPSG', 3538, 'PROJCS[\"NAD83(NSRS2007) / Iowa South\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3539, 1, 'NAD83(NSRS2007) / Iowa South (ftUS)', 'EPSG', 3539, 'PROJCS[\"NAD83(NSRS2007) / Iowa"
  " South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41."
  "7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3540, 1, 'NAD83(NSRS2007) / Kansas North', 'EPSG', 3540, 'PROJCS[\"NAD83(NSRS2007) / Kansas No"
  "rth\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.79444444444"
  "44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false o"
  "rigin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"3540\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3541, 1, 'NAD83(NSRS2007) / Kansas North (ftUS)', 'EPSG', 3541, 'PROJCS[\"NAD83(NSRS2007) / Ka"
  "nsas North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",39.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3541\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3542, 1, 'NAD83(NSRS2007) / Kansas South', 'EPSG', 3542, 'PROJCS[\"NAD83(NSRS2007) / Kansas So"
  "uth\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "38.5666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"North"
  "ing at false origin\",400000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3542\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3543, 1, 'NAD83(NSRS2007) / Kansas South (ftUS)', 'EPSG', 3543, 'PROJCS[\"NAD83(NSRS2007) / Ka"
  "nsas South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",38.5666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777"
  "777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3543\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3544, 1, 'NAD83(NSRS2007) / Kentucky North', 'EPSG', 3544, 'PROJCS[\"NAD83(NSRS2007) / Kentuck"
  "y North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-84.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.977"
  "7777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3544\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3545, 1, 'NAD83(NSRS2007) / Kentucky North (ftUS)', 'EPSG', 3545, 'PROJCS[\"NAD83(NSRS2007) / "
  "Kentucky North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-84.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.977777777777"
  "8,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAM"
  "ETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3545\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3546, 1, 'NAD83(NSRS2007) / Kentucky Single Zone', 'EPSG', 3546, 'PROJCS[\"NAD83(NSRS2007) / K"
  "entucky Single Zone\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PA"
  "RAMETER[\"Longitude of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paral"
  "lel\",37.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778"
  ",AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3546\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3547, 1, 'NAD83(NSRS2007) / Kentucky Single Zone (ftUS)', 'EPSG', 3547, 'PROJCS[\"NAD83(NSRS20"
  "07) / Kentucky Single Zone (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st s"
  "tandard parallel\",37.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38."
  "6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006"
  "0960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3547\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3548, 1, 'NAD83(NSRS2007) / Kentucky South', 'EPSG', 3548, 'PROJCS[\"NAD83(NSRS2007) / Kentuck"
  "y South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.933"
  "3333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3548\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3549, 1, 'NAD83(NSRS2007) / Kentucky South (ftUS)', 'EPSG', 3549, 'PROJCS[\"NAD83(NSRS2007) / "
  "Kentucky South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",37.9333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.744444444444"
  "4,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAM"
  "ETER[\"Northing at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601219"
  "24,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3549\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3550, 1, 'NAD83(NSRS2007) / Louisiana North', 'EPSG', 3550, 'PROJCS[\"NAD83(NSRS2007) / Louisi"
  "ana North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.677"
  "7777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3550\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3551, 1, 'NAD83(NSRS2007) / Louisiana North (ftUS)', 'EPSG', 3551, 'PROJCS[\"NAD83(NSRS2007) /"
  " Louisiana North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard pa"
  "rallel\",32.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666"
  "667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3551\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3552, 1, 'NAD83(NSRS2007) / Louisiana South', 'EPSG', 3552, 'PROJCS[\"NAD83(NSRS2007) / Louisi"
  "ana South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-91.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3552\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3553, 1, 'NAD83(NSRS2007) / Louisiana South (ftUS)', 'EPSG', 3553, 'PROJCS[\"NAD83(NSRS2007) /"
  " Louisiana South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",-91.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard pa"
  "rallel\",30.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3553\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3554, 1, 'NAD83(NSRS2007) / Maine CS2000 Central', 'EPSG', 3554, 'PROJCS[\"NAD83(NSRS2007) / M"
  "aine CS2000 Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-69.125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3554\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3555, 1, 'NAD83(NSRS2007) / Maine CS2000 East', 'EPSG', 3555, 'PROJCS[\"NAD83(NSRS2007) / Main"
  "e CS2000 East\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-67.875,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"3555\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3556, 1, 'NAD83(NSRS2007) / Maine CS2000 West', 'EPSG', 3556, 'PROJCS[\"NAD83(NSRS2007) / Main"
  "e CS2000 West\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-70.375,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"3556\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3557, 1, 'NAD83(NSRS2007) / Maine East', 'EPSG', 3557, 'PROJCS[\"NAD83(NSRS2007) / Maine East\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3557\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3558, 1, 'NAD83(NSRS2007) / Maine West', 'EPSG', 3558, 'PROJCS[\"NAD83(NSRS2007) / Maine West\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-70.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"3558\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3559, 1, 'NAD83(NSRS2007) / Maryland', 'EPSG', 3559, 'PROJCS[\"NAD83(NSRS2007) / Maryland\",GE"
  "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802"
  "\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of f"
  "alse origin\",-77,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"3559\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3560, 1, 'NAD83 / Utah North (ftUS)', 'EPSG', 3560, 'PROJCS[\"NAD83 / Utah North (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280833.333"
  "3,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3560\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3561, 1, 'Old Hawaiian / Hawaii zone 1', 'EPSG', 3561, 'PROJCS[\"Old Hawaiian / Hawaii zone 1\""
  ",GEOGCS[\"Old Hawaiian\",DATUM[\"Old Hawaiian\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG"
  "\",\"7008\"]],TOWGS84[61,-285,-181,0,0,0,0],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4135\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",18.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-155.5,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"3561\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3562, 1, 'Old Hawaiian / Hawaii zone 2', 'EPSG', 3562, 'PROJCS[\"Old Hawaiian / Hawaii zone 2\""
  ",GEOGCS[\"Old Hawaiian\",DATUM[\"Old Hawaiian\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG"
  "\",\"7008\"]],TOWGS84[61,-285,-181,0,0,0,0],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4135\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",20.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-156.666666"
  "666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3562\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3563, 1, 'Old Hawaiian / Hawaii zone 3', 'EPSG', 3563, 'PROJCS[\"Old Hawaiian / Hawaii zone 3\""
  ",GEOGCS[\"Old Hawaiian\",DATUM[\"Old Hawaiian\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG"
  "\",\"7008\"]],TOWGS84[61,-285,-181,0,0,0,0],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4135\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",21.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-158,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"3563\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3564, 1, 'Old Hawaiian / Hawaii zone 4', 'EPSG', 3564, 'PROJCS[\"Old Hawaiian / Hawaii zone 4\""
  ",GEOGCS[\"Old Hawaiian\",DATUM[\"Old Hawaiian\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG"
  "\",\"7008\"]],TOWGS84[61,-285,-181,0,0,0,0],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4135\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",21.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159.5,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"3564\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3565, 1, 'Old Hawaiian / Hawaii zone 5', 'EPSG', 3565, 'PROJCS[\"Old Hawaiian / Hawaii zone 5\""
  ",GEOGCS[\"Old Hawaiian\",DATUM[\"Old Hawaiian\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG"
  "\",\"7008\"]],TOWGS84[61,-285,-181,0,0,0,0],AUTHORITY[\"EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4135\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",21.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-160.177777"
  "777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3565\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3566, 1, 'NAD83 / Utah Central (ftUS)', 'EPSG', 3566, 'PROJCS[\"NAD83 / Utah Central (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.5111111111"
  "11,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6561666.6667,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3566\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3567, 1, 'NAD83 / Utah South (ftUS)', 'EPSG', 3567, 'PROJCS[\"NAD83 / Utah South (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fa"
  "lse origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",9842500,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3567\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3568, 1, 'NAD83(HARN) / Utah North (ftUS)', 'EPSG', 3568, 'PROJCS[\"NAD83(HARN) / Utah North ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.794444"
  "4444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3568\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3569, 1, 'NAD83(HARN) / Utah Central (ftUS)', 'EPSG', 3569, 'PROJCS[\"NAD83(HARN) / Utah Centr"
  "al (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.6"
  "5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fal"
  "se origin\",6561666.6667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3569\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3570, 1, 'NAD83(HARN) / Utah South (ftUS)', 'EPSG', 3570, 'PROJCS[\"NAD83(HARN) / Utah South ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false o"
  "rigin\",9842500,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3570\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3571, 1, 'WGS 84 / North Pole LAEA Bering Sea', 'EPSG', 3571, 'PROJCS[\"WGS 84 / North Pole LA"
  "EA Bering Sea\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTH"
  "ORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4326\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural "
  "origin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",180,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3571\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3572, 1, 'WGS 84 / North Pole LAEA Alaska', 'EPSG', 3572, 'PROJCS[\"WGS 84 / North Pole LAEA A"
  "laska\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\""
  "]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural origin\""
  ",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-150,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3572\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3573, 1, 'WGS 84 / North Pole LAEA Canada', 'EPSG', 3573, 'PROJCS[\"WGS 84 / North Pole LAEA C"
  "anada\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\""
  "]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural origin\""
  ",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-100,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3573\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3574, 1, 'WGS 84 / North Pole LAEA Atlantic', 'EPSG', 3574, 'PROJCS[\"WGS 84 / North Pole LAEA"
  " Atlantic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"43"
  "26\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural orig"
  "in\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-40,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3574\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3575, 1, 'WGS 84 / North Pole LAEA Europe', 'EPSG', 3575, 'PROJCS[\"WGS 84 / North Pole LAEA E"
  "urope\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\""
  "]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural origin\""
  ",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3575\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3576, 1, 'WGS 84 / North Pole LAEA Russia', 'EPSG', 3576, 'PROJCS[\"WGS 84 / North Pole LAEA R"
  "ussia\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\""
  "]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural origin\""
  ",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3576\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3577, 1, 'GDA94 / Australian Albers', 'EPSG', 3577, 'PROJCS[\"GDA94 / Australian Albers\",GEOG"
  "CS[\"GDA94\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS"
  "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4283\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",132,AUTHORITY[\"EPSG\",\"8822\"]],PARA"
  "METER[\"Latitude of 1st standard parallel\",-18,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard pa"
  "rallel\",-36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3577\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3578, 1, 'NAD83 / Yukon Albers', 'EPSG', 3578, 'PROJCS[\"NAD83 / Yukon Albers\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",59,AUTHORITY"
  "[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-132.5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitud"
  "e of 1st standard parallel\",61.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard para"
  "llel\",68,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3578\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3579, 1, 'NAD83(CSRS) / Yukon Albers', 'EPSG', 3579, 'PROJCS[\"NAD83(CSRS) / Yukon Albers\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"A"
  "lbers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",59,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-132.5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",61.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",68,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3579\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3580, 1, 'NAD83 / NWT Lambert', 'EPSG', 3580, 'PROJCS[\"NAD83 / NWT Lambert\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-112,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",62,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",70,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"N"
  "orthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",E"
  "AST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3580\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3581, 1, 'NAD83(CSRS) / NWT Lambert', 'EPSG', 3581, 'PROJCS[\"NAD83(CSRS) / NWT Lambert\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Lam"
  "bert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8821\"]],PARAMETER[\"Longitude of false origin\",-112,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",62,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"3581\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3582, 1, 'NAD83(NSRS2007) / Maryland (ftUS)', 'EPSG', 3582, 'PROJCS[\"NAD83(NSRS2007) / Maryla"
  "nd (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-77,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AU"
  "THORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"882"
  "4\"]],PARAMETER[\"Easting at false origin\",1312333.333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false o"
  "rigin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3582\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3583, 1, 'NAD83(NSRS2007) / Massachusetts Island', 'EPSG', 3583, 'PROJCS[\"NAD83(NSRS2007) / M"
  "assachusetts Island\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",41.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3583\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3584, 1, 'NAD83(NSRS2007) / Massachusetts Island (ftUS)', 'EPSG', 3584, 'PROJCS[\"NAD83(NSRS20"
  "07) / Massachusetts Island (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PAR"
  "AMETER[\"Longitude of false origin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",41.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.283"
  "3333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924"
  ",AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3584\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3585, 1, 'NAD83(NSRS2007) / Massachusetts Mainland', 'EPSG', 3585, 'PROJCS[\"NAD83(NSRS2007) /"
  " Massachusetts Mainland\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",42.6944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.727777777777"
  "8,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",750000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3585\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3586, 1, 'NAD83(NSRS2007) / Massachusetts Mainland (ftUS)', 'EPSG', 3586, 'PROJCS[\"NAD83(NSRS"
  "2007) / Massachusetts Mainland (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "007)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic C"
  "onformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st"
  " standard parallel\",42.6944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",4"
  "1.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",656166.667,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",2460625,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3586\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3587, 1, 'NAD83(NSRS2007) / Michigan Central', 'EPSG', 3587, 'PROJCS[\"NAD83(NSRS2007) / Michi"
  "gan Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",45.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY["
  "\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing "
  "at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3587\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3588, 1, 'NAD83(NSRS2007) / Michigan Central (ft)', 'EPSG', 3588, 'PROJCS[\"NAD83(NSRS2007) / "
  "Michigan Central (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",45.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444"
  ",AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",19685039.37,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3588\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3589, 1, 'NAD83(NSRS2007) / Michigan North', 'EPSG', 3589, 'PROJCS[\"NAD83(NSRS2007) / Michiga"
  "n North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-87,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.094444"
  "4444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",8000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3589\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3590, 1, 'NAD83(NSRS2007) / Michigan North (ft)', 'EPSG', 3590, 'PROJCS[\"NAD83(NSRS2007) / Mi"
  "chigan North (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-87,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",47.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",26246719.16,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3590\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3591, 1, 'NAD83(NSRS2007) / Michigan Oblique Mercator', 'EPSG', 3591, 'PROJCS[\"NAD83(NSRS2007"
  ") / Michigan Oblique Mercator\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Hotine Oblique Mercator"
  " (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",45.3091666666667,AUTHORITY[\""
  "EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",-86,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth "
  "of initial line\",337.25556,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",337.25556,A"
  "UTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0.9996,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETE"
  "R[\"False easting\",2546731.496,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4354009.816,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3591\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3592, 1, 'NAD83(NSRS2007) / Michigan South', 'EPSG', 3592, 'PROJCS[\"NAD83(NSRS2007) / Michiga"
  "n South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",43.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3592\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3593, 1, 'NAD83(NSRS2007) / Michigan South (ft)', 'EPSG', 3593, 'PROJCS[\"NAD83(NSRS2007) / Mi"
  "chigan South (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",43.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",13123359.58,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3593\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3594, 1, 'NAD83(NSRS2007) / Minnesota Central', 'EPSG', 3594, 'PROJCS[\"NAD83(NSRS2007) / Minn"
  "esota Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-94.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORIT"
  "Y[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],"
  "PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100"
  "000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"3594\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3595, 1, 'NAD83(NSRS2007) / Minnesota North', 'EPSG', 3595, 'PROJCS[\"NAD83(NSRS2007) / Minnes"
  "ota North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-93.1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.633"
  "3333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3595\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3596, 1, 'NAD83(NSRS2007) / Minnesota South', 'EPSG', 3596, 'PROJCS[\"NAD83(NSRS2007) / Minnes"
  "ota South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fa"
  "lse origin\",-94,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "100000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3596\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3597, 1, 'NAD83(NSRS2007) / Mississippi East', 'EPSG', 3597, 'PROJCS[\"NAD83(NSRS2007) / Missi"
  "ssippi East\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-88.8333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3597\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3598, 1, 'NAD83(NSRS2007) / Mississippi East (ftUS)', 'EPSG', 3598, 'PROJCS[\"NAD83(NSRS2007) "
  "/ Mississippi East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",-88.8333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3598\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3599, 1, 'NAD83(NSRS2007) / Mississippi West', 'EPSG', 3599, 'PROJCS[\"NAD83(NSRS2007) / Missi"
  "ssippi West\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-90.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3599\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3600, 1, 'NAD83(NSRS2007) / Mississippi West (ftUS)', 'EPSG', 3600, 'PROJCS[\"NAD83(NSRS2007) "
  "/ Mississippi West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",-90.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3601, 1, 'NAD83(NSRS2007) / Missouri Central', 'EPSG', 3601, 'PROJCS[\"NAD83(NSRS2007) / Misso"
  "uri Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0."
  "999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3601\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3602, 1, 'NAD83(NSRS2007) / Missouri East', 'EPSG', 3602, 'PROJCS[\"NAD83(NSRS2007) / Missouri"
  " East\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-90.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  "33333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3602\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3603, 1, 'NAD83(NSRS2007) / Missouri West', 'EPSG', 3603, 'PROJCS[\"NAD83(NSRS2007) / Missouri"
  " West\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",36.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-94.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  "41177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",850000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3603\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3604, 1, 'NAD83(NSRS2007) / Montana', 'EPSG', 3604, 'PROJCS[\"NAD83(NSRS2007) / Montana\",GEOG"
  "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-109.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3604\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3605, 1, 'NAD83(NSRS2007) / Montana (ft)', 'EPSG', 3605, 'PROJCS[\"NAD83(NSRS2007) / Montana ("
  "ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-109.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PAR"
  "AMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"3605\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3606, 1, 'NAD83(NSRS2007) / Nebraska', 'EPSG', 3606, 'PROJCS[\"NAD83(NSRS2007) / Nebraska\",GE"
  "OGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802"
  "\"]],PARAMETER[\"Latitude of false origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of f"
  "alse origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3606\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3607, 1, 'NAD83(NSRS2007) / Nevada Central', 'EPSG', 3607, 'PROJCS[\"NAD83(NSRS2007) / Nevada "
  "Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-116.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6"
  "000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3607\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3608, 1, 'NAD83(NSRS2007) / Nevada Central (ftUS)', 'EPSG', 3608, 'PROJCS[\"NAD83(NSRS2007) / "
  "Nevada Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY["
  "\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-116.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0."
  "9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",19685000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3608\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3609, 1, 'NAD83(NSRS2007) / Nevada East', 'EPSG', 3609, 'PROJCS[\"NAD83(NSRS2007) / Nevada Eas"
  "t\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-115.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8000000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3609\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3610, 1, 'NAD83(NSRS2007) / Nevada East (ftUS)', 'EPSG', 3610, 'PROJCS[\"NAD83(NSRS2007) / Nev"
  "ada East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-115.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",26246666.6667,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3610\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3611, 1, 'NAD83(NSRS2007) / Nevada West', 'EPSG', 3611, 'PROJCS[\"NAD83(NSRS2007) / Nevada Wes"
  "t\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-118.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4000000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3611\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3612, 1, 'NAD83(NSRS2007) / Nevada West (ftUS)', 'EPSG', 3612, 'PROJCS[\"NAD83(NSRS2007) / Nev"
  "ada West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-118.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",13123333.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3612\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3613, 1, 'NAD83(NSRS2007) / New Hampshire', 'EPSG', 3613, 'PROJCS[\"NAD83(NSRS2007) / New Hamp"
  "shire\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-71.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  "66667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3613\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3614, 1, 'NAD83(NSRS2007) / New Hampshire (ftUS)', 'EPSG', 3614, 'PROJCS[\"NAD83(NSRS2007) / N"
  "ew Hampshire (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-71.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3614\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3615, 1, 'NAD83(NSRS2007) / New Jersey', 'EPSG', 3615, 'PROJCS[\"NAD83(NSRS2007) / New Jersey\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3615\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3616, 1, 'NAD83(NSRS2007) / New Jersey (ftUS)', 'EPSG', 3616, 'PROJCS[\"NAD83(NSRS2007) / New "
  "Jersey (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3616\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3617, 1, 'NAD83(NSRS2007) / New Mexico Central', 'EPSG', 3617, 'PROJCS[\"NAD83(NSRS2007) / New"
  " Mexico Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",-106.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"3617\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3618, 1, 'NAD83(NSRS2007) / New Mexico Central (ftUS)', 'EPSG', 3618, 'PROJCS[\"NAD83(NSRS2007"
  ") / New Mexico Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-106.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3618\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3619, 1, 'NAD83(NSRS2007) / New Mexico East', 'EPSG', 3619, 'PROJCS[\"NAD83(NSRS2007) / New Me"
  "xico East\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-104.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",165000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"3619\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3620, 1, 'NAD83(NSRS2007) / New Mexico East (ftUS)', 'EPSG', 3620, 'PROJCS[\"NAD83(NSRS2007) /"
  " New Mexico East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-104.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0."
  "999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",541337.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3620\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3621, 1, 'NAD83(NSRS2007) / New Mexico West', 'EPSG', 3621, 'PROJCS[\"NAD83(NSRS2007) / New Me"
  "xico West\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-107.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",830000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"3621\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3622, 1, 'NAD83(NSRS2007) / New Mexico West (ftUS)', 'EPSG', 3622, 'PROJCS[\"NAD83(NSRS2007) /"
  " New Mexico West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-107.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0."
  "999916667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2723091.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\","
  "\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3622\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3623, 1, 'NAD83(NSRS2007) / New York Central', 'EPSG', 3623, 'PROJCS[\"NAD83(NSRS2007) / New Y"
  "ork Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-76.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3623\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3624, 1, 'NAD83(NSRS2007) / New York Central (ftUS)', 'EPSG', 3624, 'PROJCS[\"NAD83(NSRS2007) "
  "/ New York Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-76.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\","
  "0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",820208.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3624\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3625, 1, 'NAD83(NSRS2007) / New York East', 'EPSG', 3625, 'PROJCS[\"NAD83(NSRS2007) / New York"
  " East\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"3625\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3626, 1, 'NAD83(NSRS2007) / New York East (ftUS)', 'EPSG', 3626, 'PROJCS[\"NAD83(NSRS2007) / N"
  "ew York East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3626\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3627, 1, 'NAD83(NSRS2007) / New York Long Island', 'EPSG', 3627, 'PROJCS[\"NAD83(NSRS2007) / N"
  "ew York Long Island\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PA"
  "RAMETER[\"Longitude of false origin\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",41.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3627\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3628, 1, 'NAD83(NSRS2007) / New York Long Island (ftUS)', 'EPSG', 3628, 'PROJCS[\"NAD83(NSRS20"
  "07) / New York Long Island (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",41.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.677"
  "7777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],P"
  "ARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3628\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3629, 1, 'NAD83(NSRS2007) / New York West', 'EPSG', 3629, 'PROJCS[\"NAD83(NSRS2007) / New York"
  " West\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-78.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",350000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3629\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3630, 1, 'NAD83(NSRS2007) / New York West (ftUS)', 'EPSG', 3630, 'PROJCS[\"NAD83(NSRS2007) / N"
  "ew York West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-78.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99993"
  "75,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1148291.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3630\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3631, 1, 'NAD83(NSRS2007) / North Carolina', 'EPSG', 3631, 'PROJCS[\"NAD83(NSRS2007) / North C"
  "arolina\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-79,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",609601.22,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3631\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3632, 1, 'NAD83(NSRS2007) / North Carolina (ftUS)', 'EPSG', 3632, 'PROJCS[\"NAD83(NSRS2007) / "
  "North Carolina (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-79,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666"
  "666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3632\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3633, 1, 'NAD83(NSRS2007) / North Dakota North', 'EPSG', 3633, 'PROJCS[\"NAD83(NSRS2007) / Nor"
  "th Dakota North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitud"
  "e of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",4"
  "8.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHOR"
  "ITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northi"
  "ng at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3633\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3634, 1, 'NAD83(NSRS2007) / North Dakota North (ft)', 'EPSG', 3634, 'PROJCS[\"NAD83(NSRS2007) "
  "/ North Dakota North (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard pa"
  "rallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444"
  "444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3634\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3635, 1, 'NAD83(NSRS2007) / North Dakota South', 'EPSG', 3635, 'PROJCS[\"NAD83(NSRS2007) / Nor"
  "th Dakota South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",47.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.194444"
  "4444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3635\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3636, 1, 'NAD83(NSRS2007) / North Dakota South (ft)', 'EPSG', 3636, 'PROJCS[\"NAD83(NSRS2007) "
  "/ North Dakota South (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",47.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "46.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG"
  "\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3636\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3637, 1, 'NAD83(NSRS2007) / Ohio North', 'EPSG', 3637, 'PROJCS[\"NAD83(NSRS2007) / Ohio North\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7"
  ",AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"3637\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3638, 1, 'NAD83(NSRS2007) / Ohio South', 'EPSG', 3638, 'PROJCS[\"NAD83(NSRS2007) / Ohio South\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3638\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3639, 1, 'NAD83(NSRS2007) / Oklahoma North', 'EPSG', 3639, 'PROJCS[\"NAD83(NSRS2007) / Oklahom"
  "a North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"3639\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3640, 1, 'NAD83(NSRS2007) / Oklahoma North (ftUS)', 'EPSG', 3640, 'PROJCS[\"NAD83(NSRS2007) / "
  "Oklahoma North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lon"
  "gitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666"
  "666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3640\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3641, 1, 'NAD83(NSRS2007) / Oklahoma South', 'EPSG', 3641, 'PROJCS[\"NAD83(NSRS2007) / Oklahom"
  "a South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.233333"
  "3333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3641\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3642, 1, 'NAD83(NSRS2007) / Oklahoma South (ftUS)', 'EPSG', 3642, 'PROJCS[\"NAD83(NSRS2007) / "
  "Oklahoma South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paralle"
  "l\",35.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,A"
  "UTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EP"
  "SG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3642\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3643, 1, 'NAD83(NSRS2007) / Oregon LCC (m)', 'EPSG', 3643, 'PROJCS[\"NAD83(NSRS2007) / Oregon "
  "LCC (m)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3643\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3644, 1, 'NAD83(NSRS2007) / Oregon GIC Lambert (ft)', 'EPSG', 3644, 'PROJCS[\"NAD83(NSRS2007) "
  "/ Oregon GIC Lambert (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMET"
  "ER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard"
  " parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1312335.958,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3644\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3645, 1, 'NAD83(NSRS2007) / Oregon North', 'EPSG', 3645, 'PROJCS[\"NAD83(NSRS2007) / Oregon No"
  "rth\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "46,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3645\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3646, 1, 'NAD83(NSRS2007) / Oregon North (ft)', 'EPSG', 3646, 'PROJCS[\"NAD83(NSRS2007) / Oreg"
  "on North (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",46,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",8202099.738,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3646\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3647, 1, 'NAD83(NSRS2007) / Oregon South', 'EPSG', 3647, 'PROJCS[\"NAD83(NSRS2007) / Oregon So"
  "uth\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3647\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3648, 1, 'NAD83(NSRS2007) / Oregon South (ft)', 'EPSG', 3648, 'PROJCS[\"NAD83(NSRS2007) / Oreg"
  "on South (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4921259.843,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3648\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3649, 1, 'NAD83(NSRS2007) / Pennsylvania North', 'EPSG', 3649, 'PROJCS[\"NAD83(NSRS2007) / Pen"
  "nsylvania North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",41.95,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fal"
  "se origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3649\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3650, 1, 'NAD83(NSRS2007) / Pennsylvania North (ftUS)', 'EPSG', 3650, 'PROJCS[\"NAD83(NSRS2007"
  ") / Pennsylvania North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",41.95,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EP"
  "SG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3650\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3651, 1, 'NAD83(NSRS2007) / Pennsylvania South', 'EPSG', 3651, 'PROJCS[\"NAD83(NSRS2007) / Pen"
  "nsylvania South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",40.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3651\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3652, 1, 'NAD83(NSRS2007) / Pennsylvania South (ftUS)', 'EPSG', 3652, 'PROJCS[\"NAD83(NSRS2007"
  ") / Pennsylvania South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",40.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.933"
  "3333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AU"
  "THORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3652\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3653, 1, 'NAD83(NSRS2007) / Rhode Island', 'EPSG', 3653, 'PROJCS[\"NAD83(NSRS2007) / Rhode Isl"
  "and\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999993"
  "75,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"3653\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3654, 1, 'NAD83(NSRS2007) / Rhode Island (ftUS)', 'EPSG', 3654, 'PROJCS[\"NAD83(NSRS2007) / Rh"
  "ode Island (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.99999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3654\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3655, 1, 'NAD83(NSRS2007) / South Carolina', 'EPSG', 3655, 'PROJCS[\"NAD83(NSRS2007) / South C"
  "arolina\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.833333"
  "3333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",609600,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3655\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3656, 1, 'NAD83(NSRS2007) / South Carolina (ft)', 'EPSG', 3656, 'PROJCS[\"NAD83(NSRS2007) / So"
  "uth Carolina (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",34.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3656\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3657, 1, 'NAD83(NSRS2007) / South Dakota North', 'EPSG', 3657, 'PROJCS[\"NAD83(NSRS2007) / Sou"
  "th Dakota North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "45.6944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"North"
  "ing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3657\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3658, 1, 'NAD83(NSRS2007) / South Dakota North (ftUS)', 'EPSG', 3658, 'PROJCS[\"NAD83(NSRS2007"
  ") / South Dakota North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standa"
  "rd parallel\",45.6944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.41666"
  "66666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3658\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3659, 1, 'NAD83(NSRS2007) / South Dakota South', 'EPSG', 3659, 'PROJCS[\"NAD83(NSRS2007) / Sou"
  "th Dakota South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",44.4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3659\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3660, 1, 'NAD83(NSRS2007) / South Dakota South (ftUS)', 'EPSG', 3660, 'PROJCS[\"NAD83(NSRS2007"
  ") / South Dakota South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",44.4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333"
  "333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],P"
  "ARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3660\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3661, 1, 'NAD83(NSRS2007) / Tennessee', 'EPSG', 3661, 'PROJCS[\"NAD83(NSRS2007) / Tennessee\","
  "GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"98"
  "02\"]],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of"
  " false origin\",-86,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AU"
  "THORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"882"
  "4\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin"
  "\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"3661\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3662, 1, 'NAD83(NSRS2007) / Tennessee (ftUS)', 'EPSG', 3662, 'PROJCS[\"NAD83(NSRS2007) / Tenne"
  "ssee (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-86,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166"
  "666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3662\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3663, 1, 'NAD83(NSRS2007) / Texas Central', 'EPSG', 3663, 'PROJCS[\"NAD83(NSRS2007) / Texas Ce"
  "ntral\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",31.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",3000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3663\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3664, 1, 'NAD83(NSRS2007) / Texas Central (ftUS)', 'EPSG', 3664, 'PROJCS[\"NAD83(NSRS2007) / T"
  "exas Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PA"
  "RAMETER[\"Longitude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st sta"
  "ndard parallel\",31.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.11"
  "66666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2296583.333,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",9842500,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3664\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3665, 1, 'NAD83(NSRS2007) / Texas Centric Albers Equal Area', 'EPSG', 3665, 'PROJCS[\"NAD83(NS"
  "RS2007) / Texas Centric Albers Equal Area\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2007)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Albers Equa"
  "l Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "27.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",600"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"3665\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3666, 1, 'NAD83(NSRS2007) / Texas Centric Lambert Conformal', 'EPSG', 3666, 'PROJCS[\"NAD83(NS"
  "RS2007) / Texas Centric Lambert Conformal\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2007)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Con"
  "ic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"882"
  "1\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard"
  " parallel\",27.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",5000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3666\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3667, 1, 'NAD83(NSRS2007) / Texas North', 'EPSG', 3667, 'PROJCS[\"NAD83(NSRS2007) / Texas Nort"
  "h\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-101.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PAR"
  "AMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000"
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"3667\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3668, 1, 'NAD83(NSRS2007) / Texas North (ftUS)', 'EPSG', 3668, 'PROJCS[\"NAD83(NSRS2007) / Tex"
  "as North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitud"
  "e of false origin\",-101.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",3"
  "6.1944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3668\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3669, 1, 'NAD83(NSRS2007) / Texas North Central', 'EPSG', 3669, 'PROJCS[\"NAD83(NSRS2007) / Te"
  "xas North Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",33.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.1444"
  "444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3669\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3670, 1, 'NAD83(NSRS2007) / Texas North Central (ftUS)', 'EPSG', 3670, 'PROJCS[\"NAD83(NSRS200"
  "7) / Texas North Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conform"
  "al (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\""
  "8821\"]],PARAMETER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",33.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",32.1444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\""
  ",\"8826\"]],PARAMETER[\"Northing at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\","
  "0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3670\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3671, 1, 'NAD83(NSRS2007) / Texas South', 'EPSG', 3671, 'PROJCS[\"NAD83(NSRS2007) / Texas Sout"
  "h\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\""
  "9802\"]],PARAMETER[\"Latitude of false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27."
  "8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",5000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3671\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3672, 1, 'NAD83(NSRS2007) / Texas South (ftUS)', 'EPSG', 3672, 'PROJCS[\"NAD83(NSRS2007) / Tex"
  "as South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",27.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.166666"
  "6666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",16404166.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012"
  "1924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3672\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3673, 1, 'NAD83(NSRS2007) / Texas South Central', 'EPSG', 3673, 'PROJCS[\"NAD83(NSRS2007) / Te"
  "xas South Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",30.2833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",4000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3673\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3674, 1, 'NAD83(NSRS2007) / Texas South Central (ftUS)', 'EPSG', 3674, 'PROJCS[\"NAD83(NSRS200"
  "7) / Texas South Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conform"
  "al (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\""
  "8821\"]],PARAMETER[\"Longitude of false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standa"
  "rd parallel\",30.2833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.38333"
  "33333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",13123333.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3674\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3675, 1, 'NAD83(NSRS2007) / Utah Central', 'EPSG', 3675, 'PROJCS[\"NAD83(NSRS2007) / Utah Cent"
  "ral\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitu"
  "de of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "40.65,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3675\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3676, 1, 'NAD83(NSRS2007) / Utah Central (ft)', 'EPSG', 3676, 'PROJCS[\"NAD83(NSRS2007) / Utah"
  " Central (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",40.65,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",6561679.79,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3676\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3677, 1, 'NAD83(NSRS2007) / Utah Central (ftUS)', 'EPSG', 3677, 'PROJCS[\"NAD83(NSRS2007) / Ut"
  "ah Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",40.65,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,A"
  "UTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",6561666.6667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012192"
  "4,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3677\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3678, 1, 'NAD83(NSRS2007) / Utah North', 'EPSG', 3678, 'PROJCS[\"NAD83(NSRS2007) / Utah North\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7"
  "944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing "
  "at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3678\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3679, 1, 'NAD83(NSRS2007) / Utah North (ft)', 'EPSG', 3679, 'PROJCS[\"NAD83(NSRS2007) / Utah N"
  "orth (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paral"
  "lel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778"
  ",AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",3280839.895,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\""
  "9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3679\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3680, 1, 'NAD83(NSRS2007) / Utah North (ftUS)', 'EPSG', 3680, 'PROJCS[\"NAD83(NSRS2007) / Utah"
  " North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.72777777"
  "77778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3680\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3681, 1, 'NAD83(NSRS2007) / Utah South', 'EPSG', 3681, 'PROJCS[\"NAD83(NSRS2007) / Utah South\""
  ",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.3"
  "5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",3000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"3681\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3682, 1, 'NAD83(NSRS2007) / Utah South (ft)', 'EPSG', 3682, 'PROJCS[\"NAD83(NSRS2007) / Utah S"
  "outh (ft)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paral"
  "lel\",38.35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY["
  "\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640419.948,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"North"
  "ing at false origin\",9842519.685,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3682\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3683, 1, 'NAD83(NSRS2007) / Utah South (ftUS)', 'EPSG', 3683, 'PROJCS[\"NAD83(NSRS2007) / Utah"
  " South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",38.35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",9842500,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3683\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3684, 1, 'NAD83(NSRS2007) / Vermont', 'EPSG', 3684, 'PROJCS[\"NAD83(NSRS2007) / Vermont\",GEOG"
  "CS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-72.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"3684\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3685, 1, 'NAD83(NSRS2007) / Virginia North', 'EPSG', 3685, 'PROJCS[\"NAD83(NSRS2007) / Virgini"
  "a North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",39.2111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3685\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3686, 1, 'NAD83(NSRS2007) / Virginia North (ftUS)', 'EPSG', 3686, 'PROJCS[\"NAD83(NSRS2007) / "
  "Virginia North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",39.2111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0"
  "333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304"
  "80060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3686\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3687, 1, 'NAD83(NSRS2007) / Virginia South', 'EPSG', 3687, 'PROJCS[\"NAD83(NSRS2007) / Virgini"
  "a South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"E"
  "PSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Lo"
  "ngitude of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3687\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3688, 1, 'NAD83(NSRS2007) / Virginia South (ftUS)', 'EPSG', 3688, 'PROJCS[\"NAD83(NSRS2007) / "
  "Virginia South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7"
  "666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304"
  "80060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3688\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3689, 1, 'NAD83(NSRS2007) / Washington North', 'EPSG', 3689, 'PROJCS[\"NAD83(NSRS2007) / Washi"
  "ngton North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-120.833333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.744"
  "4444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3689\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3690, 1, 'NAD83(NSRS2007) / Washington North (ftUS)', 'EPSG', 3690, 'PROJCS[\"NAD83(NSRS2007) "
  "/ Washington North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",-120.833333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard pa"
  "rallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111"
  "111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3690\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3691, 1, 'NAD83(NSRS2007) / Washington South', 'EPSG', 3691, 'PROJCS[\"NAD83(NSRS2007) / Washi"
  "ngton South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",47.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.833333333333"
  "3,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3691\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3692, 1, 'NAD83(NSRS2007) / Washington South (ftUS)', 'EPSG', 3692, 'PROJCS[\"NAD83(NSRS2007) "
  "/ Washington South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",47.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "45.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3693, 1, 'NAD83(NSRS2007) / West Virginia North', 'EPSG', 3693, 'PROJCS[\"NAD83(NSRS2007) / We"
  "st Virginia North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",40.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3693\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3694, 1, 'NAD83(NSRS2007) / West Virginia South', 'EPSG', 3694, 'PROJCS[\"NAD83(NSRS2007) / We"
  "st Virginia South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longit"
  "ude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333"
  "333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3694\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3695, 1, 'NAD83(NSRS2007) / Wisconsin Central', 'EPSG', 3695, 'PROJCS[\"NAD83(NSRS2007) / Wisc"
  "onsin Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45."
  "5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3695\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3696, 1, 'NAD83(NSRS2007) / Wisconsin Central (ftUS)', 'EPSG', 3696, 'PROJCS[\"NAD83(NSRS2007)"
  " / Wisconsin Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"67"
  "59\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard p"
  "arallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.261111111"
  "1111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3696\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3697, 1, 'NAD83(NSRS2007) / Wisconsin North', 'EPSG', 3697, 'PROJCS[\"NAD83(NSRS2007) / Wiscon"
  "sin North\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.76666"
  "66666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"E"
  "PSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3697\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3698, 1, 'NAD83(NSRS2007) / Wisconsin North (ftUS)', 'EPSG', 3698, 'PROJCS[\"NAD83(NSRS2007) /"
  " Wisconsin North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",46.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.566666666666"
  "7,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3698\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3699, 1, 'NAD83(NSRS2007) / Wisconsin South', 'EPSG', 3699, 'PROJCS[\"NAD83(NSRS2007) / Wiscon"
  "sin South\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fa"
  "lse origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"3699\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3700, 1, 'NAD83(NSRS2007) / Wisconsin South (ftUS)', 'EPSG', 3700, 'PROJCS[\"NAD83(NSRS2007) /"
  " Wisconsin South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777"
  "777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3701, 1, 'NAD83(NSRS2007) / Wisconsin Transverse Mercator', 'EPSG', 3701, 'PROJCS[\"NAD83(NSRS"
  "2007) / Wisconsin Transverse Mercator\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "007)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",520000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",-4480000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3701\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3702, 1, 'NAD83(NSRS2007) / Wyoming East', 'EPSG', 3702, 'PROJCS[\"NAD83(NSRS2007) / Wyoming E"
  "ast\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-105.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999937"
  "5,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"3702\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3703, 1, 'NAD83(NSRS2007) / Wyoming East Central', 'EPSG', 3703, 'PROJCS[\"NAD83(NSRS2007) / W"
  "yoming East Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-107.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3703\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3704, 1, 'NAD83(NSRS2007) / Wyoming West Central', 'EPSG', 3704, 'PROJCS[\"NAD83(NSRS2007) / W"
  "yoming West Central\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-108.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "99375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3704\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3705, 1, 'NAD83(NSRS2007) / Wyoming West', 'EPSG', 3705, 'PROJCS[\"NAD83(NSRS2007) / Wyoming W"
  "est\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-110.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999937"
  "5,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3705\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3706, 1, 'NAD83(NSRS2007) / UTM zone 59N', 'EPSG', 3706, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "59N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3706\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3707, 1, 'NAD83(NSRS2007) / UTM zone 60N', 'EPSG', 3707, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "60N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3707\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3708, 1, 'NAD83(NSRS2007) / UTM zone 1N', 'EPSG', 3708, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 1"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "77,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3708\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3709, 1, 'NAD83(NSRS2007) / UTM zone 2N', 'EPSG', 3709, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 2"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "71,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3709\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3710, 1, 'NAD83(NSRS2007) / UTM zone 3N', 'EPSG', 3710, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 3"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "65,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3710\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3711, 1, 'NAD83(NSRS2007) / UTM zone 4N', 'EPSG', 3711, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 4"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "59,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3711\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3712, 1, 'NAD83(NSRS2007) / UTM zone 5N', 'EPSG', 3712, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 5"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "53,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3712\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3713, 1, 'NAD83(NSRS2007) / UTM zone 6N', 'EPSG', 3713, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 6"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "47,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3713\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3714, 1, 'NAD83(NSRS2007) / UTM zone 7N', 'EPSG', 3714, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 7"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "41,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3714\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3715, 1, 'NAD83(NSRS2007) / UTM zone 8N', 'EPSG', 3715, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 8"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "35,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3715\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3716, 1, 'NAD83(NSRS2007) / UTM zone 9N', 'EPSG', 3716, 'PROJCS[\"NAD83(NSRS2007) / UTM zone 9"
  "N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "29,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3716\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3717, 1, 'NAD83(NSRS2007) / UTM zone 10N', 'EPSG', 3717, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "10N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3717\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3718, 1, 'NAD83(NSRS2007) / UTM zone 11N', 'EPSG', 3718, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "11N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3718\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3719, 1, 'NAD83(NSRS2007) / UTM zone 12N', 'EPSG', 3719, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "12N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3719\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3720, 1, 'NAD83(NSRS2007) / UTM zone 13N', 'EPSG', 3720, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "13N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3720\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3721, 1, 'NAD83(NSRS2007) / UTM zone 14N', 'EPSG', 3721, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "14N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3721\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3722, 1, 'NAD83(NSRS2007) / UTM zone 15N', 'EPSG', 3722, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "15N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3722\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3723, 1, 'NAD83(NSRS2007) / UTM zone 16N', 'EPSG', 3723, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "16N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3723\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3724, 1, 'NAD83(NSRS2007) / UTM zone 17N', 'EPSG', 3724, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "17N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3724\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3725, 1, 'NAD83(NSRS2007) / UTM zone 18N', 'EPSG', 3725, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "18N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3725\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3726, 1, 'NAD83(NSRS2007) / UTM zone 19N', 'EPSG', 3726, 'PROJCS[\"NAD83(NSRS2007) / UTM zone "
  "19N\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3726\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3727, 1, 'Reunion 1947 / TM Reunion', 'EPSG', 3727, 'PROJCS[\"Reunion 1947 / TM Reunion\",GEOG"
  "CS[\"Reunion 1947\",DATUM[\"Reunion 1947\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],"
  "TOWGS84[94,-948,-1262,0,0,0,0],AUTHORITY[\"EPSG\",\"6626\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4626\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",-21.1166666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",55.5333333333333,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",160000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",50000,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3727\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3728, 1, 'NAD83(NSRS2007) / Ohio North (ftUS)', 'EPSG', 3728, 'PROJCS[\"NAD83(NSRS2007) / Ohio"
  " North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETE"
  "R[\"Longitude of false origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",41.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHOR"
  "ITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"North"
  "ing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\","
  "\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3728\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3729, 1, 'NAD83(NSRS2007) / Ohio South (ftUS)', 'EPSG', 3729, 'PROJCS[\"NAD83(NSRS2007) / Ohio"
  " South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORI"
  "TY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40."
  "0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORIT"
  "Y[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northin"
  "g at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3729\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3730, 1, 'NAD83(NSRS2007) / Wyoming East (ftUS)', 'EPSG', 3730, 'PROJCS[\"NAD83(NSRS2007) / Wy"
  "oming East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-105.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"E"
  "PSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3730\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3731, 1, 'NAD83(NSRS2007) / Wyoming East Central (ftUS)', 'EPSG', 3731, 'PROJCS[\"NAD83(NSRS20"
  "07) / Wyoming East Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]"
  "],PARAMETER[\"Longitude of natural origin\",-107.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1312333.3333,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",328083.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3731\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3732, 1, 'NAD83(NSRS2007) / Wyoming West Central (ftUS)', 'EPSG', 3732, 'PROJCS[\"NAD83(NSRS20"
  "07) / Wyoming West Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]"
  "],PARAMETER[\"Longitude of natural origin\",-108.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1968500,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"E"
  "PSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3732\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3733, 1, 'NAD83(NSRS2007) / Wyoming West (ftUS)', 'EPSG', 3733, 'PROJCS[\"NAD83(NSRS2007) / Wy"
  "oming West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-110.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",328083.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AU"
  "THORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3733\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3734, 1, 'NAD83 / Ohio North (ftUS)', 'EPSG', 3734, 'PROJCS[\"NAD83 / Ohio North (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"3734\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3735, 1, 'NAD83 / Ohio South (ftUS)', 'EPSG', 3735, 'PROJCS[\"NAD83 / Ohio South (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3735\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3736, 1, 'NAD83 / Wyoming East (ftUS)', 'EPSG', 3736, 'PROJCS[\"NAD83 / Wyoming East (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105.177777777778,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"3736\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3737, 1, 'NAD83 / Wyoming East Central (ftUS)', 'EPSG', 3737, 'PROJCS[\"NAD83 / Wyoming East C"
  "entral (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107"
  ".333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",1312333.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",32808"
  "3.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3737\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3738, 1, 'NAD83 / Wyoming West Central (ftUS)', 'EPSG', 3738, 'PROJCS[\"NAD83 / Wyoming West C"
  "entral (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-108"
  ".75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",1968500,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"3738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3739, 1, 'NAD83 / Wyoming West (ftUS)', 'EPSG', 3739, 'PROJCS[\"NAD83 / Wyoming West (ftUS)\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.094444444444,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",328083.3333,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3739\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3740, 1, 'NAD83(HARN) / UTM zone 10N', 'EPSG', 3740, 'PROJCS[\"NAD83(HARN) / UTM zone 10N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3740\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3741, 1, 'NAD83(HARN) / UTM zone 11N', 'EPSG', 3741, 'PROJCS[\"NAD83(HARN) / UTM zone 11N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3741\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3742, 1, 'NAD83(HARN) / UTM zone 12N', 'EPSG', 3742, 'PROJCS[\"NAD83(HARN) / UTM zone 12N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3742\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3743, 1, 'NAD83(HARN) / UTM zone 13N', 'EPSG', 3743, 'PROJCS[\"NAD83(HARN) / UTM zone 13N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3743\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3744, 1, 'NAD83(HARN) / UTM zone 14N', 'EPSG', 3744, 'PROJCS[\"NAD83(HARN) / UTM zone 14N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3744\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3745, 1, 'NAD83(HARN) / UTM zone 15N', 'EPSG', 3745, 'PROJCS[\"NAD83(HARN) / UTM zone 15N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3745\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3746, 1, 'NAD83(HARN) / UTM zone 16N', 'EPSG', 3746, 'PROJCS[\"NAD83(HARN) / UTM zone 16N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3746\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3747, 1, 'NAD83(HARN) / UTM zone 17N', 'EPSG', 3747, 'PROJCS[\"NAD83(HARN) / UTM zone 17N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3747\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3748, 1, 'NAD83(HARN) / UTM zone 18N', 'EPSG', 3748, 'PROJCS[\"NAD83(HARN) / UTM zone 18N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3748\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3749, 1, 'NAD83(HARN) / UTM zone 19N', 'EPSG', 3749, 'PROJCS[\"NAD83(HARN) / UTM zone 19N\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3749\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3750, 1, 'NAD83(HARN) / UTM zone 4N', 'EPSG', 3750, 'PROJCS[\"NAD83(HARN) / UTM zone 4N\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3750\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3751, 1, 'NAD83(HARN) / UTM zone 5N', 'EPSG', 3751, 'PROJCS[\"NAD83(HARN) / UTM zone 5N\",GEOG"
  "CS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3751\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3753, 1, 'NAD83(HARN) / Ohio North (ftUS)', 'EPSG', 3753, 'PROJCS[\"NAD83(HARN) / Ohio North ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3753\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3754, 1, 'NAD83(HARN) / Ohio South (ftUS)', 'EPSG', 3754, 'PROJCS[\"NAD83(HARN) / Ohio South ("
  "ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-8"
  "2.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3754\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3755, 1, 'NAD83(HARN) / Wyoming East (ftUS)', 'EPSG', 3755, 'PROJCS[\"NAD83(HARN) / Wyoming Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-105.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3755\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3756, 1, 'NAD83(HARN) / Wyoming East Central (ftUS)', 'EPSG', 3756, 'PROJCS[\"NAD83(HARN) / Wy"
  "oming East Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-107.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1312333.3333,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",328083.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3756\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3757, 1, 'NAD83(HARN) / Wyoming West Central (ftUS)', 'EPSG', 3757, 'PROJCS[\"NAD83(HARN) / Wy"
  "oming West Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-108.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99993"
  "75,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1968500,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3757\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3758, 1, 'NAD83(HARN) / Wyoming West (ftUS)', 'EPSG', 3758, 'PROJCS[\"NAD83(HARN) / Wyoming We"
  "st (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-110.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",328083.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3758\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3759, 1, 'NAD83 / Hawaii zone 3 (ftUS)', 'EPSG', 3759, 'PROJCS[\"NAD83 / Hawaii zone 3 (ftUS)\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",21.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-158,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"3759\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3760, 1, 'NAD83(HARN) / Hawaii zone 3 (ftUS)', 'EPSG', 3760, 'PROJCS[\"NAD83(HARN) / Hawaii zo"
  "ne 3 (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",21.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-158,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3760\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3761, 1, 'NAD83(CSRS) / UTM zone 22N', 'EPSG', 3761, 'PROJCS[\"NAD83(CSRS) / UTM zone 22N\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3761\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3762, 1, 'WGS 84 / South Georgia Lambert', 'EPSG', 3762, 'PROJCS[\"WGS 84 / South Georgia Lamb"
  "ert\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-5"
  "5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-37,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",-54,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",-54.75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3762\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3763, 1, 'ETRS89 / Portugal TM06', 'EPSG', 3763, 'PROJCS[\"ETRS89 / Portugal TM06\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",39.6682583333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8.1331083333"
  "3334,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3763\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3764, 1, 'NZGD2000 / Chatham Island Circuit 2000', 'EPSG', 3764, 'PROJCS[\"NZGD2000 / Chatham "
  "Island Circuit 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",-44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-17"
  "6.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3764\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3765, 1, 'HTRS96 / Croatia TM', 'EPSG', 3765, 'PROJCS[\"HTRS96 / Croatia TM\",GEOGCS[\"HTRS96\""
  ",DATUM[\"Croatian Terrestrial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4761\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3765\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3766, 1, 'HTRS96 / Croatia LCC', 'EPSG', 3766, 'PROJCS[\"HTRS96 / Croatia LCC\",GEOGCS[\"HTRS9"
  "6\",DATUM[\"Croatian Terrestrial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4761\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",16.5,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",45.9277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",43.0944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,A"
  "UTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3766\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3767, 1, 'HTRS96 / UTM zone 33N', 'EPSG', 3767, 'PROJCS[\"HTRS96 / UTM zone 33N\",GEOGCS[\"HTR"
  "S96\",DATUM[\"Croatian Terrestrial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4761\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3767\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3768, 1, 'HTRS96 / UTM zone 34N', 'EPSG', 3768, 'PROJCS[\"HTRS96 / UTM zone 34N\",GEOGCS[\"HTR"
  "S96\",DATUM[\"Croatian Terrestrial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6761\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4761\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3768\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3769, 1, 'Bermuda 1957 / UTM zone 20N', 'EPSG', 3769, 'PROJCS[\"Bermuda 1957 / UTM zone 20N\","
  "GEOGCS[\"Bermuda 1957\",DATUM[\"Bermuda 1957\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-292.295,248.758,429.447,4.9971,2.99,6.6906,1.0289],AUTHORITY[\"EPSG\",\"6216\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4216\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"376"
  "9\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3770, 1, 'BDA2000 / Bermuda 2000 National Grid', 'EPSG', 3770, 'PROJCS[\"BDA2000 / Bermuda 200"
  "0 National Grid\",GEOGCS[\"BDA2000\",DATUM[\"Bermuda 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPS"
  "G\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6762\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4762\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",32,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-64.75,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",550000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3770\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3771, 1, 'NAD27 / Alberta 3TM ref merid 111 W', 'EPSG', 3771, 'PROJCS[\"NAD27 / Alberta 3TM re"
  "f merid 111 W\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213"
  "8982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049"
  "164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3771\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3772, 1, 'NAD27 / Alberta 3TM ref merid 114 W', 'EPSG', 3772, 'PROJCS[\"NAD27 / Alberta 3TM re"
  "f merid 114 W\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213"
  "8982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049"
  "164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3772\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3773, 1, 'NAD27 / Alberta 3TM ref merid 117 W', 'EPSG', 3773, 'PROJCS[\"NAD27 / Alberta 3TM re"
  "f merid 117 W\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213"
  "8982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049"
  "164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3773\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3775, 1, 'NAD83 / Alberta 3TM ref merid 111 W', 'EPSG', 3775, 'PROJCS[\"NAD83 / Alberta 3TM re"
  "f merid 111 W\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3775\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3776, 1, 'NAD83 / Alberta 3TM ref merid 114 W', 'EPSG', 3776, 'PROJCS[\"NAD83 / Alberta 3TM re"
  "f merid 114 W\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3776\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3777, 1, 'NAD83 / Alberta 3TM ref merid 117 W', 'EPSG', 3777, 'PROJCS[\"NAD83 / Alberta 3TM re"
  "f merid 117 W\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3777\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3779, 1, 'NAD83(CSRS) / Alberta 3TM ref merid 111 W', 'EPSG', 3779, 'PROJCS[\"NAD83(CSRS) / Al"
  "berta 3TM ref merid 111 W\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.009650098"
  "96027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4617\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3779\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3780, 1, 'NAD83(CSRS) / Alberta 3TM ref merid 114 W', 'EPSG', 3780, 'PROJCS[\"NAD83(CSRS) / Al"
  "berta 3TM ref merid 114 W\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.009650098"
  "96027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4617\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3780\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3781, 1, 'NAD83(CSRS) / Alberta 3TM ref merid 117 W', 'EPSG', 3781, 'PROJCS[\"NAD83(CSRS) / Al"
  "berta 3TM ref merid 117 W\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.009650098"
  "96027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4617\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3781\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3783, 1, 'Pitcairn 2006 / Pitcairn TM 2006', 'EPSG', 3783, 'PROJCS[\"Pitcairn 2006 / Pitcairn "
  "TM 2006\",GEOGCS[\"Pitcairn 2006\",DATUM[\"Pitcairn 2006\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG"
  "\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6763\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4763\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",-25.0685526111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-130.11296711111"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",14200,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",15500,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3783\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3784, 1, 'Pitcairn 1967 / UTM zone 9S', 'EPSG', 3784, 'PROJCS[\"Pitcairn 1967 / UTM zone 9S\","
  "GEOGCS[\"Pitcairn 1967\",DATUM[\"Pitcairn 1967\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"702"
  "2\"]],TOWGS84[185,165,42,0,0,0,0],AUTHORITY[\"EPSG\",\"6729\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4729\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3784\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3788, 1, 'NZGD2000 / Auckland Islands TM 2000', 'EPSG', 3788, 'PROJCS[\"NZGD2000 / Auckland Is"
  "lands TM 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",166,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3788\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3789, 1, 'NZGD2000 / Campbell Island TM 2000', 'EPSG', 3789, 'PROJCS[\"NZGD2000 / Campbell Isl"
  "and TM 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",169,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3789\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3790, 1, 'NZGD2000 / Antipodes Islands TM 2000', 'EPSG', 3790, 'PROJCS[\"NZGD2000 / Antipodes "
  "Islands TM 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",179,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3790\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3791, 1, 'NZGD2000 / Raoul Island TM 2000', 'EPSG', 3791, 'PROJCS[\"NZGD2000 / Raoul Island TM"
  " 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT"
  "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-178,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3793, 1, 'NZGD2000 / Chatham Islands TM 2000', 'EPSG', 3793, 'PROJCS[\"NZGD2000 / Chatham Isla"
  "nds TM 2000\",GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-176.5,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3793\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3794, 1, 'Slovenia 1996 / Slovene National Grid', 'EPSG', 3794, 'PROJCS[\"Slovenia 1996 / Slov"
  "ene National Grid\",GEOGCS[\"Slovenia 1996\",DATUM[\"Slovenia Geodetic Datum 1996\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6765\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4765\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3794\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3795, 1, 'NAD27 / Cuba Norte', 'EPSG', 3795, 'PROJCS[\"NAD27 / Cuba Norte\",GEOGCS[\"NAD27\",D"
  "ATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]]"
  ",TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",22.35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",23,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",21.7,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETE"
  "R[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",280296.016,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST],AUTHO"
  "RITY[\"EPSG\",\"3795\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3796, 1, 'NAD27 / Cuba Sur', 'EPSG', 3796, 'PROJCS[\"NAD27 / Cuba Sur\",GEOGCS[\"NAD27\",DATUM"
  "[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOW"
  "GS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUT"
  "HORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",20.7277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARA"
  "METER[\"Longitude of false origin\",-76.8333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",21.3111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",20.1444"
  "444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",229126.939,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST],AUTHORITY[\"EPSG\",\"3796\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3797, 1, 'NAD27 / MTQ Lambert', 'EPSG', 3797, 'PROJCS[\"NAD27 / MTQ Lambert\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-70,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",50,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"3797\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3798, 1, 'NAD83 / MTQ Lambert', 'EPSG', 3798, 'PROJCS[\"NAD83 / MTQ Lambert\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "44,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-70,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",50,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETE"
  "R[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3798\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3799, 1, 'NAD83(CSRS) / MTQ Lambert', 'EPSG', 3799, 'PROJCS[\"NAD83(CSRS) / MTQ Lambert\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Lam"
  "bert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44,AUTHORITY[\"EPSG"
  "\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-70,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st s"
  "tandard parallel\",50,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3799\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3800, 1, 'NAD27 / Alberta 3TM ref merid 120 W', 'EPSG', 3800, 'PROJCS[\"NAD27 / Alberta 3TM re"
  "f merid 120 W\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213"
  "8982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049"
  "164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3800\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3801, 1, 'NAD83 / Alberta 3TM ref merid 120 W', 'EPSG', 3801, 'PROJCS[\"NAD83 / Alberta 3TM re"
  "f merid 120 W\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-120,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3801\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3802, 1, 'NAD83(CSRS) / Alberta 3TM ref merid 120 W', 'EPSG', 3802, 'PROJCS[\"NAD83(CSRS) / Al"
  "berta 3TM ref merid 120 W\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.009650098"
  "96027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4617\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-120,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3802\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3812, 1, 'ETRS89 / Belgian Lambert 2008', 'EPSG', 3812, 'PROJCS[\"ETRS89 / Belgian Lambert 200"
  "8\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",50.797815,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\","
  "4.35921583333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",51.1666666666667,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",649328,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\","
  "665262,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"3812\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3814, 1, 'NAD83 / Mississippi TM', 'EPSG', 3814, 'PROJCS[\"NAD83 / Mississippi TM\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",32"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.75,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9998335,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3814\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3815, 1, 'NAD83(HARN) / Mississippi TM', 'EPSG', 3815, 'PROJCS[\"NAD83(HARN) / Mississippi TM\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-89.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9998335,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1300000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"3815\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3816, 1, 'NAD83(NSRS2007) / Mississippi TM', 'EPSG', 3816, 'PROJCS[\"NAD83(NSRS2007) / Mississ"
  "ippi TM\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude "
  "of natural origin\",-89.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9998335,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",1300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"3816\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3819, 1, 'HD1909', 'EPSG', 3819, 'GEOGCS[\"HD1909\",DATUM[\"Hungarian Datum 1909\",SPHEROID[\""
  "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[595.48,121.69,515.35,4.115,-2.9383,0.853,"
  "-3.408],AUTHORITY[\"EPSG\",\"1024\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3819\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3821, 1, 'TWD67', 'EPSG', 3821, 'GEOGCS[\"TWD67\",DATUM[\"Taiwan Datum 1967\",SPHEROID[\"GRS 1"
  "967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORITY[\"EPSG\",\"1025\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3821\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3824, 1, 'TWD97', 'EPSG', 3824, 'GEOGCS[\"TWD97\",DATUM[\"Taiwan Datum 1997\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1026\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3824\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3825, 1, 'TWD97 / TM2 zone 119', 'EPSG', 3825, 'PROJCS[\"TWD97 / TM2 zone 119\",GEOGCS[\"TWD97"
  "\",DATUM[\"Taiwan Datum 1997\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,"
  "0,0,0,0,0],AUTHORITY[\"EPSG\",\"1026\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3824\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3825\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3826, 1, 'TWD97 / TM2 zone 121', 'EPSG', 3826, 'PROJCS[\"TWD97 / TM2 zone 121\",GEOGCS[\"TWD97"
  "\",DATUM[\"Taiwan Datum 1997\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,"
  "0,0,0,0,0],AUTHORITY[\"EPSG\",\"1026\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3824\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3826\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3827, 1, 'TWD67 / TM2 zone 119', 'EPSG', 3827, 'PROJCS[\"TWD67 / TM2 zone 119\",GEOGCS[\"TWD67"
  "\",DATUM[\"Taiwan Datum 1967\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORITY["
  "\"EPSG\",\"1025\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3821\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\","
  "0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3827\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3828, 1, 'TWD67 / TM2 zone 121', 'EPSG', 3828, 'PROJCS[\"TWD67 / TM2 zone 121\",GEOGCS[\"TWD67"
  "\",DATUM[\"Taiwan Datum 1967\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],AUTHORITY["
  "\"EPSG\",\"1025\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3821\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\","
  "0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3828\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3829, 1, 'Hu Tzu Shan 1950 / UTM zone 51N', 'EPSG', 3829, 'PROJCS[\"Hu Tzu Shan 1950 / UTM zon"
  "e 51N\",GEOGCS[\"Hu Tzu Shan 1950\",DATUM[\"Hu Tzu Shan 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY["
  "\"EPSG\",\"7022\"]],TOWGS84[-637,-549,-203,0,0,0,0],AUTHORITY[\"EPSG\",\"6236\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4236\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3829\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3832, 1, 'WGS 84 / PDC Mercator', 'EPSG', 3832, 'PROJCS[\"WGS 84 / PDC Mercator\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Merc"
  "ator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N"
  "\",NORTH],AUTHORITY[\"EPSG\",\"3832\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3833, 1, 'Pulkovo 1942(58) / Gauss-Kruger zone 2', 'EPSG', 3833, 'PROJCS[\"Pulkovo 1942(58) / "
  "Gauss-Kruger zone 2\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"3833\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3834, 1, 'Pulkovo 1942(83) / Gauss-Kruger zone 2', 'EPSG', 3834, 'PROJCS[\"Pulkovo 1942(83) / "
  "Gauss-Kruger zone 2\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3834\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3835, 1, 'Pulkovo 1942(83) / Gauss-Kruger zone 3', 'EPSG', 3835, 'PROJCS[\"Pulkovo 1942(83) / "
  "Gauss-Kruger zone 3\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3835\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3836, 1, 'Pulkovo 1942(83) / Gauss-Kruger zone 4', 'EPSG', 3836, 'PROJCS[\"Pulkovo 1942(83) / "
  "Gauss-Kruger zone 4\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3836\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3837, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 3', 'EPSG', 3837, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 3\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST"
  "],AUTHORITY[\"EPSG\",\"3837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3838, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 4', 'EPSG', 3838, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 4\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3838\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3839, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 9', 'EPSG', 3839, 'PROJCS[\"Pulkovo 19"
  "42(58) / 3-degree Gauss-Kruger zone 9\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"E"
  "PSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAS"
  "T],AUTHORITY[\"EPSG\",\"3839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3840, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 10', 'EPSG', 3840, 'PROJCS[\"Pulkovo 1"
  "942(58) / 3-degree Gauss-Kruger zone 10\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsk"
  "y 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\""
  "EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",E"
  "AST],AUTHORITY[\"EPSG\",\"3840\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3841, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 6', 'EPSG', 3841, 'PROJCS[\"Pulkovo 19"
  "42(83) / 3-degree Gauss-Kruger zone 6\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"617"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",18,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"3841\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3844, 1, 'Pulkovo 1942(58) / Stereo70', 'EPSG', 3844, 'PROJCS[\"Pulkovo 1942(58) / Stereo70\","
  "GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\","
  "\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"9809\"]],P"
  "ARAMETER[\"Latitude of natural origin\",46,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99975,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"384"
  "4\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3845, 1, 'SWEREF99 / RT90 7.5 gon V emulation', 'EPSG', 3845, 'PROJCS[\"SWEREF99 / RT90 7.5 go"
  "n V emulation\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11.30625,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1.000006,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "1500025.141,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-667.282,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3845\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3846, 1, 'SWEREF99 / RT90 5 gon V emulation', 'EPSG', 3846, 'PROJCS[\"SWEREF99 / RT90 5 gon V "
  "emulation\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13.5562666666667,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000058,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",1500044.695,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-667.13,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3846\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3847, 1, 'SWEREF99 / RT90 2.5 gon V emulation', 'EPSG', 3847, 'PROJCS[\"SWEREF99 / RT90 2.5 go"
  "n V emulation\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15.8062845294444,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00000561024,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",1500064.274,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-667.711,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3847\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3848, 1, 'SWEREF99 / RT90 0 gon emulation', 'EPSG', 3848, 'PROJCS[\"SWEREF99 / RT90 0 gon emul"
  "ation\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18.0563,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1.0000054,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500083.5"
  "21,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-668.844,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3848\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3849, 1, 'SWEREF99 / RT90 2.5 gon O emulation', 'EPSG', 3849, 'PROJCS[\"SWEREF99 / RT90 2.5 go"
  "n O emulation\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20.3063166666667,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000052,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",1500102.765,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-670.706,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3849\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3850, 1, 'SWEREF99 / RT90 5 gon O emulation', 'EPSG', 3850, 'PROJCS[\"SWEREF99 / RT90 5 gon O "
  "emulation\",GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4619\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22.5563333333333,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000049,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",1500121.846,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-672.557,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3850\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3851, 1, 'NZGD2000 / NZCS2000', 'EPSG', 3851, 'PROJCS[\"NZGD2000 / NZCS2000\",GEOGCS[\"NZGD200"
  "0\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4167\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false o"
  "rigin\",-41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",173,AUTHORITY[\"EPSG\",\"8822\"]],P"
  "ARAMETER[\"Latitude of 1st standard parallel\",-37.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude o"
  "f 2nd standard parallel\",-44.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",30000"
  "00,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",7000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3851\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3852, 1, 'RSRGD2000 / DGLC2000', 'EPSG', 3852, 'PROJCS[\"RSRGD2000 / DGLC2000\",GEOGCS[\"RSRGD"
  "2000\",DATUM[\"Ross Sea Region Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4764\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",-90,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",157,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3852\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3854, 1, 'County ST74', 'EPSG', 3854, 'PROJCS[\"County ST74\",GEOGCS[\"SWEREF99\",DATUM[\"SWER"
  "EF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"6619\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",18.05787,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.99999506,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100182.7406,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",-6500620.1207,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"3854\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3857, 1, 'WGS 84 / Pseudo-Mercator', 'EPSG', 3857, 'PROJCS[\"WGS 84 / Pseudo-Mercator\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Popular Visualisation Pseudo Mercator\",AUTHORITY[\"EPSG\",\"1024\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False"
  " easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3857\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3873, 1, 'ETRS89 / GK19FIN', 'EPSG', 3873, 'PROJCS[\"ETRS89 / GK19FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",19500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3873\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3874, 1, 'ETRS89 / GK20FIN', 'EPSG', 3874, 'PROJCS[\"ETRS89 / GK20FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3874\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3875, 1, 'ETRS89 / GK21FIN', 'EPSG', 3875, 'PROJCS[\"ETRS89 / GK21FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",21500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3875\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3876, 1, 'ETRS89 / GK22FIN', 'EPSG', 3876, 'PROJCS[\"ETRS89 / GK22FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",22500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3876\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3877, 1, 'ETRS89 / GK23FIN', 'EPSG', 3877, 'PROJCS[\"ETRS89 / GK23FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",23500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3877\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3878, 1, 'ETRS89 / GK24FIN', 'EPSG', 3878, 'PROJCS[\"ETRS89 / GK24FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",24500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3878\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3879, 1, 'ETRS89 / GK25FIN', 'EPSG', 3879, 'PROJCS[\"ETRS89 / GK25FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",25500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3879\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3880, 1, 'ETRS89 / GK26FIN', 'EPSG', 3880, 'PROJCS[\"ETRS89 / GK26FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",26500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3880\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3881, 1, 'ETRS89 / GK27FIN', 'EPSG', 3881, 'PROJCS[\"ETRS89 / GK27FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",27500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3881\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3882, 1, 'ETRS89 / GK28FIN', 'EPSG', 3882, 'PROJCS[\"ETRS89 / GK28FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",28500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3882\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3883, 1, 'ETRS89 / GK29FIN', 'EPSG', 3883, 'PROJCS[\"ETRS89 / GK29FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",29,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",29500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3883\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3884, 1, 'ETRS89 / GK30FIN', 'EPSG', 3884, 'PROJCS[\"ETRS89 / GK30FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3884\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3885, 1, 'ETRS89 / GK31FIN', 'EPSG', 3885, 'PROJCS[\"ETRS89 / GK31FIN\",GEOGCS[\"ETRS89\",DATU"
  "M[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",31500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"3885\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3889, 1, 'IGRS', 'EPSG', 3889, 'GEOGCS[\"IGRS\",DATUM[\"Iraqi Geospatial Reference System\",SP"
  "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1"
  "029\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3889\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3890, 1, 'IGRS / UTM zone 37N', 'EPSG', 3890, 'PROJCS[\"IGRS / UTM zone 37N\",GEOGCS[\"IGRS\","
  "DATUM[\"Iraqi Geospatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1029\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "3889\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3890\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3891, 1, 'IGRS / UTM zone 38N', 'EPSG', 3891, 'PROJCS[\"IGRS / UTM zone 38N\",GEOGCS[\"IGRS\","
  "DATUM[\"Iraqi Geospatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1029\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "3889\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3891\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3892, 1, 'IGRS / UTM zone 39N', 'EPSG', 3892, 'PROJCS[\"IGRS / UTM zone 39N\",GEOGCS[\"IGRS\","
  "DATUM[\"Iraqi Geospatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1029\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "3889\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3892\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3893, 1, 'ED50 / Iraq National Grid', 'EPSG', 3893, 'PROJCS[\"ED50 / Iraq National Grid\",GEOG"
  "CS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],T"
  "OWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",29.0262683333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",46.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9994,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"3893\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3906, 1, 'MGI 1901', 'EPSG', 3906, 'GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1"
  "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"3906\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3907, 1, 'MGI 1901 / Balkans zone 5 (3907 deprecated)', 'EPSG', 3907, 'PROJCS[\"MGI 1901 / Bal"
  "kans zone 5\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3907\"]]', 'Deprecated since EP"
  "SG Dataset 9.5. Superseded by SRID 8677.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3908, 1, 'MGI 1901 / Balkans zone 6 (3908 deprecated)', 'EPSG', 3908, 'PROJCS[\"MGI 1901 / Bal"
  "kans zone 6\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3908\"]]', 'Deprecated since EP"
  "SG Dataset 9.5. Superseded by SRID 8678.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3909, 1, 'MGI 1901 / Balkans zone 7 (3909 deprecated)', 'EPSG', 3909, 'PROJCS[\"MGI 1901 / Bal"
  "kans zone 7\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3909\"]]', 'Deprecated since EP"
  "SG Dataset 9.5. Superseded by SRID 6316.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3910, 1, 'MGI 1901 / Balkans zone 8 (3910 deprecated)', 'EPSG', 3910, 'PROJCS[\"MGI 1901 / Bal"
  "kans zone 8\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3910\"]]', 'Deprecated since EP"
  "SG Dataset 9.5. Superseded by SRID 8679.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3911, 1, 'MGI 1901 / Slovenia Grid (3911 deprecated)', 'EPSG', 3911, 'PROJCS[\"MGI 1901 / Slov"
  "enia Grid\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\""
  ",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"3911\"]]', 'Deprecated since EPSG "
  "Dataset 9.5. Superseded by SRID 8686.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3912, 1, 'MGI 1901 / Slovene National Grid', 'EPSG', 3912, 'PROJCS[\"MGI 1901 / Slovene Nation"
  "al Grid\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\","
  "\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"3912\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3920, 1, 'Puerto Rico / UTM zone 20N', 'EPSG', 3920, 'PROJCS[\"Puerto Rico / UTM zone 20N\",GE"
  "OGCS[\"Puerto Rico\",DATUM[\"Puerto Rico\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\",\"6139\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4139\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3920\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3942, 1, 'RGF93 / CC42', 'EPSG', 3942, 'PROJCS[\"RGF93 / CC42\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",41.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",1200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3942\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3943, 1, 'RGF93 / CC43', 'EPSG', 3943, 'PROJCS[\"RGF93 / CC43\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",42.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",43.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",2200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3943\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3944, 1, 'RGF93 / CC44', 'EPSG', 3944, 'PROJCS[\"RGF93 / CC44\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",43.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",3200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3944\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3945, 1, 'RGF93 / CC45', 'EPSG', 3945, 'PROJCS[\"RGF93 / CC45\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",44.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",4200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3945\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3946, 1, 'RGF93 / CC46', 'EPSG', 3946, 'PROJCS[\"RGF93 / CC46\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",46,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",45.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",5200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3946\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3947, 1, 'RGF93 / CC47', 'EPSG', 3947, 'PROJCS[\"RGF93 / CC47\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",46.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",6200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3947\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3948, 1, 'RGF93 / CC48', 'EPSG', 3948, 'PROJCS[\"RGF93 / CC48\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",48,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",47.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",48.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",7200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3948\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3949, 1, 'RGF93 / CC49', 'EPSG', 3949, 'PROJCS[\"RGF93 / CC49\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",49,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",48.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",49.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",8200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3949\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3950, 1, 'RGF93 / CC50', 'EPSG', 3950, 'PROJCS[\"RGF93 / CC50\",GEOGCS[\"RGF93\",DATUM[\"Resea"
  "u Geodesique Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]],PROJEC"
  "TION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",50,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",3,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",49.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",50.75,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",9200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3950\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3968, 1, 'NAD83 / Virginia Lambert', 'EPSG', 3968, 'PROJCS[\"NAD83 / Virginia Lambert\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79.5,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",37,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standar"
  "d parallel\",39.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]]"
  ",PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3968\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3969, 1, 'NAD83(HARN) / Virginia Lambert', 'EPSG', 3969, 'PROJCS[\"NAD83(HARN) / Virginia Lamb"
  "ert\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79"
  ".5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",39.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3969\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3970, 1, 'NAD83(NSRS2007) / Virginia Lambert', 'EPSG', 3970, 'PROJCS[\"NAD83(NSRS2007) / Virgi"
  "nia Lambert\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-79.5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3970\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3976, 1, 'WGS 84 / NSIDC Sea Ice Polar Stereographic South', 'EPSG', 3976, 'PROJCS[\"WGS 84 / "
  "NSIDC Sea Ice Polar Stereographic South\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\","
  "6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PAR"
  "AMETER[\"Latitude of standard parallel\",-70,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8833\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"3976\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3978, 1, 'NAD83 / Canada Atlas Lambert', 'EPSG', 3978, 'PROJCS[\"NAD83 / Canada Atlas Lambert\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",49,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-95,AUTHORITY[\"EPSG\",\"88"
  "22\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd s"
  "tandard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3978\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3979, 1, 'NAD83(CSRS) / Canada Atlas Lambert', 'EPSG', 3979, 'PROJCS[\"NAD83(CSRS) / Canada At"
  "las Lambert\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.011"
  "6599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]"
  "],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",4"
  "9,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-95,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3979\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3986, 1, 'Katanga 1955 / Katanga Gauss zone A', 'EPSG', 3986, 'PROJCS[\"Katanga 1955 / Katanga"
  " Gauss zone A\",GEOGCS[\"Katanga 1955\",DATUM[\"Katanga 1955\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,A"
  "UTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-102.283,-10.277,-257.396,3.976,0.002,6.203,12.315],AUTHORITY[\"EPSG\",\"6695\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4695\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"3986\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3987, 1, 'Katanga 1955 / Katanga Gauss zone B', 'EPSG', 3987, 'PROJCS[\"Katanga 1955 / Katanga"
  " Gauss zone B\",GEOGCS[\"Katanga 1955\",DATUM[\"Katanga 1955\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,A"
  "UTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-102.283,-10.277,-257.396,3.976,0.002,6.203,12.315],AUTHORITY[\"EPSG\",\"6695\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4695\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",28,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"3987\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3988, 1, 'Katanga 1955 / Katanga Gauss zone C', 'EPSG', 3988, 'PROJCS[\"Katanga 1955 / Katanga"
  " Gauss zone C\",GEOGCS[\"Katanga 1955\",DATUM[\"Katanga 1955\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,A"
  "UTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-102.283,-10.277,-257.396,3.976,0.002,6.203,12.315],AUTHORITY[\"EPSG\",\"6695\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4695\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",26,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"3988\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3989, 1, 'Katanga 1955 / Katanga Gauss zone D', 'EPSG', 3989, 'PROJCS[\"Katanga 1955 / Katanga"
  " Gauss zone D\",GEOGCS[\"Katanga 1955\",DATUM[\"Katanga 1955\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,A"
  "UTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-102.283,-10.277,-257.396,3.976,0.002,6.203,12.315],AUTHORITY[\"EPSG\",\"6695\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4695\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"3989\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3991, 1, 'Puerto Rico State Plane CS of 1927', 'EPSG', 3991, 'PROJCS[\"Puerto Rico State Plane"
  " CS of 1927\",GEOGCS[\"Puerto Rico\",DATUM[\"Puerto Rico\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHO"
  "RITY[\"EPSG\",\"7008\"]],TOWGS84[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\",\"6139\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4139\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",17.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-66.4333333333334,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",18.4333333333333,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",18.0444444444444,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3992, 1, 'Puerto Rico / St. Croix', 'EPSG', 3992, 'PROJCS[\"Puerto Rico / St. Croix\",GEOGCS[\""
  "Puerto Rico\",DATUM[\"Puerto Rico\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]"
  "],TOWGS84[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\",\"6139\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4139\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",17.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-66.4333333333334,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",18.4333333333333,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",18.0444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"3992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3993, 1, 'Guam 1963 / Guam SPCS', 'EPSG', 3993, 'PROJCS[\"Guam 1963 / Guam SPCS\",GEOGCS[\"Gua"
  "m 1963\",DATUM[\"Guam 1963\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS"
  "84[-100,-248,259,0,0,0,0],AUTHORITY[\"EPSG\",\"6675\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4675\"]],PROJECTION[\"Guam Projection\",AUTHORITY[\"EPSG\",\"9831\"]],PARAMETER[\"Latitude of natural origin\",13.4"
  "724663527778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",144.748750705556,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",50000,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"3993\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3994, 1, 'WGS 84 / Mercator 41', 'EPSG', 3994, 'PROJCS[\"WGS 84 / Mercator 41\",GEOGCS[\"WGS 8"
  "4\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTH"
  "ORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Mercat"
  "or (variant B)\",AUTHORITY[\"EPSG\",\"9805\"]],PARAMETER[\"Latitude of 1st standard parallel\",-41,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Longitude of natural origin\",100,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"3994\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3995, 1, 'WGS 84 / Arctic Polar Stereographic', 'EPSG', 3995, 'PROJCS[\"WGS 84 / Arctic Polar "
  "Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTH"
  "ORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4326\"]],PROJECTION[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of stand"
  "ard parallel\",71,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",0,AUTHORITY[\"EPSG\",\"8833\"]],PAR"
  "AMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3995\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3996, 1, 'WGS 84 / IBCAO Polar Stereographic', 'EPSG', 3996, 'PROJCS[\"WGS 84 / IBCAO Polar St"
  "ereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHOR"
  "ITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4326\"]],PROJECTION[\"Polar Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standar"
  "d parallel\",75,AUTHORITY[\"EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",0,AUTHORITY[\"EPSG\",\"8833\"]],PARAM"
  "ETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"3996\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (3997, 1, 'WGS 84 / Dubai Local TM', 'EPSG', 3997, 'PROJCS[\"WGS 84 / Dubai Local TM\",GEOGCS[\""
  "WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",55.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"3997\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4001, 1, 'Unknown datum based upon the Airy 1830 ellipsoid', 'EPSG', 4001, 'GEOGCS[\"Unknown d"
  "atum based upon the Airy 1830 ellipsoid\",DATUM[\"Not specified (based on Airy 1830 ellipsoid)\",SPHEROID[\"Airy 1830"
  "\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],AUTHORITY[\"EPSG\",\"6001\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4001\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4002, 1, 'Unknown datum based upon the Airy Modified 1849 ellipsoid', 'EPSG', 4002, 'GEOGCS[\""
  "Unknown datum based upon the Airy Modified 1849 ellipsoid\",DATUM[\"Not specified (based on Airy Modified 1849 ellips"
  "oid)\",SPHEROID[\"Airy Modified 1849\",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],AUTHORITY[\"EPSG\",\"600"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4002\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4003, 1, 'Unknown datum based upon the Australian National Spheroid', 'EPSG', 4003, 'GEOGCS[\""
  "Unknown datum based upon the Australian National Spheroid\",DATUM[\"Not specified (based on Australian National Spher"
  "oid)\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7003\"]],AUTHORITY[\"EPSG\",\"60"
  "03\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4003\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4004, 1, 'Unknown datum based upon the Bessel 1841 ellipsoid', 'EPSG', 4004, 'GEOGCS[\"Unknown"
  " datum based upon the Bessel 1841 ellipsoid\",DATUM[\"Not specified (based on Bessel 1841 ellipsoid)\",SPHEROID[\"Bes"
  "sel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6004\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4004\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4005, 1, 'Unknown datum based upon the Bessel Modified ellipsoid', 'EPSG', 4005, 'GEOGCS[\"Unk"
  "nown datum based upon the Bessel Modified ellipsoid\",DATUM[\"Not specified (based on Bessel Modified ellipsoid)\",SP"
  "HEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6005\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4005\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4006, 1, 'Unknown datum based upon the Bessel Namibia ellipsoid', 'EPSG', 4006, 'GEOGCS[\"Unkn"
  "own datum based upon the Bessel Namibia ellipsoid\",DATUM[\"Not specified (based on Bessel Namibia ellipsoid)\",SPHER"
  "OID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7046\"]],AUTHORITY[\"EPSG\",\"6006\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4006\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4007, 1, 'Unknown datum based upon the Clarke 1858 ellipsoid', 'EPSG', 4007, 'GEOGCS[\"Unknown"
  " datum based upon the Clarke 1858 ellipsoid\",DATUM[\"Not specified (based on Clarke 1858 ellipsoid)\",SPHEROID[\"Cla"
  "rke 1858\",6378293.645208759,294.26067636926103,AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6007\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4007\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4008, 1, 'Unknown datum based upon the Clarke 1866 ellipsoid', 'EPSG', 4008, 'GEOGCS[\"Unknown"
  " datum based upon the Clarke 1866 ellipsoid\",DATUM[\"Not specified (based on Clarke 1866 ellipsoid)\",SPHEROID[\"Cla"
  "rke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6008\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4008\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4010, 1, 'Unknown datum based upon the Clarke 1880 (Benoit) ellipsoid', 'EPSG', 4010, 'GEOGCS["
  "\"Unknown datum based upon the Clarke 1880 (Benoit) ellipsoid\",DATUM[\"Not specified (based on Clarke 1880 (Benoit) "
  "ellipsoid)\",SPHEROID[\"Clarke 1880 (Benoit)\",6378300.789,293.4663155389811,AUTHORITY[\"EPSG\",\"7010\"]],AUTHORITY["
  "\"EPSG\",\"6010\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4010\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4011, 1, 'Unknown datum based upon the Clarke 1880 (IGN) ellipsoid', 'EPSG', 4011, 'GEOGCS[\"U"
  "nknown datum based upon the Clarke 1880 (IGN) ellipsoid\",DATUM[\"Not specified (based on Clarke 1880 (IGN) ellipsoid"
  ")\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"601"
  "1\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4011\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4012, 1, 'Unknown datum based upon the Clarke 1880 (RGS) ellipsoid', 'EPSG', 4012, 'GEOGCS[\"U"
  "nknown datum based upon the Clarke 1880 (RGS) ellipsoid\",DATUM[\"Not specified (based on Clarke 1880 (RGS) ellipsoid"
  ")\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6012\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4012\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4013, 1, 'Unknown datum based upon the Clarke 1880 (Arc) ellipsoid', 'EPSG', 4013, 'GEOGCS[\"U"
  "nknown datum based upon the Clarke 1880 (Arc) ellipsoid\",DATUM[\"Not specified (based on Clarke 1880 (Arc) ellipsoid"
  ")\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],AUTHORITY[\"EPSG\",\"6013\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4013\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4014, 1, 'Unknown datum based upon the Clarke 1880 (SGA 1922) ellipsoid', 'EPSG', 4014, 'GEOGC"
  "S[\"Unknown datum based upon the Clarke 1880 (SGA 1922) ellipsoid\",DATUM[\"Not specified (based on Clarke 1880 (SGA "
  "1922) ellipsoid)\",SPHEROID[\"Clarke 1880 (SGA 1922)\",6378249.2,293.46598,AUTHORITY[\"EPSG\",\"7014\"]],AUTHORITY[\""
  "EPSG\",\"6014\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4015, 1, 'Unknown datum based upon the Everest 1830 (1937 Adjustment) ellipsoid', 'EPSG', 4015"
  ", 'GEOGCS[\"Unknown datum based upon the Everest 1830 (1937 Adjustment) ellipsoid\",DATUM[\"Not specified (based on E"
  "verest 1830 (1937 Adjustment) ellipsoid)\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY"
  "[\"EPSG\",\"7015\"]],AUTHORITY[\"EPSG\",\"6015\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"401"
  "5\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4016, 1, 'Unknown datum based upon the Everest 1830 (1967 Definition) ellipsoid', 'EPSG', 4016"
  ", 'GEOGCS[\"Unknown datum based upon the Everest 1830 (1967 Definition) ellipsoid\",DATUM[\"Not specified (based on E"
  "verest 1830 (1967 Definition) ellipsoid)\",SPHEROID[\"Everest 1830 (1967 Definition)\",6377298.556,300.8017,AUTHORITY"
  "[\"EPSG\",\"7016\"]],AUTHORITY[\"EPSG\",\"6016\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"401"
  "6\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4018, 1, 'Unknown datum based upon the Everest 1830 Modified ellipsoid', 'EPSG', 4018, 'GEOGCS"
  "[\"Unknown datum based upon the Everest 1830 Modified ellipsoid\",DATUM[\"Not specified (based on Everest 1830 Modifi"
  "ed ellipsoid)\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7018\"]],AUTHORITY[\"EPS"
  "G\",\"6018\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4018\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4019, 1, 'Unknown datum based upon the GRS 1980 ellipsoid', 'EPSG', 4019, 'GEOGCS[\"Unknown da"
  "tum based upon the GRS 1980 ellipsoid\",DATUM[\"Not specified (based on GRS 1980 ellipsoid)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6019\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4019\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4020, 1, 'Unknown datum based upon the Helmert 1906 ellipsoid', 'EPSG', 4020, 'GEOGCS[\"Unknow"
  "n datum based upon the Helmert 1906 ellipsoid\",DATUM[\"Not specified (based on Helmert 1906 ellipsoid)\",SPHEROID[\""
  "Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"6020\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4020\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4021, 1, 'Unknown datum based upon the Indonesian National Spheroid', 'EPSG', 4021, 'GEOGCS[\""
  "Unknown datum based upon the Indonesian National Spheroid\",DATUM[\"Not specified (based on Indonesian National Spher"
  "oid)\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],AUTHORITY[\"EPSG\",\"6"
  "021\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4021\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4022, 1, 'Unknown datum based upon the International 1924 ellipsoid', 'EPSG', 4022, 'GEOGCS[\""
  "Unknown datum based upon the International 1924 ellipsoid\",DATUM[\"Not specified (based on International 1924 ellips"
  "oid)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6022\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4022\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4023, 1, 'MOLDREF99', 'EPSG', 4023, 'GEOGCS[\"MOLDREF99\",DATUM[\"MOLDREF99\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1032\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4023\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4024, 1, 'Unknown datum based upon the Krassowsky 1940 ellipsoid', 'EPSG', 4024, 'GEOGCS[\"Unk"
  "nown datum based upon the Krassowsky 1940 ellipsoid\",DATUM[\"Not specified (based on Krassowsky 1940 ellipsoid)\",SP"
  "HEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6024\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4024\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4025, 1, 'Unknown datum based upon the NWL 9D ellipsoid', 'EPSG', 4025, 'GEOGCS[\"Unknown datu"
  "m based upon the NWL 9D ellipsoid\",DATUM[\"Not specified (based on NWL 9D ellipsoid)\",SPHEROID[\"NWL 9D\",6378145,2"
  "98.25,AUTHORITY[\"EPSG\",\"7025\"]],AUTHORITY[\"EPSG\",\"6025\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4025\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4026, 1, 'MOLDREF99 / Moldova TM', 'EPSG', 4026, 'PROJCS[\"MOLDREF99 / Moldova TM\",GEOGCS[\"M"
  "OLDREF99\",DATUM[\"MOLDREF99\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,"
  "0,0,0,0,0],AUTHORITY[\"EPSG\",\"1032\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4023\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28.4111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.99994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4026\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4027, 1, 'Unknown datum based upon the Plessis 1817 ellipsoid', 'EPSG', 4027, 'GEOGCS[\"Unknow"
  "n datum based upon the Plessis 1817 ellipsoid\",DATUM[\"Not specified (based on Plessis 1817 ellipsoid)\",SPHEROID[\""
  "Plessis 1817\",6376523,308.64,AUTHORITY[\"EPSG\",\"7027\"]],AUTHORITY[\"EPSG\",\"6027\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4027\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4028, 1, 'Unknown datum based upon the Struve 1860 ellipsoid', 'EPSG', 4028, 'GEOGCS[\"Unknown"
  " datum based upon the Struve 1860 ellipsoid\",DATUM[\"Not specified (based on Struve 1860 ellipsoid)\",SPHEROID[\"Str"
  "uve 1860\",6378298.3,294.73,AUTHORITY[\"EPSG\",\"7028\"]],AUTHORITY[\"EPSG\",\"6028\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4029, 1, 'Unknown datum based upon the War Office ellipsoid', 'EPSG', 4029, 'GEOGCS[\"Unknown "
  "datum based upon the War Office ellipsoid\",DATUM[\"Not specified (based on War Office ellipsoid)\",SPHEROID[\"War Of"
  "fice\",6378300,296,AUTHORITY[\"EPSG\",\"7029\"]],AUTHORITY[\"EPSG\",\"6029\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4029\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4030, 1, 'Unknown datum based upon the WGS 84 ellipsoid', 'EPSG', 4030, 'GEOGCS[\"Unknown datu"
  "m based upon the WGS 84 ellipsoid\",DATUM[\"Not specified (based on WGS 84 ellipsoid)\",SPHEROID[\"WGS 84\",6378137,2"
  "98.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6030\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4030\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4031, 1, 'Unknown datum based upon the GEM 10C ellipsoid', 'EPSG', 4031, 'GEOGCS[\"Unknown dat"
  "um based upon the GEM 10C ellipsoid\",DATUM[\"Not specified (based on GEM 10C ellipsoid)\",SPHEROID[\"GEM 10C\",63781"
  "37,298.257223563,AUTHORITY[\"EPSG\",\"7031\"]],AUTHORITY[\"EPSG\",\"6031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4031\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4032, 1, 'Unknown datum based upon the OSU86F ellipsoid', 'EPSG', 4032, 'GEOGCS[\"Unknown datu"
  "m based upon the OSU86F ellipsoid\",DATUM[\"Not specified (based on OSU86F ellipsoid)\",SPHEROID[\"OSU86F\",6378136.2"
  ",298.257223563,AUTHORITY[\"EPSG\",\"7032\"]],AUTHORITY[\"EPSG\",\"6032\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4032\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4033, 1, 'Unknown datum based upon the OSU91A ellipsoid', 'EPSG', 4033, 'GEOGCS[\"Unknown datu"
  "m based upon the OSU91A ellipsoid\",DATUM[\"Not specified (based on OSU91A ellipsoid)\",SPHEROID[\"OSU91A\",6378136.3"
  ",298.257223563,AUTHORITY[\"EPSG\",\"7033\"]],AUTHORITY[\"EPSG\",\"6033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4033\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4034, 1, 'Unknown datum based upon the Clarke 1880 ellipsoid', 'EPSG', 4034, 'GEOGCS[\"Unknown"
  " datum based upon the Clarke 1880 ellipsoid\",DATUM[\"Not specified (based on Clarke 1880 ellipsoid)\",SPHEROID[\"Cla"
  "rke 1880\",6378249.144808011,293.46630765562986,AUTHORITY[\"EPSG\",\"7034\"]],AUTHORITY[\"EPSG\",\"6034\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4034\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4036, 1, 'Unknown datum based upon the GRS 1967 ellipsoid', 'EPSG', 4036, 'GEOGCS[\"Unknown da"
  "tum based upon the GRS 1967 ellipsoid\",DATUM[\"Not specified (based on GRS 1967 ellipsoid)\",SPHEROID[\"GRS 1967\",6"
  "378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"]],AUTHORITY[\"EPSG\",\"6036\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4036\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4037, 1, 'WGS 84 / TMzn35N', 'EPSG', 4037, 'PROJCS[\"WGS 84 / TMzn35N\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"4037\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4038, 1, 'WGS 84 / TMzn36N', 'EPSG', 4038, 'PROJCS[\"WGS 84 / TMzn36N\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"4038\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4041, 1, 'Unknown datum based upon the Average Terrestrial System 1977 ellipsoid', 'EPSG', 404"
  "1, 'GEOGCS[\"Unknown datum based upon the Average Terrestrial System 1977 ellipsoid\",DATUM[\"Not specified (based on"
  " Average Terrestrial System 1977 ellipsoid)\",SPHEROID[\"Average Terrestrial System 1977\",6378135,298.257,AUTHORITY["
  "\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4041\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4042, 1, 'Unknown datum based upon the Everest (1830 Definition) ellipsoid', 'EPSG', 4042, 'GE"
  "OGCS[\"Unknown datum based upon the Everest (1830 Definition) ellipsoid\",DATUM[\"Not specified (based on Everest (18"
  "30 Definition) ellipsoid)\",SPHEROID[\"Everest (1830 Definition)\",6377299.36559538,300.80172554336184,AUTHORITY[\"EP"
  "SG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4042\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4043, 1, 'Unknown datum based upon the WGS 72 ellipsoid', 'EPSG', 4043, 'GEOGCS[\"Unknown datu"
  "m based upon the WGS 72 ellipsoid\",DATUM[\"Not specified (based on WGS 72 ellipsoid)\",SPHEROID[\"WGS 72\",6378135,2"
  "98.26,AUTHORITY[\"EPSG\",\"7043\"]],AUTHORITY[\"EPSG\",\"6043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4043\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4044, 1, 'Unknown datum based upon the Everest 1830 (1962 Definition) ellipsoid', 'EPSG', 4044"
  ", 'GEOGCS[\"Unknown datum based upon the Everest 1830 (1962 Definition) ellipsoid\",DATUM[\"Not specified (based on E"
  "verest 1830 (1962 Definition) ellipsoid)\",SPHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,300.8017255,AUTHOR"
  "ITY[\"EPSG\",\"7044\"]],AUTHORITY[\"EPSG\",\"6044\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4044\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4045, 1, 'Unknown datum based upon the Everest 1830 (1975 Definition) ellipsoid', 'EPSG', 4045"
  ", 'GEOGCS[\"Unknown datum based upon the Everest 1830 (1975 Definition) ellipsoid\",DATUM[\"Not specified (based on E"
  "verest 1830 (1975 Definition) ellipsoid)\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300.8017255,AUTHOR"
  "ITY[\"EPSG\",\"7045\"]],AUTHORITY[\"EPSG\",\"6045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4045\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4046, 1, 'RGRDC 2005', 'EPSG', 4046, 'GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RD"
  "C 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4046\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4047, 1, 'Unspecified datum based upon the GRS 1980 Authalic Sphere', 'EPSG', 4047, 'GEOGCS[\""
  "Unspecified datum based upon the GRS 1980 Authalic Sphere\",DATUM[\"Not specified (based on GRS 1980 Authalic Sphere)"
  "\",SPHEROID[\"GRS 1980 Authalic Sphere\",6371007,0,AUTHORITY[\"EPSG\",\"7048\"]],AUTHORITY[\"EPSG\",\"6047\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4047\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4048, 1, 'RGRDC 2005 / Congo TM zone 12', 'EPSG', 4048, 'PROJCS[\"RGRDC 2005 / Congo TM zone 1"
  "2\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4048\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4049, 1, 'RGRDC 2005 / Congo TM zone 14', 'EPSG', 4049, 'PROJCS[\"RGRDC 2005 / Congo TM zone 1"
  "4\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",14,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4049\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4050, 1, 'RGRDC 2005 / Congo TM zone 16', 'EPSG', 4050, 'PROJCS[\"RGRDC 2005 / Congo TM zone 1"
  "6\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4050\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4051, 1, 'RGRDC 2005 / Congo TM zone 18', 'EPSG', 4051, 'PROJCS[\"RGRDC 2005 / Congo TM zone 1"
  "8\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4051\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4052, 1, 'Unspecified datum based upon the Clarke 1866 Authalic Sphere', 'EPSG', 4052, 'GEOGCS"
  "[\"Unspecified datum based upon the Clarke 1866 Authalic Sphere\",DATUM[\"Not specified (based on Clarke 1866 Authali"
  "c Sphere)\",SPHEROID[\"Clarke 1866 Authalic Sphere\",6370997,0,AUTHORITY[\"EPSG\",\"7052\"]],AUTHORITY[\"EPSG\",\"605"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4052\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4053, 1, 'Unspecified datum based upon the International 1924 Authalic Sphere', 'EPSG', 4053, "
  "'GEOGCS[\"Unspecified datum based upon the International 1924 Authalic Sphere\",DATUM[\"Not specified (based on Inter"
  "national 1924 Authalic Sphere)\",SPHEROID[\"International 1924 Authalic Sphere\",6371228,0,AUTHORITY[\"EPSG\",\"7057\""
  "]],AUTHORITY[\"EPSG\",\"6053\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4053\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4054, 1, 'Unspecified datum based upon the Hughes 1980 ellipsoid', 'EPSG', 4054, 'GEOGCS[\"Uns"
  "pecified datum based upon the Hughes 1980 ellipsoid\",DATUM[\"Not specified (based on Hughes 1980 ellipsoid)\",SPHERO"
  "ID[\"Hughes 1980\",6378273,298.279411123064,AUTHORITY[\"EPSG\",\"7058\"]],AUTHORITY[\"EPSG\",\"6054\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4054\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4056, 1, 'RGRDC 2005 / Congo TM zone 20', 'EPSG', 4056, 'PROJCS[\"RGRDC 2005 / Congo TM zone 2"
  "0\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4056\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4057, 1, 'RGRDC 2005 / Congo TM zone 22', 'EPSG', 4057, 'PROJCS[\"RGRDC 2005 / Congo TM zone 2"
  "2\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4058, 1, 'RGRDC 2005 / Congo TM zone 24', 'EPSG', 4058, 'PROJCS[\"RGRDC 2005 / Congo TM zone 2"
  "4\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4058\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4059, 1, 'RGRDC 2005 / Congo TM zone 26', 'EPSG', 4059, 'PROJCS[\"RGRDC 2005 / Congo TM zone 2"
  "6\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",26,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4059\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4060, 1, 'RGRDC 2005 / Congo TM zone 28', 'EPSG', 4060, 'PROJCS[\"RGRDC 2005 / Congo TM zone 2"
  "8\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4060\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4061, 1, 'RGRDC 2005 / UTM zone 33S', 'EPSG', 4061, 'PROJCS[\"RGRDC 2005 / UTM zone 33S\",GEOG"
  "CS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4061\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4062, 1, 'RGRDC 2005 / UTM zone 34S', 'EPSG', 4062, 'PROJCS[\"RGRDC 2005 / UTM zone 34S\",GEOG"
  "CS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4062\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4063, 1, 'RGRDC 2005 / UTM zone 35S', 'EPSG', 4063, 'PROJCS[\"RGRDC 2005 / UTM zone 35S\",GEOG"
  "CS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4063\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4071, 1, 'Chua / UTM zone 23S', 'EPSG', 4071, 'PROJCS[\"Chua / UTM zone 23S\",GEOGCS[\"Chua\","
  "DATUM[\"Chua\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-144.35,242.88,-33.2"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6224\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4224\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4071\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4075, 1, 'SREF98', 'EPSG', 4075, 'GEOGCS[\"SREF98\",DATUM[\"Serbian Reference Network 1998\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "1034\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4075\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4081, 1, 'REGCAN95', 'EPSG', 4081, 'GEOGCS[\"REGCAN95\",DATUM[\"Red Geodesica de Canarias 1995"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"1035\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4081\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4082, 1, 'REGCAN95 / UTM zone 27N', 'EPSG', 4082, 'PROJCS[\"REGCAN95 / UTM zone 27N\",GEOGCS[\""
  "REGCAN95\",DATUM[\"Red Geodesica de Canarias 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1035\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4081\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4082\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4083, 1, 'REGCAN95 / UTM zone 28N', 'EPSG', 4083, 'PROJCS[\"REGCAN95 / UTM zone 28N\",GEOGCS[\""
  "REGCAN95\",DATUM[\"Red Geodesica de Canarias 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1035\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4081\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4083\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4087, 1, 'WGS 84 / World Equidistant Cylindrical', 'EPSG', 4087, 'PROJCS[\"WGS 84 / World Equi"
  "distant Cylindrical\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.25722356"
  "3,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4326\"]],PROJECTION[\"Equidistant Cylindrical\",AUTHORITY[\"EPSG\",\"1028\"]],PARAMETER[\"Latitude of 1st sta"
  "ndard parallel\",0,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4087\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4088, 1, 'World Equidistant Cylindrical (Sphere)', 'EPSG', 4088, 'PROJCS[\"World Equidistant C"
  "ylindrical (Sphere)\",GEOGCS[\"Unspecified datum based upon the GRS 1980 Authalic Sphere\",DATUM[\"Not specified (bas"
  "ed on GRS 1980 Authalic Sphere)\",SPHEROID[\"GRS 1980 Authalic Sphere\",6371007,0,AUTHORITY[\"EPSG\",\"7048\"]],AUTHO"
  "RITY[\"EPSG\",\"6047\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4047\"]],PROJECTION[\"Equidis"
  "tant Cylindrical (Spherical)\",AUTHORITY[\"EPSG\",\"1029\"]],PARAMETER[\"Latitude of 1st standard parallel\",0,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4088\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4093, 1, 'ETRS89 / DKTM1', 'EPSG', 4093, 'PROJCS[\"ETRS89 / DKTM1\",GEOGCS[\"ETRS89\",DATUM[\""
  "European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4093\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4094, 1, 'ETRS89 / DKTM2', 'EPSG', 4094, 'PROJCS[\"ETRS89 / DKTM2\",GEOGCS[\"ETRS89\",DATUM[\""
  "European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4094\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4095, 1, 'ETRS89 / DKTM3', 'EPSG', 4095, 'PROJCS[\"ETRS89 / DKTM3\",GEOGCS[\"ETRS89\",DATUM[\""
  "European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",600000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4096, 1, 'ETRS89 / DKTM4', 'EPSG', 4096, 'PROJCS[\"ETRS89 / DKTM4\",GEOGCS[\"ETRS89\",DATUM[\""
  "European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4096\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4120, 1, 'Greek', 'EPSG', 4120, 'GEOGCS[\"Greek\",DATUM[\"Greek\",SPHEROID[\"Bessel 1841\",637"
  "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4121, 1, 'GGRS87', 'EPSG', 4121, 'GEOGCS[\"GGRS87\",DATUM[\"Greek Geodetic Reference System 19"
  "87\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-199.87,74.79,246.62,0,0,0,0],"
  "AUTHORITY[\"EPSG\",\"6121\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4122, 1, 'ATS77', 'EPSG', 4122, 'GEOGCS[\"ATS77\",DATUM[\"Average Terrestrial System 1977\",SP"
  "HEROID[\"Average Terrestrial System 1977\",6378135,298.257,AUTHORITY[\"EPSG\",\"7041\"]],AUTHORITY[\"EPSG\",\"6122\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4123, 1, 'KKJ', 'EPSG', 4123, 'GEOGCS[\"KKJ\",DATUM[\"Kartastokoordinaattijarjestelma (1966)\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-96.062,-82.428,-121.753,4.801,0.3"
  "45,-1.376,1.496],AUTHORITY[\"EPSG\",\"6123\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4123\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4124, 1, 'RT90', 'EPSG', 4124, 'GEOGCS[\"RT90\",DATUM[\"Rikets koordinatsystem 1990\",SPHEROID"
  "[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[414.1,41.3,603.1,-0.855,2.141,-7.023,0"
  "],AUTHORITY[\"EPSG\",\"6124\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4127, 1, 'Tete', 'EPSG', 4127, 'GEOGCS[\"Tete\",DATUM[\"Tete\",SPHEROID[\"Clarke 1866\",637820"
  "6.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115.064,-87.39,-101.716,-0.058,4.001,-2.062,9.366],AUTHO"
  "RITY[\"EPSG\",\"6127\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4128, 1, 'Madzansua', 'EPSG', 4128, 'GEOGCS[\"Madzansua\",DATUM[\"Madzansua\",SPHEROID[\"Clark"
  "e 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6128\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4129, 1, 'Observatario', 'EPSG', 4129, 'GEOGCS[\"Observatario\",DATUM[\"Observatario\",SPHEROI"
  "D[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-132,-110,-335,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6129\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4129\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4130, 1, 'Moznet', 'EPSG', 4130, 'GEOGCS[\"Moznet\",DATUM[\"Moznet (ITRF94)\",SPHEROID[\"WGS 8"
  "4\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6130\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4130\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4131, 1, 'Indian 1960', 'EPSG', 4131, 'GEOGCS[\"Indian 1960\",DATUM[\"Indian 1960\",SPHEROID[\""
  "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[182,915,344,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"6131\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4131\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4132, 1, 'FD58', 'EPSG', 4132, 'GEOGCS[\"FD58\",DATUM[\"Final Datum 1958\",SPHEROID[\"Clarke 1"
  "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-244.72,-162.773,400.75,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4132\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4133, 1, 'EST92', 'EPSG', 4133, 'GEOGCS[\"EST92\",DATUM[\"Estonia 1992\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0.055,-0.541,-0.185,0.0183,-0.0003,-0.007,-0.014],AUTHOR"
  "ITY[\"EPSG\",\"6133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4133\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4134, 1, 'PSD93', 'EPSG', 4134, 'GEOGCS[\"PSD93\",DATUM[\"PDO Survey Datum 1993\",SPHEROID[\"C"
  "larke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-180.624,-225.516,173.919,-0.81,-1.898,8"
  ".336,16.71006],AUTHORITY[\"EPSG\",\"6134\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4134\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4135, 1, 'Old Hawaiian', 'EPSG', 4135, 'GEOGCS[\"Old Hawaiian\",DATUM[\"Old Hawaiian\",SPHEROI"
  "D[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[61,-285,-181,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4135\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4136, 1, 'St. Lawrence Island', 'EPSG', 4136, 'GEOGCS[\"St. Lawrence Island\",DATUM[\"St. Lawr"
  "ence Island\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\""
  "6136\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4137, 1, 'St. Paul Island', 'EPSG', 4137, 'GEOGCS[\"St. Paul Island\",DATUM[\"St. Paul Island\""
  ",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6137\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4137\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4138, 1, 'St. George Island', 'EPSG', 4138, 'GEOGCS[\"St. George Island\",DATUM[\"St. George I"
  "sland\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6138\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4138\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4139, 1, 'Puerto Rico', 'EPSG', 4139, 'GEOGCS[\"Puerto Rico\",DATUM[\"Puerto Rico\",SPHEROID[\""
  "Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[11,72,-101,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6139\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4139\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4141, 1, 'Israel 1993', 'EPSG', 4141, 'GEOGCS[\"Israel 1993\",DATUM[\"Israel 1993\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-48,55,52,0,0,0,0],AUTHORITY[\"EPSG\",\"6141\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4141\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4142, 1, 'Locodjo 1965', 'EPSG', 4142, 'GEOGCS[\"Locodjo 1965\",DATUM[\"Locodjo 1965\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-125,53,467,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4142\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4143, 1, 'Abidjan 1987', 'EPSG', 4143, 'GEOGCS[\"Abidjan 1987\",DATUM[\"Abidjan 1987\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-124.76,53,466.79,0,0,0,0],AUTHORIT"
  "Y[\"EPSG\",\"6143\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4143\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4144, 1, 'Kalianpur 1937', 'EPSG', 4144, 'GEOGCS[\"Kalianpur 1937\",DATUM[\"Kalianpur 1937\",S"
  "PHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[214,804,268,0,0"
  ",0,0],AUTHORITY[\"EPSG\",\"6144\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4144\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4145, 1, 'Kalianpur 1962', 'EPSG', 4145, 'GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur 1962\",S"
  "PHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,300.8017255,AUTHORITY[\"EPSG\",\"7044\"]],TOWGS84[230.25,632.7"
  "6,161.03,-1.114,1.115,1.212,12.584],AUTHORITY[\"EPSG\",\"6145\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4145\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4146, 1, 'Kalianpur 1975', 'EPSG', 4146, 'GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",S"
  "PHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300.8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4147, 1, 'Hanoi 1972', 'EPSG', 4147, 'GEOGCS[\"Hanoi 1972\",DATUM[\"Hanoi 1972\",SPHEROID[\"Kr"
  "assowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-17.51,-108.32,-62.39,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4147\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4148, 1, 'Hartebeesthoek94', 'EPSG', 4148, 'GEOGCS[\"Hartebeesthoek94\",DATUM[\"Hartebeesthoek"
  "94\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6148\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4148\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4149, 1, 'CH1903', 'EPSG', 4149, 'GEOGCS[\"CH1903\",DATUM[\"CH1903\",SPHEROID[\"Bessel 1841\","
  "6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[660.077,13.551,369.344,-0.804816,-0.577692,-0.952236,5."
  "66],AUTHORITY[\"EPSG\",\"6149\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4149\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4150, 1, 'CH1903+', 'EPSG', 4150, 'GEOGCS[\"CH1903+\",DATUM[\"CH1903+\",SPHEROID[\"Bessel 1841"
  "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[674.374,15.056,405.346,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6150\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4150\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4151, 1, 'CHTRF95', 'EPSG', 4151, 'GEOGCS[\"CHTRF95\",DATUM[\"Swiss Terrestrial Reference Fram"
  "e 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6151\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4151\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4152, 1, 'NAD83(HARN)', 'EPSG', 4152, 'GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Ref"
  "erence Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AU"
  "THORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4153, 1, 'Rassadiran', 'EPSG', 4153, 'GEOGCS[\"Rassadiran\",DATUM[\"Rassadiran\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-133.63,-157.5,-158.62,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"6153\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4153\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4154, 1, 'ED50(ED77)', 'EPSG', 4154, 'GEOGCS[\"ED50(ED77)\",DATUM[\"European Datum 1950(1977)\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-110.33,-97.73,-119.85,0.3423,1.16"
  "34,0.2715,0.063],AUTHORITY[\"EPSG\",\"6154\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4154\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4155, 1, 'Dabola 1981', 'EPSG', 4155, 'GEOGCS[\"Dabola 1981\",DATUM[\"Dabola 1981\",SPHEROID[\""
  "Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-83,37,124,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6155\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4155\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4156, 1, 'S-JTSK', 'EPSG', 4156, 'GEOGCS[\"S-JTSK\",DATUM[\"System of the Unified Trigonometri"
  "cal Cadastral Network\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[589,76"
  ",480,0,0,0,0],AUTHORITY[\"EPSG\",\"6156\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4156\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4157, 1, 'Mount Dillon', 'EPSG', 4157, 'GEOGCS[\"Mount Dillon\",DATUM[\"Mount Dillon\",SPHEROI"
  "D[\"Clarke 1858\",6378293.645208759,294.26067636926103,AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6157\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4157\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4158, 1, 'Naparima 1955', 'EPSG', 4158, 'GEOGCS[\"Naparima 1955\",DATUM[\"Naparima 1955\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-0.465,372.095,171.736,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6158\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4158\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4159, 1, 'ELD79', 'EPSG', 4159, 'GEOGCS[\"ELD79\",DATUM[\"European Libyan Datum 1979\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-194.513,-63.978,-25.759,-3.4027,3.756,-3."
  "352,-0.9175],AUTHORITY[\"EPSG\",\"6159\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4159\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4160, 1, 'Chos Malal 1914', 'EPSG', 4160, 'GEOGCS[\"Chos Malal 1914\",DATUM[\"Chos Malal 1914\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6160\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4160\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4161, 1, 'Pampa del Castillo', 'EPSG', 4161, 'GEOGCS[\"Pampa del Castillo\",DATUM[\"Pampa del "
  "Castillo\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[27.5,14,186.4,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"6161\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4161\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4162, 1, 'Korean 1985', 'EPSG', 4162, 'GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHE"
  "ROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2."
  "347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4163, 1, 'Yemen NGN96', 'EPSG', 4163, 'GEOGCS[\"Yemen NGN96\",DATUM[\"Yemen National Geodetic "
  "Network 1996\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4163\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4164, 1, 'South Yemen', 'EPSG', 4164, 'GEOGCS[\"South Yemen\",DATUM[\"South Yemen\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-76,-138,67,0,0,0,0],AUTHORITY[\"EPSG\",\"6164\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4164\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4165, 1, 'Bissau', 'EPSG', 4165, 'GEOGCS[\"Bissau\",DATUM[\"Bissau\",SPHEROID[\"International "
  "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-173,253,27,0,0,0,0],AUTHORITY[\"EPSG\",\"6165\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4165\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4166, 1, 'Korean 1995', 'EPSG', 4166, 'GEOGCS[\"Korean 1995\",DATUM[\"Korean Datum 1995\",SPHE"
  "ROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6166\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4166\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4167, 1, 'NZGD2000', 'EPSG', 4167, 'GEOGCS[\"NZGD2000\",DATUM[\"New Zealand Geodetic Datum 200"
  "0\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4167\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4168, 1, 'Accra', 'EPSG', 4168, 'GEOGCS[\"Accra\",DATUM[\"Accra\",SPHEROID[\"War Office\",6378"
  "300,296,AUTHORITY[\"EPSG\",\"7029\"]],TOWGS84[-199,32,322,0,0,0,0],AUTHORITY[\"EPSG\",\"6168\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4168\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4169, 1, 'American Samoa 1962', 'EPSG', 4169, 'GEOGCS[\"American Samoa 1962\",DATUM[\"American"
  " Samoa 1962\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-115,118,426"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6169\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4169\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4170, 1, 'SIRGAS 1995', 'EPSG', 4170, 'GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Ge"
  "ocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS"
  "84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4171, 1, 'RGF93', 'EPSG', 4171, 'GEOGCS[\"RGF93\",DATUM[\"Reseau Geodesique Francais 1993\",SP"
  "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6"
  "171\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4173, 1, 'IRENET95', 'EPSG', 4173, 'GEOGCS[\"IRENET95\",DATUM[\"IRENET95\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6173\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4174, 1, 'Sierra Leone 1924', 'EPSG', 4174, 'GEOGCS[\"Sierra Leone 1924\",DATUM[\"Sierra Leone"
  " Colony 1924\",SPHEROID[\"War Office\",6378300,296,AUTHORITY[\"EPSG\",\"7029\"]],AUTHORITY[\"EPSG\",\"6174\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4174\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4175, 1, 'Sierra Leone 1968', 'EPSG', 4175, 'GEOGCS[\"Sierra Leone 1968\",DATUM[\"Sierra Leone"
  " 1968\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-88,4,101,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"6175\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4175\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4176, 1, 'Australian Antarctic', 'EPSG', 4176, 'GEOGCS[\"Australian Antarctic\",DATUM[\"Austra"
  "lian Antarctic Datum 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0"
  ",0,0,0],AUTHORITY[\"EPSG\",\"6176\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4176\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4178, 1, 'Pulkovo 1942(83)', 'EPSG', 4178, 'GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(8"
  "3)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1."
  "1],AUTHORITY[\"EPSG\",\"6178\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4179, 1, 'Pulkovo 1942(58)', 'EPSG', 4179, 'GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(5"
  "8)\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053"
  ",0.844,-0.84],AUTHORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4180, 1, 'EST97', 'EPSG', 4180, 'GEOGCS[\"EST97\",DATUM[\"Estonia 1997\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6180\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4180\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4181, 1, 'Luxembourg 1930', 'EPSG', 4181, 'GEOGCS[\"Luxembourg 1930\",DATUM[\"Luxembourg 1930\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-189.6806,18.3463,-42.7695,-0.3374"
  "6,-3.09264,2.53861,0.4598],AUTHORITY[\"EPSG\",\"6181\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4181\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4182, 1, 'Azores Occidental 1939', 'EPSG', 4182, 'GEOGCS[\"Azores Occidental 1939\",DATUM[\"Az"
  "ores Occidental Islands 1939\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-422"
  ".651,-172.995,84.02,0,0,0,0],AUTHORITY[\"EPSG\",\"6182\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4182\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4183, 1, 'Azores Central 1948', 'EPSG', 4183, 'GEOGCS[\"Azores Central 1948\",DATUM[\"Azores C"
  "entral Islands 1948\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-104,167,-38,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6183\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4183\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4184, 1, 'Azores Oriental 1940', 'EPSG', 4184, 'GEOGCS[\"Azores Oriental 1940\",DATUM[\"Azores"
  " Oriental Islands 1940\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-203,141,5"
  "3,0,0,0,0],AUTHORITY[\"EPSG\",\"6184\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4184\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4188, 1, 'OSNI 1952', 'EPSG', 4188, 'GEOGCS[\"OSNI 1952\",DATUM[\"OSNI 1952\",SPHEROID[\"Airy "
  "1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],TOWGS84[482.5,-130.6,564.6,-1.042,-0.214,-0.631,8.15],AU"
  "THORITY[\"EPSG\",\"6188\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4188\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4189, 1, 'REGVEN', 'EPSG', 4189, 'GEOGCS[\"REGVEN\",DATUM[\"Red Geodesica Venezolana\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6189\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4189\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4190, 1, 'POSGAR 98', 'EPSG', 4190, 'GEOGCS[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argent"
  "inas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORIT"
  "Y[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4190\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4191, 1, 'Albanian 1987', 'EPSG', 4191, 'GEOGCS[\"Albanian 1987\",DATUM[\"Albanian 1987\",SPHE"
  "ROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-44.183,-0.58,-38.489,2.3867,2.7072,-3.5"
  "196,-8.2703],AUTHORITY[\"EPSG\",\"6191\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4191\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4192, 1, 'Douala 1948', 'EPSG', 4192, 'GEOGCS[\"Douala 1948\",DATUM[\"Douala 1948\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.1,-174.7,-87.7,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6192\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4193, 1, 'Manoca 1962', 'EPSG', 4193, 'GEOGCS[\"Manoca 1962\",DATUM[\"Manoca 1962\",SPHEROID[\""
  "Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-70.9,-151.8,-41.4,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6193\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4194, 1, 'Qornoq 1927', 'EPSG', 4194, 'GEOGCS[\"Qornoq 1927\",DATUM[\"Qornoq 1927\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[163.511,127.533,-159.789,0,0,0.814,-0.6],AUTHO"
  "RITY[\"EPSG\",\"6194\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4194\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4195, 1, 'Scoresbysund 1952', 'EPSG', 4195, 'GEOGCS[\"Scoresbysund 1952\",DATUM[\"Scoresbysund"
  " 1952\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[105,326,-102.5,0,0,0.814,-0"
  ".6],AUTHORITY[\"EPSG\",\"6195\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4196, 1, 'Ammassalik 1958', 'EPSG', 4196, 'GEOGCS[\"Ammassalik 1958\",DATUM[\"Ammassalik 1958\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-45,417,-3.5,0,0,0.814,-0.6],AUTHO"
  "RITY[\"EPSG\",\"6196\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4196\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4197, 1, 'Garoua', 'EPSG', 4197, 'GEOGCS[\"Garoua\",DATUM[\"Garoua\",SPHEROID[\"Clarke 1880 (R"
  "GS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6197\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4197\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4198, 1, 'Kousseri', 'EPSG', 4198, 'GEOGCS[\"Kousseri\",DATUM[\"Kousseri\",SPHEROID[\"Clarke 1"
  "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6198\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4198\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4199, 1, 'Egypt 1930', 'EPSG', 4199, 'GEOGCS[\"Egypt 1930\",DATUM[\"Egypt 1930\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6199\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4199\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4200, 1, 'Pulkovo 1995', 'EPSG', 4200, 'GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROI"
  "D[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHOR"
  "ITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4201, 1, 'Adindan', 'EPSG', 4201, 'GEOGCS[\"Adindan\",DATUM[\"Adindan\",SPHEROID[\"Clarke 1880"
  " (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-166,-15,204,0,0,0,0],AUTHORITY[\"EPSG\",\"6201\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4201\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4202, 1, 'AGD66', 'EPSG', 4202, 'GEOGCS[\"AGD66\",DATUM[\"Australian Geodetic Datum 1966\",SPH"
  "EROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0"
  ".008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4202\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4203, 1, 'AGD84', 'EPSG', 4203, 'GEOGCS[\"AGD84\",DATUM[\"Australian Geodetic Datum 1984\",SPH"
  "EROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0"
  ".292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4203\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4204, 1, 'Ain el Abd', 'EPSG', 4204, 'GEOGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID"
  "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,"
  "3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292"
  "519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4205, 1, 'Afgooye', 'EPSG', 4205, 'GEOGCS[\"Afgooye\",DATUM[\"Afgooye\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-43,-163,45,0,0,0,0],AUTHORITY[\"EPSG\",\"6205\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4205\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4206, 1, 'Agadez', 'EPSG', 4206, 'GEOGCS[\"Agadez\",DATUM[\"Agadez\",SPHEROID[\"Clarke 1880 (I"
  "GN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6206\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4206\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4207, 1, 'Lisbon', 'EPSG', 4207, 'GEOGCS[\"Lisbon\",DATUM[\"Lisbon 1937\",SPHEROID[\"Internati"
  "onal 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-288.885,-91.744,126.244,-1.691,0.41,-0.211,-4.598],AUT"
  "HORITY[\"EPSG\",\"6207\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4207\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4208, 1, 'Aratu', 'EPSG', 4208, 'GEOGCS[\"Aratu\",DATUM[\"Aratu\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-151.99,287.04,-147.45,0,0,0,0],AUTHORITY[\"EPSG\",\"6208\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4208\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4209, 1, 'Arc 1950', 'EPSG', 4209, 'GEOGCS[\"Arc 1950\",DATUM[\"Arc 1950\",SPHEROID[\"Clarke 1"
  "880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-138,-105,-289,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4209\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4210, 1, 'Arc 1960', 'EPSG', 4210, 'GEOGCS[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1"
  "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4210\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4211, 1, 'Batavia', 'EPSG', 4211, 'GEOGCS[\"Batavia\",DATUM[\"Batavia\",SPHEROID[\"Bessel 1841"
  "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-377.7,675.1,-52.2,0,0,0,0],AUTHORITY[\"EPSG\",\"621"
  "1\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4211\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4212, 1, 'Barbados 1938', 'EPSG', 4212, 'GEOGCS[\"Barbados 1938\",DATUM[\"Barbados 1938\",SPHE"
  "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[31.95,300.99,419.19,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6212\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4212\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4213, 1, 'Beduaram', 'EPSG', 4213, 'GEOGCS[\"Beduaram\",DATUM[\"Beduaram\",SPHEROID[\"Clarke 1"
  "880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-106,-87,188,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6213\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4213\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4214, 1, 'Beijing 1954', 'EPSG', 4214, 'GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROI"
  "D[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY["
  "\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4215, 1, 'Belge 1950', 'EPSG', 4215, 'GEOGCS[\"Belge 1950\",DATUM[\"Reseau National Belge 1950"
  "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6215\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4215\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4216, 1, 'Bermuda 1957', 'EPSG', 4216, 'GEOGCS[\"Bermuda 1957\",DATUM[\"Bermuda 1957\",SPHEROI"
  "D[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-292.295,248.758,429.447,4.9971,2"
  ".99,6.6906,1.0289],AUTHORITY[\"EPSG\",\"6216\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4216\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4217, 1, 'NAD83 / BLM 59N (ftUS)', 'EPSG', 4217, 'PROJCS[\"NAD83 / BLM 59N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4217\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4218, 1, 'Bogota 1975', 'EPSG', 4218, 'GEOGCS[\"Bogota 1975\",DATUM[\"Bogota 1975\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"621"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4218\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4219, 1, 'Bukit Rimpah', 'EPSG', 4219, 'GEOGCS[\"Bukit Rimpah\",DATUM[\"Bukit Rimpah\",SPHEROI"
  "D[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-384,664,-48,0,0,0,0],AUTHORITY[\"EPS"
  "G\",\"6219\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4219\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4220, 1, 'Camacupa', 'EPSG', 4220, 'GEOGCS[\"Camacupa\",DATUM[\"Camacupa\",SPHEROID[\"Clarke 1"
  "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-41.057,-374.564,-226.287,0,0,0.554,0.219],AUTH"
  "ORITY[\"EPSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4221, 1, 'Campo Inchauspe', 'EPSG', 4221, 'GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4221\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4222, 1, 'Cape', 'EPSG', 4222, 'GEOGCS[\"Cape\",DATUM[\"Cape\",SPHEROID[\"Clarke 1880 (Arc)\","
  "6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0],AUTHORITY[\"EPSG\",\"6222\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4223, 1, 'Carthage', 'EPSG', 4223, 'GEOGCS[\"Carthage\",DATUM[\"Carthage\",SPHEROID[\"Clarke 1"
  "880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-263,6,431,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4223\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4224, 1, 'Chua', 'EPSG', 4224, 'GEOGCS[\"Chua\",DATUM[\"Chua\",SPHEROID[\"International 1924\""
  ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-144.35,242.88,-33.2,0,0,0,0],AUTHORITY[\"EPSG\",\"6224\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4224\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4225, 1, 'Corrego Alegre 1970-72', 'EPSG', 4225, 'GEOGCS[\"Corrego Alegre 1970-72\",DATUM[\"Co"
  "rrego Alegre 1970-72\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.05,168."
  "28,-3.82,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4225\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4227, 1, 'Deir ez Zor', 'EPSG', 4227, 'GEOGCS[\"Deir ez Zor\",DATUM[\"Deir ez Zor\",SPHEROID[\""
  "Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-83.58,-397.54,458.78,-17.595,-"
  "2.847,4.256,3.225],AUTHORITY[\"EPSG\",\"6227\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4227\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4229, 1, 'Egypt 1907', 'EPSG', 4229, 'GEOGCS[\"Egypt 1907\",DATUM[\"Egypt 1907\",SPHEROID[\"He"
  "lmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-127.535,113.495,-12.7,-1.603747,0.153612,5.364408,5"
  ".33745],AUTHORITY[\"EPSG\",\"6229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4229\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4230, 1, 'ED50', 'EPSG', 4230, 'GEOGCS[\"ED50\",DATUM[\"European Datum 1950\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUT"
  "HORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4230\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4231, 1, 'ED87', 'EPSG', 4231, 'GEOGCS[\"ED87\",DATUM[\"European Datum 1987\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-83.11,-97.38,-117.22,0.00569290865241987,-0.044697"
  "5835137458,0.0442850539012516,0.1218],AUTHORITY[\"EPSG\",\"6231\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4231\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4232, 1, 'Fahud', 'EPSG', 4232, 'GEOGCS[\"Fahud\",DATUM[\"Fahud\",SPHEROID[\"Clarke 1880 (RGS)"
  "\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-333.102,-11.02,230.69,0,0,0.554,0.219],AUTHORITY[\"EPSG"
  "\",\"6232\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4232\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4236, 1, 'Hu Tzu Shan 1950', 'EPSG', 4236, 'GEOGCS[\"Hu Tzu Shan 1950\",DATUM[\"Hu Tzu Shan 19"
  "50\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-637,-549,-203,0,0,0,0],AUTHOR"
  "ITY[\"EPSG\",\"6236\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4236\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4237, 1, 'HD72', 'EPSG', 4237, 'GEOGCS[\"HD72\",DATUM[\"Hungarian Datum 1972\",SPHEROID[\"GRS "
  "1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"]],TOWGS84[52.684,-71.194,-13.975,-0.312,-0.1063,-0.3729,1.01"
  "91],AUTHORITY[\"EPSG\",\"6237\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4237\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4238, 1, 'ID74', 'EPSG', 4238, 'GEOGCS[\"ID74\",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Ind"
  "onesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\"]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0."
  "689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4239, 1, 'Indian 1954', 'EPSG', 4239, 'GEOGCS[\"Indian 1954\",DATUM[\"Indian 1954\",SPHEROID[\""
  "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4239\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4240, 1, 'Indian 1975', 'EPSG', 4240, 'GEOGCS[\"Indian 1975\",DATUM[\"Indian 1975\",SPHEROID[\""
  "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[210,814,289,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"6240\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4240\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4241, 1, 'Jamaica 1875', 'EPSG', 4241, 'GEOGCS[\"Jamaica 1875\",DATUM[\"Jamaica 1875\",SPHEROI"
  "D[\"Clarke 1880\",6378249.144808011,293.46630765562986,AUTHORITY[\"EPSG\",\"7034\"]],AUTHORITY[\"EPSG\",\"6241\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4241\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4242, 1, 'JAD69', 'EPSG', 4242, 'GEOGCS[\"JAD69\",DATUM[\"Jamaica 1969\",SPHEROID[\"Clarke 186"
  "6\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-33.722,153.789,94.959,-8.581,-4.478,4.54,8.95]"
  ",AUTHORITY[\"EPSG\",\"6242\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4242\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4243, 1, 'Kalianpur 1880', 'EPSG', 4243, 'GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",S"
  "PHEROID[\"Everest (1830 Definition)\",6377299.36559538,300.80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"E"
  "PSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4243\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4244, 1, 'Kandawala', 'EPSG', 4244, 'GEOGCS[\"Kandawala\",DATUM[\"Kandawala\",SPHEROID[\"Evere"
  "st 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[-97,787,86,0,0,0,0],AUTHORITY["
  "\"EPSG\",\"6244\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4244\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4245, 1, 'Kertau 1968', 'EPSG', 4245, 'GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\""
  "Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPS"
  "G\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4245\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4246, 1, 'KOC', 'EPSG', 4246, 'GEOGCS[\"KOC\",DATUM[\"Kuwait Oil Company\",SPHEROID[\"Clarke 1"
  "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-294.7,-200.1,525.5,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6246\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4247, 1, 'La Canoa', 'EPSG', 4247, 'GEOGCS[\"La Canoa\",DATUM[\"La Canoa\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-270.933,115.599,-360.226,5.266,1.238,-2.381,-5.109],A"
  "UTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4247\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4248, 1, 'PSAD56', 'EPSG', 4248, 'GEOGCS[\"PSAD56\",DATUM[\"Provisional South American Datum 1"
  "956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,"
  "3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4249, 1, 'Lake', 'EPSG', 4249, 'GEOGCS[\"Lake\",DATUM[\"Lake\",SPHEROID[\"International 1924\""
  ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6249\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4250, 1, 'Leigon', 'EPSG', 4250, 'GEOGCS[\"Leigon\",DATUM[\"Leigon\",SPHEROID[\"Clarke 1880 (R"
  "GS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-130,29,364,0,0,0,0],AUTHORITY[\"EPSG\",\"6250\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4250\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4251, 1, 'Liberia 1964', 'EPSG', 4251, 'GEOGCS[\"Liberia 1964\",DATUM[\"Liberia 1964\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-90,40,88,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"6251\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4251\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4252, 1, 'Lome', 'EPSG', 4252, 'GEOGCS[\"Lome\",DATUM[\"Lome\",SPHEROID[\"Clarke 1880 (IGN)\","
  "6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6252\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4253, 1, 'Luzon 1911', 'EPSG', 4253, 'GEOGCS[\"Luzon 1911\",DATUM[\"Luzon 1911\",SPHEROID[\"Cl"
  "arke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-133,-77,-51,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6253\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4254, 1, 'Hito XVIII 1963', 'EPSG', 4254, 'GEOGCS[\"Hito XVIII 1963\",DATUM[\"Hito XVIII 1963\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[18.38,192.45,96.82,0.056,-0.142,-0"
  ".2,-0.0013],AUTHORITY[\"EPSG\",\"6254\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4254\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4255, 1, 'Herat North', 'EPSG', 4255, 'GEOGCS[\"Herat North\",DATUM[\"Herat North\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-333,-222,114,0,0,0,0],AUTHORITY[\"EPSG\",\"62"
  "55\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4256, 1, 'Mahe 1971', 'EPSG', 4256, 'GEOGCS[\"Mahe 1971\",DATUM[\"Mahe 1971\",SPHEROID[\"Clark"
  "e 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[41,-220,-134,0,0,0,0],AUTHORITY[\"EPSG\",\"6"
  "256\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4257, 1, 'Makassar', 'EPSG', 4257, 'GEOGCS[\"Makassar\",DATUM[\"Makassar\",SPHEROID[\"Bessel 1"
  "841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-587.8,519.75,145.76,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6257\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4258, 1, 'ETRS89', 'EPSG', 4258, 'GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference Sys"
  "tem 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4259, 1, 'Malongo 1987', 'EPSG', 4259, 'GEOGCS[\"Malongo 1987\",DATUM[\"Malongo 1987\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-254.1,-5.36,-100.29,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6259\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4261, 1, 'Merchich', 'EPSG', 4261, 'GEOGCS[\"Merchich\",DATUM[\"Merchich\",SPHEROID[\"Clarke 1"
  "880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6261\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4261\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4262, 1, 'Massawa', 'EPSG', 4262, 'GEOGCS[\"Massawa\",DATUM[\"Massawa\",SPHEROID[\"Bessel 1841"
  "\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[639,405,60,0,0,0,0],AUTHORITY[\"EPSG\",\"6262\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4262\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4263, 1, 'Minna', 'EPSG', 4263, 'GEOGCS[\"Minna\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)"
  "\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-92,-93,122,0,0,0,0],AUTHORITY[\"EPSG\",\"6263\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4263\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4265, 1, 'Monte Mario', 'EPSG', 4265, 'GEOGCS[\"Monte Mario\",DATUM[\"Monte Mario\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-104.1,-49.1,-9.9,0.971,-2.917,0.714,-11.68],A"
  "UTHORITY[\"EPSG\",\"6265\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4265\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4266, 1, 'M\\'poraloko', 'EPSG', 4266, 'GEOGCS[\"M\\'poraloko\",DATUM[\"M\\'poraloko\",SPHEROID[\""
  "Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-80.7,-132.5,41.1,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"6266\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4266\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4267, 1, 'NAD27', 'EPSG', 4267, 'GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID"
  "[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.844257"
  "7,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4267\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4269, 1, 'NAD83', 'EPSG', 4269, 'GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID"
  "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4270, 1, 'Nahrwan 1967', 'EPSG', 4270, 'GEOGCS[\"Nahrwan 1967\",DATUM[\"Nahrwan 1967\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-253.4392,-148.452,386.5267,0.15605"
  ",0.43,-0.1013,-0.0424],AUTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degr"
  "ee\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4"
  "270\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4271, 1, 'Naparima 1972', 'EPSG', 4271, 'GEOGCS[\"Naparima 1972\",DATUM[\"Naparima 1972\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-2,374,172,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6271\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4271\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4272, 1, 'NZGD49', 'EPSG', 4272, 'GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\","
  "SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,"
  "-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4273, 1, 'NGO 1948', 'EPSG', 4273, 'GEOGCS[\"NGO 1948\",DATUM[\"NGO 1948\",SPHEROID[\"Bessel M"
  "odified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"7005\"]],TOWGS84[278.3,93,474.5,7.889,0.05,-6.61,6.21],AUTHORI"
  "TY[\"EPSG\",\"6273\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4273\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4274, 1, 'Datum 73', 'EPSG', 4274, 'GEOGCS[\"Datum 73\",DATUM[\"Datum 73\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-239.749,88.181,30.488,0.263,0.082,1.211,2.229],AUTHOR"
  "ITY[\"EPSG\",\"6274\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4274\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4275, 1, 'NTF', 'EPSG', 4275, 'GEOGCS[\"NTF\",DATUM[\"Nouvelle Triangulation Francaise\",SPHER"
  "OID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-168,-60,320,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6275\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4275\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4276, 1, 'NSWC 9Z-2', 'EPSG', 4276, 'GEOGCS[\"NSWC 9Z-2\",DATUM[\"NSWC 9Z-2\",SPHEROID[\"NWL 9"
  "D\",6378145,298.25,AUTHORITY[\"EPSG\",\"7025\"]],AUTHORITY[\"EPSG\",\"6276\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4276\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4277, 1, 'OSGB 1936', 'EPSG', 4277, 'GEOGCS[\"OSGB 1936\",DATUM[\"OSGB 1936\",SPHEROID[\"Airy "
  "1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],TOWGS84[446.448,-125.157,542.06,0.15,0.247,0.842,-20.489"
  "],AUTHORITY[\"EPSG\",\"6277\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4277\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4278, 1, 'OSGB70', 'EPSG', 4278, 'GEOGCS[\"OSGB70\",DATUM[\"OSGB 1970 (SN)\",SPHEROID[\"Airy 1"
  "830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],AUTHORITY[\"EPSG\",\"6278\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4278\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4279, 1, 'OS(SN)80', 'EPSG', 4279, 'GEOGCS[\"OS(SN)80\",DATUM[\"OS (SN) 1980\",SPHEROID[\"Airy"
  " 1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\",\"7001\"]],AUTHORITY[\"EPSG\",\"6279\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4279\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4280, 1, 'Padang', 'EPSG', 4280, 'GEOGCS[\"Padang\",DATUM[\"Padang 1884\",SPHEROID[\"Bessel 18"
  "41\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-377,681,-50,0,0,0,0],AUTHORITY[\"EPSG\",\"6280\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4280\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4281, 1, 'Palestine 1923', 'EPSG', 4281, 'GEOGCS[\"Palestine 1923\",DATUM[\"Palestine 1923\",S"
  "PHEROID[\"Clarke 1880 (Benoit)\",6378300.789,293.4663155389811,AUTHORITY[\"EPSG\",\"7010\"]],TOWGS84[-275.7224,94.782"
  "4,340.8944,-8.001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\"6281\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4281\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4282, 1, 'Pointe Noire', 'EPSG', 4282, 'GEOGCS[\"Pointe Noire\",DATUM[\"Congo 1960 Pointe Noir"
  "e\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-148,51,-291,0,0"
  ",0,0],AUTHORITY[\"EPSG\",\"6282\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4282\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4283, 1, 'GDA94', 'EPSG', 4283, 'GEOGCS[\"GDA94\",DATUM[\"Geocentric Datum of Australia 1994\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4284, 1, 'Pulkovo 1942', 'EPSG', 4284, 'GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROI"
  "D[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"E"
  "PSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4285, 1, 'Qatar 1974', 'EPSG', 4285, 'GEOGCS[\"Qatar 1974\",DATUM[\"Qatar 1974\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-128.033,-283.697,21.052,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6285\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4285\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4286, 1, 'Qatar 1948', 'EPSG', 4286, 'GEOGCS[\"Qatar 1948\",DATUM[\"Qatar 1948\",SPHEROID[\"He"
  "lmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"6286\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4286\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4288, 1, 'Loma Quintana', 'EPSG', 4288, 'GEOGCS[\"Loma Quintana\",DATUM[\"Loma Quintana\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6288\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4288\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4289, 1, 'Amersfoort', 'EPSG', 4289, 'GEOGCS[\"Amersfoort\",DATUM[\"Amersfoort\",SPHEROID[\"Be"
  "ssel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[565.4171,50.3319,465.5524,-0.39895738824313"
  "4,0.343987817378283,-1.87740163998045,4.0725],AUTHORITY[\"EPSG\",\"6289\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4289\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4292, 1, 'Sapper Hill 1943', 'EPSG', 4292, 'GEOGCS[\"Sapper Hill 1943\",DATUM[\"Sapper Hill 19"
  "43\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-355,21,72,0,0,0,0],AUTHORITY["
  "\"EPSG\",\"6292\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4292\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4293, 1, 'Schwarzeck', 'EPSG', 4293, 'GEOGCS[\"Schwarzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Be"
  "ssel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4293\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4295, 1, 'Serindung', 'EPSG', 4295, 'GEOGCS[\"Serindung\",DATUM[\"Serindung\",SPHEROID[\"Besse"
  "l 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6295\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4295\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4297, 1, 'Tananarive', 'EPSG', 4297, 'GEOGCS[\"Tananarive\",DATUM[\"Tananarive 1925\",SPHEROID"
  "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-198.383,-240.517,-107.909,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6297\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4297\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4298, 1, 'Timbalai 1948', 'EPSG', 4298, 'GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai 1948\",SPHE"
  "ROID[\"Everest 1830 (1967 Definition)\",6377298.556,300.8017,AUTHORITY[\"EPSG\",\"7016\"]],TOWGS84[-679,669,-48,0,0,0"
  ",0],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4298\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4299, 1, 'TM65', 'EPSG', 4299, 'GEOGCS[\"TM65\",DATUM[\"TM65\",SPHEROID[\"Airy Modified 1849\""
  ",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],TOWGS84[482.5,-130.6,564.6,-1.042,-0.214,-0.631,8.15],AUTHORIT"
  "Y[\"EPSG\",\"6299\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4299\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4300, 1, 'TM75', 'EPSG', 4300, 'GEOGCS[\"TM75\",DATUM[\"Geodetic Datum of 1965\",SPHEROID[\"Ai"
  "ry Modified 1849\",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],TOWGS84[482.5,-130.6,564.6,-1.042,-0.214,-0."
  "631,8.15],AUTHORITY[\"EPSG\",\"6300\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4300\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4301, 1, 'Tokyo', 'EPSG', 4301, 'GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",637"
  "7397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4301\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4302, 1, 'Trinidad 1903', 'EPSG', 4302, 'GEOGCS[\"Trinidad 1903\",DATUM[\"Trinidad 1903\",SPHE"
  "ROID[\"Clarke 1858\",6378293.645208759,294.26067636926103,AUTHORITY[\"EPSG\",\"7007\"]],TOWGS84[-61.702,284.488,472.0"
  "52,0,0,0,0],AUTHORITY[\"EPSG\",\"6302\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4302\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4303, 1, 'TC(1948)', 'EPSG', 4303, 'GEOGCS[\"TC(1948)\",DATUM[\"Trucial Coast 1948\",SPHEROID["
  "\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTHORITY[\"EPSG\",\"6303\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4303\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4304, 1, 'Voirol 1875', 'EPSG', 4304, 'GEOGCS[\"Voirol 1875\",DATUM[\"Voirol 1875\",SPHEROID[\""
  "Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-73,-247,227,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6304\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4304\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4306, 1, 'Bern 1938', 'EPSG', 4306, 'GEOGCS[\"Bern 1938\",DATUM[\"Bern 1938\",SPHEROID[\"Besse"
  "l 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6306\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4306\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4307, 1, 'Nord Sahara 1959', 'EPSG', 4307, 'GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 19"
  "59\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404."
  "6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"6307\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4307\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4308, 1, 'RT38', 'EPSG', 4308, 'GEOGCS[\"RT38\",DATUM[\"Stockholm 1938\",SPHEROID[\"Bessel 184"
  "1\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6308\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4308\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4309, 1, 'Yacare', 'EPSG', 4309, 'GEOGCS[\"Yacare\",DATUM[\"Yacare\",SPHEROID[\"International "
  "1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-124.45,183.74,44.64,-0.4384,0.5446,-0.9706,-2.1365],AUTHORI"
  "TY[\"EPSG\",\"6309\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4309\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4310, 1, 'Yoff', 'EPSG', 4310, 'GEOGCS[\"Yoff\",DATUM[\"Yoff\",SPHEROID[\"Clarke 1880 (IGN)\","
  "6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-30,190,89,0,0,0,0],AUTHORITY[\"EPSG\",\"6310\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4310\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4311, 1, 'Zanderij', 'EPSG', 4311, 'GEOGCS[\"Zanderij\",DATUM[\"Zanderij\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,120,-358,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4312, 1, 'MGI', 'EPSG', 4312, 'GEOGCS[\"MGI\",DATUM[\"Militar-Geographische Institut\",SPHEROI"
  "D[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5."
  "297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4313, 1, 'Belge 1972', 'EPSG', 4313, 'GEOGCS[\"Belge 1972\",DATUM[\"Reseau National Belge 1972"
  "\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-106.8686,52.2978,-103.7239,0.33"
  "66,-0.457,1.8422,-1.2747],AUTHORITY[\"EPSG\",\"6313\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4313\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4314, 1, 'DHDN', 'EPSG', 4314, 'GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID"
  "[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6."
  "7],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4315, 1, 'Conakry 1905', 'EPSG', 4315, 'GEOGCS[\"Conakry 1905\",DATUM[\"Conakry 1905\",SPHEROI"
  "D[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4315\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4316, 1, 'Dealul Piscului 1930', 'EPSG', 4316, 'GEOGCS[\"Dealul Piscului 1930\",DATUM[\"Dealul"
  " Piscului 1930\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[103.25,-100.4,-307"
  ".19,0,0,0,0],AUTHORITY[\"EPSG\",\"6316\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4316\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4318, 1, 'NGN', 'EPSG', 4318, 'GEOGCS[\"NGN\",DATUM[\"National Geodetic Network\",SPHEROID[\"W"
  "GS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-3.2,-5.7,2.8,0,0,0,0],AUTHORITY[\"EPSG\",\"6318\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4319, 1, 'KUDAMS', 'EPSG', 4319, 'GEOGCS[\"KUDAMS\",DATUM[\"Kuwait Utility\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[226.702,-193.337,-35.371,-2.229,-4.391,9.238,0.9798]"
  ",AUTHORITY[\"EPSG\",\"6319\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4319\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4322, 1, 'WGS 72', 'EPSG', 4322, 'GEOGCS[\"WGS 72\",DATUM[\"World Geodetic System 1972\",SPHER"
  "OID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"632"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4324, 1, 'WGS 72BE', 'EPSG', 4324, 'GEOGCS[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephem"
  "eris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\""
  "EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4326, 1, 'WGS 84', 'EPSG', 4326, 'GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHER"
  "OID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4390, 1, 'Kertau 1968 / Johor Grid', 'EPSG', 4390, 'PROJCS[\"Kertau 1968 / Johor Grid\",GEOGCS"
  "[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\"7"
  "018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural orig"
  "in\",2.04258333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",103.562758333333,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"4390\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4391, 1, 'Kertau 1968 / Sembilan and Melaka Grid', 'EPSG', 4391, 'PROJCS[\"Kertau 1968 / Sembi"
  "lan and Melaka Grid\",GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300"
  ".8017,AUTHORITY[\"EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETE"
  "R[\"Latitude of natural origin\",2.71228333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",101.941166666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",-242.005,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",-948.547,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4392, 1, 'Kertau 1968 / Pahang Grid', 'EPSG', 4392, 'PROJCS[\"Kertau 1968 / Pahang Grid\",GEOG"
  "CS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\""
  "7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",3.71097222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102.436177777778,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"4392\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4393, 1, 'Kertau 1968 / Selangor Grid', 'EPSG', 4393, 'PROJCS[\"Kertau 1968 / Selangor Grid\","
  "GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG"
  "\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natura"
  "l origin\",3.68034444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",101.508244444444,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",-21759.438,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",55960.906,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AX"
  "IS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4394, 1, 'Kertau 1968 / Terengganu Grid', 'EPSG', 4394, 'PROJCS[\"Kertau 1968 / Terengganu Gri"
  "d\",GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\""
  "EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of na"
  "tural origin\",4.94614166666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102.895208333"
  "333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORT"
  "H],AUTHORITY[\"EPSG\",\"4394\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4395, 1, 'Kertau 1968 / Pinang Grid', 'EPSG', 4395, 'PROJCS[\"Kertau 1968 / Pinang Grid\",GEOG"
  "CS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG\",\""
  "7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",5.421325,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",100.345869444444,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"4395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4396, 1, 'Kertau 1968 / Kedah and Perlis Grid', 'EPSG', 4396, 'PROJCS[\"Kertau 1968 / Kedah an"
  "d Perlis Grid\",GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,"
  "AUTHORITY[\"EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"La"
  "titude of natural origin\",5.96514722222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",1"
  "00.637594444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS"
  "[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4396\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4397, 1, 'Kertau 1968 / Perak Revised Grid', 'EPSG', 4397, 'PROJCS[\"Kertau 1968 / Perak Revis"
  "ed Grid\",GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHOR"
  "ITY[\"EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude"
  " of natural origin\",4.85938055555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",100.816"
  "766666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",133453.669,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],"
  "AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4397\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4398, 1, 'Kertau 1968 / Kelantan Grid', 'EPSG', 4398, 'PROJCS[\"Kertau 1968 / Kelantan Grid\","
  "GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG"
  "\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natura"
  "l origin\",5.89392222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102.177291666667,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"4398\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4399, 1, 'NAD27 / BLM 59N (ftUS)', 'EPSG', 4399, 'PROJCS[\"NAD27 / BLM 59N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4399\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4400, 1, 'NAD27 / BLM 60N (ftUS)', 'EPSG', 4400, 'PROJCS[\"NAD27 / BLM 60N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4400\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4401, 1, 'NAD27 / BLM 1N (ftUS)', 'EPSG', 4401, 'PROJCS[\"NAD27 / BLM 1N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4401\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4402, 1, 'NAD27 / BLM 2N (ftUS)', 'EPSG', 4402, 'PROJCS[\"NAD27 / BLM 2N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4402\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4403, 1, 'NAD27 / BLM 3N (ftUS)', 'EPSG', 4403, 'PROJCS[\"NAD27 / BLM 3N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4403\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4404, 1, 'NAD27 / BLM 4N (ftUS)', 'EPSG', 4404, 'PROJCS[\"NAD27 / BLM 4N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4404\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4405, 1, 'NAD27 / BLM 5N (ftUS)', 'EPSG', 4405, 'PROJCS[\"NAD27 / BLM 5N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4405\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4406, 1, 'NAD27 / BLM 6N (ftUS)', 'EPSG', 4406, 'PROJCS[\"NAD27 / BLM 6N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4406\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4407, 1, 'NAD27 / BLM 7N (ftUS)', 'EPSG', 4407, 'PROJCS[\"NAD27 / BLM 7N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4407\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4408, 1, 'NAD27 / BLM 8N (ftUS)', 'EPSG', 4408, 'PROJCS[\"NAD27 / BLM 8N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4408\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4409, 1, 'NAD27 / BLM 9N (ftUS)', 'EPSG', 4409, 'PROJCS[\"NAD27 / BLM 9N (ftUS)\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4409\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4410, 1, 'NAD27 / BLM 10N (ftUS)', 'EPSG', 4410, 'PROJCS[\"NAD27 / BLM 10N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4410\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4411, 1, 'NAD27 / BLM 11N (ftUS)', 'EPSG', 4411, 'PROJCS[\"NAD27 / BLM 11N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4411\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4412, 1, 'NAD27 / BLM 12N (ftUS)', 'EPSG', 4412, 'PROJCS[\"NAD27 / BLM 12N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4412\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4413, 1, 'NAD27 / BLM 13N (ftUS)', 'EPSG', 4413, 'PROJCS[\"NAD27 / BLM 13N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4413\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4414, 1, 'NAD83(HARN) / Guam Map Grid', 'EPSG', 4414, 'PROJCS[\"NAD83(HARN) / Guam Map Grid\","
  "GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",13.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",144.7611111111"
  "11,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4414\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4415, 1, 'Katanga 1955 / Katanga Lambert', 'EPSG', 4415, 'PROJCS[\"Katanga 1955 / Katanga Lamb"
  "ert\",GEOGCS[\"Katanga 1955\",DATUM[\"Katanga 1955\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-102.283,-10.277,-257.396,3.976,0.002,6.203,12.315],AUTHORITY[\"EPSG\",\"6695\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4695\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",-9,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",26,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-6.5,AUTHORITY[\"EPSG"
  "\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",-11.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "4415\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4417, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 7', 'EPSG', 4417, 'PROJCS[\"Pulkovo 19"
  "42(83) / 3-degree Gauss-Kruger zone 7\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"617"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4417\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4418, 1, 'NAD27 / BLM 18N (ftUS)', 'EPSG', 4418, 'PROJCS[\"NAD27 / BLM 18N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4418\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4419, 1, 'NAD27 / BLM 19N (ftUS)', 'EPSG', 4419, 'PROJCS[\"NAD27 / BLM 19N (ftUS)\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4419\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4420, 1, 'NAD83 / BLM 60N (ftUS)', 'EPSG', 4420, 'PROJCS[\"NAD83 / BLM 60N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4420\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4421, 1, 'NAD83 / BLM 1N (ftUS)', 'EPSG', 4421, 'PROJCS[\"NAD83 / BLM 1N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4421\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4422, 1, 'NAD83 / BLM 2N (ftUS)', 'EPSG', 4422, 'PROJCS[\"NAD83 / BLM 2N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4422\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4423, 1, 'NAD83 / BLM 3N (ftUS)', 'EPSG', 4423, 'PROJCS[\"NAD83 / BLM 3N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4423\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4424, 1, 'NAD83 / BLM 4N (ftUS)', 'EPSG', 4424, 'PROJCS[\"NAD83 / BLM 4N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4424\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4425, 1, 'NAD83 / BLM 5N (ftUS)', 'EPSG', 4425, 'PROJCS[\"NAD83 / BLM 5N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4425\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4426, 1, 'NAD83 / BLM 6N (ftUS)', 'EPSG', 4426, 'PROJCS[\"NAD83 / BLM 6N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4426\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4427, 1, 'NAD83 / BLM 7N (ftUS)', 'EPSG', 4427, 'PROJCS[\"NAD83 / BLM 7N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4427\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4428, 1, 'NAD83 / BLM 8N (ftUS)', 'EPSG', 4428, 'PROJCS[\"NAD83 / BLM 8N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4428\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4429, 1, 'NAD83 / BLM 9N (ftUS)', 'EPSG', 4429, 'PROJCS[\"NAD83 / BLM 9N (ftUS)\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4429\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4430, 1, 'NAD83 / BLM 10N (ftUS)', 'EPSG', 4430, 'PROJCS[\"NAD83 / BLM 10N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4430\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4431, 1, 'NAD83 / BLM 11N (ftUS)', 'EPSG', 4431, 'PROJCS[\"NAD83 / BLM 11N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4431\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4432, 1, 'NAD83 / BLM 12N (ftUS)', 'EPSG', 4432, 'PROJCS[\"NAD83 / BLM 12N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4432\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4433, 1, 'NAD83 / BLM 13N (ftUS)', 'EPSG', 4433, 'PROJCS[\"NAD83 / BLM 13N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4433\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4434, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 8', 'EPSG', 4434, 'PROJCS[\"Pulkovo 19"
  "42(83) / 3-degree Gauss-Kruger zone 8\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky "
  "1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"617"
  "8\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4434\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4437, 1, 'NAD83(NSRS2007) / Puerto Rico and Virgin Is.', 'EPSG', 4437, 'PROJCS[\"NAD83(NSRS200"
  "7) / Puerto Rico and Virgin Is.\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conform"
  "al (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",17.8333333333333,AUTHORITY[\"EPSG\",\""
  "8821\"]],PARAMETER[\"Longitude of false origin\",-66.4333333333334,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude"
  " of 1st standard parallel\",18.4333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",18.0444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4437\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4438, 1, 'NAD83 / BLM 18N (ftUS)', 'EPSG', 4438, 'PROJCS[\"NAD83 / BLM 18N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4438\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4439, 1, 'NAD83 / BLM 19N (ftUS)', 'EPSG', 4439, 'PROJCS[\"NAD83 / BLM 19N (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4439\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4455, 1, 'NAD27 / Pennsylvania South', 'EPSG', 4455, 'PROJCS[\"NAD27 / Pennsylvania South\",GE"
  "OGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"E"
  "PSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"E"
  "PSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic C"
  "onformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",40.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",39.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012"
  "1924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4455\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4456, 1, 'NAD27 / New York Long Island', 'EPSG', 4456, 'PROJCS[\"NAD27 / New York Long Island\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.5111111111111,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1"
  "st standard parallel\",41.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",40.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"4456\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4457, 1, 'NAD83 / South Dakota North (ftUS)', 'EPSG', 4457, 'PROJCS[\"NAD83 / South Dakota Nor"
  "th (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-10"
  "0,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.6944444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"4457\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4462, 1, 'WGS 84 / Australian Centre for Remote Sensing Lambert', 'EPSG', 4462, 'PROJCS[\"WGS "
  "84 / Australian Centre for Remote Sensing Lambert\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\""
  "WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",-27,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",1"
  "32,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-18,AUTHORITY[\"EPSG\",\"8823\"]],PA"
  "RAMETER[\"Latitude of 2nd standard parallel\",-36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4462\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4463, 1, 'RGSPM06', 'EPSG', 4463, 'GEOGCS[\"RGSPM06\",DATUM[\"Reseau Geodesique de Saint Pierr"
  "e et Miquelon 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"1038\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4463\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4467, 1, 'RGSPM06 / UTM zone 21N', 'EPSG', 4467, 'PROJCS[\"RGSPM06 / UTM zone 21N\",GEOGCS[\"R"
  "GSPM06\",DATUM[\"Reseau Geodesique de Saint Pierre et Miquelon 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT"
  "HORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1038\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4463\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4467\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4470, 1, 'RGM04', 'EPSG', 4470, 'GEOGCS[\"RGM04\",DATUM[\"Reseau Geodesique de Mayotte 2004\","
  "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "1036\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4470\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4471, 1, 'RGM04 / UTM zone 38S', 'EPSG', 4471, 'PROJCS[\"RGM04 / UTM zone 38S\",GEOGCS[\"RGM04"
  "\",DATUM[\"Reseau Geodesique de Mayotte 2004\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1036\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4470\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4471\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4475, 1, 'Cadastre 1997', 'EPSG', 4475, 'GEOGCS[\"Cadastre 1997\",DATUM[\"Cadastre 1997\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-381.788,-57.501,-256.673,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"1037\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4475\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4483, 1, 'Mexico ITRF92', 'EPSG', 4483, 'GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"104"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4483\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4484, 1, 'Mexico ITRF92 / UTM zone 11N', 'EPSG', 4484, 'PROJCS[\"Mexico ITRF92 / UTM zone 11N\""
  ",GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4483\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4484\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4485, 1, 'Mexico ITRF92 / UTM zone 12N', 'EPSG', 4485, 'PROJCS[\"Mexico ITRF92 / UTM zone 12N\""
  ",GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4483\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4485\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4486, 1, 'Mexico ITRF92 / UTM zone 13N', 'EPSG', 4486, 'PROJCS[\"Mexico ITRF92 / UTM zone 13N\""
  ",GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4483\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4486\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4487, 1, 'Mexico ITRF92 / UTM zone 14N', 'EPSG', 4487, 'PROJCS[\"Mexico ITRF92 / UTM zone 14N\""
  ",GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4483\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4487\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4488, 1, 'Mexico ITRF92 / UTM zone 15N', 'EPSG', 4488, 'PROJCS[\"Mexico ITRF92 / UTM zone 15N\""
  ",GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4483\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4488\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4489, 1, 'Mexico ITRF92 / UTM zone 16N', 'EPSG', 4489, 'PROJCS[\"Mexico ITRF92 / UTM zone 16N\""
  ",GEOGCS[\"Mexico ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4483\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4489\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4490, 1, 'China Geodetic Coordinate System 2000', 'EPSG', 4490, 'GEOGCS[\"China Geodetic Coord"
  "inate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AU"
  "THORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4490\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4491, 1, 'CGCS2000 / Gauss-Kruger zone 13', 'EPSG', 4491, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 13\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",13500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4491\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4492, 1, 'CGCS2000 / Gauss-Kruger zone 14', 'EPSG', 4492, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 14\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",14500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4492\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4493, 1, 'CGCS2000 / Gauss-Kruger zone 15', 'EPSG', 4493, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 15\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",15500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4493\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4494, 1, 'CGCS2000 / Gauss-Kruger zone 16', 'EPSG', 4494, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 16\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",16500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4494\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4495, 1, 'CGCS2000 / Gauss-Kruger zone 17', 'EPSG', 4495, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 17\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",17500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4495\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4496, 1, 'CGCS2000 / Gauss-Kruger zone 18', 'EPSG', 4496, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 18\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",18500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4496\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4497, 1, 'CGCS2000 / Gauss-Kruger zone 19', 'EPSG', 4497, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 19\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",19500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4497\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4498, 1, 'CGCS2000 / Gauss-Kruger zone 20', 'EPSG', 4498, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 20\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4498\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4499, 1, 'CGCS2000 / Gauss-Kruger zone 21', 'EPSG', 4499, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 21\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",21500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4499\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4500, 1, 'CGCS2000 / Gauss-Kruger zone 22', 'EPSG', 4500, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 22\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",22500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4500\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4501, 1, 'CGCS2000 / Gauss-Kruger zone 23', 'EPSG', 4501, 'PROJCS[\"CGCS2000 / Gauss-Kruger zo"
  "ne 23\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",23500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4501\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4502, 1, 'CGCS2000 / Gauss-Kruger CM 75E', 'EPSG', 4502, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM "
  "75E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4502\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4503, 1, 'CGCS2000 / Gauss-Kruger CM 81E', 'EPSG', 4503, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM "
  "81E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4503\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4504, 1, 'CGCS2000 / Gauss-Kruger CM 87E', 'EPSG', 4504, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM "
  "87E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4504\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4505, 1, 'CGCS2000 / Gauss-Kruger CM 93E', 'EPSG', 4505, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM "
  "93E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4505\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4506, 1, 'CGCS2000 / Gauss-Kruger CM 99E', 'EPSG', 4506, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM "
  "99E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4506\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4507, 1, 'CGCS2000 / Gauss-Kruger CM 105E', 'EPSG', 4507, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM"
  " 105E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4507\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4508, 1, 'CGCS2000 / Gauss-Kruger CM 111E', 'EPSG', 4508, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM"
  " 111E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4508\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4509, 1, 'CGCS2000 / Gauss-Kruger CM 117E', 'EPSG', 4509, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM"
  " 117E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4509\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4510, 1, 'CGCS2000 / Gauss-Kruger CM 123E', 'EPSG', 4510, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM"
  " 123E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4510\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4511, 1, 'CGCS2000 / Gauss-Kruger CM 129E', 'EPSG', 4511, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM"
  " 129E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4511\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4512, 1, 'CGCS2000 / Gauss-Kruger CM 135E', 'EPSG', 4512, 'PROJCS[\"CGCS2000 / Gauss-Kruger CM"
  " 135E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4512\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4513, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 25', 'EPSG', 4513, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 25\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4513\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4514, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 26', 'EPSG', 4514, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 26\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",78,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",26500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4514\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4515, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 27', 'EPSG', 4515, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 27\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",27500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4515\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4516, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 28', 'EPSG', 4516, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 28\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",84,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",28500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4516\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4517, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 29', 'EPSG', 4517, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 29\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",29500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4517\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4518, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 30', 'EPSG', 4518, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 30\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",30500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4518\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4519, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 31', 'EPSG', 4519, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 31\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4520, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 32', 'EPSG', 4520, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 32\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4520\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4521, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 33', 'EPSG', 4521, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 33\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",33500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4522, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 34', 'EPSG', 4522, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 34\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",34500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4523, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 35', 'EPSG', 4523, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 35\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",35500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4524, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 36', 'EPSG', 4524, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 36\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",108,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",36500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4524\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4525, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 37', 'EPSG', 4525, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 37\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",37500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4525\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4526, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 38', 'EPSG', 4526, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 38\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",38500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4526\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4527, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 39', 'EPSG', 4527, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 39\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",39500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4527\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4528, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 40', 'EPSG', 4528, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 40\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",120,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",40500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4529, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 41', 'EPSG', 4529, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 41\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",41500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4530, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 42', 'EPSG', 4530, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 42\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",126,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",42500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4530\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4531, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 43', 'EPSG', 4531, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 43\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",43500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4532, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 44', 'EPSG', 4532, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 44\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",44500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4532\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4533, 1, 'CGCS2000 / 3-degree Gauss-Kruger zone 45', 'EPSG', 4533, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger zone 45\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",45500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4534, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 75E', 'EPSG', 4534, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 75E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4534\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4535, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 78E', 'EPSG', 4535, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 78E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",78,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4536, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 81E', 'EPSG', 4536, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 81E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4536\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4537, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 84E', 'EPSG', 4537, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 84E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",84,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4537\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4538, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 87E', 'EPSG', 4538, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 87E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4539, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 90E', 'EPSG', 4539, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 90E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4540, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 93E', 'EPSG', 4540, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 93E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4540\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4541, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 96E', 'EPSG', 4541, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 96E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4541\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4542, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 99E', 'EPSG', 4542, 'PROJCS[\"CGCS2000 / 3-degre"
  "e Gauss-Kruger CM 99E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4542\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4543, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 102E', 'EPSG', 4543, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 102E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4543\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4544, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 105E', 'EPSG', 4544, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 105E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4544\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4545, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 108E', 'EPSG', 4545, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 108E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",108,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4545\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4546, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 111E', 'EPSG', 4546, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 111E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4546\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4547, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 114E', 'EPSG', 4547, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 114E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4547\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4548, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 117E', 'EPSG', 4548, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 117E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4548\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4549, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 120E', 'EPSG', 4549, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 120E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",120,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4549\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4550, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 123E', 'EPSG', 4550, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 123E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4550\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4551, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 126E', 'EPSG', 4551, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 126E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",126,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4551\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4552, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 129E', 'EPSG', 4552, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 129E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4552\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4553, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 132E', 'EPSG', 4553, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 132E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4553\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4554, 1, 'CGCS2000 / 3-degree Gauss-Kruger CM 135E', 'EPSG', 4554, 'PROJCS[\"CGCS2000 / 3-degr"
  "ee Gauss-Kruger CM 135E\",GEOGCS[\"China Geodetic Coordinate System 2000\",DATUM[\"China 2000\",SPHEROID[\"CGCS2000\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"1024\"]],AUTHORITY[\"EPSG\",\"1043\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4490\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4554\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4555, 1, 'New Beijing', 'EPSG', 4555, 'GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4558, 1, 'RRAF 1991', 'EPSG', 4558, 'GEOGCS[\"RRAF 1991\",DATUM[\"Reseau de Reference des Anti"
  "lles Francaises 1991\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"1047\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4558\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4559, 1, 'RRAF 1991 / UTM zone 20N', 'EPSG', 4559, 'PROJCS[\"RRAF 1991 / UTM zone 20N\",GEOGCS"
  "[\"RRAF 1991\",DATUM[\"Reseau de Reference des Antilles Francaises 1991\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1047\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4558\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4559\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4568, 1, 'New Beijing / Gauss-Kruger zone 13', 'EPSG', 4568, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 13\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",13500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4568\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4569, 1, 'New Beijing / Gauss-Kruger zone 14', 'EPSG', 4569, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 14\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",14500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4569\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4570, 1, 'New Beijing / Gauss-Kruger zone 15', 'EPSG', 4570, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 15\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",15500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4570\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4571, 1, 'New Beijing / Gauss-Kruger zone 16', 'EPSG', 4571, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 16\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",16500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4571\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4572, 1, 'New Beijing / Gauss-Kruger zone 17', 'EPSG', 4572, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 17\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",17500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4572\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4573, 1, 'New Beijing / Gauss-Kruger zone 18', 'EPSG', 4573, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 18\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4573\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4574, 1, 'New Beijing / Gauss-Kruger zone 19', 'EPSG', 4574, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 19\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",19500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4574\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4575, 1, 'New Beijing / Gauss-Kruger zone 20', 'EPSG', 4575, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 20\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4575\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4576, 1, 'New Beijing / Gauss-Kruger zone 21', 'EPSG', 4576, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 21\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",21500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4576\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4577, 1, 'New Beijing / Gauss-Kruger zone 22', 'EPSG', 4577, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 22\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",22500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4577\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4578, 1, 'New Beijing / Gauss-Kruger zone 23', 'EPSG', 4578, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger zone 23\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",23500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4578\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4579, 1, 'New Beijing / Gauss-Kruger CM 75E', 'EPSG', 4579, 'PROJCS[\"New Beijing / Gauss-Krug"
  "er CM 75E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4579\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4580, 1, 'New Beijing / Gauss-Kruger CM 81E', 'EPSG', 4580, 'PROJCS[\"New Beijing / Gauss-Krug"
  "er CM 81E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4580\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4581, 1, 'New Beijing / Gauss-Kruger CM 87E', 'EPSG', 4581, 'PROJCS[\"New Beijing / Gauss-Krug"
  "er CM 87E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4581\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4582, 1, 'New Beijing / Gauss-Kruger CM 93E', 'EPSG', 4582, 'PROJCS[\"New Beijing / Gauss-Krug"
  "er CM 93E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4582\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4583, 1, 'New Beijing / Gauss-Kruger CM 99E', 'EPSG', 4583, 'PROJCS[\"New Beijing / Gauss-Krug"
  "er CM 99E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\""
  ",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4583\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4584, 1, 'New Beijing / Gauss-Kruger CM 105E', 'EPSG', 4584, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger CM 105E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4584\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4585, 1, 'New Beijing / Gauss-Kruger CM 111E', 'EPSG', 4585, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger CM 111E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4585\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4586, 1, 'New Beijing / Gauss-Kruger CM 117E', 'EPSG', 4586, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger CM 117E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4586\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4587, 1, 'New Beijing / Gauss-Kruger CM 123E', 'EPSG', 4587, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger CM 123E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4587\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4588, 1, 'New Beijing / Gauss-Kruger CM 129E', 'EPSG', 4588, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger CM 129E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4588\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4589, 1, 'New Beijing / Gauss-Kruger CM 135E', 'EPSG', 4589, 'PROJCS[\"New Beijing / Gauss-Kru"
  "ger CM 135E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPS"
  "G\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4555\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4589\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4600, 1, 'Anguilla 1957', 'EPSG', 4600, 'GEOGCS[\"Anguilla 1957\",DATUM[\"Anguilla 1957\",SPHE"
  "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6600\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4601, 1, 'Antigua 1943', 'EPSG', 4601, 'GEOGCS[\"Antigua 1943\",DATUM[\"Antigua 1943\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-255,-15,71,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6601\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4601\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4602, 1, 'Dominica 1945', 'EPSG', 4602, 'GEOGCS[\"Dominica 1945\",DATUM[\"Dominica 1945\",SPHE"
  "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[725,685,536,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6602\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4602\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4603, 1, 'Grenada 1953', 'EPSG', 4603, 'GEOGCS[\"Grenada 1953\",DATUM[\"Grenada 1953\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[72,213.7,93,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6603\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4603\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4604, 1, 'Montserrat 1958', 'EPSG', 4604, 'GEOGCS[\"Montserrat 1958\",DATUM[\"Montserrat 1958\""
  ",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[174,359,365,0,0,0,0],AUTHOR"
  "ITY[\"EPSG\",\"6604\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4604\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4605, 1, 'St. Kitts 1955', 'EPSG', 4605, 'GEOGCS[\"St. Kitts 1955\",DATUM[\"St. Kitts 1955\",S"
  "PHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[9,183,236,0,0,0,0],AUTHORITY["
  "\"EPSG\",\"6605\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4605\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4606, 1, 'St. Lucia 1955', 'EPSG', 4606, 'GEOGCS[\"St. Lucia 1955\",DATUM[\"St. Lucia 1955\",S"
  "PHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-153,153,307,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6606\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4606\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4607, 1, 'St. Vincent 1945', 'EPSG', 4607, 'GEOGCS[\"St. Vincent 1945\",DATUM[\"St. Vincent 19"
  "45\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[195.671,332.517,274.607"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6607\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4607\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4608, 1, 'NAD27(76)', 'EPSG', 4608, 'GEOGCS[\"NAD27(76)\",DATUM[\"North American Datum 1927 (1"
  "976)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"6608\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4608\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4609, 1, 'NAD27(CGQ77)', 'EPSG', 4609, 'GEOGCS[\"NAD27(CGQ77)\",DATUM[\"North American Datum 1"
  "927 (CGQ77)\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\""
  "6609\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4609\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4610, 1, 'Xian 1980', 'EPSG', 4610, 'GEOGCS[\"Xian 1980\",DATUM[\"Xian 1980\",SPHEROID[\"IAG 1"
  "975\",6378140,298.257,AUTHORITY[\"EPSG\",\"7049\"]],AUTHORITY[\"EPSG\",\"6610\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4610\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4611, 1, 'Hong Kong 1980', 'EPSG', 4611, 'GEOGCS[\"Hong Kong 1980\",DATUM[\"Hong Kong 1980\",S"
  "PHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-162.619,-276.959,-161.764,0.067753,"
  "-2.243649,-1.158827,-1.094246],AUTHORITY[\"EPSG\",\"6611\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4611\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4612, 1, 'JGD2000', 'EPSG', 4612, 'GEOGCS[\"JGD2000\",DATUM[\"Japanese Geodetic Datum 2000\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6612\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4612\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4613, 1, 'Segara', 'EPSG', 4613, 'GEOGCS[\"Segara\",DATUM[\"Gunung Segara\",SPHEROID[\"Bessel "
  "1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-403,684,41,0,0,0,0],AUTHORITY[\"EPSG\",\"6613\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4613\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4614, 1, 'QND95', 'EPSG', 4614, 'GEOGCS[\"QND95\",DATUM[\"Qatar National Datum 1995\",SPHEROID"
  "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-119.4248,-303.65872,-11.00061,1.164298,0.1"
  "74458,1.096259,3.657065],AUTHORITY[\"EPSG\",\"6614\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4614\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4615, 1, 'Porto Santo', 'EPSG', 4615, 'GEOGCS[\"Porto Santo\",DATUM[\"Porto Santo 1936\",SPHER"
  "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-499,-249,314,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6615\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4615\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4616, 1, 'Selvagem Grande', 'EPSG', 4616, 'GEOGCS[\"Selvagem Grande\",DATUM[\"Selvagem Grande\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-289,-124,60,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6616\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4616\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4617, 1, 'NAD83(CSRS)', 'EPSG', 4617, 'GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial R"
  "eference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5"
  "129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4618, 1, 'SAD69', 'EPSG', 4618, 'GEOGCS[\"SAD69\",DATUM[\"South American Datum 1969\",SPHEROID"
  "[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6"
  "618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4618\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4619, 1, 'SWEREF99', 'EPSG', 4619, 'GEOGCS[\"SWEREF99\",DATUM[\"SWEREF99\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6619\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4619\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4620, 1, 'Point 58', 'EPSG', 4620, 'GEOGCS[\"Point 58\",DATUM[\"Point 58\",SPHEROID[\"Clarke 1"
  "880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-106,-129,165,0,0,0,0],AUTHORITY[\"EPSG\",\"662"
  "0\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4620\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4621, 1, 'Fort Marigot', 'EPSG', 4621, 'GEOGCS[\"Fort Marigot\",DATUM[\"Fort Marigot\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[137,248,-430,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6621\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4621\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4622, 1, 'Guadeloupe 1948', 'EPSG', 4622, 'GEOGCS[\"Guadeloupe 1948\",DATUM[\"Guadeloupe 1948\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-472.29,-5.63,-304.12,0.4362,-0.83"
  "74,0.2563,1.8984],AUTHORITY[\"EPSG\",\"6622\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4622\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4623, 1, 'CSG67', 'EPSG', 4623, 'GEOGCS[\"CSG67\",DATUM[\"Centre Spatial Guyanais 1967\",SPHER"
  "OID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-186,230,110,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6623\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4623\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4624, 1, 'RGFG95', 'EPSG', 4624, 'GEOGCS[\"RGFG95\",DATUM[\"Reseau Geodesique Francais Guyane "
  "1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"6624\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4624\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4625, 1, 'Martinique 1938', 'EPSG', 4625, 'GEOGCS[\"Martinique 1938\",DATUM[\"Martinique 1938\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[126.93,547.94,130.41,-2.7867,5.161"
  "2,-0.8584,13.8227],AUTHORITY[\"EPSG\",\"6625\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4625\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4626, 1, 'Reunion 1947', 'EPSG', 4626, 'GEOGCS[\"Reunion 1947\",DATUM[\"Reunion 1947\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[94,-948,-1262,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6626\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4626\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4627, 1, 'RGR92', 'EPSG', 4627, 'GEOGCS[\"RGR92\",DATUM[\"Reseau Geodesique de la Reunion 1992"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6627\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4627\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4628, 1, 'Tahiti 52', 'EPSG', 4628, 'GEOGCS[\"Tahiti 52\",DATUM[\"Tahiti 52\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[162,117,154,0,0,0,0],AUTHORITY[\"EPSG\",\"6628\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4628\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4629, 1, 'Tahaa 54', 'EPSG', 4629, 'GEOGCS[\"Tahaa 54\",DATUM[\"Tahaa 54\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[72.51,345.411,79.241,-1.5862,-0.8826,-0.5495,1.3653],A"
  "UTHORITY[\"EPSG\",\"6629\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4629\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4630, 1, 'IGN72 Nuku Hiva', 'EPSG', 4630, 'GEOGCS[\"IGN72 Nuku Hiva\",DATUM[\"IGN72 Nuku Hiva\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[84,274,65,0,0,0,0],AUTHORITY[\"EPS"
  "G\",\"6630\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4630\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4632, 1, 'Combani 1950', 'EPSG', 4632, 'GEOGCS[\"Combani 1950\",DATUM[\"Combani 1950\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-382,-59,-262,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6632\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4632\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4633, 1, 'IGN56 Lifou', 'EPSG', 4633, 'GEOGCS[\"IGN56 Lifou\",DATUM[\"IGN56 Lifou\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[137.092,131.66,91.475,-1.9436,-11.5993,-4.3321"
  ",-7.4824],AUTHORITY[\"EPSG\",\"6633\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4633\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4636, 1, 'Petrels 1972', 'EPSG', 4636, 'GEOGCS[\"Petrels 1972\",DATUM[\"Petrels 1972\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[365,194,166,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6636\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4636\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4637, 1, 'Perroud 1950', 'EPSG', 4637, 'GEOGCS[\"Perroud 1950\",DATUM[\"Pointe Geologie Perrou"
  "d 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[325,154,172,0,0,0,0],AUTHO"
  "RITY[\"EPSG\",\"6637\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4637\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4638, 1, 'Saint Pierre et Miquelon 1950', 'EPSG', 4638, 'GEOGCS[\"Saint Pierre et Miquelon 195"
  "0\",DATUM[\"Saint Pierre et Miquelon 1950\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[11.363,424.148,373.13,0,0,0,0],AUTHORITY[\"EPSG\",\"6638\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4638\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4639, 1, 'MOP78', 'EPSG', 4639, 'GEOGCS[\"MOP78\",DATUM[\"MOP78\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[253,-132,-127,0,0,0,0],AUTHORITY[\"EPSG\",\"6639\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4639\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4641, 1, 'IGN53 Mare', 'EPSG', 4641, 'GEOGCS[\"IGN53 Mare\",DATUM[\"IGN53 Mare\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-408.809,366.856,-412.987,1.8842,-0.5308,2.1655,"
  "-121.0993],AUTHORITY[\"EPSG\",\"6641\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4641\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4642, 1, 'ST84 Ile des Pins', 'EPSG', 4642, 'GEOGCS[\"ST84 Ile des Pins\",DATUM[\"ST84 Ile des"
  " Pins\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-13,-348,292,0,0,0,0],AUTHO"
  "RITY[\"EPSG\",\"6642\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4642\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4643, 1, 'ST71 Belep', 'EPSG', 4643, 'GEOGCS[\"ST71 Belep\",DATUM[\"ST71 Belep\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-480.26,-438.32,-643.429,16.3119,20.1721,-4.0349"
  ",-111.7002],AUTHORITY[\"EPSG\",\"6643\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4643\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4644, 1, 'NEA74 Noumea', 'EPSG', 4644, 'GEOGCS[\"NEA74 Noumea\",DATUM[\"NEA74 Noumea\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-10.18,-350.43,291.37,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6644\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4644\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4646, 1, 'Grand Comoros', 'EPSG', 4646, 'GEOGCS[\"Grand Comoros\",DATUM[\"Grand Comoros\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-963,510,-359,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"6646\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4646\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4647, 1, 'ETRS89 / UTM zone 32N (zE-N)', 'EPSG', 4647, 'PROJCS[\"ETRS89 / UTM zone 32N (zE-N)\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"4647\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4652, 1, 'New Beijing / 3-degree Gauss-Kruger zone 25', 'EPSG', 4652, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 25\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",25500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4652\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4653, 1, 'New Beijing / 3-degree Gauss-Kruger zone 26', 'EPSG', 4653, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 26\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",26500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4653\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4654, 1, 'New Beijing / 3-degree Gauss-Kruger zone 27', 'EPSG', 4654, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 27\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",27500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4654\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4655, 1, 'New Beijing / 3-degree Gauss-Kruger zone 28', 'EPSG', 4655, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 28\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",28500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4655\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4656, 1, 'New Beijing / 3-degree Gauss-Kruger zone 29', 'EPSG', 4656, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 29\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",29500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4656\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4657, 1, 'Reykjavik 1900', 'EPSG', 4657, 'GEOGCS[\"Reykjavik 1900\",DATUM[\"Reykjavik 1900\",S"
  "PHEROID[\"Danish 1876\",6377019.27,300,AUTHORITY[\"EPSG\",\"7051\"]],TOWGS84[-28,199,5,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6657\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4657\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4658, 1, 'Hjorsey 1955', 'EPSG', 4658, 'GEOGCS[\"Hjorsey 1955\",DATUM[\"Hjorsey 1955\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-73,47,-83,0,0,0,0],AUTHORITY[\"EPSG\",\"6"
  "658\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4658\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4659, 1, 'ISN93', 'EPSG', 4659, 'GEOGCS[\"ISN93\",DATUM[\"Islands Net 1993\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6659\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4659\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4660, 1, 'Helle 1954', 'EPSG', 4660, 'GEOGCS[\"Helle 1954\",DATUM[\"Helle 1954\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[982.6087,552.753,-540.873,6.68162662527695,-31.6"
  "114924086423,-19.8481610048169,16.805],AUTHORITY[\"EPSG\",\"6660\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4660\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4661, 1, 'LKS92', 'EPSG', 4661, 'GEOGCS[\"LKS92\",DATUM[\"Latvia 1992\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6661\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4661\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4662, 1, 'IGN72 Grande Terre', 'EPSG', 4662, 'GEOGCS[\"IGN72 Grande Terre\",DATUM[\"IGN72 Gran"
  "de Terre\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[97.295,-263.247,310.882,"
  "-1.5999,0.8386,3.1409,13.3259],AUTHORITY[\"EPSG\",\"6634\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4662\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4663, 1, 'Porto Santo 1995', 'EPSG', 4663, 'GEOGCS[\"Porto Santo 1995\",DATUM[\"Porto Santo 19"
  "95\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-210.502,-66.902,-48.476,2.094"
  ",-15.067,-5.817,0.485],AUTHORITY[\"EPSG\",\"6663\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degr"
  "ee\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4"
  "663\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4664, 1, 'Azores Oriental 1995', 'EPSG', 4664, 'GEOGCS[\"Azores Oriental 1995\",DATUM[\"Azores"
  " Oriental Islands 1995\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-208.719,1"
  "29.685,52.092,-0.195,-0.014,0.327,0.198],AUTHORITY[\"EPSG\",\"6664\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4664\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4665, 1, 'Azores Central 1995', 'EPSG', 4665, 'GEOGCS[\"Azores Central 1995\",DATUM[\"Azores C"
  "entral Islands 1995\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-103.088,162."
  "481,-28.276,0.167,0.082,0.168,-1.504],AUTHORITY[\"EPSG\",\"6665\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4665\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4666, 1, 'Lisbon 1890', 'EPSG', 4666, 'GEOGCS[\"Lisbon 1890\",DATUM[\"Lisbon 1890\",SPHEROID[\""
  "Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[631.392,-66.551,481.442,1.09,-4.445,-4.48"
  "7,-4.43],AUTHORITY[\"EPSG\",\"6666\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4666\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4667, 1, 'IKBD-92', 'EPSG', 4667, 'GEOGCS[\"IKBD-92\",DATUM[\"Iraq-Kuwait Boundary Datum 1992\""
  ",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6667\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4667\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4668, 1, 'ED79', 'EPSG', 4668, 'GEOGCS[\"ED79\",DATUM[\"European Datum 1979\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-86,-98,-119,0,0,0,0],AUTHORITY[\"EPSG\",\"6668\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4668\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4669, 1, 'LKS94', 'EPSG', 4669, 'GEOGCS[\"LKS94\",DATUM[\"Lithuania 1994 (ETRS89)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6126\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4669\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4670, 1, 'IGM95', 'EPSG', 4670, 'GEOGCS[\"IGM95\",DATUM[\"Istituto Geografico Militaire 1995\""
  ",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6670\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4670\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4671, 1, 'Voirol 1879', 'EPSG', 4671, 'GEOGCS[\"Voirol 1879\",DATUM[\"Voirol 1879\",SPHEROID[\""
  "Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6671\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4671\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4672, 1, 'Chatham Islands 1971', 'EPSG', 4672, 'GEOGCS[\"Chatham Islands 1971\",DATUM[\"Chatha"
  "m Islands Datum 1971\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[175,-38,113,"
  "0,0,0,0],AUTHORITY[\"EPSG\",\"6672\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4672\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4673, 1, 'Chatham Islands 1979', 'EPSG', 4673, 'GEOGCS[\"Chatham Islands 1979\",DATUM[\"Chatha"
  "m Islands Datum 1979\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[174.05,-25.4"
  "9,112.57,0,0,0.554,0.2263],AUTHORITY[\"EPSG\",\"6673\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4673\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4674, 1, 'SIRGAS 2000', 'EPSG', 4674, 'GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Ge"
  "ocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84["
  "0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4675, 1, 'Guam 1963', 'EPSG', 4675, 'GEOGCS[\"Guam 1963\",DATUM[\"Guam 1963\",SPHEROID[\"Clark"
  "e 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-100,-248,259,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6675\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4675\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4676, 1, 'Vientiane 1982', 'EPSG', 4676, 'GEOGCS[\"Vientiane 1982\",DATUM[\"Vientiane 1982\",S"
  "PHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6676\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4676\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4677, 1, 'Lao 1993', 'EPSG', 4677, 'GEOGCS[\"Lao 1993\",DATUM[\"Lao 1993\",SPHEROID[\"Krassows"
  "ky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"6677\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4677\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4678, 1, 'Lao 1997', 'EPSG', 4678, 'GEOGCS[\"Lao 1997\",DATUM[\"Lao National Datum 1997\",SPHE"
  "ROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[44.585,-131.212,-39.544,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6678\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4678\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4679, 1, 'Jouik 1961', 'EPSG', 4679, 'GEOGCS[\"Jouik 1961\",DATUM[\"Jouik 1961\",SPHEROID[\"Cl"
  "arke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-80.01,253.26,291.19,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6679\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4679\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4680, 1, 'Nouakchott 1965', 'EPSG', 4680, 'GEOGCS[\"Nouakchott 1965\",DATUM[\"Nouakchott 1965\""
  ",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[124.5,-63.5,-281,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"6680\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4680\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4682, 1, 'Gulshan 303', 'EPSG', 4682, 'GEOGCS[\"Gulshan 303\",DATUM[\"Gulshan 303\",SPHEROID[\""
  "Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[283.7,735.9,261.1,0,0,0,0"
  "],AUTHORITY[\"EPSG\",\"6682\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4682\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4683, 1, 'PRS92', 'EPSG', 4683, 'GEOGCS[\"PRS92\",DATUM[\"Philippine Reference System 1992\",S"
  "PHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-127.62,-67.24,-47.04,-3.06"
  "8,4.903,1.578,-1.06],AUTHORITY[\"EPSG\",\"6683\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"468"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4684, 1, 'Gan 1970', 'EPSG', 4684, 'GEOGCS[\"Gan 1970\",DATUM[\"Gan 1970\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-133,-321,50,0,0,0,0],AUTHORITY[\"EPSG\",\"6684\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4684\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4686, 1, 'MAGNA-SIRGAS', 'EPSG', 4686, 'GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Naci"
  "onal de Referencia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4687, 1, 'RGPF', 'EPSG', 4687, 'GEOGCS[\"RGPF\",DATUM[\"Reseau Geodesique de la Polynesie Fran"
  "caise\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0.072,-0.507,-0.245,0.0183,"
  "-0.0003,0.007,-0.0093],AUTHORITY[\"EPSG\",\"6687\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degr"
  "ee\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4"
  "687\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4688, 1, 'Fatu Iva 72', 'EPSG', 4688, 'GEOGCS[\"Fatu Iva 72\",DATUM[\"Fatu Iva 72\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[347.175,1077.618,2623.677,33.9058,-70.6776,9.4"
  "013,186.0647],AUTHORITY[\"EPSG\",\"6688\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4688\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4689, 1, 'IGN63 Hiva Oa', 'EPSG', 4689, 'GEOGCS[\"IGN63 Hiva Oa\",DATUM[\"IGN63 Hiva Oa\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[374.787,-58.914,-1.202,-16.1928,-11.462"
  "9,-5.5287,-0.5502],AUTHORITY[\"EPSG\",\"6689\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4689\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4690, 1, 'Tahiti 79', 'EPSG', 4690, 'GEOGCS[\"Tahiti 79\",DATUM[\"Tahiti 79\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[221.597,152.441,176.523,2.403,1.3893,0.884,11.4648]"
  ",AUTHORITY[\"EPSG\",\"6690\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4690\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4691, 1, 'Moorea 87', 'EPSG', 4691, 'GEOGCS[\"Moorea 87\",DATUM[\"Moorea 87\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[218.769,150.75,176.75,3.5231,2.0037,1.288,10.9817],"
  "AUTHORITY[\"EPSG\",\"6691\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4691\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4692, 1, 'Maupiti 83', 'EPSG', 4692, 'GEOGCS[\"Maupiti 83\",DATUM[\"Maupiti 83\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[217.109,86.452,23.711,0.0183,-0.0003,0.007,-0.00"
  "93],AUTHORITY[\"EPSG\",\"6692\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4693, 1, 'Nakhl-e Ghanem', 'EPSG', 4693, 'GEOGCS[\"Nakhl-e Ghanem\",DATUM[\"Nakhl-e Ghanem\",S"
  "PHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,-0.15,0.68,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"6693\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4693\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4694, 1, 'POSGAR 94', 'EPSG', 4694, 'GEOGCS[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argent"
  "inas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["
  "\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4694\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4695, 1, 'Katanga 1955', 'EPSG', 4695, 'GEOGCS[\"Katanga 1955\",DATUM[\"Katanga 1955\",SPHEROI"
  "D[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-102.283,-10.277,-257.396,3.976,0"
  ".002,6.203,12.315],AUTHORITY[\"EPSG\",\"6695\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4695\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4696, 1, 'Kasai 1953', 'EPSG', 4696, 'GEOGCS[\"Kasai 1953\",DATUM[\"Kasai 1953\",SPHEROID[\"Cl"
  "arke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6696\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4696\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4697, 1, 'IGC 1962 6th Parallel South', 'EPSG', 4697, 'GEOGCS[\"IGC 1962 6th Parallel South\","
  "DATUM[\"IGC 1962 Arc of the 6th Parallel South\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\""
  ",\"7012\"]],AUTHORITY[\"EPSG\",\"6697\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4697\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4698, 1, 'IGN 1962 Kerguelen', 'EPSG', 4698, 'GEOGCS[\"IGN 1962 Kerguelen\",DATUM[\"IGN 1962 K"
  "erguelen\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[145,-187,103,0,0,0,0],AU"
  "THORITY[\"EPSG\",\"6698\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4698\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4699, 1, 'Le Pouce 1934', 'EPSG', 4699, 'GEOGCS[\"Le Pouce 1934\",DATUM[\"Le Pouce 1934\",SPHE"
  "ROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-770.1,158.4,-498.2,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6699\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4699\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4700, 1, 'IGN Astro 1960', 'EPSG', 4700, 'GEOGCS[\"IGN Astro 1960\",DATUM[\"IGN Astro 1960\",S"
  "PHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\",\"6700\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4701, 1, 'IGCB 1955', 'EPSG', 4701, 'GEOGCS[\"IGCB 1955\",DATUM[\"Institut Geographique du Con"
  "go Belge 1955\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79.9,-158,-"
  "168.9,0,0,0,0],AUTHORITY[\"EPSG\",\"6701\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4701\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4702, 1, 'Mauritania 1999', 'EPSG', 4702, 'GEOGCS[\"Mauritania 1999\",DATUM[\"Mauritania 1999\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6702\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4702\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4703, 1, 'Mhast 1951', 'EPSG', 4703, 'GEOGCS[\"Mhast 1951\",DATUM[\"Missao Hidrografico Angola"
  " y Sao Tome 1951\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],AUTHORITY[\"EPSG\""
  ",\"6703\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4703\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4704, 1, 'Mhast (onshore)', 'EPSG', 4704, 'GEOGCS[\"Mhast (onshore)\",DATUM[\"Mhast (onshore)\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6704\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4704\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4705, 1, 'Mhast (offshore)', 'EPSG', 4705, 'GEOGCS[\"Mhast (offshore)\",DATUM[\"Mhast (offshor"
  "e)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6705\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4705\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4706, 1, 'Egypt Gulf of Suez S-650 TL', 'EPSG', 4706, 'GEOGCS[\"Egypt Gulf of Suez S-650 TL\","
  "DATUM[\"Egypt Gulf of Suez S-650 TL\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-"
  "146.21,112.63,4.05,0,0,0,0],AUTHORITY[\"EPSG\",\"6706\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4706\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4707, 1, 'Tern Island 1961', 'EPSG', 4707, 'GEOGCS[\"Tern Island 1961\",DATUM[\"Tern Island 19"
  "61\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[114,-116,-333,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6707\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4707\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4708, 1, 'Cocos Islands 1965', 'EPSG', 4708, 'GEOGCS[\"Cocos Islands 1965\",DATUM[\"Cocos Isla"
  "nds 1965\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG\",\"7003\"]],TOWGS84[-491,-22,43"
  "5,0,0,0,0],AUTHORITY[\"EPSG\",\"6708\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4708\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4709, 1, 'Iwo Jima 1945', 'EPSG', 4709, 'GEOGCS[\"Iwo Jima 1945\",DATUM[\"Iwo Jima 1945\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[145,75,-272,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6709\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4709\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4710, 1, 'Astro DOS 71', 'EPSG', 4710, 'GEOGCS[\"Astro DOS 71\",DATUM[\"Astro DOS 71\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-323.65,551.39,-491.22,0,0,0,0],AUTHORITY["
  "\"EPSG\",\"6710\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4710\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4711, 1, 'Marcus Island 1952', 'EPSG', 4711, 'GEOGCS[\"Marcus Island 1952\",DATUM[\"Marcus Isl"
  "and 1952\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[124,-234,-25,0,0,0,0],AU"
  "THORITY[\"EPSG\",\"6711\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4711\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4712, 1, 'Ascension Island 1958', 'EPSG', 4712, 'GEOGCS[\"Ascension Island 1958\",DATUM[\"Asce"
  "nsion Island 1958\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-205,107,53,0,0"
  ",0,0],AUTHORITY[\"EPSG\",\"6712\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4712\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4713, 1, 'Ayabelle Lighthouse', 'EPSG', 4713, 'GEOGCS[\"Ayabelle Lighthouse\",DATUM[\"Ayabelle"
  " Lighthouse\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-79,-129,145,0"
  ",0,0,0],AUTHORITY[\"EPSG\",\"6713\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4713\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4714, 1, 'Bellevue', 'EPSG', 4714, 'GEOGCS[\"Bellevue\",DATUM[\"Bellevue\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-127,-769,472,0,0,0,0],AUTHORITY[\"EPSG\",\"6714\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4714\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4715, 1, 'Camp Area Astro', 'EPSG', 4715, 'GEOGCS[\"Camp Area Astro\",DATUM[\"Camp Area Astro\""
  ",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-104,-129,239,0,0,0,0],AUTHORITY[\""
  "EPSG\",\"6715\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4715\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4716, 1, 'Phoenix Islands 1966', 'EPSG', 4716, 'GEOGCS[\"Phoenix Islands 1966\",DATUM[\"Phoeni"
  "x Islands 1966\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[298,-304,-375,0,0,"
  "0,0],AUTHORITY[\"EPSG\",\"6716\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4716\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4717, 1, 'Cape Canaveral', 'EPSG', 4717, 'GEOGCS[\"Cape Canaveral\",DATUM[\"Cape Canaveral\",S"
  "PHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-2,151,181,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"6717\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4717\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4718, 1, 'Solomon 1968', 'EPSG', 4718, 'GEOGCS[\"Solomon 1968\",DATUM[\"Solomon 1968\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[252,-209,-751,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6718\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4718\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4719, 1, 'Easter Island 1967', 'EPSG', 4719, 'GEOGCS[\"Easter Island 1967\",DATUM[\"Easter Isl"
  "and 1967\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[211,147,111,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6719\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4719\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4720, 1, 'Fiji 1986', 'EPSG', 4720, 'GEOGCS[\"Fiji 1986\",DATUM[\"Fiji Geodetic Datum 1986\",S"
  "PHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0,0,4.5,0,0,0.554,0.2263],AUTHORITY[\"EPSG\","
  "\"6720\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4720\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4721, 1, 'Fiji 1956', 'EPSG', 4721, 'GEOGCS[\"Fiji 1956\",DATUM[\"Fiji 1956\",SPHEROID[\"Inter"
  "national 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[265.025,384.929,-194.046,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6721\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4721\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4722, 1, 'South Georgia 1968', 'EPSG', 4722, 'GEOGCS[\"South Georgia 1968\",DATUM[\"South Geor"
  "gia 1968\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-794,119,-298,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"6722\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4722\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4723, 1, 'GCGD59', 'EPSG', 4723, 'GEOGCS[\"GCGD59\",DATUM[\"Grand Cayman Geodetic Datum 1959\""
  ",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-179.483,-69.379,-27.584,"
  "-7.862,8.163,6.042,-13.925],AUTHORITY[\"EPSG\",\"6723\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4723\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4724, 1, 'Diego Garcia 1969', 'EPSG', 4724, 'GEOGCS[\"Diego Garcia 1969\",DATUM[\"Diego Garcia"
  " 1969\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[208,-435,-229,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"6724\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4724\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4725, 1, 'Johnston Island 1961', 'EPSG', 4725, 'GEOGCS[\"Johnston Island 1961\",DATUM[\"Johnst"
  "on Island 1961\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[189,-79,-202,0,0,0"
  ",0],AUTHORITY[\"EPSG\",\"6725\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4725\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4726, 1, 'SIGD61', 'EPSG', 4726, 'GEOGCS[\"SIGD61\",DATUM[\"Sister Islands Geodetic Datum 1961"
  "\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[8.853,-52.644,180.304,-"
  "0.393,-2.323,2.96,-24.081],AUTHORITY[\"EPSG\",\"6726\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4726\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4727, 1, 'Midway 1961', 'EPSG', 4727, 'GEOGCS[\"Midway 1961\",DATUM[\"Midway 1961\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[403,-81,277,0,0,0,0],AUTHORITY[\"EPSG\",\"6727"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4727\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4728, 1, 'Pico de las Nieves 1984', 'EPSG', 4728, 'GEOGCS[\"Pico de las Nieves 1984\",DATUM[\""
  "Pico de las Nieves 1984\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-307,-92,"
  "127,0,0,0,0],AUTHORITY[\"EPSG\",\"6728\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4728\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4729, 1, 'Pitcairn 1967', 'EPSG', 4729, 'GEOGCS[\"Pitcairn 1967\",DATUM[\"Pitcairn 1967\",SPHE"
  "ROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[185,165,42,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6729\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4729\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4730, 1, 'Santo 1965', 'EPSG', 4730, 'GEOGCS[\"Santo 1965\",DATUM[\"Santo 1965\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[170,42,84,0,0,0,0],AUTHORITY[\"EPSG\",\"6730\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4730\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4732, 1, 'Marshall Islands 1960', 'EPSG', 4732, 'GEOGCS[\"Marshall Islands 1960\",DATUM[\"Mars"
  "hall Islands 1960\",SPHEROID[\"Hough 1960\",6378270,297,AUTHORITY[\"EPSG\",\"7053\"]],TOWGS84[102,52,-38,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6732\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4732\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4733, 1, 'Wake Island 1952', 'EPSG', 4733, 'GEOGCS[\"Wake Island 1952\",DATUM[\"Wake Island 19"
  "52\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[276,-57,149,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6733\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4733\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4734, 1, 'Tristan 1968', 'EPSG', 4734, 'GEOGCS[\"Tristan 1968\",DATUM[\"Tristan 1968\",SPHEROI"
  "D[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-632,438,-609,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6734\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4734\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4735, 1, 'Kusaie 1951', 'EPSG', 4735, 'GEOGCS[\"Kusaie 1951\",DATUM[\"Kusaie 1951\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[647,1777,-1124,0,0,0,0],AUTHORITY[\"EPSG\",\"6"
  "735\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4735\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4736, 1, 'Deception Island', 'EPSG', 4736, 'GEOGCS[\"Deception Island\",DATUM[\"Deception Isla"
  "nd\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[260,12,-147,0,0,0,0],AU"
  "THORITY[\"EPSG\",\"6736\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4736\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4737, 1, 'Korea 2000', 'EPSG', 4737, 'GEOGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4737\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4738, 1, 'Hong Kong 1963', 'EPSG', 4738, 'GEOGCS[\"Hong Kong 1963\",DATUM[\"Hong Kong 1963\",S"
  "PHEROID[\"Clarke 1858\",6378293.645208759,294.26067636926103,AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"6738\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4739, 1, 'Hong Kong 1963(67)', 'EPSG', 4739, 'GEOGCS[\"Hong Kong 1963(67)\",DATUM[\"Hong Kong "
  "1963(67)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-156,-271,-189,0,0,0,0],"
  "AUTHORITY[\"EPSG\",\"6739\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4739\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4740, 1, 'PZ-90', 'EPSG', 4740, 'GEOGCS[\"PZ-90\",DATUM[\"Parametry Zemli 1990\",SPHEROID[\"PZ"
  "-90\",6378136,298.257839303,AUTHORITY[\"EPSG\",\"7054\"]],TOWGS84[0,0,1.5,0,0,0.076,0],AUTHORITY[\"EPSG\",\"6740\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4740\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4741, 1, 'FD54', 'EPSG', 4741, 'GEOGCS[\"FD54\",DATUM[\"Faroe Datum 1954\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6741\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4741\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4742, 1, 'GDM2000', 'EPSG', 4742, 'GEOGCS[\"GDM2000\",DATUM[\"Geodetic Datum of Malaysia 2000\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6742\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4742\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4743, 1, 'Karbala 1979', 'EPSG', 4743, 'GEOGCS[\"Karbala 1979\",DATUM[\"Karbala 1979\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[70.995,-335.916,262.898,0,0,0,0],AU"
  "THORITY[\"EPSG\",\"6743\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4743\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4744, 1, 'Nahrwan 1934', 'EPSG', 4744, 'GEOGCS[\"Nahrwan 1934\",DATUM[\"Nahrwan 1934\",SPHEROI"
  "D[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-242.2,-144.9,370.3,0,0,0,0],AUTHOR"
  "ITY[\"EPSG\",\"6744\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4744\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4745, 1, 'RD/83', 'EPSG', 4745, 'GEOGCS[\"RD/83\",DATUM[\"Rauenberg Datum/83\",SPHEROID[\"Bess"
  "el 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4745\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4746, 1, 'PD/83', 'EPSG', 4746, 'GEOGCS[\"PD/83\",DATUM[\"Potsdam Datum/83\",SPHEROID[\"Bessel"
  " 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4746\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4747, 1, 'GR96', 'EPSG', 4747, 'GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4747\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4748, 1, 'Vanua Levu 1915', 'EPSG', 4748, 'GEOGCS[\"Vanua Levu 1915\",DATUM[\"Vanua Levu 1915\""
  ",SPHEROID[\"Clarke 1880 (international foot)\",6378306.3696,293.46630765562986,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84["
  "51,391,-36,0,0,0,0],AUTHORITY[\"EPSG\",\"6748\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4748\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4749, 1, 'RGNC91-93', 'EPSG', 4749, 'GEOGCS[\"RGNC91-93\",DATUM[\"Reseau Geodesique de Nouvell"
  "e Caledonie 91-93\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],"
  "AUTHORITY[\"EPSG\",\"6749\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4749\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4750, 1, 'ST87 Ouvea', 'EPSG', 4750, 'GEOGCS[\"ST87 Ouvea\",DATUM[\"ST87 Ouvea\",SPHEROID[\"WG"
  "S 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-56.263,16.136,-22.856,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6750\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4750\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4751, 1, 'Kertau (RSO)', 'EPSG', 4751, 'GEOGCS[\"Kertau (RSO)\",DATUM[\"Kertau (RSO)\",SPHEROI"
  "D[\"Everest 1830 (RSO 1969)\",6377295.664,300.8017,AUTHORITY[\"EPSG\",\"7056\"]],AUTHORITY[\"EPSG\",\"6751\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4751\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4752, 1, 'Viti Levu 1912', 'EPSG', 4752, 'GEOGCS[\"Viti Levu 1912\",DATUM[\"Viti Levu 1912\",S"
  "PHEROID[\"Clarke 1880 (international foot)\",6378306.3696,293.46630765562986,AUTHORITY[\"EPSG\",\"7055\"]],TOWGS84[98"
  ",390,-22,0,0,0,0],AUTHORITY[\"EPSG\",\"6752\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4752\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4753, 1, 'fk89', 'EPSG', 4753, 'GEOGCS[\"fk89\",DATUM[\"fk89\",SPHEROID[\"International 1924\""
  ",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6753\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4753\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4754, 1, 'LGD2006', 'EPSG', 4754, 'GEOGCS[\"LGD2006\",DATUM[\"Libyan Geodetic Datum 2006\",SPH"
  "EROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-208.4058,-109.8777,-2.5764,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"6754\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4754\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4755, 1, 'DGN95', 'EPSG', 4755, 'GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHERO"
  "ID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4755\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4756, 1, 'VN-2000', 'EPSG', 4756, 'GEOGCS[\"VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84"
  "\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-191.90441429,-39.30318279,-111.45032835,0.00928836,-0"
  ".01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4756\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4757, 1, 'SVY21', 'EPSG', 4757, 'GEOGCS[\"SVY21\",DATUM[\"SVY21\",SPHEROID[\"WGS 84\",6378137,"
  "298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6757\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4757\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4758, 1, 'JAD2001', 'EPSG', 4758, 'GEOGCS[\"JAD2001\",DATUM[\"Jamaica 2001\",SPHEROID[\"WGS 84"
  "\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6758\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4758\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4759, 1, 'NAD83(NSRS2007)', 'EPSG', 4759, 'GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National "
  "Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,"
  "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4760, 1, 'WGS 66', 'EPSG', 4760, 'GEOGCS[\"WGS 66\",DATUM[\"World Geodetic System 1966\",SPHER"
  "OID[\"NWL 9D\",6378145,298.25,AUTHORITY[\"EPSG\",\"7025\"]],AUTHORITY[\"EPSG\",\"6760\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4760\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4761, 1, 'HTRS96', 'EPSG', 4761, 'GEOGCS[\"HTRS96\",DATUM[\"Croatian Terrestrial Reference Sys"
  "tem\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6761\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4761\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4762, 1, 'BDA2000', 'EPSG', 4762, 'GEOGCS[\"BDA2000\",DATUM[\"Bermuda 2000\",SPHEROID[\"WGS 84"
  "\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6762\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4762\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4763, 1, 'Pitcairn 2006', 'EPSG', 4763, 'GEOGCS[\"Pitcairn 2006\",DATUM[\"Pitcairn 2006\",SPHE"
  "ROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6763\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4763\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4764, 1, 'RSRGD2000', 'EPSG', 4764, 'GEOGCS[\"RSRGD2000\",DATUM[\"Ross Sea Region Geodetic Dat"
  "um 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY["
  "\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4764\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4765, 1, 'Slovenia 1996', 'EPSG', 4765, 'GEOGCS[\"Slovenia 1996\",DATUM[\"Slovenia Geodetic Da"
  "tum 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY"
  "[\"EPSG\",\"6765\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4765\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4766, 1, 'New Beijing / 3-degree Gauss-Kruger zone 30', 'EPSG', 4766, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 30\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4766\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4767, 1, 'New Beijing / 3-degree Gauss-Kruger zone 31', 'EPSG', 4767, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 31\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",31500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4767\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4768, 1, 'New Beijing / 3-degree Gauss-Kruger zone 32', 'EPSG', 4768, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 32\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",32500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4768\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4769, 1, 'New Beijing / 3-degree Gauss-Kruger zone 33', 'EPSG', 4769, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 33\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",33500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4769\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4770, 1, 'New Beijing / 3-degree Gauss-Kruger zone 34', 'EPSG', 4770, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 34\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",34500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4770\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4771, 1, 'New Beijing / 3-degree Gauss-Kruger zone 35', 'EPSG', 4771, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 35\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",35500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4771\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4772, 1, 'New Beijing / 3-degree Gauss-Kruger zone 36', 'EPSG', 4772, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 36\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",36500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4772\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4773, 1, 'New Beijing / 3-degree Gauss-Kruger zone 37', 'EPSG', 4773, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 37\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",37500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4773\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4774, 1, 'New Beijing / 3-degree Gauss-Kruger zone 38', 'EPSG', 4774, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 38\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",38500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4774\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4775, 1, 'New Beijing / 3-degree Gauss-Kruger zone 39', 'EPSG', 4775, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 39\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",39500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4775\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4776, 1, 'New Beijing / 3-degree Gauss-Kruger zone 40', 'EPSG', 4776, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 40\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4776\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4777, 1, 'New Beijing / 3-degree Gauss-Kruger zone 41', 'EPSG', 4777, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 41\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",41500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4777\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4778, 1, 'New Beijing / 3-degree Gauss-Kruger zone 42', 'EPSG', 4778, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 42\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",42500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4778\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4779, 1, 'New Beijing / 3-degree Gauss-Kruger zone 43', 'EPSG', 4779, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 43\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",43500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4779\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4780, 1, 'New Beijing / 3-degree Gauss-Kruger zone 44', 'EPSG', 4780, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 44\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",44500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4780\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4781, 1, 'New Beijing / 3-degree Gauss-Kruger zone 45', 'EPSG', 4781, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger zone 45\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",45500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4781\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4782, 1, 'New Beijing / 3-degree Gauss-Kruger CM 75E', 'EPSG', 4782, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 75E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4782\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4783, 1, 'New Beijing / 3-degree Gauss-Kruger CM 78E', 'EPSG', 4783, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 78E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",78,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4783\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4784, 1, 'New Beijing / 3-degree Gauss-Kruger CM 81E', 'EPSG', 4784, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 81E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4784\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4785, 1, 'New Beijing / 3-degree Gauss-Kruger CM 84E', 'EPSG', 4785, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 84E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",84,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4785\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4786, 1, 'New Beijing / 3-degree Gauss-Kruger CM 87E', 'EPSG', 4786, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 87E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4786\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4787, 1, 'New Beijing / 3-degree Gauss-Kruger CM 90E', 'EPSG', 4787, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 90E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4787\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4788, 1, 'New Beijing / 3-degree Gauss-Kruger CM 93E', 'EPSG', 4788, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 93E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4788\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4789, 1, 'New Beijing / 3-degree Gauss-Kruger CM 96E', 'EPSG', 4789, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 96E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4789\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4790, 1, 'New Beijing / 3-degree Gauss-Kruger CM 99E', 'EPSG', 4790, 'PROJCS[\"New Beijing / 3"
  "-degree Gauss-Kruger CM 99E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298.3"
  ",AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4790\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4791, 1, 'New Beijing / 3-degree Gauss-Kruger CM 102E', 'EPSG', 4791, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 102E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4792, 1, 'New Beijing / 3-degree Gauss-Kruger CM 105E', 'EPSG', 4792, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 105E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4792\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4793, 1, 'New Beijing / 3-degree Gauss-Kruger CM 108E', 'EPSG', 4793, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 108E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4793\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4794, 1, 'New Beijing / 3-degree Gauss-Kruger CM 111E', 'EPSG', 4794, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 111E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4794\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4795, 1, 'New Beijing / 3-degree Gauss-Kruger CM 114E', 'EPSG', 4795, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 114E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4795\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4796, 1, 'New Beijing / 3-degree Gauss-Kruger CM 117E', 'EPSG', 4796, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 117E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4796\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4797, 1, 'New Beijing / 3-degree Gauss-Kruger CM 120E', 'EPSG', 4797, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 120E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",120,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4797\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4798, 1, 'New Beijing / 3-degree Gauss-Kruger CM 123E', 'EPSG', 4798, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 123E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4798\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4799, 1, 'New Beijing / 3-degree Gauss-Kruger CM 126E', 'EPSG', 4799, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 126E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",126,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4799\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4800, 1, 'New Beijing / 3-degree Gauss-Kruger CM 129E', 'EPSG', 4800, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 129E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4800\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4801, 1, 'Bern 1898 (Bern)', 'EPSG', 4801, 'GEOGCS[\"Bern 1898 (Bern)\",DATUM[\"CH1903 (Bern)\""
  ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6801\"]],PRIMEM["
  "\"Bern\",7.439583333333334,AUTHORITY[\"EPSG\",\"8907\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"91"
  "22\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4801\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4802, 1, 'Bogota 1975 (Bogota)', 'EPSG', 4802, 'GEOGCS[\"Bogota 1975 (Bogota)\",DATUM[\"Bogota"
  " 1975 (Bogota)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6802\""
  "]],PRIMEM[\"Bogota\",-74.08091666666667,AUTHORITY[\"EPSG\",\"8904\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4802\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4803, 1, 'Lisbon (Lisbon)', 'EPSG', 4803, 'GEOGCS[\"Lisbon (Lisbon)\",DATUM[\"Lisbon 1937 (Lis"
  "bon)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6803\"]],PRIMEM"
  "[\"Lisbon\",-9.131906111111112,AUTHORITY[\"EPSG\",\"8902\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4803\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4804, 1, 'Makassar (Jakarta)', 'EPSG', 4804, 'GEOGCS[\"Makassar (Jakarta)\",DATUM[\"Makassar ("
  "Jakarta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6804\""
  "]],PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4804\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4805, 1, 'MGI (Ferro)', 'EPSG', 4805, 'GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische In"
  "stitut (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\""
  "6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4806, 1, 'Monte Mario (Rome)', 'EPSG', 4806, 'GEOGCS[\"Monte Mario (Rome)\",DATUM[\"Monte Mari"
  "o (Rome)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6806\"]],PR"
  "IMEM[\"Rome\",12.452333333333332,AUTHORITY[\"EPSG\",\"8906\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4806\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4807, 1, 'NTF (Paris)', 'EPSG', 4807, 'GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation F"
  "rancaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY"
  "[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHO"
  "RITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4807\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4808, 1, 'Padang (Jakarta)', 'EPSG', 4808, 'GEOGCS[\"Padang (Jakarta)\",DATUM[\"Padang 1884 (J"
  "akarta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6808\"]"
  "],PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4808\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4809, 1, 'Belge 1950 (Brussels)', 'EPSG', 4809, 'GEOGCS[\"Belge 1950 (Brussels)\",DATUM[\"Rese"
  "au National Belge 1950 (Brussels)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORI"
  "TY[\"EPSG\",\"6809\"]],PRIMEM[\"Brussels\",4.3679749999999995,AUTHORITY[\"EPSG\",\"8910\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4809\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4810, 1, 'Tananarive (Paris)', 'EPSG', 4810, 'GEOGCS[\"Tananarive (Paris)\",DATUM[\"Tananarive"
  " 1925 (Paris)\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6810\""
  "]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"91"
  "05\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4810\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4811, 1, 'Voirol 1875 (Paris)', 'EPSG', 4811, 'GEOGCS[\"Voirol 1875 (Paris)\",DATUM[\"Voirol 1"
  "875 (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EP"
  "SG\",\"6811\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY["
  "\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4811\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4812, 1, 'New Beijing / 3-degree Gauss-Kruger CM 132E', 'EPSG', 4812, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 132E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4812\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4813, 1, 'Batavia (Jakarta)', 'EPSG', 4813, 'GEOGCS[\"Batavia (Jakarta)\",DATUM[\"Batavia (Jak"
  "arta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6813\"]],"
  "PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4813\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4814, 1, 'RT38 (Stockholm)', 'EPSG', 4814, 'GEOGCS[\"RT38 (Stockholm)\",DATUM[\"Stockholm 1938"
  " (Stockholm)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"68"
  "14\"]],PRIMEM[\"Stockholm\",18.058277777777775,AUTHORITY[\"EPSG\",\"8911\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4814\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4815, 1, 'Greek (Athens)', 'EPSG', 4815, 'GEOGCS[\"Greek (Athens)\",DATUM[\"Greek (Athens)\",S"
  "PHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6815\"]],PRIMEM[\""
  "Athens\",23.716337499999998,AUTHORITY[\"EPSG\",\"8912\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4815\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4816, 1, 'Carthage (Paris)', 'EPSG', 4816, 'GEOGCS[\"Carthage (Paris)\",DATUM[\"Carthage (Pari"
  "s)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"68"
  "16\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\","
  "\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4816\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4817, 1, 'NGO 1948 (Oslo)', 'EPSG', 4817, 'GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\""
  ",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHORITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRI"
  "MEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4817\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4818, 1, 'S-JTSK (Ferro)', 'EPSG', 4818, 'GEOGCS[\"S-JTSK (Ferro)\",DATUM[\"System of the Unif"
  "ied Trigonometrical Cadastral Network (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\""
  "7004\"]],AUTHORITY[\"EPSG\",\"6818\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degr"
  "ee\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4"
  "818\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4820, 1, 'Segara (Jakarta)', 'EPSG', 4820, 'GEOGCS[\"Segara (Jakarta)\",DATUM[\"Gunung Segara "
  "(Jakarta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6820\""
  "]],PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4820\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4821, 1, 'Voirol 1879 (Paris)', 'EPSG', 4821, 'GEOGCS[\"Voirol 1879 (Paris)\",DATUM[\"Voirol 1"
  "879 (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EP"
  "SG\",\"6821\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY["
  "\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4821\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4822, 1, 'New Beijing / 3-degree Gauss-Kruger CM 135E', 'EPSG', 4822, 'PROJCS[\"New Beijing / "
  "3-degree Gauss-Kruger CM 135E\",GEOGCS[\"New Beijing\",DATUM[\"New Beijing\",SPHEROID[\"Krassowsky 1940\",6378245,298"
  ".3,AUTHORITY[\"EPSG\",\"7024\"]],AUTHORITY[\"EPSG\",\"1045\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4555\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"4822\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4823, 1, 'Sao Tome', 'EPSG', 4823, 'GEOGCS[\"Sao Tome\",DATUM[\"Sao Tome\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1044\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4823\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4824, 1, 'Principe', 'EPSG', 4824, 'GEOGCS[\"Principe\",DATUM[\"Principe\",SPHEROID[\"Internat"
  "ional 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1046\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4824\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4826, 1, 'WGS 84 / Cape Verde National', 'EPSG', 4826, 'PROJCS[\"WGS 84 / Cape Verde National\""
  ",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\","
  "\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJ"
  "ECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",15.8333"
  "333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-24,AUTHORITY[\"EPSG\",\"8822\"]],PAR"
  "AMETER[\"Latitude of 1st standard parallel\",15,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard pa"
  "rallel\",16.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",161587.83,AUTHORITY[\"E"
  "PSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",128511.202,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"M\",EAST],AXIS[\"P\",NORTH],AUTHORITY[\"EPSG\",\"4826\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4839, 1, 'ETRS89 / LCC Germany (N-E)', 'EPSG', 4839, 'PROJCS[\"ETRS89 / LCC Germany (N-E)\",GE"
  "OGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",51,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10.5,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",53.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"4839\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4901, 1, 'ATF (Paris)', 'EPSG', 4901, 'GEOGCS[\"ATF (Paris)\",DATUM[\"Ancienne Triangulation F"
  "rancaise (Paris)\",SPHEROID[\"Plessis 1817\",6376523,308.64,AUTHORITY[\"EPSG\",\"7027\"]],AUTHORITY[\"EPSG\",\"6901\""
  "]],PRIMEM[\"Paris RGS\",2.5968981481481492,AUTHORITY[\"EPSG\",\"8914\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY["
  "\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4901\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4903, 1, 'Madrid 1870 (Madrid)', 'EPSG', 4903, 'GEOGCS[\"Madrid 1870 (Madrid)\",DATUM[\"Madrid"
  " 1870 (Madrid)\",SPHEROID[\"Struve 1860\",6378298.3,294.73,AUTHORITY[\"EPSG\",\"7028\"]],AUTHORITY[\"EPSG\",\"6903\"]"
  "],PRIMEM[\"Madrid\",-3.6879388888888895,AUTHORITY[\"EPSG\",\"8905\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4903\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (4904, 1, 'Lisbon 1890 (Lisbon)', 'EPSG', 4904, 'GEOGCS[\"Lisbon 1890 (Lisbon)\",DATUM[\"Lisbon"
  " 1890 (Lisbon)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\""
  "6904\"]],PRIMEM[\"Lisbon\",-9.131906111111112,AUTHORITY[\"EPSG\",\"8902\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4904\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5013, 1, 'PTRA08', 'EPSG', 5013, 'GEOGCS[\"PTRA08\",DATUM[\"Autonomous Regions of Portugal 200"
  "8\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"1041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5013\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5014, 1, 'PTRA08 / UTM zone 25N', 'EPSG', 5014, 'PROJCS[\"PTRA08 / UTM zone 25N\",GEOGCS[\"PTR"
  "A08\",DATUM[\"Autonomous Regions of Portugal 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5013\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5015, 1, 'PTRA08 / UTM zone 26N', 'EPSG', 5015, 'PROJCS[\"PTRA08 / UTM zone 26N\",GEOGCS[\"PTR"
  "A08\",DATUM[\"Autonomous Regions of Portugal 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5013\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5015\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5016, 1, 'PTRA08 / UTM zone 28N', 'EPSG', 5016, 'PROJCS[\"PTRA08 / UTM zone 28N\",GEOGCS[\"PTR"
  "A08\",DATUM[\"Autonomous Regions of Portugal 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5013\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5016\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5017, 1, 'Lisbon 1890 / Portugal Bonne New', 'EPSG', 5017, 'PROJCS[\"Lisbon 1890 / Portugal Bo"
  "nne New\",GEOGCS[\"Lisbon 1890\",DATUM[\"Lisbon 1890\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[631.392,-66.551,481.442,1.09,-4.445,-4.487,-4.43],AUTHORITY[\"EPSG\",\"6666\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4666\"]],PROJECTION[\"Bonne (South Orientated)\",AUTHORITY[\"EPSG\""
  ",\"9828\"]],PARAMETER[\"Latitude of natural origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-8.13190611111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"P\",SOUTH],AXIS[\"M\",WEST],AUTHORITY[\"EPSG\",\"5017\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5018, 1, 'Lisbon / Portuguese Grid New', 'EPSG', 5018, 'PROJCS[\"Lisbon / Portuguese Grid New\""
  ",GEOGCS[\"Lisbon\",DATUM[\"Lisbon 1937\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TO"
  "WGS84[-288.885,-91.744,126.244,-1.691,0.41,-0.211,-4.598],AUTHORITY[\"EPSG\",\"6207\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4207\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-8.13190611111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "5018\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5041, 1, 'WGS 84 / UPS North (E,N)', 'EPSG', 5041, 'PROJCS[\"WGS 84 / UPS North (E,N)\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitude of natural origin\",90,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fact"
  "or at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2000000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",SOUTH],AXIS[\"N\",SOUTH],AUTHORITY[\"EPSG\",\"5041\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5042, 1, 'WGS 84 / UPS South (E,N)', 'EPSG', 5042, 'PROJCS[\"WGS 84 / UPS South (E,N)\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitude of natural origin\",-90,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2000000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",NORTH],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5042\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5048, 1, 'ETRS89 / TM35FIN(N,E)', 'EPSG', 5048, 'PROJCS[\"ETRS89 / TM35FIN(N,E)\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5048\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5069, 1, 'NAD27 / Conus Albers', 'EPSG', 5069, 'PROJCS[\"NAD27 / Conus Albers\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Albers Equal Area\",AUTHORI"
  "TY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",23,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude "
  "of false origin\",-96,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",29.5,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"5069\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5070, 1, 'NAD83 / Conus Albers', 'EPSG', 5070, 'PROJCS[\"NAD83 / Conus Albers\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",23,AUTHORITY"
  "[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-96,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",29.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.51111"
  "11111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETE"
  "R[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5070\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5071, 1, 'NAD83(HARN) / Conus Albers', 'EPSG', 5071, 'PROJCS[\"NAD83(HARN) / Conus Albers\",GE"
  "OGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitu"
  "de of false origin\",23,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-96,AUTHORITY[\"EPSG\","
  "\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",29.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of"
  " 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5071\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5072, 1, 'NAD83(NSRS2007) / Conus Albers', 'EPSG', 5072, 'PROJCS[\"NAD83(NSRS2007) / Conus Alb"
  "ers\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]]"
  ",PARAMETER[\"Latitude of false origin\",23,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-96,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",29.5,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5072\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5105, 1, 'ETRS89 / NTM zone 5', 'EPSG', 5105, 'PROJCS[\"ETRS89 / NTM zone 5\",GEOGCS[\"ETRS89\""
  ",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",5.51111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5105\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5106, 1, 'ETRS89 / NTM zone 6', 'EPSG', 5106, 'PROJCS[\"ETRS89 / NTM zone 6\",GEOGCS[\"ETRS89\""
  ",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",6.51111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5106\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5107, 1, 'ETRS89 / NTM zone 7', 'EPSG', 5107, 'PROJCS[\"ETRS89 / NTM zone 7\",GEOGCS[\"ETRS89\""
  ",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",7.51111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5107\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5108, 1, 'ETRS89 / NTM zone 8', 'EPSG', 5108, 'PROJCS[\"ETRS89 / NTM zone 8\",GEOGCS[\"ETRS89\""
  ",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",8.5,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5108\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5109, 1, 'ETRS89 / NTM zone 9', 'EPSG', 5109, 'PROJCS[\"ETRS89 / NTM zone 9\",GEOGCS[\"ETRS89\""
  ",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9.5,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5109\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5110, 1, 'ETRS89 / NTM zone 10', 'EPSG', 5110, 'PROJCS[\"ETRS89 / NTM zone 10\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5110\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5111, 1, 'ETRS89 / NTM zone 11', 'EPSG', 5111, 'PROJCS[\"ETRS89 / NTM zone 11\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5111\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5112, 1, 'ETRS89 / NTM zone 12', 'EPSG', 5112, 'PROJCS[\"ETRS89 / NTM zone 12\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5112\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5113, 1, 'ETRS89 / NTM zone 13', 'EPSG', 5113, 'PROJCS[\"ETRS89 / NTM zone 13\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5113\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5114, 1, 'ETRS89 / NTM zone 14', 'EPSG', 5114, 'PROJCS[\"ETRS89 / NTM zone 14\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",14.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5114\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5115, 1, 'ETRS89 / NTM zone 15', 'EPSG', 5115, 'PROJCS[\"ETRS89 / NTM zone 15\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5115\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5116, 1, 'ETRS89 / NTM zone 16', 'EPSG', 5116, 'PROJCS[\"ETRS89 / NTM zone 16\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5116\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5117, 1, 'ETRS89 / NTM zone 17', 'EPSG', 5117, 'PROJCS[\"ETRS89 / NTM zone 17\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5117\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5118, 1, 'ETRS89 / NTM zone 18', 'EPSG', 5118, 'PROJCS[\"ETRS89 / NTM zone 18\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5118\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5119, 1, 'ETRS89 / NTM zone 19', 'EPSG', 5119, 'PROJCS[\"ETRS89 / NTM zone 19\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5120, 1, 'ETRS89 / NTM zone 20', 'EPSG', 5120, 'PROJCS[\"ETRS89 / NTM zone 20\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5121, 1, 'ETRS89 / NTM zone 21', 'EPSG', 5121, 'PROJCS[\"ETRS89 / NTM zone 21\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5122, 1, 'ETRS89 / NTM zone 22', 'EPSG', 5122, 'PROJCS[\"ETRS89 / NTM zone 22\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",22.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5123, 1, 'ETRS89 / NTM zone 23', 'EPSG', 5123, 'PROJCS[\"ETRS89 / NTM zone 23\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5123\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5124, 1, 'ETRS89 / NTM zone 24', 'EPSG', 5124, 'PROJCS[\"ETRS89 / NTM zone 24\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5125, 1, 'ETRS89 / NTM zone 25', 'EPSG', 5125, 'PROJCS[\"ETRS89 / NTM zone 25\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5125\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5126, 1, 'ETRS89 / NTM zone 26', 'EPSG', 5126, 'PROJCS[\"ETRS89 / NTM zone 26\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",26.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5126\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5127, 1, 'ETRS89 / NTM zone 27', 'EPSG', 5127, 'PROJCS[\"ETRS89 / NTM zone 27\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5128, 1, 'ETRS89 / NTM zone 28', 'EPSG', 5128, 'PROJCS[\"ETRS89 / NTM zone 28\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5129, 1, 'ETRS89 / NTM zone 29', 'EPSG', 5129, 'PROJCS[\"ETRS89 / NTM zone 29\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5129\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5130, 1, 'ETRS89 / NTM zone 30', 'EPSG', 5130, 'PROJCS[\"ETRS89 / NTM zone 30\",GEOGCS[\"ETRS8"
  "9\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30.5,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5130\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5132, 1, 'Tokyo 1892', 'EPSG', 5132, 'GEOGCS[\"Tokyo 1892\",DATUM[\"Tokyo 1892\",SPHEROID[\"Be"
  "ssel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1048\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5132\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5167, 1, 'Korean 1985 / East Sea Belt', 'EPSG', 5167, 'PROJCS[\"Korean 1985 / East Sea Belt\","
  "GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\""
  ",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",131,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"51"
  "67\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5168, 1, 'Korean 1985 / Central Belt Jeju', 'EPSG', 5168, 'PROJCS[\"Korean 1985 / Central Belt"
  " Jeju\",GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY"
  "[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",127,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",550000,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EP"
  "SG\",\"5168\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5169, 1, 'Tokyo 1892 / Korea West Belt', 'EPSG', 5169, 'PROJCS[\"Tokyo 1892 / Korea West Belt\""
  ",GEOGCS[\"Tokyo 1892\",DATUM[\"Tokyo 1892\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700"
  "4\"]],AUTHORITY[\"EPSG\",\"1048\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5132\"]],PROJECTIO"
  "N[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5169\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5170, 1, 'Tokyo 1892 / Korea Central Belt', 'EPSG', 5170, 'PROJCS[\"Tokyo 1892 / Korea Central"
  " Belt\",GEOGCS[\"Tokyo 1892\",DATUM[\"Tokyo 1892\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\""
  ",\"7004\"]],AUTHORITY[\"EPSG\",\"1048\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5132\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale facto"
  "r at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5170\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5171, 1, 'Tokyo 1892 / Korea East Belt', 'EPSG', 5171, 'PROJCS[\"Tokyo 1892 / Korea East Belt\""
  ",GEOGCS[\"Tokyo 1892\",DATUM[\"Tokyo 1892\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700"
  "4\"]],AUTHORITY[\"EPSG\",\"1048\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5132\"]],PROJECTIO"
  "N[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at n"
  "atural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5172, 1, 'Tokyo 1892 / Korea East Sea Belt', 'EPSG', 5172, 'PROJCS[\"Tokyo 1892 / Korea East S"
  "ea Belt\",GEOGCS[\"Tokyo 1892\",DATUM[\"Tokyo 1892\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],AUTHORITY[\"EPSG\",\"1048\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5132\"]],"
  "PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]"
  "],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5172\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5173, 1, 'Korean 1985 / Modified West Belt', 'EPSG', 5173, 'PROJCS[\"Korean 1985 / Modified We"
  "st Belt\",GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORI"
  "TY[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",125.002890277778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",5"
  "00000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5174, 1, 'Korean 1985 / Modified Central Belt', 'EPSG', 5174, 'PROJCS[\"Korean 1985 / Modified"
  " Central Belt\",GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,A"
  "UTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",127.002890277778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\""
  ",EAST],AUTHORITY[\"EPSG\",\"5174\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5175, 1, 'Korean 1985 / Modified Central Belt Jeju', 'EPSG', 5175, 'PROJCS[\"Korean 1985 / Mod"
  "ified Central Belt Jeju\",GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299"
  ".1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\""
  ",\"6162\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",127.002890277778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",550000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],A"
  "XIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5175\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5176, 1, 'Korean 1985 / Modified East Belt', 'EPSG', 5176, 'PROJCS[\"Korean 1985 / Modified Ea"
  "st Belt\",GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORI"
  "TY[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",129.002890277778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",5"
  "00000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5176\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5177, 1, 'Korean 1985 / Modified East Sea Belt', 'EPSG', 5177, 'PROJCS[\"Korean 1985 / Modifie"
  "d East Sea Belt\",GEOGCS[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128"
  ",AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",131.002890277778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\""
  ",EAST],AUTHORITY[\"EPSG\",\"5177\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5178, 1, 'Korean 1985 / Unified CS', 'EPSG', 5178, 'PROJCS[\"Korean 1985 / Unified CS\",GEOGCS"
  "[\"Korean 1985\",DATUM[\"Korean Datum 1985\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"70"
  "04\"]],TOWGS84[-145.907,505.034,685.756,1.162,-2.347,-1.592,6.342],AUTHORITY[\"EPSG\",\"6162\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4162\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",1"
  "27.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5178\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5179, 1, 'Korea 2000 / Unified CS', 'EPSG', 5179, 'PROJCS[\"Korea 2000 / Unified CS\",GEOGCS[\""
  "Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127.511111111111,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5179\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5180, 1, 'Korea 2000 / West Belt', 'EPSG', 5180, 'PROJCS[\"Korea 2000 / West Belt\",GEOGCS[\"K"
  "orea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5180\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5181, 1, 'Korea 2000 / Central Belt', 'EPSG', 5181, 'PROJCS[\"Korea 2000 / Central Belt\",GEOG"
  "CS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5181\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5182, 1, 'Korea 2000 / Central Belt Jeju', 'EPSG', 5182, 'PROJCS[\"Korea 2000 / Central Belt J"
  "eju\",GEOGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",550000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5182\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5183, 1, 'Korea 2000 / East Belt', 'EPSG', 5183, 'PROJCS[\"Korea 2000 / East Belt\",GEOGCS[\"K"
  "orea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5183\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5184, 1, 'Korea 2000 / East Sea Belt', 'EPSG', 5184, 'PROJCS[\"Korea 2000 / East Sea Belt\",GE"
  "OGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5184\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5185, 1, 'Korea 2000 / West Belt 2010', 'EPSG', 5185, 'PROJCS[\"Korea 2000 / West Belt 2010\","
  "GEOGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS"
  "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",125,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5185\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5186, 1, 'Korea 2000 / Central Belt 2010', 'EPSG', 5186, 'PROJCS[\"Korea 2000 / Central Belt 2"
  "010\",GEOGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5186\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5187, 1, 'Korea 2000 / East Belt 2010', 'EPSG', 5187, 'PROJCS[\"Korea 2000 / East Belt 2010\","
  "GEOGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPS"
  "G\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5187\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5188, 1, 'Korea 2000 / East Sea Belt 2010', 'EPSG', 5188, 'PROJCS[\"Korea 2000 / East Sea Belt"
  " 2010\",GEOGCS[\"Korea 2000\",DATUM[\"Geocentric datum of Korea\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI"
  "TY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6737\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4737\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5188\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5221, 1, 'S-JTSK (Ferro) / Krovak East North', 'EPSG', 5221, 'PROJCS[\"S-JTSK (Ferro) / Krovak"
  " East North\",GEOGCS[\"S-JTSK (Ferro)\",DATUM[\"System of the Unified Trigonometrical Cadastral Network (Ferro)\",SPH"
  "EROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6818\"]],PRIMEM[\"Fe"
  "rro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4818\"]],PROJECTION[\"Krovak (North Orientated)\",A"
  "UTHORITY[\"EPSG\",\"1041\"]],PARAMETER[\"Latitude of projection centre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\""
  "]],PARAMETER[\"Longitude of origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone a"
  "xis\",30.2881397527778,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"Latitude of pseudo standard parallel\",78.5111111111"
  "111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseudo standard parallel\",0.9999,AUTHORITY[\"EPSG\",\""
  "8819\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"52"
  "21\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5223, 1, 'WGS 84 / Gabon TM', 'EPSG', 5223, 'PROJCS[\"WGS 84 / Gabon TM\",GEOGCS[\"WGS 84\",DA"
  "TUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY["
  "\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0"
  ".9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5223\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5224, 1, 'S-JTSK/05 (Ferro) / Modified Krovak', 'EPSG', 5224, 'PROJCS[\"S-JTSK/05 (Ferro) / Mo"
  "dified Krovak\",GEOGCS[\"S-JTSK/05 (Ferro)\",DATUM[\"System of the Unified Trigonometrical Cadastral Network/05 (Ferr"
  "o)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1055\"]],PRI"
  "MEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5229\"]],PROJECTION[\"Krovak Modified\",AUT"
  "HORITY[\"EPSG\",\"1042\"]],PARAMETER[\"Latitude of projection centre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\"]]"
  ",PARAMETER[\"Longitude of origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone axi"
  "s\",30.2881397222222,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"Latitude of pseudo standard parallel\",78.511111111111"
  "1,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseudo standard parallel\",0.9999,AUTHORITY[\"EPSG\",\"88"
  "19\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",5000000,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],PARAMETER[\"Ordinate 1 of evaluation point\",1089000,AUTHORITY[\"EPSG\",\"8617\"]],PARAMETER[\""
  "Ordinate 2 of evaluation point\",654000,AUTHORITY[\"EPSG\",\"8618\"]],PARAMETER[\"C1\",0.02946529277,AUTHORITY[\"EPSG"
  "\",\"1026\"]],PARAMETER[\"C2\",0.02515965696,AUTHORITY[\"EPSG\",\"1027\"]],PARAMETER[\"C3\",1.193845912e-07,AUTHORITY"
  "[\"EPSG\",\"1028\"]],PARAMETER[\"C4\",-4.668270147e-07,AUTHORITY[\"EPSG\",\"1029\"]],PARAMETER[\"C5\",9.233980362e-12"
  ",AUTHORITY[\"EPSG\",\"1030\"]],PARAMETER[\"C6\",1.523735715e-12,AUTHORITY[\"EPSG\",\"1031\"]],PARAMETER[\"C7\",1.6967"
  "80024e-18,AUTHORITY[\"EPSG\",\"1032\"]],PARAMETER[\"C8\",4.408314235e-18,AUTHORITY[\"EPSG\",\"1033\"]],PARAMETER[\"C9"
  "\",-8.331083518e-24,AUTHORITY[\"EPSG\",\"1034\"]],PARAMETER[\"C10\",-3.689471323e-24,AUTHORITY[\"EPSG\",\"1035\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"EPSG\",\"5224\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5225, 1, 'S-JTSK/05 (Ferro) / Modified Krovak East North', 'EPSG', 5225, 'PROJCS[\"S-JTSK/05 ("
  "Ferro) / Modified Krovak East North\",GEOGCS[\"S-JTSK/05 (Ferro)\",DATUM[\"System of the Unified Trigonometrical Cada"
  "stral Network/05 (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\""
  "EPSG\",\"1055\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5229\"]],PROJECTION[\""
  "Krovak Modified (North Orientated)\",AUTHORITY[\"EPSG\",\"1043\"]],PARAMETER[\"Latitude of projection centre\",49.511"
  "1111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8833"
  "\"]],PARAMETER[\"Co-latitude of cone axis\",30.2881397222222,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"Latitude of ps"
  "eudo standard parallel\",78.5111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseudo standard p"
  "arallel\",0.9999,AUTHORITY[\"EPSG\",\"8819\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",5000000,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Ordinate 1 of evaluation point\",1089000,A"
  "UTHORITY[\"EPSG\",\"8617\"]],PARAMETER[\"Ordinate 2 of evaluation point\",654000,AUTHORITY[\"EPSG\",\"8618\"]],PARAME"
  "TER[\"C1\",0.02946529277,AUTHORITY[\"EPSG\",\"1026\"]],PARAMETER[\"C2\",0.02515965696,AUTHORITY[\"EPSG\",\"1027\"]],P"
  "ARAMETER[\"C3\",1.193845912e-07,AUTHORITY[\"EPSG\",\"1028\"]],PARAMETER[\"C4\",-4.668270147e-07,AUTHORITY[\"EPSG\",\""
  "1029\"]],PARAMETER[\"C5\",9.233980362e-12,AUTHORITY[\"EPSG\",\"1030\"]],PARAMETER[\"C6\",1.523735715e-12,AUTHORITY[\""
  "EPSG\",\"1031\"]],PARAMETER[\"C7\",1.696780024e-18,AUTHORITY[\"EPSG\",\"1032\"]],PARAMETER[\"C8\",4.408314235e-18,AUT"
  "HORITY[\"EPSG\",\"1033\"]],PARAMETER[\"C9\",-8.331083518e-24,AUTHORITY[\"EPSG\",\"1034\"]],PARAMETER[\"C10\",-3.68947"
  "1323e-24,AUTHORITY[\"EPSG\",\"1035\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"5225\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5228, 1, 'S-JTSK/05', 'EPSG', 5228, 'GEOGCS[\"S-JTSK/05\",DATUM[\"System of the Unified Trigon"
  "ometrical Cadastral Network/05\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS"
  "84[572.213,85.334,461.94,4.9732,1.529,5.2484,3.5378],AUTHORITY[\"EPSG\",\"1052\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5228\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5229, 1, 'S-JTSK/05 (Ferro)', 'EPSG', 5229, 'GEOGCS[\"S-JTSK/05 (Ferro)\",DATUM[\"System of th"
  "e Unified Trigonometrical Cadastral Network/05 (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1055\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5229\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5233, 1, 'SLD99', 'EPSG', 5233, 'GEOGCS[\"SLD99\",DATUM[\"Sri Lanka Datum 1999\",SPHEROID[\"Ev"
  "erest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[-0.293,766.95,87.713,0.1957"
  "04,1.695068,3.473016,-0.039338],AUTHORITY[\"EPSG\",\"1053\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5233\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5234, 1, 'Kandawala / Sri Lanka Grid', 'EPSG', 5234, 'PROJCS[\"Kandawala / Sri Lanka Grid\",GE"
  "OGCS[\"Kandawala\",DATUM[\"Kandawala\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"E"
  "PSG\",\"7015\"]],TOWGS84[-97,787,86,0,0,0,0],AUTHORITY[\"EPSG\",\"6244\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4244\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",7.00048027777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",80.771711"
  "1111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999238418,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\""
  "EPSG\",\"5234\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5235, 1, 'SLD99 / Sri Lanka Grid 1999', 'EPSG', 5235, 'PROJCS[\"SLD99 / Sri Lanka Grid 1999\","
  "GEOGCS[\"SLD99\",DATUM[\"Sri Lanka Datum 1999\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTH"
  "ORITY[\"EPSG\",\"7015\"]],TOWGS84[-0.293,766.95,87.713,0.195704,1.695068,3.473016,-0.039338],AUTHORITY[\"EPSG\",\"105"
  "3\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5233\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",7.00047152777778,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",80.7717130833333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at nat"
  "ural origin\",0.9999238418,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]]"
  ",AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5235\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5243, 1, 'ETRS89 / LCC Germany (E-N)', 'EPSG', 5243, 'PROJCS[\"ETRS89 / LCC Germany (E-N)\",GE"
  "OGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",51,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10.5,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",53.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5243\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5246, 1, 'GDBD2009', 'EPSG', 5246, 'GEOGCS[\"GDBD2009\",DATUM[\"Geocentric Datum Brunei Daruss"
  "alam 2009\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORIT"
  "Y[\"EPSG\",\"1056\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5247, 1, 'GDBD2009 / Brunei BRSO', 'EPSG', 5247, 'PROJCS[\"GDBD2009 / Brunei BRSO\",GEOGCS[\"G"
  "DBD2009\",DATUM[\"Geocentric Datum Brunei Darussalam 2009\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1056\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"5246\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\""
  "Latitude of projection centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",115,AUTH"
  "ORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",53.31580995,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\""
  "Angle from Rectified to Skew Grid\",53.1301023611111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initia"
  "l line\",0.99984,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AX"
  "IS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5247\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5252, 1, 'TUREF', 'EPSG', 5252, 'GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5253, 1, 'TUREF / TM27', 'EPSG', 5253, 'PROJCS[\"TUREF / TM27\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5254, 1, 'TUREF / TM30', 'EPSG', 5254, 'PROJCS[\"TUREF / TM30\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5255, 1, 'TUREF / TM33', 'EPSG', 5255, 'PROJCS[\"TUREF / TM33\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5256, 1, 'TUREF / TM36', 'EPSG', 5256, 'PROJCS[\"TUREF / TM36\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5257, 1, 'TUREF / TM39', 'EPSG', 5257, 'PROJCS[\"TUREF / TM39\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5258, 1, 'TUREF / TM42', 'EPSG', 5258, 'PROJCS[\"TUREF / TM42\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",42,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5259, 1, 'TUREF / TM45', 'EPSG', 5259, 'PROJCS[\"TUREF / TM45\",GEOGCS[\"TUREF\",DATUM[\"Turki"
  "sh National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5264, 1, 'DRUKREF 03', 'EPSG', 5264, 'GEOGCS[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic D"
  "atum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5264\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5266, 1, 'DRUKREF 03 / Bhutan National Grid', 'EPSG', 5266, 'PROJCS[\"DRUKREF 03 / Bhutan Nati"
  "onal Grid\",GEOGCS[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5266\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5269, 1, 'TUREF / 3-degree Gauss-Kruger zone 9', 'EPSG', 5269, 'PROJCS[\"TUREF / 3-degree Gaus"
  "s-Kruger zone 9\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5269\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5270, 1, 'TUREF / 3-degree Gauss-Kruger zone 10', 'EPSG', 5270, 'PROJCS[\"TUREF / 3-degree Gau"
  "ss-Kruger zone 10\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5270\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5271, 1, 'TUREF / 3-degree Gauss-Kruger zone 11', 'EPSG', 5271, 'PROJCS[\"TUREF / 3-degree Gau"
  "ss-Kruger zone 11\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5271\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5272, 1, 'TUREF / 3-degree Gauss-Kruger zone 12', 'EPSG', 5272, 'PROJCS[\"TUREF / 3-degree Gau"
  "ss-Kruger zone 12\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",36,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5272\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5273, 1, 'TUREF / 3-degree Gauss-Kruger zone 13', 'EPSG', 5273, 'PROJCS[\"TUREF / 3-degree Gau"
  "ss-Kruger zone 13\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5273\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5274, 1, 'TUREF / 3-degree Gauss-Kruger zone 14', 'EPSG', 5274, 'PROJCS[\"TUREF / 3-degree Gau"
  "ss-Kruger zone 14\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",42,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5274\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5275, 1, 'TUREF / 3-degree Gauss-Kruger zone 15', 'EPSG', 5275, 'PROJCS[\"TUREF / 3-degree Gau"
  "ss-Kruger zone 15\",GEOGCS[\"TUREF\",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5275\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5292, 1, 'DRUKREF 03 / Bumthang TM', 'EPSG', 5292, 'PROJCS[\"DRUKREF 03 / Bumthang TM\",GEOGCS"
  "[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.7444444444444,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5292\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5293, 1, 'DRUKREF 03 / Chhukha TM', 'EPSG', 5293, 'PROJCS[\"DRUKREF 03 / Chhukha TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",89.5611111111111,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5293\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5294, 1, 'DRUKREF 03 / Dagana TM', 'EPSG', 5294, 'PROJCS[\"DRUKREF 03 / Dagana TM\",GEOGCS[\"D"
  "RUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",89.85,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5294\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5295, 1, 'DRUKREF 03 / Gasa TM', 'EPSG', 5295, 'PROJCS[\"DRUKREF 03 / Gasa TM\",GEOGCS[\"DRUKR"
  "EF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.0444444444444,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",25000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5295\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5296, 1, 'DRUKREF 03 / Ha TM', 'EPSG', 5296, 'PROJCS[\"DRUKREF 03 / Ha TM\",GEOGCS[\"DRUKREF 0"
  "3\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5296\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5297, 1, 'DRUKREF 03 / Lhuentse TM', 'EPSG', 5297, 'PROJCS[\"DRUKREF 03 / Lhuentse TM\",GEOGCS"
  "[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",91.1444444444444,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5297\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5298, 1, 'DRUKREF 03 / Mongar TM', 'EPSG', 5298, 'PROJCS[\"DRUKREF 03 / Mongar TM\",GEOGCS[\"D"
  "RUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",91.2333333333333,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5298\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5299, 1, 'DRUKREF 03 / Paro TM', 'EPSG', 5299, 'PROJCS[\"DRUKREF 03 / Paro TM\",GEOGCS[\"DRUKR"
  "EF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",89.3611111111111,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",25000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5299\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5300, 1, 'DRUKREF 03 / Pemagatshel TM', 'EPSG', 5300, 'PROJCS[\"DRUKREF 03 / Pemagatshel TM\","
  "GEOGCS[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",91.3611111111111,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5300\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5301, 1, 'DRUKREF 03 / Punakha TM', 'EPSG', 5301, 'PROJCS[\"DRUKREF 03 / Punakha TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",89.85,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5301\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5302, 1, 'DRUKREF 03 / Samdrup Jongkhar TM', 'EPSG', 5302, 'PROJCS[\"DRUKREF 03 / Samdrup Jong"
  "khar TM\",GEOGCS[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",91.5666666666667,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5302\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5303, 1, 'DRUKREF 03 / Samtse TM', 'EPSG', 5303, 'PROJCS[\"DRUKREF 03 / Samtse TM\",GEOGCS[\"D"
  "RUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",89.0666666666667,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5303\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5304, 1, 'DRUKREF 03 / Sarpang TM', 'EPSG', 5304, 'PROJCS[\"DRUKREF 03 / Sarpang TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.2777777777778,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5304\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5305, 1, 'DRUKREF 03 / Thimphu TM', 'EPSG', 5305, 'PROJCS[\"DRUKREF 03 / Thimphu TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",89.5611111111111,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5305\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5306, 1, 'DRUKREF 03 / Trashigang TM', 'EPSG', 5306, 'PROJCS[\"DRUKREF 03 / Trashigang TM\",GE"
  "OGCS[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",91.75,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2500"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5306\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5307, 1, 'DRUKREF 03 / Trongsa TM', 'EPSG', 5307, 'PROJCS[\"DRUKREF 03 / Trongsa TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.5111111111111,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5307\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5308, 1, 'DRUKREF 03 / Tsirang TM', 'EPSG', 5308, 'PROJCS[\"DRUKREF 03 / Tsirang TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.1777777777778,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5308\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5309, 1, 'DRUKREF 03 / Wangdue Phodrang TM', 'EPSG', 5309, 'PROJCS[\"DRUKREF 03 / Wangdue Phod"
  "rang TM\",GEOGCS[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.1277777777778,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5309\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5310, 1, 'DRUKREF 03 / Yangtse TM', 'EPSG', 5310, 'PROJCS[\"DRUKREF 03 / Yangtse TM\",GEOGCS[\""
  "DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",91.5666666666667,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5310\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5311, 1, 'DRUKREF 03 / Zhemgang TM', 'EPSG', 5311, 'PROJCS[\"DRUKREF 03 / Zhemgang TM\",GEOGCS"
  "[\"DRUKREF 03\",DATUM[\"Bhutan National Geodetic Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1058\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5264\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90.8777777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5316, 1, 'ETRS89 / Faroe TM', 'EPSG', 5316, 'PROJCS[\"ETRS89 / Faroe TM\",GEOGCS[\"ETRS89\",DA"
  "TUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-7,AUTHORITY[\"EPSG\",\"8802\"]],PARAM"
  "ETER[\"Scale factor at natural origin\",0.999997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-6000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5320, 1, 'NAD83 / Teranet Ontario Lambert', 'EPSG', 5320, 'PROJCS[\"NAD83 / Teranet Ontario La"
  "mbert\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",54.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5320\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5321, 1, 'NAD83(CSRS) / Teranet Ontario Lambert', 'EPSG', 5321, 'PROJCS[\"NAD83(CSRS) / Terane"
  "t Ontario Lambert\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041"
  ",0.0116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degr"
  "ee\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4"
  "617\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84,AUTHORITY[\"EPSG\",\"8822\"]],PARAME"
  "TER[\"Latitude of 1st standard parallel\",44.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",54.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTH"
  "ORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5321\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5324, 1, 'ISN2004', 'EPSG', 5324, 'GEOGCS[\"ISN2004\",DATUM[\"Islands Net 2004\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1060\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5325, 1, 'ISN2004 / Lambert 2004', 'EPSG', 5325, 'PROJCS[\"ISN2004 / Lambert 2004\",GEOGCS[\"I"
  "SN2004\",DATUM[\"Islands Net 2004\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84"
  "[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1060\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5324\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",65"
  ",AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-19,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",64.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel"
  "\",65.75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1700000,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",300000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5325\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5329, 1, 'Segara (Jakarta) / NEIEZ', 'EPSG', 5329, 'PROJCS[\"Segara (Jakarta) / NEIEZ\",GEOGCS"
  "[\"Segara (Jakarta)\",DATUM[\"Gunung Segara (Jakarta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6820\"]],PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4820\"]],PROJECTION[\"Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural o"
  "rigin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3.19228055555555,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",3900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5329\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5330, 1, 'Batavia (Jakarta) / NEIEZ', 'EPSG', 5330, 'PROJCS[\"Batavia (Jakarta) / NEIEZ\",GEOG"
  "CS[\"Batavia (Jakarta)\",DATUM[\"Batavia (Jakarta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],AUTHORITY[\"EPSG\",\"6813\"]],PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4813\"]],PROJECTION[\"Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural orig"
  "in\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3.19228055555555,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",3900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5331, 1, 'Makassar (Jakarta) / NEIEZ', 'EPSG', 5331, 'PROJCS[\"Makassar (Jakarta) / NEIEZ\",GE"
  "OGCS[\"Makassar (Jakarta)\",DATUM[\"Makassar (Jakarta)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6804\"]],PRIMEM[\"Jakarta\",106.80771944444443,AUTHORITY[\"EPSG\",\"8908\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4804\"]],PROJECTION[\"Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9804\"]],PARAMETER[\"Latitude of natural o"
  "rigin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3.19228055555555,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.997,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",3900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5337, 1, 'Aratu / UTM zone 25S', 'EPSG', 5337, 'PROJCS[\"Aratu / UTM zone 25S\",GEOGCS[\"Aratu"
  "\",DATUM[\"Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-151.99,287.04,-"
  "147.45,0,0,0,0],AUTHORITY[\"EPSG\",\"6208\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4208\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\","
  "\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5340, 1, 'POSGAR 2007', 'EPSG', 5340, 'GEOGCS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Ar"
  "gentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTH"
  "ORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5343, 1, 'POSGAR 2007 / Argentina 1', 'EPSG', 5343, 'PROJCS[\"POSGAR 2007 / Argentina 1\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5344, 1, 'POSGAR 2007 / Argentina 2', 'EPSG', 5344, 'PROJCS[\"POSGAR 2007 / Argentina 2\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5345, 1, 'POSGAR 2007 / Argentina 3', 'EPSG', 5345, 'PROJCS[\"POSGAR 2007 / Argentina 3\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-66,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5345\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5346, 1, 'POSGAR 2007 / Argentina 4', 'EPSG', 5346, 'PROJCS[\"POSGAR 2007 / Argentina 4\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5347, 1, 'POSGAR 2007 / Argentina 5', 'EPSG', 5347, 'PROJCS[\"POSGAR 2007 / Argentina 5\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-60,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5347\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5348, 1, 'POSGAR 2007 / Argentina 6', 'EPSG', 5348, 'PROJCS[\"POSGAR 2007 / Argentina 6\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5349, 1, 'POSGAR 2007 / Argentina 7', 'EPSG', 5349, 'PROJCS[\"POSGAR 2007 / Argentina 7\",GEOG"
  "CS[\"POSGAR 2007\",DATUM[\"Posiciones Geodesicas Argentinas 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1062\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"5340\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5349\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5354, 1, 'MARGEN', 'EPSG', 5354, 'GEOGCS[\"MARGEN\",DATUM[\"Marco Geodesico Nacional de Bolivi"
  "a\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG"
  "\",\"1063\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5355, 1, 'MARGEN / UTM zone 20S', 'EPSG', 5355, 'PROJCS[\"MARGEN / UTM zone 20S\",GEOGCS[\"MAR"
  "GEN\",DATUM[\"Marco Geodesico Nacional de Bolivia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1063\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5354\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5356, 1, 'MARGEN / UTM zone 19S', 'EPSG', 5356, 'PROJCS[\"MARGEN / UTM zone 19S\",GEOGCS[\"MAR"
  "GEN\",DATUM[\"Marco Geodesico Nacional de Bolivia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1063\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5354\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5357, 1, 'MARGEN / UTM zone 21S', 'EPSG', 5357, 'PROJCS[\"MARGEN / UTM zone 21S\",GEOGCS[\"MAR"
  "GEN\",DATUM[\"Marco Geodesico Nacional de Bolivia\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1063\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5354\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5357\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5360, 1, 'SIRGAS-Chile', 'EPSG', 5360, 'GEOGCS[\"SIRGAS-Chile\",DATUM[\"SIRGAS-Chile\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1064\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5360\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5361, 1, 'SIRGAS-Chile / UTM zone 19S', 'EPSG', 5361, 'PROJCS[\"SIRGAS-Chile / UTM zone 19S\","
  "GEOGCS[\"SIRGAS-Chile\",DATUM[\"SIRGAS-Chile\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1064\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"5360\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5361\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5362, 1, 'SIRGAS-Chile / UTM zone 18S', 'EPSG', 5362, 'PROJCS[\"SIRGAS-Chile / UTM zone 18S\","
  "GEOGCS[\"SIRGAS-Chile\",DATUM[\"SIRGAS-Chile\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1064\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"5360\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5362\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5365, 1, 'CR05', 'EPSG', 5365, 'GEOGCS[\"CR05\",DATUM[\"Costa Rica 2005\",SPHEROID[\"WGS 84\","
  "6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-0.16959,0.35312,0.51846,-0.03385,0.16325,-0.03446,0.0369"
  "3],AUTHORITY[\"EPSG\",\"1065\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5365\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5367, 1, 'CR05 / CRTM05', 'EPSG', 5367, 'PROJCS[\"CR05 / CRTM05\",GEOGCS[\"CR05\",DATUM[\"Cost"
  "a Rica 2005\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-0.16959,0.35312,0.5184"
  "6,-0.03385,0.16325,-0.03446,0.03693],AUTHORITY[\"EPSG\",\"1065\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5367\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5371, 1, 'MACARIO SOLIS', 'EPSG', 5371, 'GEOGCS[\"MACARIO SOLIS\",DATUM[\"Sistema Geodesico Na"
  "cional de Panama MACARIO SOLIS\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,"
  "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1066\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5371\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5373, 1, 'Peru96', 'EPSG', 5373, 'GEOGCS[\"Peru96\",DATUM[\"Peru96\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1067\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5373\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5381, 1, 'SIRGAS-ROU98', 'EPSG', 5381, 'GEOGCS[\"SIRGAS-ROU98\",DATUM[\"SIRGAS-ROU98\",SPHEROI"
  "D[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1068\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5381\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5382, 1, 'SIRGAS-ROU98 / UTM zone 21S', 'EPSG', 5382, 'PROJCS[\"SIRGAS-ROU98 / UTM zone 21S\","
  "GEOGCS[\"SIRGAS-ROU98\",DATUM[\"SIRGAS-ROU98\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1068\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "5381\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5382\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5383, 1, 'SIRGAS-ROU98 / UTM zone 22S', 'EPSG', 5383, 'PROJCS[\"SIRGAS-ROU98 / UTM zone 22S\","
  "GEOGCS[\"SIRGAS-ROU98\",DATUM[\"SIRGAS-ROU98\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]"
  "],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1068\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "5381\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5383\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5387, 1, 'Peru96 / UTM zone 18S', 'EPSG', 5387, 'PROJCS[\"Peru96 / UTM zone 18S\",GEOGCS[\"Per"
  "u96\",DATUM[\"Peru96\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"1067\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5373\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5387\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5389, 1, 'Peru96 / UTM zone 19S', 'EPSG', 5389, 'PROJCS[\"Peru96 / UTM zone 19S\",GEOGCS[\"Per"
  "u96\",DATUM[\"Peru96\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"1067\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5373\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5389\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5393, 1, 'SIRGAS_ES2007.8', 'EPSG', 5393, 'GEOGCS[\"SIRGAS_ES2007.8\",DATUM[\"SIRGAS_ES2007.8\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"1069\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5396, 1, 'SIRGAS 2000 / UTM zone 26S', 'EPSG', 5396, 'PROJCS[\"SIRGAS 2000 / UTM zone 26S\",GE"
  "OGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-2"
  "7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"53"
  "96\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5451, 1, 'Ocotepeque 1935', 'EPSG', 5451, 'GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\""
  ",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHO"
  "RITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5451\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5456, 1, 'Ocotepeque 1935 / Costa Rica Norte', 'EPSG', 5456, 'PROJCS[\"Ocotepeque 1935 / Costa"
  " Rica Norte\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.97869821389"
  "82,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",10.4777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-84.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999956"
  "96,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",271820.522,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5456\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5457, 1, 'Ocotepeque 1935 / Costa Rica Sur', 'EPSG', 5457, 'PROJCS[\"Ocotepeque 1935 / Costa R"
  "ica Sur\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,A"
  "UTHORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],"
  "PARAMETER[\"Latitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "3.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995696,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",327987.4"
  "36,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"5457\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5459, 1, 'Ocotepeque 1935 / Guatemala Sur', 'EPSG', 5459, 'PROJCS[\"Ocotepeque 1935 / Guatemal"
  "a Sur\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUT"
  "HORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",14.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-90.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99989906,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",325992.681,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"5459\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5460, 1, 'Ocotepeque 1935 / El Salvador Lambert', 'EPSG', 5460, 'PROJCS[\"Ocotepeque 1935 / El"
  " Salvador Lambert\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.97869"
  "82138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",13.7833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-89,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99996704,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",295809.184,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"5460\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5461, 1, 'Ocotepeque 1935 / Nicaragua Norte', 'EPSG', 5461, 'PROJCS[\"Ocotepeque 1935 / Nicara"
  "gua Norte\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982"
  ",AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",13.8777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-85.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999031"
  "4,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",359891.816,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5461\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5462, 1, 'Ocotepeque 1935 / Nicaragua Sur', 'EPSG', 5462, 'PROJCS[\"Ocotepeque 1935 / Nicaragu"
  "a Sur\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUT"
  "HORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",11.7444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-85.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99992228,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",288876.327,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"5462\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5463, 1, 'SAD69 / UTM zone 17N', 'EPSG', 5463, 'PROJCS[\"SAD69 / UTM zone 17N\",GEOGCS[\"SAD69"
  "\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],TO"
  "WGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5463\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5464, 1, 'Sibun Gorge 1922', 'EPSG', 5464, 'GEOGCS[\"Sibun Gorge 1922\",DATUM[\"Sibun Gorge 19"
  "22\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.26067636926103,AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\""
  "1071\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5464\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5467, 1, 'Panama-Colon 1911', 'EPSG', 5467, 'GEOGCS[\"Panama-Colon 1911\",DATUM[\"Panama-Colon"
  " 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"1072\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5467\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5469, 1, 'Panama-Colon 1911 / Panama Lambert', 'EPSG', 5469, 'PROJCS[\"Panama-Colon 1911 / Pan"
  "ama Lambert\",GEOGCS[\"Panama-Colon 1911\",DATUM[\"Panama-Colon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982"
  "138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"1072\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5467\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",8.41666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-80,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99989909,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",294865.303,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "5469\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5472, 1, 'Panama-Colon 1911 / Panama Polyconic', 'EPSG', 5472, 'PROJCS[\"Panama-Colon 1911 / P"
  "anama Polyconic\",GEOGCS[\"Panama-Colon 1911\",DATUM[\"Panama-Colon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.978"
  "6982138982,AUTHORITY[\"EPSG\",\"7008\"]],AUTHORITY[\"EPSG\",\"1072\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"5467\"]],PROJECTION[\"American Polyconic\",AUTHORITY[\"EPSG\",\"9818\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",8.25,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1092972.1,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"Clarke\\'s yard\",0.9143917962,AUTHORITY[\"EPSG\",\"9037\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"5472\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5479, 1, 'RSRGD2000 / MSLC2000', 'EPSG', 5479, 'PROJCS[\"RSRGD2000 / MSLC2000\",GEOGCS[\"RSRGD"
  "2000\",DATUM[\"Ross Sea Region Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4764\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",-78,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",163,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",-76.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",-79.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",7000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000000,AUTHORITY[\"EPSG\",\"8827\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5479\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5480, 1, 'RSRGD2000 / BCLC2000', 'EPSG', 5480, 'PROJCS[\"RSRGD2000 / BCLC2000\",GEOGCS[\"RSRGD"
  "2000\",DATUM[\"Ross Sea Region Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4764\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",-74.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",165,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",-75.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",5000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"54"
  "80\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5481, 1, 'RSRGD2000 / PCLC2000', 'EPSG', 5481, 'PROJCS[\"RSRGD2000 / PCLC2000\",GEOGCS[\"RSRGD"
  "2000\",DATUM[\"Ross Sea Region Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4764\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",166,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-70.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",-72.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"54"
  "81\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5482, 1, 'RSRGD2000 / RSPS2000', 'EPSG', 5482, 'PROJCS[\"RSRGD2000 / RSPS2000\",GEOGCS[\"RSRGD"
  "2000\",DATUM[\"Ross Sea Region Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6764\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4764\"]],PROJECTION[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitude of"
  " natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",180,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",NORTH],AUTHORITY[\"EPSG\",\"5482\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5489, 1, 'RGAF09', 'EPSG', 5489, 'GEOGCS[\"RGAF09\",DATUM[\"Reseau Geodesique des Antilles Fra"
  "ncaises 2009\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHO"
  "RITY[\"EPSG\",\"1073\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5489\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5490, 1, 'RGAF09 / UTM zone 20N', 'EPSG', 5490, 'PROJCS[\"RGAF09 / UTM zone 20N\",GEOGCS[\"RGA"
  "F09\",DATUM[\"Reseau Geodesique des Antilles Francaises 2009\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1073\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5489\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5490\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5513, 1, 'S-JTSK / Krovak', 'EPSG', 5513, 'PROJCS[\"S-JTSK / Krovak\",GEOGCS[\"S-JTSK\",DATUM["
  "\"System of the Unified Trigonometrical Cadastral Network\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORIT"
  "Y[\"EPSG\",\"7004\"]],TOWGS84[589,76,480,0,0,0,0],AUTHORITY[\"EPSG\",\"6156\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4156\"]],PROJECTION[\"Krovak\",AUTHORITY[\"EPSG\",\"9819\"]],PARAMETER[\"Latitude of projec"
  "tion centre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of origin\",24.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone axis\",30.2881397527778,AUTHORITY[\"EPSG\",\"1036\"]],PARAME"
  "TER[\"Latitude of pseudo standard parallel\",78.5111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor "
  "on pseudo standard parallel\",0.9999,AUTHORITY[\"EPSG\",\"8819\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"EPSG\",\"5513\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5514, 1, 'S-JTSK / Krovak East North', 'EPSG', 5514, 'PROJCS[\"S-JTSK / Krovak East North\",GE"
  "OGCS[\"S-JTSK\",DATUM[\"System of the Unified Trigonometrical Cadastral Network\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[589,76,480,0,0,0,0],AUTHORITY[\"EPSG\",\"6156\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4156\"]],PROJECTION[\"Krovak (North Orientated)\",AUTHORITY[\"EPSG\","
  "\"1041\"]],PARAMETER[\"Latitude of projection centre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Lon"
  "gitude of origin\",24.8333333333333,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone axis\",30.288139752"
  "7778,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"Latitude of pseudo standard parallel\",78.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8818\"]],PARAMETER[\"Scale factor on pseudo standard parallel\",0.9999,AUTHORITY[\"EPSG\",\"8819\"]],PARAMETER"
  "[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5514\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5515, 1, 'S-JTSK/05 / Modified Krovak', 'EPSG', 5515, 'PROJCS[\"S-JTSK/05 / Modified Krovak\","
  "GEOGCS[\"S-JTSK/05\",DATUM[\"System of the Unified Trigonometrical Cadastral Network/05\",SPHEROID[\"Bessel 1841\",63"
  "77397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[572.213,85.334,461.94,4.9732,1.529,5.2484,3.5378],AUTHORI"
  "TY[\"EPSG\",\"1052\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5228\"]],PROJECTION[\"Krovak Mo"
  "dified\",AUTHORITY[\"EPSG\",\"1042\"]],PARAMETER[\"Latitude of projection centre\",49.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8811\"]],PARAMETER[\"Longitude of origin\",24.8333333333333,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude "
  "of cone axis\",30.2881397222222,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"Latitude of pseudo standard parallel\",78.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseudo standard parallel\",0.9999,AUTHORITY[\""
  "EPSG\",\"8819\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",5000"
  "000,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Ordinate 1 of evaluation point\",1089000,AUTHORITY[\"EPSG\",\"8617\"]],"
  "PARAMETER[\"Ordinate 2 of evaluation point\",654000,AUTHORITY[\"EPSG\",\"8618\"]],PARAMETER[\"C1\",0.02946529277,AUTH"
  "ORITY[\"EPSG\",\"1026\"]],PARAMETER[\"C2\",0.02515965696,AUTHORITY[\"EPSG\",\"1027\"]],PARAMETER[\"C3\",1.193845912e-"
  "07,AUTHORITY[\"EPSG\",\"1028\"]],PARAMETER[\"C4\",-4.668270147e-07,AUTHORITY[\"EPSG\",\"1029\"]],PARAMETER[\"C5\",9.2"
  "33980362e-12,AUTHORITY[\"EPSG\",\"1030\"]],PARAMETER[\"C6\",1.523735715e-12,AUTHORITY[\"EPSG\",\"1031\"]],PARAMETER[\""
  "C7\",1.696780024e-18,AUTHORITY[\"EPSG\",\"1032\"]],PARAMETER[\"C8\",4.408314235e-18,AUTHORITY[\"EPSG\",\"1033\"]],PAR"
  "AMETER[\"C9\",-8.331083518e-24,AUTHORITY[\"EPSG\",\"1034\"]],PARAMETER[\"C10\",-3.689471323e-24,AUTHORITY[\"EPSG\",\""
  "1035\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"EPSG\",\"5515\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5516, 1, 'S-JTSK/05 / Modified Krovak East North', 'EPSG', 5516, 'PROJCS[\"S-JTSK/05 / Modifie"
  "d Krovak East North\",GEOGCS[\"S-JTSK/05\",DATUM[\"System of the Unified Trigonometrical Cadastral Network/05\",SPHER"
  "OID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[572.213,85.334,461.94,4.9732,1.529,"
  "5.2484,3.5378],AUTHORITY[\"EPSG\",\"1052\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5228\"]],"
  "PROJECTION[\"Krovak Modified (North Orientated)\",AUTHORITY[\"EPSG\",\"1043\"]],PARAMETER[\"Latitude of projection ce"
  "ntre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of origin\",24.8333333333333,AUTHORITY[\""
  "EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone axis\",30.2881397222222,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"L"
  "atitude of pseudo standard parallel\",78.5111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseu"
  "do standard parallel\",0.9999,AUTHORITY[\"EPSG\",\"8819\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",5000000,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Ordinate 1 of evaluation poin"
  "t\",1089000,AUTHORITY[\"EPSG\",\"8617\"]],PARAMETER[\"Ordinate 2 of evaluation point\",654000,AUTHORITY[\"EPSG\",\"86"
  "18\"]],PARAMETER[\"C1\",0.02946529277,AUTHORITY[\"EPSG\",\"1026\"]],PARAMETER[\"C2\",0.02515965696,AUTHORITY[\"EPSG\""
  ",\"1027\"]],PARAMETER[\"C3\",1.193845912e-07,AUTHORITY[\"EPSG\",\"1028\"]],PARAMETER[\"C4\",-4.668270147e-07,AUTHORIT"
  "Y[\"EPSG\",\"1029\"]],PARAMETER[\"C5\",9.233980362e-12,AUTHORITY[\"EPSG\",\"1030\"]],PARAMETER[\"C6\",1.523735715e-12"
  ",AUTHORITY[\"EPSG\",\"1031\"]],PARAMETER[\"C7\",1.696780024e-18,AUTHORITY[\"EPSG\",\"1032\"]],PARAMETER[\"C8\",4.4083"
  "14235e-18,AUTHORITY[\"EPSG\",\"1033\"]],PARAMETER[\"C9\",-8.331083518e-24,AUTHORITY[\"EPSG\",\"1034\"]],PARAMETER[\"C"
  "10\",-3.689471323e-24,AUTHORITY[\"EPSG\",\"1035\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5516\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5518, 1, 'CI1971 / Chatham Islands Map Grid', 'EPSG', 5518, 'PROJCS[\"CI1971 / Chatham Islands"
  " Map Grid\",GEOGCS[\"Chatham Islands 1971\",DATUM[\"Chatham Islands Datum 1971\",SPHEROID[\"International 1924\",6378"
  "388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[175,-38,113,0,0,0,0],AUTHORITY[\"EPSG\",\"6672\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4672\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",-44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-176.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",350000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",650000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"551"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5519, 1, 'CI1979 / Chatham Islands Map Grid', 'EPSG', 5519, 'PROJCS[\"CI1979 / Chatham Islands"
  " Map Grid\",GEOGCS[\"Chatham Islands 1979\",DATUM[\"Chatham Islands Datum 1979\",SPHEROID[\"International 1924\",6378"
  "388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[174.05,-25.49,112.57,0,0,0.554,0.2263],AUTHORITY[\"EPSG\",\"6673\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4673\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-176.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",350000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",650000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"5519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5520, 1, 'DHDN / 3-degree Gauss-Kruger zone 1', 'EPSG', 5520, 'PROJCS[\"DHDN / 3-degree Gauss-"
  "Kruger zone 1\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.155,299.152812"
  "8,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"6314\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"55"
  "20\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5523, 1, 'WGS 84 / Gabon TM 2011', 'EPSG', 5523, 'PROJCS[\"WGS 84 / Gabon TM 2011\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",11.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",5500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5524, 1, 'Corrego Alegre 1961', 'EPSG', 5524, 'GEOGCS[\"Corrego Alegre 1961\",DATUM[\"Corrego "
  "Alegre 1961\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1074\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5524\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5527, 1, 'SAD69(96)', 'EPSG', 5527, 'GEOGCS[\"SAD69(96)\",DATUM[\"South American Datum 1969(96"
  ")\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5527\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5530, 1, 'SAD69(96) / Brazil Polyconic', 'EPSG', 5530, 'PROJCS[\"SAD69(96) / Brazil Polyconic\""
  ",GEOGCS[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY"
  "[\"EPSG\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"American Polyconic\",AUTHORITY[\"EPSG\",\"9818\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10"
  "000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"5530\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5531, 1, 'SAD69(96) / UTM zone 21S', 'EPSG', 5531, 'PROJCS[\"SAD69(96) / UTM zone 21S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5533, 1, 'SAD69(96) / UTM zone 23S', 'EPSG', 5533, 'PROJCS[\"SAD69(96) / UTM zone 23S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5534, 1, 'SAD69(96) / UTM zone 24S', 'EPSG', 5534, 'PROJCS[\"SAD69(96) / UTM zone 24S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5534\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5535, 1, 'SAD69(96) / UTM zone 25S', 'EPSG', 5535, 'PROJCS[\"SAD69(96) / UTM zone 25S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5536, 1, 'Corrego Alegre 1961 / UTM zone 21S', 'EPSG', 5536, 'PROJCS[\"Corrego Alegre 1961 / U"
  "TM zone 21S\",GEOGCS[\"Corrego Alegre 1961\",DATUM[\"Corrego Alegre 1961\",SPHEROID[\"International 1924\",6378388,29"
  "7,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1074\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5524\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5536\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5537, 1, 'Corrego Alegre 1961 / UTM zone 22S', 'EPSG', 5537, 'PROJCS[\"Corrego Alegre 1961 / U"
  "TM zone 22S\",GEOGCS[\"Corrego Alegre 1961\",DATUM[\"Corrego Alegre 1961\",SPHEROID[\"International 1924\",6378388,29"
  "7,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1074\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5524\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5537\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5538, 1, 'Corrego Alegre 1961 / UTM zone 23S', 'EPSG', 5538, 'PROJCS[\"Corrego Alegre 1961 / U"
  "TM zone 23S\",GEOGCS[\"Corrego Alegre 1961\",DATUM[\"Corrego Alegre 1961\",SPHEROID[\"International 1924\",6378388,29"
  "7,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1074\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5524\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5539, 1, 'Corrego Alegre 1961 / UTM zone 24S', 'EPSG', 5539, 'PROJCS[\"Corrego Alegre 1961 / U"
  "TM zone 24S\",GEOGCS[\"Corrego Alegre 1961\",DATUM[\"Corrego Alegre 1961\",SPHEROID[\"International 1924\",6378388,29"
  "7,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"1074\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"5524\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5546, 1, 'PNG94', 'EPSG', 5546, 'GEOGCS[\"PNG94\",DATUM[\"Papua New Guinea Geodetic Datum 1994"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"1076\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5546\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5550, 1, 'PNG94 / PNGMG94 zone 54', 'EPSG', 5550, 'PROJCS[\"PNG94 / PNGMG94 zone 54\",GEOGCS[\""
  "PNG94\",DATUM[\"Papua New Guinea Geodetic Datum 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1076\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"5546\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5550\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5551, 1, 'PNG94 / PNGMG94 zone 55', 'EPSG', 5551, 'PROJCS[\"PNG94 / PNGMG94 zone 55\",GEOGCS[\""
  "PNG94\",DATUM[\"Papua New Guinea Geodetic Datum 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1076\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"5546\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5551\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5552, 1, 'PNG94 / PNGMG94 zone 56', 'EPSG', 5552, 'PROJCS[\"PNG94 / PNGMG94 zone 56\",GEOGCS[\""
  "PNG94\",DATUM[\"Papua New Guinea Geodetic Datum 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1076\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"5546\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5552\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5559, 1, 'Ocotepeque 1935 / Guatemala Norte', 'EPSG', 5559, 'PROJCS[\"Ocotepeque 1935 / Guatem"
  "ala Norte\",GEOGCS[\"Ocotepeque 1935\",DATUM[\"Ocotepeque 1935\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982"
  ",AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[205,96,-98,0,0,0,0],AUTHORITY[\"EPSG\",\"1070\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5451\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",16.8277777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-90.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999222"
  "6,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",292209.579,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5559\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5561, 1, 'UCS-2000', 'EPSG', 5561, 'GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Kras"
  "sowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\""
  "1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5561\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5562, 1, 'UCS-2000 / Gauss-Kruger zone 4', 'EPSG', 5562, 'PROJCS[\"UCS-2000 / Gauss-Kruger zon"
  "e 4\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",45000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5562\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5563, 1, 'UCS-2000 / Gauss-Kruger zone 5', 'EPSG', 5563, 'PROJCS[\"UCS-2000 / Gauss-Kruger zon"
  "e 5\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",55000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5563\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5564, 1, 'UCS-2000 / Gauss-Kruger zone 6', 'EPSG', 5564, 'PROJCS[\"UCS-2000 / Gauss-Kruger zon"
  "e 6\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",65000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5564\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5565, 1, 'UCS-2000 / Gauss-Kruger zone 7', 'EPSG', 5565, 'PROJCS[\"UCS-2000 / Gauss-Kruger zon"
  "e 7\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",75000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5565\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5566, 1, 'UCS-2000 / Gauss-Kruger CM 21E', 'EPSG', 5566, 'PROJCS[\"UCS-2000 / Gauss-Kruger CM "
  "21E\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5566\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5567, 1, 'UCS-2000 / Gauss-Kruger CM 27E', 'EPSG', 5567, 'PROJCS[\"UCS-2000 / Gauss-Kruger CM "
  "27E\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5567\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5568, 1, 'UCS-2000 / Gauss-Kruger CM 33E', 'EPSG', 5568, 'PROJCS[\"UCS-2000 / Gauss-Kruger CM "
  "33E\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5568\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5569, 1, 'UCS-2000 / Gauss-Kruger CM 39E', 'EPSG', 5569, 'PROJCS[\"UCS-2000 / Gauss-Kruger CM "
  "39E\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024"
  "\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"5569\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5588, 1, 'NAD27 / New Brunswick Stereographic (NAD27)', 'EPSG', 5588, 'PROJCS[\"NAD27 / New Br"
  "unswick Stereographic (NAD27)\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206"
  ".4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,"
  "0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"426"
  "7\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"9809\"]],PARAMETER[\"Latitude of natural origin\",46."
  "5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-66.5111111111111,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999912,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5588\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5589, 1, 'Sibun Gorge 1922 / Colony Grid', 'EPSG', 5589, 'PROJCS[\"Sibun Gorge 1922 / Colony G"
  "rid\",GEOGCS[\"Sibun Gorge 1922\",DATUM[\"Sibun Gorge 1922\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.260676369"
  "26103,AUTHORITY[\"EPSG\",\"7007\"]],AUTHORITY[\"EPSG\",\"1071\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"5464\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",17.0612419444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.6318575,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",217259.26,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",445474.83,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"Clarke\\'s foot\",0.3047972654,AUTHORITY[\"EPSG\",\"9005\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\""
  "EPSG\",\"5589\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5593, 1, 'FEH2010', 'EPSG', 5593, 'GEOGCS[\"FEH2010\",DATUM[\"Fehmarnbelt Datum 2010\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1078\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5593\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5596, 1, 'FEH2010 / Fehmarnbelt TM', 'EPSG', 5596, 'PROJCS[\"FEH2010 / Fehmarnbelt TM\",GEOGCS"
  "[\"FEH2010\",DATUM[\"Fehmarnbelt Datum 2010\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1078\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"5593\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11.3444444444444,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5596\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5623, 1, 'NAD27 / Michigan East', 'EPSG', 5623, 'PROJCS[\"NAD27 / Michigan East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-83.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.999942857,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5623\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5624, 1, 'NAD27 / Michigan Old Central', 'EPSG', 5624, 'PROJCS[\"NAD27 / Michigan Old Central\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.5111111111111,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5624\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5625, 1, 'NAD27 / Michigan West', 'EPSG', 5625, 'PROJCS[\"NAD27 / Michigan West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-88.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5625\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5627, 1, 'ED50 / TM 6 NE', 'EPSG', 5627, 'PROJCS[\"ED50 / TM 6 NE\",GEOGCS[\"ED50\",DATUM[\"Eu"
  "ropean Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,"
  "-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",6,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5627\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5629, 1, 'Moznet / UTM zone 38S', 'EPSG', 5629, 'PROJCS[\"Moznet / UTM zone 38S\",GEOGCS[\"Moz"
  "net\",DATUM[\"Moznet (ITRF94)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0"
  ",0,0,0,0],AUTHORITY[\"EPSG\",\"6130\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453"
  "292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4130\"]],PROJE"
  "CTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]]"
  ",AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5629\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5631, 1, 'Pulkovo 1942(58) / Gauss-Kruger zone 2 (E-N)', 'EPSG', 5631, 'PROJCS[\"Pulkovo 1942("
  "58) / Gauss-Kruger zone 2 (E-N)\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\""
  ",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\","
  "\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"5631\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5632, 1, 'PTRA08 / LCC Europe', 'EPSG', 5632, 'PROJCS[\"PTRA08 / LCC Europe\",GEOGCS[\"PTRA08\""
  ",DATUM[\"Autonomous Regions of Portugal 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"5013\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false o"
  "rigin\",52,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10,AUTHORITY[\"EPSG\",\"8822\"]],PAR"
  "AMETER[\"Latitude of 1st standard parallel\",35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard pa"
  "rallel\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]]"
  ",PARAMETER[\"Northing at false origin\",2800000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5632\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5633, 1, 'PTRA08 / LAEA Europe', 'EPSG', 5633, 'PROJCS[\"PTRA08 / LAEA Europe\",GEOGCS[\"PTRA0"
  "8\",DATUM[\"Autonomous Regions of Portugal 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1041\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"5013\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",52,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"False easting\",4321000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3210000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST],AUTHORITY[\"EPSG\""
  ",\"5633\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5634, 1, 'REGCAN95 / LCC Europe', 'EPSG', 5634, 'PROJCS[\"REGCAN95 / LCC Europe\",GEOGCS[\"REG"
  "CAN95\",DATUM[\"Red Geodesica de Canarias 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1035\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4081\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fals"
  "e origin\",52,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10,AUTHORITY[\"EPSG\",\"8822\"]],"
  "PARAMETER[\"Latitude of 1st standard parallel\",35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard"
  " parallel\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",2800000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5634\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5635, 1, 'REGCAN95 / LAEA Europe', 'EPSG', 5635, 'PROJCS[\"REGCAN95 / LAEA Europe\",GEOGCS[\"R"
  "EGCAN95\",DATUM[\"Red Geodesica de Canarias 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1035\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"4081\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",52,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"False easting\",4321000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3210000,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EP"
  "SG\",\"5635\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5636, 1, 'TUREF / LAEA Europe', 'EPSG', 5636, 'PROJCS[\"TUREF / LAEA Europe\",GEOGCS[\"TUREF\""
  ",DATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "5252\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural or"
  "igin\",52,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"False easting\",4321000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3210000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST],AUTHORITY[\"EPSG\",\""
  "5636\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5637, 1, 'TUREF / LCC Europe', 'EPSG', 5637, 'PROJCS[\"TUREF / LCC Europe\",GEOGCS[\"TUREF\",D"
  "ATUM[\"Turkish National Reference Frame\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1057\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"52"
  "52\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",52,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10,AUTHORITY[\"EPSG\",\"8822\"]],PARAMET"
  "ER[\"Latitude of 1st standard parallel\",35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",2800000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5637\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5638, 1, 'ISN2004 / LAEA Europe', 'EPSG', 5638, 'PROJCS[\"ISN2004 / LAEA Europe\",GEOGCS[\"ISN"
  "2004\",DATUM[\"Islands Net 2004\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0"
  ",0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1060\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5324\"]],P"
  "ROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural origin\",52,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "False easting\",4321000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3210000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST],AUTHORITY[\"EPSG\",\"5638\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5639, 1, 'ISN2004 / LCC Europe', 'EPSG', 5639, 'PROJCS[\"ISN2004 / LCC Europe\",GEOGCS[\"ISN20"
  "04\",DATUM[\"Islands Net 2004\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0"
  ",0,0,0,0,0],AUTHORITY[\"EPSG\",\"1060\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5324\"]],PRO"
  "JECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",52,AUT"
  "HORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latit"
  "ude of 1st standard parallel\",35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",65,AU"
  "THORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"N"
  "orthing at false origin\",2800000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5639\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5641, 1, 'SIRGAS 2000 / Brazil Mercator', 'EPSG', 5641, 'PROJCS[\"SIRGAS 2000 / Brazil Mercato"
  "r\",GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9"
  "805\"]],PARAMETER[\"Latitude of 1st standard parallel\",-2,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",-43,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5641\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5643, 1, 'ED50 / SPBA LCC', 'EPSG', 5643, 'PROJCS[\"ED50 / SPBA LCC\",GEOGCS[\"ED50\",DATUM[\""
  "European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.1"
  "6,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4230\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",48,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",10,AUTHORITY[\"EPSG\",\"882"
  "2\"]],PARAMETER[\"Latitude of 1st standard parallel\",52.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",54.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\","
  "815000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5643\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5644, 1, 'RGR92 / UTM zone 39S', 'EPSG', 5644, 'PROJCS[\"RGR92 / UTM zone 39S\",GEOGCS[\"RGR92"
  "\",DATUM[\"Reseau Geodesique de la Reunion 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6627\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4627\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5644\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5646, 1, 'NAD83 / Vermont (ftUS)', 'EPSG', 5646, 'PROJCS[\"NAD83 / Vermont (ftUS)\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"5646\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5649, 1, 'ETRS89 / UTM zone 31N (zE-N)', 'EPSG', 5649, 'PROJCS[\"ETRS89 / UTM zone 31N (zE-N)\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5649\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5650, 1, 'ETRS89 / UTM zone 33N (zE-N)', 'EPSG', 5650, 'PROJCS[\"ETRS89 / UTM zone 33N (zE-N)\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",33500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5650\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5651, 1, 'ETRS89 / UTM zone 31N (N-zE)', 'EPSG', 5651, 'PROJCS[\"ETRS89 / UTM zone 31N (N-zE)\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5651\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5652, 1, 'ETRS89 / UTM zone 32N (N-zE)', 'EPSG', 5652, 'PROJCS[\"ETRS89 / UTM zone 32N (N-zE)\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5652\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5653, 1, 'ETRS89 / UTM zone 33N (N-zE)', 'EPSG', 5653, 'PROJCS[\"ETRS89 / UTM zone 33N (N-zE)\""
  ",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,A"
  "UTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",33500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"5653\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5654, 1, 'NAD83(HARN) / Vermont (ftUS)', 'EPSG', 5654, 'PROJCS[\"NAD83(HARN) / Vermont (ftUS)\""
  ",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-72.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5654\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5655, 1, 'NAD83(NSRS2007) / Vermont (ftUS)', 'EPSG', 5655, 'PROJCS[\"NAD83(NSRS2007) / Vermont"
  " (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude "
  "of natural origin\",-72.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "9964286,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5655\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5659, 1, 'Monte Mario / TM Emilia-Romagna', 'EPSG', 5659, 'PROJCS[\"Monte Mario / TM Emilia-Ro"
  "magna\",GEOGCS[\"Monte Mario\",DATUM[\"Monte Mario\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-104.1,-49.1,-9.9,0.971,-2.917,0.714,-11.68],AUTHORITY[\"EPSG\",\"6265\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4265\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500053,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-3999820,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5659\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5663, 1, 'Pulkovo 1942(58) / Gauss-Kruger zone 3 (E-N)', 'EPSG', 5663, 'PROJCS[\"Pulkovo 1942("
  "58) / Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\"Krassowsky 1940\""
  ",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AUTHORITY[\"EPSG\","
  "\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"5663\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5664, 1, 'Pulkovo 1942(83) / Gauss-Kruger zone 2 (E-N)', 'EPSG', 5664, 'PROJCS[\"Pulkovo 1942("
  "83) / Gauss-Kruger zone 2 (E-N)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky 1940\""
  ",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"6178\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY["
  "\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"5664\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5665, 1, 'Pulkovo 1942(83) / Gauss-Kruger zone 3 (E-N)', 'EPSG', 5665, 'PROJCS[\"Pulkovo 1942("
  "83) / Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\"Krassowsky 1940\""
  ",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"EPSG\",\"6178\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY["
  "\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"5665\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5666, 1, 'PD/83 / 3-degree Gauss-Kruger zone 3 (E-N)', 'EPSG', 5666, 'PROJCS[\"PD/83 / 3-degre"
  "e Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"PD/83\",DATUM[\"Potsdam Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299.15"
  "28128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4746\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHO"
  "RITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5666\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5667, 1, 'PD/83 / 3-degree Gauss-Kruger zone 4 (E-N)', 'EPSG', 5667, 'PROJCS[\"PD/83 / 3-degre"
  "e Gauss-Kruger zone 4 (E-N)\",GEOGCS[\"PD/83\",DATUM[\"Potsdam Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299.15"
  "28128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6746\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4746\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5667\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5668, 1, 'RD/83 / 3-degree Gauss-Kruger zone 4 (E-N)', 'EPSG', 5668, 'PROJCS[\"RD/83 / 3-degre"
  "e Gauss-Kruger zone 4 (E-N)\",GEOGCS[\"RD/83\",DATUM[\"Rauenberg Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299."
  "1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4745\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5668\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5669, 1, 'RD/83 / 3-degree Gauss-Kruger zone 5 (E-N)', 'EPSG', 5669, 'PROJCS[\"RD/83 / 3-degre"
  "e Gauss-Kruger zone 5 (E-N)\",GEOGCS[\"RD/83\",DATUM[\"Rauenberg Datum/83\",SPHEROID[\"Bessel 1841\",6377397.155,299."
  "1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6745\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4745\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5669\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5670, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 3 (E-N)', 'EPSG', 5670, 'PROJCS[\"Pulk"
  "ovo 1942(58) / 3-degree Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AU"
  "THORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"5670\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5671, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 4 (E-N)', 'EPSG', 5671, 'PROJCS[\"Pulk"
  "ovo 1942(58) / 3-degree Gauss-Kruger zone 4 (E-N)\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AU"
  "THORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS["
  "\"N\",NORTH],AUTHORITY[\"EPSG\",\"5671\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5672, 1, 'Pulkovo 1942(58) / 3-degree Gauss-Kruger zone 5 (E-N)', 'EPSG', 5672, 'PROJCS[\"Pulk"
  "ovo 1942(58) / 3-degree Gauss-Kruger zone 5 (E-N)\",GEOGCS[\"Pulkovo 1942(58)\",DATUM[\"Pulkovo 1942(58)\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[33.4,-146.6,-76.3,-0.359,-0.053,0.844,-0.84],AU"
  "THORITY[\"EPSG\",\"6179\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4179\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS["
  "\"N\",NORTH],AUTHORITY[\"EPSG\",\"5672\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5673, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 3 (E-N)', 'EPSG', 5673, 'PROJCS[\"Pulk"
  "ovo 1942(83) / 3-degree Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"E"
  "PSG\",\"6178\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH"
  "],AUTHORITY[\"EPSG\",\"5673\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5674, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 4 (E-N)', 'EPSG', 5674, 'PROJCS[\"Pulk"
  "ovo 1942(83) / 3-degree Gauss-Kruger zone 4 (E-N)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"E"
  "PSG\",\"6178\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORT"
  "H],AUTHORITY[\"EPSG\",\"5674\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5675, 1, 'Pulkovo 1942(83) / 3-degree Gauss-Kruger zone 5 (E-N)', 'EPSG', 5675, 'PROJCS[\"Pulk"
  "ovo 1942(83) / 3-degree Gauss-Kruger zone 5 (E-N)\",GEOGCS[\"Pulkovo 1942(83)\",DATUM[\"Pulkovo 1942(83)\",SPHEROID[\""
  "Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[24,-123,-94,0.02,-0.25,-0.13,1.1],AUTHORITY[\"E"
  "PSG\",\"6178\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4178\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORT"
  "H],AUTHORITY[\"EPSG\",\"5675\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5676, 1, 'DHDN / 3-degree Gauss-Kruger zone 2 (E-N)', 'EPSG', 5676, 'PROJCS[\"DHDN / 3-degree "
  "Gauss-Kruger zone 2 (E-N)\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"631"
  "4\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",6,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY"
  "[\"EPSG\",\"5676\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5677, 1, 'DHDN / 3-degree Gauss-Kruger zone 3 (E-N)', 'EPSG', 5677, 'PROJCS[\"DHDN / 3-degree "
  "Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"631"
  "4\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY"
  "[\"EPSG\",\"5677\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5678, 1, 'DHDN / 3-degree Gauss-Kruger zone 4 (E-N)', 'EPSG', 5678, 'PROJCS[\"DHDN / 3-degree "
  "Gauss-Kruger zone 4 (E-N)\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"631"
  "4\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"5678\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5679, 1, 'DHDN / 3-degree Gauss-Kruger zone 5 (E-N)', 'EPSG', 5679, 'PROJCS[\"DHDN / 3-degree "
  "Gauss-Kruger zone 5 (E-N)\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"631"
  "4\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"5679\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5680, 1, 'DHDN / 3-degree Gauss-Kruger zone 1 (E-N)', 'EPSG', 5680, 'PROJCS[\"DHDN / 3-degree "
  "Gauss-Kruger zone 1 (E-N)\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"631"
  "4\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY"
  "[\"EPSG\",\"5680\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5681, 1, 'DB_REF', 'EPSG', 5681, 'GEOGCS[\"DB_REF\",DATUM[\"Deutsche Bahn Reference System\",S"
  "PHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1081\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5681\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5682, 1, 'DB_REF / 3-degree Gauss-Kruger zone 2 (E-N)', 'EPSG', 5682, 'PROJCS[\"DB_REF / 3-deg"
  "ree Gauss-Kruger zone 2 (E-N)\",GEOGCS[\"DB_REF\",DATUM[\"Deutsche Bahn Reference System\",SPHEROID[\"Bessel 1841\",6"
  "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1081\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5681\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",6,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5682\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5683, 1, 'DB_REF / 3-degree Gauss-Kruger zone 3 (E-N)', 'EPSG', 5683, 'PROJCS[\"DB_REF / 3-deg"
  "ree Gauss-Kruger zone 3 (E-N)\",GEOGCS[\"DB_REF\",DATUM[\"Deutsche Bahn Reference System\",SPHEROID[\"Bessel 1841\",6"
  "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1081\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5681\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5683\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5684, 1, 'DB_REF / 3-degree Gauss-Kruger zone 4 (E-N)', 'EPSG', 5684, 'PROJCS[\"DB_REF / 3-deg"
  "ree Gauss-Kruger zone 4 (E-N)\",GEOGCS[\"DB_REF\",DATUM[\"Deutsche Bahn Reference System\",SPHEROID[\"Bessel 1841\",6"
  "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1081\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5681\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5684\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5685, 1, 'DB_REF / 3-degree Gauss-Kruger zone 5 (E-N)', 'EPSG', 5685, 'PROJCS[\"DB_REF / 3-deg"
  "ree Gauss-Kruger zone 5 (E-N)\",GEOGCS[\"DB_REF\",DATUM[\"Deutsche Bahn Reference System\",SPHEROID[\"Bessel 1841\",6"
  "377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"1081\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"5681\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5685\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5700, 1, 'NZGD2000 / UTM zone 1S', 'EPSG', 5700, 'PROJCS[\"NZGD2000 / UTM zone 1S\",GEOGCS[\"N"
  "ZGD2000\",DATUM[\"New Zealand Geodetic Datum 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4167\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5825, 1, 'AGD66 / ACT Standard Grid', 'EPSG', 5825, 'PROJCS[\"AGD66 / ACT Standard Grid\",GEOG"
  "CS[\"AGD66\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHOR"
  "ITY[\"EPSG\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-35.3177362777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",149.009294830556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",1.000086,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5825\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5836, 1, 'Yemen NGN96 / UTM zone 37N', 'EPSG', 5836, 'PROJCS[\"Yemen NGN96 / UTM zone 37N\",GE"
  "OGCS[\"Yemen NGN96\",DATUM[\"Yemen National Geodetic Network 1996\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORI"
  "TY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4163\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5836\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5837, 1, 'Yemen NGN96 / UTM zone 40N', 'EPSG', 5837, 'PROJCS[\"Yemen NGN96 / UTM zone 40N\",GE"
  "OGCS[\"Yemen NGN96\",DATUM[\"Yemen National Geodetic Network 1996\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORI"
  "TY[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6163\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4163\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5839, 1, 'Peru96 / UTM zone 17S', 'EPSG', 5839, 'PROJCS[\"Peru96 / UTM zone 17S\",GEOGCS[\"Per"
  "u96\",DATUM[\"Peru96\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"1067\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5373\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5842, 1, 'WGS 84 / TM 12 SE', 'EPSG', 5842, 'PROJCS[\"WGS 84 / TM 12 SE\",GEOGCS[\"WGS 84\",DA"
  "TUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY["
  "\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0"
  ".9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],A"
  "XIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5842\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5844, 1, 'RGRDC 2005 / Congo TM zone 30', 'EPSG', 5844, 'PROJCS[\"RGRDC 2005 / Congo TM zone 3"
  "0\",GEOGCS[\"RGRDC 2005\",DATUM[\"Reseau Geodesique de la RDC 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1033\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4046\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5844\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5858, 1, 'SAD69(96) / UTM zone 22S', 'EPSG', 5858, 'PROJCS[\"SAD69(96) / UTM zone 22S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5858\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5875, 1, 'SAD69(96) / UTM zone 18S', 'EPSG', 5875, 'PROJCS[\"SAD69(96) / UTM zone 18S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5875\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5876, 1, 'SAD69(96) / UTM zone 19S', 'EPSG', 5876, 'PROJCS[\"SAD69(96) / UTM zone 19S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5876\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5877, 1, 'SAD69(96) / UTM zone 20S', 'EPSG', 5877, 'PROJCS[\"SAD69(96) / UTM zone 20S\",GEOGCS"
  "[\"SAD69(96)\",DATUM[\"South American Datum 1969(96)\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7050\"]],TOWGS84[-67.35,3.88,-38.22,0,0,0,0],AUTHORITY[\"EPSG\",\"1075\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"5527\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5877\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5879, 1, 'Cadastre 1997 / UTM zone 38S', 'EPSG', 5879, 'PROJCS[\"Cadastre 1997 / UTM zone 38S\""
  ",GEOGCS[\"Cadastre 1997\",DATUM[\"Cadastre 1997\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[-381.788,-57.501,-256.673,0,0,0,0],AUTHORITY[\"EPSG\",\"1037\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4475\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5879\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5880, 1, 'SIRGAS 2000 / Brazil Polyconic', 'EPSG', 5880, 'PROJCS[\"SIRGAS 2000 / Brazil Polyco"
  "nic\",GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"American Polyconic\",AUTHORITY[\"EPSG\",\"9"
  "818\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-54,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"5880\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5886, 1, 'TGD2005', 'EPSG', 5886, 'GEOGCS[\"TGD2005\",DATUM[\"Tonga Geodetic Datum 2005\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1095\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5886\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5887, 1, 'TGD2005 / Tonga Map Grid', 'EPSG', 5887, 'PROJCS[\"TGD2005 / Tonga Map Grid\",GEOGCS"
  "[\"TGD2005\",DATUM[\"Tonga Geodetic Datum 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],AUTHORITY[\"EPSG\",\"1095\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"5886\"]],PROJECTIO"
  "N[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at "
  "natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5887\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5890, 1, 'JAXA Snow Depth Polar Stereographic North', 'EPSG', 5890, 'PROJCS[\"JAXA Snow Depth "
  "Polar Stereographic North\",GEOGCS[\"Unspecified datum based upon the Hughes 1980 ellipsoid\",DATUM[\"Not specified ("
  "based on Hughes 1980 ellipsoid)\",SPHEROID[\"Hughes 1980\",6378273,298.279411123064,AUTHORITY[\"EPSG\",\"7058\"]],AUT"
  "HORITY[\"EPSG\",\"6054\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4054\"]],PROJECTION[\"Polar"
  " Stereographic (variant B)\",AUTHORITY[\"EPSG\",\"9829\"]],PARAMETER[\"Latitude of standard parallel\",70,AUTHORITY[\""
  "EPSG\",\"8832\"]],PARAMETER[\"Longitude of origin\",90,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"False easting\",0,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"5890\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5896, 1, 'VN-2000 / TM-3 zone 481', 'EPSG', 5896, 'PROJCS[\"VN-2000 / TM-3 zone 481\",GEOGCS[\""
  "VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-191"
  ".90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"6756\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\""
  "EPSG\",\"5896\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5897, 1, 'VN-2000 / TM-3 zone 482', 'EPSG', 5897, 'PROJCS[\"VN-2000 / TM-3 zone 482\",GEOGCS[\""
  "VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-191"
  ".90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"6756\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\""
  "EPSG\",\"5897\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5898, 1, 'VN-2000 / TM-3 zone 491', 'EPSG', 5898, 'PROJCS[\"VN-2000 / TM-3 zone 491\",GEOGCS[\""
  "VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-191"
  ".90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"6756\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\""
  "EPSG\",\"5898\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5899, 1, 'VN-2000 / TM-3 Da Nang zone', 'EPSG', 5899, 'PROJCS[\"VN-2000 / TM-3 Da Nang zone\","
  "GEOGCS[\"VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOW"
  "GS84[-191.90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EPSG\",\"675"
  "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",107.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"5899\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5921, 1, 'WGS 84 / EPSG Arctic Regional zone A1', 'EPSG', 5921, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone A1\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",81.317226,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",85,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd"
  " standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5921\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5922, 1, 'WGS 84 / EPSG Arctic Regional zone A2', 'EPSG', 5922, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone A2\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",81.317226,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-39,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",85,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd"
  " standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5922\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5923, 1, 'WGS 84 / EPSG Arctic Regional zone A3', 'EPSG', 5923, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone A3\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",81.317226,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",33,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",85,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5923\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5924, 1, 'WGS 84 / EPSG Arctic Regional zone A4', 'EPSG', 5924, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone A4\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",81.317226,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",105,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",85,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd"
  " standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5924\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5925, 1, 'WGS 84 / EPSG Arctic Regional zone A5', 'EPSG', 5925, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone A5\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",81.317226,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",177,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",85,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd"
  " standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5925\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5926, 1, 'WGS 84 / EPSG Arctic Regional zone B1', 'EPSG', 5926, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone B1\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",73.1557408611111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111,AUTHORITY[\"E"
  "PSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",69,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5926\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5927, 1, 'WGS 84 / EPSG Arctic Regional zone B2', 'EPSG', 5927, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone B2\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",73.1557408611111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-39,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",69,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5927\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5928, 1, 'WGS 84 / EPSG Arctic Regional zone B3', 'EPSG', 5928, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone B3\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",73.1557408611111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",33,AUTHORITY[\"EPS"
  "G\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",69,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5928\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5929, 1, 'WGS 84 / EPSG Arctic Regional zone B4', 'EPSG', 5929, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone B4\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",73.1557408611111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",105,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",69,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5929\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5930, 1, 'WGS 84 / EPSG Arctic Regional zone B5', 'EPSG', 5930, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone B5\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",73.1557408611111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",177,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",69,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5930\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5931, 1, 'WGS 84 / EPSG Arctic Regional zone C1', 'EPSG', 5931, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone C1\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",65.1012708888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111,AUTHORITY[\"E"
  "PSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",69,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",61,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5931\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5932, 1, 'WGS 84 / EPSG Arctic Regional zone C2', 'EPSG', 5932, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone C2\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",65.1012708888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-39,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",69,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",61,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5932\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5933, 1, 'WGS 84 / EPSG Arctic Regional zone C3', 'EPSG', 5933, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone C3\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",65.1012708888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",33,AUTHORITY[\"EPS"
  "G\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",69,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",61,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5933\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5934, 1, 'WGS 84 / EPSG Arctic Regional zone C4', 'EPSG', 5934, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone C4\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",65.1012708888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",105,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",69,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",61,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5934\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5935, 1, 'WGS 84 / EPSG Arctic Regional zone C5', 'EPSG', 5935, 'PROJCS[\"WGS 84 / EPSG Arctic"
  " Regional zone C5\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,"
  "AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4326\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",65.1012708888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",177,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",69,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",61,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"5935\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5936, 1, 'WGS 84 / EPSG Alaska Polar Stereographic', 'EPSG', 5936, 'PROJCS[\"WGS 84 / EPSG Ala"
  "ska Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.2572"
  "23563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitud"
  "e of natural origin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-150,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"5936\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5937, 1, 'WGS 84 / EPSG Canada Polar Stereographic', 'EPSG', 5937, 'PROJCS[\"WGS 84 / EPSG Can"
  "ada Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.2572"
  "23563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitud"
  "e of natural origin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-100,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"5937\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5938, 1, 'WGS 84 / EPSG Greenland Polar Stereographic', 'EPSG', 5938, 'PROJCS[\"WGS 84 / EPSG "
  "Greenland Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,29"
  "8.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"L"
  "atitude of natural origin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"5938\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5939, 1, 'WGS 84 / EPSG Norway Polar Stereographic', 'EPSG', 5939, 'PROJCS[\"WGS 84 / EPSG Nor"
  "way Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.2572"
  "23563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitud"
  "e of natural origin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"5939\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (5940, 1, 'WGS 84 / EPSG Russia Polar Stereographic', 'EPSG', 5940, 'PROJCS[\"WGS 84 / EPSG Rus"
  "sia Polar Stereographic\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.2572"
  "23563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4326\"]],PROJECTION[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitud"
  "e of natural origin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"5940\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6050, 1, 'GR96 / EPSG Arctic zone 1-25', 'EPSG', 6050, 'PROJCS[\"GR96 / EPSG Arctic zone 1-25\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",85.4371183333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-30,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",87,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",25500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6050\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6051, 1, 'GR96 / EPSG Arctic zone 2-18', 'EPSG', 6051, 'PROJCS[\"GR96 / EPSG Arctic zone 2-18\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",82.0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-52,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",18500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6051\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6052, 1, 'GR96 / EPSG Arctic zone 2-20', 'EPSG', 6052, 'PROJCS[\"GR96 / EPSG Arctic zone 2-20\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",82.0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",20500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6052\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6053, 1, 'GR96 / EPSG Arctic zone 3-29', 'EPSG', 6053, 'PROJCS[\"GR96 / EPSG Arctic zone 3-29\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-69,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",29500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6053\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6054, 1, 'GR96 / EPSG Arctic zone 3-31', 'EPSG', 6054, 'PROJCS[\"GR96 / EPSG Arctic zone 3-31\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-39,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",31500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6054\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6055, 1, 'GR96 / EPSG Arctic zone 3-33', 'EPSG', 6055, 'PROJCS[\"GR96 / EPSG Arctic zone 3-33\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-10,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",33500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6055\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6056, 1, 'GR96 / EPSG Arctic zone 4-20', 'EPSG', 6056, 'PROJCS[\"GR96 / EPSG Arctic zone 4-20\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-64,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",20500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6056\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6057, 1, 'GR96 / EPSG Arctic zone 4-22', 'EPSG', 6057, 'PROJCS[\"GR96 / EPSG Arctic zone 4-22\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-39,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",22500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6058, 1, 'GR96 / EPSG Arctic zone 4-24', 'EPSG', 6058, 'PROJCS[\"GR96 / EPSG Arctic zone 4-24\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-14,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",24500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6058\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6059, 1, 'GR96 / EPSG Arctic zone 5-41', 'EPSG', 6059, 'PROJCS[\"GR96 / EPSG Arctic zone 5-41\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-62,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",41500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6059\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6060, 1, 'GR96 / EPSG Arctic zone 5-43', 'EPSG', 6060, 'PROJCS[\"GR96 / EPSG Arctic zone 5-43\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-42,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",43500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6060\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6061, 1, 'GR96 / EPSG Arctic zone 5-45', 'EPSG', 6061, 'PROJCS[\"GR96 / EPSG Arctic zone 5-45\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-22,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",45500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6061\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6062, 1, 'GR96 / EPSG Arctic zone 6-26', 'EPSG', 6062, 'PROJCS[\"GR96 / EPSG Arctic zone 6-26\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-56,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",26500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6062\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6063, 1, 'GR96 / EPSG Arctic zone 6-28', 'EPSG', 6063, 'PROJCS[\"GR96 / EPSG Arctic zone 6-28\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-38,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",28500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6063\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6064, 1, 'GR96 / EPSG Arctic zone 6-30', 'EPSG', 6064, 'PROJCS[\"GR96 / EPSG Arctic zone 6-30\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-20,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",30500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6064\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6065, 1, 'GR96 / EPSG Arctic zone 7-11', 'EPSG', 6065, 'PROJCS[\"GR96 / EPSG Arctic zone 7-11\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",65.3510393055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-51,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",67,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",63.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",11500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",7500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6066, 1, 'GR96 / EPSG Arctic zone 7-13', 'EPSG', 6066, 'PROJCS[\"GR96 / EPSG Arctic zone 7-13\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",65.3510393055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-34,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",67,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd "
  "standard parallel\",63.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",13500000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",7500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6067, 1, 'GR96 / EPSG Arctic zone 8-20', 'EPSG', 6067, 'PROJCS[\"GR96 / EPSG Arctic zone 8-20\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",62.0153068888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-52,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",63.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",60.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",20500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",8500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6067\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6068, 1, 'GR96 / EPSG Arctic zone 8-22', 'EPSG', 6068, 'PROJCS[\"GR96 / EPSG Arctic zone 8-22\""
  ",GEOGCS[\"GR96\",DATUM[\"Greenland 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6747\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"47"
  "47\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",62.0153068888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-37,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",63.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",60.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",22500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",8500000,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6068\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6069, 1, 'ETRS89 / EPSG Arctic zone 2-22', 'EPSG', 6069, 'PROJCS[\"ETRS89 / EPSG Arctic zone 2"
  "-22\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",82.0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",16,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",22500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",250"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"6069\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6070, 1, 'ETRS89 / EPSG Arctic zone 3-11', 'EPSG', 6070, 'PROJCS[\"ETRS89 / EPSG Arctic zone 3"
  "-11\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",21,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",11500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"6070\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6071, 1, 'ETRS89 / EPSG Arctic zone 4-26', 'EPSG', 6071, 'PROJCS[\"ETRS89 / EPSG Arctic zone 4"
  "-26\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",10,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"882"
  "3\"]],PARAMETER[\"Latitude of 2nd standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",26500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"6071\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6072, 1, 'ETRS89 / EPSG Arctic zone 4-28', 'EPSG', 6072, 'PROJCS[\"ETRS89 / EPSG Arctic zone 4"
  "-28\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",34,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"882"
  "3\"]],PARAMETER[\"Latitude of 2nd standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",28500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EP"
  "SG\",\"6072\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6073, 1, 'ETRS89 / EPSG Arctic zone 5-11', 'EPSG', 6073, 'PROJCS[\"ETRS89 / EPSG Arctic zone 5"
  "-11\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",14,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",11500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",550"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"6073\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6074, 1, 'ETRS89 / EPSG Arctic zone 5-13', 'EPSG', 6074, 'PROJCS[\"ETRS89 / EPSG Arctic zone 5"
  "-13\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",34,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",13500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",550"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"6074\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6075, 1, 'WGS 84 / EPSG Arctic zone 2-24', 'EPSG', 6075, 'PROJCS[\"WGS 84 / EPSG Arctic zone 2"
  "-24\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",82"
  ".0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",53,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",24500"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6075\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6076, 1, 'WGS 84 / EPSG Arctic zone 2-26', 'EPSG', 6076, 'PROJCS[\"WGS 84 / EPSG Arctic zone 2"
  "-26\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",82"
  ".0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",93,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",26500"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6076\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6077, 1, 'WGS 84 / EPSG Arctic zone 3-13', 'EPSG', 6077, 'PROJCS[\"WGS 84 / EPSG Arctic zone 3"
  "-13\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",78"
  ".7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",52,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",13500000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6077\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6078, 1, 'WGS 84 / EPSG Arctic zone 3-15', 'EPSG', 6078, 'PROJCS[\"WGS 84 / EPSG Arctic zone 3"
  "-15\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",78"
  ".7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",83,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",15500000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6078\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6079, 1, 'WGS 84 / EPSG Arctic zone 3-17', 'EPSG', 6079, 'PROJCS[\"WGS 84 / EPSG Arctic zone 3"
  "-17\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",78"
  ".7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",114,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",17500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6079\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6080, 1, 'WGS 84 / EPSG Arctic zone 3-19', 'EPSG', 6080, 'PROJCS[\"WGS 84 / EPSG Arctic zone 3"
  "-19\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",78"
  ".7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",145,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",19500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6080\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6081, 1, 'WGS 84 / EPSG Arctic zone 4-30', 'EPSG', 6081, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-30\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",58,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standar"
  "d parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",30500000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6081\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6082, 1, 'WGS 84 / EPSG Arctic zone 4-32', 'EPSG', 6082, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-32\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",82,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standar"
  "d parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",32500000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6082\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6083, 1, 'WGS 84 / EPSG Arctic zone 4-34', 'EPSG', 6083, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-34\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",106,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",34500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6083\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6084, 1, 'WGS 84 / EPSG Arctic zone 4-36', 'EPSG', 6084, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-36\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",130,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",36500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6084\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6085, 1, 'WGS 84 / EPSG Arctic zone 4-38', 'EPSG', 6085, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-38\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",154,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",38500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6085\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6086, 1, 'WGS 84 / EPSG Arctic zone 4-40', 'EPSG', 6086, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-40\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",179,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",40500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6086\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6087, 1, 'WGS 84 / EPSG Arctic zone 5-15', 'EPSG', 6087, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-15\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",54,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",15500"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6087\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6088, 1, 'WGS 84 / EPSG Arctic zone 5-17', 'EPSG', 6088, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-17\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",74,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",17500"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6088\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6089, 1, 'WGS 84 / EPSG Arctic zone 5-19', 'EPSG', 6089, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-19\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",95,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",19500"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6089\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6090, 1, 'WGS 84 / EPSG Arctic zone 5-21', 'EPSG', 6090, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-21\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",116,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2150"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6090\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6091, 1, 'WGS 84 / EPSG Arctic zone 5-23', 'EPSG', 6091, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-23\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",137,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2350"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6091\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6092, 1, 'WGS 84 / EPSG Arctic zone 5-25', 'EPSG', 6092, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-25\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",158,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2550"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6092\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6093, 1, 'WGS 84 / EPSG Arctic zone 5-27', 'EPSG', 6093, 'PROJCS[\"WGS 84 / EPSG Arctic zone 5"
  "-27\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72"
  ".0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",179,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2750"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6093\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6094, 1, 'NAD83(NSRS2007) / EPSG Arctic zone 5-29', 'EPSG', 6094, 'PROJCS[\"NAD83(NSRS2007) / "
  "EPSG Arctic zone 5-29\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-163,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",29500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6094\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6095, 1, 'NAD83(NSRS2007) / EPSG Arctic zone 5-31', 'EPSG', 6095, 'PROJCS[\"NAD83(NSRS2007) / "
  "EPSG Arctic zone 5-31\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-147,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",31500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER["
  "\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6096, 1, 'NAD83(NSRS2007) / EPSG Arctic zone 6-14', 'EPSG', 6096, 'PROJCS[\"NAD83(NSRS2007) / "
  "EPSG Arctic zone 6-14\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-165,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",67,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",14500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],"
  "AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6096\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6097, 1, 'NAD83(NSRS2007) / EPSG Arctic zone 6-16', 'EPSG', 6097, 'PROJCS[\"NAD83(NSRS2007) / "
  "EPSG Arctic zone 6-16\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\","
  "AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],P"
  "ARAMETER[\"Longitude of false origin\",-147,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parall"
  "el\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",67,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",16500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],"
  "AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6097\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6098, 1, 'NAD83(CSRS) / EPSG Arctic zone 1-23', 'EPSG', 6098, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 1-23\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "85.4371183333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",87,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stand"
  "ard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",23500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6098\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6099, 1, 'NAD83(CSRS) / EPSG Arctic zone 2-14', 'EPSG', 6099, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 2-14\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "82.0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-115,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1"
  "4500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6099\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6100, 1, 'NAD83(CSRS) / EPSG Arctic zone 2-16', 'EPSG', 6100, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 2-16\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "82.0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-75,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",165"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6100\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6101, 1, 'NAD83(CSRS) / EPSG Arctic zone 3-25', 'EPSG', 6101, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 3-25\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-129,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",25500000,AUTHOR"
  "ITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6101\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6102, 1, 'NAD83(CSRS) / EPSG Arctic zone 3-27', 'EPSG', 6102, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 3-27\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-99,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",27500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6102\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6103, 1, 'NAD83(CSRS) / EPSG Arctic zone 3-29', 'EPSG', 6103, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 3-29\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "78.7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-69,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",29500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6103\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6104, 1, 'NAD83(CSRS) / EPSG Arctic zone 4-14', 'EPSG', 6104, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 4-14\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-129,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",14500000,AUTHOR"
  "ITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6104\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6105, 1, 'NAD83(CSRS) / EPSG Arctic zone 4-16', 'EPSG', 6105, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 4-16\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-104,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",16500000,AUTHOR"
  "ITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6105\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6106, 1, 'NAD83(CSRS) / EPSG Arctic zone 4-18', 'EPSG', 6106, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 4-18\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "75.3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stand"
  "ard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",18500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6106\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6107, 1, 'NAD83(CSRS) / EPSG Arctic zone 5-33', 'EPSG', 6107, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 5-33\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-131,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3"
  "3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6107\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6108, 1, 'NAD83(CSRS) / EPSG Arctic zone 5-35', 'EPSG', 6108, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 5-35\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3"
  "5500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6108\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6109, 1, 'NAD83(CSRS) / EPSG Arctic zone 5-37', 'EPSG', 6109, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 5-37\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",375"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6109\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6110, 1, 'NAD83(CSRS) / EPSG Arctic zone 5-39', 'EPSG', 6110, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 5-39\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-71,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",395"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6110\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6111, 1, 'NAD83(CSRS) / EPSG Arctic zone 6-18', 'EPSG', 6111, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 6-18\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-132,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",18500000,AUTHOR"
  "ITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6111\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6112, 1, 'NAD83(CSRS) / EPSG Arctic zone 6-20', 'EPSG', 6112, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 6-20\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-113,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",20500000,AUTHOR"
  "ITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6112\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6113, 1, 'NAD83(CSRS) / EPSG Arctic zone 6-22', 'EPSG', 6113, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 6-22\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",22500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6113\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6114, 1, 'NAD83(CSRS) / EPSG Arctic zone 6-24', 'EPSG', 6114, 'PROJCS[\"NAD83(CSRS) / EPSG Arc"
  "tic zone 6-24\",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0"
  "116599432323421,0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-75,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",24500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6114\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6115, 1, 'WGS 84 / EPSG Arctic zone 1-27', 'EPSG', 6115, 'PROJCS[\"WGS 84 / EPSG Arctic zone 1"
  "-27\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",85"
  ".4371183333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",30,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",87,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standar"
  "d parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",27500000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6115\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6116, 1, 'WGS 84 / EPSG Arctic zone 1-29', 'EPSG', 6116, 'PROJCS[\"WGS 84 / EPSG Arctic zone 1"
  "-29\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",85"
  ".4371183333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",90,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",87,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standar"
  "d parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",29500000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6116\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6117, 1, 'WGS 84 / EPSG Arctic zone 1-31', 'EPSG', 6117, 'PROJCS[\"WGS 84 / EPSG Arctic zone 1"
  "-31\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",85"
  ".4371183333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",150,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",87,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",31500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6117\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6118, 1, 'WGS 84 / EPSG Arctic zone 1-21', 'EPSG', 6118, 'PROJCS[\"WGS 84 / EPSG Arctic zone 1"
  "-21\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",85"
  ".4371183333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-150,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",87,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stand"
  "ard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",21500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6118\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6119, 1, 'WGS 84 / EPSG Arctic zone 2-28', 'EPSG', 6119, 'PROJCS[\"WGS 84 / EPSG Arctic zone 2"
  "-28\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",82"
  ".0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",133,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2850"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6120, 1, 'WGS 84 / EPSG Arctic zone 2-10', 'EPSG', 6120, 'PROJCS[\"WGS 84 / EPSG Arctic zone 2"
  "-10\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",82"
  ".0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",166,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1050"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6121, 1, 'WGS 84 / EPSG Arctic zone 2-12', 'EPSG', 6121, 'PROJCS[\"WGS 84 / EPSG Arctic zone 2"
  "-12\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",82"
  ".0584248888889,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-154,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",83.6666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",125"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6122, 1, 'WGS 84 / EPSG Arctic zone 3-21', 'EPSG', 6122, 'PROJCS[\"WGS 84 / EPSG Arctic zone 3"
  "-21\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",78"
  ".7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",176,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",21500000,AUTHORITY"
  "[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6123, 1, 'WGS 84 / EPSG Arctic zone 3-23', 'EPSG', 6123, 'PROJCS[\"WGS 84 / EPSG Arctic zone 3"
  "-23\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",78"
  ".7073375277778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-153,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",80.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",77,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",23500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6123\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6124, 1, 'WGS 84 / EPSG Arctic zone 4-12', 'EPSG', 6124, 'PROJCS[\"WGS 84 / EPSG Arctic zone 4"
  "-12\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",75"
  ".3644033055556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-155,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",77,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stand"
  "ard parallel\",73.6666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",12500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6125, 1, 'ETRS89 / EPSG Arctic zone 5-47', 'EPSG', 6125, 'PROJCS[\"ETRS89 / EPSG Arctic zone 5"
  "-47\",GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY["
  "\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PA"
  "RAMETER[\"Easting at false origin\",47500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",550"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"6125\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6128, 1, 'Grand Cayman National Grid 1959', 'EPSG', 6128, 'PROJCS[\"Grand Cayman National Grid"
  " 1959\",GEOGCS[\"GCGD59\",DATUM[\"Grand Cayman Geodetic Datum 1959\",SPHEROID[\"Clarke 1866\",6378206.4,294.978698213"
  "8982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[-179.483,-69.379,-27.584,-7.862,8.163,6.042,-13.925],AUTHORITY[\"EPSG\",\""
  "6723\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4723\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640419.9475,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"E\",EAST],AXIS[\"N"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6129, 1, 'Sister Islands National Grid 1961', 'EPSG', 6129, 'PROJCS[\"Sister Islands National "
  "Grid 1961\",GEOGCS[\"SIGD61\",DATUM[\"Sister Islands Geodetic Datum 1961\",SPHEROID[\"Clarke 1866\",6378206.4,294.978"
  "6982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOWGS84[8.853,-52.644,180.304,-0.393,-2.323,2.96,-24.081],AUTHORITY[\"EPSG\""
  ",\"6726\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4726\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640419.9475,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"6129\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6135, 1, 'CIGD11', 'EPSG', 6135, 'GEOGCS[\"CIGD11\",DATUM[\"Cayman Islands Geodetic Datum 2011"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"1100\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6135\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6201, 1, 'NAD27 / Michigan Central', 'EPSG', 6201, 'PROJCS[\"NAD27 / Michigan Central\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP Michigan)\",AUTHORITY[\"EPSG\",\"1051\"]],PARAMETER[\"Latitude of false origin\",43.3277777777778,AUTHORITY["
  "\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER"
  "[\"Latitude of 1st standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",45.7,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],PARAMETER[\"Ellipsoid scaling facto"
  "r\",1.0000382,AUTHORITY[\"EPSG\",\"1038\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6201\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6202, 1, 'NAD27 / Michigan South', 'EPSG', 6202, 'PROJCS[\"NAD27 / Michigan South\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP Michigan)\",AUTHORITY[\"EPSG\",\"1051\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",42.1,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",43.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],PARAMETER[\"Ellipsoid scaling factor\","
  "1.0000382,AUTHORITY[\"EPSG\",\"1038\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6202\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6204, 1, 'Macedonia State Coordinate System', 'EPSG', 6204, 'PROJCS[\"Macedonia State Coordina"
  "te System\",GEOGCS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\""
  ",\"7004\"]],TOWGS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"6204\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6207, 1, 'Nepal 1981', 'EPSG', 6207, 'GEOGCS[\"Nepal 1981\",DATUM[\"Nepal 1981\",SPHEROID[\"Ev"
  "erest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[293.17,726.18,245.36,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"1111\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6207\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6210, 1, 'SIRGAS 2000 / UTM zone 23N', 'EPSG', 6210, 'PROJCS[\"SIRGAS 2000 / UTM zone 23N\",GE"
  "OGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-4"
  "5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6210\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6211, 1, 'SIRGAS 2000 / UTM zone 24N', 'EPSG', 6211, 'PROJCS[\"SIRGAS 2000 / UTM zone 24N\",GE"
  "OGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3"
  "9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6211\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6244, 1, 'MAGNA-SIRGAS / Arauca urban grid', 'EPSG', 6244, 'PROJCS[\"MAGNA-SIRGAS / Arauca urb"
  "an grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMET"
  "ER[\"Latitude of natural origin\",7.08760639166667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-70.7583096555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1035263.443,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",1275526.621,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin hei"
  "ght\",100,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"6244\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6245, 1, 'MAGNA-SIRGAS / Armenia urban grid', 'EPSG', 6245, 'PROJCS[\"MAGNA-SIRGAS / Armenia u"
  "rban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAM"
  "ETER[\"Latitude of natural origin\",4.532325,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-75.6734891666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1155824.666,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",993087.465,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\",1"
  "470,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],A"
  "UTHORITY[\"EPSG\",\"6245\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6246, 1, 'MAGNA-SIRGAS / Barranquilla urban grid', 'EPSG', 6246, 'PROJCS[\"MAGNA-SIRGAS / Barr"
  "anquilla urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\""
  "]],PARAMETER[\"Latitude of natural origin\",10.9231830833333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-74.8343313333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",917264.406,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",1699839.935,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane o"
  "rigin height\",100,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXI"
  "S[\"E\",EAST],AUTHORITY[\"EPSG\",\"6246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6247, 1, 'MAGNA-SIRGAS / Bogota urban grid', 'EPSG', 6247, 'PROJCS[\"MAGNA-SIRGAS / Bogota urb"
  "an grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMET"
  "ER[\"Latitude of natural origin\",4.68048611111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-74.1465916666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",92334.879,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",109320.965,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height"
  "\",2550,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAS"
  "T],AUTHORITY[\"EPSG\",\"6247\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6248, 1, 'MAGNA-SIRGAS / Bucaramanga urban grid', 'EPSG', 6248, 'PROJCS[\"MAGNA-SIRGAS / Bucar"
  "amanga urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\""
  "]],PARAMETER[\"Latitude of natural origin\",7.07888714166667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-73.1973432222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1097241.305,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",1274642.278,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane "
  "origin height\",931,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AX"
  "IS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6248\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6249, 1, 'MAGNA-SIRGAS / Cali urban grid', 'EPSG', 6249, 'PROJCS[\"MAGNA-SIRGAS / Cali urban g"
  "rid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER[\""
  "Latitude of natural origin\",3.44188333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-76.5205625,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1061900.18,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",872364.63,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\",1000,AUT"
  "HORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6250, 1, 'MAGNA-SIRGAS / Cartagena urban grid', 'EPSG', 6250, 'PROJCS[\"MAGNA-SIRGAS / Cartage"
  "na urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],P"
  "ARAMETER[\"Latitude of natural origin\",10.3970475,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-75.5112069444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",842981.41,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",1641887.09,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height"
  "\",0,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],"
  "AUTHORITY[\"EPSG\",\"6250\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6251, 1, 'MAGNA-SIRGAS / Cucuta urban grid', 'EPSG', 6251, 'PROJCS[\"MAGNA-SIRGAS / Cucuta urb"
  "an grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMET"
  "ER[\"Latitude of natural origin\",7.88893673611111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-72.50287095,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",842805.406,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",1364404.57,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\",3"
  "08,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AU"
  "THORITY[\"EPSG\",\"6251\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6252, 1, 'MAGNA-SIRGAS / Florencia urban grid', 'EPSG', 6252, 'PROJCS[\"MAGNA-SIRGAS / Florenc"
  "ia urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],P"
  "ARAMETER[\"Latitude of natural origin\",1.62101229444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",-75.6191176027778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1162300.348,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",671068.716,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin"
  " height\",300,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E"
  "\",EAST],AUTHORITY[\"EPSG\",\"6252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6253, 1, 'MAGNA-SIRGAS / Ibague urban grid', 'EPSG', 6253, 'PROJCS[\"MAGNA-SIRGAS / Ibague urb"
  "an grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMET"
  "ER[\"Latitude of natural origin\",4.41941282777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-75.1799259333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",877634.33,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",980541.348,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height"
  "\",1100,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAS"
  "T],AUTHORITY[\"EPSG\",\"6253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6254, 1, 'MAGNA-SIRGAS / Inirida urban grid', 'EPSG', 6254, 'PROJCS[\"MAGNA-SIRGAS / Inirida u"
  "rban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAM"
  "ETER[\"Latitude of natural origin\",3.84543818333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-67.9052320888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1019177.687,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",491791.326,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin he"
  "ight\",96,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"6254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6255, 1, 'MAGNA-SIRGAS / Leticia urban grid', 'EPSG', 6255, 'PROJCS[\"MAGNA-SIRGAS / Leticia u"
  "rban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAM"
  "ETER[\"Latitude of natural origin\",-4.19768404722222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-69.9428110583333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",25978.217,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",27501.365,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin heig"
  "ht\",89.7,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"6255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6256, 1, 'MAGNA-SIRGAS / Manizales urban grid', 'EPSG', 6256, 'PROJCS[\"MAGNA-SIRGAS / Manizal"
  "es urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],P"
  "ARAMETER[\"Latitude of natural origin\",5.06815388888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",-75.5110947222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1173727.04,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",1052391.13,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin"
  " height\",2100,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\""
  "E\",EAST],AUTHORITY[\"EPSG\",\"6256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6257, 1, 'MAGNA-SIRGAS / Medellin urban grid', 'EPSG', 6257, 'PROJCS[\"MAGNA-SIRGAS / Medellin"
  " urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PAR"
  "AMETER[\"Latitude of natural origin\",6.22920888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-75.5648869444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",835378.647,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",1180816.875,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin h"
  "eight\",1510,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"6257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6258, 1, 'MAGNA-SIRGAS / Mitu urban grid', 'EPSG', 6258, 'PROJCS[\"MAGNA-SIRGAS / Mitu urban g"
  "rid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER[\""
  "Latitude of natural origin\",1.24996936666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-70.2354616555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1093717.398,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",629997.236,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\","
  "170,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],A"
  "UTHORITY[\"EPSG\",\"6258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6259, 1, 'MAGNA-SIRGAS / Mocoa urban grid', 'EPSG', 6259, 'PROJCS[\"MAGNA-SIRGAS / Mocoa urban"
  " grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER"
  "[\"Latitude of natural origin\",1.14002335833333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-76.6510212194445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1047467.388,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",617828.474,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height"
  "\",655.2,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6260, 1, 'MAGNA-SIRGAS / Monteria urban grid', 'EPSG', 6260, 'PROJCS[\"MAGNA-SIRGAS / Monteria"
  " urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PAR"
  "AMETER[\"Latitude of natural origin\",8.77308575555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-75.8795533305556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1131814.934,AUTHORITY[\"EPSG\","
  "\"8806\"]],PARAMETER[\"False northing\",1462131.119,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin"
  " height\",15,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"6260\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6261, 1, 'MAGNA-SIRGAS / Neiva urban grid', 'EPSG', 6261, 'PROJCS[\"MAGNA-SIRGAS / Neiva urban"
  " grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER"
  "[\"Latitude of natural origin\",2.94241505555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-75.2964367222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",864476.923,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",817199.827,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\""
  ",430,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],"
  "AUTHORITY[\"EPSG\",\"6261\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6262, 1, 'MAGNA-SIRGAS / Pasto urban grid', 'EPSG', 6262, 'PROJCS[\"MAGNA-SIRGAS / Pasto urban"
  " grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER"
  "[\"Latitude of natural origin\",1.20098951388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-77.2531256333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",980469.695,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",624555.332,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\""
  ",2530,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6262\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6263, 1, 'MAGNA-SIRGAS / Pereira urban grid', 'EPSG', 6263, 'PROJCS[\"MAGNA-SIRGAS / Pereira u"
  "rban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAM"
  "ETER[\"Latitude of natural origin\",4.81359361111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-75.6939513888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1153492.012,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",1024195.255,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin h"
  "eight\",1500,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"6263\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6264, 1, 'MAGNA-SIRGAS / Popayan urban grid', 'EPSG', 6264, 'PROJCS[\"MAGNA-SIRGAS / Popayan u"
  "rban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAM"
  "ETER[\"Latitude of natural origin\",2.45615988333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-76.6060916361111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1052430.525,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",763366.548,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin he"
  "ight\",1740,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"6264\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6265, 1, 'MAGNA-SIRGAS / Puerto Carreno urban grid', 'EPSG', 6265, 'PROJCS[\"MAGNA-SIRGAS / Pu"
  "erto Carreno urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\""
  "1052\"]],PARAMETER[\"Latitude of natural origin\",6.18072141388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-67.5007502472222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1063834.703,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1175257.481,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection "
  "plane origin height\",51.58,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",N"
  "ORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6265\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6266, 1, 'MAGNA-SIRGAS / Quibdo urban grid', 'EPSG', 6266, 'PROJCS[\"MAGNA-SIRGAS / Quibdo urb"
  "an grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMET"
  "ER[\"Latitude of natural origin\",5.69424766111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-76.6507538583334,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1047273.617,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",1121443.09,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin heig"
  "ht\",44,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAS"
  "T],AUTHORITY[\"EPSG\",\"6266\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6267, 1, 'MAGNA-SIRGAS / Riohacha urban grid', 'EPSG', 6267, 'PROJCS[\"MAGNA-SIRGAS / Riohacha"
  " urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PAR"
  "AMETER[\"Latitude of natural origin\",11.5369133277778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-72.9027688694445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1128154.73,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",1767887.914,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin h"
  "eight\",6,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"6267\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6268, 1, 'MAGNA-SIRGAS / San Andres urban grid', 'EPSG', 6268, 'PROJCS[\"MAGNA-SIRGAS / San An"
  "dres urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],"
  "PARAMETER[\"Latitude of natural origin\",12.523794325,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-81.72937595,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",820439.298,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",1877357.828,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\""
  ",6,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AU"
  "THORITY[\"EPSG\",\"6268\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6269, 1, 'MAGNA-SIRGAS / San Jose del Guaviare urban grid', 'EPSG', 6269, 'PROJCS[\"MAGNA-SIRG"
  "AS / San Jose del Guaviare urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SP"
  "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6"
  "686\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORI"
  "TY[\"EPSG\",\"1052\"]],PARAMETER[\"Latitude of natural origin\",2.56407894166667,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-72.640033325,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1159876.6"
  "2,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",775380.342,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Pro"
  "jection plane origin height\",185,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6269\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6270, 1, 'MAGNA-SIRGAS / Santa Marta urban grid', 'EPSG', 6270, 'PROJCS[\"MAGNA-SIRGAS / Santa"
  " Marta urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\""
  "]],PARAMETER[\"Latitude of natural origin\",11.2196430555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-74.2250052777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",983892.409,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",1732533.518,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane o"
  "rigin height\",29,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS"
  "[\"E\",EAST],AUTHORITY[\"EPSG\",\"6270\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6271, 1, 'MAGNA-SIRGAS / Sucre urban grid', 'EPSG', 6271, 'PROJCS[\"MAGNA-SIRGAS / Sucre urban"
  " grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER"
  "[\"Latitude of natural origin\",8.81055036666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-74.722466825,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",929043.607,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",1466125.658,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\",2"
  "0,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUT"
  "HORITY[\"EPSG\",\"6271\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6272, 1, 'MAGNA-SIRGAS / Tunja urban grid', 'EPSG', 6272, 'PROJCS[\"MAGNA-SIRGAS / Tunja urban"
  " grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER"
  "[\"Latitude of natural origin\",5.53419473888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-73.3519389,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1080514.91,AUTHORITY[\"EPSG\",\"8806\"]],PA"
  "RAMETER[\"False northing\",1103772.028,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\",280"
  "0,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUT"
  "HORITY[\"EPSG\",\"6272\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6273, 1, 'MAGNA-SIRGAS / Valledupar urban grid', 'EPSG', 6273, 'PROJCS[\"MAGNA-SIRGAS / Valled"
  "upar urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],"
  "PARAMETER[\"Latitude of natural origin\",10.4472611111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-73.2465713888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1090979.66,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",1647208.93,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin"
  " height\",200,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E"
  "\",EAST],AUTHORITY[\"EPSG\",\"6273\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6274, 1, 'MAGNA-SIRGAS / Villavicencio urban grid', 'EPSG', 6274, 'PROJCS[\"MAGNA-SIRGAS / Vil"
  "lavicencio urban grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 198"
  "0\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052"
  "\"]],PARAMETER[\"Latitude of natural origin\",4.1553751,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-73.6244859861111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1050678.757,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",950952.124,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin"
  " height\",427.19,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS["
  "\"E\",EAST],AUTHORITY[\"EPSG\",\"6274\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6275, 1, 'MAGNA-SIRGAS / Yopal urban grid', 'EPSG', 6275, 'PROJCS[\"MAGNA-SIRGAS / Yopal urban"
  " grid\",GEOGCS[\"MAGNA-SIRGAS\",DATUM[\"Marco Geocentrico Nacional de Referencia\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6686\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4686\"]],PROJECTION[\"Colombia Urban\",AUTHORITY[\"EPSG\",\"1052\"]],PARAMETER"
  "[\"Latitude of natural origin\",5.35392722222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-72.4200402777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",851184.177,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",1083954.137,AUTHORITY[\"EPSG\",\"8807\"]],PARAMETER[\"Projection plane origin height\""
  ",300,AUTHORITY[\"EPSG\",\"1039\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],"
  "AUTHORITY[\"EPSG\",\"6275\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6307, 1, 'NAD83(CORS96) / Puerto Rico and Virgin Is.', 'EPSG', 6307, 'PROJCS[\"NAD83(CORS96) /"
  " Puerto Rico and Virgin Is.\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",17.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"L"
  "ongitude of false origin\",-66.4333333333334,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard paral"
  "lel\",18.4333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",18.0444444444444"
  ",AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6307\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6311, 1, 'CGRS93', 'EPSG', 6311, 'GEOGCS[\"CGRS93\",DATUM[\"Cyprus Geodetic Reference System 1"
  "993\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[8.846,-4.394,-1.122,-0.00237,-0"
  ".146528,0.130428,0.783926],AUTHORITY[\"EPSG\",\"1112\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6312, 1, 'CGRS93 / Cyprus Local Transverse Mercator', 'EPSG', 6312, 'PROJCS[\"CGRS93 / Cyprus "
  "Local Transverse Mercator\",GEOGCS[\"CGRS93\",DATUM[\"Cyprus Geodetic Reference System 1993\",SPHEROID[\"WGS 84\",637"
  "8137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[8.846,-4.394,-1.122,-0.00237,-0.146528,0.130428,0.783926],AU"
  "THORITY[\"EPSG\",\"1112\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6311\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",-3500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6316, 1, 'MGI 1901 / Balkans zone 7', 'EPSG', 6316, 'PROJCS[\"MGI 1901 / Balkans zone 7\",GEOG"
  "CS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW"
  "GS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",7500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"6316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6318, 1, 'NAD83(2011)', 'EPSG', 6318, 'GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial "
  "Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\""
  ",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6322, 1, 'NAD83(PA11)', 'EPSG', 6322, 'GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial "
  "Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\""
  ",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6325, 1, 'NAD83(MA11)', 'EPSG', 6325, 'GEOGCS[\"NAD83(MA11)\",DATUM[\"NAD83 (National Spatial "
  "Reference System MA11)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\""
  ",\"1118\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6325\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6328, 1, 'NAD83(2011) / UTM zone 59N', 'EPSG', 6328, 'PROJCS[\"NAD83(2011) / UTM zone 59N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6328\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6329, 1, 'NAD83(2011) / UTM zone 60N', 'EPSG', 6329, 'PROJCS[\"NAD83(2011) / UTM zone 60N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6329\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6330, 1, 'NAD83(2011) / UTM zone 1N', 'EPSG', 6330, 'PROJCS[\"NAD83(2011) / UTM zone 1N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6331, 1, 'NAD83(2011) / UTM zone 2N', 'EPSG', 6331, 'PROJCS[\"NAD83(2011) / UTM zone 2N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6332, 1, 'NAD83(2011) / UTM zone 3N', 'EPSG', 6332, 'PROJCS[\"NAD83(2011) / UTM zone 3N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6333, 1, 'NAD83(2011) / UTM zone 4N', 'EPSG', 6333, 'PROJCS[\"NAD83(2011) / UTM zone 4N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6334, 1, 'NAD83(2011) / UTM zone 5N', 'EPSG', 6334, 'PROJCS[\"NAD83(2011) / UTM zone 5N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6334\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6335, 1, 'NAD83(2011) / UTM zone 6N', 'EPSG', 6335, 'PROJCS[\"NAD83(2011) / UTM zone 6N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6335\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6336, 1, 'NAD83(2011) / UTM zone 7N', 'EPSG', 6336, 'PROJCS[\"NAD83(2011) / UTM zone 7N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6336\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6337, 1, 'NAD83(2011) / UTM zone 8N', 'EPSG', 6337, 'PROJCS[\"NAD83(2011) / UTM zone 8N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6338, 1, 'NAD83(2011) / UTM zone 9N', 'EPSG', 6338, 'PROJCS[\"NAD83(2011) / UTM zone 9N\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6338\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6339, 1, 'NAD83(2011) / UTM zone 10N', 'EPSG', 6339, 'PROJCS[\"NAD83(2011) / UTM zone 10N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6340, 1, 'NAD83(2011) / UTM zone 11N', 'EPSG', 6340, 'PROJCS[\"NAD83(2011) / UTM zone 11N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6341, 1, 'NAD83(2011) / UTM zone 12N', 'EPSG', 6341, 'PROJCS[\"NAD83(2011) / UTM zone 12N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6341\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6342, 1, 'NAD83(2011) / UTM zone 13N', 'EPSG', 6342, 'PROJCS[\"NAD83(2011) / UTM zone 13N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6342\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6343, 1, 'NAD83(2011) / UTM zone 14N', 'EPSG', 6343, 'PROJCS[\"NAD83(2011) / UTM zone 14N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6344, 1, 'NAD83(2011) / UTM zone 15N', 'EPSG', 6344, 'PROJCS[\"NAD83(2011) / UTM zone 15N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6345, 1, 'NAD83(2011) / UTM zone 16N', 'EPSG', 6345, 'PROJCS[\"NAD83(2011) / UTM zone 16N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6345\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6346, 1, 'NAD83(2011) / UTM zone 17N', 'EPSG', 6346, 'PROJCS[\"NAD83(2011) / UTM zone 17N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6347, 1, 'NAD83(2011) / UTM zone 18N', 'EPSG', 6347, 'PROJCS[\"NAD83(2011) / UTM zone 18N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6347\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6348, 1, 'NAD83(2011) / UTM zone 19N', 'EPSG', 6348, 'PROJCS[\"NAD83(2011) / UTM zone 19N\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6350, 1, 'NAD83(2011) / Conus Albers', 'EPSG', 6350, 'PROJCS[\"NAD83(2011) / Conus Albers\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false orig"
  "in\",23,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-96,AUTHORITY[\"EPSG\",\"8822\"]],PARAM"
  "ETER[\"Latitude of 1st standard parallel\",29.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard pa"
  "rallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6351, 1, 'NAD83(2011) / EPSG Arctic zone 5-29', 'EPSG', 6351, 'PROJCS[\"NAD83(2011) / EPSG Arc"
  "tic zone 5-29\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-163,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"E"
  "PSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAM"
  "ETER[\"Easting at false origin\",29500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",550000"
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"6351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6352, 1, 'NAD83(2011) / EPSG Arctic zone 5-31', 'EPSG', 6352, 'PROJCS[\"NAD83(2011) / EPSG Arc"
  "tic zone 5-31\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",72.0250091944445,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-147,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",73.6666666666667,AUTHORITY[\"E"
  "PSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",70.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAM"
  "ETER[\"Easting at false origin\",31500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",550000"
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"6352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6353, 1, 'NAD83(2011) / EPSG Arctic zone 6-14', 'EPSG', 6353, 'PROJCS[\"NAD83(2011) / EPSG Arc"
  "tic zone 6-14\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-165,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"E"
  "PSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",14500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6354, 1, 'NAD83(2011) / EPSG Arctic zone 6-16', 'EPSG', 6354, 'PROJCS[\"NAD83(2011) / EPSG Arc"
  "tic zone 6-16\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",68.6874755555556,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-147,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",70.3333333333333,AUTHORITY[\"E"
  "PSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",67,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",16500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6500000,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6355, 1, 'NAD83(2011) / Alabama East', 'EPSG', 6355, 'PROJCS[\"NAD83(2011) / Alabama East\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",30.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.8333333333333,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6356, 1, 'NAD83(2011) / Alabama West', 'EPSG', 6356, 'PROJCS[\"NAD83(2011) / Alabama West\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5111111111111,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6362, 1, 'Mexico ITRF92 / LCC', 'EPSG', 6362, 'PROJCS[\"Mexico ITRF92 / LCC\",GEOGCS[\"Mexico "
  "ITRF92\",DATUM[\"Mexico ITRF92\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,"
  "0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1042\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4483\"]],PR"
  "OJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",12,AU"
  "THORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-102,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"La"
  "titude of 1st standard parallel\",17.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "29.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6362\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6365, 1, 'Mexico ITRF2008', 'EPSG', 6365, 'GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6365\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6366, 1, 'Mexico ITRF2008 / UTM zone 11N', 'EPSG', 6366, 'PROJCS[\"Mexico ITRF2008 / UTM zone "
  "11N\",GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6366\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6367, 1, 'Mexico ITRF2008 / UTM zone 12N', 'EPSG', 6367, 'PROJCS[\"Mexico ITRF2008 / UTM zone "
  "12N\",GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6367\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6368, 1, 'Mexico ITRF2008 / UTM zone 13N', 'EPSG', 6368, 'PROJCS[\"Mexico ITRF2008 / UTM zone "
  "13N\",GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6368\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6369, 1, 'Mexico ITRF2008 / UTM zone 14N', 'EPSG', 6369, 'PROJCS[\"Mexico ITRF2008 / UTM zone "
  "14N\",GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6369\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6370, 1, 'Mexico ITRF2008 / UTM zone 15N', 'EPSG', 6370, 'PROJCS[\"Mexico ITRF2008 / UTM zone "
  "15N\",GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6370\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6371, 1, 'Mexico ITRF2008 / UTM zone 16N', 'EPSG', 6371, 'PROJCS[\"Mexico ITRF2008 / UTM zone "
  "16N\",GEOGCS[\"Mexico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6365\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6371\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6372, 1, 'Mexico ITRF2008 / LCC', 'EPSG', 6372, 'PROJCS[\"Mexico ITRF2008 / LCC\",GEOGCS[\"Mex"
  "ico ITRF2008\",DATUM[\"Mexico ITRF2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1120\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"636"
  "5\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin"
  "\",12,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-102,AUTHORITY[\"EPSG\",\"8822\"]],PARAME"
  "TER[\"Latitude of 1st standard parallel\",17.5,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard par"
  "allel\",29.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]"
  "],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6372\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6381, 1, 'UCS-2000 / Ukraine TM zone 7', 'EPSG', 6381, 'PROJCS[\"UCS-2000 / Ukraine TM zone 7\""
  ",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],"
  "TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6381\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6382, 1, 'UCS-2000 / Ukraine TM zone 8', 'EPSG', 6382, 'PROJCS[\"UCS-2000 / Ukraine TM zone 8\""
  ",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],"
  "TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6382\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6383, 1, 'UCS-2000 / Ukraine TM zone 9', 'EPSG', 6383, 'PROJCS[\"UCS-2000 / Ukraine TM zone 9\""
  ",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],"
  "TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6383\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6384, 1, 'UCS-2000 / Ukraine TM zone 10', 'EPSG', 6384, 'PROJCS[\"UCS-2000 / Ukraine TM zone 1"
  "0\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6384\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6385, 1, 'UCS-2000 / Ukraine TM zone 11', 'EPSG', 6385, 'PROJCS[\"UCS-2000 / Ukraine TM zone 1"
  "1\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6385\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6386, 1, 'UCS-2000 / Ukraine TM zone 12', 'EPSG', 6386, 'PROJCS[\"UCS-2000 / Ukraine TM zone 1"
  "2\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6386\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6387, 1, 'UCS-2000 / Ukraine TM zone 13', 'EPSG', 6387, 'PROJCS[\"UCS-2000 / Ukraine TM zone 1"
  "3\",GEOGCS[\"UCS-2000\",DATUM[\"Ukraine 2000\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"1077\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"5561\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6387\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6391, 1, 'Cayman Islands National Grid 2011', 'EPSG', 6391, 'PROJCS[\"Cayman Islands National "
  "Grid 2011\",GEOGCS[\"CIGD11\",DATUM[\"Cayman Islands Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.25722210"
  "1,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1100\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6135\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",19.3444444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",-80.5666666666667,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",19.34444444444"
  "44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",19.7,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",2950000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",190"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"E\",EAST],AXIS[\"N\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6393, 1, 'NAD83(2011) / Alaska Albers', 'EPSG', 6393, 'PROJCS[\"NAD83(2011) / Alaska Albers\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false or"
  "igin\",50,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-154,AUTHORITY[\"EPSG\",\"8822\"]],PA"
  "RAMETER[\"Latitude of 1st standard parallel\",55,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard p"
  "arallel\",65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6394, 1, 'NAD83(2011) / Alaska zone 1', 'EPSG', 6394, 'PROJCS[\"NAD83(2011) / Alaska zone 1\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"La"
  "titude of projection centre\",57,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",-133.6666"
  "66666667,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",323.130102361111,AUTHORITY[\"EPSG\",\"88"
  "13\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",323.130102361111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Sca"
  "le factor on initial line\",0.9999,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6394\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6395, 1, 'NAD83(2011) / Alaska zone 2', 'EPSG', 6395, 'PROJCS[\"NAD83(2011) / Alaska zone 2\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-142,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6396, 1, 'NAD83(2011) / Alaska zone 3', 'EPSG', 6396, 'PROJCS[\"NAD83(2011) / Alaska zone 3\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-146,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6396\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6397, 1, 'NAD83(2011) / Alaska zone 4', 'EPSG', 6397, 'PROJCS[\"NAD83(2011) / Alaska zone 4\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-150,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6397\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6398, 1, 'NAD83(2011) / Alaska zone 5', 'EPSG', 6398, 'PROJCS[\"NAD83(2011) / Alaska zone 5\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-154,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6398\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6399, 1, 'NAD83(2011) / Alaska zone 6', 'EPSG', 6399, 'PROJCS[\"NAD83(2011) / Alaska zone 6\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-158,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6399\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6400, 1, 'NAD83(2011) / Alaska zone 7', 'EPSG', 6400, 'PROJCS[\"NAD83(2011) / Alaska zone 7\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-162,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6400\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6401, 1, 'NAD83(2011) / Alaska zone 8', 'EPSG', 6401, 'PROJCS[\"NAD83(2011) / Alaska zone 8\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-166,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6401\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6402, 1, 'NAD83(2011) / Alaska zone 9', 'EPSG', 6402, 'PROJCS[\"NAD83(2011) / Alaska zone 9\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-170,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6402\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6403, 1, 'NAD83(2011) / Alaska zone 10', 'EPSG', 6403, 'PROJCS[\"NAD83(2011) / Alaska zone 10\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",51,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-176,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",53.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",51.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6403\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6404, 1, 'NAD83(2011) / Arizona Central', 'EPSG', 6404, 'PROJCS[\"NAD83(2011) / Arizona Centra"
  "l\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111.927777777778,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6404\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6405, 1, 'NAD83(2011) / Arizona Central (ft)', 'EPSG', 6405, 'PROJCS[\"NAD83(2011) / Arizona C"
  "entral (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111.927777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6405\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6406, 1, 'NAD83(2011) / Arizona East', 'EPSG', 6406, 'PROJCS[\"NAD83(2011) / Arizona East\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.177777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6406\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6407, 1, 'NAD83(2011) / Arizona East (ft)', 'EPSG', 6407, 'PROJCS[\"NAD83(2011) / Arizona East"
  " (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.177777777778,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6407\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6408, 1, 'NAD83(2011) / Arizona West', 'EPSG', 6408, 'PROJCS[\"NAD83(2011) / Arizona West\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.75,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6408\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6409, 1, 'NAD83(2011) / Arizona West (ft)', 'EPSG', 6409, 'PROJCS[\"NAD83(2011) / Arizona West"
  " (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.75,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6409\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6410, 1, 'NAD83(2011) / Arkansas North', 'EPSG', 6410, 'PROJCS[\"NAD83(2011) / Arkansas North\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6410\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6411, 1, 'NAD83(2011) / Arkansas North (ftUS)', 'EPSG', 6411, 'PROJCS[\"NAD83(2011) / Arkansas"
  " North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.2333333333333,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,A"
  "UTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6411\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6412, 1, 'NAD83(2011) / Arkansas South', 'EPSG', 6412, 'PROJCS[\"NAD83(2011) / Arkansas South\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"64"
  "12\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6413, 1, 'NAD83(2011) / Arkansas South (ftUS)', 'EPSG', 6413, 'PROJCS[\"NAD83(2011) / Arkansas"
  " South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.7666666666667,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",131"
  "2333.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6413\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6414, 1, 'NAD83(2011) / California Albers', 'EPSG', 6414, 'PROJCS[\"NAD83(2011) / California A"
  "lbers\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of "
  "false origin\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",34,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",40.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",-4000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6414\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6415, 1, 'NAD83(2011) / California zone 1', 'EPSG', 6415, 'PROJCS[\"NAD83(2011) / California z"
  "one 1\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "2,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.6777777777778,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6415\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6416, 1, 'NAD83(2011) / California zone 1 (ftUS)', 'EPSG', 6416, 'PROJCS[\"NAD83(2011) / Calif"
  "ornia zone 1 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.6777777777778,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.66"
  "7,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6416\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6417, 1, 'NAD83(2011) / California zone 2', 'EPSG', 6417, 'PROJCS[\"NAD83(2011) / California z"
  "one 2\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "2,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.8333333333333,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6417\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6418, 1, 'NAD83(2011) / California zone 2 (ftUS)', 'EPSG', 6418, 'PROJCS[\"NAD83(2011) / Calif"
  "ornia zone 2 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.8333333333333,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6418\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6419, 1, 'NAD83(2011) / California zone 3', 'EPSG', 6419, 'PROJCS[\"NAD83(2011) / California z"
  "one 3\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "0.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5"
  "00000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"6419\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6420, 1, 'NAD83(2011) / California zone 3 (ftUS)', 'EPSG', 6420, 'PROJCS[\"NAD83(2011) / Calif"
  "ornia zone 3 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444"
  "444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.0777777777778,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing a"
  "t false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EP"
  "SG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6420\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6421, 1, 'NAD83(2011) / California zone 4', 'EPSG', 6421, 'PROJCS[\"NAD83(2011) / California z"
  "one 4\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "9,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.2611111111111,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6421\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6422, 1, 'NAD83(2011) / California zone 4 (ftUS)', 'EPSG', 6422, 'PROJCS[\"NAD83(2011) / Calif"
  "ornia zone 4 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-119,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.2611111111111,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1640416.66"
  "7,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6422\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6423, 1, 'NAD83(2011) / California zone 5', 'EPSG', 6423, 'PROJCS[\"NAD83(2011) / California z"
  "one 5\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "8,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.4666666666667,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6423\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6424, 1, 'NAD83(2011) / California zone 5 (ftUS)', 'EPSG', 6424, 'PROJCS[\"NAD83(2011) / Calif"
  "ornia zone 5 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-118,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.4666666666667,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]"
  "],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6424\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6425, 1, 'NAD83(2011) / California zone 6', 'EPSG', 6425, 'PROJCS[\"NAD83(2011) / California z"
  "one 6\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "6.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.8833333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"6425\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6426, 1, 'NAD83(2011) / California zone 6 (ftUS)', 'EPSG', 6426, 'PROJCS[\"NAD83(2011) / Calif"
  "ornia zone 6 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-116.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.8833333333333,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",1640416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6426\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6427, 1, 'NAD83(2011) / Colorado Central', 'EPSG', 6427, 'PROJCS[\"NAD83(2011) / Colorado Cent"
  "ral\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105."
  "511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.75,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",304800.6096,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"6427\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6428, 1, 'NAD83(2011) / Colorado Central (ftUS)', 'EPSG', 6428, 'PROJCS[\"NAD83(2011) / Colora"
  "do Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.75,AUTHOR"
  "ITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER"
  "[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUT"
  "HORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6428\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6429, 1, 'NAD83(2011) / Colorado North', 'EPSG', 6429, 'PROJCS[\"NAD83(2011) / Colorado North\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.7944444444444,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",30480"
  "0.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6429\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6430, 1, 'NAD83(2011) / Colorado North (ftUS)', 'EPSG', 6430, 'PROJCS[\"NAD83(2011) / Colorado"
  " North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.7944444444444,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6430\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6431, 1, 'NAD83(2011) / Colorado South', 'EPSG', 6431, 'PROJCS[\"NAD83(2011) / Colorado South\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",30480"
  "0.6096,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6431\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6432, 1, 'NAD83(2011) / Colorado South (ftUS)', 'EPSG', 6432, 'PROJCS[\"NAD83(2011) / Colorado"
  " South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false ori"
  "gin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6432\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6433, 1, 'NAD83(2011) / Connecticut', 'EPSG', 6433, 'PROJCS[\"NAD83(2011) / Connecticut\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-72.75,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.8666666666667,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",304800.6096,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",152400.3048,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6433\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6434, 1, 'NAD83(2011) / Connecticut (ftUS)', 'EPSG', 6434, 'PROJCS[\"NAD83(2011) / Connecticut"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-72"
  ".75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.8666666666667,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6434\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6435, 1, 'NAD83(2011) / Delaware', 'EPSG', 6435, 'PROJCS[\"NAD83(2011) / Delaware\",GEOGCS[\"N"
  "AD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT"
  "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.4166666666667,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6435\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6436, 1, 'NAD83(2011) / Delaware (ftUS)', 'EPSG', 6436, 'PROJCS[\"NAD83(2011) / Delaware (ftUS"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.4166666666667,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"6436\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6437, 1, 'NAD83(2011) / Florida East', 'EPSG', 6437, 'PROJCS[\"NAD83(2011) / Florida East\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6437\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6438, 1, 'NAD83(2011) / Florida East (ftUS)', 'EPSG', 6438, 'PROJCS[\"NAD83(2011) / Florida Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6438\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6439, 1, 'NAD83(2011) / Florida GDL Albers', 'EPSG', 6439, 'PROJCS[\"NAD83(2011) / Florida GDL"
  " Albers\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of"
  " false origin\",24,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84,AUTHORITY[\"EPSG\",\"882"
  "2\"]],PARAMETER[\"Latitude of 1st standard parallel\",24,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd st"
  "andard parallel\",31.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6439\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6440, 1, 'NAD83(2011) / Florida North', 'EPSG', 6440, 'PROJCS[\"NAD83(2011) / Florida North\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.5111111111111,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7611111111111,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6440\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6441, 1, 'NAD83(2011) / Florida North (ftUS)', 'EPSG', 6441, 'PROJCS[\"NAD83(2011) / Florida N"
  "orth (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7611111111111,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6441\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6442, 1, 'NAD83(2011) / Florida West', 'EPSG', 6442, 'PROJCS[\"NAD83(2011) / Florida West\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6442\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6443, 1, 'NAD83(2011) / Florida West (ftUS)', 'EPSG', 6443, 'PROJCS[\"NAD83(2011) / Florida We"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6443\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6444, 1, 'NAD83(2011) / Georgia East', 'EPSG', 6444, 'PROJCS[\"NAD83(2011) / Georgia East\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.1777777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6444\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6445, 1, 'NAD83(2011) / Georgia East (ftUS)', 'EPSG', 6445, 'PROJCS[\"NAD83(2011) / Georgia Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.1777777777778,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"6445\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6446, 1, 'NAD83(2011) / Georgia West', 'EPSG', 6446, 'PROJCS[\"NAD83(2011) / Georgia West\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.1777777777778,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6446\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6447, 1, 'NAD83(2011) / Georgia West (ftUS)', 'EPSG', 6447, 'PROJCS[\"NAD83(2011) / Georgia We"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.1777777777778,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"6447\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6448, 1, 'NAD83(2011) / Idaho Central', 'EPSG', 6448, 'PROJCS[\"NAD83(2011) / Idaho Central\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6448\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6449, 1, 'NAD83(2011) / Idaho Central (ftUS)', 'EPSG', 6449, 'PROJCS[\"NAD83(2011) / Idaho Cen"
  "tral (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-11"
  "4,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6449\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6450, 1, 'NAD83(2011) / Idaho East', 'EPSG', 6450, 'PROJCS[\"NAD83(2011) / Idaho East\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.177777777778,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6450\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6451, 1, 'NAD83(2011) / Idaho East (ftUS)', 'EPSG', 6451, 'PROJCS[\"NAD83(2011) / Idaho East ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.1777"
  "77777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",656166.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6451\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6452, 1, 'NAD83(2011) / Idaho West', 'EPSG', 6452, 'PROJCS[\"NAD83(2011) / Idaho West\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.75,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6452\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6453, 1, 'NAD83(2011) / Idaho West (ftUS)', 'EPSG', 6453, 'PROJCS[\"NAD83(2011) / Idaho West ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.75,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",2624666.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6453\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6454, 1, 'NAD83(2011) / Illinois East', 'EPSG', 6454, 'PROJCS[\"NAD83(2011) / Illinois East\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.3333333333333"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999975,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6454\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6455, 1, 'NAD83(2011) / Illinois East (ftUS)', 'EPSG', 6455, 'PROJCS[\"NAD83(2011) / Illinois "
  "East (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88"
  ".3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999975,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6455\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6456, 1, 'NAD83(2011) / Illinois West', 'EPSG', 6456, 'PROJCS[\"NAD83(2011) / Illinois West\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.1777777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6456\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6457, 1, 'NAD83(2011) / Illinois West (ftUS)', 'EPSG', 6457, 'PROJCS[\"NAD83(2011) / Illinois "
  "West (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90"
  ".1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",2296583.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6457\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6458, 1, 'NAD83(2011) / Indiana East', 'EPSG', 6458, 'PROJCS[\"NAD83(2011) / Indiana East\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.6666666666667,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"64"
  "58\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6459, 1, 'NAD83(2011) / Indiana East (ftUS)', 'EPSG', 6459, 'PROJCS[\"NAD83(2011) / Indiana Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.6"
  "666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",328083.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",820208."
  "333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6459\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6460, 1, 'NAD83(2011) / Indiana West', 'EPSG', 6460, 'PROJCS[\"NAD83(2011) / Indiana West\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.0944444444444,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"64"
  "60\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6461, 1, 'NAD83(2011) / Indiana West (ftUS)', 'EPSG', 6461, 'PROJCS[\"NAD83(2011) / Indiana We"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.0"
  "944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",820208.333"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6461\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6462, 1, 'NAD83(2011) / Iowa North', 'EPSG', 6462, 'PROJCS[\"NAD83(2011) / Iowa North\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.2777777777778,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6462\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6463, 1, 'NAD83(2011) / Iowa North (ftUS)', 'EPSG', 6463, 'PROJCS[\"NAD83(2011) / Iowa North ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.2777777777778,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3"
  "280833.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6463\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6464, 1, 'NAD83(2011) / Iowa South', 'EPSG', 6464, 'PROJCS[\"NAD83(2011) / Iowa South\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111,AUTHORITY[\"E"
  "PSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAM"
  "ETER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6464\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6465, 1, 'NAD83(2011) / Iowa South (ftUS)', 'EPSG', 6465, 'PROJCS[\"NAD83(2011) / Iowa South ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6465\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6466, 1, 'NAD83(2011) / Kansas North', 'EPSG', 6466, 'PROJCS[\"NAD83(2011) / Kansas North\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6466\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6467, 1, 'NAD83(2011) / Kansas North (ftUS)', 'EPSG', 6467, 'PROJCS[\"NAD83(2011) / Kansas Nor"
  "th (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.7944444444444,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6467\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6468, 1, 'NAD83(2011) / Kansas South', 'EPSG', 6468, 'PROJCS[\"NAD83(2011) / Kansas South\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.511111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.5666666666667,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6468\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6469, 1, 'NAD83(2011) / Kansas South (ftUS)', 'EPSG', 6469, 'PROJCS[\"NAD83(2011) / Kansas Sou"
  "th (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.5666666666667,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false o"
  "rigin\",1312333.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6469\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6470, 1, 'NAD83(2011) / Kentucky North', 'EPSG', 6470, 'PROJCS[\"NAD83(2011) / Kentucky North\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.25,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6470\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6471, 1, 'NAD83(2011) / Kentucky North (ftUS)', 'EPSG', 6471, 'PROJCS[\"NAD83(2011) / Kentucky"
  " North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-84.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6471\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6472, 1, 'NAD83(2011) / Kentucky Single Zone', 'EPSG', 6472, 'PROJCS[\"NAD83(2011) / Kentucky "
  "Single Zone\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.0944444444444,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000"
  "0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"6472\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6473, 1, 'NAD83(2011) / Kentucky Single Zone (ftUS)', 'EPSG', 6473, 'PROJCS[\"NAD83(2011) / Ke"
  "ntucky Single Zone (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.094444444444"
  "4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.6777777777778,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",4921250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6473\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6474, 1, 'NAD83(2011) / Kentucky South', 'EPSG', 6474, 'PROJCS[\"NAD83(2011) / Kentucky South\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-85.75,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9333333333333,AUTHORITY[\"EPSG\",\"8823\""
  "]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "6474\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6475, 1, 'NAD83(2011) / Kentucky South (ftUS)', 'EPSG', 6475, 'PROJCS[\"NAD83(2011) / Kentucky"
  " South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9333333333333,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",16"
  "40416.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6475\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6476, 1, 'NAD83(2011) / Louisiana North', 'EPSG', 6476, 'PROJCS[\"NAD83(2011) / Louisiana Nort"
  "h\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92.5111111111111,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.6777777777778,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "6476\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6477, 1, 'NAD83(2011) / Louisiana North (ftUS)', 'EPSG', 6477, 'PROJCS[\"NAD83(2011) / Louisia"
  "na North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.6777777777778,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6477\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6478, 1, 'NAD83(2011) / Louisiana South', 'EPSG', 6478, 'PROJCS[\"NAD83(2011) / Louisiana Sout"
  "h\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3333333333333,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAM"
  "ETER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6478\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6479, 1, 'NAD83(2011) / Louisiana South (ftUS)', 'EPSG', 6479, 'PROJCS[\"NAD83(2011) / Louisia"
  "na South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3"
  "333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",3280833.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6479\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6480, 1, 'NAD83(2011) / Maine CS2000 Central', 'EPSG', 6480, 'PROJCS[\"NAD83(2011) / Maine CS2"
  "000 Central\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69"
  ".125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6480\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6481, 1, 'NAD83(2011) / Maine CS2000 East', 'EPSG', 6481, 'PROJCS[\"NAD83(2011) / Maine CS2000"
  " East\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-67.875,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6481\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6482, 1, 'NAD83(2011) / Maine CS2000 West', 'EPSG', 6482, 'PROJCS[\"NAD83(2011) / Maine CS2000"
  " West\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.375,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99998,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6482\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6483, 1, 'NAD83(2011) / Maine East', 'EPSG', 6483, 'PROJCS[\"NAD83(2011) / Maine East\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-68.5111111111111,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6483\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6484, 1, 'NAD83(2011) / Maine East (ftUS)', 'EPSG', 6484, 'PROJCS[\"NAD83(2011) / Maine East ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-68.51111"
  "11111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6484\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6485, 1, 'NAD83(2011) / Maine West', 'EPSG', 6485, 'PROJCS[\"NAD83(2011) / Maine West\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.1777777777778,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6485\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6486, 1, 'NAD83(2011) / Maine West (ftUS)', 'EPSG', 6486, 'PROJCS[\"NAD83(2011) / Maine West ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.17777"
  "77777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6486\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6487, 1, 'NAD83(2011) / Maryland', 'EPSG', 6487, 'PROJCS[\"NAD83(2011) / Maryland\",GEOGCS[\"N"
  "AD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT"
  "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude "
  "of 2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",40000"
  "0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6487\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6488, 1, 'NAD83(2011) / Maryland (ftUS)', 'EPSG', 6488, 'PROJCS[\"NAD83(2011) / Maryland (ftUS"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77,AUT"
  "HORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",1312333.333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6488\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6489, 1, 'NAD83(2011) / Massachusetts Island', 'EPSG', 6489, 'PROJCS[\"NAD83(2011) / Massachus"
  "etts Island\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-70.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.4944444444444,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6489\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6490, 1, 'NAD83(2011) / Massachusetts Island (ftUS)', 'EPSG', 6490, 'PROJCS[\"NAD83(2011) / Ma"
  "ssachusetts Island (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.4944444444444,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6490\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6491, 1, 'NAD83(2011) / Massachusetts Mainland', 'EPSG', 6491, 'PROJCS[\"NAD83(2011) / Massach"
  "usetts Mainland\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-71.511"
  "1111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.6944444444444,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",750000,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"6491\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6492, 1, 'NAD83(2011) / Massachusetts Mainland (ftUS)', 'EPSG', 6492, 'PROJCS[\"NAD83(2011) / "
  "Massachusetts Mainland (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.69444444444"
  "44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.7277777777778,AUTHORITY[\"EPSG\","
  "\"8824\"]],PARAMETER[\"Easting at false origin\",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fal"
  "se origin\",2460625,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"90"
  "03\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6492\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6493, 1, 'NAD83(2011) / Michigan Central', 'EPSG', 6493, 'PROJCS[\"NAD83(2011) / Michigan Cent"
  "ral\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3"
  "777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.7,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",6000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6493\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6494, 1, 'NAD83(2011) / Michigan Central (ft)', 'EPSG', 6494, 'PROJCS[\"NAD83(2011) / Michigan"
  " Central (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.7,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",19685039.37,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"6494\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6495, 1, 'NAD83(2011) / Michigan North', 'EPSG', 6495, 'PROJCS[\"NAD83(2011) / Michigan North\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-87,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",8000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6495\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6496, 1, 'NAD83(2011) / Michigan North (ft)', 'EPSG', 6496, 'PROJCS[\"NAD83(2011) / Michigan N"
  "orth (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-87,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.0944444444444,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",26246719.16,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"6496\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6497, 1, 'NAD83(2011) / Michigan Oblique Mercator', 'EPSG', 6497, 'PROJCS[\"NAD83(2011) / Mich"
  "igan Oblique Mercator\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\""
  "9812\"]],PARAMETER[\"Latitude of projection centre\",45.3091666666667,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longi"
  "tude of projection centre\",-86,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",337.25556,AUTHORI"
  "TY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",337.25556,AUTHORITY[\"EPSG\",\"8814\"]],PARAME"
  "TER[\"Scale factor on initial line\",0.9996,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",2546731.496,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4354009.816,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6497\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6498, 1, 'NAD83(2011) / Michigan South', 'EPSG', 6498, 'PROJCS[\"NAD83(2011) / Michigan South\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3777777"
  "777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.6777777777778,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting"
  " at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\","
  "\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"649"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6499, 1, 'NAD83(2011) / Michigan South (ft)', 'EPSG', 6499, 'PROJCS[\"NAD83(2011) / Michigan S"
  "outh (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-84.3777777777778,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.6777777777778,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",13123359.58,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUT"
  "HORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"6499\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6500, 1, 'NAD83(2011) / Minnesota Central', 'EPSG', 6500, 'PROJCS[\"NAD83(2011) / Minnesota Ce"
  "ntral\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94.25,AUTHORITY["
  "\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6500\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6501, 1, 'NAD83(2011) / Minnesota Central (ftUS)', 'EPSG', 6501, 'PROJCS[\"NAD83(2011) / Minne"
  "sota Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94"
  ".25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",328083.3333,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6501\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6502, 1, 'NAD83(2011) / Minnesota North', 'EPSG', 6502, 'PROJCS[\"NAD83(2011) / Minnesota Nort"
  "h\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.1,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6333333333333,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6502\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6503, 1, 'NAD83(2011) / Minnesota North (ftUS)', 'EPSG', 6503, 'PROJCS[\"NAD83(2011) / Minneso"
  "ta North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",-93.1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6333333333333,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",328083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6503\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6504, 1, 'NAD83(2011) / Minnesota South', 'EPSG', 6504, 'PROJCS[\"NAD83(2011) / Minnesota Sout"
  "h\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6504\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6505, 1, 'NAD83(2011) / Minnesota South (ftUS)', 'EPSG', 6505, 'PROJCS[\"NAD83(2011) / Minneso"
  "ta South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94,AUT"
  "HORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easti"
  "ng at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",328083.3333,AU"
  "THORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6505\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6506, 1, 'NAD83(2011) / Mississippi East', 'EPSG', 6506, 'PROJCS[\"NAD83(2011) / Mississippi E"
  "ast\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6506\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6507, 1, 'NAD83(2011) / Mississippi East (ftUS)', 'EPSG', 6507, 'PROJCS[\"NAD83(2011) / Missis"
  "sippi East (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.83333"
  "33333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"6507\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6508, 1, 'NAD83(2011) / Mississippi TM', 'EPSG', 6508, 'PROJCS[\"NAD83(2011) / Mississippi TM\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.75,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9998335,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1300000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6508\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6509, 1, 'NAD83(2011) / Mississippi West', 'EPSG', 6509, 'PROJCS[\"NAD83(2011) / Mississippi W"
  "est\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.3333333333333,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6509\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6510, 1, 'NAD83(2011) / Mississippi West (ftUS)', 'EPSG', 6510, 'PROJCS[\"NAD83(2011) / Missis"
  "sippi West (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.33333"
  "33333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",2296583.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6510\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6511, 1, 'NAD83(2011) / Missouri Central', 'EPSG', 6511, 'PROJCS[\"NAD83(2011) / Missouri Cent"
  "ral\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"6511\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6512, 1, 'NAD83(2011) / Missouri East', 'EPSG', 6512, 'PROJCS[\"NAD83(2011) / Missouri East\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6512\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6513, 1, 'NAD83(2011) / Missouri West', 'EPSG', 6513, 'PROJCS[\"NAD83(2011) / Missouri West\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-94.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",850000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6513\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6514, 1, 'NAD83(2011) / Montana', 'EPSG', 6514, 'PROJCS[\"NAD83(2011) / Montana\",GEOGCS[\"NAD"
  "83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO"
  "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-109.511111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\""
  "Latitude of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6514\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6515, 1, 'NAD83(2011) / Montana (ft)', 'EPSG', 6515, 'PROJCS[\"NAD83(2011) / Montana (ft)\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-109.51111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNI"
  "T[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6515\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6516, 1, 'NAD83(2011) / Nebraska', 'EPSG', 6516, 'PROJCS[\"NAD83(2011) / Nebraska\",GEOGCS[\"N"
  "AD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUT"
  "HORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG"
  "\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude o"
  "f 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"E"
  "PSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6516\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6518, 1, 'NAD83(2011) / Nevada Central', 'EPSG', 6518, 'PROJCS[\"NAD83(2011) / Nevada Central\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-116.666666666667,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6000000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6518\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6519, 1, 'NAD83(2011) / Nevada Central (ftUS)', 'EPSG', 6519, 'PROJCS[\"NAD83(2011) / Nevada C"
  "entral (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-116.66666666"
  "6667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",19685000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6520, 1, 'NAD83(2011) / Nevada East', 'EPSG', 6520, 'PROJCS[\"NAD83(2011) / Nevada East\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.594444444444,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6520\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6521, 1, 'NAD83(2011) / Nevada East (ftUS)', 'EPSG', 6521, 'PROJCS[\"NAD83(2011) / Nevada East"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.594444444444,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",26246666.6667,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"6521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6522, 1, 'NAD83(2011) / Nevada West', 'EPSG', 6522, 'PROJCS[\"NAD83(2011) / Nevada West\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118.594444444444,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6523, 1, 'NAD83(2011) / Nevada West (ftUS)', 'EPSG', 6523, 'PROJCS[\"NAD83(2011) / Nevada West"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118.594444444444,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",13123333.3333,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6524, 1, 'NAD83(2011) / New Hampshire', 'EPSG', 6524, 'PROJCS[\"NAD83(2011) / New Hampshire\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.6666666666667"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6524\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6525, 1, 'NAD83(2011) / New Hampshire (ftUS)', 'EPSG', 6525, 'PROJCS[\"NAD83(2011) / New Hamps"
  "hire (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71"
  ".6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6525\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6526, 1, 'NAD83(2011) / New Jersey', 'EPSG', 6526, 'PROJCS[\"NAD83(2011) / New Jersey\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural orig"
  "in\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111111111111,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6526\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6527, 1, 'NAD83(2011) / New Jersey (ftUS)', 'EPSG', 6527, 'PROJCS[\"NAD83(2011) / New Jersey ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.51111"
  "11111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6527\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6528, 1, 'NAD83(2011) / New Mexico Central', 'EPSG', 6528, 'PROJCS[\"NAD83(2011) / New Mexico "
  "Central\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-106.25,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6529, 1, 'NAD83(2011) / New Mexico Central (ftUS)', 'EPSG', 6529, 'PROJCS[\"NAD83(2011) / New "
  "Mexico Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-106.25"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"6529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6530, 1, 'NAD83(2011) / New Mexico East', 'EPSG', 6530, 'PROJCS[\"NAD83(2011) / New Mexico Eas"
  "t\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-104.333333333333,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",165000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6530\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6531, 1, 'NAD83(2011) / New Mexico East (ftUS)', 'EPSG', 6531, 'PROJCS[\"NAD83(2011) / New Mex"
  "ico East (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-104.33333333"
  "3333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",541337.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6532, 1, 'NAD83(2011) / New Mexico West', 'EPSG', 6532, 'PROJCS[\"NAD83(2011) / New Mexico Wes"
  "t\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.833333333333,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",830000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6532\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6533, 1, 'NAD83(2011) / New Mexico West (ftUS)', 'EPSG', 6533, 'PROJCS[\"NAD83(2011) / New Mex"
  "ico West (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.83333333"
  "3333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",2723091.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6534, 1, 'NAD83(2011) / New York Central', 'EPSG', 6534, 'PROJCS[\"NAD83(2011) / New York Cent"
  "ral\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5944444444444,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6534\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6535, 1, 'NAD83(2011) / New York Central (ftUS)', 'EPSG', 6535, 'PROJCS[\"NAD83(2011) / New Yo"
  "rk Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5944444"
  "444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",820208.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6536, 1, 'NAD83(2011) / New York East', 'EPSG', 6536, 'PROJCS[\"NAD83(2011) / New York East\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6536\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6537, 1, 'NAD83(2011) / New York East (ftUS)', 'EPSG', 6537, 'PROJCS[\"NAD83(2011) / New York "
  "East (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",492125,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6537\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6538, 1, 'NAD83(2011) / New York Long Island', 'EPSG', 6538, 'PROJCS[\"NAD83(2011) / New York "
  "Long Island\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.0333333333333,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6539, 1, 'NAD83(2011) / New York Long Island (ftUS)', 'EPSG', 6539, 'PROJCS[\"NAD83(2011) / Ne"
  "w York Long Island (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",-74,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.0333333333333,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6540, 1, 'NAD83(2011) / New York West', 'EPSG', 6540, 'PROJCS[\"NAD83(2011) / New York West\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-78.5944444444444,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",350000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6540\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6541, 1, 'NAD83(2011) / New York West (ftUS)', 'EPSG', 6541, 'PROJCS[\"NAD83(2011) / New York "
  "West (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-78.5944444444444"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",1148291.667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6541\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6542, 1, 'NAD83(2011) / North Carolina', 'EPSG', 6542, 'PROJCS[\"NAD83(2011) / North Carolina\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",609601.22,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6542\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6543, 1, 'NAD83(2011) / North Carolina (ftUS)', 'EPSG', 6543, 'PROJCS[\"NAD83(2011) / North Ca"
  "rolina (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79,AUT"
  "HORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easti"
  "ng at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6543\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6544, 1, 'NAD83(2011) / North Dakota North', 'EPSG', 6544, 'PROJCS[\"NAD83(2011) / North Dakot"
  "a North\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.511111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"6544\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6545, 1, 'NAD83(2011) / North Dakota North (ft)', 'EPSG', 6545, 'PROJCS[\"NAD83(2011) / North "
  "Dakota North (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100."
  "511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6545\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6546, 1, 'NAD83(2011) / North Dakota South', 'EPSG', 6546, 'PROJCS[\"NAD83(2011) / North Dakot"
  "a South\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-10"
  "0.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.4944444444444,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,"
  "AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"6546\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6547, 1, 'NAD83(2011) / North Dakota South (ft)', 'EPSG', 6547, 'PROJCS[\"NAD83(2011) / North "
  "Dakota South (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.494444444"
  "4444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.1944444444444,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",1968503.937,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6547\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6548, 1, 'NAD83(2011) / Ohio North', 'EPSG', 6548, 'PROJCS[\"NAD83(2011) / Ohio North\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fa"
  "lse origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6548\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6549, 1, 'NAD83(2011) / Ohio North (ftUS)', 'EPSG', 6549, 'PROJCS[\"NAD83(2011) / Ohio North ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6549\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6550, 1, 'NAD83(2011) / Ohio South', 'EPSG', 6550, 'PROJCS[\"NAD83(2011) / Ohio South\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111,AUTHORITY[\"E"
  "PSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAM"
  "ETER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6550\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6551, 1, 'NAD83(2011) / Ohio South (ftUS)', 'EPSG', 6551, 'PROJCS[\"NAD83(2011) / Ohio South ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6551\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6552, 1, 'NAD83(2011) / Oklahoma North', 'EPSG', 6552, 'PROJCS[\"NAD83(2011) / Oklahoma North\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6552\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6553, 1, 'NAD83(2011) / Oklahoma North (ftUS)', 'EPSG', 6553, 'PROJCS[\"NAD83(2011) / Oklahoma"
  " North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"6553\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6554, 1, 'NAD83(2011) / Oklahoma South', 'EPSG', 6554, 'PROJCS[\"NAD83(2011) / Oklahoma South\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6554\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6555, 1, 'NAD83(2011) / Oklahoma South (ftUS)', 'EPSG', 6555, 'PROJCS[\"NAD83(2011) / Oklahoma"
  " South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.2333333333333,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6555\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6556, 1, 'NAD83(2011) / Oregon LCC (m)', 'EPSG', 6556, 'PROJCS[\"NAD83(2011) / Oregon LCC (m)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTH"
  "ORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6556\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6557, 1, 'NAD83(2011) / Oregon GIC Lambert (ft)', 'EPSG', 6557, 'PROJCS[\"NAD83(2011) / Oregon"
  " GIC Lambert (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-1"
  "20.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1312335.958,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"6557\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6558, 1, 'NAD83(2011) / Oregon North', 'EPSG', 6558, 'PROJCS[\"NAD83(2011) / Oregon North\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.51111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6558\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6559, 1, 'NAD83(2011) / Oregon North (ft)', 'EPSG', 6559, 'PROJCS[\"NAD83(2011) / Oregon North"
  " (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "0.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",8202099.738,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"6559\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6560, 1, 'NAD83(2011) / Oregon South', 'EPSG', 6560, 'PROJCS[\"NAD83(2011) / Oregon South\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.51111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6560\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6561, 1, 'NAD83(2011) / Oregon South (ft)', 'EPSG', 6561, 'PROJCS[\"NAD83(2011) / Oregon South"
  " (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "0.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",4921259.843,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"6561\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6562, 1, 'NAD83(2011) / Pennsylvania North', 'EPSG', 6562, 'PROJCS[\"NAD83(2011) / Pennsylvani"
  "a North\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77"
  ".75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.95,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6562\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6563, 1, 'NAD83(2011) / Pennsylvania North (ftUS)', 'EPSG', 6563, 'PROJCS[\"NAD83(2011) / Penn"
  "sylvania North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.95,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6563\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6564, 1, 'NAD83(2011) / Pennsylvania South', 'EPSG', 6564, 'PROJCS[\"NAD83(2011) / Pennsylvani"
  "a South\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77"
  ".75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.9777777777778,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6564\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6565, 1, 'NAD83(2011) / Pennsylvania South (ftUS)', 'EPSG', 6565, 'PROJCS[\"NAD83(2011) / Penn"
  "sylvania South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.9777777777778,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin"
  "\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6565\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6566, 1, 'NAD83(2011) / Puerto Rico and Virgin Is.', 'EPSG', 6566, 'PROJCS[\"NAD83(2011) / Pue"
  "rto Rico and Virgin Is.\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",17.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-66.4333333333334,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",18.433"
  "3333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",18.0444444444444,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6566\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6567, 1, 'NAD83(2011) / Rhode Island', 'EPSG', 6567, 'PROJCS[\"NAD83(2011) / Rhode Island\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.5111111111111,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999375,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6567\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6568, 1, 'NAD83(2011) / Rhode Island (ftUS)', 'EPSG', 6568, 'PROJCS[\"NAD83(2011) / Rhode Isla"
  "nd (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999375,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6568\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6569, 1, 'NAD83(2011) / South Carolina', 'EPSG', 6569, 'PROJCS[\"NAD83(2011) / South Carolina\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",609600,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6569\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6570, 1, 'NAD83(2011) / South Carolina (ft)', 'EPSG', 6570, 'PROJCS[\"NAD83(2011) / South Caro"
  "lina (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.8333333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6570\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6571, 1, 'NAD83(2011) / South Dakota North', 'EPSG', 6571, 'PROJCS[\"NAD83(2011) / South Dakot"
  "a North\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-10"
  "0,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.6944444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"6571\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6572, 1, 'NAD83(2011) / South Dakota North (ftUS)', 'EPSG', 6572, 'PROJCS[\"NAD83(2011) / Sout"
  "h Dakota North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.6944444444444,AUTHO"
  "RITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6572\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6573, 1, 'NAD83(2011) / South Dakota South', 'EPSG', 6573, 'PROJCS[\"NAD83(2011) / South Dakot"
  "a South\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-10"
  "0.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.4,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6573\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6574, 1, 'NAD83(2011) / South Dakota South (ftUS)', 'EPSG', 6574, 'PROJCS[\"NAD83(2011) / Sout"
  "h Dakota South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\""
  "]],PARAMETER[\"Latitude of false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fal"
  "se origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.4,AUTH"
  "ORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6574\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6575, 1, 'NAD83(2011) / Tennessee', 'EPSG', 6575, 'PROJCS[\"NAD83(2011) / Tennessee\",GEOGCS[\""
  "NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-86,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6575\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6576, 1, 'NAD83(2011) / Tennessee (ftUS)', 'EPSG', 6576, 'PROJCS[\"NAD83(2011) / Tennessee (ft"
  "US)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-86,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"6576\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6577, 1, 'NAD83(2011) / Texas Central', 'EPSG', 6577, 'PROJCS[\"NAD83(2011) / Texas Central\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.333333"
  "333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",31.8833333333333,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AU"
  "THORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"6577\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6578, 1, 'NAD83(2011) / Texas Central (ftUS)', 'EPSG', 6578, 'PROJCS[\"NAD83(2011) / Texas Cen"
  "tral (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAME"
  "TER[\"Latitude of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin"
  "\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",31.8833333333333,A"
  "UTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AUTHORITY[\"EPSG\",\"88"
  "24\"]],PARAMETER[\"Easting at false origin\",2296583.333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false "
  "origin\",9842500,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6578\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6579, 1, 'NAD83(2011) / Texas Centric Albers Equal Area', 'EPSG', 6579, 'PROJCS[\"NAD83(2011) "
  "/ Texas Centric Albers Equal Area\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\""
  "]],PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-1"
  "00,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.5,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",6000000,AUTHORITY[\"EPSG\",\"8827\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6579\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6580, 1, 'NAD83(2011) / Texas Centric Lambert Conformal', 'EPSG', 6580, 'PROJCS[\"NAD83(2011) "
  "/ Texas Centric Lambert Conformal\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",18,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.5,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fa"
  "lse origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000000,AUTHORITY[\"EPSG\","
  "\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"658"
  "0\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6581, 1, 'NAD83(2011) / Texas North', 'EPSG', 6581, 'PROJCS[\"NAD83(2011) / Texas North\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-101.511111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6581\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6582, 1, 'NAD83(2011) / Texas North (ftUS)', 'EPSG', 6582, 'PROJCS[\"NAD83(2011) / Texas North"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-101.511111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",656166.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3280833.333,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6582\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6583, 1, 'NAD83(2011) / Texas North Central', 'EPSG', 6583, 'PROJCS[\"NAD83(2011) / Texas Nort"
  "h Central\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.9777777777778,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.1444444444444,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6583\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6584, 1, 'NAD83(2011) / Texas North Central (ftUS)', 'EPSG', 6584, 'PROJCS[\"NAD83(2011) / Tex"
  "as North Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.977"
  "7777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.1444444444444,AUTHORITY[\""
  "EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at"
  " false origin\",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6584\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6585, 1, 'NAD83(2011) / Texas South', 'EPSG', 6585, 'PROJCS[\"NAD83(2011) / Texas South\",GEOG"
  "CS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.51111111111"
  "11,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000000,AUTHORI"
  "TY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6585\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6586, 1, 'NAD83(2011) / Texas South (ftUS)', 'EPSG', 6586, 'PROJCS[\"NAD83(2011) / Texas South"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",984250,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",16"
  "404166.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6586\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6587, 1, 'NAD83(2011) / Texas South Central', 'EPSG', 6587, 'PROJCS[\"NAD83(2011) / Texas Sout"
  "h Central\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.2833333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4000000,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"6587\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6588, 1, 'NAD83(2011) / Texas South Central (ftUS)', 'EPSG', 6588, 'PROJCS[\"NAD83(2011) / Tex"
  "as South Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.2833333333333,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin"
  "\",13123333.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6588\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6589, 1, 'NAD83(2011) / Vermont', 'EPSG', 6589, 'PROJCS[\"NAD83(2011) / Vermont\",GEOGCS[\"NAD"
  "83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHO"
  "RITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6589\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6590, 1, 'NAD83(2011) / Vermont (ftUS)', 'EPSG', 6590, 'PROJCS[\"NAD83(2011) / Vermont (ftUS)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72.5111111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6590\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6591, 1, 'NAD83(2011) / Virginia Lambert', 'EPSG', 6591, 'PROJCS[\"NAD83(2011) / Virginia Lamb"
  "ert\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79.5,AUTHORITY[\"E"
  "PSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",39.5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6591\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6592, 1, 'NAD83(2011) / Virginia North', 'EPSG', 6592, 'PROJCS[\"NAD83(2011) / Virginia North\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.2111111111111,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2000000,A"
  "UTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"6592\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6593, 1, 'NAD83(2011) / Virginia North (ftUS)', 'EPSG', 6593, 'PROJCS[\"NAD83(2011) / Virginia"
  " North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.2111111111111,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",6561666.667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6593\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6594, 1, 'NAD83(2011) / Virginia South', 'EPSG', 6594, 'PROJCS[\"NAD83(2011) / Virginia South\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111"
  "111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\"EPS"
  "G\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMET"
  "ER[\"Easting at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,A"
  "UTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"6594\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6595, 1, 'NAD83(2011) / Virginia South (ftUS)', 'EPSG', 6595, 'PROJCS[\"NAD83(2011) / Virginia"
  " South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origi"
  "n\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",11482916.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fals"
  "e origin\",3280833.333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6595\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6596, 1, 'NAD83(2011) / Washington North', 'EPSG', 6596, 'PROJCS[\"NAD83(2011) / Washington No"
  "rth\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.833333333333,A"
  "UTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"88"
  "23\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Eas"
  "ting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "6596\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6597, 1, 'NAD83(2011) / Washington North (ftUS)', 'EPSG', 6597, 'PROJCS[\"NAD83(2011) / Washin"
  "gton North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120."
  "833333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6597\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6598, 1, 'NAD83(2011) / Washington South', 'EPSG', 6598, 'PROJCS[\"NAD83(2011) / Washington So"
  "uth\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"La"
  "titude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120."
  "511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.3333333333333,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AU"
  "THORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"6598\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6599, 1, 'NAD83(2011) / Washington South (ftUS)', 'EPSG', 6599, 'PROJCS[\"NAD83(2011) / Washin"
  "gton South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.333333333"
  "3333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.8333333333333,AUTHORITY[\"EPSG\""
  ",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6599\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6600, 1, 'NAD83(2011) / West Virginia North', 'EPSG', 6600, 'PROJCS[\"NAD83(2011) / West Virgi"
  "nia North\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.2611111111111,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER"
  "[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"6600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6601, 1, 'NAD83(2011) / West Virginia North (ftUS)', 'EPSG', 6601, 'PROJCS[\"NAD83(2011) / Wes"
  "t Virginia North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of "
  "false origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.261"
  "1111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTHORITY[\"EPSG\",\"8824"
  "\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin"
  "\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6601\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6602, 1, 'NAD83(2011) / West Virginia South', 'EPSG', 6602, 'PROJCS[\"NAD83(2011) / West Virgi"
  "nia South\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6602\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6603, 1, 'NAD83(2011) / West Virginia South (ftUS)', 'EPSG', 6603, 'PROJCS[\"NAD83(2011) / Wes"
  "t Virginia South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"980"
  "2\"]],PARAMETER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333333,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY["
  "\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"6603\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6605, 1, 'NAD83(2011) / Wisconsin Central (ftUS)', 'EPSG', 6605, 'PROJCS[\"NAD83(2011) / Wisco"
  "nsin Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.5111111111111,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6605\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6606, 1, 'NAD83(2011) / Wisconsin North', 'EPSG', 6606, 'PROJCS[\"NAD83(2011) / Wisconsin Nort"
  "h\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUT"
  "HORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.7666666666667,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easti"
  "ng at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"66"
  "06\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6607, 1, 'NAD83(2011) / Wisconsin North (ftUS)', 'EPSG', 6607, 'PROJCS[\"NAD83(2011) / Wiscons"
  "in North (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false or"
  "igin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.7666666666667,AUTHORITY[\""
  "EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARA"
  "METER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6607\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6608, 1, 'NAD83(2011) / Wisconsin South', 'EPSG', 6608, 'PROJCS[\"NAD83(2011) / Wisconsin Sout"
  "h\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6608\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6609, 1, 'NAD83(2011) / Wisconsin South (ftUS)', 'EPSG', 6609, 'PROJCS[\"NAD83(2011) / Wiscons"
  "in South (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PA"
  "RAMETER[\"Latitude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUT"
  "HORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easti"
  "ng at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6609\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6610, 1, 'NAD83(2011) / Wisconsin Transverse Mercator', 'EPSG', 6610, 'PROJCS[\"NAD83(2011) / "
  "Wisconsin Transverse Mercator\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",520000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-4480000,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "6610\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6611, 1, 'NAD83(2011) / Wyoming East', 'EPSG', 6611, 'PROJCS[\"NAD83(2011) / Wyoming East\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105.177777777778,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6611\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6612, 1, 'NAD83(2011) / Wyoming East (ftUS)', 'EPSG', 6612, 'PROJCS[\"NAD83(2011) / Wyoming Ea"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105."
  "177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",656166.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6612\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6613, 1, 'NAD83(2011) / Wyoming East Central', 'EPSG', 6613, 'PROJCS[\"NAD83(2011) / Wyoming E"
  "ast Central\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-10"
  "7.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"6613\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6614, 1, 'NAD83(2011) / Wyoming East Central (ftUS)', 'EPSG', 6614, 'PROJCS[\"NAD83(2011) / Wy"
  "oming East Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-107.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1312333.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",328083.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6614\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6615, 1, 'NAD83(2011) / Wyoming West', 'EPSG', 6615, 'PROJCS[\"NAD83(2011) / Wyoming West\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.094444444444,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"661"
  "5\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6616, 1, 'NAD83(2011) / Wyoming West (ftUS)', 'EPSG', 6616, 'PROJCS[\"NAD83(2011) / Wyoming We"
  "st (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110."
  "094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",2624666.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",328083"
  ".3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6616\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6617, 1, 'NAD83(2011) / Wyoming West Central', 'EPSG', 6617, 'PROJCS[\"NAD83(2011) / Wyoming W"
  "est Central\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-10"
  "8.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"66"
  "17\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6618, 1, 'NAD83(2011) / Wyoming West Central (ftUS)', 'EPSG', 6618, 'PROJCS[\"NAD83(2011) / Wy"
  "oming West Central (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-108.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",1968500,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6618\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6619, 1, 'NAD83(2011) / Utah Central', 'EPSG', 6619, 'PROJCS[\"NAD83(2011) / Utah Central\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.51111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2000000,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "6619\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6620, 1, 'NAD83(2011) / Utah North', 'EPSG', 6620, 'PROJCS[\"NAD83(2011) / Utah North\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6620\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6621, 1, 'NAD83(2011) / Utah South', 'EPSG', 6621, 'PROJCS[\"NAD83(2011) / Utah South\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHORITY[\"EPSG\",\"8823\"]],PA"
  "RAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHORITY[\"EPSG\","
  "\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"662"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6622, 1, 'NAD83(CSRS) / Quebec Lambert', 'EPSG', 6622, 'PROJCS[\"NAD83(CSRS) / Quebec Lambert\""
  ",GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421"
  ",0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION["
  "\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",60,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6622\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6623, 1, 'NAD83 / Quebec Albers', 'EPSG', 6623, 'PROJCS[\"NAD83 / Quebec Albers\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",44,AUTHO"
  "RITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARA"
  "METER[\"Latitude of 1st standard parallel\",60,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard par"
  "allel\",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6623\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6624, 1, 'NAD83(CSRS) / Quebec Albers', 'EPSG', 6624, 'PROJCS[\"NAD83(CSRS) / Quebec Albers\","
  "GEOGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,"
  "AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,"
  "0],AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\""
  "Albers Equal Area\",AUTHORITY[\"EPSG\",\"9822\"]],PARAMETER[\"Latitude of false origin\",44,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",60,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fa"
  "lse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6624\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6625, 1, 'NAD83(2011) / Utah Central (ftUS)', 'EPSG', 6625, 'PROJCS[\"NAD83(2011) / Utah Centr"
  "al (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETE"
  "R[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUTHORITY[\"EP"
  "SG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAME"
  "TER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",656"
  "1666.6667,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6625\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6626, 1, 'NAD83(2011) / Utah North (ftUS)', 'EPSG', 6626, 'PROJCS[\"NAD83(2011) / Utah North ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "1.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origi"
  "n\",3280833.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6626\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6627, 1, 'NAD83(2011) / Utah South (ftUS)', 'EPSG', 6627, 'PROJCS[\"NAD83(2011) / Utah South ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-11"
  "1.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",9842500"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6627\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6628, 1, 'NAD83(PA11) / Hawaii zone 1', 'EPSG', 6628, 'PROJCS[\"NAD83(PA11) / Hawaii zone 1\","
  "GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",18.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-155.5,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6628\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6629, 1, 'NAD83(PA11) / Hawaii zone 2', 'EPSG', 6629, 'PROJCS[\"NAD83(PA11) / Hawaii zone 2\","
  "GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",20.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-156.666666666667"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6629\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6630, 1, 'NAD83(PA11) / Hawaii zone 3', 'EPSG', 6630, 'PROJCS[\"NAD83(PA11) / Hawaii zone 3\","
  "GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",21.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-158,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6630\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6631, 1, 'NAD83(PA11) / Hawaii zone 4', 'EPSG', 6631, 'PROJCS[\"NAD83(PA11) / Hawaii zone 4\","
  "GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",21.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159.5,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6631\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6632, 1, 'NAD83(PA11) / Hawaii zone 5', 'EPSG', 6632, 'PROJCS[\"NAD83(PA11) / Hawaii zone 5\","
  "GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",21.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-160.177777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6632\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6633, 1, 'NAD83(PA11) / Hawaii zone 3 (ftUS)', 'EPSG', 6633, 'PROJCS[\"NAD83(PA11) / Hawaii zo"
  "ne 3 (ftUS)\",GEOGCS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",21.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15"
  "8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",1640416.6667,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6633\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6634, 1, 'NAD83(PA11) / UTM zone 4N', 'EPSG', 6634, 'PROJCS[\"NAD83(PA11) / UTM zone 4N\",GEOG"
  "CS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6634\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6635, 1, 'NAD83(PA11) / UTM zone 5N', 'EPSG', 6635, 'PROJCS[\"NAD83(PA11) / UTM zone 5N\",GEOG"
  "CS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6635\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6636, 1, 'NAD83(PA11) / UTM zone 2S', 'EPSG', 6636, 'PROJCS[\"NAD83(PA11) / UTM zone 2S\",GEOG"
  "CS[\"NAD83(PA11)\",DATUM[\"NAD83 (National Spatial Reference System PA11)\",SPHEROID[\"GRS 1980\",6378137,298.2572221"
  "01,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1117\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"6322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,A"
  "UTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6636\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6637, 1, 'NAD83(MA11) / Guam Map Grid', 'EPSG', 6637, 'PROJCS[\"NAD83(MA11) / Guam Map Grid\","
  "GEOGCS[\"NAD83(MA11)\",DATUM[\"NAD83 (National Spatial Reference System MA11)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1118\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6325\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",13.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",144.761111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6637\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6646, 1, 'Karbala 1979 / Iraq National Grid', 'EPSG', 6646, 'PROJCS[\"Karbala 1979 / Iraq Nati"
  "onal Grid\",GEOGCS[\"Karbala 1979\",DATUM[\"Karbala 1979\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORI"
  "TY[\"EPSG\",\"7012\"]],TOWGS84[70.995,-335.916,262.898,0,0,0,0],AUTHORITY[\"EPSG\",\"6743\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4743\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",29.0262683333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9994,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"6646\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6668, 1, 'JGD2011', 'EPSG', 6668, 'GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6668\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6669, 1, 'JGD2011 / Japan Plane Rectangular CS I', 'EPSG', 6669, 'PROJCS[\"JGD2011 / Japan Pla"
  "ne Rectangular CS I\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129.5,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6669\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6670, 1, 'JGD2011 / Japan Plane Rectangular CS II', 'EPSG', 6670, 'PROJCS[\"JGD2011 / Japan Pl"
  "ane Rectangular CS II\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6670\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6671, 1, 'JGD2011 / Japan Plane Rectangular CS III', 'EPSG', 6671, 'PROJCS[\"JGD2011 / Japan P"
  "lane Rectangular CS III\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132.177777777778,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6671\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6672, 1, 'JGD2011 / Japan Plane Rectangular CS IV', 'EPSG', 6672, 'PROJCS[\"JGD2011 / Japan Pl"
  "ane Rectangular CS IV\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",133.5,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6672\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6673, 1, 'JGD2011 / Japan Plane Rectangular CS V', 'EPSG', 6673, 'PROJCS[\"JGD2011 / Japan Pla"
  "ne Rectangular CS V\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",134.344444444444,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6673\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6674, 1, 'JGD2011 / Japan Plane Rectangular CS VI', 'EPSG', 6674, 'PROJCS[\"JGD2011 / Japan Pl"
  "ane Rectangular CS VI\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",136,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6674\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6675, 1, 'JGD2011 / Japan Plane Rectangular CS VII', 'EPSG', 6675, 'PROJCS[\"JGD2011 / Japan P"
  "lane Rectangular CS VII\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",137.177777777778,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6675\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6676, 1, 'JGD2011 / Japan Plane Rectangular CS VIII', 'EPSG', 6676, 'PROJCS[\"JGD2011 / Japan "
  "Plane Rectangular CS VIII\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",138.5,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6676\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6677, 1, 'JGD2011 / Japan Plane Rectangular CS IX', 'EPSG', 6677, 'PROJCS[\"JGD2011 / Japan Pl"
  "ane Rectangular CS IX\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",139.833333333333,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6677\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6678, 1, 'JGD2011 / Japan Plane Rectangular CS X', 'EPSG', 6678, 'PROJCS[\"JGD2011 / Japan Pla"
  "ne Rectangular CS X\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",140.833333333333,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6678\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6679, 1, 'JGD2011 / Japan Plane Rectangular CS XI', 'EPSG', 6679, 'PROJCS[\"JGD2011 / Japan Pl"
  "ane Rectangular CS XI\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",140.25,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6679\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6680, 1, 'JGD2011 / Japan Plane Rectangular CS XII', 'EPSG', 6680, 'PROJCS[\"JGD2011 / Japan P"
  "lane Rectangular CS XII\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",142.25,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6680\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6681, 1, 'JGD2011 / Japan Plane Rectangular CS XIII', 'EPSG', 6681, 'PROJCS[\"JGD2011 / Japan "
  "Plane Rectangular CS XIII\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",144.25,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6681\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6682, 1, 'JGD2011 / Japan Plane Rectangular CS XIV', 'EPSG', 6682, 'PROJCS[\"JGD2011 / Japan P"
  "lane Rectangular CS XIV\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",142,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6682\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6683, 1, 'JGD2011 / Japan Plane Rectangular CS XV', 'EPSG', 6683, 'PROJCS[\"JGD2011 / Japan Pl"
  "ane Rectangular CS XV\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.2"
  "57222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127.511111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6683\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6684, 1, 'JGD2011 / Japan Plane Rectangular CS XVI', 'EPSG', 6684, 'PROJCS[\"JGD2011 / Japan P"
  "lane Rectangular CS XVI\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",124,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6684\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6685, 1, 'JGD2011 / Japan Plane Rectangular CS XVII', 'EPSG', 6685, 'PROJCS[\"JGD2011 / Japan "
  "Plane Rectangular CS XVII\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6685\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6686, 1, 'JGD2011 / Japan Plane Rectangular CS XVIII', 'EPSG', 6686, 'PROJCS[\"JGD2011 / Japan"
  " Plane Rectangular CS XVIII\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",20,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",136,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6686\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6687, 1, 'JGD2011 / Japan Plane Rectangular CS XIX', 'EPSG', 6687, 'PROJCS[\"JGD2011 / Japan P"
  "lane Rectangular CS XIX\",GEOGCS[\"JGD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6668\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",154,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6687\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6688, 1, 'JGD2011 / UTM zone 51N', 'EPSG', 6688, 'PROJCS[\"JGD2011 / UTM zone 51N\",GEOGCS[\"J"
  "GD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6668\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6688\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6689, 1, 'JGD2011 / UTM zone 52N', 'EPSG', 6689, 'PROJCS[\"JGD2011 / UTM zone 52N\",GEOGCS[\"J"
  "GD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6668\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6689\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6690, 1, 'JGD2011 / UTM zone 53N', 'EPSG', 6690, 'PROJCS[\"JGD2011 / UTM zone 53N\",GEOGCS[\"J"
  "GD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6668\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6690\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6691, 1, 'JGD2011 / UTM zone 54N', 'EPSG', 6691, 'PROJCS[\"JGD2011 / UTM zone 54N\",GEOGCS[\"J"
  "GD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6668\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6691\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6692, 1, 'JGD2011 / UTM zone 55N', 'EPSG', 6692, 'PROJCS[\"JGD2011 / UTM zone 55N\",GEOGCS[\"J"
  "GD2011\",DATUM[\"Japanese Geodetic Datum 2011\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],AUTHORITY[\"EPSG\",\"1128\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251"
  "9943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6668\"]],PROJECTION"
  "[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6703, 1, 'WGS 84 / TM 60 SW', 'EPSG', 6703, 'PROJCS[\"WGS 84 / TM 60 SW\",GEOGCS[\"WGS 84\",DA"
  "TUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY["
  "\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",-60,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\","
  "0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],"
  "AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6703\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6706, 1, 'RDN2008', 'EPSG', 6706, 'GEOGCS[\"RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6706\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6707, 1, 'RDN2008 / UTM zone 32N (N-E)', 'EPSG', 6707, 'PROJCS[\"RDN2008 / UTM zone 32N (N-E)\""
  ",GEOGCS[\"RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6707\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6708, 1, 'RDN2008 / UTM zone 33N (N-E)', 'EPSG', 6708, 'PROJCS[\"RDN2008 / UTM zone 33N (N-E)\""
  ",GEOGCS[\"RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6708\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6709, 1, 'RDN2008 / UTM zone 34N (N-E)', 'EPSG', 6709, 'PROJCS[\"RDN2008 / UTM zone 34N (N-E)\""
  ",GEOGCS[\"RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6709\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6720, 1, 'WGS 84 / CIG92', 'EPSG', 6720, 'PROJCS[\"WGS 84 / CIG92\",GEOGCS[\"WGS 84\",DATUM[\""
  "World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG"
  "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercato"
  "r\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER["
  "\"Longitude of natural origin\",105.625,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "000024,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",1300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AX"
  "IS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6720\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6721, 1, 'GDA94 / CIG94', 'EPSG', 6721, 'PROJCS[\"GDA94 / CIG94\",GEOGCS[\"GDA94\",DATUM[\"Geo"
  "centric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0"
  ",0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY["
  "\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105.625,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",1.00002514,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",1300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6721\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6722, 1, 'WGS 84 / CKIG92', 'EPSG', 6722, 'PROJCS[\"WGS 84 / CKIG92\",GEOGCS[\"WGS 84\",DATUM["
  "\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP"
  "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",96.875,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",1400000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6722\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6723, 1, 'GDA94 / CKIG94', 'EPSG', 6723, 'PROJCS[\"GDA94 / CKIG94\",GEOGCS[\"GDA94\",DATUM[\"G"
  "eocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84"
  "[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96.875,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.99999387,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6723\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6736, 1, 'GDA94 / MGA zone 46', 'EPSG', 6736, 'PROJCS[\"GDA94 / MGA zone 46\",GEOGCS[\"GDA94\""
  ",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6736\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6737, 1, 'GDA94 / MGA zone 47', 'EPSG', 6737, 'PROJCS[\"GDA94 / MGA zone 47\",GEOGCS[\"GDA94\""
  ",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6737\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6738, 1, 'GDA94 / MGA zone 59', 'EPSG', 6738, 'PROJCS[\"GDA94 / MGA zone 59\",GEOGCS[\"GDA94\""
  ",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6783, 1, 'NAD83(CORS96)', 'EPSG', 6783, 'GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously "
  "Operating Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI"
  "TY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6784, 1, 'NAD83(CORS96) / Oregon Baker zone (m)', 'EPSG', 6784, 'PROJCS[\"NAD83(CORS96) / Oreg"
  "on Baker zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID"
  "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",-117.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00016,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6784\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6785, 1, 'NAD83(CORS96) / Oregon Baker zone (ft)', 'EPSG', 6785, 'PROJCS[\"NAD83(CORS96) / Ore"
  "gon Baker zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHERO"
  "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-117.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00016,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6785\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6786, 1, 'NAD83(2011) / Oregon Baker zone (m)', 'EPSG', 6786, 'PROJCS[\"NAD83(2011) / Oregon B"
  "aker zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "17.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00016,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"6786\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6787, 1, 'NAD83(2011) / Oregon Baker zone (ft)', 'EPSG', 6787, 'PROJCS[\"NAD83(2011) / Oregon "
  "Baker zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-117.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00016,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6787\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6788, 1, 'NAD83(CORS96) / Oregon Bend-Klamath Falls zone (m)', 'EPSG', 6788, 'PROJCS[\"NAD83(C"
  "ORS96) / Oregon Bend-Klamath Falls zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Referenc"
  "e Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-121.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0002,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"6788\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6789, 1, 'NAD83(CORS96) / Oregon Bend-Klamath Falls zone (ft)', 'EPSG', 6789, 'PROJCS[\"NAD83("
  "CORS96) / Oregon Bend-Klamath Falls zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refere"
  "nce Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"11"
  "33\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-121.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0002"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6789\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6790, 1, 'NAD83(2011) / Oregon Bend-Klamath Falls zone (m)', 'EPSG', 6790, 'PROJCS[\"NAD83(201"
  "1) / Oregon Bend-Klamath Falls zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 201"
  "1)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-121.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0002,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6790\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6791, 1, 'NAD83(2011) / Oregon Bend-Klamath Falls zone (ft)', 'EPSG', 6791, 'PROJCS[\"NAD83(20"
  "11) / Oregon Bend-Klamath Falls zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-121.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0002,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"6791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6792, 1, 'NAD83(CORS96) / Oregon Bend-Redmond-Prineville zone (m)', 'EPSG', 6792, 'PROJCS[\"NA"
  "D83(CORS96) / Oregon Bend-Redmond-Prineville zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operatin"
  "g Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Con"
  "formal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6777777777778,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-121.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",130000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6792\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6793, 1, 'NAD83(CORS96) / Oregon Bend-Redmond-Prineville zone (ft)', 'EPSG', 6793, 'PROJCS[\"N"
  "AD83(CORS96) / Oregon Bend-Redmond-Prineville zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operat"
  "ing Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"E"
  "PSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic C"
  "onformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6777777777778,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-121.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",426509.1864,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORIT"
  "Y[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6793\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6794, 1, 'NAD83(2011) / Oregon Bend-Redmond-Prineville zone (m)', 'EPSG', 6794, 'PROJCS[\"NAD8"
  "3(2011) / Oregon Bend-Redmond-Prineville zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference "
  "System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\""
  ",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]"
  "],PARAMETER[\"Longitude of natural origin\",-121.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",130000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6794\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6795, 1, 'NAD83(2011) / Oregon Bend-Redmond-Prineville zone (ft)', 'EPSG', 6795, 'PROJCS[\"NAD"
  "83(2011) / Oregon Bend-Redmond-Prineville zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Referenc"
  "e System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)"
  "\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6777777777778,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-121.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",426509.1864,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"90"
  "02\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6795\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6796, 1, 'NAD83(CORS96) / Oregon Bend-Burns zone (m)', 'EPSG', 6796, 'PROJCS[\"NAD83(CORS96) /"
  " Oregon Bend-Burns zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\""
  "EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-119.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00"
  "02,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",60000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"6796\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6797, 1, 'NAD83(CORS96) / Oregon Bend-Burns zone (ft)', 'EPSG', 6797, 'PROJCS[\"NAD83(CORS96) "
  "/ Oregon Bend-Burns zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996"
  ")\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY["
  "\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-119.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1"
  ".0002,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",393700.7874,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",196850.3937,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6797\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6798, 1, 'NAD83(2011) / Oregon Bend-Burns zone (m)', 'EPSG', 6798, 'PROJCS[\"NAD83(2011) / Ore"
  "gon Bend-Burns zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-119.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0002,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6"
  "0000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"6798\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6799, 1, 'NAD83(2011) / Oregon Bend-Burns zone (ft)', 'EPSG', 6799, 'PROJCS[\"NAD83(2011) / Or"
  "egon Bend-Burns zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-119.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0002,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",393700.7874,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",196850.3937,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6799\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6800, 1, 'NAD83(CORS96) / Oregon Canyonville-Grants Pass zone (m)', 'EPSG', 6800, 'PROJCS[\"NA"
  "D83(CORS96) / Oregon Canyonville-Grants Pass zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operatin"
  "g Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",1.00007,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6800\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6801, 1, 'NAD83(CORS96) / Oregon Canyonville-Grants Pass zone (ft)', 'EPSG', 6801, 'PROJCS[\"N"
  "AD83(CORS96) / Oregon Canyonville-Grants Pass zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operat"
  "ing Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"E"
  "PSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Merc"
  "ator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",1.00007,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\""
  ",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6801\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6802, 1, 'NAD83(2011) / Oregon Canyonville-Grants Pass zone (m)', 'EPSG', 6802, 'PROJCS[\"NAD8"
  "3(2011) / Oregon Canyonville-Grants Pass zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference "
  "System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",1.00007,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6802\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6803, 1, 'NAD83(2011) / Oregon Canyonville-Grants Pass zone (ft)', 'EPSG', 6803, 'PROJCS[\"NAD"
  "83(2011) / Oregon Canyonville-Grants Pass zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Referenc"
  "e System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",1.00007,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6803\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6804, 1, 'NAD83(CORS96) / Oregon Columbia River East zone (m)', 'EPSG', 6804, 'PROJCS[\"NAD83("
  "CORS96) / Oregon Columbia River East zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refere"
  "nce Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"11"
  "33\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (1"
  "SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1.000008,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",30000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6804\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6805, 1, 'NAD83(CORS96) / Oregon Columbia River East zone (ft)', 'EPSG', 6805, 'PROJCS[\"NAD83"
  "(CORS96) / Oregon Columbia River East zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refe"
  "rence Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal"
  " (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.6777777777778,AUTHORITY[\"EPSG\",\""
  "8801\"]],PARAMETER[\"Longitude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1.000008,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",492125.9843,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",98425.1969,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORIT"
  "Y[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6805\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6806, 1, 'NAD83(2011) / Oregon Columbia River East zone (m)', 'EPSG', 6806, 'PROJCS[\"NAD83(20"
  "11) / Oregon Columbia River East zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAM"
  "ETER[\"Longitude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natu"
  "ral origin\",1.000008,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",30000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6806\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6807, 1, 'NAD83(2011) / Oregon Columbia River East zone (ft)', 'EPSG', 6807, 'PROJCS[\"NAD83(2"
  "011) / Oregon Columbia River East zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTH"
  "ORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",1.000008,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",492125.9843,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",98425.1969,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\","
  "\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6807\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6808, 1, 'NAD83(CORS96) / Oregon Columbia River West zone (m)', 'EPSG', 6808, 'PROJCS[\"NAD83("
  "CORS96) / Oregon Columbia River West zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refere"
  "nce Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"11"
  "33\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Hotine Oblique Mercator (v"
  "ariant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",45.9277777777778,AUTHORITY[\"EP"
  "SG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",-123,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth o"
  "f initial line\",295,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",295,AUTHORITY[\"EP"
  "SG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\""
  ",7000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-3000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6808\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6809, 1, 'NAD83(CORS96) / Oregon Columbia River West zone (ft)', 'EPSG', 6809, 'PROJCS[\"NAD83"
  "(CORS96) / Oregon Columbia River West zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refe"
  "rence Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Hotine Oblique Mercator"
  " (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",45.9277777777778,AUTHORITY[\""
  "EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",-123,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth"
  " of initial line\",295,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",295,AUTHORITY[\""
  "EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False eastin"
  "g\",22965879.2651,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-9842519.685,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6809\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6810, 1, 'NAD83(2011) / Oregon Columbia River West zone (m)', 'EPSG', 6810, 'PROJCS[\"NAD83(20"
  "11) / Oregon Columbia River West zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\","
  "AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",45.9277777777778,AUTHORITY[\"EPSG\",\"8811\""
  "]],PARAMETER[\"Longitude of projection centre\",-123,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial lin"
  "e\",295,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",295,AUTHORITY[\"EPSG\",\"8814\""
  "]],PARAMETER[\"Scale factor on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",7000000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-3000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6810\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6811, 1, 'NAD83(2011) / Oregon Columbia River West zone (ft)', 'EPSG', 6811, 'PROJCS[\"NAD83(2"
  "011) / Oregon Columbia River West zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\""
  ",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",45.9277777777778,AUTHORITY[\"EPSG\",\"8811"
  "\"]],PARAMETER[\"Longitude of projection centre\",-123,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial l"
  "ine\",295,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",295,AUTHORITY[\"EPSG\",\"8814"
  "\"]],PARAMETER[\"Scale factor on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",22965879."
  "2651,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-9842519.685,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foo"
  "t\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6811\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6812, 1, 'NAD83(CORS96) / Oregon Cottage Grove-Canyonville zone (m)', 'EPSG', 6812, 'PROJCS[\""
  "NAD83(CORS96) / Oregon Cottage Grove-Canyonville zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Oper"
  "ating Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\""
  "EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\""
  "8801\"]],PARAMETER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1.000023,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6812\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6813, 1, 'NAD83(CORS96) / Oregon Cottage Grove-Canyonville zone (ft)', 'EPSG', 6813, 'PROJCS[\""
  "NAD83(CORS96) / Oregon Cottage Grove-Canyonville zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Ope"
  "rating Reference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY["
  "\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse M"
  "ercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\","
  "\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1.000023,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EP"
  "SG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6813\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6814, 1, 'NAD83(2011) / Oregon Cottage Grove-Canyonville zone (m)', 'EPSG', 6814, 'PROJCS[\"NA"
  "D83(2011) / Oregon Cottage Grove-Canyonville zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Refere"
  "nce System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"111"
  "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",1.000023,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6814\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6815, 1, 'NAD83(2011) / Oregon Cottage Grove-Canyonville zone (ft)', 'EPSG', 6815, 'PROJCS[\"N"
  "AD83(2011) / Oregon Cottage Grove-Canyonville zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Refe"
  "rence System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1"
  "116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AU"
  "THORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at "
  "natural origin\",1.000023,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6815\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6816, 1, 'NAD83(CORS96) / Oregon Dufur-Madras zone (m)', 'EPSG', 6816, 'PROJCS[\"NAD83(CORS96)"
  " / Oregon Dufur-Madras zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 19"
  "96)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00011,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6816\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6817, 1, 'NAD83(CORS96) / Oregon Dufur-Madras zone (ft)', 'EPSG', 6817, 'PROJCS[\"NAD83(CORS96"
  ") / Oregon Dufur-Madras zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station "
  "1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00011,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"6817\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6818, 1, 'NAD83(2011) / Oregon Dufur-Madras zone (m)', 'EPSG', 6818, 'PROJCS[\"NAD83(2011) / O"
  "regon Dufur-Madras zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00011,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6818\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6819, 1, 'NAD83(2011) / Oregon Dufur-Madras zone (ft)', 'EPSG', 6819, 'PROJCS[\"NAD83(2011) / "
  "Oregon Dufur-Madras zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",44.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-121,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00011,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"6819\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6820, 1, 'NAD83(CORS96) / Oregon Eugene zone (m)', 'EPSG', 6820, 'PROJCS[\"NAD83(CORS96) / Ore"
  "gon Eugene zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHERO"
  "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",43.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-123.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6820\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6821, 1, 'NAD83(CORS96) / Oregon Eugene zone (ft)', 'EPSG', 6821, 'PROJCS[\"NAD83(CORS96) / Or"
  "egon Eugene zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",43.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-123.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6821\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6822, 1, 'NAD83(2011) / Oregon Eugene zone (m)', 'EPSG', 6822, 'PROJCS[\"NAD83(2011) / Oregon "
  "Eugene zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",43.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123.17777"
  "7777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"68"
  "22\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6823, 1, 'NAD83(2011) / Oregon Eugene zone (ft)', 'EPSG', 6823, 'PROJCS[\"NAD83(2011) / Oregon"
  " Eugene zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",43.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123.177"
  "777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"6823\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6824, 1, 'NAD83(CORS96) / Oregon Grants Pass-Ashland zone (m)', 'EPSG', 6824, 'PROJCS[\"NAD83("
  "CORS96) / Oregon Grants Pass-Ashland zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refere"
  "nce Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"11"
  "33\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",1.000043,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6824\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6825, 1, 'NAD83(CORS96) / Oregon Grants Pass-Ashland zone (ft)', 'EPSG', 6825, 'PROJCS[\"NAD83"
  "(CORS96) / Oregon Grants Pass-Ashland zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refe"
  "rence Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1.000043,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PA"
  "RAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6825\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6826, 1, 'NAD83(2011) / Oregon Grants Pass-Ashland zone (m)', 'EPSG', 6826, 'PROJCS[\"NAD83(20"
  "11) / Oregon Grants Pass-Ashland zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000043"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"6826\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6827, 1, 'NAD83(2011) / Oregon Grants Pass-Ashland zone (ft)', 'EPSG', 6827, 'PROJCS[\"NAD83(2"
  "011) / Oregon Grants Pass-Ashland zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-123.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "43,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6827\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6828, 1, 'NAD83(CORS96) / Oregon Gresham-Warm Springs zone (m)', 'EPSG', 6828, 'PROJCS[\"NAD83"
  "(CORS96) / Oregon Gresham-Warm Springs zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refe"
  "rence Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-122.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",1.00005,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",10000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6828\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6829, 1, 'NAD83(CORS96) / Oregon Gresham-Warm Springs zone (ft)', 'EPSG', 6829, 'PROJCS[\"NAD8"
  "3(CORS96) / Oregon Gresham-Warm Springs zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Re"
  "ference Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\","
  "\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-122.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1.00005,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",32808.399,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6829\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6830, 1, 'NAD83(2011) / Oregon Gresham-Warm Springs zone (m)', 'EPSG', 6830, 'PROJCS[\"NAD83(2"
  "011) / Oregon Gresham-Warm Springs zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-122.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00005,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",10000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6830\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6831, 1, 'NAD83(2011) / Oregon Gresham-Warm Springs zone (ft)', 'EPSG', 6831, 'PROJCS[\"NAD83("
  "2011) / Oregon Gresham-Warm Springs zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-122.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00005"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",32808.399,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"6831\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6832, 1, 'NAD83(CORS96) / Oregon La Grande zone (m)', 'EPSG', 6832, 'PROJCS[\"NAD83(CORS96) / "
  "Oregon La Grande zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\","
  "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-118,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00013,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "6832\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6833, 1, 'NAD83(CORS96) / Oregon La Grande zone (ft)', 'EPSG', 6833, 'PROJCS[\"NAD83(CORS96) /"
  " Oregon La Grande zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-118,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00013,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"6833\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6834, 1, 'NAD83(2011) / Oregon La Grande zone (m)', 'EPSG', 6834, 'PROJCS[\"NAD83(2011) / Oreg"
  "on La Grande zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00013,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6834\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6835, 1, 'NAD83(2011) / Oregon La Grande zone (ft)', 'EPSG', 6835, 'PROJCS[\"NAD83(2011) / Ore"
  "gon La Grande zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00013,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6"
  "835\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6836, 1, 'NAD83(CORS96) / Oregon Ontario zone (m)', 'EPSG', 6836, 'PROJCS[\"NAD83(CORS96) / Or"
  "egon Ontario zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",43.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6836\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6837, 1, 'NAD83(CORS96) / Oregon Ontario zone (ft)', 'EPSG', 6837, 'PROJCS[\"NAD83(CORS96) / O"
  "regon Ontario zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SP"
  "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",43.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"6837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6838, 1, 'NAD83(2011) / Oregon Ontario zone (m)', 'EPSG', 6838, 'PROJCS[\"NAD83(2011) / Oregon"
  " Ontario zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",43.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"683"
  "8\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6839, 1, 'NAD83(2011) / Oregon Ontario zone (ft)', 'EPSG', 6839, 'PROJCS[\"NAD83(2011) / Orego"
  "n Ontario zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",43.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"6839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6840, 1, 'NAD83(CORS96) / Oregon Coast zone (m)', 'EPSG', 6840, 'PROJCS[\"NAD83(CORS96) / Oreg"
  "on Coast zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID"
  "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPS"
  "G\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",44.75,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude "
  "of projection centre\",-124.05,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",5,AUTHORITY[\"EPSG"
  "\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",5,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale facto"
  "r on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",-300000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",-4600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6840\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6841, 1, 'NAD83(CORS96) / Oregon Coast zone (ft)', 'EPSG', 6841, 'PROJCS[\"NAD83(CORS96) / Ore"
  "gon Coast zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHERO"
  "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EP"
  "SG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",44.75,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude"
  " of projection centre\",-124.05,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",5,AUTHORITY[\"EPS"
  "G\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",5,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale fact"
  "or on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",-984251.9685,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",-15091863.5171,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPS"
  "G\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6841\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6842, 1, 'NAD83(2011) / Oregon Coast zone (m)', 'EPSG', 6842, 'PROJCS[\"NAD83(2011) / Oregon C"
  "oast zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]]"
  ",PARAMETER[\"Latitude of projection centre\",44.75,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection "
  "centre\",-124.05,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",5,AUTHORITY[\"EPSG\",\"8813\"]],"
  "PARAMETER[\"Angle from Rectified to Skew Grid\",5,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial l"
  "ine\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",-300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",-4600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6842\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6843, 1, 'NAD83(2011) / Oregon Coast zone (ft)', 'EPSG', 6843, 'PROJCS[\"NAD83(2011) / Oregon "
  "Coast zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\""
  "]],PARAMETER[\"Latitude of projection centre\",44.75,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projectio"
  "n centre\",-124.05,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",5,AUTHORITY[\"EPSG\",\"8813\"]"
  "],PARAMETER[\"Angle from Rectified to Skew Grid\",5,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial"
  " line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",-984251.9685,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",-15091863.5171,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6843\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6844, 1, 'NAD83(CORS96) / Oregon Pendleton zone (m)', 'EPSG', 6844, 'PROJCS[\"NAD83(CORS96) / "
  "Oregon Pendleton zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\","
  "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude "
  "of natural origin\",-119.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00"
  "0045,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6844\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6845, 1, 'NAD83(CORS96) / Oregon Pendleton zone (ft)', 'EPSG', 6845, 'PROJCS[\"NAD83(CORS96) /"
  " Oregon Pendleton zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-119.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "00045,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",196850.3937,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6845\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6846, 1, 'NAD83(2011) / Oregon Pendleton zone (m)', 'EPSG', 6846, 'PROJCS[\"NAD83(2011) / Oreg"
  "on Pendleton zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-119.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000045,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"6846\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6847, 1, 'NAD83(2011) / Oregon Pendleton zone (ft)', 'EPSG', 6847, 'PROJCS[\"NAD83(2011) / Ore"
  "gon Pendleton zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-119.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000045,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",196850.3937,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"6847\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6848, 1, 'NAD83(CORS96) / Oregon Pendleton-La Grande zone (m)', 'EPSG', 6848, 'PROJCS[\"NAD83("
  "CORS96) / Oregon Pendleton-La Grande zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refere"
  "nce Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"11"
  "33\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-118.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",1.000175,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6848\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6849, 1, 'NAD83(CORS96) / Oregon Pendleton-La Grande zone (ft)', 'EPSG', 6849, 'PROJCS[\"NAD83"
  "(CORS96) / Oregon Pendleton-La Grande zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Refe"
  "rence Station 1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-118.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1.000175,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",98425.1969,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6849\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6850, 1, 'NAD83(2011) / Oregon Pendleton-La Grande zone (m)', 'EPSG', 6850, 'PROJCS[\"NAD83(20"
  "11) / Oregon Pendleton-La Grande zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-118.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.000175,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6850\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6851, 1, 'NAD83(2011) / Oregon Pendleton-La Grande zone (ft)', 'EPSG', 6851, 'PROJCS[\"NAD83(2"
  "011) / Oregon Pendleton-La Grande zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",45.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",-118.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",1.000175,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",98425.1969,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6851\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6852, 1, 'NAD83(CORS96) / Oregon Portland zone (m)', 'EPSG', 6852, 'PROJCS[\"NAD83(CORS96) / O"
  "regon Portland zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SP"
  "HEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPS"
  "G\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-122.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "02,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",50000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"6852\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6853, 1, 'NAD83(CORS96) / Oregon Portland zone (ft)', 'EPSG', 6853, 'PROJCS[\"NAD83(CORS96) / "
  "Oregon Portland zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\","
  "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"E"
  "PSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-122.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00"
  "0002,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.9895,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",164041.9948,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6853\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6854, 1, 'NAD83(2011) / Oregon Portland zone (m)', 'EPSG', 6854, 'PROJCS[\"NAD83(2011) / Orego"
  "n Portland zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",45.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-122.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000002,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500"
  "00,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6854\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6855, 1, 'NAD83(2011) / Oregon Portland zone (ft)', 'EPSG', 6855, 'PROJCS[\"NAD83(2011) / Oreg"
  "on Portland zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",45.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-122.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000002,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.9895,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",164041.9948,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6855\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6856, 1, 'NAD83(CORS96) / Oregon Salem zone (m)', 'EPSG', 6856, 'PROJCS[\"NAD83(CORS96) / Oreg"
  "on Salem zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID"
  "[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",44.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natur"
  "al origin\",-123.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00001,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6856\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6857, 1, 'NAD83(CORS96) / Oregon Salem zone (ft)', 'EPSG', 6857, 'PROJCS[\"NAD83(CORS96) / Ore"
  "gon Salem zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHERO"
  "ID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-123.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00001,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"6857\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6858, 1, 'NAD83(2011) / Oregon Salem zone (m)', 'EPSG', 6858, 'PROJCS[\"NAD83(2011) / Oregon S"
  "alem zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",44.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "23.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00001,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"6858\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6859, 1, 'NAD83(2011) / Oregon Salem zone (ft)', 'EPSG', 6859, 'PROJCS[\"NAD83(2011) / Oregon "
  "Salem zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",44.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-123.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00001,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"6859\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6860, 1, 'NAD83(CORS96) / Oregon Santiam Pass zone (m)', 'EPSG', 6860, 'PROJCS[\"NAD83(CORS96)"
  " / Oregon Santiam Pass zone (m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 19"
  "96)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM["
  "\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS"
  "[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\""
  ",\"9807\"]],PARAMETER[\"Latitude of natural origin\",44.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-122.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1.000155,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"6860\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6861, 1, 'NAD83(CORS96) / Oregon Santiam Pass zone (ft)', 'EPSG', 6861, 'PROJCS[\"NAD83(CORS96"
  ") / Oregon Santiam Pass zone (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station "
  "1996)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",44.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-122.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",1.000155,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6861\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6862, 1, 'NAD83(2011) / Oregon Santiam Pass zone (m)', 'EPSG', 6862, 'PROJCS[\"NAD83(2011) / O"
  "regon Santiam Pass zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-122.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000155,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"6862\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6863, 1, 'NAD83(2011) / Oregon Santiam Pass zone (ft)', 'EPSG', 6863, 'PROJCS[\"NAD83(2011) / "
  "Oregon Santiam Pass zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",44.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-122.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000155,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6863\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6867, 1, 'NAD83(CORS96) / Oregon LCC (m)', 'EPSG', 6867, 'PROJCS[\"NAD83(CORS96) / Oregon LCC "
  "(m)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAM"
  "ETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.51"
  "1111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823"
  "\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easti"
  "ng at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"68"
  "67\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6868, 1, 'NAD83(CORS96) / Oregon GIC Lambert (ft)', 'EPSG', 6868, 'PROJCS[\"NAD83(CORS96) / Or"
  "egon GIC Lambert (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\""
  ",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORIT"
  "Y[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],"
  "PARAMETER[\"Easting at false origin\",1312335.958,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6868\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6870, 1, 'ETRS89 / Albania TM 2010', 'EPSG', 6870, 'PROJCS[\"ETRS89 / Albania TM 2010\",GEOGCS"
  "[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",20,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6870\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6875, 1, 'RDN2008 / Italy zone (N-E)', 'EPSG', 6875, 'PROJCS[\"RDN2008 / Italy zone (N-E)\",GE"
  "OGCS[\"RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9985,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",7000000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6875\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6876, 1, 'RDN2008 / Zone 12 (N-E)', 'EPSG', 6876, 'PROJCS[\"RDN2008 / Zone 12 (N-E)\",GEOGCS[\""
  "RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3000000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"6876\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6879, 1, 'NAD83(2011) / Wisconsin Central', 'EPSG', 6879, 'PROJCS[\"NAD83(2011) / Wisconsin Ce"
  "ntral\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90"
  ",AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"6879\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6880, 1, 'NAD83(2011) / Nebraska (ftUS)', 'EPSG', 6880, 'PROJCS[\"NAD83(2011) / Nebraska (ftUS"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lati"
  "tude of false origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AU"
  "THORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETE"
  "R[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",164041"
  "6.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US"
  " survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6880\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6881, 1, 'Aden 1925', 'EPSG', 6881, 'GEOGCS[\"Aden 1925\",DATUM[\"Aden 1925\",SPHEROID[\"Clark"
  "e 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-24,-203,268,0,0,0,0],AUTHORITY[\"EPSG\",\"1"
  "135\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6881\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6882, 1, 'Bekaa Valley 1920', 'EPSG', 6882, 'GEOGCS[\"Bekaa Valley 1920\",DATUM[\"Bekaa Valley"
  " 1920\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-183,-15,273,0,0,0,0"
  "],AUTHORITY[\"EPSG\",\"1137\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994"
  "3278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6882\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6883, 1, 'Bioko', 'EPSG', 6883, 'GEOGCS[\"Bioko\",DATUM[\"Bioko\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-235,-110,393,0,0,0,0],AUTHORITY[\"EPSG\",\"1136\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6883\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6884, 1, 'NAD83(CORS96) / Oregon North', 'EPSG', 6884, 'PROJCS[\"NAD83(CORS96) / Oregon North\""
  ",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "0.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6884\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6885, 1, 'NAD83(CORS96) / Oregon North (ft)', 'EPSG', 6885, 'PROJCS[\"NAD83(CORS96) / Oregon N"
  "orth (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",8202099.738,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6885\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6886, 1, 'NAD83(CORS96) / Oregon South', 'EPSG', 6886, 'PROJCS[\"NAD83(CORS96) / Oregon South\""
  ",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-12"
  "0.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"6886\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6887, 1, 'NAD83(CORS96) / Oregon South (ft)', 'EPSG', 6887, 'PROJCS[\"NAD83(CORS96) / Oregon S"
  "outh (ft)\",GEOGCS[\"NAD83(CORS96)\",DATUM[\"NAD83 (Continuously Operating Reference Station 1996)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1133\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6783\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],"
  "PARAMETER[\"Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false "
  "origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY"
  "[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],P"
  "ARAMETER[\"Easting at false origin\",4921259.843,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"6887\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6892, 1, 'South East Island 1943', 'EPSG', 6892, 'GEOGCS[\"South East Island 1943\",DATUM[\"So"
  "uth East Island 1943\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-76.2"
  "69,-16.683,68.562,-6.275,10.536,-4.286,-13.686],AUTHORITY[\"EPSG\",\"1138\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"6892\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6894, 1, 'Gambia', 'EPSG', 6894, 'GEOGCS[\"Gambia\",DATUM[\"Gambia\",SPHEROID[\"Clarke 1880 (R"
  "GS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-63,176,185,0,0,0,0],AUTHORITY[\"EPSG\",\"1139\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6894\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6915, 1, 'South East Island 1943 / UTM zone 40N', 'EPSG', 6915, 'PROJCS[\"South East Island 19"
  "43 / UTM zone 40N\",GEOGCS[\"South East Island 1943\",DATUM[\"South East Island 1943\",SPHEROID[\"Clarke 1880 (RGS)\""
  ",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-76.269,-16.683,68.562,-6.275,10.536,-4.286,-13.686],AUTHO"
  "RITY[\"EPSG\",\"1138\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6892\"]],PROJECTION[\"Transve"
  "rse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS"
  "[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6915\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6922, 1, 'NAD83 / Kansas LCC', 'EPSG', 6922, 'PROJCS[\"NAD83 / Kansas LCC\",GEOGCS[\"NAD83\",D"
  "ATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84["
  "1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36"
  ",AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER"
  "[\"Latitude of 1st standard parallel\",39.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",37.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6922\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6923, 1, 'NAD83 / Kansas LCC (ftUS)', 'EPSG', 6923, 'PROJCS[\"NAD83 / Kansas LCC (ftUS)\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.25,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latit"
  "ude of 2nd standard parallel\",37.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1"
  "312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT"
  "[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"6923\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6924, 1, 'NAD83(2011) / Kansas LCC', 'EPSG', 6924, 'PROJCS[\"NAD83(2011) / Kansas LCC\",GEOGCS"
  "[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101"
  ",AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNI"
  "T[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EP"
  "SG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fa"
  "lse origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.25,AUTHORITY[\"EPSG\",\"882"
  "2\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",37.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\","
  "400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6924\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6925, 1, 'NAD83(2011) / Kansas LCC (ftUS)', 'EPSG', 6925, 'PROJCS[\"NAD83(2011) / Kansas LCC ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",36,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.25,AUTHORITY["
  "\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PA"
  "RAMETER[\"Latitude of 2nd standard parallel\",37.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",1312333.3333,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\","
  "\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"6925\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6931, 1, 'WGS 84 / NSIDC EASE-Grid 2.0 North', 'EPSG', 6931, 'PROJCS[\"WGS 84 / NSIDC EASE-Gri"
  "d 2.0 North\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHOR"
  "ITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4326\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural or"
  "igin\",90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",SOUTH],AUTHORITY[\"EPSG\",\"6931\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6932, 1, 'WGS 84 / NSIDC EASE-Grid 2.0 South', 'EPSG', 6932, 'PROJCS[\"WGS 84 / NSIDC EASE-Gri"
  "d 2.0 South\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHOR"
  "ITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4326\"]],PROJECTION[\"Lambert Azimuthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural or"
  "igin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6932\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6933, 1, 'WGS 84 / NSIDC EASE-Grid 2.0 Global', 'EPSG', 6933, 'PROJCS[\"WGS 84 / NSIDC EASE-Gr"
  "id 2.0 Global\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTH"
  "ORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4326\"]],PROJECTION[\"Lambert Cylindrical Equal Area\",AUTHORITY[\"EPSG\",\"9835\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",30,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"69"
  "33\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6956, 1, 'VN-2000 / TM-3 zone 481 (6956 deprecated)', 'EPSG', 6956, 'PROJCS[\"VN-2000 / TM-3 z"
  "one 481\",GEOGCS[\"VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"70"
  "30\"]],TOWGS84[-191.90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EP"
  "SG\",\"6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",102,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999"
  "9,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"6956\"]]', 'Deprecated since EPSG Dataset 9.3. Superseded by SRID 5896.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6957, 1, 'VN-2000 / TM-3 zone 482 (6957 deprecated)', 'EPSG', 6957, 'PROJCS[\"VN-2000 / TM-3 z"
  "one 482\",GEOGCS[\"VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"70"
  "30\"]],TOWGS84[-191.90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EP"
  "SG\",\"6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999"
  "9,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"6957\"]]', 'Deprecated since EPSG Dataset 9.3. Superseded by SRID 5897.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6958, 1, 'VN-2000 / TM-3 zone 491 (6958 deprecated)', 'EPSG', 6958, 'PROJCS[\"VN-2000 / TM-3 z"
  "one 491\",GEOGCS[\"VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"70"
  "30\"]],TOWGS84[-191.90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHORITY[\"EP"
  "SG\",\"6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",108,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999"
  "9,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"6958\"]]', 'Deprecated since EPSG Dataset 9.3. Superseded by SRID 5898.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6959, 1, 'VN-2000 / TM-3 Da Nang zone (6959 deprecated)', 'EPSG', 6959, 'PROJCS[\"VN-2000 / TM"
  "-3 Da Nang zone\",GEOGCS[\"VN-2000\",DATUM[\"Vietnam 2000\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPS"
  "G\",\"7030\"]],TOWGS84[-191.90441429,-39.30318279,-111.45032835,0.00928836,-0.01975479,0.00427372,0.252906278],AUTHOR"
  "ITY[\"EPSG\",\"6756\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4756\"]],PROJECTION[\"Transver"
  "se Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",107.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural or"
  "igin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],"
  "AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6959\"]]', 'Deprecated since EPSG Dataset 9.3. Superseded by SRID 5899.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6962, 1, 'ETRS89 / Albania LCC 2010', 'EPSG', 6962, 'PROJCS[\"ETRS89 / Albania LCC 2010\",GEOG"
  "CS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",20,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",43,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"6962\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6966, 1, 'NAD27 / Michigan North', 'EPSG', 6966, 'PROJCS[\"NAD27 / Michigan North\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP Michigan)\",AUTHORITY[\"EPSG\",\"1051\"]],PARAMETER[\"Latitude of false origin\",44.7944444444444,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-87,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "47.0944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],PARAMETER[\"Ellipsoid scaling factor\",1"
  ".0000382,AUTHORITY[\"EPSG\",\"1038\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"6966\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6983, 1, 'IG05 Intermediate CRS', 'EPSG', 6983, 'GEOGCS[\"IG05 Intermediate CRS\",DATUM[\"IG05"
  " Intermediate Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6983\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6984, 1, 'Israeli Grid 05', 'EPSG', 6984, 'PROJCS[\"Israeli Grid 05\",GEOGCS[\"IG05 Intermedia"
  "te CRS\",DATUM[\"IG05 Intermediate Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "AUTHORITY[\"EPSG\",\"1142\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6983\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31.7343936111111,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",35.2045169444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1.0000067,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",219529.584,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",626907.39,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6984\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6990, 1, 'IG05/12 Intermediate CRS', 'EPSG', 6990, 'GEOGCS[\"IG05/12 Intermediate CRS\",DATUM["
  "\"IG05/12 Intermediate Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\""
  "EPSG\",\"1144\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6990\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (6991, 1, 'Israeli Grid 05/12', 'EPSG', 6991, 'PROJCS[\"Israeli Grid 05/12\",GEOGCS[\"IG05/12 I"
  "ntermediate CRS\",DATUM[\"IG05/12 Intermediate Datum\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],AUTHORITY[\"EPSG\",\"1144\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6990\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31.73439361111"
  "11,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",35.2045169444444,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",1.0000067,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",219529.584,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",626907.39,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"6991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7005, 1, 'Nahrwan 1934 / UTM zone 37N', 'EPSG', 7005, 'PROJCS[\"Nahrwan 1934 / UTM zone 37N\","
  "GEOGCS[\"Nahrwan 1934\",DATUM[\"Nahrwan 1934\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-242.2,-144.9,370.3,0,0,0,0],AUTHORITY[\"EPSG\",\"6744\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4744\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7005\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7006, 1, 'Nahrwan 1934 / UTM zone 38N', 'EPSG', 7006, 'PROJCS[\"Nahrwan 1934 / UTM zone 38N\","
  "GEOGCS[\"Nahrwan 1934\",DATUM[\"Nahrwan 1934\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-242.2,-144.9,370.3,0,0,0,0],AUTHORITY[\"EPSG\",\"6744\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4744\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7006\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7007, 1, 'Nahrwan 1934 / UTM zone 39N', 'EPSG', 7007, 'PROJCS[\"Nahrwan 1934 / UTM zone 39N\","
  "GEOGCS[\"Nahrwan 1934\",DATUM[\"Nahrwan 1934\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\","
  "\"7012\"]],TOWGS84[-242.2,-144.9,370.3,0,0,0,0],AUTHORITY[\"EPSG\",\"6744\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4744\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7007\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7035, 1, 'RGSPM06 (lon-lat)', 'EPSG', 7035, 'GEOGCS[\"RGSPM06 (lon-lat)\",DATUM[\"Reseau Geode"
  "sique de Saint Pierre et Miquelon 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1038\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7037, 1, 'RGR92 (lon-lat)', 'EPSG', 7037, 'GEOGCS[\"RGR92 (lon-lat)\",DATUM[\"Reseau Geodesiqu"
  "e de la Reunion 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "6627\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7037\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7039, 1, 'RGM04 (lon-lat)', 'EPSG', 7039, 'GEOGCS[\"RGM04 (lon-lat)\",DATUM[\"Reseau Geodesiqu"
  "e de Mayotte 2004\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"10"
  "36\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7039\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7041, 1, 'RGFG95 (lon-lat)', 'EPSG', 7041, 'GEOGCS[\"RGFG95 (lon-lat)\",DATUM[\"Reseau Geodesi"
  "que Francais Guyane 1995\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG"
  "\",\"6624\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7041\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7057, 1, 'NAD83(2011) / IaRCS zone 1', 'EPSG', 7057, 'PROJCS[\"NAD83(2011) / IaRCS zone 1\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude o"
  "f natural origin\",43.2111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-95.25,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000052,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",9600000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7058, 1, 'NAD83(2011) / IaRCS zone 2', 'EPSG', 7058, 'PROJCS[\"NAD83(2011) / IaRCS zone 2\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude o"
  "f natural origin\",43.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.75,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000043,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",9800000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7058\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7059, 1, 'NAD83(2011) / IaRCS zone 3', 'EPSG', 7059, 'PROJCS[\"NAD83(2011) / IaRCS zone 3\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.2,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000035,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8300000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7059\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7060, 1, 'NAD83(2011) / IaRCS zone 4', 'EPSG', 7060, 'PROJCS[\"NAD83(2011) / IaRCS zone 4\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude o"
  "f natural origin\",42.5333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-94.83333"
  "33333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000045,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8600000,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7060\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7061, 1, 'NAD83(2011) / IaRCS zone 5', 'EPSG', 7061, 'PROJCS[\"NAD83(2011) / IaRCS zone 5\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude o"
  "f natural origin\",42.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.25,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000032,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8900000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7061\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7062, 1, 'NAD83(2011) / IaRCS zone 6', 'EPSG', 7062, 'PROJCS[\"NAD83(2011) / IaRCS zone 6\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-95.7444444444444,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000039,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6600000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7062\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7063, 1, 'NAD83(2011) / IaRCS zone 7', 'EPSG', 7063, 'PROJCS[\"NAD83(2011) / IaRCS zone 7\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-94.6444444444444,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000045,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6800000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7063\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7064, 1, 'NAD83(2011) / IaRCS zone 8', 'EPSG', 7064, 'PROJCS[\"NAD83(2011) / IaRCS zone 8\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93.7166666666667,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000033,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",7000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7064\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7065, 1, 'NAD83(2011) / IaRCS zone 9', 'EPSG', 7065, 'PROJCS[\"NAD83(2011) / IaRCS zone 9\",GE"
  "OGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.8277777777778,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",7200000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7066, 1, 'NAD83(2011) / IaRCS zone 10', 'EPSG', 7066, 'PROJCS[\"NAD83(2011) / IaRCS zone 10\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",41.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.666"
  "6666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8000000,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7067, 1, 'NAD83(2011) / IaRCS zone 11', 'EPSG', 7067, 'PROJCS[\"NAD83(2011) / IaRCS zone 11\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.5444444444444"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",7600000,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7067\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7068, 1, 'NAD83(2011) / IaRCS zone 12', 'EPSG', 7068, 'PROJCS[\"NAD83(2011) / IaRCS zone 12\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",40.9277777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93.75,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000037,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6200000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7068\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7069, 1, 'NAD83(2011) / IaRCS zone 13', 'EPSG', 7069, 'PROJCS[\"NAD83(2011) / IaRCS zone 13\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.9277777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6400000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7069\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7070, 1, 'NAD83(2011) / IaRCS zone 14', 'EPSG', 7070, 'PROJCS[\"NAD83(2011) / IaRCS zone 14\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.25,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000018,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",24500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6200000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7070\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7073, 1, 'RGTAAF07', 'EPSG', 7073, 'GEOGCS[\"RGTAAF07\",DATUM[\"Reseau Geodesique des Terres A"
  "ustrales et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7073\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7074, 1, 'RGTAAF07 / UTM zone 37S', 'EPSG', 7074, 'PROJCS[\"RGTAAF07 / UTM zone 37S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7074\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7075, 1, 'RGTAAF07 / UTM zone 38S', 'EPSG', 7075, 'PROJCS[\"RGTAAF07 / UTM zone 38S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7075\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7076, 1, 'RGTAAF07 / UTM zone 39S', 'EPSG', 7076, 'PROJCS[\"RGTAAF07 / UTM zone 39S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7076\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7077, 1, 'RGTAAF07 / UTM zone 40S', 'EPSG', 7077, 'PROJCS[\"RGTAAF07 / UTM zone 40S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7077\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7078, 1, 'RGTAAF07 / UTM zone 41S', 'EPSG', 7078, 'PROJCS[\"RGTAAF07 / UTM zone 41S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7078\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7079, 1, 'RGTAAF07 / UTM zone 42S', 'EPSG', 7079, 'PROJCS[\"RGTAAF07 / UTM zone 42S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7079\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7080, 1, 'RGTAAF07 / UTM zone 43S', 'EPSG', 7080, 'PROJCS[\"RGTAAF07 / UTM zone 43S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7080\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7081, 1, 'RGTAAF07 / UTM zone 44S', 'EPSG', 7081, 'PROJCS[\"RGTAAF07 / UTM zone 44S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7081\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7082, 1, 'RGTAAF07 / Terre Adelie Polar Stereographic (7082 deprecated)', 'EPSG', 7082, 'PROJC"
  "S[\"RGTAAF07 / Terre Adelie Polar Stereographic\",GEOGCS[\"RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes "
  "et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY"
  "[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Polar Stere"
  "ographic (variant C)\",AUTHORITY[\"EPSG\",\"9830\"]],PARAMETER[\"Latitude of standard parallel\",-67,AUTHORITY[\"EPSG"
  "\",\"8832\"]],PARAMETER[\"Longitude of origin\",140,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Easting at false origin"
  "\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",200000,AUTHORITY[\"EPSG\",\"8827\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7082\"]]', 'Dep"
  "recated since EPSG Dataset 9.3.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7084, 1, 'RGF93 (lon-lat)', 'EPSG', 7084, 'GEOGCS[\"RGF93 (lon-lat)\",DATUM[\"Reseau Geodesiqu"
  "e Francais 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6171"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7084\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7086, 1, 'RGAF09 (lon-lat)', 'EPSG', 7086, 'GEOGCS[\"RGAF09 (lon-lat)\",DATUM[\"Reseau Geodesi"
  "que des Antilles Francaises 2009\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT"
  "Y[\"EPSG\",\"1073\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7086\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7109, 1, 'NAD83(2011) / RMTCRS St Mary (m)', 'EPSG', 7109, 'PROJCS[\"NAD83(2011) / RMTCRS St M"
  "ary (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",48.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.511"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00016,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "7109\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7110, 1, 'NAD83(2011) / RMTCRS Blackfeet (m)', 'EPSG', 7110, 'PROJCS[\"NAD83(2011) / RMTCRS Bl"
  "ackfeet (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",48,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.511111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00019,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7110\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7111, 1, 'NAD83(2011) / RMTCRS Milk River (m)', 'EPSG', 7111, 'PROJCS[\"NAD83(2011) / RMTCRS M"
  "ilk River (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAM"
  "ETER[\"Latitude of natural origin\",48.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000145,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"7111\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7112, 1, 'NAD83(2011) / RMTCRS Fort Belknap (m)', 'EPSG', 7112, 'PROJCS[\"NAD83(2011) / RMTCRS"
  " Fort Belknap (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],"
  "PARAMETER[\"Latitude of natural origin\",48.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-108.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00012,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",150000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7112\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7113, 1, 'NAD83(2011) / RMTCRS Fort Peck Assiniboine (m)', 'EPSG', 7113, 'PROJCS[\"NAD83(2011)"
  " / RMTCRS Fort Peck Assiniboine (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\""
  "EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",48.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7113\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7114, 1, 'NAD83(2011) / RMTCRS Fort Peck Sioux (m)', 'EPSG', 7114, 'PROJCS[\"NAD83(2011) / RMT"
  "CRS Fort Peck Sioux (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",48.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000"
  "09,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",50000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"7114\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7115, 1, 'NAD83(2011) / RMTCRS Crow (m)', 'EPSG', 7115, 'PROJCS[\"NAD83(2011) / RMTCRS Crow (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",44.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.75,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000148,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7115\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7116, 1, 'NAD83(2011) / RMTCRS Bobcat (m)', 'EPSG', 7116, 'PROJCS[\"NAD83(2011) / RMTCRS Bobca"
  "t (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",46.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-111.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000185,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"7116\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7117, 1, 'NAD83(2011) / RMTCRS Billings (m)', 'EPSG', 7117, 'PROJCS[\"NAD83(2011) / RMTCRS Bil"
  "lings (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",45.7944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-108.416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001515,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",50"
  "000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"7117\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7118, 1, 'NAD83(2011) / RMTCRS Wind River (m)', 'EPSG', 7118, 'PROJCS[\"NAD83(2011) / RMTCRS W"
  "ind River (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",42.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "08.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00024,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EP"
  "SG\",\"7118\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7119, 1, 'NAD83(2011) / RMTCRS St Mary (ft)', 'EPSG', 7119, 'PROJCS[\"NAD83(2011) / RMTCRS St "
  "Mary (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",48.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112."
  "511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00016,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",492125.9843,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"7119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7120, 1, 'NAD83(2011) / RMTCRS Blackfeet (ft)', 'EPSG', 7120, 'PROJCS[\"NAD83(2011) / RMTCRS B"
  "lackfeet (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",48,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.51111111111"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00019,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",328083.9895,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"7120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7121, 1, 'NAD83(2011) / RMTCRS Milk River (ft)', 'EPSG', 7121, 'PROJCS[\"NAD83(2011) / RMTCRS "
  "Milk River (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",48.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000145,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",492125.9843,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",656167"
  ".979,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7122, 1, 'NAD83(2011) / RMTCRS Fort Belknap (ft)', 'EPSG', 7122, 'PROJCS[\"NAD83(2011) / RMTCR"
  "S Fort Belknap (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",48.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-108.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00012,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656167.979,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",492125.9843,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7123, 1, 'NAD83(2011) / RMTCRS Fort Peck Assiniboine (ft)', 'EPSG', 7123, 'PROJCS[\"NAD83(2011"
  ") / RMTCRS Fort Peck Assiniboine (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY["
  "\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",48.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656167.979,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",328083.9895,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7123\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7124, 1, 'NAD83(2011) / RMTCRS Fort Peck Sioux (ft)', 'EPSG', 7124, 'PROJCS[\"NAD83(2011) / RM"
  "TCRS Fort Peck Sioux (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",48.3333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "0009,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.9895,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",164041.9938,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7125, 1, 'NAD83(2011) / RMTCRS Crow (ft)', 'EPSG', 7125, 'PROJCS[\"NAD83(2011) / RMTCRS Crow ("
  "ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",44.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.75,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000148,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",656167.979,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7125\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7126, 1, 'NAD83(2011) / RMTCRS Bobcat (ft)', 'EPSG', 7126, 'PROJCS[\"NAD83(2011) / RMTCRS Bobc"
  "at (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",46.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-111.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000185,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",328083.9895,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",328083.9895,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"7126\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7127, 1, 'NAD83(2011) / RMTCRS Billings (ft)', 'EPSG', 7127, 'PROJCS[\"NAD83(2011) / RMTCRS Bi"
  "llings (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",45.7944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-108.416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001515,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",656167.979,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",164041.9948,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7128, 1, 'NAD83(2011) / RMTCRS Wind River (ftUS)', 'EPSG', 7128, 'PROJCS[\"NAD83(2011) / RMTCR"
  "S Wind River (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",42.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-108.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00024,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7131, 1, 'NAD83(2011) / San Francisco CS13', 'EPSG', 7131, 'PROJCS[\"NAD83(2011) / San Francis"
  "co CS13\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",37.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-122.45,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000007,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",48000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",24000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7131\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7132, 1, 'NAD83(2011) / San Francisco CS13 (ftUS)', 'EPSG', 7132, 'PROJCS[\"NAD83(2011) / San "
  "Francisco CS13 (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",37.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-122"
  ".45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000007,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",157480,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",78740,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7132\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7133, 1, 'RGTAAF07 (lon-lat)', 'EPSG', 7133, 'GEOGCS[\"RGTAAF07 (lon-lat)\",DATUM[\"Reseau Geo"
  "desique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"7133\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7136, 1, 'IGD05', 'EPSG', 7136, 'GEOGCS[\"IGD05\",DATUM[\"Israeli Geodetic Datum 2005\",SPHERO"
  "ID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1114\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7139, 1, 'IGD05/12', 'EPSG', 7139, 'GEOGCS[\"IGD05/12\",DATUM[\"Israeli Geodetic Datum 2005(20"
  "12)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1115\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7139\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7142, 1, 'Palestine 1923 / Palestine Grid modified', 'EPSG', 7142, 'PROJCS[\"Palestine 1923 / "
  "Palestine Grid modified\",GEOGCS[\"Palestine 1923\",DATUM[\"Palestine 1923\",SPHEROID[\"Clarke 1880 (Benoit)\",637830"
  "0.789,293.4663155389811,AUTHORITY[\"EPSG\",\"7010\"]],TOWGS84[-275.7224,94.7824,340.8944,-8.001,-4.42,-11.821,1],AUTH"
  "ORITY[\"EPSG\",\"6281\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4281\"]],PROJECTION[\"Transv"
  "erse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31.7340969444444,AUTHORITY[\"E"
  "PSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",35.2120805555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",170251.555,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",126867.909,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7142\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7257, 1, 'NAD83(2011) / InGCS Adams (m)', 'EPSG', 7257, 'PROJCS[\"NAD83(2011) / InGCS Adams (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",40.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.961111111"
  "1111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "7257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7258, 1, 'NAD83(2011) / InGCS Adams (ftUS)', 'EPSG', 7258, 'PROJCS[\"NAD83(2011) / InGCS Adams"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",40.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.9611"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7259, 1, 'NAD83(2011) / InGCS Allen (m)', 'EPSG', 7259, 'PROJCS[\"NAD83(2011) / InGCS Allen (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",40.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.05,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7259\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7260, 1, 'NAD83(2011) / InGCS Allen (ftUS)', 'EPSG', 7260, 'PROJCS[\"NAD83(2011) / InGCS Allen"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",40.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.05,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"7260\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7261, 1, 'NAD83(2011) / InGCS Bartholomew (m)', 'EPSG', 7261, 'PROJCS[\"NAD83(2011) / InGCS Ba"
  "rtholomew (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",39,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.85,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7261\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7262, 1, 'NAD83(2011) / InGCS Bartholomew (ftUS)', 'EPSG', 7262, 'PROJCS[\"NAD83(2011) / InGCS"
  " Bartholomew (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.85,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"7262\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7263, 1, 'NAD83(2011) / InGCS Benton (m)', 'EPSG', 7263, 'PROJCS[\"NAD83(2011) / InGCS Benton "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",40.45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.3,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000029,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7263\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7264, 1, 'NAD83(2011) / InGCS Benton (ftUS)', 'EPSG', 7264, 'PROJCS[\"NAD83(2011) / InGCS Bent"
  "on (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",40.45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.3,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000029,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7264\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7265, 1, 'NAD83(2011) / InGCS Blackford-Delaware (m)', 'EPSG', 7265, 'PROJCS[\"NAD83(2011) / I"
  "nGCS Blackford-Delaware (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",40.05,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-85.4111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7265\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7266, 1, 'NAD83(2011) / InGCS Blackford-Delaware (ftUS)', 'EPSG', 7266, 'PROJCS[\"NAD83(2011) "
  "/ InGCS Blackford-Delaware (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",40.05,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-85.4111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1"
  "18110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7266\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7267, 1, 'NAD83(2011) / InGCS Boone-Hendricks (m)', 'EPSG', 7267, 'PROJCS[\"NAD83(2011) / InGC"
  "S Boone-Hendricks (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",39.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-86.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3"
  "6000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7267\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7268, 1, 'NAD83(2011) / InGCS Boone-Hendricks (ftUS)', 'EPSG', 7268, 'PROJCS[\"NAD83(2011) / I"
  "nGCS Boone-Hendricks (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",39.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-86.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7268\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7269, 1, 'NAD83(2011) / InGCS Brown (m)', 'EPSG', 7269, 'PROJCS[\"NAD83(2011) / InGCS Brown (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",39,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.3,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7269\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7270, 1, 'NAD83(2011) / InGCS Brown (ftUS)', 'EPSG', 7270, 'PROJCS[\"NAD83(2011) / InGCS Brown"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",39,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.3,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"7270\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7271, 1, 'NAD83(2011) / InGCS Carroll (m)', 'EPSG', 7271, 'PROJCS[\"NAD83(2011) / InGCS Carrol"
  "l (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",40.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.65,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7271\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7272, 1, 'NAD83(2011) / InGCS Carroll (ftUS)', 'EPSG', 7272, 'PROJCS[\"NAD83(2011) / InGCS Car"
  "roll (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.65,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7272\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7273, 1, 'NAD83(2011) / InGCS Cass (m)', 'EPSG', 7273, 'PROJCS[\"NAD83(2011) / InGCS Cass (m)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.4111111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000028,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"727"
  "3\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7274, 1, 'NAD83(2011) / InGCS Cass (ftUS)', 'EPSG', 7274, 'PROJCS[\"NAD83(2011) / InGCS Cass ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",40.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.41111"
  "11111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000028,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"7274\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7275, 1, 'NAD83(2011) / InGCS Clark-Floyd-Scott (m)', 'EPSG', 7275, 'PROJCS[\"NAD83(2011) / In"
  "GCS Clark-Floyd-Scott (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",38.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-85.6111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000021,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7275\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7276, 1, 'NAD83(2011) / InGCS Clark-Floyd-Scott (ftUS)', 'EPSG', 7276, 'PROJCS[\"NAD83(2011) /"
  " InGCS Clark-Floyd-Scott (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",38.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-85.6111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000021,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118"
  "110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7276\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7277, 1, 'NAD83(2011) / InGCS Clay (m)', 'EPSG', 7277, 'PROJCS[\"NAD83(2011) / InGCS Clay (m)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",39.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.15,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000024,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7277\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7278, 1, 'NAD83(2011) / InGCS Clay (ftUS)', 'EPSG', 7278, 'PROJCS[\"NAD83(2011) / InGCS Clay ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",39.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.15,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000024,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7278\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7279, 1, 'NAD83(2011) / InGCS Clinton (m)', 'EPSG', 7279, 'PROJCS[\"NAD83(2011) / InGCS Clinto"
  "n (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",40.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.6111111111111,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000032,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7279\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7280, 1, 'NAD83(2011) / InGCS Clinton (ftUS)', 'EPSG', 7280, 'PROJCS[\"NAD83(2011) / InGCS Cli"
  "nton (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",40.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.6111111111"
  "111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000032,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7280\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7281, 1, 'NAD83(2011) / InGCS Crawford-Lawrence-Orange (m)', 'EPSG', 7281, 'PROJCS[\"NAD83(201"
  "1) / InGCS Crawford-Lawrence-Orange (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 201"
  "1)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",38.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-86.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000025,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"7281\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7282, 1, 'NAD83(2011) / InGCS Crawford-Lawrence-Orange (ftUS)', 'EPSG', 7282, 'PROJCS[\"NAD83("
  "2011) / InGCS Crawford-Lawrence-Orange (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-86.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000"
  "025,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7282\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7283, 1, 'NAD83(2011) / InGCS Daviess-Greene (m)', 'EPSG', 7283, 'PROJCS[\"NAD83(2011) / InGCS"
  " Daviess-Greene (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",38.45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.1,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000018,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7283\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7284, 1, 'NAD83(2011) / InGCS Daviess-Greene (ftUS)', 'EPSG', 7284, 'PROJCS[\"NAD83(2011) / In"
  "GCS Daviess-Greene (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",38.45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-87.1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000018,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7284\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7285, 1, 'NAD83(2011) / InGCS Dearborn-Ohio-Switzerland (m)', 'EPSG', 7285, 'PROJCS[\"NAD83(20"
  "11) / InGCS Dearborn-Ohio-Switzerland (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-84.9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000029,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"7285\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7286, 1, 'NAD83(2011) / InGCS Dearborn-Ohio-Switzerland (ftUS)', 'EPSG', 7286, 'PROJCS[\"NAD83"
  "(2011) / InGCS Dearborn-Ohio-Switzerland (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Sy"
  "stem 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY["
  "\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-84.9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000029,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7286\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7287, 1, 'NAD83(2011) / InGCS Decatur-Rush (m)', 'EPSG', 7287, 'PROJCS[\"NAD83(2011) / InGCS D"
  "ecatur-Rush (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",39.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.65,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7287\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7288, 1, 'NAD83(2011) / InGCS Decatur-Rush (ftUS)', 'EPSG', 7288, 'PROJCS[\"NAD83(2011) / InGC"
  "S Decatur-Rush (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",39.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.6"
  "5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7288\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7289, 1, 'NAD83(2011) / InGCS DeKalb (m)', 'EPSG', 7289, 'PROJCS[\"NAD83(2011) / InGCS DeKalb "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",41.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.9611111"
  "111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"7289\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7290, 1, 'NAD83(2011) / InGCS DeKalb (ftUS)', 'EPSG', 7290, 'PROJCS[\"NAD83(2011) / InGCS DeKa"
  "lb (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",41.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.9"
  "611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7290\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7291, 1, 'NAD83(2011) / InGCS Dubois-Martin (m)', 'EPSG', 7291, 'PROJCS[\"NAD83(2011) / InGCS "
  "Dubois-Martin (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",38.2111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-86.9611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7291\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7292, 1, 'NAD83(2011) / InGCS Dubois-Martin (ftUS)', 'EPSG', 7292, 'PROJCS[\"NAD83(2011) / InG"
  "CS Dubois-Martin (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",38.2111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-86.9611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7292\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7293, 1, 'NAD83(2011) / InGCS Elkhart-Kosciusko-Wabash (m)', 'EPSG', 7293, 'PROJCS[\"NAD83(201"
  "1) / InGCS Elkhart-Kosciusko-Wabash (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 201"
  "1)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",40.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-85.85,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000033,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7293\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7294, 1, 'NAD83(2011) / InGCS Elkhart-Kosciusko-Wabash (ftUS)', 'EPSG', 7294, 'PROJCS[\"NAD83("
  "2011) / InGCS Elkhart-Kosciusko-Wabash (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude "
  "of natural origin\",-85.85,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000033,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7294\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7295, 1, 'NAD83(2011) / InGCS Fayette-Franklin-Union (m)', 'EPSG', 7295, 'PROJCS[\"NAD83(2011)"
  " / InGCS Fayette-Franklin-Union (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",39.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-85.05,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"7295\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7296, 1, 'NAD83(2011) / InGCS Fayette-Franklin-Union (ftUS)', 'EPSG', 7296, 'PROJCS[\"NAD83(20"
  "11) / InGCS Fayette-Franklin-Union (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",39.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-85.05,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7296\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7297, 1, 'NAD83(2011) / InGCS Fountain-Warren (m)', 'EPSG', 7297, 'PROJCS[\"NAD83(2011) / InGC"
  "S Fountain-Warren (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",39.95,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87."
  "3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000025,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"72"
  "97\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7298, 1, 'NAD83(2011) / InGCS Fountain-Warren (ftUS)', 'EPSG', 7298, 'PROJCS[\"NAD83(2011) / I"
  "nGCS Fountain-Warren (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",39.95,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-87.3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000025,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7298\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7299, 1, 'NAD83(2011) / InGCS Fulton-Marshall-St. Joseph (m)', 'EPSG', 7299, 'PROJCS[\"NAD83(2"
  "011) / InGCS Fulton-Marshall-St. Joseph (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"L"
  "ongitude of natural origin\",-86.3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003"
  "1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"7299\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7300, 1, 'NAD83(2011) / InGCS Fulton-Marshall-St. Joseph (ftUS)', 'EPSG', 7300, 'PROJCS[\"NAD8"
  "3(2011) / InGCS Fulton-Marshall-St. Joseph (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference "
  "System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORIT"
  "Y[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",-86.3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1"
  ".000031,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7300\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7301, 1, 'NAD83(2011) / InGCS Gibson (m)', 'EPSG', 7301, 'PROJCS[\"NAD83(2011) / InGCS Gibson "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",38.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.65,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000013,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7301\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7302, 1, 'NAD83(2011) / InGCS Gibson (ftUS)', 'EPSG', 7302, 'PROJCS[\"NAD83(2011) / InGCS Gibs"
  "on (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",38.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.65,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000013,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7302\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7303, 1, 'NAD83(2011) / InGCS Grant (m)', 'EPSG', 7303, 'PROJCS[\"NAD83(2011) / InGCS Grant (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",40.35,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.7,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7303\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7304, 1, 'NAD83(2011) / InGCS Grant (ftUS)', 'EPSG', 7304, 'PROJCS[\"NAD83(2011) / InGCS Grant"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",40.35,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.7,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7304\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7305, 1, 'NAD83(2011) / InGCS Hamilton-Tipton (m)', 'EPSG', 7305, 'PROJCS[\"NAD83(2011) / InGC"
  "S Hamilton-Tipton (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",39.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-86,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EP"
  "SG\",\"7305\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7306, 1, 'NAD83(2011) / InGCS Hamilton-Tipton (ftUS)', 'EPSG', 7306, 'PROJCS[\"NAD83(2011) / I"
  "nGCS Hamilton-Tipton (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",39.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-86,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7306\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7307, 1, 'NAD83(2011) / InGCS Hancock-Madison (m)', 'EPSG', 7307, 'PROJCS[\"NAD83(2011) / InGC"
  "S Hancock-Madison (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",39.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85."
  "8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"73"
  "07\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7308, 1, 'NAD83(2011) / InGCS Hancock-Madison (ftUS)', 'EPSG', 7308, 'PROJCS[\"NAD83(2011) / I"
  "nGCS Hancock-Madison (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",39.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-85.8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7308\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7309, 1, 'NAD83(2011) / InGCS Harrison-Washington (m)', 'EPSG', 7309, 'PROJCS[\"NAD83(2011) / "
  "InGCS Harrison-Washington (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",37.95,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-86.15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EP"
  "SG\",\"7309\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7310, 1, 'NAD83(2011) / InGCS Harrison-Washington (ftUS)', 'EPSG', 7310, 'PROJCS[\"NAD83(2011)"
  " / InGCS Harrison-Washington (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",37.95,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-86.15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7310\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7311, 1, 'NAD83(2011) / InGCS Henry (m)', 'EPSG', 7311, 'PROJCS[\"NAD83(2011) / InGCS Henry (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",39.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.4611111111111,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000043,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7311\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7312, 1, 'NAD83(2011) / InGCS Henry (ftUS)', 'EPSG', 7312, 'PROJCS[\"NAD83(2011) / InGCS Henry"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",39.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.4611111111111,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000043,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"7312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7313, 1, 'NAD83(2011) / InGCS Howard-Miami (m)', 'EPSG', 7313, 'PROJCS[\"NAD83(2011) / InGCS H"
  "oward-Miami (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",40.35,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.15,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7313\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7314, 1, 'NAD83(2011) / InGCS Howard-Miami (ftUS)', 'EPSG', 7314, 'PROJCS[\"NAD83(2011) / InGC"
  "S Howard-Miami (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",40.35,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86."
  "15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7314\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7315, 1, 'NAD83(2011) / InGCS Huntington-Whitley (m)', 'EPSG', 7315, 'PROJCS[\"NAD83(2011) / I"
  "nGCS Huntington-Whitley (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",40.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-85.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7315\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7316, 1, 'NAD83(2011) / InGCS Huntington-Whitley (ftUS)', 'EPSG', 7316, 'PROJCS[\"NAD83(2011) "
  "/ InGCS Huntington-Whitley (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",40.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-85.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1"
  "18110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7317, 1, 'NAD83(2011) / InGCS Jackson (m)', 'EPSG', 7317, 'PROJCS[\"NAD83(2011) / InGCS Jackso"
  "n (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",38.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.9611111111111,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000022,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7317\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7318, 1, 'NAD83(2011) / InGCS Jackson (ftUS)', 'EPSG', 7318, 'PROJCS[\"NAD83(2011) / InGCS Jac"
  "kson (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",38.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.96111111111"
  "11,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000022,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7319, 1, 'NAD83(2011) / InGCS Jasper-Porter (m)', 'EPSG', 7319, 'PROJCS[\"NAD83(2011) / InGCS "
  "Jasper-Porter (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",40.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.1,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"88"
  "07\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7319\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7320, 1, 'NAD83(2011) / InGCS Jasper-Porter (ftUS)', 'EPSG', 7320, 'PROJCS[\"NAD83(2011) / InG"
  "CS Jasper-Porter (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",40.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87"
  ".1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7320\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7321, 1, 'NAD83(2011) / InGCS Jay (m)', 'EPSG', 7321, 'PROJCS[\"NAD83(2011) / InGCS Jay (m)\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.3111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7321\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7322, 1, 'NAD83(2011) / InGCS Jay (ftUS)', 'EPSG', 7322, 'PROJCS[\"NAD83(2011) / InGCS Jay (ft"
  "US)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",40.3111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7323, 1, 'NAD83(2011) / InGCS Jefferson (m)', 'EPSG', 7323, 'PROJCS[\"NAD83(2011) / InGCS Jeff"
  "erson (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",38.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.3"
  "611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000028,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"7323\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7324, 1, 'NAD83(2011) / InGCS Jefferson (ftUS)', 'EPSG', 7324, 'PROJCS[\"NAD83(2011) / InGCS J"
  "efferson (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",38.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-85.3611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000028,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7325, 1, 'NAD83(2011) / InGCS Jennings (m)', 'EPSG', 7325, 'PROJCS[\"NAD83(2011) / InGCS Jenni"
  "ngs (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",38.8111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.8,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000025,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7325\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7326, 1, 'NAD83(2011) / InGCS Jennings (ftUS)', 'EPSG', 7326, 'PROJCS[\"NAD83(2011) / InGCS Je"
  "nnings (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",38.8111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "5.8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000025,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7326\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7327, 1, 'NAD83(2011) / InGCS Johnson-Marion (m)', 'EPSG', 7327, 'PROJCS[\"NAD83(2011) / InGCS"
  " Johnson-Marion (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",39.3111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-86.15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"7327\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7328, 1, 'NAD83(2011) / InGCS Johnson-Marion (ftUS)', 'EPSG', 7328, 'PROJCS[\"NAD83(2011) / In"
  "GCS Johnson-Marion (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",39.3111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-86.15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7328\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7329, 1, 'NAD83(2011) / InGCS Knox (m)', 'EPSG', 7329, 'PROJCS[\"NAD83(2011) / InGCS Knox (m)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",38.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.4611111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7329\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7330, 1, 'NAD83(2011) / InGCS Knox (ftUS)', 'EPSG', 7330, 'PROJCS[\"NAD83(2011) / InGCS Knox ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",38.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.4611111111111,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7331, 1, 'NAD83(2011) / InGCS LaGrange-Noble (m)', 'EPSG', 7331, 'PROJCS[\"NAD83(2011) / InGCS"
  " LaGrange-Noble (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",41.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-85.4611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000037,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",360"
  "00,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"7331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7332, 1, 'NAD83(2011) / InGCS LaGrange-Noble (ftUS)', 'EPSG', 7332, 'PROJCS[\"NAD83(2011) / In"
  "GCS LaGrange-Noble (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",41.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-85.4611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000037,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7333, 1, 'NAD83(2011) / InGCS Lake-Newton (m)', 'EPSG', 7333, 'PROJCS[\"NAD83(2011) / InGCS La"
  "ke-Newton (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",40.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.4111111111"
  "111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "7333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7334, 1, 'NAD83(2011) / InGCS Lake-Newton (ftUS)', 'EPSG', 7334, 'PROJCS[\"NAD83(2011) / InGCS"
  " Lake-Newton (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",40.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.411"
  "1111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7334\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7335, 1, 'NAD83(2011) / InGCS LaPorte-Pulaski-Starke (m)', 'EPSG', 7335, 'PROJCS[\"NAD83(2011)"
  " / InGCS LaPorte-Pulaski-Starke (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",40.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-86.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"7335\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7336, 1, 'NAD83(2011) / InGCS LaPorte-Pulaski-Starke (ftUS)', 'EPSG', 7336, 'PROJCS[\"NAD83(20"
  "11) / InGCS LaPorte-Pulaski-Starke (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-86.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000027"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7336\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7337, 1, 'NAD83(2011) / InGCS Monroe-Morgan (m)', 'EPSG', 7337, 'PROJCS[\"NAD83(2011) / InGCS "
  "Monroe-Morgan (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",38.95,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.5111"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000028,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"7337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7338, 1, 'NAD83(2011) / InGCS Monroe-Morgan (ftUS)', 'EPSG', 7338, 'PROJCS[\"NAD83(2011) / InG"
  "CS Monroe-Morgan (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",38.95,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "6.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000028,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7338\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7339, 1, 'NAD83(2011) / InGCS Montgomery-Putnam (m)', 'EPSG', 7339, 'PROJCS[\"NAD83(2011) / In"
  "GCS Montgomery-Putnam (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",39.45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-86.9611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7340, 1, 'NAD83(2011) / InGCS Montgomery-Putnam (ftUS)', 'EPSG', 7340, 'PROJCS[\"NAD83(2011) /"
  " InGCS Montgomery-Putnam (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",39.45,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-86.9611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118"
  "110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7341, 1, 'NAD83(2011) / InGCS Owen (m)', 'EPSG', 7341, 'PROJCS[\"NAD83(2011) / InGCS Owen (m)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",39.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.9,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7341\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7342, 1, 'NAD83(2011) / InGCS Owen (ftUS)', 'EPSG', 7342, 'PROJCS[\"NAD83(2011) / InGCS Owen ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",39.15,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.9,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"7342\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7343, 1, 'NAD83(2011) / InGCS Parke-Vermillion (m)', 'EPSG', 7343, 'PROJCS[\"NAD83(2011) / InG"
  "CS Parke-Vermillion (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",39.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-87.3611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000022,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"7343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7344, 1, 'NAD83(2011) / InGCS Parke-Vermillion (ftUS)', 'EPSG', 7344, 'PROJCS[\"NAD83(2011) / "
  "InGCS Parke-Vermillion (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",39.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-87.3611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000022,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",118110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7345, 1, 'NAD83(2011) / InGCS Perry (m)', 'EPSG', 7345, 'PROJCS[\"NAD83(2011) / InGCS Perry (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",37.8111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.7,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7345\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7346, 1, 'NAD83(2011) / InGCS Perry (ftUS)', 'EPSG', 7346, 'PROJCS[\"NAD83(2011) / InGCS Perry"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",37.8111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-86.7,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7347, 1, 'NAD83(2011) / InGCS Pike-Warrick (m)', 'EPSG', 7347, 'PROJCS[\"NAD83(2011) / InGCS P"
  "ike-Warrick (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",37.8611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-87.3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"7347\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7348, 1, 'NAD83(2011) / InGCS Pike-Warrick (ftUS)', 'EPSG', 7348, 'PROJCS[\"NAD83(2011) / InGC"
  "S Pike-Warrick (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",37.8611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-87.3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7349, 1, 'NAD83(2011) / InGCS Posey (m)', 'EPSG', 7349, 'PROJCS[\"NAD83(2011) / InGCS Posey (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",37.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.9611111111111,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000013,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7349\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7350, 1, 'NAD83(2011) / InGCS Posey (ftUS)', 'EPSG', 7350, 'PROJCS[\"NAD83(2011) / InGCS Posey"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",37.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.9611111111111,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000013,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"7350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7351, 1, 'NAD83(2011) / InGCS Randolph-Wayne (m)', 'EPSG', 7351, 'PROJCS[\"NAD83(2011) / InGCS"
  " Randolph-Wayne (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",39.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.05,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000044,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7351\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7352, 1, 'NAD83(2011) / InGCS Randolph-Wayne (ftUS)', 'EPSG', 7352, 'PROJCS[\"NAD83(2011) / In"
  "GCS Randolph-Wayne (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",39.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-"
  "85.05,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000044,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7353, 1, 'NAD83(2011) / InGCS Ripley (m)', 'EPSG', 7353, 'PROJCS[\"NAD83(2011) / InGCS Ripley "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",38.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.3,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7353\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7354, 1, 'NAD83(2011) / InGCS Ripley (ftUS)', 'EPSG', 7354, 'PROJCS[\"NAD83(2011) / InGCS Ripl"
  "ey (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",38.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.3"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7355, 1, 'NAD83(2011) / InGCS Shelby (m)', 'EPSG', 7355, 'PROJCS[\"NAD83(2011) / InGCS Shelby "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",39.3111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.9,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7355\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7356, 1, 'NAD83(2011) / InGCS Shelby (ftUS)', 'EPSG', 7356, 'PROJCS[\"NAD83(2011) / InGCS Shel"
  "by (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",39.3111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.9"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"7356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7357, 1, 'NAD83(2011) / InGCS Spencer (m)', 'EPSG', 7357, 'PROJCS[\"NAD83(2011) / InGCS Spence"
  "r (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",37.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.05,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000014,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7357\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7358, 1, 'NAD83(2011) / InGCS Spencer (ftUS)', 'EPSG', 7358, 'PROJCS[\"NAD83(2011) / InGCS Spe"
  "ncer (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",37.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.05,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000014,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7358\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7359, 1, 'NAD83(2011) / InGCS Steuben (m)', 'EPSG', 7359, 'PROJCS[\"NAD83(2011) / InGCS Steube"
  "n (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000041,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7359\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7360, 1, 'NAD83(2011) / InGCS Steuben (ftUS)', 'EPSG', 7360, 'PROJCS[\"NAD83(2011) / InGCS Ste"
  "uben (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000041,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7360\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7361, 1, 'NAD83(2011) / InGCS Sullivan (m)', 'EPSG', 7361, 'PROJCS[\"NAD83(2011) / InGCS Sulli"
  "van (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",38.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5111"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000017,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"7361\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7362, 1, 'NAD83(2011) / InGCS Sullivan (ftUS)', 'EPSG', 7362, 'PROJCS[\"NAD83(2011) / InGCS Su"
  "llivan (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",38.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "7.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000017,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7362\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7363, 1, 'NAD83(2011) / InGCS Tippecanoe-White (m)', 'EPSG', 7363, 'PROJCS[\"NAD83(2011) / InG"
  "CS Tippecanoe-White (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",40.2111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-86.9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7363\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7364, 1, 'NAD83(2011) / InGCS Tippecanoe-White (ftUS)', 'EPSG', 7364, 'PROJCS[\"NAD83(2011) / "
  "InGCS Tippecanoe-White (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",40.2111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-86.9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000026,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",11811"
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7364\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7365, 1, 'NAD83(2011) / InGCS Vanderburgh (m)', 'EPSG', 7365, 'PROJCS[\"NAD83(2011) / InGCS Va"
  "nderburgh (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",37.8111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "7.5611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7365\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7366, 1, 'NAD83(2011) / InGCS Vanderburgh (ftUS)', 'EPSG', 7366, 'PROJCS[\"NAD83(2011) / InGCS"
  " Vanderburgh (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",37.8111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-87.5611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000015,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118"
  "110,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7366\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7367, 1, 'NAD83(2011) / InGCS Vigo (m)', 'EPSG', 7367, 'PROJCS[\"NAD83(2011) / InGCS Vigo (m)\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",39.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.4611111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7367\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7368, 1, 'NAD83(2011) / InGCS Vigo (ftUS)', 'EPSG', 7368, 'PROJCS[\"NAD83(2011) / InGCS Vigo ("
  "ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",39.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.46111"
  "11111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"7368\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7369, 1, 'NAD83(2011) / InGCS Wells (m)', 'EPSG', 7369, 'PROJCS[\"NAD83(2011) / InGCS Wells (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",40.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.25,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",240000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",36000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7369\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7370, 1, 'NAD83(2011) / InGCS Wells (ftUS)', 'EPSG', 7370, 'PROJCS[\"NAD83(2011) / InGCS Wells"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",40.5611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.25,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",787400,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",118110,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"7370\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7373, 1, 'ONGD14', 'EPSG', 7373, 'GEOGCS[\"ONGD14\",DATUM[\"Oman National Geodetic Datum 2014\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"1147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7373\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7374, 1, 'ONGD14 / UTM zone 39N', 'EPSG', 7374, 'PROJCS[\"ONGD14 / UTM zone 39N\",GEOGCS[\"ONG"
  "D14\",DATUM[\"Oman National Geodetic Datum 2014\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"7373\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7374\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7375, 1, 'ONGD14 / UTM zone 40N', 'EPSG', 7375, 'PROJCS[\"ONGD14 / UTM zone 40N\",GEOGCS[\"ONG"
  "D14\",DATUM[\"Oman National Geodetic Datum 2014\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"7373\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7375\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7376, 1, 'ONGD14 / UTM zone 41N', 'EPSG', 7376, 'PROJCS[\"ONGD14 / UTM zone 41N\",GEOGCS[\"ONG"
  "D14\",DATUM[\"Oman National Geodetic Datum 2014\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1147\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"7373\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7376\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7528, 1, 'NAD83(2011) / WISCRS Adams and Juneau (m)', 'EPSG', 7528, 'PROJCS[\"NAD83(2011) / WI"
  "SCRS Adams and Juneau (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43.3777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000365285,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",147218.6942,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "0037,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7529, 1, 'NAD83(2011) / WISCRS Ashland (m)', 'EPSG', 7529, 'PROJCS[\"NAD83(2011) / WISCRS Ashl"
  "and (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",45.7061111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.6222"
  "222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495683,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",172821.9461,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0017,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7530, 1, 'NAD83(2011) / WISCRS Barron (m)', 'EPSG', 7530, 'PROJCS[\"NAD83(2011) / WISCRS Barro"
  "n (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",45.1444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.85,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000486665,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",93150,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0029,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"75"
  "30\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7531, 1, 'NAD83(2011) / WISCRS Bayfield (m)', 'EPSG', 7531, 'PROJCS[\"NAD83(2011) / WISCRS Bay"
  "field (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",46.6696483772222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-91.1527777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000331195,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",228600.4575,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",148551.4837,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7532, 1, 'NAD83(2011) / WISCRS Brown (m)', 'EPSG', 7532, 'PROJCS[\"NAD83(2011) / WISCRS Brown "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",43,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",31600,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4600,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7532\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7533, 1, 'NAD83(2011) / WISCRS Buffalo (m)', 'EPSG', 7533, 'PROJCS[\"NAD83(2011) / WISCRS Buff"
  "alo (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",43.4813888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.7972"
  "222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000382778,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",175260.3502,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0048,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7534, 1, 'NAD83(2011) / WISCRS Burnett (m)', 'EPSG', 7534, 'PROJCS[\"NAD83(2011) / WISCRS Burn"
  "ett (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",45.8987148658333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-92.4577777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000383841,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",64008.1276,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",59445.9043,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7534\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7535, 1, 'NAD83(2011) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (m)', 'EPSG', 753"
  "5, 'PROJCS[\"NAD83(2011) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\""
  "NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTIO"
  "N[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.7194444444444,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",1.0000286569,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "44754.8893,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0049,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7536, 1, 'NAD83(2011) / WISCRS Chippewa (m)', 'EPSG', 7536, 'PROJCS[\"NAD83(2011) / WISCRS Chi"
  "ppewa (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",44.9778568986111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-91.2944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000391127,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60045.72,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",44091.4346,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7536\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7537, 1, 'NAD83(2011) / WISCRS Clark (m)', 'EPSG', 7537, 'PROJCS[\"NAD83(2011) / WISCRS Clark "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",43.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.7083333"
  "333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000463003,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",199949.1989,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0086,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"7537\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7538, 1, 'NAD83(2011) / WISCRS Columbia (m)', 'EPSG', 7538, 'PROJCS[\"NAD83(2011) / WISCRS Col"
  "umbia (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",43.4625466458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-89.3944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003498,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",169164.3381,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",111569.6134,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"7538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7539, 1, 'NAD83(2011) / WISCRS Crawford (m)', 'EPSG', 7539, 'PROJCS[\"NAD83(2011) / WISCRS Cra"
  "wford (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",43.200055605,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-90.9388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000349151,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",113690.6274,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",53703.1201,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7540, 1, 'NAD83(2011) / WISCRS Dane (m)', 'EPSG', 7540, 'PROJCS[\"NAD83(2011) / WISCRS Dane (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Lati"
  "tude of natural origin\",43.0695160375,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.42"
  "22222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000384786,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",247193.2944,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",146591"
  ".9896,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"7540\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7541, 1, 'NAD83(2011) / WISCRS Dodge and Jefferson (m)', 'EPSG', 7541, 'PROJCS[\"NAD83(2011) /"
  " WISCRS Dodge and Jefferson (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",41.4722222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-88.775,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000346418,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",263347.7263,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0.0076,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7541\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7542, 1, 'NAD83(2011) / WISCRS Door (m)', 'EPSG', 7542, 'PROJCS[\"NAD83(2011) / WISCRS Door (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",44.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.2722222222222,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000187521,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",158801.1176,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0023,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"75"
  "42\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7543, 1, 'NAD83(2011) / WISCRS Douglas (m)', 'EPSG', 7543, 'PROJCS[\"NAD83(2011) / WISCRS Doug"
  "las (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",45.8833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.9277"
  "777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000385418,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",59131.3183,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0041,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7543\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7544, 1, 'NAD83(2011) / WISCRS Dunn (m)', 'EPSG', 7544, 'PROJCS[\"NAD83(2011) / WISCRS Dunn (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",44.4083333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.894444444"
  "4445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000410324,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",51816.104,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.003,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"7544\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7545, 1, 'NAD83(2011) / WISCRS Eau Claire (m)', 'EPSG', 7545, 'PROJCS[\"NAD83(2011) / WISCRS E"
  "au Claire (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAM"
  "ETER[\"Latitude of natural origin\",44.8722811263889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-91.2888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000035079,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120091.4402,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",91687.9239,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7545\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7546, 1, 'NAD83(2011) / WISCRS Florence (m)', 'EPSG', 7546, 'PROJCS[\"NAD83(2011) / WISCRS Flo"
  "rence (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",45.4388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.1"
  "416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000552095,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",133502.6683,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.00"
  "63,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"7546\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7547, 1, 'NAD83(2011) / WISCRS Forest (m)', 'EPSG', 7547, 'PROJCS[\"NAD83(2011) / WISCRS Fores"
  "t (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",44.0055555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.64444"
  "44444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000673004,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",275844.5533,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0157,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7547\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7548, 1, 'NAD83(2011) / WISCRS Grant (m)', 'EPSG', 7548, 'PROJCS[\"NAD83(2011) / WISCRS Grant "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",41.4111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.8,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000349452,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",242316.4841,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.01,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"75"
  "48\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7549, 1, 'NAD83(2011) / WISCRS Green and Lafayette (m)', 'EPSG', 7549, 'PROJCS[\"NAD83(2011) /"
  " WISCRS Green and Lafayette (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.6375622769444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-89.8388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000390487,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",170078.7403,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",45830.2947,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7549\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7550, 1, 'NAD83(2011) / WISCRS Green Lake and Marquette (m)', 'EPSG', 7550, 'PROJCS[\"NAD83(20"
  "11) / WISCRS Green Lake and Marquette (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.8070001177778,AUTHORITY[\"EPSG\",\"8801\"]],PARAM"
  "ETER[\"Longitude of natural origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natu"
  "ral origin\",1.0000344057,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150876.3018,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",79170.7795,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7550\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7551, 1, 'NAD83(2011) / WISCRS Iowa (m)', 'EPSG', 7551, 'PROJCS[\"NAD83(2011) / WISCRS Iowa (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",42.5388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.161111111"
  "1111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000394961,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",113081.0261,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0045,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7551\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7552, 1, 'NAD83(2011) / WISCRS Iron (m)', 'EPSG', 7552, 'PROJCS[\"NAD83(2011) / WISCRS Iron (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",45.4444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.255555555"
  "5556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000677153,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",220980.4419,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0085,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7552\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7553, 1, 'NAD83(2011) / WISCRS Jackson (m)', 'EPSG', 7553, 'PROJCS[\"NAD83(2011) / WISCRS Jack"
  "son (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44.2533351277778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.8442"
  "965194444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000353,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",27000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",25000,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"7553\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7554, 1, 'NAD83(2011) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (m)', 'EPSG', 7554, 'PRO"
  "JCS[\"NAD83(2011) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (Nationa"
  "l Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT"
  "Y[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.2166666666667,AUTHORITY[\"EPSG\""
  ",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.8944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",1.0000260649,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",185928.3728,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0009,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORIT"
  "Y[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7554\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7555, 1, 'NAD83(2011) / WISCRS Kewaunee, Manitowoc and Sheboygan (m)', 'EPSG', 7555, 'PROJCS[\""
  "NAD83(2011) / WISCRS Kewaunee, Manitowoc and Sheboygan (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial R"
  "eference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\","
  "\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.2777777777778,AUTHORITY[\"EPSG\",\"8801\"]"
  "],PARAMETER[\"Longitude of natural origin\",-87.5611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor "
  "at natural origin\",1.0000233704,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",79857.7614,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",0.0012,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7555\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7556, 1, 'NAD83(2011) / WISCRS La Crosse (m)', 'EPSG', 7556, 'PROJCS[\"NAD83(2011) / WISCRS La"
  " Crosse (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",43.4511111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91"
  ".3277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000319985,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",130454.6598,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "0033,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7556\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7557, 1, 'NAD83(2011) / WISCRS Langlade (m)', 'EPSG', 7557, 'PROJCS[\"NAD83(2011) / WISCRS Lan"
  "glade (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",45.1542371052778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-89.0444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000627024,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",198425.197,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",105279.7829,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"7557\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7558, 1, 'NAD83(2011) / WISCRS Lincoln (m)', 'EPSG', 7558, 'PROJCS[\"NAD83(2011) / WISCRS Linc"
  "oln (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44.8444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.7444"
  "444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000599003,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",116129.0323,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0058,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"7558\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7559, 1, 'NAD83(2011) / WISCRS Marathon (m)', 'EPSG', 7559, 'PROJCS[\"NAD83(2011) / WISCRS Mar"
  "athon (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",44.9009044236111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-89.77,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000053289,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",74676.1493,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",55049.2"
  "669,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"7559\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7560, 1, 'NAD83(2011) / WISCRS Marinette (m)', 'EPSG', 7560, 'PROJCS[\"NAD83(2011) / WISCRS Ma"
  "rinette (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44.6916666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87"
  ".7111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000234982,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",238658.8794,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "0032,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7560\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7561, 1, 'NAD83(2011) / WISCRS Menominee (m)', 'EPSG', 7561, 'PROJCS[\"NAD83(2011) / WISCRS Me"
  "nominee (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44.7277777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88"
  ".4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000362499,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",105461.0121,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "0029,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7561\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7562, 1, 'NAD83(2011) / WISCRS Monroe (m)', 'EPSG', 7562, 'PROJCS[\"NAD83(2011) / WISCRS Monro"
  "e (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",44.0000739286111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-90.6416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000434122,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",204521.209,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",121923.9861,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7562\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7563, 1, 'NAD83(2011) / WISCRS Oconto (m)', 'EPSG', 7563, 'PROJCS[\"NAD83(2011) / WISCRS Ocont"
  "o (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",44.3972222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.90833"
  "33333334,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000236869,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",182880.3676,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0033,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7563\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7564, 1, 'NAD83(2011) / WISCRS Oneida (m)', 'EPSG', 7564, 'PROJCS[\"NAD83(2011) / WISCRS Oneid"
  "a (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",45.7042237702778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-89.5444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000686968,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",70104.1401,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",57588.0346,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7564\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7565, 1, 'NAD83(2011) / WISCRS Pepin and Pierce (m)', 'EPSG', 7565, 'PROJCS[\"NAD83(2011) / WI"
  "SCRS Pepin and Pierce (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",44.6361488719444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-92.2277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000362977,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",167640.3354,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",86033.0876,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7565\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7566, 1, 'NAD83(2011) / WISCRS Polk (m)', 'EPSG', 7566, 'PROJCS[\"NAD83(2011) / WISCRS Polk (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",44.6611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.644444444"
  "4444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000433849,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",141732.2823,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0059,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7566\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7567, 1, 'NAD83(2011) / WISCRS Portage (m)', 'EPSG', 7567, 'PROJCS[\"NAD83(2011) / WISCRS Port"
  "age (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",44.4168239752778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-89.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000039936,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",56388.1128,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",5"
  "0022.1874,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"7567\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7568, 1, 'NAD83(2011) / WISCRS Price (m)', 'EPSG', 7568, 'PROJCS[\"NAD83(2011) / WISCRS Price "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",44.5555555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.4888888"
  "888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000649554,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",227990.8546,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0109,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"7568\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7569, 1, 'NAD83(2011) / WISCRS Richland (m)', 'EPSG', 7569, 'PROJCS[\"NAD83(2011) / WISCRS Ric"
  "hland (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",43.3223129275,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-90.4305555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000375653,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",202387.6048,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",134255.4253,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7569\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7570, 1, 'NAD83(2011) / WISCRS Rock (m)', 'EPSG', 7570, 'PROJCS[\"NAD83(2011) / WISCRS Rock (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",41.9444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.072222222"
  "2222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000337311,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",146304.2926,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0068,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7570\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7571, 1, 'NAD83(2011) / WISCRS Rusk (m)', 'EPSG', 7571, 'PROJCS[\"NAD83(2011) / WISCRS Rusk (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",43.9194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.066666666"
  "6667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495976,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",250546.1013,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0234,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7571\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7572, 1, 'NAD83(2011) / WISCRS Sauk (m)', 'EPSG', 7572, 'PROJCS[\"NAD83(2011) / WISCRS Sauk (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",42.8194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.9,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000373868,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",185623.5716,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0051,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"75"
  "72\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7573, 1, 'NAD83(2011) / WISCRS Sawyer (m)', 'EPSG', 7573, 'PROJCS[\"NAD83(2011) / WISCRS Sawye"
  "r (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",45.9000991313889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-91.1277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000573461,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",216713.2336,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",120734.1631,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7573\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7574, 1, 'NAD83(2011) / WISCRS Shawano (m)', 'EPSG', 7574, 'PROJCS[\"NAD83(2011) / WISCRS Shaw"
  "ano (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.6055"
  "555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000032144,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",262433.3253,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0096,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"7574\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7575, 1, 'NAD83(2011) / WISCRS St. Croix (m)', 'EPSG', 7575, 'PROJCS[\"NAD83(2011) / WISCRS St"
  ". Croix (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92"
  ".6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000381803,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",165506.7302,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "0103,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7575\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7576, 1, 'NAD83(2011) / WISCRS Taylor (m)', 'EPSG', 7576, 'PROJCS[\"NAD83(2011) / WISCRS Taylo"
  "r (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",45.1778220858333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-90.4833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000597566,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",187147.5744,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",107746.7522,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7576\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7577, 1, 'NAD83(2011) / WISCRS Trempealeau (m)', 'EPSG', 7577, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Trempealeau (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",43.1611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-91.3777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000361538,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",256946.9138,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0.0041,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"7577\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7578, 1, 'NAD83(2011) / WISCRS Vernon (m)', 'EPSG', 7578, 'PROJCS[\"NAD83(2011) / WISCRS Verno"
  "n (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137"
  ",298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",43.5750329397222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-90.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000408158,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",222504.4451,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",47532.0602,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"7578\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7579, 1, 'NAD83(2011) / WISCRS Vilas (m)', 'EPSG', 7579, 'PROJCS[\"NAD83(2011) / WISCRS Vilas "
  "(m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"La"
  "titude of natural origin\",46.0778440905556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-"
  "89.4888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000730142,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",134417.0689,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "50337.1092,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"7579\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7580, 1, 'NAD83(2011) / WISCRS Walworth (m)', 'EPSG', 7580, 'PROJCS[\"NAD83(2011) / WISCRS Wal"
  "worth (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",42.6694620969444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-88.5416666666666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000367192,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",232562.8651,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",111088.2224,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7580\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7581, 1, 'NAD83(2011) / WISCRS Washburn (m)', 'EPSG', 7581, 'PROJCS[\"NAD83(2011) / WISCRS Was"
  "hburn (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",45.9612198333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-91.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000475376,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",234086.8682,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",188358.6058,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7581\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7582, 1, 'NAD83(2011) / WISCRS Washington (m)', 'EPSG', 7582, 'PROJCS[\"NAD83(2011) / WISCRS W"
  "ashington (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",42.9180555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "8.0638888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003738,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",120091.4415,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0"
  "03,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"7582\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7583, 1, 'NAD83(2011) / WISCRS Waukesha (m)', 'EPSG', 7583, 'PROJCS[\"NAD83(2011) / WISCRS Wau"
  "kesha (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",42.5694444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.2"
  "25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000346179,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",208788.418,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0034,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EP"
  "SG\",\"7583\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7584, 1, 'NAD83(2011) / WISCRS Waupaca (m)', 'EPSG', 7584, 'PROJCS[\"NAD83(2011) / WISCRS Waup"
  "aca (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",43.4202777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.8277"
  "777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000333645,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",185013.9709,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.007,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7584\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7585, 1, 'NAD83(2011) / WISCRS Waushara (m)', 'EPSG', 7585, 'PROJCS[\"NAD83(2011) / WISCRS Wau"
  "shara (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",44.1139440458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000392096,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120091.4402,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",45069.7587,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"7585\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7586, 1, 'NAD83(2011) / WISCRS Wood (m)', 'EPSG', 7586, 'PROJCS[\"NAD83(2011) / WISCRS Wood (m"
  ")\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298"
  ".257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44.3625954694444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000421209,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",208483.6173,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",134589.754,AUTHO"
  "RITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"7586\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7587, 1, 'NAD83(2011) / WISCRS Adams and Juneau (ftUS)', 'EPSG', 7587, 'PROJCS[\"NAD83(2011) /"
  " WISCRS Adams and Juneau (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",43.3777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000365285,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",482999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0.012,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7587\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7588, 1, 'NAD83(2011) / WISCRS Ashland (ftUS)', 'EPSG', 7588, 'PROJCS[\"NAD83(2011) / WISCRS A"
  "shland (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",45.7061111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "0.6222222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495683,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",567000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "006,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7588\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7589, 1, 'NAD83(2011) / WISCRS Barron (ftUS)', 'EPSG', 7589, 'PROJCS[\"NAD83(2011) / WISCRS Ba"
  "rron (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",45.1444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91"
  ".85,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000486665,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",305609.625,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.01,AUTHORITY"
  "[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7589\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7590, 1, 'NAD83(2011) / WISCRS Bayfield (ftUS)', 'EPSG', 7590, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Bayfield (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",46.6696483772222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-91.1527777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000331195,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",750000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",487372.659,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7590\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7591, 1, 'NAD83(2011) / WISCRS Brown (ftUS)', 'EPSG', 7591, 'PROJCS[\"NAD83(2011) / WISCRS Bro"
  "wn (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",43,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",103674.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",15091.833,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"7591\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7592, 1, 'NAD83(2011) / WISCRS Buffalo (ftUS)', 'EPSG', 7592, 'PROJCS[\"NAD83(2011) / WISCRS B"
  "uffalo (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",43.4813888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "1.7972222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000382778,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",574999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "016,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7592\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7593, 1, 'NAD83(2011) / WISCRS Burnett (ftUS)', 'EPSG', 7593, 'PROJCS[\"NAD83(2011) / WISCRS B"
  "urnett (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAM"
  "ETER[\"Latitude of natural origin\",45.8987148658333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-92.4577777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000383841,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",209999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",195032.104,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7593\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7594, 1, 'NAD83(2011) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (ftUS)', 'EPSG', "
  "7594, 'PROJCS[\"NAD83(2011) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (ftUS)\",GEOGCS[\"NAD83(2011)\",DA"
  "TUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.71944444444"
  "44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.5111111111111,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",1.0000286569,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",802999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.016,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPS"
  "G\",\"7594\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7595, 1, 'NAD83(2011) / WISCRS Chippewa (ftUS)', 'EPSG', 7595, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Chippewa (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",44.9778568986111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-91.2944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000391127,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",197000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",144656.648,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7595\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7596, 1, 'NAD83(2011) / WISCRS Clark (ftUS)', 'EPSG', 7596, 'PROJCS[\"NAD83(2011) / WISCRS Cla"
  "rk (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",43.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.7"
  "083333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000463003,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",655999.997,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.028"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7596\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7597, 1, 'NAD83(2011) / WISCRS Columbia (ftUS)', 'EPSG', 7597, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Columbia (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43.4625466458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-89.3944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003498,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",554999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",366041.307,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"90"
  "03\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7597\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7598, 1, 'NAD83(2011) / WISCRS Crawford (ftUS)', 'EPSG', 7598, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Crawford (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43.200055605,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-90.9388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000349151,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",373000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",176190.987,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7598\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7599, 1, 'NAD83(2011) / WISCRS Dane (ftUS)', 'EPSG', 7599, 'PROJCS[\"NAD83(2011) / WISCRS Dane"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",43.0695160375,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "9.4222222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000384786,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",811000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",480943"
  ".886,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7599\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7600, 1, 'NAD83(2011) / WISCRS Dodge and Jefferson (ftUS)', 'EPSG', 7600, 'PROJCS[\"NAD83(2011"
  ") / WISCRS Dodge and Jefferson (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",41.4722222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-88.775,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000346418"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",863999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",0.025,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7601, 1, 'NAD83(2011) / WISCRS Door (ftUS)', 'EPSG', 7601, 'PROJCS[\"NAD83(2011) / WISCRS Door"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.2722222222222,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000187521,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",521000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.008,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"7601\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7602, 1, 'NAD83(2011) / WISCRS Douglas (ftUS)', 'EPSG', 7602, 'PROJCS[\"NAD83(2011) / WISCRS D"
  "ouglas (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",45.8833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "1.9277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000385418,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",194000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.013,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7602\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7603, 1, 'NAD83(2011) / WISCRS Dunn (ftUS)', 'EPSG', 7603, 'PROJCS[\"NAD83(2011) / WISCRS Dunn"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44.4083333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.8944"
  "444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000410324,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",170000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.01,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7603\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7604, 1, 'NAD83(2011) / WISCRS Eau Claire (ftUS)', 'EPSG', 7604, 'PROJCS[\"NAD83(2011) / WISCR"
  "S Eau Claire (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",44.8722811263889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-91.2888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000350"
  "79,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",394000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",300812.797,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7604\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7605, 1, 'NAD83(2011) / WISCRS Florence (ftUS)', 'EPSG', 7605, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Florence (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",45.4388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-88.1416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000552095,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",438000.004,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0.021,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7605\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7606, 1, 'NAD83(2011) / WISCRS Forest (ftUS)', 'EPSG', 7606, 'PROJCS[\"NAD83(2011) / WISCRS Fo"
  "rest (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44.0055555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88"
  ".6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000673004,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",905000.005,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0"
  "52,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7606\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7607, 1, 'NAD83(2011) / WISCRS Grant (ftUS)', 'EPSG', 7607, 'PROJCS[\"NAD83(2011) / WISCRS Gra"
  "nt (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",41.4111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.8"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000349452,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",794999.998,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.033,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"7607\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7608, 1, 'NAD83(2011) / WISCRS Green and Lafayette (ftUS)', 'EPSG', 7608, 'PROJCS[\"NAD83(2011"
  ") / WISCRS Green and Lafayette (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY["
  "\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.6375622769444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-89.8388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1.0000390487,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",558000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",150361.559,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7608\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7609, 1, 'NAD83(2011) / WISCRS Green Lake and Marquette (ftUS)', 'EPSG', 7609, 'PROJCS[\"NAD83"
  "(2011) / WISCRS Green Lake and Marquette (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Sy"
  "stem 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\","
  "AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.8070001177778,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",1.0000344057,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",495000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",259746.132,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800609601"
  "21924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7609\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7610, 1, 'NAD83(2011) / WISCRS Iowa (ftUS)', 'EPSG', 7610, 'PROJCS[\"NAD83(2011) / WISCRS Iowa"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",42.5388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.1611"
  "111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000394961,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",371000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.015,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7610\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7611, 1, 'NAD83(2011) / WISCRS Iron (ftUS)', 'EPSG', 7611, 'PROJCS[\"NAD83(2011) / WISCRS Iron"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",45.4444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.2555"
  "555555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000677153,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",725000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.028,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7611\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7612, 1, 'NAD83(2011) / WISCRS Jackson (ftUS)', 'EPSG', 7612, 'PROJCS[\"NAD83(2011) / WISCRS J"
  "ackson (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",44.2533351277778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9"
  "0.8442965194444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000353,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",88582.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",82020.83"
  "3,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7612\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7613, 1, 'NAD83(2011) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (ftUS)', 'EPSG', 7613, '"
  "PROJCS[\"NAD83(2011) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (N"
  "ational Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.2166666666667,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.8944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1.0000260649,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",610000.003"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.003,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot"
  "\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7613\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7614, 1, 'NAD83(2011) / WISCRS Kewaunee, Manitowoc and Sheboygan (ftUS)', 'EPSG', 7614, 'PROJC"
  "S[\"NAD83(2011) / WISCRS Kewaunee, Manitowoc and Sheboygan (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Sp"
  "atial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\""
  "EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.2777777777778,AUTHORITY[\"EPSG\",\""
  "8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1.0000233704,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262000.006,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.004,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.3048"
  "0060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7614\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7615, 1, 'NAD83(2011) / WISCRS La Crosse (ftUS)', 'EPSG', 7615, 'PROJCS[\"NAD83(2011) / WISCRS"
  " La Crosse (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",43.4511111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-91.3277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000319985,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",427999.996,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0.011,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7615\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7616, 1, 'NAD83(2011) / WISCRS Langlade (ftUS)', 'EPSG', 7616, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Langlade (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",45.1542371052778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-89.0444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000627024,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",651000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",345405.421,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7616\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7617, 1, 'NAD83(2011) / WISCRS Lincoln (ftUS)', 'EPSG', 7617, 'PROJCS[\"NAD83(2011) / WISCRS L"
  "incoln (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",44.8444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "9.7444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000599003,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",381000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.019,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7617\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7618, 1, 'NAD83(2011) / WISCRS Marathon (ftUS)', 'EPSG', 7618, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Marathon (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",44.9009044236111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-89.77,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000053289,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",245000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",18060"
  "7.47,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7618\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7619, 1, 'NAD83(2011) / WISCRS Marinette (ftUS)', 'EPSG', 7619, 'PROJCS[\"NAD83(2011) / WISCRS"
  " Marinette (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",44.6916666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-87.7111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000234982,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",783000.007,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7619\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7620, 1, 'NAD83(2011) / WISCRS Menominee (ftUS)', 'EPSG', 7620, 'PROJCS[\"NAD83(2011) / WISCRS"
  " Menominee (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",44.7277777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-88.4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000362499,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",346000.004,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7620\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7621, 1, 'NAD83(2011) / WISCRS Monroe (ftUS)', 'EPSG', 7621, 'PROJCS[\"NAD83(2011) / WISCRS Mo"
  "nroe (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",44.0000739286111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-90.6416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000434122,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",671000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",400012.278,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7621\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7622, 1, 'NAD83(2011) / WISCRS Oconto (ftUS)', 'EPSG', 7622, 'PROJCS[\"NAD83(2011) / WISCRS Oc"
  "onto (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44.3972222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87"
  ".9083333333334,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000236869,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",600000.006,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0"
  "11,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7622\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7623, 1, 'NAD83(2011) / WISCRS Oneida (ftUS)', 'EPSG', 7623, 'PROJCS[\"NAD83(2011) / WISCRS On"
  "eida (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",45.7042237702778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-89.5444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000686968,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",230000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",188936.744,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7623\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7624, 1, 'NAD83(2011) / WISCRS Pepin and Pierce (ftUS)', 'EPSG', 7624, 'PROJCS[\"NAD83(2011) /"
  " WISCRS Pepin and Pierce (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6361488719444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-92.2277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000362977,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",550000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",282260.222,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7624\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7625, 1, 'NAD83(2011) / WISCRS Polk (ftUS)', 'EPSG', 7625, 'PROJCS[\"NAD83(2011) / WISCRS Polk"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44.6611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.6444"
  "444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000433849,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",464999.996,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.019,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7625\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7626, 1, 'NAD83(2011) / WISCRS Portage (ftUS)', 'EPSG', 7626, 'PROJCS[\"NAD83(2011) / WISCRS P"
  "ortage (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAM"
  "ETER[\"Latitude of natural origin\",44.4168239752778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-89.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000039936,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",185000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",164114.46,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7626\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7627, 1, 'NAD83(2011) / WISCRS Price (ftUS)', 'EPSG', 7627, 'PROJCS[\"NAD83(2011) / WISCRS Pri"
  "ce (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",44.5555555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.4"
  "888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000649554,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",747999.995,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.036"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7627\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7628, 1, 'NAD83(2011) / WISCRS Richland (ftUS)', 'EPSG', 7628, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Richland (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43.3223129275,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-90.4305555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000375653,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",664000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",440469.675,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7628\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7629, 1, 'NAD83(2011) / WISCRS Rock (ftUS)', 'EPSG', 7629, 'PROJCS[\"NAD83(2011) / WISCRS Rock"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",41.9444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.0722"
  "222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000337311,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",480000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.022,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7629\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7630, 1, 'NAD83(2011) / WISCRS Rusk (ftUS)', 'EPSG', 7630, 'PROJCS[\"NAD83(2011) / WISCRS Rusk"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",43.9194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-91.0666"
  "666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495976,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",822000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.077,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7630\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7631, 1, 'NAD83(2011) / WISCRS Sauk (ftUS)', 'EPSG', 7631, 'PROJCS[\"NAD83(2011) / WISCRS Sauk"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",42.8194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.9,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000373868,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",609000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.017,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7631\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7632, 1, 'NAD83(2011) / WISCRS Sawyer (ftUS)', 'EPSG', 7632, 'PROJCS[\"NAD83(2011) / WISCRS Sa"
  "wyer (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",45.9000991313889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-91.1277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000573461,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",711000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",396108.667,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7632\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7633, 1, 'NAD83(2011) / WISCRS Shawano (ftUS)', 'EPSG', 7633, 'PROJCS[\"NAD83(2011) / WISCRS S"
  "hawano (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "8.6055555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000032144,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",861000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0"
  "31,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7633\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7634, 1, 'NAD83(2011) / WISCRS St. Croix (ftUS)', 'EPSG', 7634, 'PROJCS[\"NAD83(2011) / WISCRS"
  " St. Croix (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-92.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000381803,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",542999.997,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0.034,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7634\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7635, 1, 'NAD83(2011) / WISCRS Taylor (ftUS)', 'EPSG', 7635, 'PROJCS[\"NAD83(2011) / WISCRS Ta"
  "ylor (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",45.1778220858333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-90.4833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000597566,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",614000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",353499.136,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7635\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7636, 1, 'NAD83(2011) / WISCRS Trempealeau (ftUS)', 'EPSG', 7636, 'PROJCS[\"NAD83(2011) / WISC"
  "RS Trempealeau (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",43.1611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-91.3777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000361538,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",843000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0.013,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7636\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7637, 1, 'NAD83(2011) / WISCRS Vernon (ftUS)', 'EPSG', 7637, 'PROJCS[\"NAD83(2011) / WISCRS Ve"
  "rnon (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6"
  "378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",43.5750329397222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-90.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000408158,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",730000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",155944.768,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7637\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7638, 1, 'NAD83(2011) / WISCRS Vilas (ftUS)', 'EPSG', 7638, 'PROJCS[\"NAD83(2011) / WISCRS Vil"
  "as (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",46.0778440905556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-89.4888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000730142,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",441000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",165147.666,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7638\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7639, 1, 'NAD83(2011) / WISCRS Walworth (ftUS)', 'EPSG', 7639, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Walworth (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",42.6694620969444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-88.5416666666666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000367192,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",763000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",364461.943,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7639\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7640, 1, 'NAD83(2011) / WISCRS Washburn (ftUS)', 'EPSG', 7640, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Washburn (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",45.9612198333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-91.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000475376,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",768000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",617973.193,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7640\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7641, 1, 'NAD83(2011) / WISCRS Washington (ftUS)', 'EPSG', 7641, 'PROJCS[\"NAD83(2011) / WISCR"
  "S Washington (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",42.9180555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-88.0638888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003738,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",394000.004,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7641\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7642, 1, 'NAD83(2011) / WISCRS Waukesha (ftUS)', 'EPSG', 7642, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Waukesha (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",42.5694444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-88.225,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000346179,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",685000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.011,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7642\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7643, 1, 'NAD83(2011) / WISCRS Waupaca (ftUS)', 'EPSG', 7643, 'PROJCS[\"NAD83(2011) / WISCRS W"
  "aupaca (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",43.4202777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "8.8277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000333645,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",607000.003,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0."
  "023,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7643\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7644, 1, 'NAD83(2011) / WISCRS Waushara (ftUS)', 'EPSG', 7644, 'PROJCS[\"NAD83(2011) / WISCRS "
  "Waushara (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",44.1139440458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000392096,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",394000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",147866.367,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003"
  "\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7644\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7645, 1, 'NAD83(2011) / WISCRS Wood (ftUS)', 'EPSG', 7645, 'PROJCS[\"NAD83(2011) / WISCRS Wood"
  " (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",44.3625954694444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000421209,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",684000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",441566.551,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST"
  "],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7645\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7683, 1, 'GSK-2011', 'EPSG', 7683, 'GEOGCS[\"GSK-2011\",DATUM[\"Geodezicheskaya Sistema Koordi"
  "nat 2011\",SPHEROID[\"GSK-2011\",6378136.5,298.2564151,AUTHORITY[\"EPSG\",\"1025\"]],AUTHORITY[\"EPSG\",\"1159\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7683\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7686, 1, 'Kyrg-06', 'EPSG', 7686, 'GEOGCS[\"Kyrg-06\",DATUM[\"Kyrgyzstan Geodetic Datum 2006\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1160\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7686\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7692, 1, 'Kyrg-06 / zone 1', 'EPSG', 7692, 'PROJCS[\"Kyrg-06 / zone 1\",GEOGCS[\"Kyrg-06\",DAT"
  "UM[\"Kyrgyzstan Geodetic Datum 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO"
  "RITY[\"EPSG\",\"1160\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7686\"]],PROJECTION[\"Transve"
  "rse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",68.5166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1300000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",14743.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7693, 1, 'Kyrg-06 / zone 2', 'EPSG', 7693, 'PROJCS[\"Kyrg-06 / zone 2\",GEOGCS[\"Kyrg-06\",DAT"
  "UM[\"Kyrgyzstan Geodetic Datum 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO"
  "RITY[\"EPSG\",\"1160\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7686\"]],PROJECTION[\"Transve"
  "rse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",71.5166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2300000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",14743.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7693\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7694, 1, 'Kyrg-06 / zone 3', 'EPSG', 7694, 'PROJCS[\"Kyrg-06 / zone 3\",GEOGCS[\"Kyrg-06\",DAT"
  "UM[\"Kyrgyzstan Geodetic Datum 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO"
  "RITY[\"EPSG\",\"1160\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7686\"]],PROJECTION[\"Transve"
  "rse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",74.5166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3300000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",14743.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7694\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7695, 1, 'Kyrg-06 / zone 4', 'EPSG', 7695, 'PROJCS[\"Kyrg-06 / zone 4\",GEOGCS[\"Kyrg-06\",DAT"
  "UM[\"Kyrgyzstan Geodetic Datum 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO"
  "RITY[\"EPSG\",\"1160\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7686\"]],PROJECTION[\"Transve"
  "rse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",77.5166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4300000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",14743.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7695\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7696, 1, 'Kyrg-06 / zone 5', 'EPSG', 7696, 'PROJCS[\"Kyrg-06 / zone 5\",GEOGCS[\"Kyrg-06\",DAT"
  "UM[\"Kyrgyzstan Geodetic Datum 2006\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHO"
  "RITY[\"EPSG\",\"1160\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AU"
  "THORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7686\"]],PROJECTION[\"Transve"
  "rse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]]"
  ",PARAMETER[\"Longitude of natural origin\",80.5166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5300000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",14743.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7696\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7755, 1, 'WGS 84 / India NSF LCC', 'EPSG', 7755, 'PROJCS[\"WGS 84 / India NSF LCC\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"La"
  "mbert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",24,AUTHORITY[\"EPS"
  "G\",\"8821\"]],PARAMETER[\"Longitude of false origin\",80,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st s"
  "tandard parallel\",12.472955,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.1728044"
  "444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",4000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7755\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7756, 1, 'WGS 84 / Andhra Pradesh', 'EPSG', 7756, 'PROJCS[\"WGS 84 / Andhra Pradesh\",GEOGCS[\""
  "WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"L"
  "ambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",16.25543298,AUTHO"
  "RITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",80.875,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Lat"
  "itude of 1st standard parallel\",13.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "18.75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAM"
  "ETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7756\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7757, 1, 'WGS 84 / Arunachal Pradesh', 'EPSG', 7757, 'PROJCS[\"WGS 84 / Arunachal Pradesh\",GE"
  "OGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7"
  "030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECT"
  "ION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",28.0015789"
  "7,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",94.5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",27,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",29,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7757\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7758, 1, 'WGS 84 / Assam', 'EPSG', 7758, 'PROJCS[\"WGS 84 / Assam\",GEOGCS[\"WGS 84\",DATUM[\""
  "World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG"
  "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic Conf"
  "ormal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",26.00257703,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",92.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stand"
  "ard parallel\",24.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",27.3444"
  "444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],P"
  "ARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7758\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7759, 1, 'WGS 84 / Bihar', 'EPSG', 7759, 'PROJCS[\"WGS 84 / Bihar\",GEOGCS[\"WGS 84\",DATUM[\""
  "World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG"
  "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic Conf"
  "ormal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",25.87725247,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",85.875,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",24.625,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",27.125,AUTHORITY"
  "[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7759\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7760, 1, 'WGS 84 / Delhi', 'EPSG', 7760, 'PROJCS[\"WGS 84 / Delhi\",GEOGCS[\"WGS 84\",DATUM[\""
  "World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG"
  "\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"E"
  "PSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic Conf"
  "ormal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",28.62510126,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",77,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard"
  " parallel\",28.375,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",28.875,AUTHORITY[\"E"
  "PSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at "
  "false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7760\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7761, 1, 'WGS 84 / Gujarat', 'EPSG', 7761, 'PROJCS[\"WGS 84 / Gujarat\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",22.37807121,AUTHORITY[\"EPSG\""
  ",\"8821\"]],PARAMETER[\"Longitude of false origin\",71.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st "
  "standard parallel\",20.7916666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",23"
  ".9583333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7761\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7762, 1, 'WGS 84 / Haryana', 'EPSG', 7762, 'PROJCS[\"WGS 84 / Haryana\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29.25226266,AUTHORITY[\"EPSG\""
  ",\"8821\"]],PARAMETER[\"Longitude of false origin\",76,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",28.0833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",30.416"
  "6666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7762\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7763, 1, 'WGS 84 / Himachal Pradesh', 'EPSG', 7763, 'PROJCS[\"WGS 84 / Himachal Pradesh\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.75183497,"
  "AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",77.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",30.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",32.75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7763\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7764, 1, 'WGS 84 / Jammu and Kashmir', 'EPSG', 7764, 'PROJCS[\"WGS 84 / Jammu and Kashmir\",GE"
  "OGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7"
  "030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECT"
  "ION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.7557087"
  "4,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",76.5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER["
  "\"Latitude of 1st standard parallel\",33.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stan"
  "dard parallel\",36.4166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7764\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7765, 1, 'WGS 84 / Jharkhand', 'EPSG', 7765, 'PROJCS[\"WGS 84 / Jharkhand\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Co"
  "nic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",23.62652682,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",85.625,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",22.5416666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",24.7083333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7765\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7766, 1, 'WGS 84 / Madhya Pradesh', 'EPSG', 7766, 'PROJCS[\"WGS 84 / Madhya Pradesh\",GEOGCS[\""
  "WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]]"
  ",AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"L"
  "ambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",24.00529821,AUTHO"
  "RITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",78.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Lat"
  "itude of 1st standard parallel\",22,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26,"
  "AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS["
  "\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7766\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7767, 1, 'WGS 84 / Maharashtra', 'EPSG', 7767, 'PROJCS[\"WGS 84 / Maharashtra\",GEOGCS[\"WGS 8"
  "4\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTH"
  "ORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lamber"
  "t Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",18.88015774,AUTHORITY["
  "\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",76.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",16.625,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",21.12"
  "5,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7767\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7768, 1, 'WGS 84 / Manipur', 'EPSG', 7768, 'PROJCS[\"WGS 84 / Manipur\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",24.75060911,AUTHORITY[\"EPSG\""
  ",\"8821\"]],PARAMETER[\"Longitude of false origin\",94,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",24.0833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",25.416"
  "6666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7768\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7769, 1, 'WGS 84 / Meghalaya', 'EPSG', 7769, 'PROJCS[\"WGS 84 / Meghalaya\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Co"
  "nic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",25.62524747,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",91.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",25.2083333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",26.0416666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7769\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7770, 1, 'WGS 84 / Nagaland', 'EPSG', 7770, 'PROJCS[\"WGS 84 / Nagaland\",GEOGCS[\"WGS 84\",DA"
  "TUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY["
  "\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",26.12581974,AUTHORITY[\"EPSG"
  "\",\"8821\"]],PARAMETER[\"Longitude of false origin\",94.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",25.375,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",26.875,AUT"
  "HORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"No"
  "rthing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7770\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7771, 1, 'WGS 84 / India Northeast', 'EPSG', 7771, 'PROJCS[\"WGS 84 / India Northeast\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",25.63452135,AUTH"
  "ORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",93.5,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Lati"
  "tude of 1st standard parallel\",23.0416666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard p"
  "arallel\",28.2083333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EP"
  "SG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7771\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7772, 1, 'WGS 84 / Orissa', 'EPSG', 7772, 'PROJCS[\"WGS 84 / Orissa\",GEOGCS[\"WGS 84\",DATUM["
  "\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP"
  "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic Con"
  "formal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",20.25305174,AUTHORITY[\"EPSG\",\""
  "8821\"]],PARAMETER[\"Longitude of false origin\",84.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st sta"
  "ndard parallel\",18.5833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",21.91"
  "66666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]]"
  ",PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7772\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7773, 1, 'WGS 84 / Punjab', 'EPSG', 7773, 'PROJCS[\"WGS 84 / Punjab\",GEOGCS[\"WGS 84\",DATUM["
  "\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP"
  "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Conic Con"
  "formal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.00178226,AUTHORITY[\"EPSG\",\""
  "8821\"]],PARAMETER[\"Longitude of false origin\",75.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st sta"
  "ndard parallel\",30,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32,AUTHORITY[\"EPSG"
  "\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at fal"
  "se origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7773\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7774, 1, 'WGS 84 / Rajasthan', 'EPSG', 7774, 'PROJCS[\"WGS 84 / Rajasthan\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lambert Co"
  "nic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",26.88505546,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",73.875,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",24.2916666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",29.4583333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7774\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7775, 1, 'WGS 84 / Uttar Pradesh', 'EPSG', 7775, 'PROJCS[\"WGS 84 / Uttar Pradesh\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"La"
  "mbert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",27.13270823,AUTHOR"
  "ITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",80.875,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Lati"
  "tude of 1st standard parallel\",24.875,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "29.375,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7775\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7776, 1, 'WGS 84 / Uttaranchal', 'EPSG', 7776, 'PROJCS[\"WGS 84 / Uttaranchal\",GEOGCS[\"WGS 8"
  "4\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTH"
  "ORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Lamber"
  "t Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",30.0017132,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",79.375,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",29,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",31,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northi"
  "ng at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7776\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7777, 1, 'WGS 84 / Andaman and Nicobar', 'EPSG', 7777, 'PROJCS[\"WGS 84 / Andaman and Nicobar\""
  ",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\","
  "\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745"
  "3292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJ"
  "ECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",10.25,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",0.9999428,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7777\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7778, 1, 'WGS 84 / Chhattisgarh', 'EPSG', 7778, 'PROJCS[\"WGS 84 / Chhattisgarh\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",82.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9998332,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7778\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7779, 1, 'WGS 84 / Goa', 'EPSG', 7779, 'PROJCS[\"WGS 84 / Goa\",GEOGCS[\"WGS 84\",DATUM[\"Worl"
  "d Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\""
  "6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",15.375,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER["
  "\"Longitude of natural origin\",74,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999"
  "13,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7779\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7780, 1, 'WGS 84 / Karnataka', 'EPSG', 7780, 'PROJCS[\"WGS 84 / Karnataka\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",15.125,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",76.375,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9998012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7780\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7781, 1, 'WGS 84 / Kerala', 'EPSG', 7781, 'PROJCS[\"WGS 84 / Kerala\",GEOGCS[\"WGS 84\",DATUM["
  "\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP"
  "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",10.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",76,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "999177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7781\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7782, 1, 'WGS 84 / Lakshadweep', 'EPSG', 7782, 'PROJCS[\"WGS 84 / Lakshadweep\",GEOGCS[\"WGS 8"
  "4\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTH"
  "ORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transv"
  "erse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",73.125,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9999536,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7782\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7783, 1, 'WGS 84 / Mizoram', 'EPSG', 7783, 'PROJCS[\"WGS 84 / Mizoram\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",23.125,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",92.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",0.9999821,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7783\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7784, 1, 'WGS 84 / Sikkim', 'EPSG', 7784, 'PROJCS[\"WGS 84 / Sikkim\",GEOGCS[\"WGS 84\",DATUM["
  "\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EP"
  "SG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",27.625,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",88.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9999926,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7784\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7785, 1, 'WGS 84 / Tamil Nadu', 'EPSG', 7785, 'PROJCS[\"WGS 84 / Tamil Nadu\",GEOGCS[\"WGS 84\""
  ",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORI"
  "TY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",10.875,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",78.375,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9997942,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7785\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7786, 1, 'WGS 84 / Tripura', 'EPSG', 7786, 'PROJCS[\"WGS 84 / Tripura\",GEOGCS[\"WGS 84\",DATU"
  "M[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\""
  "EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",23.75,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",91.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.9999822,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7786\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7787, 1, 'WGS 84 / West Bengal', 'EPSG', 7787, 'PROJCS[\"WGS 84 / West Bengal\",GEOGCS[\"WGS 8"
  "4\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTH"
  "ORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transv"
  "erse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.375,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",87.875,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at nat"
  "ural origin\",0.9998584,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"7787\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7791, 1, 'RDN2008 / UTM zone 32N', 'EPSG', 7791, 'PROJCS[\"RDN2008 / UTM zone 32N\",GEOGCS[\"R"
  "DN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7792, 1, 'RDN2008 / UTM zone 33N', 'EPSG', 7792, 'PROJCS[\"RDN2008 / UTM zone 33N\",GEOGCS[\"R"
  "DN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7792\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7793, 1, 'RDN2008 / UTM zone 34N', 'EPSG', 7793, 'PROJCS[\"RDN2008 / UTM zone 34N\",GEOGCS[\"R"
  "DN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7793\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7794, 1, 'RDN2008 / Italy zone (E-N)', 'EPSG', 7794, 'PROJCS[\"RDN2008 / Italy zone (E-N)\",GE"
  "OGCS[\"RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\""
  ",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"Scale factor at natural origin\",0.9985,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",7000000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7794\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7795, 1, 'RDN2008 / Zone 12 (E-N)', 'EPSG', 7795, 'PROJCS[\"RDN2008 / Zone 12 (E-N)\",GEOGCS[\""
  "RDN2008\",DATUM[\"Rete Dinamica Nazionale 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1132\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"6706\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3000000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7795\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7798, 1, 'BGS2005', 'EPSG', 7798, 'GEOGCS[\"BGS2005\",DATUM[\"Bulgaria Geodetic System 2005\","
  "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7798\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7799, 1, 'BGS2005 / UTM zone 34N (N-E)', 'EPSG', 7799, 'PROJCS[\"BGS2005 / UTM zone 34N (N-E)\""
  ",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria Geodetic System 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"7798\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"7799\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7800, 1, 'BGS2005 / UTM zone 35N (N-E)', 'EPSG', 7800, 'PROJCS[\"BGS2005 / UTM zone 35N (N-E)\""
  ",GEOGCS[\"BGS2005\",DATUM[\"Bulgaria Geodetic System 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"E"
  "PSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"7798\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"7800\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7801, 1, 'BGS2005 / CCS2005', 'EPSG', 7801, 'PROJCS[\"BGS2005 / CCS2005\",GEOGCS[\"BGS2005\",D"
  "ATUM[\"Bulgaria Geodetic System 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7798\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "42.6678756833333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",25.5,AUTHORITY[\"EPSG\",\"8822"
  "\"]],PARAMETER[\"Latitude of 1st standard parallel\",42,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd sta"
  "ndard parallel\",43.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4725824.3591,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"7801\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7803, 1, 'BGS2005 / UTM zone 34N', 'EPSG', 7803, 'PROJCS[\"BGS2005 / UTM zone 34N\",GEOGCS[\"B"
  "GS2005\",DATUM[\"Bulgaria Geodetic System 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"7798\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7803\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7804, 1, 'BGS2005 / UTM zone 35N', 'EPSG', 7804, 'PROJCS[\"BGS2005 / UTM zone 35N\",GEOGCS[\"B"
  "GS2005\",DATUM[\"Bulgaria Geodetic System 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"7798\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7804\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7805, 1, 'BGS2005 / UTM zone 36N', 'EPSG', 7805, 'PROJCS[\"BGS2005 / UTM zone 36N\",GEOGCS[\"B"
  "GS2005\",DATUM[\"Bulgaria Geodetic System 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1167\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"7798\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin"
  "\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7805\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7825, 1, 'Pulkovo 1942 / CS63 zone X1', 'EPSG', 7825, 'PROJCS[\"Pulkovo 1942 / CS63 zone X1\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23.5,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",1300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7825\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7826, 1, 'Pulkovo 1942 / CS63 zone X2', 'EPSG', 7826, 'PROJCS[\"Pulkovo 1942 / CS63 zone X2\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",26.5,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",2300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7826\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7827, 1, 'Pulkovo 1942 / CS63 zone X3', 'EPSG', 7827, 'PROJCS[\"Pulkovo 1942 / CS63 zone X3\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",29.5,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",3300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7827\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7828, 1, 'Pulkovo 1942 / CS63 zone X4', 'EPSG', 7828, 'PROJCS[\"Pulkovo 1942 / CS63 zone X4\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",32.51111111"
  "11111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",4300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7828\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7829, 1, 'Pulkovo 1942 / CS63 zone X5', 'EPSG', 7829, 'PROJCS[\"Pulkovo 1942 / CS63 zone X5\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",35.51111111"
  "11111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",5300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7829\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7830, 1, 'Pulkovo 1942 / CS63 zone X6', 'EPSG', 7830, 'PROJCS[\"Pulkovo 1942 / CS63 zone X6\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",38.51111111"
  "11111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",6300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7830\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7831, 1, 'Pulkovo 1942 / CS63 zone X7', 'EPSG', 7831, 'PROJCS[\"Pulkovo 1942 / CS63 zone X7\","
  "GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\""
  "]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0.0833333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",41.51111111"
  "11111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",7300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"7831\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7844, 1, 'GDA2020', 'EPSG', 7844, 'GEOGCS[\"GDA2020\",DATUM[\"Geocentric Datum of Australia 20"
  "20\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPS"
  "G\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7844\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7845, 1, 'GDA2020 / GA LCC', 'EPSG', 7845, 'PROJCS[\"GDA2020 / GA LCC\",GEOGCS[\"GDA2020\",DAT"
  "UM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"78"
  "44\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",134,AUTHORITY[\"EPSG\",\"8822\"]],PARAMET"
  "ER[\"Latitude of 1st standard parallel\",-18,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paral"
  "lel\",-36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7845\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7846, 1, 'GDA2020 / MGA zone 46', 'EPSG', 7846, 'PROJCS[\"GDA2020 / MGA zone 46\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7846\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7847, 1, 'GDA2020 / MGA zone 47', 'EPSG', 7847, 'PROJCS[\"GDA2020 / MGA zone 47\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7847\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7848, 1, 'GDA2020 / MGA zone 48', 'EPSG', 7848, 'PROJCS[\"GDA2020 / MGA zone 48\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7848\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7849, 1, 'GDA2020 / MGA zone 49', 'EPSG', 7849, 'PROJCS[\"GDA2020 / MGA zone 49\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7849\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7850, 1, 'GDA2020 / MGA zone 50', 'EPSG', 7850, 'PROJCS[\"GDA2020 / MGA zone 50\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7850\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7851, 1, 'GDA2020 / MGA zone 51', 'EPSG', 7851, 'PROJCS[\"GDA2020 / MGA zone 51\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7851\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7852, 1, 'GDA2020 / MGA zone 52', 'EPSG', 7852, 'PROJCS[\"GDA2020 / MGA zone 52\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7852\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7853, 1, 'GDA2020 / MGA zone 53', 'EPSG', 7853, 'PROJCS[\"GDA2020 / MGA zone 53\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7853\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7854, 1, 'GDA2020 / MGA zone 54', 'EPSG', 7854, 'PROJCS[\"GDA2020 / MGA zone 54\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7854\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7855, 1, 'GDA2020 / MGA zone 55', 'EPSG', 7855, 'PROJCS[\"GDA2020 / MGA zone 55\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7855\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7856, 1, 'GDA2020 / MGA zone 56', 'EPSG', 7856, 'PROJCS[\"GDA2020 / MGA zone 56\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7856\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7857, 1, 'GDA2020 / MGA zone 57', 'EPSG', 7857, 'PROJCS[\"GDA2020 / MGA zone 57\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7857\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7858, 1, 'GDA2020 / MGA zone 58', 'EPSG', 7858, 'PROJCS[\"GDA2020 / MGA zone 58\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7858\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7859, 1, 'GDA2020 / MGA zone 59', 'EPSG', 7859, 'PROJCS[\"GDA2020 / MGA zone 59\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7859\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7877, 1, 'Astro DOS 71 / SHLG71', 'EPSG', 7877, 'PROJCS[\"Astro DOS 71 / SHLG71\",GEOGCS[\"Ast"
  "ro DOS 71\",DATUM[\"Astro DOS 71\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84["
  "-323.65,551.39,-491.22,0,0,0,0],AUTHORITY[\"EPSG\",\"6710\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4710\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",-15.9666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5.72777777777778,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7877\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7878, 1, 'Astro DOS 71 / UTM zone 30S', 'EPSG', 7878, 'PROJCS[\"Astro DOS 71 / UTM zone 30S\","
  "GEOGCS[\"Astro DOS 71\",DATUM[\"Astro DOS 71\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-323.65,551.39,-491.22,0,0,0,0],AUTHORITY[\"EPSG\",\"6710\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4710\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7878\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7881, 1, 'St. Helena Tritan', 'EPSG', 7881, 'GEOGCS[\"St. Helena Tritan\",DATUM[\"St. Helena T"
  "ritan\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-0.077,0.079,0.086,0,0,0,0],A"
  "UTHORITY[\"EPSG\",\"1173\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7881\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7882, 1, 'St. Helena Tritan / SHLG(Tritan)', 'EPSG', 7882, 'PROJCS[\"St. Helena Tritan / SHLG("
  "Tritan)\",GEOGCS[\"St. Helena Tritan\",DATUM[\"St. Helena Tritan\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[-0.077,0.079,0.086,0,0,0,0],AUTHORITY[\"EPSG\",\"1173\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"7881\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",-15.9666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-5.72777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",299483.737,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2000527"
  ".879,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"7882\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7883, 1, 'St. Helena Tritan / UTM zone 30S', 'EPSG', 7883, 'PROJCS[\"St. Helena Tritan / UTM z"
  "one 30S\",GEOGCS[\"St. Helena Tritan\",DATUM[\"St. Helena Tritan\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[-0.077,0.079,0.086,0,0,0,0],AUTHORITY[\"EPSG\",\"1173\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"7881\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7883\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7886, 1, 'SHGD2015', 'EPSG', 7886, 'GEOGCS[\"SHGD2015\",DATUM[\"St. Helena Geodetic Datum 2015"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"1174\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7886\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7887, 1, 'SHMG2015', 'EPSG', 7887, 'PROJCS[\"SHMG2015\",GEOGCS[\"SHGD2015\",DATUM[\"St. Helena"
  " Geodetic Datum 2015\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,"
  "0],AUTHORITY[\"EPSG\",\"1174\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"7886\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7887\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7899, 1, 'GDA2020 / Vicgrid', 'EPSG', 7899, 'PROJCS[\"GDA2020 / Vicgrid\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",-37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",145,AUTHORITY[\"EPSG\",\"8822\"]],PAR"
  "AMETER[\"Latitude of 1st standard parallel\",-36,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard p"
  "arallel\",-38,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2500000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",2500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7899\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7991, 1, 'NAD27 / MTM zone 10', 'EPSG', 7991, 'PROJCS[\"NAD27 / MTM zone 10\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"7991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (7992, 1, 'Malongo 1987 / UTM zone 33S', 'EPSG', 7992, 'PROJCS[\"Malongo 1987 / UTM zone 33S\","
  "GEOGCS[\"Malongo 1987\",DATUM[\"Malongo 1987\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-254.1,-5.36,-100.29,0,0,0,0],AUTHORITY[\"EPSG\",\"6259\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4259\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"7992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8013, 1, 'GDA2020 / ALB2020', 'EPSG', 8013, 'PROJCS[\"GDA2020 / ALB2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117.883333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.0000044,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8013\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8014, 1, 'GDA2020 / BIO2020', 'EPSG', 8014, 'PROJCS[\"GDA2020 / BIO2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",1.0000022,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8015, 1, 'GDA2020 / BRO2020', 'EPSG', 8015, 'PROJCS[\"GDA2020 / BRO2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",122.333333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.00000298,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8015\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8016, 1, 'GDA2020 / BCG2020', 'EPSG', 8016, 'PROJCS[\"GDA2020 / BCG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.433333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.99999592,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8016\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8017, 1, 'GDA2020 / CARN2020', 'EPSG', 8017, 'PROJCS[\"GDA2020 / CARN2020\",GEOGCS[\"GDA2020\""
  ",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",113.666666666667,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.99999796,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3050000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8017\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8018, 1, 'GDA2020 / CIG2020', 'EPSG', 8018, 'PROJCS[\"GDA2020 / CIG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105.625,AUTHORITY[\"EPSG\",\"8802\"]],PARAMET"
  "ER[\"Scale factor at natural origin\",1.00002514,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1400000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8018\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8019, 1, 'GDA2020 / CKIG2020', 'EPSG', 8019, 'PROJCS[\"GDA2020 / CKIG2020\",GEOGCS[\"GDA2020\""
  ",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",96.875,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.99999387,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1600000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8019\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8020, 1, 'GDA2020 / COL2020', 'EPSG', 8020, 'PROJCS[\"GDA2020 / COL2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.933333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.000019,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",400"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8020\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8021, 1, 'GDA2020 / ESP2020', 'EPSG', 8021, 'PROJCS[\"GDA2020 / ESP2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",121.883333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.0000055,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4050000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8021\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8022, 1, 'GDA2020 / EXM2020', 'EPSG', 8022, 'PROJCS[\"GDA2020 / EXM2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114.066666666667,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.00000236,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2750000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8022\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8023, 1, 'GDA2020 / GCG2020', 'EPSG', 8023, 'PROJCS[\"GDA2020 / GCG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114.594444444444,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.00000628,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3450000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8023\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8024, 1, 'GDA2020 / GOLD2020', 'EPSG', 8024, 'PROJCS[\"GDA2020 / GOLD2020\",GEOGCS[\"GDA2020\""
  ",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",121.511111111111,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1.00004949,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "60000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8024\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8025, 1, 'GDA2020 / JCG2020', 'EPSG', 8025, 'PROJCS[\"GDA2020 / JCG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114.983333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.00000314,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3650000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8025\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8026, 1, 'GDA2020 / KALB2020', 'EPSG', 8026, 'PROJCS[\"GDA2020 / KALB2020\",GEOGCS[\"GDA2020\""
  ",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",114.315277777778,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1.000014,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",55"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8026\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8027, 1, 'GDA2020 / KAR2020', 'EPSG', 8027, 'PROJCS[\"GDA2020 / KAR2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",116.933333333333,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9999989,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2550000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8027\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8028, 1, 'GDA2020 / KUN2020', 'EPSG', 8028, 'PROJCS[\"GDA2020 / KUN2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",128.761111111111,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.0000165,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8029, 1, 'GDA2020 / LCG2020', 'EPSG', 8029, 'PROJCS[\"GDA2020 / LCG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.377777777778,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.00000157,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3750000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8029\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8030, 1, 'GDA2020 / MRCG2020', 'EPSG', 8030, 'PROJCS[\"GDA2020 / MRCG2020\",GEOGCS[\"GDA2020\""
  ",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.177777777778,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1.0000055,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4050000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8030\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8031, 1, 'GDA2020 / PCG2020', 'EPSG', 8031, 'PROJCS[\"GDA2020 / PCG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.827777777778,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.99999906,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3900000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8031\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8032, 1, 'GDA2020 / PHG2020', 'EPSG', 8032, 'PROJCS[\"GDA2020 / PHG2020\",GEOGCS[\"GDA2020\",D"
  "ATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "7844\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",118.611111111111,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.00000135,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8032\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8035, 1, 'WGS 84 / TM Zone 20N (ftUS)', 'EPSG', 8035, 'PROJCS[\"WGS 84 / TM Zone 20N (ftUS)\","
  "GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\""
  "7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJEC"
  "TION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPS"
  "G\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8036, 1, 'WGS 84 / TM Zone 21N (ftUS)', 'EPSG', 8036, 'PROJCS[\"WGS 84 / TM Zone 21N (ftUS)\","
  "GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\""
  "7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532"
  "92519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJEC"
  "TION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPS"
  "G\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor a"
  "t natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.667,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8036\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8042, 1, 'Gusterberg (Ferro)', 'EPSG', 8042, 'GEOGCS[\"Gusterberg (Ferro)\",DATUM[\"Gusterberg"
  " (Ferro)\",SPHEROID[\"Zach 1812\",6376045,310,AUTHORITY[\"EPSG\",\"1026\"]],AUTHORITY[\"EPSG\",\"1188\"]],PRIMEM[\"Fe"
  "rro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8042\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8043, 1, 'St. Stephen (Ferro)', 'EPSG', 8043, 'GEOGCS[\"St. Stephen (Ferro)\",DATUM[\"St. Step"
  "hen (Ferro)\",SPHEROID[\"Zach 1812\",6376045,310,AUTHORITY[\"EPSG\",\"1026\"]],AUTHORITY[\"EPSG\",\"1189\"]],PRIMEM[\""
  "Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8043\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8044, 1, 'Gusterberg Grid (Ferro)', 'EPSG', 8044, 'PROJCS[\"Gusterberg Grid (Ferro)\",GEOGCS[\""
  "Gusterberg (Ferro)\",DATUM[\"Gusterberg (Ferro)\",SPHEROID[\"Zach 1812\",6376045,310,AUTHORITY[\"EPSG\",\"1026\"]],AU"
  "THORITY[\"EPSG\",\"1188\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8042\"]],PRO"
  "JECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",48.0384638888889,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31.8041805555556,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"EPSG\",\"8044\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8045, 1, 'St. Stephen Grid (Ferro)', 'EPSG', 8045, 'PROJCS[\"St. Stephen Grid (Ferro)\",GEOGCS"
  "[\"St. Stephen (Ferro)\",DATUM[\"St. Stephen (Ferro)\",SPHEROID[\"Zach 1812\",6376045,310,AUTHORITY[\"EPSG\",\"1026\""
  "]],AUTHORITY[\"EPSG\",\"1189\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8043\"]"
  "],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",48.2087611111"
  "111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",34.0409222222222,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"EPSG\",\"8045\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8058, 1, 'GDA2020 / NSW Lambert', 'EPSG', 8058, 'PROJCS[\"GDA2020 / NSW Lambert\",GEOGCS[\"GDA"
  "2020\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"7844\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",-33.25,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",147,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",-30.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of "
  "2nd standard parallel\",-35.75,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",9300000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8058\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8059, 1, 'GDA2020 / SA Lambert', 'EPSG', 8059, 'PROJCS[\"GDA2020 / SA Lambert\",GEOGCS[\"GDA20"
  "20\",DATUM[\"Geocentric Datum of Australia 2020\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1168\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"7844\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",-32,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",135,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",-28,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stan"
  "dard parallel\",-36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8059\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8065, 1, 'NAD83(2011) / PCCS zone 1 (ft)', 'EPSG', 8065, 'PROJCS[\"NAD83(2011) / PCCS zone 1 ("
  "ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETE"
  "R[\"Latitude of projection centre\",32.2611111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projectio"
  "n centre\",-111.411111111111,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",45,AUTHORITY[\"EPSG\""
  ",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",45,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor"
  " on initial line\",1.00011,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection centre\",160000,AUTHORITY["
  "\"EPSG\",\"8816\"]],PARAMETER[\"Northing at projection centre\",800000,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"foot\",0."
  "3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8066, 1, 'NAD83(2011) / PCCS zone 2 (ft)', 'EPSG', 8066, 'PROJCS[\"NAD83(2011) / PCCS zone 2 ("
  "ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",31.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.177777"
  "777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00009,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",1800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"8066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8067, 1, 'NAD83(2011) / PCCS zone 3 (ft)', 'EPSG', 8067, 'PROJCS[\"NAD83(2011) / PCCS zone 3 ("
  "ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",31.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.177777777778,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000055,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8067\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8068, 1, 'NAD83(2011) / PCCS zone 4 (ft)', 'EPSG', 8068, 'PROJCS[\"NAD83(2011) / PCCS zone 4 ("
  "ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,2"
  "98.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"La"
  "titude of natural origin\",30.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.75,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9998,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-620000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8068\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8082, 1, 'NAD83(CSRS)v6 / MTM Nova Scotia zone 4', 'EPSG', 8082, 'PROJCS[\"NAD83(CSRS)v6 / MTM"
  " Nova Scotia zone 4\",GEOGCS[\"NAD83(CSRS)v6\",DATUM[\"North American Datum of 1983 (CSRS) version 6\",SPHEROID[\"GRS"
  " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1197\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-61.51111"
  "11111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",24500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "8082\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8083, 1, 'NAD83(CSRS)v6 / MTM Nova Scotia zone 5', 'EPSG', 8083, 'PROJCS[\"NAD83(CSRS)v6 / MTM"
  " Nova Scotia zone 5\",GEOGCS[\"NAD83(CSRS)v6\",DATUM[\"North American Datum of 1983 (CSRS) version 6\",SPHEROID[\"GRS"
  " 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1197\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8252\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-64.51111"
  "11111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "8083\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8086, 1, 'ISN2016', 'EPSG', 8086, 'GEOGCS[\"ISN2016\",DATUM[\"Islands Net 2016\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1187\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8086\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8088, 1, 'ISN2016 / Lambert 2016', 'EPSG', 8088, 'PROJCS[\"ISN2016 / Lambert 2016\",GEOGCS[\"I"
  "SN2016\",DATUM[\"Islands Net 2016\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORI"
  "TY[\"EPSG\",\"1187\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8086\"]],PROJECTION[\"Lambert C"
  "onic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",65,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-19,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",64.25,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",65.75,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",300000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8088\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8090, 1, 'NAD83(HARN) / WISCRS Florence (m)', 'EPSG', 8090, 'PROJCS[\"NAD83(HARN) / WISCRS Flo"
  "rence (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",45.4388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-88.1416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000552095,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",133502.6683,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",0.0063,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8090\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8091, 1, 'NAD83(HARN) / WISCRS Florence (ftUS)', 'EPSG', 8091, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Florence (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",45.4388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-88.1416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000055"
  "2095,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",438000.004,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",0.021,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8091\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8092, 1, 'NAD83(HARN) / WISCRS Eau Claire (m)', 'EPSG', 8092, 'PROJCS[\"NAD83(HARN) / WISCRS E"
  "au Claire (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.8722811263889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-91.2888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1.000035079,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120091.4402,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",91687.9239,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8092\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8093, 1, 'NAD83(HARN) / WISCRS Eau Claire (ftUS)', 'EPSG', 8093, 'PROJCS[\"NAD83(HARN) / WISCR"
  "S Eau Claire (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\""
  "EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.8722811263889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-91.2888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1.000035079,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",394000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",300812.797,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8093\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8095, 1, 'NAD83(HARN) / WISCRS Wood (m)', 'EPSG', 8095, 'PROJCS[\"NAD83(HARN) / WISCRS Wood (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.3625954694444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000421209,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",208483.6173,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "134589.754,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8096, 1, 'NAD83(HARN) / WISCRS Wood (ftUS)', 'EPSG', 8096, 'PROJCS[\"NAD83(HARN) / WISCRS Wood"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",44.3625954694444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000421209,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",684000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",441566.551,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8096\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8097, 1, 'NAD83(HARN) / WISCRS Waushara (m)', 'EPSG', 8097, 'PROJCS[\"NAD83(HARN) / WISCRS Wau"
  "shara (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",44.1139440458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000392096,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120091.4402,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",45069.7587,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8097\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8098, 1, 'NAD83(HARN) / WISCRS Waushara (ftUS)', 'EPSG', 8098, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Waushara (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.1139440458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000392096,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",394000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",147866.367,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8098\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8099, 1, 'NAD83(HARN) / WISCRS Waupaca (m)', 'EPSG', 8099, 'PROJCS[\"NAD83(HARN) / WISCRS Waup"
  "aca (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",43.4202777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-88.8277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000333645,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",185013.9709,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0.007,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8099\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8100, 1, 'NAD83(HARN) / WISCRS Waupaca (ftUS)', 'EPSG', 8100, 'PROJCS[\"NAD83(HARN) / WISCRS W"
  "aupaca (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",43.4202777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-88.8277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000333"
  "645,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",607000.003,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0.023,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8100\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8101, 1, 'NAD83(HARN) / WISCRS Waukesha (m)', 'EPSG', 8101, 'PROJCS[\"NAD83(HARN) / WISCRS Wau"
  "kesha (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",42.5694444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-88.225,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000346179,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",208788.418,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0.0034,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"8101\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8102, 1, 'NAD83(HARN) / WISCRS Waukesha (ftUS)', 'EPSG', 8102, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Waukesha (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",42.5694444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-88.225,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000346179,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",685000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0.011,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8102\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8103, 1, 'NAD83(HARN) / WISCRS Washington (m)', 'EPSG', 8103, 'PROJCS[\"NAD83(HARN) / WISCRS W"
  "ashington (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",42.9180555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-88.0638888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000373"
  "8,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",120091.4415,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.003,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8103\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8104, 1, 'NAD83(HARN) / WISCRS Washington (ftUS)', 'EPSG', 8104, 'PROJCS[\"NAD83(HARN) / WISCR"
  "S Washington (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\","
  "6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",42.9180555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-88.0638888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "0003738,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",394000.004,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\","
  "\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8104\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8105, 1, 'NAD83(HARN) / WISCRS Washburn (m)', 'EPSG', 8105, 'PROJCS[\"NAD83(HARN) / WISCRS Was"
  "hburn (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",45.9612198333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-91.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000475376,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",234086.8682,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",188358.6058,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8105\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8106, 1, 'NAD83(HARN) / WISCRS Washburn (ftUS)', 'EPSG', 8106, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Washburn (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.9612198333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-91.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000475376,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",768000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",617973.193,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8106\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8107, 1, 'NAD83(HARN) / WISCRS Walworth (m)', 'EPSG', 8107, 'PROJCS[\"NAD83(HARN) / WISCRS Wal"
  "worth (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",42.6694620969444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-88.5416666666666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000367192,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",232562.8651,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",111088.2224,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8107\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8108, 1, 'NAD83(HARN) / WISCRS Walworth (ftUS)', 'EPSG', 8108, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Walworth (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.6694620969444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-88.5416666666666,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000367192,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",763000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",364461.943,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8108\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8109, 1, 'NAD83(HARN) / WISCRS Vilas (m)', 'EPSG', 8109, 'PROJCS[\"NAD83(HARN) / WISCRS Vilas "
  "(m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",46.0778440905556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-89.4888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000730"
  "142,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",134417.0689,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",50337.1092,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8109\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8110, 1, 'NAD83(HARN) / WISCRS Vilas (ftUS)', 'EPSG', 8110, 'PROJCS[\"NAD83(HARN) / WISCRS Vil"
  "as (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",46.0778440905556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-89.4888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000730142,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",441000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",165147.666,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8110\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8111, 1, 'NAD83(HARN) / WISCRS Vernon (m)', 'EPSG', 8111, 'PROJCS[\"NAD83(HARN) / WISCRS Verno"
  "n (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",43.5750329397222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-90.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000040"
  "8158,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",222504.4451,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",47532.0602,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8111\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8112, 1, 'NAD83(HARN) / WISCRS Vernon (ftUS)', 'EPSG', 8112, 'PROJCS[\"NAD83(HARN) / WISCRS Ve"
  "rnon (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",43.5750329397222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-90.7944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "0000408158,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",730000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",155944.768,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8112\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8113, 1, 'NAD83(HARN) / WISCRS Trempealeau (m)', 'EPSG', 8113, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Trempealeau (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",43.1611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-91.3777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036"
  "1538,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",256946.9138,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0.0041,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8113\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8114, 1, 'NAD83(HARN) / WISCRS Trempealeau (ftUS)', 'EPSG', 8114, 'PROJCS[\"NAD83(HARN) / WISC"
  "RS Trempealeau (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",43.1611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-91.3777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "0000361538,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",843000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0.013,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8114\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8115, 1, 'NAD83(HARN) / WISCRS Taylor (m)', 'EPSG', 8115, 'PROJCS[\"NAD83(HARN) / WISCRS Taylo"
  "r (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",45.1778220858333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-90.4833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000059"
  "7566,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",187147.5744,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",107746.7522,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8115\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8116, 1, 'NAD83(HARN) / WISCRS Taylor (ftUS)', 'EPSG', 8116, 'PROJCS[\"NAD83(HARN) / WISCRS Ta"
  "ylor (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",45.1778220858333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-90.4833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "0000597566,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",614000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",353499.136,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8116\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8117, 1, 'NAD83(HARN) / WISCRS St. Croix (m)', 'EPSG', 8117, 'PROJCS[\"NAD83(HARN) / WISCRS St"
  ". Croix (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-92.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038180"
  "3,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",165506.7302,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.0103,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8117\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8118, 1, 'NAD83(HARN) / WISCRS St. Croix (ftUS)', 'EPSG', 8118, 'PROJCS[\"NAD83(HARN) / WISCRS"
  " St. Croix (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-92.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "381803,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",542999.997,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0.034,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8118\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8119, 1, 'NAD83(HARN) / WISCRS Shawano (m)', 'EPSG', 8119, 'PROJCS[\"NAD83(HARN) / WISCRS Shaw"
  "ano (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-88.6055555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000032144,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262433.3253,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0.0096,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8120, 1, 'NAD83(HARN) / WISCRS Shawano (ftUS)', 'EPSG', 8120, 'PROJCS[\"NAD83(HARN) / WISCRS S"
  "hawano (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",44.0361111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-88.6055555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000321"
  "44,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",861000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.031,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"90"
  "03\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8121, 1, 'NAD83(HARN) / WISCRS Sawyer (m)', 'EPSG', 8121, 'PROJCS[\"NAD83(HARN) / WISCRS Sawye"
  "r (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",45.9000991313889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-91.1277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000057"
  "3461,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",216713.2336,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",120734.1631,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8122, 1, 'NAD83(HARN) / WISCRS Sawyer (ftUS)', 'EPSG', 8122, 'PROJCS[\"NAD83(HARN) / WISCRS Sa"
  "wyer (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",45.9000991313889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-91.1277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "0000573461,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",711000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",396108.667,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8123, 1, 'NAD83(HARN) / WISCRS Sauk (m)', 'EPSG', 8123, 'PROJCS[\"NAD83(HARN) / WISCRS Sauk (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",42.8194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-89.9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000373868,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",185623.5716,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0051,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"8123\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8124, 1, 'NAD83(HARN) / WISCRS Sauk (ftUS)', 'EPSG', 8124, 'PROJCS[\"NAD83(HARN) / WISCRS Sauk"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",42.8194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-89.9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000373868,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",609000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0"
  "17,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8125, 1, 'NAD83(HARN) / WISCRS Rusk (m)', 'EPSG', 8125, 'PROJCS[\"NAD83(HARN) / WISCRS Rusk (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",43.9194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-91.0666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495976,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250546.1013,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0.0234,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8125\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8126, 1, 'NAD83(HARN) / WISCRS Rusk (ftUS)', 'EPSG', 8126, 'PROJCS[\"NAD83(HARN) / WISCRS Rusk"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",43.9194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-91.0666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495976,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",822000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0.077,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8126\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8127, 1, 'NAD83(HARN) / WISCRS Rock (m)', 'EPSG', 8127, 'PROJCS[\"NAD83(HARN) / WISCRS Rock (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",41.9444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-89.0722222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000337311,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",146304.2926,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0.0068,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8128, 1, 'NAD83(HARN) / WISCRS Rock (ftUS)', 'EPSG', 8128, 'PROJCS[\"NAD83(HARN) / WISCRS Rock"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",41.9444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-89.0722222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000337311,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",480000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0.022,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8129, 1, 'NAD83(HARN) / WISCRS Richland (m)', 'EPSG', 8129, 'PROJCS[\"NAD83(HARN) / WISCRS Ric"
  "hland (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",43.3223129275,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-90.4305555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "375653,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",202387.6048,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",134255.4253,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8129\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8130, 1, 'NAD83(HARN) / WISCRS Richland (ftUS)', 'EPSG', 8130, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Richland (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.3223129275,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-90.4305555555555,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\","
  "1.0000375653,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",664000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",440469.675,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8130\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8131, 1, 'NAD83(HARN) / WISCRS Price (m)', 'EPSG', 8131, 'PROJCS[\"NAD83(HARN) / WISCRS Price "
  "(m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",44.5555555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-90.4888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000649554,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",227990.8546,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0.0109,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"8131\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8132, 1, 'NAD83(HARN) / WISCRS Price (ftUS)', 'EPSG', 8132, 'PROJCS[\"NAD83(HARN) / WISCRS Pri"
  "ce (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",44.5555555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-90.4888888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000649554,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",747999.995,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0.036,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8132\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8133, 1, 'NAD83(HARN) / WISCRS Portage (m)', 'EPSG', 8133, 'PROJCS[\"NAD83(HARN) / WISCRS Port"
  "age (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",44.4168239752778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-89.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000"
  "039936,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",56388.1128,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",50022.1874,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8133\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8134, 1, 'NAD83(HARN) / WISCRS Portage (ftUS)', 'EPSG', 8134, 'PROJCS[\"NAD83(HARN) / WISCRS P"
  "ortage (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.4168239752778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-89.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1.000039936,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",185000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",164114.46,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"E"
  "PSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8134\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8135, 1, 'NAD83(HARN) / WISCRS Polk (m)', 'EPSG', 8135, 'PROJCS[\"NAD83(HARN) / WISCRS Polk (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",44.6611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-92.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000433849,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",141732.2823,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0.0059,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8135\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8136, 1, 'NAD83(HARN) / WISCRS Polk (ftUS)', 'EPSG', 8136, 'PROJCS[\"NAD83(HARN) / WISCRS Polk"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",44.6611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-92.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000433849,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",464999.996,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0.019,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8137, 1, 'NAD83(HARN) / WISCRS Pepin and Pierce (m)', 'EPSG', 8137, 'PROJCS[\"NAD83(HARN) / WI"
  "SCRS Pepin and Pierce (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6361488719444,AUTHORITY[\"EPSG\",\"8801\"]],PARAM"
  "ETER[\"Longitude of natural origin\",-92.2277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natu"
  "ral origin\",1.0000362977,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",167640.3354,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",86033.0876,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8137\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8138, 1, 'NAD83(HARN) / WISCRS Pepin and Pierce (ftUS)', 'EPSG', 8138, 'PROJCS[\"NAD83(HARN) /"
  " WISCRS Pepin and Pierce (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",A"
  "UTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.6361488719444,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-92.2277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1.0000362977,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",550000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",282260.222,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.304800609601"
  "21924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8138\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8139, 1, 'NAD83(HARN) / WISCRS Oneida (m)', 'EPSG', 8139, 'PROJCS[\"NAD83(HARN) / WISCRS Oneid"
  "a (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",45.7042237702778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-89.5444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000068"
  "6968,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",70104.1401,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",57588.0346,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EA"
  "ST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8139\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8140, 1, 'NAD83(HARN) / WISCRS Oneida (ftUS)', 'EPSG', 8140, 'PROJCS[\"NAD83(HARN) / WISCRS On"
  "eida (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",45.7042237702778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-89.5444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "0000686968,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",230000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",188936.744,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8140\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8141, 1, 'NAD83(HARN) / WISCRS Oconto (m)', 'EPSG', 8141, 'PROJCS[\"NAD83(HARN) / WISCRS Ocont"
  "o (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",44.3972222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-87.9083333333334,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000236869,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",182880.3676,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0.0033,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"8141\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8142, 1, 'NAD83(HARN) / WISCRS Oconto (ftUS)', 'EPSG', 8142, 'PROJCS[\"NAD83(HARN) / WISCRS Oc"
  "onto (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.3972222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-87.9083333333334,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000023686"
  "9,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",600000.006,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0.011,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8142\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8143, 1, 'NAD83(HARN) / WISCRS Monroe (m)', 'EPSG', 8143, 'PROJCS[\"NAD83(HARN) / WISCRS Monro"
  "e (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",44.0000739286111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-90.6416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000043"
  "4122,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",204521.209,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",121923.9861,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8143\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8144, 1, 'NAD83(HARN) / WISCRS Monroe (ftUS)', 'EPSG', 8144, 'PROJCS[\"NAD83(HARN) / WISCRS Mo"
  "nroe (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",44.0000739286111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-90.6416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1."
  "0000434122,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",671000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",400012.278,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8144\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8145, 1, 'NAD83(HARN) / WISCRS Menominee (m)', 'EPSG', 8145, 'PROJCS[\"NAD83(HARN) / WISCRS Me"
  "nominee (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.7277777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-88.4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036249"
  "9,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",105461.0121,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.0029,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8145\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8146, 1, 'NAD83(HARN) / WISCRS Menominee (ftUS)', 'EPSG', 8146, 'PROJCS[\"NAD83(HARN) / WISCRS"
  " Menominee (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",44.7277777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-88.4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "362499,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",346000.004,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8146\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8147, 1, 'NAD83(HARN) / WISCRS Marinette (m)', 'EPSG', 8147, 'PROJCS[\"NAD83(HARN) / WISCRS Ma"
  "rinette (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.6916666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-87.7111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000023498"
  "2,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",238658.8794,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.0032,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8147\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8148, 1, 'NAD83(HARN) / WISCRS Marinette (ftUS)', 'EPSG', 8148, 'PROJCS[\"NAD83(HARN) / WISCRS"
  " Marinette (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",44.6916666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-87.7111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "234982,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",783000.007,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8148\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8149, 1, 'NAD83(HARN) / WISCRS Marathon (m)', 'EPSG', 8149, 'PROJCS[\"NAD83(HARN) / WISCRS Mar"
  "athon (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",44.9009044236111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-89.77,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000053289,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",74676.1493,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",55049.2669,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8149\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8150, 1, 'NAD83(HARN) / WISCRS Marathon (ftUS)', 'EPSG', 8150, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Marathon (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.9009044236111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-89.77,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000053"
  "289,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",245000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",180607.47,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8150\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8151, 1, 'NAD83(HARN) / WISCRS Lincoln (m)', 'EPSG', 8151, 'PROJCS[\"NAD83(HARN) / WISCRS Linc"
  "oln (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",44.8444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-89.7444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000599003,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",116129.0323,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0.0058,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8151\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8152, 1, 'NAD83(HARN) / WISCRS Lincoln (ftUS)', 'EPSG', 8152, 'PROJCS[\"NAD83(HARN) / WISCRS L"
  "incoln (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",44.8444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-89.7444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000599"
  "003,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",381000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0.019,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8152\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8153, 1, 'NAD83(HARN) / WISCRS Langlade (m)', 'EPSG', 8153, 'PROJCS[\"NAD83(HARN) / WISCRS Lan"
  "glade (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",45.1542371052778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-89.0444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000627024,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",198425.197,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",105279.7829,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8153\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8154, 1, 'NAD83(HARN) / WISCRS Langlade (ftUS)', 'EPSG', 8154, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Langlade (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.1542371052778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-89.0444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000627024,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",651000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",345405.421,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8154\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8155, 1, 'NAD83(HARN) / WISCRS La Crosse (m)', 'EPSG', 8155, 'PROJCS[\"NAD83(HARN) / WISCRS La"
  " Crosse (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",43.4511111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-91.3277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031998"
  "5,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",130454.6598,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.0033,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8155\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8156, 1, 'NAD83(HARN) / WISCRS La Crosse (ftUS)', 'EPSG', 8156, 'PROJCS[\"NAD83(HARN) / WISCRS"
  " La Crosse (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",43.4511111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of"
  " natural origin\",-91.3277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000"
  "319985,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",427999.996,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0.011,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8156\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8157, 1, 'NAD83(HARN) / WISCRS Kewaunee, Manitowoc and Sheboygan (m)', 'EPSG', 8157, 'PROJCS[\""
  "NAD83(HARN) / WISCRS Kewaunee, Manitowoc and Sheboygan (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Refe"
  "rence Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Trans"
  "verse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.2777777777778,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",1.0000233704,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",79857.7614"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0012,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8157\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8158, 1, 'NAD83(HARN) / WISCRS Kewaunee, Manitowoc and Sheboygan (ftUS)', 'EPSG', 8158, 'PROJC"
  "S[\"NAD83(HARN) / WISCRS Kewaunee, Manitowoc and Sheboygan (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accura"
  "cy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0"
  ",0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519"
  "943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION["
  "\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.2777777777778,AUTHO"
  "RITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",1.0000233704,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",262"
  "000.006,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.004,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US surv"
  "ey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8"
  "158\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8159, 1, 'NAD83(HARN) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (m)', 'EPSG', 8159, 'PRO"
  "JCS[\"NAD83(HARN) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Ac"
  "curacy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0"
  ",0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329"
  "2519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECT"
  "ION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.2166666666667,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.8944444444445,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",1.0000260649,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",185928.3728,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0009,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8159\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8160, 1, 'NAD83(HARN) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (ftUS)', 'EPSG', 8160, '"
  "PROJCS[\"NAD83(HARN) / WISCRS Kenosha, Milwaukee, Ozaukee and Racine (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (H"
  "igh Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0"
  ",0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.216666666"
  "6667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.8944444444445,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000260649,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",610000.003,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.003,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"8160\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8161, 1, 'NAD83(HARN) / WISCRS Jackson (m)', 'EPSG', 8161, 'PROJCS[\"NAD83(HARN) / WISCRS Jack"
  "son (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",44.2533351277778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-90.8442965194444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000353,AUTHO"
  "RITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",27000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",25000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"8161\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8162, 1, 'NAD83(HARN) / WISCRS Jackson (ftUS)', 'EPSG', 8162, 'PROJCS[\"NAD83(HARN) / WISCRS J"
  "ackson (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",44.2533351277778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-90.8442965194444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000353"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",88582.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",82020.833,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8162\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8163, 1, 'NAD83(HARN) / WISCRS Iron (m)', 'EPSG', 8163, 'PROJCS[\"NAD83(HARN) / WISCRS Iron (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",45.4444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-90.2555555555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000677153,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",220980.4419,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0.0085,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8163\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8164, 1, 'NAD83(HARN) / WISCRS Iron (ftUS)', 'EPSG', 8164, 'PROJCS[\"NAD83(HARN) / WISCRS Iron"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",45.4444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-90.2555555555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000677153,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",725000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0.028,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8164\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8165, 1, 'NAD83(HARN) / WISCRS Iowa (m)', 'EPSG', 8165, 'PROJCS[\"NAD83(HARN) / WISCRS Iowa (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",42.5388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-90.1611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000394961,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",113081.0261,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0.0045,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8165\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8166, 1, 'NAD83(HARN) / WISCRS Iowa (ftUS)', 'EPSG', 8166, 'PROJCS[\"NAD83(HARN) / WISCRS Iowa"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",42.5388888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-90.1611111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000394961,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",371000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0.015,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8166\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8167, 1, 'NAD83(HARN) / WISCRS Green Lake and Marquette (m)', 'EPSG', 8167, 'PROJCS[\"NAD83(HA"
  "RN) / WISCRS Green Lake and Marquette (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\","
  "SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal"
  " (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.8070001177778,AUTHORITY[\"EPSG\",\""
  "8801\"]],PARAMETER[\"Longitude of natural origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",1.0000344057,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150876.3018,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",79170.7795,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8167\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8168, 1, 'NAD83(HARN) / WISCRS Green Lake and Marquette (ftUS)', 'EPSG', 8168, 'PROJCS[\"NAD83"
  "(HARN) / WISCRS Green Lake and Marquette (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Netwo"
  "rk)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EP"
  "SG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Con"
  "formal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.8070001177778,AUTHORITY[\"EP"
  "SG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-89.2416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",1.0000344057,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",495000,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",259746.132,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\""
  ",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8168\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8169, 1, 'NAD83(HARN) / WISCRS Green and Lafayette (m)', 'EPSG', 8169, 'PROJCS[\"NAD83(HARN) /"
  " WISCRS Green and Lafayette (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",A"
  "UTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.6375622769444,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-89.8388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",1.0000390487,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",170078.7403,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",45830.2947,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8169\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8170, 1, 'NAD83(HARN) / WISCRS Green and Lafayette (ftUS)', 'EPSG', 8170, 'PROJCS[\"NAD83(HARN"
  ") / WISCRS Green and Lafayette (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"615"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1"
  "SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.6375622769444,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-89.8388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",1.0000390487,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",558000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",150361.559,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480"
  "060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8170\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8171, 1, 'NAD83(HARN) / WISCRS Grant (m)', 'EPSG', 8171, 'PROJCS[\"NAD83(HARN) / WISCRS Grant "
  "(m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",41.4111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-90.8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000349452,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",242316.4841,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.01,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"8171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8172, 1, 'NAD83(HARN) / WISCRS Grant (ftUS)', 'EPSG', 8172, 'PROJCS[\"NAD83(HARN) / WISCRS Gra"
  "nt (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",41.4111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-90.8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000349452,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",794999.998,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ".033,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8172\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8173, 1, 'NAD83(HARN) / WISCRS Forest (m)', 'EPSG', 8173, 'PROJCS[\"NAD83(HARN) / WISCRS Fores"
  "t (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",44.0055555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-88.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000673004,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",275844.5533,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0.0157,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"8173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8177, 1, 'NAD83(HARN) / WISCRS Forest (ftUS)', 'EPSG', 8177, 'PROJCS[\"NAD83(HARN) / WISCRS Fo"
  "rest (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",44.0055555555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-88.6444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000067300"
  "4,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",905000.005,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",0.052,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8177\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8179, 1, 'NAD83(HARN) / WISCRS Dunn (m)', 'EPSG', 8179, 'PROJCS[\"NAD83(HARN) / WISCRS Dunn (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",44.4083333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-91.8944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000410324,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",51816.104,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0.003,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"8179\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8180, 1, 'NAD83(HARN) / WISCRS Dunn (ftUS)', 'EPSG', 8180, 'PROJCS[\"NAD83(HARN) / WISCRS Dunn"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",44.4083333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-91.8944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000410324,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",170000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8180\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8181, 1, 'NAD83(HARN) / WISCRS Douglas (m)', 'EPSG', 8181, 'PROJCS[\"NAD83(HARN) / WISCRS Doug"
  "las (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",45.8833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-91.9277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000385418,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",59131.3183,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0.0041,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8181\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8182, 1, 'NAD83(HARN) / WISCRS Douglas (ftUS)', 'EPSG', 8182, 'PROJCS[\"NAD83(HARN) / WISCRS D"
  "ouglas (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",45.8833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-91.9277777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000385"
  "418,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",194000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0.013,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8182\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8184, 1, 'NAD83(HARN) / WISCRS Door (m)', 'EPSG', 8184, 'PROJCS[\"NAD83(HARN) / WISCRS Door (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",44.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.27222"
  "22222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000187521,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",158801.1176,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0023,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"8184\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8185, 1, 'NAD83(HARN) / WISCRS Door (ftUS)', 'EPSG', 8185, 'PROJCS[\"NAD83(HARN) / WISCRS Door"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",44.4,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87"
  ".2722222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000187521,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",521000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.008,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8185\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8187, 1, 'NAD83(HARN) / WISCRS Dodge and Jefferson (m)', 'EPSG', 8187, 'PROJCS[\"NAD83(HARN) /"
  " WISCRS Dodge and Jefferson (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.4722222222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-88.775,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00"
  "00346418,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",263347.7263,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0.0076,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8187\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8189, 1, 'NAD83(HARN) / WISCRS Dodge and Jefferson (ftUS)', 'EPSG', 8189, 'PROJCS[\"NAD83(HARN"
  ") / WISCRS Dodge and Jefferson (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHE"
  "ROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"615"
  "2\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.4722222222222,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-88.775,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",1.0000346418,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",863999.999,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0.025,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8189\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8191, 1, 'NAD83(HARN) / WISCRS Dane (m)', 'EPSG', 8191, 'PROJCS[\"NAD83(HARN) / WISCRS Dane (m"
  ")\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222"
  "101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],"
  "PARAMETER[\"Latitude of natural origin\",43.0695160375,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-89.4222222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000384786,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",247193.2944,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",146591.9896,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8193, 1, 'NAD83(HARN) / WISCRS Dane (ftUS)', 'EPSG', 8193, 'PROJCS[\"NAD83(HARN) / WISCRS Dane"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",43.0695160375,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-89.4222222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000038"
  "4786,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",811000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",480943.886,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8196, 1, 'NAD83(HARN) / WISCRS Crawford (m)', 'EPSG', 8196, 'PROJCS[\"NAD83(HARN) / WISCRS Cra"
  "wford (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",43.200055605,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-90.9388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00003"
  "49151,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",113690.6274,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",53703.1201,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8196\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8197, 1, 'NAD83(HARN) / WISCRS Crawford (ftUS)', 'EPSG', 8197, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Crawford (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.200055605,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-90.9388888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1"
  ".0000349151,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",373000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",176190.987,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"E"
  "PSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8197\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8198, 1, 'NAD83(HARN) / WISCRS Columbia (m)', 'EPSG', 8198, 'PROJCS[\"NAD83(HARN) / WISCRS Col"
  "umbia (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",43.4625466458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-89.3944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "0003498,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",169164.3381,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",111569.6134,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8198\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8200, 1, 'NAD83(HARN) / WISCRS Columbia (ftUS)', 'EPSG', 8200, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Columbia (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.4625466458333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-89.3944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.00003498,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",554999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",366041.307,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8200\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8201, 1, 'NAD83(HARN) / WISCRS Clark (m)', 'EPSG', 8201, 'PROJCS[\"NAD83(HARN) / WISCRS Clark "
  "(m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",43.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-90.7083333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000463003,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",199949.1989,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",0.0086,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"8201\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8202, 1, 'NAD83(HARN) / WISCRS Clark (ftUS)', 'EPSG', 8202, 'PROJCS[\"NAD83(HARN) / WISCRS Cla"
  "rk (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43.6111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-90.7083333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000463003,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",655999.997,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0.028,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8202\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8203, 1, 'NAD83(HARN) / WISCRS Chippewa (m)', 'EPSG', 8203, 'PROJCS[\"NAD83(HARN) / WISCRS Chi"
  "ppewa (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",44.9778568986111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-91.2944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000391127,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60045.72,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",44091.4346,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8203\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8204, 1, 'NAD83(HARN) / WISCRS Chippewa (ftUS)', 'EPSG', 8204, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Chippewa (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44.9778568986111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-91.2944444444445,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000391127,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",197000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",144656.648,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8204\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8205, 1, 'NAD83(HARN) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (m)', 'EPSG', 820"
  "5, 'PROJCS[\"NAD83(HARN) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\""
  "NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],T"
  "OWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"41"
  "52\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.7"
  "194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000286569,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",244754.8893,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0049,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8205\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8206, 1, 'NAD83(HARN) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (ftUS)', 'EPSG', "
  "8206, 'PROJCS[\"NAD83(HARN) / WISCRS Calumet, Fond du Lac, Outagamie and Winnebago (ftUS)\",GEOGCS[\"NAD83(HARN)\",DA"
  "TUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",42.7194444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.5111111111111,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000286569,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",802999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.016,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"8206\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8207, 1, 'NAD83(HARN) / WISCRS Burnett (m)', 'EPSG', 8207, 'PROJCS[\"NAD83(HARN) / WISCRS Burn"
  "ett (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",45.8987148658333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-92.4577777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000"
  "0383841,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",64008.1276,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",59445.9043,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8207\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8208, 1, 'NAD83(HARN) / WISCRS Burnett (ftUS)', 'EPSG', 8208, 'PROJCS[\"NAD83(HARN) / WISCRS B"
  "urnett (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.8987148658333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-92.4577777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1.0000383841,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",209999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",195032.104,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8208\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8209, 1, 'NAD83(HARN) / WISCRS Buffalo (m)', 'EPSG', 8209, 'PROJCS[\"NAD83(HARN) / WISCRS Buff"
  "alo (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",43.4813888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-91.7972222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000382778,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",175260.3502,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0.0048,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8209\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8210, 1, 'NAD83(HARN) / WISCRS Buffalo (ftUS)', 'EPSG', 8210, 'PROJCS[\"NAD83(HARN) / WISCRS B"
  "uffalo (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",43.4813888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-91.7972222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000382"
  "778,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",574999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0.016,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8210\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8212, 1, 'NAD83(HARN) / WISCRS Brown (m)', 'EPSG', 8212, 'PROJCS[\"NAD83(HARN) / WISCRS Brown "
  "(m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",43,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",31600,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4600,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8212\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8213, 1, 'NAD83(HARN) / WISCRS Brown (ftUS)', 'EPSG', 8213, 'PROJCS[\"NAD83(HARN) / WISCRS Bro"
  "wn (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",43,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",103674.333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",15091.833,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"8213\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8214, 1, 'NAD83(HARN) / WISCRS Bayfield (m)', 'EPSG', 8214, 'PROJCS[\"NAD83(HARN) / WISCRS Bay"
  "field (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,29"
  "8.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",46.6696483772222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-91.1527777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0"
  "000331195,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",228600.4575,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",148551.4837,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8214\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8216, 1, 'NAD83(HARN) / WISCRS Bayfield (ftUS)', 'EPSG', 8216, 'PROJCS[\"NAD83(HARN) / WISCRS "
  "Bayfield (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378"
  "137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",46.6696483772222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-91.1527777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.0000331195,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",750000.001,AUTHORITY[\"EPSG\",\"8806\"]],PA"
  "RAMETER[\"False northing\",487372.659,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8216\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8218, 1, 'NAD83(HARN) / WISCRS Barron (m)', 'EPSG', 8218, 'PROJCS[\"NAD83(HARN) / WISCRS Barro"
  "n (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",45.1444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-91.85,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000486665,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",93150,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0.0029,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"8218\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8220, 1, 'NAD83(HARN) / WISCRS Barron (ftUS)', 'EPSG', 8220, 'PROJCS[\"NAD83(HARN) / WISCRS Ba"
  "rron (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,"
  "298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",45.1444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-91.85,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000486665,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",305609.625,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0.01,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8222, 1, 'NAD83(HARN) / WISCRS Ashland (m)', 'EPSG', 8222, 'PROJCS[\"NAD83(HARN) / WISCRS Ashl"
  "and (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",45.7061111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-90.6222222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495683,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",172821.9461,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0.0017,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8222\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8224, 1, 'NAD83(HARN) / WISCRS Ashland (ftUS)', 'EPSG', 8224, 'PROJCS[\"NAD83(HARN) / WISCRS A"
  "shland (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",45.7061111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-90.6222222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0000495"
  "683,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",567000.001,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0.006,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9"
  "003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8224\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8225, 1, 'NAD83(HARN) / WISCRS Adams and Juneau (m)', 'EPSG', 8225, 'PROJCS[\"NAD83(HARN) / WI"
  "SCRS Adams and Juneau (m)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG"
  "\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.3777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036528"
  "5,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",147218.6942,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fal"
  "se northing\",0.0037,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8225\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8226, 1, 'NAD83(HARN) / WISCRS Adams and Juneau (ftUS)', 'EPSG', 8226, 'PROJCS[\"NAD83(HARN) /"
  " WISCRS Adams and Juneau (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.3777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000036"
  "5285,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",482999.999,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",0.012,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8226\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8232, 1, 'NAD83(CSRS96)', 'EPSG', 8232, 'GEOGCS[\"NAD83(CSRS96)\",DATUM[\"North American Datum"
  " of 1983 (CSRS96)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"11"
  "92\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8232\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8237, 1, 'NAD83(CSRS)v2', 'EPSG', 8237, 'GEOGCS[\"NAD83(CSRS)v2\",DATUM[\"North American Datum"
  " of 1983 (CSRS) version 2\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1193\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8237\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8240, 1, 'NAD83(CSRS)v3', 'EPSG', 8240, 'GEOGCS[\"NAD83(CSRS)v3\",DATUM[\"North American Datum"
  " of 1983 (CSRS) version 3\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1194\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8240\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8246, 1, 'NAD83(CSRS)v4', 'EPSG', 8246, 'GEOGCS[\"NAD83(CSRS)v4\",DATUM[\"North American Datum"
  " of 1983 (CSRS) version 4\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1195\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8249, 1, 'NAD83(CSRS)v5', 'EPSG', 8249, 'GEOGCS[\"NAD83(CSRS)v5\",DATUM[\"North American Datum"
  " of 1983 (CSRS) version 5\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1196\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8252, 1, 'NAD83(CSRS)v6', 'EPSG', 8252, 'GEOGCS[\"NAD83(CSRS)v6\",DATUM[\"North American Datum"
  " of 1983 (CSRS) version 6\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1197\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8255, 1, 'NAD83(CSRS)v7', 'EPSG', 8255, 'GEOGCS[\"NAD83(CSRS)v7\",DATUM[\"North American Datum"
  " of 1983 (CSRS) version 7\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPS"
  "G\",\"1198\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8311, 1, 'NAD83(2011) / Oregon Burns-Harper zone (m)', 'EPSG', 8311, 'PROJCS[\"NAD83(2011) / O"
  "regon Burns-Harper zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-117.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00014,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",90000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8312, 1, 'NAD83(2011) / Oregon Burns-Harper zone (ft)', 'EPSG', 8312, 'PROJCS[\"NAD83(2011) / "
  "Oregon Burns-Harper zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-117.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00014,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",295275.5906,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8313, 1, 'NAD83(2011) / Oregon Canyon City-Burns zone (m)', 'EPSG', 8313, 'PROJCS[\"NAD83(2011"
  ") / Oregon Canyon City-Burns zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00022,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"8313\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8314, 1, 'NAD83(2011) / Oregon Canyon City-Burns zone (ft)', 'EPSG', 8314, 'PROJCS[\"NAD83(201"
  "1) / Oregon Canyon City-Burns zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 201"
  "1)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",-119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00022,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",65616.7979,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"8314\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8315, 1, 'NAD83(2011) / Oregon Coast Range North zone (m)', 'EPSG', 8315, 'PROJCS[\"NAD83(2011"
  ") / Oregon Coast Range North zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY["
  "\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.5833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-123.416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1.000045,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",20000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8315\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8316, 1, 'NAD83(2011) / Oregon Coast Range North zone (ft)', 'EPSG', 8316, 'PROJCS[\"NAD83(201"
  "1) / Oregon Coast Range North zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 201"
  "1)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY["
  "\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",45.5833333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-123.416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1.000045,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",98425.1969,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",65616.7979,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8317, 1, 'NAD83(2011) / Oregon Dayville-Prairie City zone (m)', 'EPSG', 8317, 'PROJCS[\"NAD83("
  "2011) / Oregon Dayville-Prairie City zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-119.644444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8317\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8318, 1, 'NAD83(2011) / Oregon Dayville-Prairie City zone (ft)', 'EPSG', 8318, 'PROJCS[\"NAD83"
  "(2011) / Oregon Dayville-Prairie City zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Sy"
  "stem 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],"
  "PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY["
  "\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",44.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-119.644444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",1.00012,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",65616.7979,AUTHORITY[\"EPSG\",\"8806\"]],P"
  "ARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8319, 1, 'NAD83(2011) / Oregon Denio-Burns zone (m)', 'EPSG', 8319, 'PROJCS[\"NAD83(2011) / Or"
  "egon Denio-Burns zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID["
  "\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-118.416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00019,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",80000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"8319\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8320, 1, 'NAD83(2011) / Oregon Denio-Burns zone (ft)', 'EPSG', 8320, 'PROJCS[\"NAD83(2011) / O"
  "regon Denio-Burns zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROI"
  "D[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-118.416666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00019,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",262467.1916,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"8320\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8321, 1, 'NAD83(2011) / Oregon Halfway zone (m)', 'EPSG', 8321, 'PROJCS[\"NAD83(2011) / Oregon"
  " Halfway zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],"
  "PARAMETER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-117.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000085,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",70000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"8321\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8322, 1, 'NAD83(2011) / Oregon Halfway zone (ft)', 'EPSG', 8322, 'PROJCS[\"NAD83(2011) / Orego"
  "n Halfway zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-117.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000085,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",229658.7927,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8323, 1, 'NAD83(2011) / Oregon Medford-Diamond Lake zone (m)', 'EPSG', 8323, 'PROJCS[\"NAD83(2"
  "011) / Oregon Medford-Diamond Lake zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTH"
  "ORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-122.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00004,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",-60000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"8323\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8324, 1, 'NAD83(2011) / Oregon Medford-Diamond Lake zone (ft)', 'EPSG', 8324, 'PROJCS[\"NAD83("
  "2011) / Oregon Medford-Diamond Lake zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference Syst"
  "em 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AU"
  "THORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-122.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00004"
  ",AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",196850.3937,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fals"
  "e northing\",-196850.3937,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\""
  ",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8325, 1, 'NAD83(2011) / Oregon Mitchell zone (m)', 'EPSG', 8325, 'PROJCS[\"NAD83(2011) / Orego"
  "n Mitchell zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",47,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-120.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99927,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",290000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"8325\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8326, 1, 'NAD83(2011) / Oregon Mitchell zone (ft)', 'EPSG', 8326, 'PROJCS[\"NAD83(2011) / Oreg"
  "on Mitchell zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",47,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-120.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99927,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",98425.1969,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",951443.5696,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"8326\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8327, 1, 'NAD83(2011) / Oregon North Central zone (m)', 'EPSG', 8327, 'PROJCS[\"NAD83(2011) / "
  "Oregon North Central zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",46.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",140000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8327\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8328, 1, 'NAD83(2011) / Oregon North Central zone (ft)', 'EPSG', 8328, 'PROJCS[\"NAD83(2011) /"
  " Oregon North Central zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",46.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",328083.9895,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",459317.5853,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8328\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8329, 1, 'NAD83(2011) / Oregon Ochoco Summit zone (m)', 'EPSG', 8329, 'PROJCS[\"NAD83(2011) / "
  "Oregon Ochoco Summit zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1.00006,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",-80000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],A"
  "XIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8329\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8330, 1, 'NAD83(2011) / Oregon Ochoco Summit zone (ft)', 'EPSG', 8330, 'PROJCS[\"NAD83(2011) /"
  " Oregon Ochoco Summit zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",43.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.00006,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",-262467.1916,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],"
  "AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8331, 1, 'NAD83(2011) / Oregon Owyhee zone (m)', 'EPSG', 8331, 'PROJCS[\"NAD83(2011) / Oregon "
  "Owyhee zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117.59444"
  "4444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00018,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",70000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"833"
  "1\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8332, 1, 'NAD83(2011) / Oregon Owyhee zone (ft)', 'EPSG', 8332, 'PROJCS[\"NAD83(2011) / Oregon"
  " Owyhee zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117.594"
  "444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00018,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",229658.7927,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"8332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8333, 1, 'NAD83(2011) / Oregon Pilot Rock-Ukiah zone (m)', 'EPSG', 8333, 'PROJCS[\"NAD83(2011)"
  " / Oregon Pilot Rock-Ukiah zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\""
  "EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",46.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00002"
  "5,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",130000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y"
  "\",NORTH],AUTHORITY[\"EPSG\",\"8333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8334, 1, 'NAD83(2011) / Oregon Pilot Rock-Ukiah zone (ft)', 'EPSG', 8334, 'PROJCS[\"NAD83(2011"
  ") / Oregon Pilot Rock-Ukiah zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY["
  "\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",46.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER"
  "[\"Longitude of natural origin\",-119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00"
  "0025,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",164041.9948,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",426509.1864,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8334\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8335, 1, 'NAD83(2011) / Oregon Prairie City-Brogan zone (m)', 'EPSG', 8335, 'PROJCS[\"NAD83(20"
  "11) / Oregon Prairie City-Brogan zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2"
  "011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-118,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00017,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",60000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"8335\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8336, 1, 'NAD83(2011) / Oregon Prairie City-Brogan zone (ft)', 'EPSG', 8336, 'PROJCS[\"NAD83(2"
  "011) / Oregon Prairie City-Brogan zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System"
  " 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTH"
  "ORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-118,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00017,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",196850.3937,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nor"
  "thing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"8336\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8337, 1, 'NAD83(2011) / Oregon Riley-Lakeview zone (m)', 'EPSG', 8337, 'PROJCS[\"NAD83(2011) /"
  " Oregon Riley-Lakeview zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-120.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000215,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",70000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"8337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8338, 1, 'NAD83(2011) / Oregon Riley-Lakeview zone (ft)', 'EPSG', 8338, 'PROJCS[\"NAD83(2011) "
  "/ Oregon Riley-Lakeview zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",41.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-120.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000215,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",229658.7927,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"8338\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8339, 1, 'NAD83(2011) / Oregon Siskiyou Pass zone (m)', 'EPSG', 8339, 'PROJCS[\"NAD83(2011) / "
  "Oregon Siskiyou Pass zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-122.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",1.00015,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",10000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",60000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8340, 1, 'NAD83(2011) / Oregon Siskiyou Pass zone (ft)', 'EPSG', 8340, 'PROJCS[\"NAD83(2011) /"
  " Oregon Siskiyou Pass zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-122.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",1.00015,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",32808.399,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",196850.3937,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8341, 1, 'NAD83(2011) / Oregon Ukiah-Fox zone (m)', 'EPSG', 8341, 'PROJCS[\"NAD83(2011) / Oreg"
  "on Ukiah-Fox zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GR"
  "S 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00014,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",90000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"8341\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8342, 1, 'NAD83(2011) / Oregon Ukiah-Fox zone (ft)', 'EPSG', 8342, 'PROJCS[\"NAD83(2011) / Ore"
  "gon Ukiah-Fox zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"980"
  "1\"]],PARAMETER[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",-119,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00014,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",98425.1969,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",295275.5906,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8342\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8343, 1, 'NAD83(2011) / Oregon Wallowa zone (m)', 'EPSG', 8343, 'PROJCS[\"NAD83(2011) / Oregon"
  " Wallowa zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-117.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000195,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",60000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"8343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8344, 1, 'NAD83(2011) / Oregon Wallowa zone (ft)', 'EPSG', 8344, 'PROJCS[\"NAD83(2011) / Orego"
  "n Wallowa zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",45.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural orig"
  "in\",-117.511111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000195,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",196850.3937,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH"
  "],AUTHORITY[\"EPSG\",\"8344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8345, 1, 'NAD83(2011) / Oregon Warner Highway zone (m)', 'EPSG', 8345, 'PROJCS[\"NAD83(2011) /"
  " Oregon Warner Highway zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPH"
  "EROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000245,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",40000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",60000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"8345\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8346, 1, 'NAD83(2011) / Oregon Warner Highway zone (ft)', 'EPSG', 8346, 'PROJCS[\"NAD83(2011) "
  "/ Oregon Warner Highway zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPS"
  "G\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-120,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000245,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",131233.5958,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",196850.3937,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8347, 1, 'NAD83(2011) / Oregon Willamette Pass zone (m)', 'EPSG', 8347, 'PROJCS[\"NAD83(2011) "
  "/ Oregon Willamette Pass zone (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",43,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-122,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000223,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",20000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "8347\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8348, 1, 'NAD83(2011) / Oregon Willamette Pass zone (ft)', 'EPSG', 8348, 'PROJCS[\"NAD83(2011)"
  " / Oregon Willamette Pass zone (ft)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",43,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-122,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000223,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",65616.7979,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"foot\",0.3048,AUTHORITY[\"EPSG\",\"9002\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY"
  "[\"EPSG\",\"8348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8351, 1, 'S-JTSK [JTSK03]', 'EPSG', 8351, 'GEOGCS[\"S-JTSK [JTSK03]\",DATUM[\"System of the Un"
  "ified Trigonometrical Cadastral Network [JTSK03]\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\""
  ",\"7004\"]],TOWGS84[485.021,169.465,483.839,7.786342,4.397554,4.102655,0],AUTHORITY[\"EPSG\",\"1201\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8352, 1, 'S-JTSK [JTSK03] / Krovak', 'EPSG', 8352, 'PROJCS[\"S-JTSK [JTSK03] / Krovak\",GEOGCS"
  "[\"S-JTSK [JTSK03]\",DATUM[\"System of the Unified Trigonometrical Cadastral Network [JTSK03]\",SPHEROID[\"Bessel 184"
  "1\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[485.021,169.465,483.839,7.786342,4.397554,4.102655,"
  "0],AUTHORITY[\"EPSG\",\"1201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8351\"]],PROJECTION[\""
  "Krovak\",AUTHORITY[\"EPSG\",\"9819\"]],PARAMETER[\"Latitude of projection centre\",49.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8811\"]],PARAMETER[\"Longitude of origin\",24.8333333333333,AUTHORITY[\"EPSG\",\"8833\"]],PARAMETER[\"Co-latitude "
  "of cone axis\",30.2881397527778,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"Latitude of pseudo standard parallel\",78.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseudo standard parallel\",0.9999,AUTHORITY[\""
  "EPSG\",\"8819\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",SOUTH],AXIS[\"Y\",WEST],AUTHORITY[\"E"
  "PSG\",\"8352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8353, 1, 'S-JTSK [JTSK03] / Krovak East North', 'EPSG', 8353, 'PROJCS[\"S-JTSK [JTSK03] / Krov"
  "ak East North\",GEOGCS[\"S-JTSK [JTSK03]\",DATUM[\"System of the Unified Trigonometrical Cadastral Network [JTSK03]\""
  ",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[485.021,169.465,483.839,7.786"
  "342,4.397554,4.102655,0],AUTHORITY[\"EPSG\",\"1201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "8351\"]],PROJECTION[\"Krovak (North Orientated)\",AUTHORITY[\"EPSG\",\"1041\"]],PARAMETER[\"Latitude of projection ce"
  "ntre\",49.5111111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of origin\",24.8333333333333,AUTHORITY[\""
  "EPSG\",\"8833\"]],PARAMETER[\"Co-latitude of cone axis\",30.2881397527778,AUTHORITY[\"EPSG\",\"1036\"]],PARAMETER[\"L"
  "atitude of pseudo standard parallel\",78.5111111111111,AUTHORITY[\"EPSG\",\"8818\"]],PARAMETER[\"Scale factor on pseu"
  "do standard parallel\",0.9999,AUTHORITY[\"EPSG\",\"8819\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8379, 1, 'NAD83 / NCRS Las Vegas (m)', 'EPSG', 8379, 'PROJCS[\"NAD83 / NCRS Las Vegas (m)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114.977777777778,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8379\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8380, 1, 'NAD83 / NCRS Las Vegas (ftUS)', 'EPSG', 8380, 'PROJCS[\"NAD83 / NCRS Las Vegas (ftUS"
  ")\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114.977777777778"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",656166.6667,AUTHORITY["
  "\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\""
  "Y\",NORTH],AUTHORITY[\"EPSG\",\"8380\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8381, 1, 'NAD83 / NCRS Las Vegas high (m)', 'EPSG', 8381, 'PROJCS[\"NAD83 / NCRS Las Vegas hig"
  "h (m)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of n"
  "atural origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114.9777777"
  "77778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000135,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",400000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"8381\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8382, 1, 'NAD83 / NCRS Las Vegas high (ftUS)', 'EPSG', 8382, 'PROJCS[\"NAD83 / NCRS Las Vegas "
  "high (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHOR"
  "ITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114.9"
  "77777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000135,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1312333.3333,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAS"
  "T],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8382\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8383, 1, 'NAD83(2011) / NCRS Las Vegas (m)', 'EPSG', 8383, 'PROJCS[\"NAD83(2011) / NCRS Las Ve"
  "gas (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114.977"
  "777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"8383\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8384, 1, 'NAD83(2011) / NCRS Las Vegas (ftUS)', 'EPSG', 8384, 'PROJCS[\"NAD83(2011) / NCRS Las"
  " Vegas (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1"
  "14.977777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.0001,AUTHORITY[\"EPSG\","
  "\"8805\"]],PARAMETER[\"False easting\",328083.3333,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",656166."
  "6667,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X"
  "\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8384\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8385, 1, 'NAD83(2011) / NCRS Las Vegas high (m)', 'EPSG', 8385, 'PROJCS[\"NAD83(2011) / NCRS L"
  "as Vegas high (m)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 19"
  "80\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-114.977777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000135,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",40000"
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUT"
  "HORITY[\"EPSG\",\"8385\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8387, 1, 'NAD83(2011) / NCRS Las Vegas high (ftUS)', 'EPSG', 8387, 'PROJCS[\"NAD83(2011) / NCR"
  "S Las Vegas high (ftUS)\",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\""
  "GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",36.2611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-114.977777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000135,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",1312333.3333,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]]"
  ",AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8387\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8391, 1, 'GDA94 / WEIPA94', 'EPSG', 8391, 'PROJCS[\"GDA94 / WEIPA94\",GEOGCS[\"GDA94\",DATUM[\""
  "Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0"
  ".017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4283\"]"
  "],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORI"
  "TY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",0.999929,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8395, 1, 'ETRS89 / Gauss-Kruger CM 9E', 'EPSG', 8395, 'PROJCS[\"ETRS89 / Gauss-Kruger CM 9E\","
  "GEOGCS[\"ETRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8427, 1, 'Hong Kong Geodetic CS', 'EPSG', 8427, 'GEOGCS[\"Hong Kong Geodetic CS\",DATUM[\"Hong"
  " Kong Geodetic\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUT"
  "HORITY[\"EPSG\",\"1209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8427\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8428, 1, 'Macao 1920', 'EPSG', 8428, 'GEOGCS[\"Macao 1920\",DATUM[\"Macao 1920\",SPHEROID[\"In"
  "ternational 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-202.865,-303.99,-155.873,34.079,-76.126,-32.66,"
  "6.096],AUTHORITY[\"EPSG\",\"1207\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292"
  "519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8428\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8431, 1, 'Macao 2008', 'EPSG', 8431, 'GEOGCS[\"Macao 2008\",DATUM[\"Macao Geodetic Datum 2008\""
  ",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\","
  "\"1208\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8431\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8433, 1, 'Macao 1920 / Macao Grid', 'EPSG', 8433, 'PROJCS[\"Macao 1920 / Macao Grid\",GEOGCS[\""
  "Macao 1920\",DATUM[\"Macao 1920\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-"
  "202.865,-303.99,-155.873,34.079,-76.126,-32.66,6.096],AUTHORITY[\"EPSG\",\"1207\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"8428\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",22.2123972222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "113.536469444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",20000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",20000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY[\"EPSG\""
  ",\"8433\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8441, 1, 'Tananarive / Laborde Grid', 'EPSG', 8441, 'PROJCS[\"Tananarive / Laborde Grid\",GEOG"
  "CS[\"Tananarive\",DATUM[\"Tananarive 1925\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]]"
  ",TOWGS84[-198.383,-240.517,-107.909,0,0,0,0],AUTHORITY[\"EPSG\",\"6297\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4297\"]],PROJECTION[\"Laborde Oblique Mercator\",AUTHORITY[\"EPSG\",\"9813\"]],PARAMETER[\"Latit"
  "ude of projection centre\",-18.9111111111111,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre"
  "\",46.4372291666667,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",18.9111111111111,AUTHORITY[\""
  "EPSG\",\"8813\"]],PARAMETER[\"Scale factor on initial line\",0.9995,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False e"
  "asting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"8441\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8449, 1, 'NAD83(FBN) (8449 deprecated)', 'EPSG', 8449, 'GEOGCS[\"NAD83(FBN)\",DATUM[\"NAD83 (H"
  "igh Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY"
  "[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8449\"]]', 'Deprecated since EPSG"
  " Dataset 9.6. Superseded by SRID 8860.');\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8455, 1, 'RGTAAF07 / UTM zone 53S', 'EPSG', 8455, 'PROJCS[\"RGTAAF07 / UTM zone 53S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8455\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8456, 1, 'RGTAAF07 / UTM zone 54S', 'EPSG', 8456, 'PROJCS[\"RGTAAF07 / UTM zone 54S\",GEOGCS[\""
  "RGTAAF07\",DATUM[\"Reseau Geodesique des Terres Australes et Antarctiques Francaises 2007\",SPHEROID[\"GRS 1980\",637"
  "8137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1113\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"7073\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitu"
  "de of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8456\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8518, 1, 'NAD83(2011) / KS RCS zone 1', 'EPSG', 8518, 'PROJCS[\"NAD83(2011) / KS RCS zone 1\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-101.611111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000156,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8518\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8519, 1, 'NAD83(2011) / KS RCS zone 2', 'EPSG', 8519, 'PROJCS[\"NAD83(2011) / KS RCS zone 2\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-100.961111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000134,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8520, 1, 'NAD83(2011) / KS RCS zone 3', 'EPSG', 8520, 'PROJCS[\"NAD83(2011) / KS RCS zone 3\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-100.361111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000116,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8520\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8521, 1, 'NAD83(2011) / KS RCS zone 4', 'EPSG', 8521, 'PROJCS[\"NAD83(2011) / KS RCS zone 4\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99.4611111111111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000082,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8522, 1, 'NAD83(2011) / KS RCS zone 5', 'EPSG', 8522, 'PROJCS[\"NAD83(2011) / KS RCS zone 5\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-98.6666666666667"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000078,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8523, 1, 'NAD83(2011) / KS RCS zone 6', 'EPSG', 8523, 'PROJCS[\"NAD83(2011) / KS RCS zone 6\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-98.15,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000068,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"8523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8524, 1, 'NAD83(2011) / KS RCS zone 7', 'EPSG', 8524, 'PROJCS[\"NAD83(2011) / KS RCS zone 7\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-97.3333333333333"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000049,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8"
  "807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8524\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8525, 1, 'NAD83(2011) / KS RCS zone 8', 'EPSG', 8525, 'PROJCS[\"NAD83(2011) / KS RCS zone 8\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",39.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-96.511"
  "1111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000044,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXI"
  "S[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8525\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8526, 1, 'NAD83(2011) / KS RCS zone 9', 'EPSG', 8526, 'PROJCS[\"NAD83(2011) / KS RCS zone 9\","
  "GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.257"
  "222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-96.511"
  "1111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00005,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS"
  "[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8526\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8527, 1, 'NAD83(2011) / KS RCS zone 10', 'EPSG', 8527, 'PROJCS[\"NAD83(2011) / KS RCS zone 10\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",39.6333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-95.75,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00004,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NOR"
  "TH],AUTHORITY[\"EPSG\",\"8527\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8528, 1, 'NAD83(2011) / KS RCS zone 11', 'EPSG', 8528, 'PROJCS[\"NAD83(2011) / KS RCS zone 11\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",39.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-95.25,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000033,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",600000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"8528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8529, 1, 'NAD83(2011) / KS RCS zone 12', 'EPSG', 8529, 'PROJCS[\"NAD83(2011) / KS RCS zone 12\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-101.416666666667,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.00014,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"8529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8531, 1, 'NAD83(2011) / KS RCS zone 13', 'EPSG', 8531, 'PROJCS[\"NAD83(2011) / KS RCS zone 13\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-100.411111111111,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000109,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"8531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8533, 1, 'NAD83(2011) / KS RCS zone 14', 'EPSG', 8533, 'PROJCS[\"NAD83(2011) / KS RCS zone 14\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99.6666666666667,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000097,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"8533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8534, 1, 'NAD83(2011) / KS RCS zone 15', 'EPSG', 8534, 'PROJCS[\"NAD83(2011) / KS RCS zone 15\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99.2,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1.000087,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US surv"
  "ey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8"
  "534\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8535, 1, 'NAD83(2011) / KS RCS zone 16', 'EPSG', 8535, 'PROJCS[\"NAD83(2011) / KS RCS zone 16\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-98.5611111111111,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000069,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"8535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8536, 1, 'NAD83(2011) / KS RCS zone 17', 'EPSG', 8536, 'PROJCS[\"NAD83(2011) / KS RCS zone 17\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",37.7666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-97.511"
  "1111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000059,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",400000,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8536\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8538, 1, 'NAD83(2011) / KS RCS zone 18', 'EPSG', 8538, 'PROJCS[\"NAD83(2011) / KS RCS zone 18\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",37.1944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-97.511"
  "1111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000055,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHOR"
  "ITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AX"
  "IS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8539, 1, 'NAD83(2011) / KS RCS zone 19', 'EPSG', 8539, 'PROJCS[\"NAD83(2011) / KS RCS zone 19\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-95.9777777777778,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000034,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"8539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8540, 1, 'NAD83(2011) / KS RCS zone 20', 'EPSG', 8540, 'PROJCS[\"NAD83(2011) / KS RCS zone 20\""
  ",GEOGCS[\"NAD83(2011)\",DATUM[\"NAD83 (National Spatial Reference System 2011)\",SPHEROID[\"GRS 1980\",6378137,298.25"
  "7222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1116\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"6318\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",36.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-95.0944444444444,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000031,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"8540\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8545, 1, 'NAD83(HARN Corrected)', 'EPSG', 8545, 'GEOGCS[\"NAD83(HARN Corrected)\",DATUM[\"NAD8"
  "3 (High Accuracy Reference Network - Corrected)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"70"
  "19\"]],AUTHORITY[\"EPSG\",\"1212\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292"
  "519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8545\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8677, 1, 'MGI 1901 / Balkans zone 5', 'EPSG', 8677, 'PROJCS[\"MGI 1901 / Balkans zone 5\",GEOG"
  "CS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW"
  "GS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"8677\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8678, 1, 'MGI 1901 / Balkans zone 6', 'EPSG', 8678, 'PROJCS[\"MGI 1901 / Balkans zone 6\",GEOG"
  "CS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW"
  "GS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",6500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"8678\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8679, 1, 'MGI 1901 / Balkans zone 8', 'EPSG', 8679, 'PROJCS[\"MGI 1901 / Balkans zone 8\",GEOG"
  "CS[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOW"
  "GS84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",8500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"8679\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8682, 1, 'SRB_ETRS89 / UTM zone 34N', 'EPSG', 8682, 'PROJCS[\"SRB_ETRS89 / UTM zone 34N\",GEOG"
  "CS[\"SRB_ETRS89\",DATUM[\"Serbian Spatial Reference System 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORI"
  "TY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1214\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"8685\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8682\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8685, 1, 'SRB_ETRS89', 'EPSG', 8685, 'GEOGCS[\"SRB_ETRS89\",DATUM[\"Serbian Spatial Reference "
  "System 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHOR"
  "ITY[\"EPSG\",\"1214\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUT"
  "HORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8685\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8686, 1, 'MGI 1901 / Slovenia Grid', 'EPSG', 8686, 'PROJCS[\"MGI 1901 / Slovenia Grid\",GEOGCS"
  "[\"MGI 1901\",DATUM[\"MGI 1901\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS"
  "84[682,-203,480,0,0,0,0],AUTHORITY[\"EPSG\",\"1031\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "3906\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"8686\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8687, 1, 'Slovenia 1996 / UTM zone 33N', 'EPSG', 8687, 'PROJCS[\"Slovenia 1996 / UTM zone 33N\""
  ",GEOGCS[\"Slovenia 1996\",DATUM[\"Slovenia Geodetic Datum 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6765\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4765\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8687\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8692, 1, 'NAD83(MA11) / UTM zone 54N', 'EPSG', 8692, 'PROJCS[\"NAD83(MA11) / UTM zone 54N\",GE"
  "OGCS[\"NAD83(MA11)\",DATUM[\"NAD83 (National Spatial Reference System MA11)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1118\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6325\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8693, 1, 'NAD83(MA11) / UTM zone 55N', 'EPSG', 8693, 'PROJCS[\"NAD83(MA11) / UTM zone 55N\",GE"
  "OGCS[\"NAD83(MA11)\",DATUM[\"NAD83 (National Spatial Reference System MA11)\",SPHEROID[\"GRS 1980\",6378137,298.25722"
  "2101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1118\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"6325\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],"
  "PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,"
  "AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8693\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8818, 1, 'MTRF-2000', 'EPSG', 8818, 'GEOGCS[\"MTRF-2000\",DATUM[\"MOMRA Terrestrial Reference "
  "Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORI"
  "TY[\"EPSG\",\"1218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8818\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8826, 1, 'NAD83 / Idaho Transverse Mercator', 'EPSG', 8826, 'PROJCS[\"NAD83 / Idaho Transverse"
  " Mercator\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",42,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1200000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"8826\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8836, 1, 'MTRF-2000 / UTM zone 36N', 'EPSG', 8836, 'PROJCS[\"MTRF-2000 / UTM zone 36N\",GEOGCS"
  "[\"MTRF-2000\",DATUM[\"MOMRA Terrestrial Reference Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"8818\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8836\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8837, 1, 'MTRF-2000 / UTM zone 37N', 'EPSG', 8837, 'PROJCS[\"MTRF-2000 / UTM zone 37N\",GEOGCS"
  "[\"MTRF-2000\",DATUM[\"MOMRA Terrestrial Reference Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"8818\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8838, 1, 'MTRF-2000 / UTM zone 38N', 'EPSG', 8838, 'PROJCS[\"MTRF-2000 / UTM zone 38N\",GEOGCS"
  "[\"MTRF-2000\",DATUM[\"MOMRA Terrestrial Reference Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"8818\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8838\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8839, 1, 'MTRF-2000 / UTM zone 39N', 'EPSG', 8839, 'PROJCS[\"MTRF-2000 / UTM zone 39N\",GEOGCS"
  "[\"MTRF-2000\",DATUM[\"MOMRA Terrestrial Reference Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"8818\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8840, 1, 'MTRF-2000 / UTM zone 40N', 'EPSG', 8840, 'PROJCS[\"MTRF-2000 / UTM zone 40N\",GEOGCS"
  "[\"MTRF-2000\",DATUM[\"MOMRA Terrestrial Reference Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY"
  "[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"8818\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8840\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8857, 1, 'WGS 84 / Equal Earth Greenwich', 'EPSG', 8857, 'PROJCS[\"WGS 84 / Equal Earth Greenw"
  "ich\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EP"
  "SG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]]"
  ",PROJECTION[\"Equal Earth\",AUTHORITY[\"EPSG\",\"1078\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"8857\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8858, 1, 'WGS 84 / Equal Earth Americas', 'EPSG', 8858, 'PROJCS[\"WGS 84 / Equal Earth America"
  "s\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG"
  "\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],P"
  "ROJECTION[\"Equal Earth\",AUTHORITY[\"EPSG\",\"1078\"]],PARAMETER[\"Longitude of natural origin\",-90,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"8858\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8859, 1, 'WGS 84 / Equal Earth Asia-Pacific', 'EPSG', 8859, 'PROJCS[\"WGS 84 / Equal Earth Asi"
  "a-Pacific\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"43"
  "26\"]],PROJECTION[\"Equal Earth\",AUTHORITY[\"EPSG\",\"1078\"]],PARAMETER[\"Longitude of natural origin\",150,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUT"
  "HORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"8859\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8860, 1, 'NAD83(FBN)', 'EPSG', 8860, 'GEOGCS[\"NAD83(FBN)\",DATUM[\"NAD83 (Federal Base Networ"
  "k)\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1211\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8860\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8888, 1, 'WGS 84 (Transit)', 'EPSG', 8888, 'GEOGCS[\"WGS 84 (Transit)\",DATUM[\"World Geodetic"
  " System 1984 (Transit)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\""
  "1166\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8888\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8900, 1, 'RGWF96', 'EPSG', 8900, 'GEOGCS[\"RGWF96\",DATUM[\"Reseau Geodesique de Wallis et Fut"
  "una 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1223\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8900\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8902, 1, 'RGWF96 (lon-lat)', 'EPSG', 8902, 'GEOGCS[\"RGWF96 (lon-lat)\",DATUM[\"Reseau Geodesi"
  "que de Wallis et Futuna 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\""
  "EPSG\",\"1223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lon\",EAST],AXIS[\"Lat\",NORTH],AUTHORITY[\"EPSG\",\"8902\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8903, 1, 'RGWF96 / UTM zone 1S', 'EPSG', 8903, 'PROJCS[\"RGWF96 / UTM zone 1S\",GEOGCS[\"RGWF9"
  "6\",DATUM[\"Reseau Geodesique de Wallis et Futuna 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],AUTHORITY[\"EPSG\",\"1223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8900\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY["
  "\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"8903\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8907, 1, 'CR-SIRGAS', 'EPSG', 8907, 'GEOGCS[\"CR-SIRGAS\",DATUM[\"CR-SIRGAS\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1225\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"8907\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8908, 1, 'CR-SIRGAS / CRTM05', 'EPSG', 8908, 'PROJCS[\"CR-SIRGAS / CRTM05\",GEOGCS[\"CR-SIRGAS"
  "\",DATUM[\"CR-SIRGAS\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\""
  "1225\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8907\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"8908\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8909, 1, 'CR-SIRGAS / UTM zone 16N', 'EPSG', 8909, 'PROJCS[\"CR-SIRGAS / UTM zone 16N\",GEOGCS"
  "[\"CR-SIRGAS\",DATUM[\"CR-SIRGAS\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT"
  "Y[\"EPSG\",\"1225\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8907\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"8909\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8910, 1, 'CR-SIRGAS / UTM zone 17N', 'EPSG', 8910, 'PROJCS[\"CR-SIRGAS / UTM zone 17N\",GEOGCS"
  "[\"CR-SIRGAS\",DATUM[\"CR-SIRGAS\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORIT"
  "Y[\"EPSG\",\"1225\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8907\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"8910\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8972, 1, 'SIRGAS-CON DGF00P01', 'EPSG', 8972, 'GEOGCS[\"SIRGAS-CON DGF00P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF00P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1227\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8972\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8973, 1, 'SIRGAS-CON DGF01P01', 'EPSG', 8973, 'GEOGCS[\"SIRGAS-CON DGF01P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF01P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1228\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8973\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8974, 1, 'SIRGAS-CON DGF01P02', 'EPSG', 8974, 'GEOGCS[\"SIRGAS-CON DGF01P02\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF01P02\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8974\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8975, 1, 'SIRGAS-CON DGF02P01', 'EPSG', 8975, 'GEOGCS[\"SIRGAS-CON DGF02P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF02P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8975\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8976, 1, 'SIRGAS-CON DGF04P01', 'EPSG', 8976, 'GEOGCS[\"SIRGAS-CON DGF04P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF04P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1231\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8976\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8977, 1, 'SIRGAS-CON DGF05P01', 'EPSG', 8977, 'GEOGCS[\"SIRGAS-CON DGF05P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF05P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1232\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8977\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8978, 1, 'SIRGAS-CON DGF06P01', 'EPSG', 8978, 'GEOGCS[\"SIRGAS-CON DGF06P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF06P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1233\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8978\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8979, 1, 'SIRGAS-CON DGF07P01', 'EPSG', 8979, 'GEOGCS[\"SIRGAS-CON DGF07P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF07P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1234\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8979\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8980, 1, 'SIRGAS-CON DGF08P01', 'EPSG', 8980, 'GEOGCS[\"SIRGAS-CON DGF08P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network DGF08P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1235\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8980\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8981, 1, 'SIRGAS-CON SIR09P01', 'EPSG', 8981, 'GEOGCS[\"SIRGAS-CON SIR09P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR09P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1236\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8981\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8982, 1, 'SIRGAS-CON SIR10P01', 'EPSG', 8982, 'GEOGCS[\"SIRGAS-CON SIR10P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR10P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1237\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8982\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8983, 1, 'SIRGAS-CON SIR11P01', 'EPSG', 8983, 'GEOGCS[\"SIRGAS-CON SIR11P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR11P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1238\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8983\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8984, 1, 'SIRGAS-CON SIR13P01', 'EPSG', 8984, 'GEOGCS[\"SIRGAS-CON SIR13P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR13P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8984\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8985, 1, 'SIRGAS-CON SIR14P01', 'EPSG', 8985, 'GEOGCS[\"SIRGAS-CON SIR14P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR14P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1240\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8985\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8986, 1, 'SIRGAS-CON SIR15P01', 'EPSG', 8986, 'GEOGCS[\"SIRGAS-CON SIR15P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR15P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1241\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8986\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8987, 1, 'SIRGAS-CON SIR17P01', 'EPSG', 8987, 'GEOGCS[\"SIRGAS-CON SIR17P01\",DATUM[\"SIRGAS C"
  "ontinuously Operating Network SIR17P01\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AU"
  "THORITY[\"EPSG\",\"1242\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8987\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8988, 1, 'ITRF88', 'EPSG', 8988, 'GEOGCS[\"ITRF88\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1988\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6647\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8988\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8989, 1, 'ITRF89', 'EPSG', 8989, 'GEOGCS[\"ITRF89\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6648\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8989\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8990, 1, 'ITRF90', 'EPSG', 8990, 'GEOGCS[\"ITRF90\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1990\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6649\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8990\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8991, 1, 'ITRF91', 'EPSG', 8991, 'GEOGCS[\"ITRF91\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1991\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6650\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8992, 1, 'ITRF92', 'EPSG', 8992, 'GEOGCS[\"ITRF92\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6651\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8993, 1, 'ITRF93', 'EPSG', 8993, 'GEOGCS[\"ITRF93\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6652\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8993\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8994, 1, 'ITRF94', 'EPSG', 8994, 'GEOGCS[\"ITRF94\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6653\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8994\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8995, 1, 'ITRF96', 'EPSG', 8995, 'GEOGCS[\"ITRF96\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6654\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8995\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8996, 1, 'ITRF97', 'EPSG', 8996, 'GEOGCS[\"ITRF97\",DATUM[\"International Terrestrial Referenc"
  "e Frame 1997\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"6655\"]"
  "],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122"
  "\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8996\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8997, 1, 'ITRF2000', 'EPSG', 8997, 'GEOGCS[\"ITRF2000\",DATUM[\"International Terrestrial Refe"
  "rence Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"665"
  "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8997\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8998, 1, 'ITRF2005', 'EPSG', 8998, 'GEOGCS[\"ITRF2005\",DATUM[\"International Terrestrial Refe"
  "rence Frame 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"689"
  "6\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8998\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (8999, 1, 'ITRF2008', 'EPSG', 8999, 'GEOGCS[\"ITRF2008\",DATUM[\"International Terrestrial Refe"
  "rence Frame 2008\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"106"
  "1\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8999\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9000, 1, 'ITRF2014', 'EPSG', 9000, 'GEOGCS[\"ITRF2014\",DATUM[\"International Terrestrial Refe"
  "rence Frame 2014\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"116"
  "5\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9000\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9003, 1, 'IGS97', 'EPSG', 9003, 'GEOGCS[\"IGS97\",DATUM[\"IGS97\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1244\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9003\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9006, 1, 'IGS00', 'EPSG', 9006, 'GEOGCS[\"IGS00\",DATUM[\"IGS00\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9006\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9009, 1, 'IGb00', 'EPSG', 9009, 'GEOGCS[\"IGb00\",DATUM[\"IGb00\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1246\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9009\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9012, 1, 'IGS05', 'EPSG', 9012, 'GEOGCS[\"IGS05\",DATUM[\"IGS05\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1247\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9012\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9014, 1, 'IGS08', 'EPSG', 9014, 'GEOGCS[\"IGS08\",DATUM[\"IGS08\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1141\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9017, 1, 'IGb08', 'EPSG', 9017, 'GEOGCS[\"IGb08\",DATUM[\"IGb08\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1248\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9017\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9019, 1, 'IGS14', 'EPSG', 9019, 'GEOGCS[\"IGS14\",DATUM[\"IGS14\",SPHEROID[\"GRS 1980\",637813"
  "7,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1191\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"9019\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9039, 1, 'ISN2016 / LAEA Europe', 'EPSG', 9039, 'PROJCS[\"ISN2016 / LAEA Europe\",GEOGCS[\"ISN"
  "2016\",DATUM[\"Islands Net 2016\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY"
  "[\"EPSG\",\"1187\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHOR"
  "ITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8086\"]],PROJECTION[\"Lambert Azi"
  "muthal Equal Area\",AUTHORITY[\"EPSG\",\"9820\"]],PARAMETER[\"Latitude of natural origin\",52,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",10,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",4321000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3210000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EAST],AUTHORITY[\"EPSG\",\"9039\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9040, 1, 'ISN2016 / LCC Europe', 'EPSG', 9040, 'PROJCS[\"ISN2016 / LCC Europe\",GEOGCS[\"ISN20"
  "16\",DATUM[\"Islands Net 2016\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\""
  "EPSG\",\"1187\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"8086\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",52,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",10,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",65,AUTHORITY[\"EPSG\",\"8824\""
  "]],PARAMETER[\"Easting at false origin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",2800000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",NORTH],AXIS[\"X\",EA"
  "ST],AUTHORITY[\"EPSG\",\"9040\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9053, 1, 'WGS 84 (G730)', 'EPSG', 9053, 'GEOGCS[\"WGS 84 (G730)\",DATUM[\"World Geodetic Syste"
  "m 1984 (G730)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1152\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9053\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9054, 1, 'WGS 84 (G873)', 'EPSG', 9054, 'GEOGCS[\"WGS 84 (G873)\",DATUM[\"World Geodetic Syste"
  "m 1984 (G873)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1153\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9054\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9055, 1, 'WGS 84 (G1150)', 'EPSG', 9055, 'GEOGCS[\"WGS 84 (G1150)\",DATUM[\"World Geodetic Sys"
  "tem 1984 (G1150)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1154\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9055\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9056, 1, 'WGS 84 (G1674)', 'EPSG', 9056, 'GEOGCS[\"WGS 84 (G1674)\",DATUM[\"World Geodetic Sys"
  "tem 1984 (G1674)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1155\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9056\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9057, 1, 'WGS 84 (G1762)', 'EPSG', 9057, 'GEOGCS[\"WGS 84 (G1762)\",DATUM[\"World Geodetic Sys"
  "tem 1984 (G1762)\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"1156\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9059, 1, 'ETRF89', 'EPSG', 9059, 'GEOGCS[\"ETRF89\",DATUM[\"European Terrestrial Reference Fra"
  "me 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1178\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9059\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9060, 1, 'ETRF90', 'EPSG', 9060, 'GEOGCS[\"ETRF90\",DATUM[\"European Terrestrial Reference Fra"
  "me 1990\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1179\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9060\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9061, 1, 'ETRF91', 'EPSG', 9061, 'GEOGCS[\"ETRF91\",DATUM[\"European Terrestrial Reference Fra"
  "me 1991\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1180\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9061\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9062, 1, 'ETRF92', 'EPSG', 9062, 'GEOGCS[\"ETRF92\",DATUM[\"European Terrestrial Reference Fra"
  "me 1992\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1181\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9062\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9063, 1, 'ETRF93', 'EPSG', 9063, 'GEOGCS[\"ETRF93\",DATUM[\"European Terrestrial Reference Fra"
  "me 1993\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1182\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9063\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9064, 1, 'ETRF94', 'EPSG', 9064, 'GEOGCS[\"ETRF94\",DATUM[\"European Terrestrial Reference Fra"
  "me 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1183\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9064\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9065, 1, 'ETRF96', 'EPSG', 9065, 'GEOGCS[\"ETRF96\",DATUM[\"European Terrestrial Reference Fra"
  "me 1996\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1184\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9066, 1, 'ETRF97', 'EPSG', 9066, 'GEOGCS[\"ETRF97\",DATUM[\"European Terrestrial Reference Fra"
  "me 1997\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1185\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9067, 1, 'ETRF2000', 'EPSG', 9067, 'GEOGCS[\"ETRF2000\",DATUM[\"European Terrestrial Reference"
  " Frame 2000\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1186\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9067\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9068, 1, 'ETRF2005', 'EPSG', 9068, 'GEOGCS[\"ETRF2005\",DATUM[\"European Terrestrial Reference"
  " Frame 2005\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1204\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9068\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9069, 1, 'ETRF2014', 'EPSG', 9069, 'GEOGCS[\"ETRF2014\",DATUM[\"European Terrestrial Reference"
  " Frame 2014\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],AUTHORITY[\"EPSG\",\"1206\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9069\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9140, 1, 'KOSOVAREF01', 'EPSG', 9140, 'GEOGCS[\"KOSOVAREF01\",DATUM[\"Kosovo Reference System "
  "2001\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"E"
  "PSG\",\"1251\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"9140\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (9141, 1, 'KOSOVAREF01 / Balkans zone 7', 'EPSG', 9141, 'PROJCS[\"KOSOVAREF01 / Balkans zone 7\""
  ",GEOGCS[\"KOSOVAREF01\",DATUM[\"Kosovo Reference System 2001\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"1251\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"9140\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"9141\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20004, 1, 'Pulkovo 1995 / Gauss-Kruger zone 4', 'EPSG', 20004, 'PROJCS[\"Pulkovo 1995 / Gauss-"
  "Kruger zone 4\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20004\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20005, 1, 'Pulkovo 1995 / Gauss-Kruger zone 5', 'EPSG', 20005, 'PROJCS[\"Pulkovo 1995 / Gauss-"
  "Kruger zone 5\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20005\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20006, 1, 'Pulkovo 1995 / Gauss-Kruger zone 6', 'EPSG', 20006, 'PROJCS[\"Pulkovo 1995 / Gauss-"
  "Kruger zone 6\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20006\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20007, 1, 'Pulkovo 1995 / Gauss-Kruger zone 7', 'EPSG', 20007, 'PROJCS[\"Pulkovo 1995 / Gauss-"
  "Kruger zone 7\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20007\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20008, 1, 'Pulkovo 1995 / Gauss-Kruger zone 8', 'EPSG', 20008, 'PROJCS[\"Pulkovo 1995 / Gauss-"
  "Kruger zone 8\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20008\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20009, 1, 'Pulkovo 1995 / Gauss-Kruger zone 9', 'EPSG', 20009, 'PROJCS[\"Pulkovo 1995 / Gauss-"
  "Kruger zone 9\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20009\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20010, 1, 'Pulkovo 1995 / Gauss-Kruger zone 10', 'EPSG', 20010, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 10\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20010\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20011, 1, 'Pulkovo 1995 / Gauss-Kruger zone 11', 'EPSG', 20011, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 11\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20011\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20012, 1, 'Pulkovo 1995 / Gauss-Kruger zone 12', 'EPSG', 20012, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 12\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20012\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20013, 1, 'Pulkovo 1995 / Gauss-Kruger zone 13', 'EPSG', 20013, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 13\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20013\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20014, 1, 'Pulkovo 1995 / Gauss-Kruger zone 14', 'EPSG', 20014, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 14\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20014\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20015, 1, 'Pulkovo 1995 / Gauss-Kruger zone 15', 'EPSG', 20015, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 15\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20015\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20016, 1, 'Pulkovo 1995 / Gauss-Kruger zone 16', 'EPSG', 20016, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 16\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20016\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20017, 1, 'Pulkovo 1995 / Gauss-Kruger zone 17', 'EPSG', 20017, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 17\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,"
  "AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20017\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20018, 1, 'Pulkovo 1995 / Gauss-Kruger zone 18', 'EPSG', 20018, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 18\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20018\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20019, 1, 'Pulkovo 1995 / Gauss-Kruger zone 19', 'EPSG', 20019, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 19\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20019\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20020, 1, 'Pulkovo 1995 / Gauss-Kruger zone 20', 'EPSG', 20020, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 20\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20020\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20021, 1, 'Pulkovo 1995 / Gauss-Kruger zone 21', 'EPSG', 20021, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 21\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20021\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20022, 1, 'Pulkovo 1995 / Gauss-Kruger zone 22', 'EPSG', 20022, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 22\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20022\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20023, 1, 'Pulkovo 1995 / Gauss-Kruger zone 23', 'EPSG', 20023, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 23\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20023\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20024, 1, 'Pulkovo 1995 / Gauss-Kruger zone 24', 'EPSG', 20024, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 24\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",24500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20024\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20025, 1, 'Pulkovo 1995 / Gauss-Kruger zone 25', 'EPSG', 20025, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 25\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20025\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20026, 1, 'Pulkovo 1995 / Gauss-Kruger zone 26', 'EPSG', 20026, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 26\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",26500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20026\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20027, 1, 'Pulkovo 1995 / Gauss-Kruger zone 27', 'EPSG', 20027, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 27\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",27500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20027\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20028, 1, 'Pulkovo 1995 / Gauss-Kruger zone 28', 'EPSG', 20028, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 28\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",28500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20028\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20029, 1, 'Pulkovo 1995 / Gauss-Kruger zone 29', 'EPSG', 20029, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 29\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",29500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20029\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20030, 1, 'Pulkovo 1995 / Gauss-Kruger zone 30', 'EPSG', 20030, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 30\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",30500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20030\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20031, 1, 'Pulkovo 1995 / Gauss-Kruger zone 31', 'EPSG', 20031, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 31\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-17"
  "7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20031\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20032, 1, 'Pulkovo 1995 / Gauss-Kruger zone 32', 'EPSG', 20032, 'PROJCS[\"Pulkovo 1995 / Gauss"
  "-Kruger zone 32\",GEOGCS[\"Pulkovo 1995\",DATUM[\"Pulkovo 1995\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[24.47,-130.89,-81.56,0,0,0.13,-0.22],AUTHORITY[\"EPSG\",\"6200\"]],PRIMEM[\"Greenwich\","
  "0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH"
  "],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4200\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],P"
  "ARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-17"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"20032\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20135, 1, 'Adindan / UTM zone 35N', 'EPSG', 20135, 'PROJCS[\"Adindan / UTM zone 35N\",GEOGCS[\""
  "Adindan\",DATUM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84["
  "-166,-15,204,0,0,0,0],AUTHORITY[\"EPSG\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "01\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20135\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20136, 1, 'Adindan / UTM zone 36N', 'EPSG', 20136, 'PROJCS[\"Adindan / UTM zone 36N\",GEOGCS[\""
  "Adindan\",DATUM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84["
  "-166,-15,204,0,0,0,0],AUTHORITY[\"EPSG\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "01\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20137, 1, 'Adindan / UTM zone 37N', 'EPSG', 20137, 'PROJCS[\"Adindan / UTM zone 37N\",GEOGCS[\""
  "Adindan\",DATUM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84["
  "-166,-15,204,0,0,0,0],AUTHORITY[\"EPSG\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "01\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20137\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20138, 1, 'Adindan / UTM zone 38N', 'EPSG', 20138, 'PROJCS[\"Adindan / UTM zone 38N\",GEOGCS[\""
  "Adindan\",DATUM[\"Adindan\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84["
  "-166,-15,204,0,0,0,0],AUTHORITY[\"EPSG\",\"6201\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "01\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20138\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20249, 1, 'AGD66 / AMG zone 49', 'EPSG', 20249, 'PROJCS[\"AGD66 / AMG zone 49\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20250, 1, 'AGD66 / AMG zone 50', 'EPSG', 20250, 'PROJCS[\"AGD66 / AMG zone 50\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20250\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20251, 1, 'AGD66 / AMG zone 51', 'EPSG', 20251, 'PROJCS[\"AGD66 / AMG zone 51\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20251\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20252, 1, 'AGD66 / AMG zone 52', 'EPSG', 20252, 'PROJCS[\"AGD66 / AMG zone 52\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20253, 1, 'AGD66 / AMG zone 53', 'EPSG', 20253, 'PROJCS[\"AGD66 / AMG zone 53\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20254, 1, 'AGD66 / AMG zone 54', 'EPSG', 20254, 'PROJCS[\"AGD66 / AMG zone 54\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20255, 1, 'AGD66 / AMG zone 55', 'EPSG', 20255, 'PROJCS[\"AGD66 / AMG zone 55\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20256, 1, 'AGD66 / AMG zone 56', 'EPSG', 20256, 'PROJCS[\"AGD66 / AMG zone 56\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20257, 1, 'AGD66 / AMG zone 57', 'EPSG', 20257, 'PROJCS[\"AGD66 / AMG zone 57\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20258, 1, 'AGD66 / AMG zone 58', 'EPSG', 20258, 'PROJCS[\"AGD66 / AMG zone 58\",GEOGCS[\"AGD66"
  "\",DATUM[\"Australian Geodetic Datum 1966\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-124.133,-42.003,137.4,-0.008,0.557,0.178,-1.854],AUTHORITY[\"EPSG\",\"6202\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4202\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20349, 1, 'AGD84 / AMG zone 49', 'EPSG', 20349, 'PROJCS[\"AGD84 / AMG zone 49\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20349\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20350, 1, 'AGD84 / AMG zone 50', 'EPSG', 20350, 'PROJCS[\"AGD84 / AMG zone 50\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20351, 1, 'AGD84 / AMG zone 51', 'EPSG', 20351, 'PROJCS[\"AGD84 / AMG zone 51\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20352, 1, 'AGD84 / AMG zone 52', 'EPSG', 20352, 'PROJCS[\"AGD84 / AMG zone 52\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20353, 1, 'AGD84 / AMG zone 53', 'EPSG', 20353, 'PROJCS[\"AGD84 / AMG zone 53\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20354, 1, 'AGD84 / AMG zone 54', 'EPSG', 20354, 'PROJCS[\"AGD84 / AMG zone 54\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20355, 1, 'AGD84 / AMG zone 55', 'EPSG', 20355, 'PROJCS[\"AGD84 / AMG zone 55\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20356, 1, 'AGD84 / AMG zone 56', 'EPSG', 20356, 'PROJCS[\"AGD84 / AMG zone 56\",GEOGCS[\"AGD84"
  "\",DATUM[\"Australian Geodetic Datum 1984\",SPHEROID[\"Australian National Spheroid\",6378160,298.25,AUTHORITY[\"EPSG"
  "\",\"7003\"]],TOWGS84[-117.763,-51.51,139.061,0.292,0.443,0.277,-0.191],AUTHORITY[\"EPSG\",\"6203\"]],PRIMEM[\"Greenw"
  "ich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4203\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"20356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20436, 1, 'Ain el Abd / UTM zone 36N', 'EPSG', 20436, 'PROJCS[\"Ain el Abd / UTM zone 36N\",GE"
  "OGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20436\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20437, 1, 'Ain el Abd / UTM zone 37N', 'EPSG', 20437, 'PROJCS[\"Ain el Abd / UTM zone 37N\",GE"
  "OGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20437\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20438, 1, 'Ain el Abd / UTM zone 38N', 'EPSG', 20438, 'PROJCS[\"Ain el Abd / UTM zone 38N\",GE"
  "OGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20438\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20439, 1, 'Ain el Abd / UTM zone 39N', 'EPSG', 20439, 'PROJCS[\"Ain el Abd / UTM zone 39N\",GE"
  "OGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20439\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20440, 1, 'Ain el Abd / UTM zone 40N', 'EPSG', 20440, 'PROJCS[\"Ain el Abd / UTM zone 40N\",GE"
  "OGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20440\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20499, 1, 'Ain el Abd / Bahrain Grid', 'EPSG', 20499, 'PROJCS[\"Ain el Abd / Bahrain Grid\",GE"
  "OGCS[\"Ain el Abd\",DATUM[\"Ain el Abd 1970\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-18.944,-379.364,-24.063,0.04,-0.764,6.431,3.657],AUTHORITY[\"EPSG\",\"6204\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4204\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20499\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20538, 1, 'Afgooye / UTM zone 38N', 'EPSG', 20538, 'PROJCS[\"Afgooye / UTM zone 38N\",GEOGCS[\""
  "Afgooye\",DATUM[\"Afgooye\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-43,-163"
  ",45,0,0,0,0],AUTHORITY[\"EPSG\",\"6205\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4205\"]],PR"
  "OJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20539, 1, 'Afgooye / UTM zone 39N', 'EPSG', 20539, 'PROJCS[\"Afgooye / UTM zone 39N\",GEOGCS[\""
  "Afgooye\",DATUM[\"Afgooye\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\"EPSG\",\"7024\"]],TOWGS84[-43,-163"
  ",45,0,0,0,0],AUTHORITY[\"EPSG\",\"6205\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4205\"]],PR"
  "OJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor"
  " at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"880"
  "6\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXI"
  "S[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20790, 1, 'Lisbon (Lisbon) / Portuguese National Grid', 'EPSG', 20790, 'PROJCS[\"Lisbon (Lisbo"
  "n) / Portuguese National Grid\",GEOGCS[\"Lisbon (Lisbon)\",DATUM[\"Lisbon 1937 (Lisbon)\",SPHEROID[\"International 19"
  "24\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6803\"]],PRIMEM[\"Lisbon\",-9.131906111111112,AUT"
  "HORITY[\"EPSG\",\"8902\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4803\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"20"
  "790\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20791, 1, 'Lisbon (Lisbon) / Portuguese Grid', 'EPSG', 20791, 'PROJCS[\"Lisbon (Lisbon) / Port"
  "uguese Grid\",GEOGCS[\"Lisbon (Lisbon)\",DATUM[\"Lisbon 1937 (Lisbon)\",SPHEROID[\"International 1924\",6378388,297,A"
  "UTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6803\"]],PRIMEM[\"Lisbon\",-9.131906111111112,AUTHORITY[\"EPSG\",\""
  "8902\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4803\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"20791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20822, 1, 'Aratu / UTM zone 22S', 'EPSG', 20822, 'PROJCS[\"Aratu / UTM zone 22S\",GEOGCS[\"Ara"
  "tu\",DATUM[\"Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-151.99,287.04"
  ",-147.45,0,0,0,0],AUTHORITY[\"EPSG\",\"6208\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4208\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20822\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20823, 1, 'Aratu / UTM zone 23S', 'EPSG', 20823, 'PROJCS[\"Aratu / UTM zone 23S\",GEOGCS[\"Ara"
  "tu\",DATUM[\"Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-151.99,287.04"
  ",-147.45,0,0,0,0],AUTHORITY[\"EPSG\",\"6208\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4208\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20823\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20824, 1, 'Aratu / UTM zone 24S', 'EPSG', 20824, 'PROJCS[\"Aratu / UTM zone 24S\",GEOGCS[\"Ara"
  "tu\",DATUM[\"Aratu\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-151.99,287.04"
  ",-147.45,0,0,0,0],AUTHORITY[\"EPSG\",\"6208\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4208\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20824\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20934, 1, 'Arc 1950 / UTM zone 34S', 'EPSG', 20934, 'PROJCS[\"Arc 1950 / UTM zone 34S\",GEOGCS"
  "[\"Arc 1950\",DATUM[\"Arc 1950\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]]"
  ",TOWGS84[-138,-105,-289,0,0,0,0],AUTHORITY[\"EPSG\",\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4209\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20934\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20935, 1, 'Arc 1950 / UTM zone 35S', 'EPSG', 20935, 'PROJCS[\"Arc 1950 / UTM zone 35S\",GEOGCS"
  "[\"Arc 1950\",DATUM[\"Arc 1950\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]]"
  ",TOWGS84[-138,-105,-289,0,0,0,0],AUTHORITY[\"EPSG\",\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4209\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20935\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (20936, 1, 'Arc 1950 / UTM zone 36S', 'EPSG', 20936, 'PROJCS[\"Arc 1950 / UTM zone 36S\",GEOGCS"
  "[\"Arc 1950\",DATUM[\"Arc 1950\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]]"
  ",TOWGS84[-138,-105,-289,0,0,0,0],AUTHORITY[\"EPSG\",\"6209\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4209\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"20936\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21035, 1, 'Arc 1960 / UTM zone 35S', 'EPSG', 21035, 'PROJCS[\"Arc 1960 / UTM zone 35S\",GEOGCS"
  "[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4210\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21036, 1, 'Arc 1960 / UTM zone 36S', 'EPSG', 21036, 'PROJCS[\"Arc 1960 / UTM zone 36S\",GEOGCS"
  "[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4210\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21036\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21037, 1, 'Arc 1960 / UTM zone 37S', 'EPSG', 21037, 'PROJCS[\"Arc 1960 / UTM zone 37S\",GEOGCS"
  "[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4210\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21037\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21095, 1, 'Arc 1960 / UTM zone 35N', 'EPSG', 21095, 'PROJCS[\"Arc 1960 / UTM zone 35N\",GEOGCS"
  "[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4210\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21096, 1, 'Arc 1960 / UTM zone 36N', 'EPSG', 21096, 'PROJCS[\"Arc 1960 / UTM zone 36N\",GEOGCS"
  "[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4210\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21096\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21097, 1, 'Arc 1960 / UTM zone 37N', 'EPSG', 21097, 'PROJCS[\"Arc 1960 / UTM zone 37N\",GEOGCS"
  "[\"Arc 1960\",DATUM[\"Arc 1960\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-157,-2,-299,0,0,0,0],AUTHORITY[\"EPSG\",\"6210\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4210\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21097\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21148, 1, 'Batavia / UTM zone 48S', 'EPSG', 21148, 'PROJCS[\"Batavia / UTM zone 48S\",GEOGCS[\""
  "Batavia\",DATUM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-3"
  "77.7,675.1,-52.2,0,0,0,0],AUTHORITY[\"EPSG\",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4211\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21148\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21149, 1, 'Batavia / UTM zone 49S', 'EPSG', 21149, 'PROJCS[\"Batavia / UTM zone 49S\",GEOGCS[\""
  "Batavia\",DATUM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-3"
  "77.7,675.1,-52.2,0,0,0,0],AUTHORITY[\"EPSG\",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4211\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21149\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21150, 1, 'Batavia / UTM zone 50S', 'EPSG', 21150, 'PROJCS[\"Batavia / UTM zone 50S\",GEOGCS[\""
  "Batavia\",DATUM[\"Batavia\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[-3"
  "77.7,675.1,-52.2,0,0,0,0],AUTHORITY[\"EPSG\",\"6211\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4211\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21150\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21291, 1, 'Barbados 1938 / British West Indies Grid', 'EPSG', 21291, 'PROJCS[\"Barbados 1938 /"
  " British West Indies Grid\",GEOGCS[\"Barbados 1938\",DATUM[\"Barbados 1938\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.1"
  "45,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[31.95,300.99,419.19,0,0,0,0],AUTHORITY[\"EPSG\",\"6212\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4212\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",-62,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPS"
  "G\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "21291\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21292, 1, 'Barbados 1938 / Barbados National Grid', 'EPSG', 21292, 'PROJCS[\"Barbados 1938 / B"
  "arbados National Grid\",GEOGCS[\"Barbados 1938\",DATUM[\"Barbados 1938\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,2"
  "93.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[31.95,300.99,419.19,0,0,0,0],AUTHORITY[\"EPSG\",\"6212\"]],PRIMEM[\"Gree"
  "nwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat"
  "\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4212\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"98"
  "07\"]],PARAMETER[\"Latitude of natural origin\",13.1763888888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude "
  "of natural origin\",-59.5597222222222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "99986,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",75000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS["
  "\"N\",NORTH],AUTHORITY[\"EPSG\",\"21292\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21413, 1, 'Beijing 1954 / Gauss-Kruger zone 13', 'EPSG', 21413, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 13\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21413\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21414, 1, 'Beijing 1954 / Gauss-Kruger zone 14', 'EPSG', 21414, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 14\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21414\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21415, 1, 'Beijing 1954 / Gauss-Kruger zone 15', 'EPSG', 21415, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 15\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21415\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21416, 1, 'Beijing 1954 / Gauss-Kruger zone 16', 'EPSG', 21416, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 16\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21416\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21417, 1, 'Beijing 1954 / Gauss-Kruger zone 17', 'EPSG', 21417, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 17\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21417\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21418, 1, 'Beijing 1954 / Gauss-Kruger zone 18', 'EPSG', 21418, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 18\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21418\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21419, 1, 'Beijing 1954 / Gauss-Kruger zone 19', 'EPSG', 21419, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 19\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21419\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21420, 1, 'Beijing 1954 / Gauss-Kruger zone 20', 'EPSG', 21420, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 20\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21420\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21421, 1, 'Beijing 1954 / Gauss-Kruger zone 21', 'EPSG', 21421, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 21\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21421\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21422, 1, 'Beijing 1954 / Gauss-Kruger zone 22', 'EPSG', 21422, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 22\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21422\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21423, 1, 'Beijing 1954 / Gauss-Kruger zone 23', 'EPSG', 21423, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger zone 23\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21423\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21453, 1, 'Beijing 1954 / Gauss-Kruger CM 75E', 'EPSG', 21453, 'PROJCS[\"Beijing 1954 / Gauss-"
  "Kruger CM 75E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21453\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21454, 1, 'Beijing 1954 / Gauss-Kruger CM 81E', 'EPSG', 21454, 'PROJCS[\"Beijing 1954 / Gauss-"
  "Kruger CM 81E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21454\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21455, 1, 'Beijing 1954 / Gauss-Kruger CM 87E', 'EPSG', 21455, 'PROJCS[\"Beijing 1954 / Gauss-"
  "Kruger CM 87E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21455\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21456, 1, 'Beijing 1954 / Gauss-Kruger CM 93E', 'EPSG', 21456, 'PROJCS[\"Beijing 1954 / Gauss-"
  "Kruger CM 93E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21456\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21457, 1, 'Beijing 1954 / Gauss-Kruger CM 99E', 'EPSG', 21457, 'PROJCS[\"Beijing 1954 / Gauss-"
  "Kruger CM 99E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21457\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21458, 1, 'Beijing 1954 / Gauss-Kruger CM 105E', 'EPSG', 21458, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger CM 105E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21458\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21459, 1, 'Beijing 1954 / Gauss-Kruger CM 111E', 'EPSG', 21459, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger CM 111E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21459\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21460, 1, 'Beijing 1954 / Gauss-Kruger CM 117E', 'EPSG', 21460, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger CM 117E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21460\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21461, 1, 'Beijing 1954 / Gauss-Kruger CM 123E', 'EPSG', 21461, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger CM 123E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21461\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21462, 1, 'Beijing 1954 / Gauss-Kruger CM 129E', 'EPSG', 21462, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger CM 129E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21462\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21463, 1, 'Beijing 1954 / Gauss-Kruger CM 135E', 'EPSG', 21463, 'PROJCS[\"Beijing 1954 / Gauss"
  "-Kruger CM 135E\",GEOGCS[\"Beijing 1954\",DATUM[\"Beijing 1954\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[18,-136.8,-73.7,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6214\"]],PRIMEM[\"Greenwich\",0,AU"
  "THORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AX"
  "IS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4214\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAM"
  "ETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21463\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21500, 1, 'Belge 1950 (Brussels) / Belge Lambert 50', 'EPSG', 21500, 'PROJCS[\"Belge 1950 (Bru"
  "ssels) / Belge Lambert 50\",GEOGCS[\"Belge 1950 (Brussels)\",DATUM[\"Reseau National Belge 1950 (Brussels)\",SPHEROID"
  "[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6809\"]],PRIMEM[\"Brussels\",4"
  ".3679749999999995,AUTHORITY[\"EPSG\",\"8910\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4809\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHO"
  "RITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",90,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitud"
  "e of false origin\",0,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49.8333333333333,"
  "AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",51.1666666666667,AUTHORITY[\"EPSG\",\"8"
  "824\"]],PARAMETER[\"Easting at false origin\",150000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false orig"
  "in\",5400000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"21500\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21780, 1, 'Bern 1898 (Bern) / LV03C', 'EPSG', 21780, 'PROJCS[\"Bern 1898 (Bern) / LV03C\",GEOG"
  "CS[\"Bern 1898 (Bern)\",DATUM[\"CH1903 (Bern)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\""
  "7004\"]],AUTHORITY[\"EPSG\",\"6801\"]],PRIMEM[\"Bern\",7.439583333333334,AUTHORITY[\"EPSG\",\"8907\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4801\""
  "]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETER[\"Latitude of projectio"
  "n centre\",46.9524055555556,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",0,AUTHORITY[\""
  "EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",90,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Recti"
  "fied to Skew Grid\",90,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",1,AUTHORITY[\"EPSG\","
  "\"8815\"]],PARAMETER[\"Easting at projection centre\",0,AUTHORITY[\"EPSG\",\"8816\"]],PARAMETER[\"Northing at project"
  "ion centre\",0,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X"
  "\",NORTH],AUTHORITY[\"EPSG\",\"21780\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21781, 1, 'CH1903 / LV03', 'EPSG', 21781, 'PROJCS[\"CH1903 / LV03\",GEOGCS[\"CH1903\",DATUM[\""
  "CH1903\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[660.077,13.551,369.34"
  "4,-0.804816,-0.577692,-0.952236,5.66],AUTHORITY[\"EPSG\",\"6149\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4149\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETER[\"La"
  "titude of projection centre\",46.9524055555556,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection cent"
  "re\",7.43958333333333,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",90,AUTHORITY[\"EPSG\",\"881"
  "3\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",90,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on in"
  "itial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection centre\",600000,AUTHORITY[\"EPSG\",\"8"
  "816\"]],PARAMETER[\"Northing at projection centre\",200000,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"21781\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21782, 1, 'CH1903 / LV03C-G', 'EPSG', 21782, 'PROJCS[\"CH1903 / LV03C-G\",GEOGCS[\"CH1903\",DA"
  "TUM[\"CH1903\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[660.077,13.551,"
  "369.344,-0.804816,-0.577692,-0.952236,5.66],AUTHORITY[\"EPSG\",\"6149\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4149\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETE"
  "R[\"Latitude of projection centre\",46.9524055555556,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projectio"
  "n centre\",7.43958333333333,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",90,AUTHORITY[\"EPSG\""
  ",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",90,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor"
  " on initial line\",1,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection centre\",0,AUTHORITY[\"EPSG\",\""
  "8816\"]],PARAMETER[\"Northing at projection centre\",0,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"21782\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21818, 1, 'Bogota 1975 / UTM zone 18N', 'EPSG', 21818, 'PROJCS[\"Bogota 1975 / UTM zone 18N\","
  "GEOGCS[\"Bogota 1975\",DATUM[\"Bogota 1975\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]"
  "],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4218\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"21818\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21896, 1, 'Bogota 1975 / Colombia West zone', 'EPSG', 21896, 'PROJCS[\"Bogota 1975 / Colombia "
  "West zone\",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota 1975\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG"
  "\",\"7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4218\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",4.59904722222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-77.0809166"
  "666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21"
  "896\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21897, 1, 'Bogota 1975 / Colombia Bogota zone', 'EPSG', 21897, 'PROJCS[\"Bogota 1975 / Colombi"
  "a Bogota zone\",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota 1975\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4218\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",4.59904722222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.08"
  "09166666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"21897\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21898, 1, 'Bogota 1975 / Colombia East Central zone', 'EPSG', 21898, 'PROJCS[\"Bogota 1975 / C"
  "olombia East Central zone\",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota 1975\",SPHEROID[\"International 1924\",6378388,297,"
  "AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4218\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",4.59904722222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-71.0809166666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHO"
  "RITY[\"EPSG\",\"21898\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (21899, 1, 'Bogota 1975 / Colombia East zone', 'EPSG', 21899, 'PROJCS[\"Bogota 1975 / Colombia "
  "East zone\",GEOGCS[\"Bogota 1975\",DATUM[\"Bogota 1975\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG"
  "\",\"7022\"]],TOWGS84[307,304,-318,0,0,0,0],AUTHORITY[\"EPSG\",\"6218\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4218\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",4.59904722222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-68.0809166"
  "666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",1000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"21"
  "899\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22032, 1, 'Camacupa / UTM zone 32S', 'EPSG', 22032, 'PROJCS[\"Camacupa / UTM zone 32S\",GEOGCS"
  "[\"Camacupa\",DATUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-41.057,-374.564,-226.287,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4220\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22032\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22033, 1, 'Camacupa / UTM zone 33S', 'EPSG', 22033, 'PROJCS[\"Camacupa / UTM zone 33S\",GEOGCS"
  "[\"Camacupa\",DATUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-41.057,-374.564,-226.287,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4220\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22033\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22091, 1, 'Camacupa / TM 11.30 SE', 'EPSG', 22091, 'PROJCS[\"Camacupa / TM 11.30 SE\",GEOGCS[\""
  "Camacupa\",DATUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS8"
  "4[-41.057,-374.564,-226.287,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4220\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11.5,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22091\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22092, 1, 'Camacupa / TM 12 SE', 'EPSG', 22092, 'PROJCS[\"Camacupa / TM 12 SE\",GEOGCS[\"Camac"
  "upa\",DATUM[\"Camacupa\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-41"
  ".057,-374.564,-226.287,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6220\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4220\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22092\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22171, 1, 'POSGAR 98 / Argentina 1', 'EPSG', 22171, 'PROJCS[\"POSGAR 98 / Argentina 1\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22172, 1, 'POSGAR 98 / Argentina 2', 'EPSG', 22172, 'PROJCS[\"POSGAR 98 / Argentina 2\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",250"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22172\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22173, 1, 'POSGAR 98 / Argentina 3', 'EPSG', 22173, 'PROJCS[\"POSGAR 98 / Argentina 3\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-66,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",350"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22174, 1, 'POSGAR 98 / Argentina 4', 'EPSG', 22174, 'PROJCS[\"POSGAR 98 / Argentina 4\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",450"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22174\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22175, 1, 'POSGAR 98 / Argentina 5', 'EPSG', 22175, 'PROJCS[\"POSGAR 98 / Argentina 5\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-60,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",550"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22175\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22176, 1, 'POSGAR 98 / Argentina 6', 'EPSG', 22176, 'PROJCS[\"POSGAR 98 / Argentina 6\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",650"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22176\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22177, 1, 'POSGAR 98 / Argentina 7', 'EPSG', 22177, 'PROJCS[\"POSGAR 98 / Argentina 7\",GEOGCS"
  "[\"POSGAR 98\",DATUM[\"Posiciones Geodesicas Argentinas 1998\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY["
  "\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6190\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4190\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",750"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22177\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22181, 1, 'POSGAR 94 / Argentina 1', 'EPSG', 22181, 'PROJCS[\"POSGAR 94 / Argentina 1\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22181\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22182, 1, 'POSGAR 94 / Argentina 2', 'EPSG', 22182, 'PROJCS[\"POSGAR 94 / Argentina 2\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22182\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22183, 1, 'POSGAR 94 / Argentina 3', 'EPSG', 22183, 'PROJCS[\"POSGAR 94 / Argentina 3\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-66,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22183\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22184, 1, 'POSGAR 94 / Argentina 4', 'EPSG', 22184, 'PROJCS[\"POSGAR 94 / Argentina 4\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",4500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22184\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22185, 1, 'POSGAR 94 / Argentina 5', 'EPSG', 22185, 'PROJCS[\"POSGAR 94 / Argentina 5\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-60,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22185\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22186, 1, 'POSGAR 94 / Argentina 6', 'EPSG', 22186, 'PROJCS[\"POSGAR 94 / Argentina 6\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",6500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22186\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22187, 1, 'POSGAR 94 / Argentina 7', 'EPSG', 22187, 'PROJCS[\"POSGAR 94 / Argentina 7\",GEOGCS"
  "[\"POSGAR 94\",DATUM[\"Posiciones Geodesicas Argentinas 1994\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\""
  "EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6694\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4694\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",7500"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22187\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22191, 1, 'Campo Inchauspe / Argentina 1', 'EPSG', 22191, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 1\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22192, 1, 'Campo Inchauspe / Argentina 2', 'EPSG', 22192, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 2\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22193, 1, 'Campo Inchauspe / Argentina 3', 'EPSG', 22193, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 3\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-66,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22194, 1, 'Campo Inchauspe / Argentina 4', 'EPSG', 22194, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 4\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22194\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22195, 1, 'Campo Inchauspe / Argentina 5', 'EPSG', 22195, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 5\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-60,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22196, 1, 'Campo Inchauspe / Argentina 6', 'EPSG', 22196, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 6\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22196\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22197, 1, 'Campo Inchauspe / Argentina 7', 'EPSG', 22197, 'PROJCS[\"Campo Inchauspe / Argentin"
  "a 7\",GEOGCS[\"Campo Inchauspe\",DATUM[\"Campo Inchauspe\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EP"
  "SG\",\"7022\"]],TOWGS84[-148,136,90,0,0,0,0],AUTHORITY[\"EPSG\",\"6221\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4221\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",-90,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"22197\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22234, 1, 'Cape / UTM zone 34S', 'EPSG', 22234, 'PROJCS[\"Cape / UTM zone 34S\",GEOGCS[\"Cape\""
  ",DATUM[\"Cape\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-10"
  "8,-292,0,0,0,0],AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22234\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22235, 1, 'Cape / UTM zone 35S', 'EPSG', 22235, 'PROJCS[\"Cape / UTM zone 35S\",GEOGCS[\"Cape\""
  ",DATUM[\"Cape\",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-10"
  "8,-292,0,0,0,0],AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22235\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22275, 1, 'Cape / Lo15', 'EPSG', 22275, 'PROJCS[\"Cape / Lo15\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22275\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22277, 1, 'Cape / Lo17', 'EPSG', 22277, 'PROJCS[\"Cape / Lo17\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22277\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22279, 1, 'Cape / Lo19', 'EPSG', 22279, 'PROJCS[\"Cape / Lo19\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22279\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22281, 1, 'Cape / Lo21', 'EPSG', 22281, 'PROJCS[\"Cape / Lo21\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22281\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22283, 1, 'Cape / Lo23', 'EPSG', 22283, 'PROJCS[\"Cape / Lo23\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22283\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22285, 1, 'Cape / Lo25', 'EPSG', 22285, 'PROJCS[\"Cape / Lo25\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22285\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22287, 1, 'Cape / Lo27', 'EPSG', 22287, 'PROJCS[\"Cape / Lo27\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22287\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22289, 1, 'Cape / Lo29', 'EPSG', 22289, 'PROJCS[\"Cape / Lo29\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",29,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22289\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22291, 1, 'Cape / Lo31', 'EPSG', 22291, 'PROJCS[\"Cape / Lo31\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22291\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22293, 1, 'Cape / Lo33', 'EPSG', 22293, 'PROJCS[\"Cape / Lo33\",GEOGCS[\"Cape\",DATUM[\"Cape\""
  ",SPHEROID[\"Clarke 1880 (Arc)\",6378249.145,293.4663077,AUTHORITY[\"EPSG\",\"7013\"]],TOWGS84[-136,-108,-292,0,0,0,0]"
  ",AUTHORITY[\"EPSG\",\"6222\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4222\"]],PROJECTION[\"T"
  "ransverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\"Latitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"22293\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22300, 1, 'Carthage (Paris) / Tunisia Mining Grid', 'EPSG', 22300, 'PROJCS[\"Carthage (Paris) "
  "/ Tunisia Mining Grid\",GEOGCS[\"Carthage (Paris)\",DATUM[\"Carthage (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249"
  ".2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6816\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY"
  "[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4816\"]],PROJECTION[\"Tunisia Mining Grid\",AUTHORITY[\"EPSG\",\"9816\"]],PARAMETER[\"Lat"
  "itude of false origin\",36.5964,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",7.83445,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Easting at false origin\",270,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing a"
  "t false origin\",360,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"kilometre\",1000,AUTHORITY[\"EPSG\",\"9036\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"22300\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22332, 1, 'Carthage / UTM zone 32N', 'EPSG', 22332, 'PROJCS[\"Carthage / UTM zone 32N\",GEOGCS"
  "[\"Carthage\",DATUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011"
  "\"]],TOWGS84[-263,6,431,0,0,0,0],AUTHORITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4223\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22391, 1, 'Carthage / Nord Tunisie', 'EPSG', 22391, 'PROJCS[\"Carthage / Nord Tunisie\",GEOGCS"
  "[\"Carthage\",DATUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011"
  "\"]],TOWGS84[-263,6,431,0,0,0,0],AUTHORITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4223\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of "
  "natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9.9,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999625544,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"22391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22392, 1, 'Carthage / Sud Tunisie', 'EPSG', 22392, 'PROJCS[\"Carthage / Sud Tunisie\",GEOGCS[\""
  "Carthage\",DATUM[\"Carthage\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]"
  "],TOWGS84[-263,6,431,0,0,0,0],AUTHORITY[\"EPSG\",\"6223\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4223\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",33.3,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9.9,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.999625769,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"22392\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22521, 1, 'Corrego Alegre 1970-72 / UTM zone 21S', 'EPSG', 22521, 'PROJCS[\"Corrego Alegre 197"
  "0-72 / UTM zone 21S\",GEOGCS[\"Corrego Alegre 1970-72\",DATUM[\"Corrego Alegre 1970-72\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.05,168.28,-3.82,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4225\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"22521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22522, 1, 'Corrego Alegre 1970-72 / UTM zone 22S', 'EPSG', 22522, 'PROJCS[\"Corrego Alegre 197"
  "0-72 / UTM zone 22S\",GEOGCS[\"Corrego Alegre 1970-72\",DATUM[\"Corrego Alegre 1970-72\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.05,168.28,-3.82,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4225\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"22522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22523, 1, 'Corrego Alegre 1970-72 / UTM zone 23S', 'EPSG', 22523, 'PROJCS[\"Corrego Alegre 197"
  "0-72 / UTM zone 23S\",GEOGCS[\"Corrego Alegre 1970-72\",DATUM[\"Corrego Alegre 1970-72\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.05,168.28,-3.82,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4225\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"22523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22524, 1, 'Corrego Alegre 1970-72 / UTM zone 24S', 'EPSG', 22524, 'PROJCS[\"Corrego Alegre 197"
  "0-72 / UTM zone 24S\",GEOGCS[\"Corrego Alegre 1970-72\",DATUM[\"Corrego Alegre 1970-72\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.05,168.28,-3.82,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4225\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"22524\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22525, 1, 'Corrego Alegre 1970-72 / UTM zone 25S', 'EPSG', 22525, 'PROJCS[\"Corrego Alegre 197"
  "0-72 / UTM zone 25S\",GEOGCS[\"Corrego Alegre 1970-72\",DATUM[\"Corrego Alegre 1970-72\",SPHEROID[\"International 192"
  "4\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-206.05,168.28,-3.82,0,0,0,0],AUTHORITY[\"EPSG\",\"6225\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4225\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of na"
  "tural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"22525\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22700, 1, 'Deir ez Zor / Levant Zone', 'EPSG', 22700, 'PROJCS[\"Deir ez Zor / Levant Zone\",GE"
  "OGCS[\"Deir ez Zor\",DATUM[\"Deir ez Zor\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPS"
  "G\",\"7011\"]],TOWGS84[-83.58,-397.54,458.78,-17.595,-2.847,4.256,3.225],AUTHORITY[\"EPSG\",\"6227\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4227\"]],PROJECTION[\"Lambert Conic Near-Conformal\",AUTHORITY[\"EPSG"
  "\",\"9817\"]],PARAMETER[\"Latitude of natural origin\",34.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",37.35,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996256,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300"
  "000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],A"
  "UTHORITY[\"EPSG\",\"22700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22770, 1, 'Deir ez Zor / Syria Lambert', 'EPSG', 22770, 'PROJCS[\"Deir ez Zor / Syria Lambert\""
  ",GEOGCS[\"Deir ez Zor\",DATUM[\"Deir ez Zor\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\""
  "EPSG\",\"7011\"]],TOWGS84[-83.58,-397.54,458.78,-17.595,-2.847,4.256,3.225],AUTHORITY[\"EPSG\",\"6227\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4227\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\""
  "EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",34.65,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude "
  "of natural origin\",37.35,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996256,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"22770\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22780, 1, 'Deir ez Zor / Levant Stereographic', 'EPSG', 22780, 'PROJCS[\"Deir ez Zor / Levant "
  "Stereographic\",GEOGCS[\"Deir ez Zor\",DATUM[\"Deir ez Zor\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602129362"
  "69,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-83.58,-397.54,458.78,-17.595,-2.847,4.256,3.225],AUTHORITY[\"EPSG\",\"6227\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4227\"]],PROJECTION[\"Oblique Stereographic\",AUTHO"
  "RITY[\"EPSG\",\"9809\"]],PARAMETER[\"Latitude of natural origin\",34.2,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",39.15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9995341,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"22780\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22991, 1, 'Egypt 1907 / Blue Belt', 'EPSG', 22991, 'PROJCS[\"Egypt 1907 / Blue Belt\",GEOGCS[\""
  "Egypt 1907\",DATUM[\"Egypt 1907\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-127."
  "535,113.495,-12.7,-1.603747,0.153612,5.364408,5.33745],AUTHORITY[\"EPSG\",\"6229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY"
  "[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4229\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",35,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1100000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22992, 1, 'Egypt 1907 / Red Belt', 'EPSG', 22992, 'PROJCS[\"Egypt 1907 / Red Belt\",GEOGCS[\"E"
  "gypt 1907\",DATUM[\"Egypt 1907\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84[-127.5"
  "35,113.495,-12.7,-1.603747,0.153612,5.364408,5.33745],AUTHORITY[\"EPSG\",\"6229\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4229\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",615000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",810000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22993, 1, 'Egypt 1907 / Purple Belt', 'EPSG', 22993, 'PROJCS[\"Egypt 1907 / Purple Belt\",GEOG"
  "CS[\"Egypt 1907\",DATUM[\"Egypt 1907\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],TOWGS84["
  "-127.535,113.495,-12.7,-1.603747,0.153612,5.364408,5.33745],AUTHORITY[\"EPSG\",\"6229\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4229\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"22993\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (22994, 1, 'Egypt 1907 / Extended Purple Belt', 'EPSG', 22994, 'PROJCS[\"Egypt 1907 / Extended "
  "Purple Belt\",GEOGCS[\"Egypt 1907\",DATUM[\"Egypt 1907\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\""
  "7020\"]],TOWGS84[-127.535,113.495,-12.7,-1.603747,0.153612,5.364408,5.33745],AUTHORITY[\"EPSG\",\"6229\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4229\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1200000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"22994\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23028, 1, 'ED50 / UTM zone 28N', 'EPSG', 23028, 'PROJCS[\"ED50 / UTM zone 28N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23029, 1, 'ED50 / UTM zone 29N', 'EPSG', 23029, 'PROJCS[\"ED50 / UTM zone 29N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23029\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23030, 1, 'ED50 / UTM zone 30N', 'EPSG', 23030, 'PROJCS[\"ED50 / UTM zone 30N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23030\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23031, 1, 'ED50 / UTM zone 31N', 'EPSG', 23031, 'PROJCS[\"ED50 / UTM zone 31N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23031\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23032, 1, 'ED50 / UTM zone 32N', 'EPSG', 23032, 'PROJCS[\"ED50 / UTM zone 32N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23032\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23033, 1, 'ED50 / UTM zone 33N', 'EPSG', 23033, 'PROJCS[\"ED50 / UTM zone 33N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23033\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23034, 1, 'ED50 / UTM zone 34N', 'EPSG', 23034, 'PROJCS[\"ED50 / UTM zone 34N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23034\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23035, 1, 'ED50 / UTM zone 35N', 'EPSG', 23035, 'PROJCS[\"ED50 / UTM zone 35N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23036, 1, 'ED50 / UTM zone 36N', 'EPSG', 23036, 'PROJCS[\"ED50 / UTM zone 36N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23036\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23037, 1, 'ED50 / UTM zone 37N', 'EPSG', 23037, 'PROJCS[\"ED50 / UTM zone 37N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23037\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23038, 1, 'ED50 / UTM zone 38N', 'EPSG', 23038, 'PROJCS[\"ED50 / UTM zone 38N\",GEOGCS[\"ED50\""
  ",DATUM[\"European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157"
  ".89,-17.16,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23038\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23090, 1, 'ED50 / TM 0 N', 'EPSG', 23090, 'PROJCS[\"ED50 / TM 0 N\",GEOGCS[\"ED50\",DATUM[\"Eu"
  "ropean Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.16,"
  "-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARA"
  "METER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23090\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23095, 1, 'ED50 / TM 5 NE', 'EPSG', 23095, 'PROJCS[\"ED50 / TM 5 NE\",GEOGCS[\"ED50\",DATUM[\""
  "European Datum 1950\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-157.89,-17.1"
  "6,-78.41,2.118,2.697,-1.434,-5.38],AUTHORITY[\"EPSG\",\"6230\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4230\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",5,AUTHORITY[\"EPSG\",\"8802\"]],PA"
  "RAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AU"
  "THORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23095\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23239, 1, 'Fahud / UTM zone 39N', 'EPSG', 23239, 'PROJCS[\"Fahud / UTM zone 39N\",GEOGCS[\"Fah"
  "ud\",DATUM[\"Fahud\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-333.10"
  "2,-11.02,230.69,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6232\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4232\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23239\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23240, 1, 'Fahud / UTM zone 40N', 'EPSG', 23240, 'PROJCS[\"Fahud / UTM zone 40N\",GEOGCS[\"Fah"
  "ud\",DATUM[\"Fahud\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-333.10"
  "2,-11.02,230.69,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6232\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4232\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23240\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23700, 1, 'HD72 / EOV', 'EPSG', 23700, 'PROJCS[\"HD72 / EOV\",GEOGCS[\"HD72\",DATUM[\"Hungaria"
  "n Datum 1972\",SPHEROID[\"GRS 1967\",6378160,298.247167427,AUTHORITY[\"EPSG\",\"7036\"]],TOWGS84[52.684,-71.194,-13.9"
  "75,-0.312,-0.1063,-0.3729,1.0191],AUTHORITY[\"EPSG\",\"6237\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],"
  "UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4237\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\",\"9815\"]],PARAMETER[\"Latitu"
  "de of projection centre\",47.1443937222222,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\""
  ",19.0485717777778,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",90,AUTHORITY[\"EPSG\",\"8813\"]"
  "],PARAMETER[\"Angle from Rectified to Skew Grid\",90,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initia"
  "l line\",0.99993,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection centre\",650000,AUTHORITY[\"EPSG\",\""
  "8816\"]],PARAMETER[\"Northing at projection centre\",200000,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"Y\",EAST],AXIS[\"X\",NORTH],AUTHORITY[\"EPSG\",\"23700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23830, 1, 'DGN95 / Indonesia TM-3 zone 46.2', 'EPSG', 23830, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 46.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",94.5,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23830\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23831, 1, 'DGN95 / Indonesia TM-3 zone 47.1', 'EPSG', 23831, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 47.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",97.5,AUTHORITY[\"EPSG\","
  "\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23831\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23832, 1, 'DGN95 / Indonesia TM-3 zone 47.2', 'EPSG', 23832, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 47.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",100.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23832\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23833, 1, 'DGN95 / Indonesia TM-3 zone 48.1', 'EPSG', 23833, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 48.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",103.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23833\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23834, 1, 'DGN95 / Indonesia TM-3 zone 48.2', 'EPSG', 23834, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 48.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",106.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23834\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23835, 1, 'DGN95 / Indonesia TM-3 zone 49.1', 'EPSG', 23835, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 49.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",109.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23835\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23836, 1, 'DGN95 / Indonesia TM-3 zone 49.2', 'EPSG', 23836, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 49.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",112.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23836\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23837, 1, 'DGN95 / Indonesia TM-3 zone 50.1', 'EPSG', 23837, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 50.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",115.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23838, 1, 'DGN95 / Indonesia TM-3 zone 50.2', 'EPSG', 23838, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 50.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",118.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23838\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23839, 1, 'DGN95 / Indonesia TM-3 zone 51.1', 'EPSG', 23839, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 51.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",121.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23840, 1, 'DGN95 / Indonesia TM-3 zone 51.2', 'EPSG', 23840, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 51.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",124.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23840\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23841, 1, 'DGN95 / Indonesia TM-3 zone 52.1', 'EPSG', 23841, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 52.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23841\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23842, 1, 'DGN95 / Indonesia TM-3 zone 52.2', 'EPSG', 23842, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 52.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",130.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23842\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23843, 1, 'DGN95 / Indonesia TM-3 zone 53.1', 'EPSG', 23843, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 53.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",133.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23843\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23844, 1, 'DGN95 / Indonesia TM-3 zone 53.2', 'EPSG', 23844, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 53.2\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",136.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23844\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23845, 1, 'DGN95 / Indonesia TM-3 zone 54.1', 'EPSG', 23845, 'PROJCS[\"DGN95 / Indonesia TM-3 "
  "zone 54.1\",GEOGCS[\"DGN95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORIT"
  "Y[\"EPSG\",\"7030\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4755\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",139.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"23845\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23846, 1, 'ID74 / UTM zone 46N', 'EPSG', 23846, 'PROJCS[\"ID74 / UTM zone 46N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23846\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23847, 1, 'ID74 / UTM zone 47N', 'EPSG', 23847, 'PROJCS[\"ID74 / UTM zone 47N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23847\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23848, 1, 'ID74 / UTM zone 48N', 'EPSG', 23848, 'PROJCS[\"ID74 / UTM zone 48N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23848\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23849, 1, 'ID74 / UTM zone 49N', 'EPSG', 23849, 'PROJCS[\"ID74 / UTM zone 49N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23849\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23850, 1, 'ID74 / UTM zone 50N', 'EPSG', 23850, 'PROJCS[\"ID74 / UTM zone 50N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23850\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23851, 1, 'ID74 / UTM zone 51N', 'EPSG', 23851, 'PROJCS[\"ID74 / UTM zone 51N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23851\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23852, 1, 'ID74 / UTM zone 52N', 'EPSG', 23852, 'PROJCS[\"ID74 / UTM zone 52N\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23852\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23866, 1, 'DGN95 / UTM zone 46N', 'EPSG', 23866, 'PROJCS[\"DGN95 / UTM zone 46N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23866\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23867, 1, 'DGN95 / UTM zone 47N', 'EPSG', 23867, 'PROJCS[\"DGN95 / UTM zone 47N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23867\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23868, 1, 'DGN95 / UTM zone 48N', 'EPSG', 23868, 'PROJCS[\"DGN95 / UTM zone 48N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23868\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23869, 1, 'DGN95 / UTM zone 49N', 'EPSG', 23869, 'PROJCS[\"DGN95 / UTM zone 49N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23869\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23870, 1, 'DGN95 / UTM zone 50N', 'EPSG', 23870, 'PROJCS[\"DGN95 / UTM zone 50N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23870\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23871, 1, 'DGN95 / UTM zone 51N', 'EPSG', 23871, 'PROJCS[\"DGN95 / UTM zone 51N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23871\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23872, 1, 'DGN95 / UTM zone 52N', 'EPSG', 23872, 'PROJCS[\"DGN95 / UTM zone 52N\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23872\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23877, 1, 'DGN95 / UTM zone 47S', 'EPSG', 23877, 'PROJCS[\"DGN95 / UTM zone 47S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23877\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23878, 1, 'DGN95 / UTM zone 48S', 'EPSG', 23878, 'PROJCS[\"DGN95 / UTM zone 48S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23878\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23879, 1, 'DGN95 / UTM zone 49S', 'EPSG', 23879, 'PROJCS[\"DGN95 / UTM zone 49S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23879\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23880, 1, 'DGN95 / UTM zone 50S', 'EPSG', 23880, 'PROJCS[\"DGN95 / UTM zone 50S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23880\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23881, 1, 'DGN95 / UTM zone 51S', 'EPSG', 23881, 'PROJCS[\"DGN95 / UTM zone 51S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23881\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23882, 1, 'DGN95 / UTM zone 52S', 'EPSG', 23882, 'PROJCS[\"DGN95 / UTM zone 52S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23882\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23883, 1, 'DGN95 / UTM zone 53S', 'EPSG', 23883, 'PROJCS[\"DGN95 / UTM zone 53S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23883\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23884, 1, 'DGN95 / UTM zone 54S', 'EPSG', 23884, 'PROJCS[\"DGN95 / UTM zone 54S\",GEOGCS[\"DGN"
  "95\",DATUM[\"Datum Geodesi Nasional 1995\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TO"
  "WGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6755\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"475"
  "5\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23884\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23887, 1, 'ID74 / UTM zone 47S', 'EPSG', 23887, 'PROJCS[\"ID74 / UTM zone 47S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23887\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23888, 1, 'ID74 / UTM zone 48S', 'EPSG', 23888, 'PROJCS[\"ID74 / UTM zone 48S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23888\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23889, 1, 'ID74 / UTM zone 49S', 'EPSG', 23889, 'PROJCS[\"ID74 / UTM zone 49S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23889\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23890, 1, 'ID74 / UTM zone 50S', 'EPSG', 23890, 'PROJCS[\"ID74 / UTM zone 50S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23890\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23891, 1, 'ID74 / UTM zone 51S', 'EPSG', 23891, 'PROJCS[\"ID74 / UTM zone 51S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23891\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23892, 1, 'ID74 / UTM zone 52S', 'EPSG', 23892, 'PROJCS[\"ID74 / UTM zone 52S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23892\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23893, 1, 'ID74 / UTM zone 53S', 'EPSG', 23893, 'PROJCS[\"ID74 / UTM zone 53S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23893\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23894, 1, 'ID74 / UTM zone 54S', 'EPSG', 23894, 'PROJCS[\"ID74 / UTM zone 54S\",GEOGCS[\"ID74\""
  ",DATUM[\"Indonesian Datum 1974\",SPHEROID[\"Indonesian National Spheroid\",6378160,298.247,AUTHORITY[\"EPSG\",\"7021\""
  "]],TOWGS84[-1.977,-13.06,-9.993,0.364,0.254,0.689,-1.037],AUTHORITY[\"EPSG\",\"6238\"]],PRIMEM[\"Greenwich\",0,AUTHOR"
  "ITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4238\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORIT"
  "Y[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23894\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23946, 1, 'Indian 1954 / UTM zone 46N', 'EPSG', 23946, 'PROJCS[\"Indian 1954 / UTM zone 46N\","
  "GEOGCS[\"Indian 1954\",DATUM[\"Indian 1954\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORI"
  "TY[\"EPSG\",\"7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTHORITY[\"EPSG\",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4239\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23946\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23947, 1, 'Indian 1954 / UTM zone 47N', 'EPSG', 23947, 'PROJCS[\"Indian 1954 / UTM zone 47N\","
  "GEOGCS[\"Indian 1954\",DATUM[\"Indian 1954\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORI"
  "TY[\"EPSG\",\"7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTHORITY[\"EPSG\",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4239\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23947\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (23948, 1, 'Indian 1954 / UTM zone 48N', 'EPSG', 23948, 'PROJCS[\"Indian 1954 / UTM zone 48N\","
  "GEOGCS[\"Indian 1954\",DATUM[\"Indian 1954\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORI"
  "TY[\"EPSG\",\"7015\"]],TOWGS84[217,823,299,0,0,0,0],AUTHORITY[\"EPSG\",\"6239\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4239\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"23948\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24047, 1, 'Indian 1975 / UTM zone 47N', 'EPSG', 24047, 'PROJCS[\"Indian 1975 / UTM zone 47N\","
  "GEOGCS[\"Indian 1975\",DATUM[\"Indian 1975\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORI"
  "TY[\"EPSG\",\"7015\"]],TOWGS84[210,814,289,0,0,0,0],AUTHORITY[\"EPSG\",\"6240\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4240\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24047\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24048, 1, 'Indian 1975 / UTM zone 48N', 'EPSG', 24048, 'PROJCS[\"Indian 1975 / UTM zone 48N\","
  "GEOGCS[\"Indian 1975\",DATUM[\"Indian 1975\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300.8017,AUTHORI"
  "TY[\"EPSG\",\"7015\"]],TOWGS84[210,814,289,0,0,0,0],AUTHORITY[\"EPSG\",\"6240\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4240\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24048\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24100, 1, 'Jamaica 1875 / Jamaica (Old Grid)', 'EPSG', 24100, 'PROJCS[\"Jamaica 1875 / Jamaica"
  " (Old Grid)\",GEOGCS[\"Jamaica 1875\",DATUM[\"Jamaica 1875\",SPHEROID[\"Clarke 1880\",6378249.144808011,293.466307655"
  "62986,AUTHORITY[\"EPSG\",\"7034\"]],AUTHORITY[\"EPSG\",\"6241\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4241\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude "
  "of natural origin\",18,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-77,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "550000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",400000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Clarke\\"
  "'s foot\",0.3047972654,AUTHORITY[\"EPSG\",\"9005\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24100\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24200, 1, 'JAD69 / Jamaica National Grid', 'EPSG', 24200, 'PROJCS[\"JAD69 / Jamaica National G"
  "rid\",GEOGCS[\"JAD69\",DATUM[\"Jamaica 1969\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-33.722,153.789,94.959,-8.581,-4.478,4.54,8.95],AUTHORITY[\"EPSG\",\"6242\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4242\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",18,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",-77,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",150000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"24200\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24305, 1, 'Kalianpur 1937 / UTM zone 45N', 'EPSG', 24305, 'PROJCS[\"Kalianpur 1937 / UTM zone "
  "45N\",GEOGCS[\"Kalianpur 1937\",DATUM[\"Kalianpur 1937\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300."
  "8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[214,804,268,0,0,0,0],AUTHORITY[\"EPSG\",\"6144\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4144\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24305\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24306, 1, 'Kalianpur 1937 / UTM zone 46N', 'EPSG', 24306, 'PROJCS[\"Kalianpur 1937 / UTM zone "
  "46N\",GEOGCS[\"Kalianpur 1937\",DATUM[\"Kalianpur 1937\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,300."
  "8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[214,804,268,0,0,0,0],AUTHORITY[\"EPSG\",\"6144\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4144\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24306\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24311, 1, 'Kalianpur 1962 / UTM zone 41N', 'EPSG', 24311, 'PROJCS[\"Kalianpur 1962 / UTM zone "
  "41N\",GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur 1962\",SPHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,300."
  "8017255,AUTHORITY[\"EPSG\",\"7044\"]],TOWGS84[230.25,632.76,161.03,-1.114,1.115,1.212,12.584],AUTHORITY[\"EPSG\",\"61"
  "45\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4145\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"24311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24312, 1, 'Kalianpur 1962 / UTM zone 42N', 'EPSG', 24312, 'PROJCS[\"Kalianpur 1962 / UTM zone "
  "42N\",GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur 1962\",SPHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,300."
  "8017255,AUTHORITY[\"EPSG\",\"7044\"]],TOWGS84[230.25,632.76,161.03,-1.114,1.115,1.212,12.584],AUTHORITY[\"EPSG\",\"61"
  "45\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4145\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"24312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24313, 1, 'Kalianpur 1962 / UTM zone 43N', 'EPSG', 24313, 'PROJCS[\"Kalianpur 1962 / UTM zone "
  "43N\",GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur 1962\",SPHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,300."
  "8017255,AUTHORITY[\"EPSG\",\"7044\"]],TOWGS84[230.25,632.76,161.03,-1.114,1.115,1.212,12.584],AUTHORITY[\"EPSG\",\"61"
  "45\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4145\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"24313\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24342, 1, 'Kalianpur 1975 / UTM zone 42N', 'EPSG', 24342, 'PROJCS[\"Kalianpur 1975 / UTM zone "
  "42N\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24342\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24343, 1, 'Kalianpur 1975 / UTM zone 43N', 'EPSG', 24343, 'PROJCS[\"Kalianpur 1975 / UTM zone "
  "43N\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24343\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24344, 1, 'Kalianpur 1975 / UTM zone 44N', 'EPSG', 24344, 'PROJCS[\"Kalianpur 1975 / UTM zone "
  "44N\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24344\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24345, 1, 'Kalianpur 1975 / UTM zone 45N', 'EPSG', 24345, 'PROJCS[\"Kalianpur 1975 / UTM zone "
  "45N\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24345\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24346, 1, 'Kalianpur 1975 / UTM zone 46N', 'EPSG', 24346, 'PROJCS[\"Kalianpur 1975 / UTM zone "
  "46N\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24346\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24347, 1, 'Kalianpur 1975 / UTM zone 47N', 'EPSG', 24347, 'PROJCS[\"Kalianpur 1975 / UTM zone "
  "47N\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24347\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24370, 1, 'Kalianpur 1880 / India zone 0', 'EPSG', 24370, 'PROJCS[\"Kalianpur 1880 / India zon"
  "e 0\",GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",SPHEROID[\"Everest (1830 Definition)\",6377299.36559538,300."
  "80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4243\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",39.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",68,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99846154,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",2355500,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2590000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"Indian yard\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],AXIS[\"E\",EAST],AXIS[\"N\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"24370\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24371, 1, 'Kalianpur 1880 / India zone I', 'EPSG', 24371, 'PROJCS[\"Kalianpur 1880 / India zon"
  "e I\",GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",SPHEROID[\"Everest (1830 Definition)\",6377299.36559538,300."
  "80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4243\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",68,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",3000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"Indian yard\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],AXIS[\"E\",EAST],AXIS[\"N\",N"
  "ORTH],AUTHORITY[\"EPSG\",\"24371\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24372, 1, 'Kalianpur 1880 / India zone IIa', 'EPSG', 24372, 'PROJCS[\"Kalianpur 1880 / India z"
  "one IIa\",GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",SPHEROID[\"Everest (1830 Definition)\",6377299.36559538,"
  "300.80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4243\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",74,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",3000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"Indian yard\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"24372\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24373, 1, 'Kalianpur 1880 / India zone IIIa', 'EPSG', 24373, 'PROJCS[\"Kalianpur 1880 / India "
  "zone IIIa\",GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",SPHEROID[\"Everest (1830 Definition)\",6377299.3655953"
  "8,300.80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4243\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",19,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",80,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",3000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"Indian yard\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],A"
  "UTHORITY[\"EPSG\",\"24373\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24374, 1, 'Kalianpur 1880 / India zone IVa', 'EPSG', 24374, 'PROJCS[\"Kalianpur 1880 / India z"
  "one IVa\",GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",SPHEROID[\"Everest (1830 Definition)\",6377299.36559538,"
  "300.80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4243\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",80,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",3000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"Indian yard\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"24374\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24375, 1, 'Kalianpur 1937 / India zone IIb', 'EPSG', 24375, 'PROJCS[\"Kalianpur 1937 / India z"
  "one IIb\",GEOGCS[\"Kalianpur 1937\",DATUM[\"Kalianpur 1937\",SPHEROID[\"Everest 1830 (1937 Adjustment)\",6377276.345,"
  "300.8017,AUTHORITY[\"EPSG\",\"7015\"]],TOWGS84[214,804,268,0,0,0,0],AUTHORITY[\"EPSG\",\"6144\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4144\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural o"
  "rigin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPSG\",\""
  "8805\"]],PARAMETER[\"False easting\",2743185.69,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",914395.23,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHO"
  "RITY[\"EPSG\",\"24375\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24376, 1, 'Kalianpur 1962 / India zone I', 'EPSG', 24376, 'PROJCS[\"Kalianpur 1962 / India zon"
  "e I\",GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur 1962\",SPHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,300."
  "8017255,AUTHORITY[\"EPSG\",\"7044\"]],TOWGS84[230.25,632.76,161.03,-1.114,1.115,1.212,12.584],AUTHORITY[\"EPSG\",\"61"
  "45\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4145\"]],PROJECTION[\"Lambert Conic Conformal (1"
  "SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",68,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.99878641,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2743196.4,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",914398.8,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24376\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24377, 1, 'Kalianpur 1962 / India zone IIa', 'EPSG', 24377, 'PROJCS[\"Kalianpur 1962 / India z"
  "one IIa\",GEOGCS[\"Kalianpur 1962\",DATUM[\"Kalianpur 1962\",SPHEROID[\"Everest 1830 (1962 Definition)\",6377301.243,"
  "300.8017255,AUTHORITY[\"EPSG\",\"7044\"]],TOWGS84[230.25,632.76,161.03,-1.114,1.115,1.212,12.584],AUTHORITY[\"EPSG\","
  "\"6145\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG"
  "\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4145\"]],PROJECTION[\"Lambert Conic Conform"
  "al (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",74,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.99878641,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2743196.4,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",914398.8,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24377\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24378, 1, 'Kalianpur 1975 / India zone I', 'EPSG', 24378, 'PROJCS[\"Kalianpur 1975 / India zon"
  "e I\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,300."
  "8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",32.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",68,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2743195.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northin"
  "g\",914398.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"24378\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24379, 1, 'Kalianpur 1975 / India zone IIa', 'EPSG', 24379, 'PROJCS[\"Kalianpur 1975 / India z"
  "one IIa\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,"
  "300.8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",74,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",2743195.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",91439"
  "8.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],A"
  "UTHORITY[\"EPSG\",\"24379\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24380, 1, 'Kalianpur 1975 / India zone IIb', 'EPSG', 24380, 'PROJCS[\"Kalianpur 1975 / India z"
  "one IIb\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,"
  "300.8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",2743195.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",91439"
  "8.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],A"
  "UTHORITY[\"EPSG\",\"24380\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24381, 1, 'Kalianpur 1975 / India zone IIIa', 'EPSG', 24381, 'PROJCS[\"Kalianpur 1975 / India "
  "zone IIIa\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.15"
  "1,300.8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPS"
  "G\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",19,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",80,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",2743195.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",9143"
  "98.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],"
  "AUTHORITY[\"EPSG\",\"24381\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24382, 1, 'Kalianpur 1880 / India zone IIb', 'EPSG', 24382, 'PROJCS[\"Kalianpur 1880 / India z"
  "one IIb\",GEOGCS[\"Kalianpur 1880\",DATUM[\"Kalianpur 1880\",SPHEROID[\"Everest (1830 Definition)\",6377299.36559538,"
  "300.80172554336184,AUTHORITY[\"EPSG\",\"7042\"]],AUTHORITY[\"EPSG\",\"6243\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPS"
  "G\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EA"
  "ST],AUTHORITY[\"EPSG\",\"4243\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETE"
  "R[\"Latitude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",90,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",3000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1000000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"Indian yard\",0.9143985307444408,AUTHORITY[\"EPSG\",\"9084\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"24382\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24383, 1, 'Kalianpur 1975 / India zone IVa', 'EPSG', 24383, 'PROJCS[\"Kalianpur 1975 / India z"
  "one IVa\",GEOGCS[\"Kalianpur 1975\",DATUM[\"Kalianpur 1975\",SPHEROID[\"Everest 1830 (1975 Definition)\",6377299.151,"
  "300.8017255,AUTHORITY[\"EPSG\",\"7045\"]],TOWGS84[295,736,257,0,0,0,0],AUTHORITY[\"EPSG\",\"6146\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4146\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG"
  "\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",12,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",80,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99878641,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",2743195.5,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",91439"
  "8.5,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],A"
  "UTHORITY[\"EPSG\",\"24383\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24500, 1, 'Kertau 1968 / Singapore Grid', 'EPSG', 24500, 'PROJCS[\"Kertau 1968 / Singapore Gri"
  "d\",GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\""
  "EPSG\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of na"
  "tural origin\",1.28764666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",103.853002222"
  "222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",30000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",30000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\""
  "N\",NORTH],AUTHORITY[\"EPSG\",\"24500\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24547, 1, 'Kertau 1968 / UTM zone 47N', 'EPSG', 24547, 'PROJCS[\"Kertau 1968 / UTM zone 47N\","
  "GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG"
  "\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24547\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24548, 1, 'Kertau 1968 / UTM zone 48N', 'EPSG', 24548, 'PROJCS[\"Kertau 1968 / UTM zone 48N\","
  "GEOGCS[\"Kertau 1968\",DATUM[\"Kertau 1968\",SPHEROID[\"Everest 1830 Modified\",6377304.063,300.8017,AUTHORITY[\"EPSG"
  "\",\"7018\"]],TOWGS84[-11,851,5,0,0,0,0],AUTHORITY[\"EPSG\",\"6245\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4245\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24548\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24600, 1, 'KOC Lambert', 'EPSG', 24600, 'PROJCS[\"KOC Lambert\",GEOGCS[\"KOC\",DATUM[\"Kuwait "
  "Oil Company\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-294.7,-200.1,"
  "525.5,0,0,0,0],AUTHORITY[\"EPSG\",\"6246\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0"
  "17453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4246\"]],"
  "PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",3"
  "2.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9987864078,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1166200,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24718, 1, 'La Canoa / UTM zone 18N', 'EPSG', 24718, 'PROJCS[\"La Canoa / UTM zone 18N\",GEOGCS"
  "[\"La Canoa\",DATUM[\"La Canoa\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-2"
  "70.933,115.599,-360.226,5.266,1.238,-2.381,-5.109],AUTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4247\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24718\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24719, 1, 'La Canoa / UTM zone 19N', 'EPSG', 24719, 'PROJCS[\"La Canoa / UTM zone 19N\",GEOGCS"
  "[\"La Canoa\",DATUM[\"La Canoa\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-2"
  "70.933,115.599,-360.226,5.266,1.238,-2.381,-5.109],AUTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4247\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24719\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24720, 1, 'La Canoa / UTM zone 20N', 'EPSG', 24720, 'PROJCS[\"La Canoa / UTM zone 20N\",GEOGCS"
  "[\"La Canoa\",DATUM[\"La Canoa\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-2"
  "70.933,115.599,-360.226,5.266,1.238,-2.381,-5.109],AUTHORITY[\"EPSG\",\"6247\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4247\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24720\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24817, 1, 'PSAD56 / UTM zone 17N', 'EPSG', 24817, 'PROJCS[\"PSAD56 / UTM zone 17N\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24817\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24818, 1, 'PSAD56 / UTM zone 18N', 'EPSG', 24818, 'PROJCS[\"PSAD56 / UTM zone 18N\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-7"
  "5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24818\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24819, 1, 'PSAD56 / UTM zone 19N', 'EPSG', 24819, 'PROJCS[\"PSAD56 / UTM zone 19N\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-6"
  "9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24819\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24820, 1, 'PSAD56 / UTM zone 20N', 'EPSG', 24820, 'PROJCS[\"PSAD56 / UTM zone 20N\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-6"
  "3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24820\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24821, 1, 'PSAD56 / UTM zone 21N', 'EPSG', 24821, 'PROJCS[\"PSAD56 / UTM zone 21N\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5"
  "7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24821\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24877, 1, 'PSAD56 / UTM zone 17S', 'EPSG', 24877, 'PROJCS[\"PSAD56 / UTM zone 17S\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24"
  "877\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24878, 1, 'PSAD56 / UTM zone 18S', 'EPSG', 24878, 'PROJCS[\"PSAD56 / UTM zone 18S\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-7"
  "5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24"
  "878\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24879, 1, 'PSAD56 / UTM zone 19S', 'EPSG', 24879, 'PROJCS[\"PSAD56 / UTM zone 19S\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-6"
  "9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24"
  "879\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24880, 1, 'PSAD56 / UTM zone 20S', 'EPSG', 24880, 'PROJCS[\"PSAD56 / UTM zone 20S\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-6"
  "3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24"
  "880\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24881, 1, 'PSAD56 / UTM zone 21S', 'EPSG', 24881, 'PROJCS[\"PSAD56 / UTM zone 21S\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5"
  "7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24"
  "881\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24882, 1, 'PSAD56 / UTM zone 22S', 'EPSG', 24882, 'PROJCS[\"PSAD56 / UTM zone 22S\",GEOGCS[\"P"
  "SAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5"
  "1,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"24"
  "882\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24891, 1, 'PSAD56 / Peru west zone', 'EPSG', 24891, 'PROJCS[\"PSAD56 / Peru west zone\",GEOGCS"
  "[\"PSAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E"
  "PSG\",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",-6,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-80.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99983008,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",222000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",14"
  "26834.743,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NO"
  "RTH],AUTHORITY[\"EPSG\",\"24891\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24892, 1, 'PSAD56 / Peru central zone', 'EPSG', 24892, 'PROJCS[\"PSAD56 / Peru central zone\","
  "GEOGCS[\"PSAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORI"
  "TY[\"EPSG\",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",-9.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",-76,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99932994,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",720000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1039979.159"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTH"
  "ORITY[\"EPSG\",\"24892\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (24893, 1, 'PSAD56 / Peru east zone', 'EPSG', 24893, 'PROJCS[\"PSAD56 / Peru east zone\",GEOGCS"
  "[\"PSAD56\",DATUM[\"Provisional South American Datum 1956\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E"
  "PSG\",\"7022\"]],TOWGS84[-60.31,245.935,31.008,12.324,3.755,-7.37,0.447],AUTHORITY[\"EPSG\",\"6248\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4248\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",-9.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99952992,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1324000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",1040084.558,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\""
  ",NORTH],AUTHORITY[\"EPSG\",\"24893\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25000, 1, 'Leigon / Ghana Metre Grid', 'EPSG', 25000, 'PROJCS[\"Leigon / Ghana Metre Grid\",GE"
  "OGCS[\"Leigon\",DATUM[\"Leigon\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOW"
  "GS84[-130,29,364,0,0,0,0],AUTHORITY[\"EPSG\",\"6250\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4250\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "4.66666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-1,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.99975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2"
  "74319.51,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25000\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25231, 1, 'Lome / UTM zone 31N', 'EPSG', 25231, 'PROJCS[\"Lome / UTM zone 31N\",GEOGCS[\"Lome\""
  ",DATUM[\"Lome\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\""
  "EPSG\",\"6252\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4252\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99"
  "96,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\","
  "NORTH],AUTHORITY[\"EPSG\",\"25231\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25391, 1, 'Luzon 1911 / Philippines zone I', 'EPSG', 25391, 'PROJCS[\"Luzon 1911 / Philippines"
  " zone I\",GEOGCS[\"Luzon 1911\",DATUM[\"Luzon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-133,-77,-51,0,0,0,0],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4253\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"25391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25392, 1, 'Luzon 1911 / Philippines zone II', 'EPSG', 25392, 'PROJCS[\"Luzon 1911 / Philippine"
  "s zone II\",GEOGCS[\"Luzon 1911\",DATUM[\"Luzon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],TOWGS84[-133,-77,-51,0,0,0,0],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4253\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",119,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"25392\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25393, 1, 'Luzon 1911 / Philippines zone III', 'EPSG', 25393, 'PROJCS[\"Luzon 1911 / Philippin"
  "es zone III\",GEOGCS[\"Luzon 1911\",DATUM[\"Luzon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORI"
  "TY[\"EPSG\",\"7008\"]],TOWGS84[-133,-77,-51,0,0,0,0],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4253\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",121,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"25393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25394, 1, 'Luzon 1911 / Philippines zone IV', 'EPSG', 25394, 'PROJCS[\"Luzon 1911 / Philippine"
  "s zone IV\",GEOGCS[\"Luzon 1911\",DATUM[\"Luzon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY"
  "[\"EPSG\",\"7008\"]],TOWGS84[-133,-77,-51,0,0,0,0],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4253\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"25394\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25395, 1, 'Luzon 1911 / Philippines zone V', 'EPSG', 25395, 'PROJCS[\"Luzon 1911 / Philippines"
  " zone V\",GEOGCS[\"Luzon 1911\",DATUM[\"Luzon 1911\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-133,-77,-51,0,0,0,0],AUTHORITY[\"EPSG\",\"6253\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4253\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",125,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"25395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25828, 1, 'ETRS89 / UTM zone 28N', 'EPSG', 25828, 'PROJCS[\"ETRS89 / UTM zone 28N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25828\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25829, 1, 'ETRS89 / UTM zone 29N', 'EPSG', 25829, 'PROJCS[\"ETRS89 / UTM zone 29N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25829\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25830, 1, 'ETRS89 / UTM zone 30N', 'EPSG', 25830, 'PROJCS[\"ETRS89 / UTM zone 30N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25830\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25831, 1, 'ETRS89 / UTM zone 31N', 'EPSG', 25831, 'PROJCS[\"ETRS89 / UTM zone 31N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25831\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25832, 1, 'ETRS89 / UTM zone 32N', 'EPSG', 25832, 'PROJCS[\"ETRS89 / UTM zone 32N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25832\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25833, 1, 'ETRS89 / UTM zone 33N', 'EPSG', 25833, 'PROJCS[\"ETRS89 / UTM zone 33N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25833\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25834, 1, 'ETRS89 / UTM zone 34N', 'EPSG', 25834, 'PROJCS[\"ETRS89 / UTM zone 34N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25834\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25835, 1, 'ETRS89 / UTM zone 35N', 'EPSG', 25835, 'PROJCS[\"ETRS89 / UTM zone 35N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25835\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25836, 1, 'ETRS89 / UTM zone 36N', 'EPSG', 25836, 'PROJCS[\"ETRS89 / UTM zone 36N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25836\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25837, 1, 'ETRS89 / UTM zone 37N', 'EPSG', 25837, 'PROJCS[\"ETRS89 / UTM zone 37N\",GEOGCS[\"E"
  "TRS89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"89"
  "01\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25837\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25884, 1, 'ETRS89 / TM Baltic93', 'EPSG', 25884, 'PROJCS[\"ETRS89 / TM Baltic93\",GEOGCS[\"ETR"
  "S89\",DATUM[\"European Terrestrial Reference System 1989\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EP"
  "SG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6258\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4258\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",24,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"25884\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (25932, 1, 'Malongo 1987 / UTM zone 32S', 'EPSG', 25932, 'PROJCS[\"Malongo 1987 / UTM zone 32S\""
  ",GEOGCS[\"Malongo 1987\",DATUM[\"Malongo 1987\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022"
  "\"]],TOWGS84[-254.1,-5.36,-100.29,0,0,0,0],AUTHORITY[\"EPSG\",\"6259\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\""
  "8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AU"
  "THORITY[\"EPSG\",\"4259\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of "
  "natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"25932\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26191, 1, 'Merchich / Nord Maroc', 'EPSG', 26191, 'PROJCS[\"Merchich / Nord Maroc\",GEOGCS[\"M"
  "erchich\",DATUM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]]"
  ",TOWGS84[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4261\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natura"
  "l origin\",33.3,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5.4,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.999625769,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26192, 1, 'Merchich / Sud Maroc', 'EPSG', 26192, 'PROJCS[\"Merchich / Sud Maroc\",GEOGCS[\"Mer"
  "chich\",DATUM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],T"
  "OWGS84[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4261\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural"
  " origin\",29.7,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5.4,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.999615596,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26194, 1, 'Merchich / Sahara Nord', 'EPSG', 26194, 'PROJCS[\"Merchich / Sahara Nord\",GEOGCS[\""
  "Merchich\",DATUM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]"
  "],TOWGS84[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT["
  "\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG"
  "\",\"4261\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",26.1,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5.4,AUTHORITY[\"EPSG\",\"8"
  "802\"]],PARAMETER[\"Scale factor at natural origin\",0.999616304,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",1200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",400000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26194\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26195, 1, 'Merchich / Sahara Sud', 'EPSG', 26195, 'PROJCS[\"Merchich / Sahara Sud\",GEOGCS[\"M"
  "erchich\",DATUM[\"Merchich\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]]"
  ",TOWGS84[31,146,47,0,0,0,0],AUTHORITY[\"EPSG\",\"6261\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4261\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natura"
  "l origin\",22.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-5.4,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.999616437,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",1500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",400000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26237, 1, 'Massawa / UTM zone 37N', 'EPSG', 26237, 'PROJCS[\"Massawa / UTM zone 37N\",GEOGCS[\""
  "Massawa\",DATUM[\"Massawa\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[63"
  "9,405,60,0,0,0,0],AUTHORITY[\"EPSG\",\"6262\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4262\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26237\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26331, 1, 'Minna / UTM zone 31N', 'EPSG', 26331, 'PROJCS[\"Minna / UTM zone 31N\",GEOGCS[\"Min"
  "na\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-92,-93"
  ",122,0,0,0,0],AUTHORITY[\"EPSG\",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4263\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY["
  "\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale facto"
  "r at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26332, 1, 'Minna / UTM zone 32N', 'EPSG', 26332, 'PROJCS[\"Minna / UTM zone 32N\",GEOGCS[\"Min"
  "na\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-92,-93"
  ",122,0,0,0,0],AUTHORITY[\"EPSG\",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4263\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY["
  "\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale facto"
  "r at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26391, 1, 'Minna / Nigeria West Belt', 'EPSG', 26391, 'PROJCS[\"Minna / Nigeria West Belt\",GE"
  "OGCS[\"Minna\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS"
  "84[-92,-93,122,0,0,0,0],AUTHORITY[\"EPSG\",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4263\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",4,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",4.51111111111111,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.99975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2307"
  "38.26,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26392, 1, 'Minna / Nigeria Mid Belt', 'EPSG', 26392, 'PROJCS[\"Minna / Nigeria Mid Belt\",GEOG"
  "CS[\"Minna\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS84"
  "[-92,-93,122,0,0,0,0],AUTHORITY[\"EPSG\",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "63\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",4,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",8.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.99975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",670553.98,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26392\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26393, 1, 'Minna / Nigeria East Belt', 'EPSG', 26393, 'PROJCS[\"Minna / Nigeria East Belt\",GE"
  "OGCS[\"Minna\",DATUM[\"Minna\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\",\"7012\"]],TOWGS"
  "84[-92,-93,122,0,0,0,0],AUTHORITY[\"EPSG\",\"6263\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4263\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",4,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",12.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.99975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1110369.7,AUTHOR"
  "ITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPS"
  "G\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26632, 1, 'M\\'poraloko / UTM zone 32N', 'EPSG', 26632, 'PROJCS[\"M\\'poraloko / UTM zone 32N\","
  "GEOGCS[\"M\\'poraloko\",DATUM[\"M\\'poraloko\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"E"
  "PSG\",\"7011\"]],TOWGS84[-80.7,-132.5,41.1,0,0,0,0],AUTHORITY[\"EPSG\",\"6266\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4266\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26632\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26692, 1, 'M\\'poraloko / UTM zone 32S', 'EPSG', 26692, 'PROJCS[\"M\\'poraloko / UTM zone 32S\","
  "GEOGCS[\"M\\'poraloko\",DATUM[\"M\\'poraloko\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"E"
  "PSG\",\"7011\"]],TOWGS84[-80.7,-132.5,41.1,0,0,0,0],AUTHORITY[\"EPSG\",\"6266\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4266\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26692\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26701, 1, 'NAD27 / UTM zone 1N', 'EPSG', 26701, 'PROJCS[\"NAD27 / UTM zone 1N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26701\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26702, 1, 'NAD27 / UTM zone 2N', 'EPSG', 26702, 'PROJCS[\"NAD27 / UTM zone 2N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26702\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26703, 1, 'NAD27 / UTM zone 3N', 'EPSG', 26703, 'PROJCS[\"NAD27 / UTM zone 3N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26703\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26704, 1, 'NAD27 / UTM zone 4N', 'EPSG', 26704, 'PROJCS[\"NAD27 / UTM zone 4N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26704\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26705, 1, 'NAD27 / UTM zone 5N', 'EPSG', 26705, 'PROJCS[\"NAD27 / UTM zone 5N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26705\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26706, 1, 'NAD27 / UTM zone 6N', 'EPSG', 26706, 'PROJCS[\"NAD27 / UTM zone 6N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26706\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26707, 1, 'NAD27 / UTM zone 7N', 'EPSG', 26707, 'PROJCS[\"NAD27 / UTM zone 7N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26707\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26708, 1, 'NAD27 / UTM zone 8N', 'EPSG', 26708, 'PROJCS[\"NAD27 / UTM zone 8N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26708\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26709, 1, 'NAD27 / UTM zone 9N', 'EPSG', 26709, 'PROJCS[\"NAD27 / UTM zone 9N\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26709\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26710, 1, 'NAD27 / UTM zone 10N', 'EPSG', 26710, 'PROJCS[\"NAD27 / UTM zone 10N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"26710\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26711, 1, 'NAD27 / UTM zone 11N', 'EPSG', 26711, 'PROJCS[\"NAD27 / UTM zone 11N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"26711\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26712, 1, 'NAD27 / UTM zone 12N', 'EPSG', 26712, 'PROJCS[\"NAD27 / UTM zone 12N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"26712\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26713, 1, 'NAD27 / UTM zone 13N', 'EPSG', 26713, 'PROJCS[\"NAD27 / UTM zone 13N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"26713\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26714, 1, 'NAD27 / UTM zone 14N', 'EPSG', 26714, 'PROJCS[\"NAD27 / UTM zone 14N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26714\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26715, 1, 'NAD27 / UTM zone 15N', 'EPSG', 26715, 'PROJCS[\"NAD27 / UTM zone 15N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26715\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26716, 1, 'NAD27 / UTM zone 16N', 'EPSG', 26716, 'PROJCS[\"NAD27 / UTM zone 16N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26716\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26717, 1, 'NAD27 / UTM zone 17N', 'EPSG', 26717, 'PROJCS[\"NAD27 / UTM zone 17N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26717\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26718, 1, 'NAD27 / UTM zone 18N', 'EPSG', 26718, 'PROJCS[\"NAD27 / UTM zone 18N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26718\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26719, 1, 'NAD27 / UTM zone 19N', 'EPSG', 26719, 'PROJCS[\"NAD27 / UTM zone 19N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26719\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26720, 1, 'NAD27 / UTM zone 20N', 'EPSG', 26720, 'PROJCS[\"NAD27 / UTM zone 20N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26720\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26721, 1, 'NAD27 / UTM zone 21N', 'EPSG', 26721, 'PROJCS[\"NAD27 / UTM zone 21N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26721\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26722, 1, 'NAD27 / UTM zone 22N', 'EPSG', 26722, 'PROJCS[\"NAD27 / UTM zone 22N\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longit"
  "ude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"26722\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26729, 1, 'NAD27 / Alabama East', 'EPSG', 26729, 'PROJCS[\"NAD27 / Alabama East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-85.8333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",0.99996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26729\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26730, 1, 'NAD27 / Alabama West', 'EPSG', 26730, 'PROJCS[\"NAD27 / Alabama West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-87.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26730\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26731, 1, 'NAD27 / Alaska zone 1', 'EPSG', 26731, 'PROJCS[\"NAD27 / Alaska zone 1\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Hotine Oblique Mercator"
  " (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of projection centre\",57,AUTHORITY[\"EPSG\",\"8811"
  "\"]],PARAMETER[\"Longitude of projection centre\",-133.666666666667,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth"
  " of initial line\",323.130102361111,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",323"
  ".130102361111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on initial line\",0.9999,AUTHORITY[\"EPSG\",\"88"
  "15\"]],PARAMETER[\"False easting\",16404166.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-16404166.6"
  "7,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\","
  "EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26731\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26732, 1, 'NAD27 / Alaska zone 2', 'EPSG', 26732, 'PROJCS[\"NAD27 / Alaska zone 2\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-142,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26732\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26733, 1, 'NAD27 / Alaska zone 3', 'EPSG', 26733, 'PROJCS[\"NAD27 / Alaska zone 3\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-146,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26733\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26734, 1, 'NAD27 / Alaska zone 4', 'EPSG', 26734, 'PROJCS[\"NAD27 / Alaska zone 4\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26734\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26735, 1, 'NAD27 / Alaska zone 5', 'EPSG', 26735, 'PROJCS[\"NAD27 / Alaska zone 5\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-154,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26735\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26736, 1, 'NAD27 / Alaska zone 6', 'EPSG', 26736, 'PROJCS[\"NAD27 / Alaska zone 6\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-158,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26736\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26737, 1, 'NAD27 / Alaska zone 7', 'EPSG', 26737, 'PROJCS[\"NAD27 / Alaska zone 7\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26737\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26738, 1, 'NAD27 / Alaska zone 8', 'EPSG', 26738, 'PROJCS[\"NAD27 / Alaska zone 8\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-166,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26739, 1, 'NAD27 / Alaska zone 9', 'EPSG', 26739, 'PROJCS[\"NAD27 / Alaska zone 9\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-170,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AU"
  "THORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26739\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26740, 1, 'NAD27 / Alaska zone 10', 'EPSG', 26740, 'PROJCS[\"NAD27 / Alaska zone 10\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",51,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-176,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\","
  "53.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",51.8333333333333,AUTHO"
  "RITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",3000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26740\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26741, 1, 'NAD27 / California zone I', 'EPSG', 26741, 'PROJCS[\"NAD27 / California zone I\",GE"
  "OGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"E"
  "PSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"E"
  "PSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic C"
  "onformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",41.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETE"
  "R[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26741\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26742, 1, 'NAD27 / California zone II', 'EPSG', 26742, 'PROJCS[\"NAD27 / California zone II\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-122,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1"
  "st standard parallel\",39.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",38.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012"
  "1924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26742\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26743, 1, 'NAD27 / California zone III', 'EPSG', 26743, 'PROJCS[\"NAD27 / California zone III\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.5111111111111,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",37.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0."
  "30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26743\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26744, 1, 'NAD27 / California zone IV', 'EPSG', 26744, 'PROJCS[\"NAD27 / California zone IV\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35.3333333333333,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-119,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1"
  "st standard parallel\",37.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26744\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26745, 1, 'NAD27 / California zone V', 'EPSG', 26745, 'PROJCS[\"NAD27 / California zone V\",GE"
  "OGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"E"
  "PSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"E"
  "PSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic C"
  "onformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-118,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",35.4666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "34.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121"
  "924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26745\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26746, 1, 'NAD27 / California zone VI', 'EPSG', 26746, 'PROJCS[\"NAD27 / California zone VI\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.1666666666667,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-116.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude o"
  "f 1st standard parallel\",33.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard paralle"
  "l\",32.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26746\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26748, 1, 'NAD27 / Arizona East', 'EPSG', 26748, 'PROJCS[\"NAD27 / Arizona East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-110.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26748\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26749, 1, 'NAD27 / Arizona Central', 'EPSG', 26749, 'PROJCS[\"NAD27 / Arizona Central\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-111.927777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26749\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26750, 1, 'NAD27 / Arizona West', 'EPSG', 26750, 'PROJCS[\"NAD27 / Arizona West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-113.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999333"
  "33,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False n"
  "orthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26750\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26751, 1, 'NAD27 / Arkansas North', 'EPSG', 26751, 'PROJCS[\"NAD27 / Arkansas North\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",36.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.933333"
  "3333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26751\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26752, 1, 'NAD27 / Arkansas South', 'EPSG', 26752, 'PROJCS[\"NAD27 / Arkansas South\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",34.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",33.311111"
  "1111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26752\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26753, 1, 'NAD27 / Colorado North', 'EPSG', 26753, 'PROJCS[\"NAD27 / Colorado North\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",39.7277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",40.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26753\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26754, 1, 'NAD27 / Colorado Central', 'EPSG', 26754, 'PROJCS[\"NAD27 / Colorado Central\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Con"
  "formal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG"
  "\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Lat"
  "itude of 1st standard parallel\",39.75,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "38.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAM"
  "ETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26754\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26755, 1, 'NAD27 / Colorado South', 'EPSG', 26755, 'PROJCS[\"NAD27 / Colorado South\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",37.2333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26755\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26756, 1, 'NAD27 / Connecticut', 'EPSG', 26756, 'PROJCS[\"NAD27 / Connecticut\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.8333333333333,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-72.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",41.8666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.21111111"
  "11111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26756\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26757, 1, 'NAD27 / Delaware', 'EPSG', 26757, 'PROJCS[\"NAD27 / Delaware\",GEOGCS[\"NAD27\",DAT"
  "UM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],T"
  "OWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",-75.4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999"
  "5,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False no"
  "rthing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AX"
  "IS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26757\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26758, 1, 'NAD27 / Florida East', 'EPSG', 26758, 'PROJCS[\"NAD27 / Florida East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26758\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26759, 1, 'NAD27 / Florida West', 'EPSG', 26759, 'PROJCS[\"NAD27 / Florida West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-82,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26759\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26760, 1, 'NAD27 / Florida North', 'EPSG', 26760, 'PROJCS[\"NAD27 / Florida North\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-84.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",30.7611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",29.583333"
  "3333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26760\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26766, 1, 'NAD27 / Georgia East', 'EPSG', 26766, 'PROJCS[\"NAD27 / Georgia East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-82.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26766\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26767, 1, 'NAD27 / Georgia West', 'EPSG', 26767, 'PROJCS[\"NAD27 / Georgia West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longi"
  "tude of natural origin\",-84.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26767\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26768, 1, 'NAD27 / Idaho East', 'EPSG', 26768, 'PROJCS[\"NAD27 / Idaho East\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-112.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26768\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26769, 1, 'NAD27 / Idaho Central', 'EPSG', 26769, 'PROJCS[\"NAD27 / Idaho Central\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26769\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26770, 1, 'NAD27 / Idaho West', 'EPSG', 26770, 'PROJCS[\"NAD27 / Idaho West\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-115.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26770\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26771, 1, 'NAD27 / Illinois East', 'EPSG', 26771, 'PROJCS[\"NAD27 / Illinois East\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-88.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.999975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26771\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26772, 1, 'NAD27 / Illinois West', 'EPSG', 26772, 'PROJCS[\"NAD27 / Illinois West\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-90.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26772\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26773, 1, 'NAD27 / Indiana East', 'EPSG', 26773, 'PROJCS[\"NAD27 / Indiana East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-85.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26773\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26774, 1, 'NAD27 / Indiana West', 'EPSG', 26774, 'PROJCS[\"NAD27 / Indiana West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-87.0944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26774\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26775, 1, 'NAD27 / Iowa North', 'EPSG', 26775, 'PROJCS[\"NAD27 / Iowa North\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st"
  " standard parallel\",43.2777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",4"
  "2.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601219"
  "24,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26775\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26776, 1, 'NAD27 / Iowa South', 'EPSG', 26776, 'PROJCS[\"NAD27 / Iowa South\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.627777777777"
  "8,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26776\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26777, 1, 'NAD27 / Kansas North', 'EPSG', 26777, 'PROJCS[\"NAD27 / Kansas North\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard p"
  "arallel\",39.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.727777777"
  "7778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26777\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26778, 1, 'NAD27 / Kansas South', 'EPSG', 26778, 'PROJCS[\"NAD27 / Kansas South\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",38.5666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",37.2777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012"
  "1924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26778\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26779, 1, 'NAD27 / Kentucky North', 'EPSG', 26779, 'PROJCS[\"NAD27 / Kentucky North\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.5111111111111,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-84.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.977"
  "7777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AU"
  "THORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26779\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26780, 1, 'NAD27 / Kentucky South', 'EPSG', 26780, 'PROJCS[\"NAD27 / Kentucky South\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-85.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",36.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.933"
  "3333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AU"
  "THORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26780\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26781, 1, 'NAD27 / Louisiana North', 'EPSG', 26781, 'PROJCS[\"NAD27 / Louisiana North\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",30.6777777777778,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitu"
  "de of 1st standard parallel\",31.1666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard par"
  "allel\",32.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26781\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26782, 1, 'NAD27 / Louisiana South', 'EPSG', 26782, 'PROJCS[\"NAD27 / Louisiana South\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",28.6777777777778,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitu"
  "de of 1st standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard par"
  "allel\",30.7,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]"
  "],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,"
  "AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26782\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26783, 1, 'NAD27 / Maine East', 'EPSG', 26783, 'PROJCS[\"NAD27 / Maine East\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PAR"
  "AMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPS"
  "G\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26783\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26784, 1, 'NAD27 / Maine West', 'EPSG', 26784, 'PROJCS[\"NAD27 / Maine West\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-70.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26784\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26785, 1, 'NAD27 / Maryland', 'EPSG', 26785, 'PROJCS[\"NAD27 / Maryland\",GEOGCS[\"NAD27\",DAT"
  "UM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],T"
  "OWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],P"
  "RIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]"
  "],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",A"
  "UTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PA"
  "RAMETER[\"Longitude of false origin\",-77,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel"
  "\",38.3111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39.45,AUTHORITY[\"E"
  "PSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at f"
  "alse origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26785\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26786, 1, 'NAD27 / Massachusetts Mainland', 'EPSG', 26786, 'PROJCS[\"NAD27 / Massachusetts Mai"
  "nland\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUT"
  "HORITY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUT"
  "HORITY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,"
  "AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambe"
  "rt Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\""
  ",\"8821\"]],PARAMETER[\"Longitude of false origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latit"
  "ude of 1st standard parallel\",41.7277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard pa"
  "rallel\",42.6944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPSG"
  "\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800"
  "60960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26786\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26787, 1, 'NAD27 / Massachusetts Island', 'EPSG', 26787, 'PROJCS[\"NAD27 / Massachusetts Islan"
  "d\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORI"
  "TY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORI"
  "TY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert C"
  "onic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",41.2833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",41.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\""
  "8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26787\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26791, 1, 'NAD27 / Minnesota North', 'EPSG', 26791, 'PROJCS[\"NAD27 / Minnesota North\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st st"
  "andard parallel\",47.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",48.6"
  "333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]"
  "],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,"
  "AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26792, 1, 'NAD27 / Minnesota Central', 'EPSG', 26792, 'PROJCS[\"NAD27 / Minnesota Central\",GE"
  "OGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"E"
  "PSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"E"
  "PSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic C"
  "onformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-94.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard p"
  "arallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.05,AUTHOR"
  "ITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"North"
  "ing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\","
  "\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26792\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26793, 1, 'NAD27 / Minnesota South', 'EPSG', 26793, 'PROJCS[\"NAD27 / Minnesota South\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PAR"
  "AMETER[\"Longitude of false origin\",-94,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",43.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.2166666666667,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26793\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26794, 1, 'NAD27 / Mississippi East', 'EPSG', 26794, 'PROJCS[\"NAD27 / Mississippi East\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-88.8333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",0.99996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUT"
  "HORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26794\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26795, 1, 'NAD27 / Mississippi West', 'EPSG', 26795, 'PROJCS[\"NAD27 / Mississippi West\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-90.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]"
  "],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY["
  "\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26795\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26796, 1, 'NAD27 / Missouri East', 'EPSG', 26796, 'PROJCS[\"NAD27 / Missouri East\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-90.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26796\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26797, 1, 'NAD27 / Missouri Central', 'EPSG', 26797, 'PROJCS[\"NAD27 / Missouri Central\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",35.8333333333333,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-92.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fac"
  "tor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.3048006096012192"
  "4,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26797\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26798, 1, 'NAD27 / Missouri West', 'EPSG', 26798, 'PROJCS[\"NAD27 / Missouri West\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36.1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-94.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26798\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26799, 1, 'NAD27 / California zone VII', 'EPSG', 26799, 'PROJCS[\"NAD27 / California zone VII\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34.1444444444444,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-118.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",34.4166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",33.8666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",4186692.58,AUTHORI"
  "TY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4160926.74,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US surv"
  "ey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"2"
  "6799\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26847, 1, 'NAD83 / Maine East (ftUS)', 'EPSG', 26847, 'PROJCS[\"NAD83 / Maine East (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-68.5111111111111,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORIT"
  "Y[\"EPSG\",\"26847\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26848, 1, 'NAD83 / Maine West (ftUS)', 'EPSG', 26848, 'PROJCS[\"NAD83 / Maine West (ftUS)\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural ori"
  "gin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.1777777777778,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARA"
  "METER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"880"
  "7\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"26848\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26849, 1, 'NAD83 / Minnesota North (ftUS)', 'EPSG', 26849, 'PROJCS[\"NAD83 / Minnesota North ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.1,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6333333333333,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",328083.3333"
  ",AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26849\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26850, 1, 'NAD83 / Minnesota Central (ftUS)', 'EPSG', 26850, 'PROJCS[\"NAD83 / Minnesota Centr"
  "al (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94.25,AUTHORITY["
  "\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",328083.3333,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]"
  ",AUTHORITY[\"EPSG\",\"26850\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26851, 1, 'NAD83 / Minnesota South (ftUS)', 'EPSG', 26851, 'PROJCS[\"NAD83 / Minnesota South ("
  "ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94,AUTHORITY[\"EPSG"
  "\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETE"
  "R[\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false o"
  "rigin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",328083.3333,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\","
  "NORTH],AUTHORITY[\"EPSG\",\"26851\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26852, 1, 'NAD83 / Nebraska (ftUS)', 'EPSG', 26852, 'PROJCS[\"NAD83 / Nebraska (ftUS)\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG"
  "\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude o"
  "f 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1640416.6667,AUTHORI"
  "TY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\""
  ",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26852\"]]'"
  ", NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26853, 1, 'NAD83 / West Virginia North (ftUS)', 'EPSG', 26853, 'PROJCS[\"NAD83 / West Virginia"
  " North (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMET"
  "ER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\""
  ",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.2611111111111,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER"
  "[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY"
  "[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26853\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26854, 1, 'NAD83 / West Virginia South (ftUS)', 'EPSG', 26854, 'PROJCS[\"NAD83 / West Virginia"
  " South (ftUS)\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMET"
  "ER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"26854\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26855, 1, 'NAD83(HARN) / Maine East (ftUS)', 'EPSG', 26855, 'PROJCS[\"NAD83(HARN) / Maine East"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORIT"
  "Y[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",E"
  "AST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26855\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26856, 1, 'NAD83(HARN) / Maine West (ftUS)', 'EPSG', 26856, 'PROJCS[\"NAD83(HARN) / Maine West"
  " (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298."
  "257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",-70.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUT"
  "HORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northi"
  "ng\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26856\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26857, 1, 'NAD83(HARN) / Minnesota North (ftUS)', 'EPSG', 26857, 'PROJCS[\"NAD83(HARN) / Minne"
  "sota North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Long"
  "itude of false origin\",-93.1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.633333"
  "3333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\"EP"
  "SG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing"
  " at false origin\",328083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26857\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26858, 1, 'NAD83(HARN) / Minnesota Central (ftUS)', 'EPSG', 26858, 'PROJCS[\"NAD83(HARN) / Min"
  "nesota Central (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\""
  ",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Gr"
  "eenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"L"
  "at\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\""
  "EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fa"
  "lse origin\",-94.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORITY[\"E"
  "PSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAM"
  "ETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",32"
  "8083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXI"
  "S[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26858\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26859, 1, 'NAD83(HARN) / Minnesota South (ftUS)', 'EPSG', 26859, 'PROJCS[\"NAD83(HARN) / Minne"
  "sota South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPS"
  "G\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false"
  " origin\",-94,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHORIT"
  "Y[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],"
  "PARAMETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin"
  "\",328083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]"
  "],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26859\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26860, 1, 'NAD83(HARN) / Nebraska (ftUS)', 'EPSG', 26860, 'PROJCS[\"NAD83(HARN) / Nebraska (ft"
  "US)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1980\",6378137,298.2572"
  "22101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]"
  "],PARAMETER[\"Latitude of false origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of fals"
  "e origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fals"
  "e origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"26860\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26861, 1, 'NAD83(HARN) / West Virginia North (ftUS)', 'EPSG', 26861, 'PROJCS[\"NAD83(HARN) / W"
  "est Virginia North (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMET"
  "ER[\"Longitude of false origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard"
  " parallel\",40.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northi"
  "ng at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26861\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26862, 1, 'NAD83(HARN) / West Virginia South (ftUS)', 'EPSG', 26862, 'PROJCS[\"NAD83(HARN) / W"
  "est Virginia South (ftUS)\",GEOGCS[\"NAD83(HARN)\",DATUM[\"NAD83 (High Accuracy Reference Network)\",SPHEROID[\"GRS 1"
  "980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6152\"]],PRIMEM"
  "[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXI"
  "S[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4152\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHOR"
  "ITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude"
  " of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333333"
  ",AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26862\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26863, 1, 'NAD83(NSRS2007) / Maine East (ftUS)', 'EPSG', 26863, 'PROJCS[\"NAD83(NSRS2007) / Ma"
  "ine East (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",984250,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26863\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26864, 1, 'NAD83(NSRS2007) / Maine West (ftUS)', 'EPSG', 26864, 'PROJCS[\"NAD83(NSRS2007) / Ma"
  "ine West (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS "
  "1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIME"
  "M[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AX"
  "IS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPS"
  "G\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-70.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2952750,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26864\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26865, 1, 'NAD83(NSRS2007) / Minnesota North (ftUS)', 'EPSG', 26865, 'PROJCS[\"NAD83(NSRS2007)"
  " / Minnesota North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-93.1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard p"
  "arallel\",48.6333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.033333333"
  "3333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],P"
  "ARAMETER[\"Northing at false origin\",328083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960"
  "121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26865\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26866, 1, 'NAD83(NSRS2007) / Minnesota Central (ftUS)', 'EPSG', 26866, 'PROJCS[\"NAD83(NSRS200"
  "7) / Minnesota Central (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",S"
  "PHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\""
  "6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-94.25,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",47.05,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.6277777777778,AUTHORITY[\"EPS"
  "G\",\"8824\"]],PARAMETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing "
  "at false origin\",328083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"E"
  "PSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26866\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26867, 1, 'NAD83(NSRS2007) / Minnesota South (ftUS)', 'EPSG', 26867, 'PROJCS[\"NAD83(NSRS2007)"
  " / Minnesota South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHER"
  "OID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",-94,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.21"
  "66666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",43.7944444444444,AUTHORITY["
  "\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2624666.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nort"
  "hing at false origin\",328083.3333,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26867\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26868, 1, 'NAD83(NSRS2007) / Nebraska (ftUS)', 'EPSG', 26868, 'PROJCS[\"NAD83(NSRS2007) / Nebr"
  "aska (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)\",SPHEROID[\"GRS 1980"
  "\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6759\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY["
  "\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43,AUT"
  "HORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER"
  "[\"Easting at false origin\",1640416.6667,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],"
  "AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26868\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26869, 1, 'NAD83(NSRS2007) / West Virginia North (ftUS)', 'EPSG', 26869, 'PROJCS[\"NAD83(NSRS2"
  "007) / West Virginia North (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitu"
  "de of 1st standard parallel\",40.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard par"
  "allel\",39,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AU"
  "THORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26869\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26870, 1, 'NAD83(NSRS2007) / West Virginia South (ftUS)', 'EPSG', 26870, 'PROJCS[\"NAD83(NSRS2"
  "007) / West Virginia South (ftUS)\",GEOGCS[\"NAD83(NSRS2007)\",DATUM[\"NAD83 (National Spatial Reference System 2007)"
  "\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\""
  ",\"6759\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4759\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PAR"
  "AMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",38.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.4944444444444,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1968500,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26870\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26891, 1, 'NAD83(CSRS) / MTM zone 11', 'EPSG', 26891, 'PROJCS[\"NAD83(CSRS) / MTM zone 11\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale fa"
  "ctor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],"
  "AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26891\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26892, 1, 'NAD83(CSRS) / MTM zone 12', 'EPSG', 26892, 'PROJCS[\"NAD83(CSRS) / MTM zone 12\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26892\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26893, 1, 'NAD83(CSRS) / MTM zone 13', 'EPSG', 26893, 'PROJCS[\"NAD83(CSRS) / MTM zone 13\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26893\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26894, 1, 'NAD83(CSRS) / MTM zone 14', 'EPSG', 26894, 'PROJCS[\"NAD83(CSRS) / MTM zone 14\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26894\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26895, 1, 'NAD83(CSRS) / MTM zone 15', 'EPSG', 26895, 'PROJCS[\"NAD83(CSRS) / MTM zone 15\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26895\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26896, 1, 'NAD83(CSRS) / MTM zone 16', 'EPSG', 26896, 'PROJCS[\"NAD83(CSRS) / MTM zone 16\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26896\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26897, 1, 'NAD83(CSRS) / MTM zone 17', 'EPSG', 26897, 'PROJCS[\"NAD83(CSRS) / MTM zone 17\",GE"
  "OGCS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AU"
  "THORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0]"
  ",AUTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943"
  "278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"T"
  "ransverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"88"
  "01\"]],PARAMETER[\"Longitude of natural origin\",-96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26897\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26898, 1, 'NAD83(CSRS) / MTM zone 1', 'EPSG', 26898, 'PROJCS[\"NAD83(CSRS) / MTM zone 1\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-53,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E(X)\",E"
  "AST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"26898\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26899, 1, 'NAD83(CSRS) / MTM zone 2', 'EPSG', 26899, 'PROJCS[\"NAD83(CSRS) / MTM zone 2\",GEOG"
  "CS[\"NAD83(CSRS)\",DATUM[\"NAD83 Canadian Spatial Reference System\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTH"
  "ORITY[\"EPSG\",\"7019\"]],TOWGS84[-0.991,1.9072,0.5129,0.0257899075194932,0.00965009896027041,0.0116599432323421,0],A"
  "UTHORITY[\"EPSG\",\"6140\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994327"
  "8,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4617\"]],PROJECTION[\"Tra"
  "nsverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801"
  "\"]],PARAMETER[\"Longitude of natural origin\",-56,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E(X)\",E"
  "AST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"26899\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26901, 1, 'NAD83 / UTM zone 1N', 'EPSG', 26901, 'PROJCS[\"NAD83 / UTM zone 1N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26901\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26902, 1, 'NAD83 / UTM zone 2N', 'EPSG', 26902, 'PROJCS[\"NAD83 / UTM zone 2N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26902\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26903, 1, 'NAD83 / UTM zone 3N', 'EPSG', 26903, 'PROJCS[\"NAD83 / UTM zone 3N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26903\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26904, 1, 'NAD83 / UTM zone 4N', 'EPSG', 26904, 'PROJCS[\"NAD83 / UTM zone 4N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26904\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26905, 1, 'NAD83 / UTM zone 5N', 'EPSG', 26905, 'PROJCS[\"NAD83 / UTM zone 5N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26905\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26906, 1, 'NAD83 / UTM zone 6N', 'EPSG', 26906, 'PROJCS[\"NAD83 / UTM zone 6N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26906\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26907, 1, 'NAD83 / UTM zone 7N', 'EPSG', 26907, 'PROJCS[\"NAD83 / UTM zone 7N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26907\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26908, 1, 'NAD83 / UTM zone 8N', 'EPSG', 26908, 'PROJCS[\"NAD83 / UTM zone 8N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26908\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26909, 1, 'NAD83 / UTM zone 9N', 'EPSG', 26909, 'PROJCS[\"NAD83 / UTM zone 9N\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scal"
  "e factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG"
  "\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001"
  "\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26909\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26910, 1, 'NAD83 / UTM zone 10N', 'EPSG', 26910, 'PROJCS[\"NAD83 / UTM zone 10N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26910\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26911, 1, 'NAD83 / UTM zone 11N', 'EPSG', 26911, 'PROJCS[\"NAD83 / UTM zone 11N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26911\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26912, 1, 'NAD83 / UTM zone 12N', 'EPSG', 26912, 'PROJCS[\"NAD83 / UTM zone 12N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26912\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26913, 1, 'NAD83 / UTM zone 13N', 'EPSG', 26913, 'PROJCS[\"NAD83 / UTM zone 13N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26913\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26914, 1, 'NAD83 / UTM zone 14N', 'EPSG', 26914, 'PROJCS[\"NAD83 / UTM zone 14N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26914\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26915, 1, 'NAD83 / UTM zone 15N', 'EPSG', 26915, 'PROJCS[\"NAD83 / UTM zone 15N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26915\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26916, 1, 'NAD83 / UTM zone 16N', 'EPSG', 26916, 'PROJCS[\"NAD83 / UTM zone 16N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26916\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26917, 1, 'NAD83 / UTM zone 17N', 'EPSG', 26917, 'PROJCS[\"NAD83 / UTM zone 17N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26917\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26918, 1, 'NAD83 / UTM zone 18N', 'EPSG', 26918, 'PROJCS[\"NAD83 / UTM zone 18N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26918\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26919, 1, 'NAD83 / UTM zone 19N', 'EPSG', 26919, 'PROJCS[\"NAD83 / UTM zone 19N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26919\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26920, 1, 'NAD83 / UTM zone 20N', 'EPSG', 26920, 'PROJCS[\"NAD83 / UTM zone 20N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26920\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26921, 1, 'NAD83 / UTM zone 21N', 'EPSG', 26921, 'PROJCS[\"NAD83 / UTM zone 21N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26921\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26922, 1, 'NAD83 / UTM zone 22N', 'EPSG', 26922, 'PROJCS[\"NAD83 / UTM zone 22N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26922\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26923, 1, 'NAD83 / UTM zone 23N', 'EPSG', 26923, 'PROJCS[\"NAD83 / UTM zone 23N\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"E"
  "PSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"26923\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26929, 1, 'NAD83 / Alabama East', 'EPSG', 26929, 'PROJCS[\"NAD83 / Alabama East\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30.5"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.8333333333333,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.99996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200"
  "000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHO"
  "RITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26929\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26930, 1, 'NAD83 / Alabama West', 'EPSG', 26930, 'PROJCS[\"NAD83 / Alabama West\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26930\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26931, 1, 'NAD83 / Alaska zone 1', 'EPSG', 26931, 'PROJCS[\"NAD83 / Alaska zone 1\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Hotine Oblique Mercator (variant A)\",AUTHORITY[\"EPSG\",\"9812\"]],PARAMETER[\"Latitude of pro"
  "jection centre\",57,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",-133.666666666667,AUTH"
  "ORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",323.130102361111,AUTHORITY[\"EPSG\",\"8813\"]],PARAME"
  "TER[\"Angle from Rectified to Skew Grid\",323.130102361111,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on "
  "initial line\",0.9999,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26931\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26932, 1, 'NAD83 / Alaska zone 2', 'EPSG', 26932, 'PROJCS[\"NAD83 / Alaska zone 2\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-142,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26932\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26933, 1, 'NAD83 / Alaska zone 3', 'EPSG', 26933, 'PROJCS[\"NAD83 / Alaska zone 3\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-146,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26933\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26934, 1, 'NAD83 / Alaska zone 4', 'EPSG', 26934, 'PROJCS[\"NAD83 / Alaska zone 4\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-150,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26934\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26935, 1, 'NAD83 / Alaska zone 5', 'EPSG', 26935, 'PROJCS[\"NAD83 / Alaska zone 5\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-154,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26935\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26936, 1, 'NAD83 / Alaska zone 6', 'EPSG', 26936, 'PROJCS[\"NAD83 / Alaska zone 6\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-158,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26936\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26937, 1, 'NAD83 / Alaska zone 7', 'EPSG', 26937, 'PROJCS[\"NAD83 / Alaska zone 7\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-162,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26937\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26938, 1, 'NAD83 / Alaska zone 8', 'EPSG', 26938, 'PROJCS[\"NAD83 / Alaska zone 8\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-166,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26938\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26939, 1, 'NAD83 / Alaska zone 9', 'EPSG', 26939, 'PROJCS[\"NAD83 / Alaska zone 9\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",54"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-170,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26939\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26940, 1, 'NAD83 / Alaska zone 10', 'EPSG', 26940, 'PROJCS[\"NAD83 / Alaska zone 10\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",51,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-176,AUTHORITY[\"EPSG\",\"8822\"]],PAR"
  "AMETER[\"Latitude of 1st standard parallel\",53.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2"
  "nd standard parallel\",51.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",1000000,A"
  "UTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26940\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26941, 1, 'NAD83 / California zone 1', 'EPSG', 26941, 'PROJCS[\"NAD83 / California zone 1\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-122,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",20"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26941\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26942, 1, 'NAD83 / California zone 2', 'EPSG', 26942, 'PROJCS[\"NAD83 / California zone 2\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-122,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",38.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26942"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26943, 1, 'NAD83 / California zone 3', 'EPSG', 26943, 'PROJCS[\"NAD83 / California zone 3\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",36.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.51111111111"
  "1,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\""
  "8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",37.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"E"
  "asting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"26943\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26944, 1, 'NAD83 / California zone 4', 'EPSG', 26944, 'PROJCS[\"NAD83 / California zone 4\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",35.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-119,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",36,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",20"
  "00000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26944\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26945, 1, 'NAD83 / California zone 5', 'EPSG', 26945, 'PROJCS[\"NAD83 / California zone 5\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",33.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-118,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.4666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",34.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26945"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26946, 1, 'NAD83 / California zone 6', 'EPSG', 26946, 'PROJCS[\"NAD83 / California zone 6\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",32.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-116.25,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",32.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26"
  "946\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26948, 1, 'NAD83 / Arizona East', 'EPSG', 26948, 'PROJCS[\"NAD83 / Arizona East\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.177777777778,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",21336"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26948\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26949, 1, 'NAD83 / Arizona Central', 'EPSG', 26949, 'PROJCS[\"NAD83 / Arizona Central\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111.927777777778,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "213360,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26949\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26950, 1, 'NAD83 / Arizona West', 'EPSG', 26950, 'PROJCS[\"NAD83 / Arizona West\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-113.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",213360,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26950\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26951, 1, 'NAD83 / Arkansas North', 'EPSG', 26951, 'PROJCS[\"NAD83 / Arkansas North\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",34.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",36.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",34.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26951\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26952, 1, 'NAD83 / Arkansas South', 'EPSG', 26952, 'PROJCS[\"NAD83 / Arkansas South\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",32.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",33.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,AUTHORITY[\"EPSG\",\"8827\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26952\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26953, 1, 'NAD83 / Colorado North', 'EPSG', 26953, 'PROJCS[\"NAD83 / Colorado North\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",39.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",304800.6096,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"26953\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26954, 1, 'NAD83 / Colorado Central', 'EPSG', 26954, 'PROJCS[\"NAD83 / Colorado Central\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",37.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.75,AUTHORITY[\"EPSG\",\"8823\"]],PAR"
  "AMETER[\"Latitude of 2nd standard parallel\",38.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",304800.6096,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26954\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26955, 1, 'NAD83 / Colorado South', 'EPSG', 26955, 'PROJCS[\"NAD83 / Colorado South\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-105.511111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",37.2333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",914401.8289,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",304800.6096,AUTHOR"
  "ITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\""
  "EPSG\",\"26955\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26956, 1, 'NAD83 / Connecticut', 'EPSG', 26956, 'PROJCS[\"NAD83 / Connecticut\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "40.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-72.75,AUTHORITY[\"EPSG\",\"88"
  "22\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.8666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lat"
  "itude of 2nd standard parallel\",41.2111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",304800.6096,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",152400.3048,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26956\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26957, 1, 'NAD83 / Delaware', 'EPSG', 26957, 'PROJCS[\"NAD83 / Delaware\",GEOGCS[\"NAD83\",DAT"
  "UM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,"
  "1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],P"
  "ROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38,AUTHORITY"
  "[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75.4166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",0.999995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000,AUTH"
  "ORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"E"
  "PSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26957\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26958, 1, 'NAD83 / Florida East', 'EPSG', 26958, 'PROJCS[\"NAD83 / Florida East\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.3"
  "444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26958\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26959, 1, 'NAD83 / Florida West', 'EPSG', 26959, 'PROJCS[\"NAD83 / Florida West\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",24.3"
  "444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26959\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26960, 1, 'NAD83 / Florida North', 'EPSG', 26960, 'PROJCS[\"NAD83 / Florida North\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",29,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.5111111111111,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",29.5833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26960\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26961, 1, 'NAD83 / Hawaii zone 1', 'EPSG', 26961, 'PROJCS[\"NAD83 / Hawaii zone 1\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",18"
  ".8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-155.5,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26961\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26962, 1, 'NAD83 / Hawaii zone 2', 'EPSG', 26962, 'PROJCS[\"NAD83 / Hawaii zone 2\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",20"
  ".3444444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-156.666666666667,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26962\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26963, 1, 'NAD83 / Hawaii zone 3', 'EPSG', 26963, 'PROJCS[\"NAD83 / Hawaii zone 3\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",21"
  ".1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-158,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26963\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26964, 1, 'NAD83 / Hawaii zone 4', 'EPSG', 26964, 'PROJCS[\"NAD83 / Hawaii zone 4\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",21"
  ".8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159.5,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26964\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26965, 1, 'NAD83 / Hawaii zone 5', 'EPSG', 26965, 'PROJCS[\"NAD83 / Hawaii zone 5\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",21"
  ".6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-160.177777777778,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26965\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26966, 1, 'NAD83 / Georgia East', 'EPSG', 26966, 'PROJCS[\"NAD83 / Georgia East\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",20000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26966\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26967, 1, 'NAD83 / Georgia West', 'EPSG', 26967, 'PROJCS[\"NAD83 / Georgia West\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",30,A"
  "UTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84.1777777777778,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",70000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26967\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26968, 1, 'NAD83 / Idaho East', 'EPSG', 26968, 'PROJCS[\"NAD83 / Idaho East\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.67777"
  "77777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-112.177777777778,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26968\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26969, 1, 'NAD83 / Idaho Central', 'EPSG', 26969, 'PROJCS[\"NAD83 / Idaho Central\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41"
  ".6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-114,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.999947368,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26969\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26970, 1, 'NAD83 / Idaho West', 'EPSG', 26970, 'PROJCS[\"NAD83 / Idaho West\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.67777"
  "77777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.75,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26970\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26971, 1, 'NAD83 / Illinois East', 'EPSG', 26971, 'PROJCS[\"NAD83 / Illinois East\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36"
  ".6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.3333333333333,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26971\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26972, 1, 'NAD83 / Illinois West', 'EPSG', 26972, 'PROJCS[\"NAD83 / Illinois West\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36"
  ".6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.1777777777778,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26972\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26973, 1, 'NAD83 / Indiana East', 'EPSG', 26973, 'PROJCS[\"NAD83 / Indiana East\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",37.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-85.6666666666667,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26973\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26974, 1, 'NAD83 / Indiana West', 'EPSG', 26974, 'PROJCS[\"NAD83 / Indiana West\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",37.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87.0944444444444,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fa"
  "lse easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26974\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26975, 1, 'NAD83 / Iowa North', 'EPSG', 26975, 'PROJCS[\"NAD83 / Iowa North\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.2777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",42.0777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26975"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26976, 1, 'NAD83 / Iowa South', 'EPSG', 26976, 'PROJCS[\"NAD83 / Iowa South\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "40,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.5111111111111,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",40.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26976\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26977, 1, 'NAD83 / Kansas North', 'EPSG', 26977, 'PROJCS[\"NAD83 / Kansas North\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8"
  "822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",38.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26977\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26978, 1, 'NAD83 / Kansas South', 'EPSG', 26978, 'PROJCS[\"NAD83 / Kansas South\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.5111111111111,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.5666666666667,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",37.2777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26"
  "978\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26980, 1, 'NAD83 / Kentucky South', 'EPSG', 26980, 'PROJCS[\"NAD83 / Kentucky South\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-85.75,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",36.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",500000,AUTHORITY[\"EPSG\",\"8827\"]"
  "],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26980\"]]', "
  "NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26981, 1, 'NAD83 / Louisiana North', 'EPSG', 26981, 'PROJCS[\"NAD83 / Louisiana North\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",30.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-92.5111111111111,AUTHORITY[\"EPS"
  "G\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",32.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMET"
  "ER[\"Latitude of 2nd standard parallel\",31.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false "
  "origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26981\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26982, 1, 'NAD83 / Louisiana South', 'EPSG', 26982, 'PROJCS[\"NAD83 / Louisiana South\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",28.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3333333333333,AUTHORITY[\"EPS"
  "G\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",29.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",100"
  "0000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26982\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26983, 1, 'NAD83 / Maine East', 'EPSG', 26983, 'PROJCS[\"NAD83 / Maine East\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",43.67777"
  "77777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-68.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26983\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26984, 1, 'NAD83 / Maine West', 'EPSG', 26984, 'PROJCS[\"NAD83 / Maine West\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.83333"
  "33333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.1777777777778,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",900000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26984\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26985, 1, 'NAD83 / Maryland', 'EPSG', 26985, 'PROJCS[\"NAD83 / Maryland\",GEOGCS[\"NAD83\",DAT"
  "UM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,"
  "1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],P"
  "ROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6"
  "777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77,AUTHORITY[\"EPSG\",\"8822\"]],"
  "PARAMETER[\"Latitude of 1st standard parallel\",39.45,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd stand"
  "ard parallel\",38.3111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",400000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY"
  "[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26985\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26986, 1, 'NAD83 / Massachusetts Mainland', 'EPSG', 26986, 'PROJCS[\"NAD83 / Massachusetts Mai"
  "nland\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\""
  "EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8"
  "901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUT"
  "HORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Lat"
  "itude of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-71.5111111111111,AU"
  "THORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",42.6944444444444,AUTHORITY[\"EPSG\",\"882"
  "3\"]],PARAMETER[\"Latitude of 2nd standard parallel\",41.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"East"
  "ing at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",750000,AUTHORITY[\""
  "EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"26986\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26987, 1, 'NAD83 / Massachusetts Island', 'EPSG', 26987, 'PROJCS[\"NAD83 / Massachusetts Islan"
  "d\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",41,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-70.5111111111111,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",41.2833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26987\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26988, 1, 'NAD83 / Michigan North', 'EPSG', 26988, 'PROJCS[\"NAD83 / Michigan North\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",44.7944444444444,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-87,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.0944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",45.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",8000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26988\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26989, 1, 'NAD83 / Michigan Central', 'EPSG', 26989, 'PROJCS[\"NAD83 / Michigan Central\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",43.3277777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3777777777778,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.7,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",44.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",6000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26989\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26990, 1, 'NAD83 / Michigan South', 'EPSG', 26990, 'PROJCS[\"NAD83 / Michigan South\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",41.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-84.3777777777778,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",43.6777777777778,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",42.1,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",4000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26990\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26991, 1, 'NAD83 / Minnesota North', 'EPSG', 26991, 'PROJCS[\"NAD83 / Minnesota North\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",46.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-93.1,AUTHORITY[\"EPS"
  "G\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.6333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMET"
  "ER[\"Latitude of 2nd standard parallel\",47.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false "
  "origin\",800000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26991\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26992, 1, 'NAD83 / Minnesota Central', 'EPSG', 26992, 'PROJCS[\"NAD83 / Minnesota Central\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",45,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94.25,AUTHORITY[\"EPSG\",\"88"
  "22\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.05,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2n"
  "d standard parallel\",45.6277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",800000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26993, 1, 'NAD83 / Minnesota South', 'EPSG', 26993, 'PROJCS[\"NAD83 / Minnesota South\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",43,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-94,AUTHORITY[\"EPSG\",\"8822\"]],P"
  "ARAMETER[\"Latitude of 1st standard parallel\",45.2166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of"
  " 2nd standard parallel\",43.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",800000,"
  "AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",100000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26993\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26994, 1, 'NAD83 / Mississippi East', 'EPSG', 26994, 'PROJCS[\"NAD83 / Mississippi East\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-88.8333333333333,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26994\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26995, 1, 'NAD83 / Mississippi West', 'EPSG', 26995, 'PROJCS[\"NAD83 / Mississippi West\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",29.5,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.3333333333333,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99995,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",700000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26995\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26996, 1, 'NAD83 / Missouri East', 'EPSG', 26996, 'PROJCS[\"NAD83 / Missouri East\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",35"
  ".8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26996\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26997, 1, 'NAD83 / Missouri Central', 'EPSG', 26997, 'PROJCS[\"NAD83 / Missouri Central\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",35.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-92.5111111111111,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999933333,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26997\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (26998, 1, 'NAD83 / Missouri West', 'EPSG', 26998, 'PROJCS[\"NAD83 / Missouri West\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",36"
  ".1666666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-94.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",850000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"26998\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27039, 1, 'Nahrwan 1967 / UTM zone 39N', 'EPSG', 27039, 'PROJCS[\"Nahrwan 1967 / UTM zone 39N\""
  ",GEOGCS[\"Nahrwan 1967\",DATUM[\"Nahrwan 1967\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\""
  ",\"7012\"]],TOWGS84[-253.4392,-148.452,386.5267,0.15605,0.43,-0.1013,-0.0424],AUTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4270\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "27039\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27040, 1, 'Nahrwan 1967 / UTM zone 40N', 'EPSG', 27040, 'PROJCS[\"Nahrwan 1967 / UTM zone 40N\""
  ",GEOGCS[\"Nahrwan 1967\",DATUM[\"Nahrwan 1967\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUTHORITY[\"EPSG\""
  ",\"7012\"]],TOWGS84[-253.4392,-148.452,386.5267,0.15605,0.43,-0.1013,-0.0424],AUTHORITY[\"EPSG\",\"6270\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4270\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805"
  "\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "27040\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27120, 1, 'Naparima 1972 / UTM zone 20N', 'EPSG', 27120, 'PROJCS[\"Naparima 1972 / UTM zone 20"
  "N\",GEOGCS[\"Naparima 1972\",DATUM[\"Naparima 1972\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\""
  "7022\"]],TOWGS84[-2,374,172,0,0,0,0],AUTHORITY[\"EPSG\",\"6271\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4271\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"27120\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27200, 1, 'NZGD49 / New Zealand Map Grid', 'EPSG', 27200, 'PROJCS[\"NZGD49 / New Zealand Map G"
  "rid\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORI"
  "TY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"New Zealand Map Grid\",AUTHORITY[\"EPSG\",\""
  "9811\"]],PARAMETER[\"Latitude of natural origin\",-41,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural "
  "origin\",173,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",2510000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",6023150,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\","
  "EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"27200\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27205, 1, 'NZGD49 / Mount Eden Circuit', 'EPSG', 27205, 'PROJCS[\"NZGD49 / Mount Eden Circuit\""
  ",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",-36.8798652777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",174.764339361111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,A"
  "UTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False north"
  "ing\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\""
  ",EAST],AUTHORITY[\"EPSG\",\"27205\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27206, 1, 'NZGD49 / Bay of Plenty Circuit', 'EPSG', 27206, 'PROJCS[\"NZGD49 / Bay of Plenty Ci"
  "rcuit\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHO"
  "RITY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"G"
  "reenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",-37.7612498055555,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",176.46619725,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"27206\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27207, 1, 'NZGD49 / Poverty Bay Circuit', 'EPSG', 27207, 'PROJCS[\"NZGD49 / Poverty Bay Circui"
  "t\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",-38.6247027777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",177.885636277778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"27207\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27208, 1, 'NZGD49 / Hawkes Bay Circuit', 'EPSG', 27208, 'PROJCS[\"NZGD49 / Hawkes Bay Circuit\""
  ",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",-39.6509293055556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",176.673680527778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAS"
  "T],AUTHORITY[\"EPSG\",\"27208\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27209, 1, 'NZGD49 / Taranaki Circuit', 'EPSG', 27209, 'PROJCS[\"NZGD49 / Taranaki Circuit\",GE"
  "OGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS"
  "G\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",-39.1357583055556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",174.22801175,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTH"
  "ORITY[\"EPSG\",\"27209\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27210, 1, 'NZGD49 / Tuhirangi Circuit', 'EPSG', 27210, 'PROJCS[\"NZGD49 / Tuhirangi Circuit\","
  "GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E"
  "PSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",-39.5124703888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",175.640036805556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST"
  "],AUTHORITY[\"EPSG\",\"27210\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27211, 1, 'NZGD49 / Wanganui Circuit', 'EPSG', 27211, 'PROJCS[\"NZGD49 / Wanganui Circuit\",GE"
  "OGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS"
  "G\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",-40.2419471388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",175.488099611111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",70"
  "0000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],"
  "AUTHORITY[\"EPSG\",\"27211\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27212, 1, 'NZGD49 / Wairarapa Circuit', 'EPSG', 27212, 'PROJCS[\"NZGD49 / Wairarapa Circuit\","
  "GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E"
  "PSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",-40.9255326388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",175.647349666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST"
  "],AUTHORITY[\"EPSG\",\"27212\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27213, 1, 'NZGD49 / Wellington Circuit', 'EPSG', 27213, 'PROJCS[\"NZGD49 / Wellington Circuit\""
  ",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",-41.3013196388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",174.776623111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAS"
  "T],AUTHORITY[\"EPSG\",\"27213\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27214, 1, 'NZGD49 / Collingwood Circuit', 'EPSG', 27214, 'PROJCS[\"NZGD49 / Collingwood Circui"
  "t\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",-40.7147590555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",172.6720465,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY"
  "[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",70"
  "0000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],"
  "AUTHORITY[\"EPSG\",\"27214\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27215, 1, 'NZGD49 / Nelson Circuit', 'EPSG', 27215, 'PROJCS[\"NZGD49 / Nelson Circuit\",GEOGCS"
  "[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",-41.2745447222222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",173.299316805556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTH"
  "ORITY[\"EPSG\",\"27215\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27216, 1, 'NZGD49 / Karamea Circuit', 'EPSG', 27216, 'PROJCS[\"NZGD49 / Karamea Circuit\",GEOG"
  "CS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",-41.2899115277778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",172.109028194444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",70000"
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUT"
  "HORITY[\"EPSG\",\"27216\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27217, 1, 'NZGD49 / Buller Circuit', 'EPSG', 27217, 'PROJCS[\"NZGD49 / Buller Circuit\",GEOGCS"
  "[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",-41.8108028611111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",171.581260055556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTH"
  "ORITY[\"EPSG\",\"27217\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27218, 1, 'NZGD49 / Grey Circuit', 'EPSG', 27218, 'PROJCS[\"NZGD49 / Grey Circuit\",GEOGCS[\"N"
  "ZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",-42.3336942777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural ori"
  "gin\",171.549771305556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORITY"
  "[\"EPSG\",\"27218\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27219, 1, 'NZGD49 / Amuri Circuit', 'EPSG', 27219, 'PROJCS[\"NZGD49 / Amuri Circuit\",GEOGCS[\""
  "NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7"
  "022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",-42.6891165833333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",173.010133388889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORI"
  "TY[\"EPSG\",\"27219\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27220, 1, 'NZGD49 / Marlborough Circuit', 'EPSG', 27220, 'PROJCS[\"NZGD49 / Marlborough Circui"
  "t\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",-41.5444866666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",173.802074111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"27220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27221, 1, 'NZGD49 / Hokitika Circuit', 'EPSG', 27221, 'PROJCS[\"NZGD49 / Hokitika Circuit\",GE"
  "OGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPS"
  "G\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],"
  "PARAMETER[\"Latitude of natural origin\",-42.8863223611111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",170.9799935,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EP"
  "SG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000,"
  "AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHO"
  "RITY[\"EPSG\",\"27221\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27222, 1, 'NZGD49 / Okarito Circuit', 'EPSG', 27222, 'PROJCS[\"NZGD49 / Okarito Circuit\",GEOG"
  "CS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,"
  "AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",-43.1101281388889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",170.260925833333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\""
  "EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",70000"
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUT"
  "HORITY[\"EPSG\",\"27222\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27223, 1, 'NZGD49 / Jacksons Bay Circuit', 'EPSG', 27223, 'PROJCS[\"NZGD49 / Jacksons Bay Circ"
  "uit\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORI"
  "TY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",-43.9778028888889,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",168.606267,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST"
  "],AUTHORITY[\"EPSG\",\"27223\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27224, 1, 'NZGD49 / Mount Pleasant Circuit', 'EPSG', 27224, 'PROJCS[\"NZGD49 / Mount Pleasant "
  "Circuit\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUT"
  "HORITY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",-43.5906375833333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",172.727193583333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E"
  "\",EAST],AUTHORITY[\"EPSG\",\"27224\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27225, 1, 'NZGD49 / Gawler Circuit', 'EPSG', 27225, 'PROJCS[\"NZGD49 / Gawler Circuit\",GEOGCS"
  "[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",-43.7487115555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",171.360748472222,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTH"
  "ORITY[\"EPSG\",\"27225\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27226, 1, 'NZGD49 / Timaru Circuit', 'EPSG', 27226, 'PROJCS[\"NZGD49 / Timaru Circuit\",GEOGCS"
  "[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\","
  "\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",-44.4022203611111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",171.057250833333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"E"
  "PSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",700000"
  ",AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTH"
  "ORITY[\"EPSG\",\"27226\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27227, 1, 'NZGD49 / Lindis Peak Circuit', 'EPSG', 27227, 'PROJCS[\"NZGD49 / Lindis Peak Circui"
  "t\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY"
  "[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",-44.7352679722222,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",169.467755083333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing"
  "\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",E"
  "AST],AUTHORITY[\"EPSG\",\"27227\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27228, 1, 'NZGD49 / Mount Nicholas Circuit', 'EPSG', 27228, 'PROJCS[\"NZGD49 / Mount Nicholas "
  "Circuit\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUT"
  "HORITY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\""
  "Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\""
  "Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\""
  "9807\"]],PARAMETER[\"Latitude of natural origin\",-45.1329025833333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitu"
  "de of natural origin\",168.398641194444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,"
  "AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E"
  "\",EAST],AUTHORITY[\"EPSG\",\"27228\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27229, 1, 'NZGD49 / Mount York Circuit', 'EPSG', 27229, 'PROJCS[\"NZGD49 / Mount York Circuit\""
  ",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\""
  "EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",-45.5637261666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of "
  "natural origin\",167.738861777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAS"
  "T],AUTHORITY[\"EPSG\",\"27229\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27230, 1, 'NZGD49 / Observation Point Circuit', 'EPSG', 27230, 'PROJCS[\"NZGD49 / Observation "
  "Point Circuit\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,2"
  "97,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PR"
  "IMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]]"
  ",AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\""
  "EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",-45.8161966111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER["
  "\"Longitude of natural origin\",170.628595166667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural or"
  "igin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"F"
  "alse northing\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH]"
  ",AXIS[\"E\",EAST],AUTHORITY[\"EPSG\",\"27230\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27231, 1, 'NZGD49 / North Taieri Circuit', 'EPSG', 27231, 'PROJCS[\"NZGD49 / North Taieri Circ"
  "uit\",GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORI"
  "TY[\"EPSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Gre"
  "enwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"La"
  "t\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9"
  "807\"]],PARAMETER[\"Latitude of natural origin\",-45.8615133611111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitud"
  "e of natural origin\",170.282589111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False"
  " northing\",700000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXI"
  "S[\"E\",EAST],AUTHORITY[\"EPSG\",\"27231\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27232, 1, 'NZGD49 / Bluff Circuit', 'EPSG', 27232, 'PROJCS[\"NZGD49 / Bluff Circuit\",GEOGCS[\""
  "NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7"
  "022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTH"
  "ORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMET"
  "ER[\"Latitude of natural origin\",-46.6000096111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",168.342872,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",300002.66,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",699999.58,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",EAST],AUTHORI"
  "TY[\"EPSG\",\"27232\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27258, 1, 'NZGD49 / UTM zone 58S', 'EPSG', 27258, 'PROJCS[\"NZGD49 / UTM zone 58S\",GEOGCS[\"N"
  "ZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"27258\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27259, 1, 'NZGD49 / UTM zone 59S', 'EPSG', 27259, 'PROJCS[\"NZGD49 / UTM zone 59S\",GEOGCS[\"N"
  "ZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"27259\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27260, 1, 'NZGD49 / UTM zone 60S', 'EPSG', 27260, 'PROJCS[\"NZGD49 / UTM zone 60S\",GEOGCS[\"N"
  "ZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"70"
  "22\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"27260\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27291, 1, 'NZGD49 / North Island Grid', 'EPSG', 27291, 'PROJCS[\"NZGD49 / North Island Grid\","
  "GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E"
  "PSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",175.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",400000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"British yard (Sears 1922)\",0.9143984146160287,AUTHORITY[\"EPSG\",\"9040\"]],AXIS[\"E\",EAST],AXIS["
  "\"N\",NORTH],AUTHORITY[\"EPSG\",\"27291\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27292, 1, 'NZGD49 / South Island Grid', 'EPSG', 27292, 'PROJCS[\"NZGD49 / South Island Grid\","
  "GEOGCS[\"NZGD49\",DATUM[\"New Zealand Geodetic Datum 1949\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"E"
  "PSG\",\"7022\"]],TOWGS84[59.47,-5.04,187.44,0.47,-0.1,1.024,-4.5993],AUTHORITY[\"EPSG\",\"6272\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4272\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",-44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",171.5,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PAR"
  "AMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",500000,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"British yard (Sears 1922)\",0.9143984146160287,AUTHORITY[\"EPSG\",\"9040\"]],AXIS[\"E\",EAST],AXIS["
  "\"N\",NORTH],AUTHORITY[\"EPSG\",\"27292\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27391, 1, 'NGO 1948 (Oslo) / NGO zone I', 'EPSG', 27391, 'PROJCS[\"NGO 1948 (Oslo) / NGO zone "
  "I\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHORI"
  "TY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-4.66666666666667,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting"
  "\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27391\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27392, 1, 'NGO 1948 (Oslo) / NGO zone II', 'EPSG', 27392, 'PROJCS[\"NGO 1948 (Oslo) / NGO zone"
  " II\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHO"
  "RITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-2.33333333333333,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27392\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27393, 1, 'NGO 1948 (Oslo) / NGO zone III', 'EPSG', 27393, 'PROJCS[\"NGO 1948 (Oslo) / NGO zon"
  "e III\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUT"
  "HORITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORI"
  "TY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG"
  "\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27393\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27394, 1, 'NGO 1948 (Oslo) / NGO zone IV', 'EPSG', 27394, 'PROJCS[\"NGO 1948 (Oslo) / NGO zone"
  " IV\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHO"
  "RITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",2.51111111111111,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27394\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27395, 1, 'NGO 1948 (Oslo) / NGO zone V', 'EPSG', 27395, 'PROJCS[\"NGO 1948 (Oslo) / NGO zone "
  "V\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHORI"
  "TY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural "
  "origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",6.17777777777778,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27395\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27396, 1, 'NGO 1948 (Oslo) / NGO zone VI', 'EPSG', 27396, 'PROJCS[\"NGO 1948 (Oslo) / NGO zone"
  " VI\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUTHO"
  "RITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10.1777777777778,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27396\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27397, 1, 'NGO 1948 (Oslo) / NGO zone VII', 'EPSG', 27397, 'PROJCS[\"NGO 1948 (Oslo) / NGO zon"
  "e VII\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,AUT"
  "HORITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"8913"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natu"
  "ral origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",14.1777777777778,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27397\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27398, 1, 'NGO 1948 (Oslo) / NGO zone VIII', 'EPSG', 27398, 'PROJCS[\"NGO 1948 (Oslo) / NGO zo"
  "ne VIII\",GEOGCS[\"NGO 1948 (Oslo)\",DATUM[\"NGO 1948 (Oslo)\",SPHEROID[\"Bessel Modified\",6377492.018,299.1528128,A"
  "UTHORITY[\"EPSG\",\"7005\"]],AUTHORITY[\"EPSG\",\"6817\"]],PRIMEM[\"Oslo\",10.722916666666666,AUTHORITY[\"EPSG\",\"89"
  "13\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTH"
  "ORITY[\"EPSG\",\"4817\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of na"
  "tural origin\",58,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",18.3444444444444,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"x\",NORTH],AXIS[\"y\",EAST],AUTHORITY[\"EPSG\",\"27398\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27429, 1, 'Datum 73 / UTM zone 29N', 'EPSG', 27429, 'PROJCS[\"Datum 73 / UTM zone 29N\",GEOGCS"
  "[\"Datum 73\",DATUM[\"Datum 73\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-2"
  "39.749,88.181,30.488,0.263,0.082,1.211,2.229],AUTHORITY[\"EPSG\",\"6274\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4274\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,A"
  "UTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"27429\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27493, 1, 'Datum 73 / Modified Portuguese Grid', 'EPSG', 27493, 'PROJCS[\"Datum 73 / Modified "
  "Portuguese Grid\",GEOGCS[\"Datum 73\",DATUM[\"Datum 73\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG"
  "\",\"7022\"]],TOWGS84[-239.749,88.181,30.488,0.263,0.082,1.211,2.229],AUTHORITY[\"EPSG\",\"6274\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4274\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of n"
  "atural origin\",-8.13190611111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",180.598,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",-86.99,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORT"
  "H],AUTHORITY[\"EPSG\",\"27493\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27500, 1, 'ATF (Paris) / Nord de Guerre', 'EPSG', 27500, 'PROJCS[\"ATF (Paris) / Nord de Guerr"
  "e\",GEOGCS[\"ATF (Paris)\",DATUM[\"Ancienne Triangulation Francaise (Paris)\",SPHEROID[\"Plessis 1817\",6376523,308.6"
  "4,AUTHORITY[\"EPSG\",\"7027\"]],AUTHORITY[\"EPSG\",\"6901\"]],PRIMEM[\"Paris RGS\",2.5968981481481492,AUTHORITY[\"EPS"
  "G\",\"8914\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4901\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\""
  "Latitude of natural origin\",55,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",6,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99950908,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"F"
  "alse easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"27500\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27561, 1, 'NTF (Paris) / Lambert Nord France', 'EPSG', 27561, 'PROJCS[\"NTF (Paris) / Lambert "
  "Nord France\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\""
  ",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,A"
  "UTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS"
  "[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\""
  "]],PARAMETER[\"Latitude of natural origin\",55,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999877341,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"27561\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27562, 1, 'NTF (Paris) / Lambert Centre France', 'EPSG', 27562, 'PROJCS[\"NTF (Paris) / Lamber"
  "t Centre France\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (I"
  "GN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969"
  "213,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9"
  "801\"]],PARAMETER[\"Latitude of natural origin\",52,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99987742,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORIT"
  "Y[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"E"
  "PSG\",\"27562\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27563, 1, 'NTF (Paris) / Lambert Sud France', 'EPSG', 27563, 'PROJCS[\"NTF (Paris) / Lambert S"
  "ud France\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\","
  "6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AU"
  "THORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]"
  "],PARAMETER[\"Latitude of natural origin\",49,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999877499,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",200000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"27563\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27564, 1, 'NTF (Paris) / Lambert Corse', 'EPSG', 27564, 'PROJCS[\"NTF (Paris) / Lambert Corse\""
  ",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,2"
  "93.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"E"
  "PSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER"
  "[\"Latitude of natural origin\",46.85,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99994471,AUTHORITY[\"EPSG\",\"8805\"]],PARAME"
  "TER[\"False easting\",234.358,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",185861.369,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"27"
  "564\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27571, 1, 'NTF (Paris) / Lambert zone I', 'EPSG', 27571, 'PROJCS[\"NTF (Paris) / Lambert zone "
  "I\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378249."
  "2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY["
  "\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAME"
  "TER[\"Latitude of natural origin\",55,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999877341,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1200000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"27571"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27572, 1, 'NTF (Paris) / Lambert zone II', 'EPSG', 27572, 'PROJCS[\"NTF (Paris) / Lambert zone"
  " II\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",637824"
  "9.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AUTHORIT"
  "Y[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAM"
  "ETER[\"Latitude of natural origin\",52,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99987742,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",2200000,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"27572"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27573, 1, 'NTF (Paris) / Lambert zone III', 'EPSG', 27573, 'PROJCS[\"NTF (Paris) / Lambert zon"
  "e III\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",6378"
  "249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AUTHOR"
  "ITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lo"
  "n\",EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PA"
  "RAMETER[\"Latitude of natural origin\",49,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999877499,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",3200000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "27573\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27574, 1, 'NTF (Paris) / Lambert zone IV', 'EPSG', 27574, 'PROJCS[\"NTF (Paris) / Lambert zone"
  " IV\",GEOGCS[\"NTF (Paris)\",DATUM[\"Nouvelle Triangulation Francaise (Paris)\",SPHEROID[\"Clarke 1880 (IGN)\",637824"
  "9.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6807\"]],PRIMEM[\"Paris\",2.5969213,AUTHORIT"
  "Y[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4807\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAM"
  "ETER[\"Latitude of natural origin\",46.85,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,A"
  "UTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99994471,AUTHORITY[\"EPSG\",\"8805\"]],PA"
  "RAMETER[\"False easting\",234.358,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4185861.369,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"27574\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (27700, 1, 'OSGB 1936 / British National Grid', 'EPSG', 27700, 'PROJCS[\"OSGB 1936 / British Na"
  "tional Grid\",GEOGCS[\"OSGB 1936\",DATUM[\"OSGB 1936\",SPHEROID[\"Airy 1830\",6377563.396,299.3249646,AUTHORITY[\"EPS"
  "G\",\"7001\"]],TOWGS84[446.448,-125.157,542.06,0.15,0.247,0.842,-20.489],AUTHORITY[\"EPSG\",\"6277\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4277\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",49,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origi"
  "n\",-2,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996012717,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-100000,AUTHORI"
  "TY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\""
  "EPSG\",\"27700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28191, 1, 'Palestine 1923 / Palestine Grid', 'EPSG', 28191, 'PROJCS[\"Palestine 1923 / Palesti"
  "ne Grid\",GEOGCS[\"Palestine 1923\",DATUM[\"Palestine 1923\",SPHEROID[\"Clarke 1880 (Benoit)\",6378300.789,293.466315"
  "5389811,AUTHORITY[\"EPSG\",\"7010\"]],TOWGS84[-275.7224,94.7824,340.8944,-8.001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\""
  "6281\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4281\"]],PROJECTION[\"Cassini-Soldner\",AUTHO"
  "RITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",31.7340969444444,AUTHORITY[\"EPSG\",\"8801\"]],PARA"
  "METER[\"Longitude of natural origin\",35.2120805555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1702"
  "51.555,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",126867.909,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"met"
  "re\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28192, 1, 'Palestine 1923 / Palestine Belt', 'EPSG', 28192, 'PROJCS[\"Palestine 1923 / Palesti"
  "ne Belt\",GEOGCS[\"Palestine 1923\",DATUM[\"Palestine 1923\",SPHEROID[\"Clarke 1880 (Benoit)\",6378300.789,293.466315"
  "5389811,AUTHORITY[\"EPSG\",\"7010\"]],TOWGS84[-275.7224,94.7824,340.8944,-8.001,-4.42,-11.821,1],AUTHORITY[\"EPSG\",\""
  "6281\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4281\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31.7340969444444,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",35.2120805555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at "
  "natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",170251.555,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",1126867.909,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]"
  "],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28193, 1, 'Palestine 1923 / Israeli CS Grid', 'EPSG', 28193, 'PROJCS[\"Palestine 1923 / Israel"
  "i CS Grid\",GEOGCS[\"Palestine 1923\",DATUM[\"Palestine 1923\",SPHEROID[\"Clarke 1880 (Benoit)\",6378300.789,293.4663"
  "155389811,AUTHORITY[\"EPSG\",\"7010\"]],TOWGS84[-275.7224,94.7824,340.8944,-8.001,-4.42,-11.821,1],AUTHORITY[\"EPSG\""
  ",\"6281\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4281\"]],PROJECTION[\"Cassini-Soldner\",AU"
  "THORITY[\"EPSG\",\"9806\"]],PARAMETER[\"Latitude of natural origin\",31.7340969444444,AUTHORITY[\"EPSG\",\"8801\"]],P"
  "ARAMETER[\"Longitude of natural origin\",35.2120805555556,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",1"
  "70251.555,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",1126867.909,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28232, 1, 'Pointe Noire / UTM zone 32S', 'EPSG', 28232, 'PROJCS[\"Pointe Noire / UTM zone 32S\""
  ",GEOGCS[\"Pointe Noire\",DATUM[\"Congo 1960 Pointe Noire\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269"
  ",AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-148,51,-291,0,0,0,0],AUTHORITY[\"EPSG\",\"6282\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4282\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28232\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28348, 1, 'GDA94 / MGA zone 48', 'EPSG', 28348, 'PROJCS[\"GDA94 / MGA zone 48\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28349, 1, 'GDA94 / MGA zone 49', 'EPSG', 28349, 'PROJCS[\"GDA94 / MGA zone 49\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28349\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28350, 1, 'GDA94 / MGA zone 50', 'EPSG', 28350, 'PROJCS[\"GDA94 / MGA zone 50\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28351, 1, 'GDA94 / MGA zone 51', 'EPSG', 28351, 'PROJCS[\"GDA94 / MGA zone 51\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28352, 1, 'GDA94 / MGA zone 52', 'EPSG', 28352, 'PROJCS[\"GDA94 / MGA zone 52\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28353, 1, 'GDA94 / MGA zone 53', 'EPSG', 28353, 'PROJCS[\"GDA94 / MGA zone 53\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28354, 1, 'GDA94 / MGA zone 54', 'EPSG', 28354, 'PROJCS[\"GDA94 / MGA zone 54\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28355, 1, 'GDA94 / MGA zone 55', 'EPSG', 28355, 'PROJCS[\"GDA94 / MGA zone 55\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28356, 1, 'GDA94 / MGA zone 56', 'EPSG', 28356, 'PROJCS[\"GDA94 / MGA zone 56\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28357, 1, 'GDA94 / MGA zone 57', 'EPSG', 28357, 'PROJCS[\"GDA94 / MGA zone 57\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28357\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28358, 1, 'GDA94 / MGA zone 58', 'EPSG', 28358, 'PROJCS[\"GDA94 / MGA zone 58\",GEOGCS[\"GDA94"
  "\",DATUM[\"Geocentric Datum of Australia 1994\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019"
  "\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6283\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4283\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"28358\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28404, 1, 'Pulkovo 1942 / Gauss-Kruger zone 4', 'EPSG', 28404, 'PROJCS[\"Pulkovo 1942 / Gauss-"
  "Kruger zone 4\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28404\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28405, 1, 'Pulkovo 1942 / Gauss-Kruger zone 5', 'EPSG', 28405, 'PROJCS[\"Pulkovo 1942 / Gauss-"
  "Kruger zone 5\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28405\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28406, 1, 'Pulkovo 1942 / Gauss-Kruger zone 6', 'EPSG', 28406, 'PROJCS[\"Pulkovo 1942 / Gauss-"
  "Kruger zone 6\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",6500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28406\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28407, 1, 'Pulkovo 1942 / Gauss-Kruger zone 7', 'EPSG', 28407, 'PROJCS[\"Pulkovo 1942 / Gauss-"
  "Kruger zone 7\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",7500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28407\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28408, 1, 'Pulkovo 1942 / Gauss-Kruger zone 8', 'EPSG', 28408, 'PROJCS[\"Pulkovo 1942 / Gauss-"
  "Kruger zone 8\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",8500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28408\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28409, 1, 'Pulkovo 1942 / Gauss-Kruger zone 9', 'EPSG', 28409, 'PROJCS[\"Pulkovo 1942 / Gauss-"
  "Kruger zone 9\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY[\""
  "EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHORIT"
  "Y[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"L"
  "on\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\""
  "Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False east"
  "ing\",9500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28409\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28410, 1, 'Pulkovo 1942 / Gauss-Kruger zone 10', 'EPSG', 28410, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 10\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",10500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28410\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28411, 1, 'Pulkovo 1942 / Gauss-Kruger zone 11', 'EPSG', 28411, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 11\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",11500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28411\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28412, 1, 'Pulkovo 1942 / Gauss-Kruger zone 12', 'EPSG', 28412, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 12\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",12500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28412\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28413, 1, 'Pulkovo 1942 / Gauss-Kruger zone 13', 'EPSG', 28413, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 13\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",13500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28413\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28414, 1, 'Pulkovo 1942 / Gauss-Kruger zone 14', 'EPSG', 28414, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 14\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",14500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28414\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28415, 1, 'Pulkovo 1942 / Gauss-Kruger zone 15', 'EPSG', 28415, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 15\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",15500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28415\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28416, 1, 'Pulkovo 1942 / Gauss-Kruger zone 16', 'EPSG', 28416, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 16\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",16500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28416\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28417, 1, 'Pulkovo 1942 / Gauss-Kruger zone 17', 'EPSG', 28417, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 17\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORI"
  "TY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",17500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28417\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28418, 1, 'Pulkovo 1942 / Gauss-Kruger zone 18', 'EPSG', 28418, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 18\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",18500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28418\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28419, 1, 'Pulkovo 1942 / Gauss-Kruger zone 19', 'EPSG', 28419, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 19\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",19500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28419\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28420, 1, 'Pulkovo 1942 / Gauss-Kruger zone 20', 'EPSG', 28420, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 20\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",20500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28420\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28421, 1, 'Pulkovo 1942 / Gauss-Kruger zone 21', 'EPSG', 28421, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 21\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",21500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28421\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28422, 1, 'Pulkovo 1942 / Gauss-Kruger zone 22', 'EPSG', 28422, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 22\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",22500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28422\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28423, 1, 'Pulkovo 1942 / Gauss-Kruger zone 23', 'EPSG', 28423, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 23\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",23500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28423\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28424, 1, 'Pulkovo 1942 / Gauss-Kruger zone 24', 'EPSG', 28424, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 24\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",24500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28424\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28425, 1, 'Pulkovo 1942 / Gauss-Kruger zone 25', 'EPSG', 28425, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 25\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",25500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28425\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28426, 1, 'Pulkovo 1942 / Gauss-Kruger zone 26', 'EPSG', 28426, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 26\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",26500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28426\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28427, 1, 'Pulkovo 1942 / Gauss-Kruger zone 27', 'EPSG', 28427, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 27\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",27500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28427\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28428, 1, 'Pulkovo 1942 / Gauss-Kruger zone 28', 'EPSG', 28428, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 28\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",28500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28428\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28429, 1, 'Pulkovo 1942 / Gauss-Kruger zone 29', 'EPSG', 28429, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 29\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",29500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28429\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28430, 1, 'Pulkovo 1942 / Gauss-Kruger zone 30', 'EPSG', 28430, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 30\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",30500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28430\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28431, 1, 'Pulkovo 1942 / Gauss-Kruger zone 31', 'EPSG', 28431, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 31\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",31500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28431\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28432, 1, 'Pulkovo 1942 / Gauss-Kruger zone 32', 'EPSG', 28432, 'PROJCS[\"Pulkovo 1942 / Gauss"
  "-Kruger zone 32\",GEOGCS[\"Pulkovo 1942\",DATUM[\"Pulkovo 1942\",SPHEROID[\"Krassowsky 1940\",6378245,298.3,AUTHORITY"
  "[\"EPSG\",\"7024\"]],TOWGS84[25,-141,-78.5,0,0.35,0.736,0],AUTHORITY[\"EPSG\",\"6284\"]],PRIMEM[\"Greenwich\",0,AUTHO"
  "RITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS["
  "\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4284\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETE"
  "R[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",32500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"28432\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28600, 1, 'Qatar 1974 / Qatar National Grid', 'EPSG', 28600, 'PROJCS[\"Qatar 1974 / Qatar Nati"
  "onal Grid\",GEOGCS[\"Qatar 1974\",DATUM[\"Qatar 1974\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\""
  ",\"7022\"]],TOWGS84[-128.033,-283.697,21.052,0,0,0,0],AUTHORITY[\"EPSG\",\"6285\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4285\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",24.4611111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "51.2166666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999,AUTHORITY[\"EPSG\""
  ",\"8805\"]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTH"
  "ORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY"
  "[\"EPSG\",\"28600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28991, 1, 'Amersfoort / RD Old', 'EPSG', 28991, 'PROJCS[\"Amersfoort / RD Old\",GEOGCS[\"Amers"
  "foort\",DATUM[\"Amersfoort\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[5"
  "65.4171,50.3319,465.5524,-0.398957388243134,0.343987817378283,-1.87740163998045,4.0725],AUTHORITY[\"EPSG\",\"6289\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4289\"]],PROJECTION[\"Oblique Stereographic\",AUTHORIT"
  "Y[\"EPSG\",\"9809\"]],PARAMETER[\"Latitude of natural origin\",52.1561605555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",5.38763888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9999079,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"28991\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (28992, 1, 'Amersfoort / RD New', 'EPSG', 28992, 'PROJCS[\"Amersfoort / RD New\",GEOGCS[\"Amers"
  "foort\",DATUM[\"Amersfoort\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[5"
  "65.4171,50.3319,465.5524,-0.398957388243134,0.343987817378283,-1.87740163998045,4.0725],AUTHORITY[\"EPSG\",\"6289\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4289\"]],PROJECTION[\"Oblique Stereographic\",AUTHORIT"
  "Y[\"EPSG\",\"9809\"]],PARAMETER[\"Latitude of natural origin\",52.1561605555556,AUTHORITY[\"EPSG\",\"8801\"]],PARAMET"
  "ER[\"Longitude of natural origin\",5.38763888888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9999079,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",155000,AUTHORITY[\"EPSG\",\"8806\"]],PA"
  "RAMETER[\"False northing\",463000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"28992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29101, 1, 'SAD69 / Brazil Polyconic', 'EPSG', 29101, 'PROJCS[\"SAD69 / Brazil Polyconic\",GEOG"
  "CS[\"SAD69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7"
  "050\"]],TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4618\"]],PROJECTION[\"American Polyconic\",AUTHORITY[\"EPSG\",\"9818\"]],PARAMETER[\"Latitude of natural o"
  "rigin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY[\"EPSG\",\"8802\"]],P"
  "ARAMETER[\"False easting\",5000000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"29101\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29168, 1, 'SAD69 / UTM zone 18N', 'EPSG', 29168, 'PROJCS[\"SAD69 / UTM zone 18N\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29168\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29169, 1, 'SAD69 / UTM zone 19N', 'EPSG', 29169, 'PROJCS[\"SAD69 / UTM zone 19N\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29169\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29170, 1, 'SAD69 / UTM zone 20N', 'EPSG', 29170, 'PROJCS[\"SAD69 / UTM zone 20N\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29170\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29171, 1, 'SAD69 / UTM zone 21N', 'EPSG', 29171, 'PROJCS[\"SAD69 / UTM zone 21N\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29172, 1, 'SAD69 / UTM zone 22N', 'EPSG', 29172, 'PROJCS[\"SAD69 / UTM zone 22N\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29172\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29187, 1, 'SAD69 / UTM zone 17S', 'EPSG', 29187, 'PROJCS[\"SAD69 / UTM zone 17S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29187\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29188, 1, 'SAD69 / UTM zone 18S', 'EPSG', 29188, 'PROJCS[\"SAD69 / UTM zone 18S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29188\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29189, 1, 'SAD69 / UTM zone 19S', 'EPSG', 29189, 'PROJCS[\"SAD69 / UTM zone 19S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29189\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29190, 1, 'SAD69 / UTM zone 20S', 'EPSG', 29190, 'PROJCS[\"SAD69 / UTM zone 20S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29190\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29191, 1, 'SAD69 / UTM zone 21S', 'EPSG', 29191, 'PROJCS[\"SAD69 / UTM zone 21S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29192, 1, 'SAD69 / UTM zone 22S', 'EPSG', 29192, 'PROJCS[\"SAD69 / UTM zone 22S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29193, 1, 'SAD69 / UTM zone 23S', 'EPSG', 29193, 'PROJCS[\"SAD69 / UTM zone 23S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29194, 1, 'SAD69 / UTM zone 24S', 'EPSG', 29194, 'PROJCS[\"SAD69 / UTM zone 24S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29194\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29195, 1, 'SAD69 / UTM zone 25S', 'EPSG', 29195, 'PROJCS[\"SAD69 / UTM zone 25S\",GEOGCS[\"SAD"
  "69\",DATUM[\"South American Datum 1969\",SPHEROID[\"GRS 1967 Modified\",6378160,298.25,AUTHORITY[\"EPSG\",\"7050\"]],"
  "TOWGS84[-64,0,-32,0,0,0,0],AUTHORITY[\"EPSG\",\"6618\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4618\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29220, 1, 'Sapper Hill 1943 / UTM zone 20S', 'EPSG', 29220, 'PROJCS[\"Sapper Hill 1943 / UTM z"
  "one 20S\",GEOGCS[\"Sapper Hill 1943\",DATUM[\"Sapper Hill 1943\",SPHEROID[\"International 1924\",6378388,297,AUTHORIT"
  "Y[\"EPSG\",\"7022\"]],TOWGS84[-355,21,72,0,0,0,0],AUTHORITY[\"EPSG\",\"6292\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4292\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29221, 1, 'Sapper Hill 1943 / UTM zone 21S', 'EPSG', 29221, 'PROJCS[\"Sapper Hill 1943 / UTM z"
  "one 21S\",GEOGCS[\"Sapper Hill 1943\",DATUM[\"Sapper Hill 1943\",SPHEROID[\"International 1924\",6378388,297,AUTHORIT"
  "Y[\"EPSG\",\"7022\"]],TOWGS84[-355,21,72,0,0,0,0],AUTHORITY[\"EPSG\",\"6292\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4292\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latit"
  "ude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29221\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29333, 1, 'Schwarzeck / UTM zone 33S', 'EPSG', 29333, 'PROJCS[\"Schwarzeck / UTM zone 33S\",GE"
  "OGCS[\"Schwarzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\""
  "EPSG\",\"7046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29371, 1, 'Schwarzeck / Lo22/11', 'EPSG', 29371, 'PROJCS[\"Schwarzeck / Lo22/11\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",11,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29371"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29373, 1, 'Schwarzeck / Lo22/13', 'EPSG', 29373, 'PROJCS[\"Schwarzeck / Lo22/13\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29373"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29375, 1, 'Schwarzeck / Lo22/15', 'EPSG', 29375, 'PROJCS[\"Schwarzeck / Lo22/15\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29375"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29377, 1, 'Schwarzeck / Lo22/17', 'EPSG', 29377, 'PROJCS[\"Schwarzeck / Lo22/17\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",17,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29377"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29379, 1, 'Schwarzeck / Lo22/19', 'EPSG', 29379, 'PROJCS[\"Schwarzeck / Lo22/19\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",19,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29379"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29381, 1, 'Schwarzeck / Lo22/21', 'EPSG', 29381, 'PROJCS[\"Schwarzeck / Lo22/21\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29381"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29383, 1, 'Schwarzeck / Lo22/23', 'EPSG', 29383, 'PROJCS[\"Schwarzeck / Lo22/23\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",23,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29383"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29385, 1, 'Schwarzeck / Lo22/25', 'EPSG', 29385, 'PROJCS[\"Schwarzeck / Lo22/25\",GEOGCS[\"Sch"
  "warzeck\",DATUM[\"Schwarzeck\",SPHEROID[\"Bessel Namibia (GLM)\",6377483.865280419,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "046\"]],TOWGS84[616,97,-251,0,0,0,0],AUTHORITY[\"EPSG\",\"6293\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4293\"]],PROJECTION[\"Transverse Mercator (South Orientated)\",AUTHORITY[\"EPSG\",\"9808\"]],PARAMETER[\""
  "Latitude of natural origin\",-22,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",25,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"German l"
  "egal metre\",1.0000135965,AUTHORITY[\"EPSG\",\"9031\"]],AXIS[\"Y\",WEST],AXIS[\"X\",SOUTH],AUTHORITY[\"EPSG\",\"29385"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29701, 1, 'Tananarive (Paris) / Laborde Grid', 'EPSG', 29701, 'PROJCS[\"Tananarive (Paris) / L"
  "aborde Grid\",GEOGCS[\"Tananarive (Paris)\",DATUM[\"Tananarive 1925 (Paris)\",SPHEROID[\"International 1924\",6378388"
  ",297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6810\"]],PRIMEM[\"Paris\",2.5969213,AUTHORITY[\"EPSG\",\"8903"
  "\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4810\"]],PROJECTION[\"Laborde Oblique Mercator\",AUTHORITY[\"EPSG\",\"9813\"]],PARAMETER[\"Latitude of pr"
  "ojection centre\",-21,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of projection centre\",49,AUTHORITY[\"EPSG\""
  ",\"8812\"]],PARAMETER[\"Azimuth of initial line\",21,AUTHORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Scale factor on initia"
  "l line\",0.9995,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",800000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\""
  ",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"29701\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29702, 1, 'Tananarive (Paris) / Laborde Grid approximation', 'EPSG', 29702, 'PROJCS[\"Tananari"
  "ve (Paris) / Laborde Grid approximation\",GEOGCS[\"Tananarive (Paris)\",DATUM[\"Tananarive 1925 (Paris)\",SPHEROID[\""
  "International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],AUTHORITY[\"EPSG\",\"6810\"]],PRIMEM[\"Paris\",2.59692"
  "13,AUTHORITY[\"EPSG\",\"8903\"]],UNIT[\"grad\",0.01570796326794895,AUTHORITY[\"EPSG\",\"9105\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4810\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG\""
  ",\"9815\"]],PARAMETER[\"Latitude of projection centre\",-21,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of pr"
  "ojection centre\",49,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",21,AUTHORITY[\"EPSG\",\"8813"
  "\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",21,AUTHORITY[\"EPSG\",\"8814\"]],PARAMETER[\"Scale factor on ini"
  "tial line\",0.9995,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection centre\",400000,AUTHORITY[\"EPSG\""
  ",\"8816\"]],PARAMETER[\"Northing at projection centre\",800000,AUTHORITY[\"EPSG\",\"8817\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"29702\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29738, 1, 'Tananarive / UTM zone 38S', 'EPSG', 29738, 'PROJCS[\"Tananarive / UTM zone 38S\",GE"
  "OGCS[\"Tananarive\",DATUM[\"Tananarive 1925\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-198.383,-240.517,-107.909,0,0,0,0],AUTHORITY[\"EPSG\",\"6297\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4297\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29739, 1, 'Tananarive / UTM zone 39S', 'EPSG', 29739, 'PROJCS[\"Tananarive / UTM zone 39S\",GE"
  "OGCS[\"Tananarive\",DATUM[\"Tananarive 1925\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\""
  "]],TOWGS84[-198.383,-240.517,-107.909,0,0,0,0],AUTHORITY[\"EPSG\",\"6297\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4297\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\""
  ",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29739\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29849, 1, 'Timbalai 1948 / UTM zone 49N', 'EPSG', 29849, 'PROJCS[\"Timbalai 1948 / UTM zone 49"
  "N\",GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai 1948\",SPHEROID[\"Everest 1830 (1967 Definition)\",6377298.556,300.8017"
  ",AUTHORITY[\"EPSG\",\"7016\"]],TOWGS84[-679,669,-48,0,0,0,0],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4298\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29849\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29850, 1, 'Timbalai 1948 / UTM zone 50N', 'EPSG', 29850, 'PROJCS[\"Timbalai 1948 / UTM zone 50"
  "N\",GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai 1948\",SPHEROID[\"Everest 1830 (1967 Definition)\",6377298.556,300.8017"
  ",AUTHORITY[\"EPSG\",\"7016\"]],TOWGS84[-679,669,-48,0,0,0,0],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"Greenwich\",0,AUT"
  "HORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXI"
  "S[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4298\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAME"
  "TER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTH"
  "ORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER"
  "[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],"
  "UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29850\"]]', NU"
  "LL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29871, 1, 'Timbalai 1948 / RSO Borneo (ch)', 'EPSG', 29871, 'PROJCS[\"Timbalai 1948 / RSO Born"
  "eo (ch)\",GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai 1948\",SPHEROID[\"Everest 1830 (1967 Definition)\",6377298.556,30"
  "0.8017,AUTHORITY[\"EPSG\",\"7016\"]],TOWGS84[-679,669,-48,0,0,0,0],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4298\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EP"
  "SG\",\"9815\"]],PARAMETER[\"Latitude of projection centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of "
  "projection centre\",115,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",53.3158204722222,AUTHORIT"
  "Y[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",53.1301023611111,AUTHORITY[\"EPSG\",\"8814\"]],"
  "PARAMETER[\"Scale factor on initial line\",0.99984,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection ce"
  "ntre\",29352.4763,AUTHORITY[\"EPSG\",\"8816\"]],PARAMETER[\"Northing at projection centre\",22014.3572,AUTHORITY[\"EP"
  "SG\",\"8817\"]],UNIT[\"British chain (Sears 1922)\",20.116765121552632,AUTHORITY[\"EPSG\",\"9042\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29871\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29872, 1, 'Timbalai 1948 / RSO Borneo (ftSe)', 'EPSG', 29872, 'PROJCS[\"Timbalai 1948 / RSO Bo"
  "rneo (ftSe)\",GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai 1948\",SPHEROID[\"Everest 1830 (1967 Definition)\",6377298.55"
  "6,300.8017,AUTHORITY[\"EPSG\",\"7016\"]],TOWGS84[-679,669,-48,0,0,0,0],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4298\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY["
  "\"EPSG\",\"9815\"]],PARAMETER[\"Latitude of projection centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude"
  " of projection centre\",115,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",53.3158204722222,AUTH"
  "ORITY[\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",53.1301023611111,AUTHORITY[\"EPSG\",\"8814\""
  "]],PARAMETER[\"Scale factor on initial line\",0.99984,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection"
  " centre\",1937263.44,AUTHORITY[\"EPSG\",\"8816\"]],PARAMETER[\"Northing at projection centre\",1452947.58,AUTHORITY[\""
  "EPSG\",\"8817\"]],UNIT[\"British foot (Sears 1922)\",0.3047994715386762,AUTHORITY[\"EPSG\",\"9041\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29872\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29873, 1, 'Timbalai 1948 / RSO Borneo (m)', 'EPSG', 29873, 'PROJCS[\"Timbalai 1948 / RSO Borne"
  "o (m)\",GEOGCS[\"Timbalai 1948\",DATUM[\"Timbalai 1948\",SPHEROID[\"Everest 1830 (1967 Definition)\",6377298.556,300."
  "8017,AUTHORITY[\"EPSG\",\"7016\"]],TOWGS84[-679,669,-48,0,0,0,0],AUTHORITY[\"EPSG\",\"6298\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4298\"]],PROJECTION[\"Hotine Oblique Mercator (variant B)\",AUTHORITY[\"EPSG"
  "\",\"9815\"]],PARAMETER[\"Latitude of projection centre\",4,AUTHORITY[\"EPSG\",\"8811\"]],PARAMETER[\"Longitude of pr"
  "ojection centre\",115,AUTHORITY[\"EPSG\",\"8812\"]],PARAMETER[\"Azimuth of initial line\",53.3158204722222,AUTHORITY["
  "\"EPSG\",\"8813\"]],PARAMETER[\"Angle from Rectified to Skew Grid\",53.1301023611111,AUTHORITY[\"EPSG\",\"8814\"]],PA"
  "RAMETER[\"Scale factor on initial line\",0.99984,AUTHORITY[\"EPSG\",\"8815\"]],PARAMETER[\"Easting at projection cent"
  "re\",590476.87,AUTHORITY[\"EPSG\",\"8816\"]],PARAMETER[\"Northing at projection centre\",442857.65,AUTHORITY[\"EPSG\""
  ",\"8817\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29"
  "873\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29901, 1, 'OSNI 1952 / Irish National Grid', 'EPSG', 29901, 'PROJCS[\"OSNI 1952 / Irish Nation"
  "al Grid\",GEOGCS[\"OSNI 1952\",DATUM[\"OSNI 1952\",SPHEROID[\"Airy 1830\",6377563.396,299.3249646,AUTHORITY[\"EPSG\","
  "\"7001\"]],TOWGS84[482.5,-130.6,564.6,-1.042,-0.214,-0.631,8.15],AUTHORITY[\"EPSG\",\"6188\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4188\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",53.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",-8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"29901\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29902, 1, 'TM65 / Irish Grid', 'EPSG', 29902, 'PROJCS[\"TM65 / Irish Grid\",GEOGCS[\"TM65\",DA"
  "TUM[\"TM65\",SPHEROID[\"Airy Modified 1849\",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],TOWGS84[482.5,-130"
  ".6,564.6,-1.042,-0.214,-0.631,8.15],AUTHORITY[\"EPSG\",\"6299\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4299\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",53.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-8,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000035,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"29902\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (29903, 1, 'TM75 / Irish Grid', 'EPSG', 29903, 'PROJCS[\"TM75 / Irish Grid\",GEOGCS[\"TM75\",DA"
  "TUM[\"Geodetic Datum of 1965\",SPHEROID[\"Airy Modified 1849\",6377340.189,299.3249646,AUTHORITY[\"EPSG\",\"7002\"]],"
  "TOWGS84[482.5,-130.6,564.6,-1.042,-0.214,-0.631,8.15],AUTHORITY[\"EPSG\",\"6300\"]],PRIMEM[\"Greenwich\",0,AUTHORITY["
  "\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon"
  "\",EAST],AUTHORITY[\"EPSG\",\"4300\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"L"
  "atitude of natural origin\",53.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "-8,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1.000035,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",250000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\""
  "29903\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30161, 1, 'Tokyo / Japan Plane Rectangular CS I', 'EPSG', 30161, 'PROJCS[\"Tokyo / Japan Plane"
  " Rectangular CS I\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129.5,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30161\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30162, 1, 'Tokyo / Japan Plane Rectangular CS II', 'EPSG', 30162, 'PROJCS[\"Tokyo / Japan Plan"
  "e Rectangular CS II\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30162\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30163, 1, 'Tokyo / Japan Plane Rectangular CS III', 'EPSG', 30163, 'PROJCS[\"Tokyo / Japan Pla"
  "ne Rectangular CS III\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",132.177777777778,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30163\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30164, 1, 'Tokyo / Japan Plane Rectangular CS IV', 'EPSG', 30164, 'PROJCS[\"Tokyo / Japan Plan"
  "e Rectangular CS IV\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",33,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",133.5,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30164\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30165, 1, 'Tokyo / Japan Plane Rectangular CS V', 'EPSG', 30165, 'PROJCS[\"Tokyo / Japan Plane"
  " Rectangular CS V\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",134.344444444444,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30165\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30166, 1, 'Tokyo / Japan Plane Rectangular CS VI', 'EPSG', 30166, 'PROJCS[\"Tokyo / Japan Plan"
  "e Rectangular CS VI\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",136,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30166\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30167, 1, 'Tokyo / Japan Plane Rectangular CS VII', 'EPSG', 30167, 'PROJCS[\"Tokyo / Japan Pla"
  "ne Rectangular CS VII\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",137.177777777778,AUT"
  "HORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETE"
  "R[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT"
  "[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30167\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30168, 1, 'Tokyo / Japan Plane Rectangular CS VIII', 'EPSG', 30168, 'PROJCS[\"Tokyo / Japan Pl"
  "ane Rectangular CS VIII\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY"
  "[\"EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",138.5,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30168\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30169, 1, 'Tokyo / Japan Plane Rectangular CS IX', 'EPSG', 30169, 'PROJCS[\"Tokyo / Japan Plan"
  "e Rectangular CS IX\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",139.833333333333,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30169\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30170, 1, 'Tokyo / Japan Plane Rectangular CS X', 'EPSG', 30170, 'PROJCS[\"Tokyo / Japan Plane"
  " Rectangular CS X\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPS"
  "G\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\","
  "\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],"
  "AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude o"
  "f natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",140.833333333333,AUTHOR"
  "ITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30170\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30171, 1, 'Tokyo / Japan Plane Rectangular CS XI', 'EPSG', 30171, 'PROJCS[\"Tokyo / Japan Plan"
  "e Rectangular CS XI\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",140.25,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False ea"
  "sting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30172, 1, 'Tokyo / Japan Plane Rectangular CS XII', 'EPSG', 30172, 'PROJCS[\"Tokyo / Japan Pla"
  "ne Rectangular CS XII\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",142.25,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30172\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30173, 1, 'Tokyo / Japan Plane Rectangular CS XIII', 'EPSG', 30173, 'PROJCS[\"Tokyo / Japan Pl"
  "ane Rectangular CS XIII\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY"
  "[\"EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",44,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",144.25,AUTHORITY["
  "\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metr"
  "e\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30173\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30174, 1, 'Tokyo / Japan Plane Rectangular CS XIV', 'EPSG', 30174, 'PROJCS[\"Tokyo / Japan Pla"
  "ne Rectangular CS XIV\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",142,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30174\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30175, 1, 'Tokyo / Japan Plane Rectangular CS XV', 'EPSG', 30175, 'PROJCS[\"Tokyo / Japan Plan"
  "e Rectangular CS XV\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"E"
  "PSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude "
  "of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",127.511111111111,AUTHO"
  "RITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER["
  "\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30175\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30176, 1, 'Tokyo / Japan Plane Rectangular CS XVI', 'EPSG', 30176, 'PROJCS[\"Tokyo / Japan Pla"
  "ne Rectangular CS XVI\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",124,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30176\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30177, 1, 'Tokyo / Japan Plane Rectangular CS XVII', 'EPSG', 30177, 'PROJCS[\"Tokyo / Japan Pl"
  "ane Rectangular CS XVII\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY"
  "[\"EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"E"
  "PSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\","
  "EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lati"
  "tude of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",131,AUTHORITY[\"E"
  "PSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False "
  "easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30177\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30178, 1, 'Tokyo / Japan Plane Rectangular CS XVIII', 'EPSG', 30178, 'PROJCS[\"Tokyo / Japan P"
  "lane Rectangular CS XVIII\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORI"
  "TY[\"EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\""
  "EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\""
  ",EAST],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Lat"
  "itude of natural origin\",20,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",136,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30178\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30179, 1, 'Tokyo / Japan Plane Rectangular CS XIX', 'EPSG', 30179, 'PROJCS[\"Tokyo / Japan Pla"
  "ne Rectangular CS XIX\",GEOGCS[\"Tokyo\",DATUM[\"Tokyo\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\""
  "EPSG\",\"7004\"]],TOWGS84[-147,506,687,0,0,0,0],AUTHORITY[\"EPSG\",\"6301\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4301\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",26,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",154,AUTHORITY[\"EPSG"
  "\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eas"
  "ting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"30179\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30200, 1, 'Trinidad 1903 / Trinidad Grid', 'EPSG', 30200, 'PROJCS[\"Trinidad 1903 / Trinidad G"
  "rid\",GEOGCS[\"Trinidad 1903\",DATUM[\"Trinidad 1903\",SPHEROID[\"Clarke 1858\",6378293.645208759,294.26067636926103,"
  "AUTHORITY[\"EPSG\",\"7007\"]],TOWGS84[-61.702,284.488,472.052,0,0,0,0],AUTHORITY[\"EPSG\",\"6302\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4302\"]],PROJECTION[\"Cassini-Soldner\",AUTHORITY[\"EPSG\",\"9806\"]],"
  "PARAMETER[\"Latitude of natural origin\",10.4416666666667,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natu"
  "ral origin\",-61.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"False easting\",430000,AUTHORITY[\"EPSG\",\""
  "8806\"]],PARAMETER[\"False northing\",325000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"Clarke\\'s link\",0.201166195164,AUT"
  "HORITY[\"EPSG\",\"9039\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"30200\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30339, 1, 'TC(1948) / UTM zone 39N', 'EPSG', 30339, 'PROJCS[\"TC(1948) / UTM zone 39N\",GEOGCS"
  "[\"TC(1948)\",DATUM[\"Trucial Coast 1948\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTH"
  "ORITY[\"EPSG\",\"6303\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4303\"]],PROJECTION[\"Transv"
  "erse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]"
  "],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AX"
  "IS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"30339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30340, 1, 'TC(1948) / UTM zone 40N', 'EPSG', 30340, 'PROJCS[\"TC(1948) / UTM zone 40N\",GEOGCS"
  "[\"TC(1948)\",DATUM[\"Trucial Coast 1948\",SPHEROID[\"Helmert 1906\",6378200,298.3,AUTHORITY[\"EPSG\",\"7020\"]],AUTH"
  "ORITY[\"EPSG\",\"6303\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,A"
  "UTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4303\"]],PROJECTION[\"Transv"
  "erse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]"
  "],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural orig"
  "in\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER["
  "\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AX"
  "IS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"30340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30491, 1, 'Voirol 1875 / Nord Algerie (ancienne)', 'EPSG', 30491, 'PROJCS[\"Voirol 1875 / Nord"
  " Algerie (ancienne)\",GEOGCS[\"Voirol 1875\",DATUM[\"Voirol 1875\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602"
  "12936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-73,-247,227,0,0,0,0],AUTHORITY[\"EPSG\",\"6304\"]],PRIMEM[\"Greenwich"
  "\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4304\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\""
  ",\"9801\"]],PARAMETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natura"
  "l origin\",2.7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999625544,AUTHORITY[\"EPS"
  "G\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"30491\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30492, 1, 'Voirol 1875 / Sud Algerie (ancienne)', 'EPSG', 30492, 'PROJCS[\"Voirol 1875 / Sud A"
  "lgerie (ancienne)\",GEOGCS[\"Voirol 1875\",DATUM[\"Voirol 1875\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212"
  "936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-73,-247,227,0,0,0,0],AUTHORITY[\"EPSG\",\"6304\"]],PRIMEM[\"Greenwich\""
  ",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORT"
  "H],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4304\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\""
  "9801\"]],PARAMETER[\"Latitude of natural origin\",33.3,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural"
  " origin\",2.7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999625769,AUTHORITY[\"EPSG"
  "\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AU"
  "THORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORI"
  "TY[\"EPSG\",\"30492\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30493, 1, 'Voirol 1879 / Nord Algerie (ancienne)', 'EPSG', 30493, 'PROJCS[\"Voirol 1879 / Nord"
  " Algerie (ancienne)\",GEOGCS[\"Voirol 1879\",DATUM[\"Voirol 1879\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.46602"
  "12936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6671\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901"
  "\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHOR"
  "ITY[\"EPSG\",\"4671\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitu"
  "de of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",2.7,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999625544,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"30493\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30494, 1, 'Voirol 1879 / Sud Algerie (ancienne)', 'EPSG', 30494, 'PROJCS[\"Voirol 1879 / Sud A"
  "lgerie (ancienne)\",GEOGCS[\"Voirol 1879\",DATUM[\"Voirol 1879\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212"
  "936269,AUTHORITY[\"EPSG\",\"7011\"]],AUTHORITY[\"EPSG\",\"6671\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4671\"]],PROJECTION[\"Lambert Conic Conformal (1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude"
  " of natural origin\",33.3,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",2.7,AUTHORITY[\"EPS"
  "G\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999625769,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",300000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"30494\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30729, 1, 'Nord Sahara 1959 / UTM zone 29N', 'EPSG', 30729, 'PROJCS[\"Nord Sahara 1959 / UTM z"
  "one 29N\",GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 1959\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,A"
  "UTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404.6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"6"
  "307\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4307\"]],PROJECTION[\"Transverse Mercator\",AU"
  "THORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"30729\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30730, 1, 'Nord Sahara 1959 / UTM zone 30N', 'EPSG', 30730, 'PROJCS[\"Nord Sahara 1959 / UTM z"
  "one 30N\",GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 1959\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,A"
  "UTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404.6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"6"
  "307\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4307\"]],PROJECTION[\"Transverse Mercator\",AU"
  "THORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHOR"
  "ITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\""
  ",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AU"
  "THORITY[\"EPSG\",\"30730\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30731, 1, 'Nord Sahara 1959 / UTM zone 31N', 'EPSG', 30731, 'PROJCS[\"Nord Sahara 1959 / UTM z"
  "one 31N\",GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 1959\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,A"
  "UTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404.6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"6"
  "307\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4307\"]],PROJECTION[\"Transverse Mercator\",AU"
  "THORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"30731\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30732, 1, 'Nord Sahara 1959 / UTM zone 32N', 'EPSG', 30732, 'PROJCS[\"Nord Sahara 1959 / UTM z"
  "one 32N\",GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 1959\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,A"
  "UTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404.6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"6"
  "307\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4307\"]],PROJECTION[\"Transverse Mercator\",AU"
  "THORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Long"
  "itude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORI"
  "TY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\","
  "0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUT"
  "HORITY[\"EPSG\",\"30732\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30791, 1, 'Nord Sahara 1959 / Nord Algerie', 'EPSG', 30791, 'PROJCS[\"Nord Sahara 1959 / Nord "
  "Algerie\",GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 1959\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,A"
  "UTHORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404.6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"6"
  "307\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\","
  "\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4307\"]],PROJECTION[\"Lambert Conic Conformal "
  "(1SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8801\"]],PARAM"
  "ETER[\"Longitude of natural origin\",2.7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0"
  ".999625544,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500135,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",300090,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"30791\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (30792, 1, 'Nord Sahara 1959 / Sud Algerie', 'EPSG', 30792, 'PROJCS[\"Nord Sahara 1959 / Sud Al"
  "gerie\",GEOGCS[\"Nord Sahara 1959\",DATUM[\"Nord Sahara 1959\",SPHEROID[\"Clarke 1880 (RGS)\",6378249.145,293.465,AUT"
  "HORITY[\"EPSG\",\"7012\"]],TOWGS84[-209.3622,-87.8162,404.6198,0.0046,3.4784,0.5805,-1.4547],AUTHORITY[\"EPSG\",\"630"
  "7\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4307\"]],PROJECTION[\"Lambert Conic Conformal (1"
  "SP)\",AUTHORITY[\"EPSG\",\"9801\"]],PARAMETER[\"Latitude of natural origin\",33.3,AUTHORITY[\"EPSG\",\"8801\"]],PARAM"
  "ETER[\"Longitude of natural origin\",2.7,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0"
  ".999625769,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500135,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",300090,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST]"
  ",AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"30792\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31028, 1, 'Yoff / UTM zone 28N', 'EPSG', 31028, 'PROJCS[\"Yoff / UTM zone 28N\",GEOGCS[\"Yoff\""
  ",DATUM[\"Yoff\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY[\"EPSG\",\"7011\"]],TOWGS84[-30,"
  "190,89,0,0,0,0],AUTHORITY[\"EPSG\",\"6310\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4310\"]]"
  ",PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORIT"
  "Y[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale f"
  "actor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\","
  "\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]"
  "],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31121, 1, 'Zanderij / UTM zone 21N', 'EPSG', 31121, 'PROJCS[\"Zanderij / UTM zone 21N\",GEOGCS"
  "[\"Zanderij\",DATUM[\"Zanderij\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-2"
  "65,120,-358,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"431"
  "1\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sc"
  "ale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EP"
  "SG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"90"
  "01\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31154, 1, 'Zanderij / TM 54 NW', 'EPSG', 31154, 'PROJCS[\"Zanderij / TM 54 NW\",GEOGCS[\"Zande"
  "rij\",DATUM[\"Zanderij\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,120,-"
  "358,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017"
  "453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4311\"]],PR"
  "OJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\""
  "EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-54,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale facto"
  "r at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"88"
  "06\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AX"
  "IS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31154\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31170, 1, 'Zanderij / Suriname Old TM', 'EPSG', 31170, 'PROJCS[\"Zanderij / Suriname Old TM\","
  "GEOGCS[\"Zanderij\",DATUM[\"Zanderij\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWG"
  "S84[-265,120,-358,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4311\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-55.6944444444444,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31170\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31171, 1, 'Zanderij / Suriname TM', 'EPSG', 31171, 'PROJCS[\"Zanderij / Suriname TM\",GEOGCS[\""
  "Zanderij\",DATUM[\"Zanderij\",SPHEROID[\"International 1924\",6378388,297,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[-265,"
  "120,-358,0,0,0,0],AUTHORITY[\"EPSG\",\"6311\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4311\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-55.6944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31171\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31251, 1, 'MGI (Ferro) / Austria GK West Zone', 'EPSG', 31251, 'PROJCS[\"MGI (Ferro) / Austria"
  " GK West Zone\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377"
  "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,"
  "AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31251\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31252, 1, 'MGI (Ferro) / Austria GK Central Zone', 'EPSG', 31252, 'PROJCS[\"MGI (Ferro) / Aust"
  "ria GK Central Zone\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\""
  ",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.67777777777"
  "7774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NO"
  "RTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]"
  "],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\","
  "31,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMET"
  "ER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31252\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31253, 1, 'MGI (Ferro) / Austria GK East Zone', 'EPSG', 31253, 'PROJCS[\"MGI (Ferro) / Austria"
  " GK East Zone\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377"
  "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,"
  "AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",34,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG\",\"8807\"]],U"
  "NIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31253\"]]', NUL"
  "L);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31254, 1, 'MGI / Austria GK West', 'EPSG', 31254, 'PROJCS[\"MGI / Austria GK West\",GEOGCS[\"M"
  "GI\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10.3"
  "444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\""
  "31254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31255, 1, 'MGI / Austria GK Central', 'EPSG', 31255, 'PROJCS[\"MGI / Austria GK Central\",GEOG"
  "CS[\"MGI\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG"
  "\",\"7004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",13.3444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8"
  "805\"]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"31255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31256, 1, 'MGI / Austria GK East', 'EPSG', 31256, 'PROJCS[\"MGI / Austria GK East\",GEOGCS[\"M"
  "GI\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PA"
  "RAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16.3"
  "444444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\""
  "]],PARAMETER[\"False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"EPSG"
  "\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\""
  "31256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31257, 1, 'MGI / Austria GK M28', 'EPSG', 31257, 'PROJCS[\"MGI / Austria GK M28\",GEOGCS[\"MGI"
  "\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700"
  "4\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10.344"
  "4444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"31257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31258, 1, 'MGI / Austria GK M31', 'EPSG', 31258, 'PROJCS[\"MGI / Austria GK M31\",GEOGCS[\"MGI"
  "\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700"
  "4\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13.344"
  "4444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",450000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"31258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31259, 1, 'MGI / Austria GK M34', 'EPSG', 31259, 'PROJCS[\"MGI / Austria GK M34\",GEOGCS[\"MGI"
  "\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"700"
  "4\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,A"
  "UTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],A"
  "XIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARA"
  "METER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16.344"
  "4444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",750000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",-5000000,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"31259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31281, 1, 'MGI (Ferro) / Austria West Zone', 'EPSG', 31281, 'PROJCS[\"MGI (Ferro) / Austria We"
  "st Zone\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHOR"
  "ITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31281\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31282, 1, 'MGI (Ferro) / Austria Central Zone', 'EPSG', 31282, 'PROJCS[\"MGI (Ferro) / Austria"
  " Central Zone\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377"
  "397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,"
  "AUTHORITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],"
  "AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PAR"
  "AMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AU"
  "THORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31282\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31283, 1, 'MGI (Ferro) / Austria East Zone', 'EPSG', 31283, 'PROJCS[\"MGI (Ferro) / Austria Ea"
  "st Zone\",GEOGCS[\"MGI (Ferro)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.15"
  "5,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHOR"
  "ITY[\"EPSG\",\"8909\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",34,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False e"
  "asting\",0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31283\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31284, 1, 'MGI / Austria M28', 'EPSG', 31284, 'PROJCS[\"MGI / Austria M28\",GEOGCS[\"MGI\",DAT"
  "UM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],"
  "TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",10.344444444"
  "4444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31284\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31285, 1, 'MGI / Austria M31', 'EPSG', 31285, 'PROJCS[\"MGI / Austria M31\",GEOGCS[\"MGI\",DAT"
  "UM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],"
  "TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",13.344444444"
  "4444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",450000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31285\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31286, 1, 'MGI / Austria M34', 'EPSG', 31286, 'PROJCS[\"MGI / Austria M34\",GEOGCS[\"MGI\",DAT"
  "UM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7004\"]],"
  "TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0,AUTHORI"
  "TY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\""
  "Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER["
  "\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",16.344444444"
  "4444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAM"
  "ETER[\"False easting\",750000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31286\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31287, 1, 'MGI / Austria Lambert', 'EPSG', 31287, 'PROJCS[\"MGI / Austria Lambert\",GEOGCS[\"M"
  "GI\",DATUM[\"Militar-Geographische Institut\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EPSG\",\"7"
  "004\"]],TOWGS84[577.326,90.129,463.919,5.137,1.474,5.297,2.4232],AUTHORITY[\"EPSG\",\"6312\"]],PRIMEM[\"Greenwich\",0"
  ",AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH]"
  ",AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4312\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9"
  "802\"]],PARAMETER[\"Latitude of false origin\",47.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude o"
  "f false origin\",13.3444444444444,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AU"
  "THORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETE"
  "R[\"Easting at false origin\",400000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",400000,AUTH"
  "ORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY"
  "[\"EPSG\",\"31287\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31288, 1, 'MGI (Ferro) / M28', 'EPSG', 31288, 'PROJCS[\"MGI (Ferro) / M28\",GEOGCS[\"MGI (Ferr"
  "o)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP"
  "SG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",28,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",150000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31288\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31289, 1, 'MGI (Ferro) / M31', 'EPSG', 31289, 'PROJCS[\"MGI (Ferro) / M31\",GEOGCS[\"MGI (Ferr"
  "o)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP"
  "SG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",450000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31289\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31290, 1, 'MGI (Ferro) / M34', 'EPSG', 31290, 'PROJCS[\"MGI (Ferro) / M34\",GEOGCS[\"MGI (Ferr"
  "o)\",DATUM[\"Militar-Geographische Institut (Ferro)\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528128,AUTHORITY[\"EP"
  "SG\",\"7004\"]],AUTHORITY[\"EPSG\",\"6805\"]],PRIMEM[\"Ferro\",-17.677777777777774,AUTHORITY[\"EPSG\",\"8909\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4805\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",34,AUTHORITY[\"EPSG\",\"8802\"]],PARAME"
  "TER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",750000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\",\"31290\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31300, 1, 'Belge 1972 / Belge Lambert 72', 'EPSG', 31300, 'PROJCS[\"Belge 1972 / Belge Lambert"
  " 72\",GEOGCS[\"Belge 1972\",DATUM[\"Reseau National Belge 1972\",SPHEROID[\"International 1924\",6378388,297,AUTHORIT"
  "Y[\"EPSG\",\"7022\"]],TOWGS84[-106.8686,52.2978,-103.7239,0.3366,-0.457,1.8422,-1.2747],AUTHORITY[\"EPSG\",\"6313\"]]"
  ",PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\""
  "]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4313\"]],PROJECTION[\"Lambert Conic Conformal (2SP Bel"
  "gium)\",AUTHORITY[\"EPSG\",\"9803\"]],PARAMETER[\"Latitude of false origin\",90,AUTHORITY[\"EPSG\",\"8821\"]],PARAMET"
  "ER[\"Longitude of false origin\",4.35693972222222,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard "
  "parallel\",49.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",51.16666666"
  "66667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",150000.01256,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",5400088.4378,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"31300\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31370, 1, 'Belge 1972 / Belgian Lambert 72', 'EPSG', 31370, 'PROJCS[\"Belge 1972 / Belgian Lam"
  "bert 72\",GEOGCS[\"Belge 1972\",DATUM[\"Reseau National Belge 1972\",SPHEROID[\"International 1924\",6378388,297,AUTH"
  "ORITY[\"EPSG\",\"7022\"]],TOWGS84[-106.8686,52.2978,-103.7239,0.3366,-0.457,1.8422,-1.2747],AUTHORITY[\"EPSG\",\"6313"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4313\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",90,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",4.36748666666667,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard par"
  "allel\",51.1666672333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",49.8333339,AUT"
  "HORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",150000.013,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\""
  "Northing at false origin\",5400088.438,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],A"
  "XIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"31370\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31466, 1, 'DHDN / 3-degree Gauss-Kruger zone 2', 'EPSG', 31466, 'PROJCS[\"DHDN / 3-degree Gaus"
  "s-Kruger zone 2\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528"
  "128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"6314\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",6,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",2500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"31466\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31467, 1, 'DHDN / 3-degree Gauss-Kruger zone 3', 'EPSG', 31467, 'PROJCS[\"DHDN / 3-degree Gaus"
  "s-Kruger zone 3\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528"
  "128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"6314\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"880"
  "5\"]],PARAMETER[\"False easting\",3500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\","
  "\"31467\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31468, 1, 'DHDN / 3-degree Gauss-Kruger zone 4', 'EPSG', 31468, 'PROJCS[\"DHDN / 3-degree Gaus"
  "s-Kruger zone 4\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528"
  "128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"6314\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",12,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",4500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"31468\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31469, 1, 'DHDN / 3-degree Gauss-Kruger zone 5', 'EPSG', 31469, 'PROJCS[\"DHDN / 3-degree Gaus"
  "s-Kruger zone 5\",GEOGCS[\"DHDN\",DATUM[\"Deutsches Hauptdreiecksnetz\",SPHEROID[\"Bessel 1841\",6377397.155,299.1528"
  "128,AUTHORITY[\"EPSG\",\"7004\"]],TOWGS84[598.1,73.7,418.2,0.202,0.045,-2.455,6.7],AUTHORITY[\"EPSG\",\"6314\"]],PRIM"
  "EM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],A"
  "XIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4314\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EP"
  "SG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of nat"
  "ural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"88"
  "05\"]],PARAMETER[\"False easting\",5500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"E"
  "PSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",NORTH],AXIS[\"Y\",EAST],AUTHORITY[\"EPSG\""
  ",\"31469\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31528, 1, 'Conakry 1905 / UTM zone 28N', 'EPSG', 31528, 'PROJCS[\"Conakry 1905 / UTM zone 28N\""
  ",GEOGCS[\"Conakry 1905\",DATUM[\"Conakry 1905\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY["
  "\"EPSG\",\"7011\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORITY[\"EPSG\",\"6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4315\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31529, 1, 'Conakry 1905 / UTM zone 29N', 'EPSG', 31529, 'PROJCS[\"Conakry 1905 / UTM zone 29N\""
  ",GEOGCS[\"Conakry 1905\",DATUM[\"Conakry 1905\",SPHEROID[\"Clarke 1880 (IGN)\",6378249.2,293.4660212936269,AUTHORITY["
  "\"EPSG\",\"7011\"]],TOWGS84[-23,259,-9,0,0,0,0],AUTHORITY[\"EPSG\",\"6315\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG"
  "\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAS"
  "T],AUTHORITY[\"EPSG\",\"4315\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitud"
  "e of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31600, 1, 'Dealul Piscului 1930 / Stereo 33', 'EPSG', 31600, 'PROJCS[\"Dealul Piscului 1930 / "
  "Stereo 33\",GEOGCS[\"Dealul Piscului 1930\",DATUM[\"Dealul Piscului 1930\",SPHEROID[\"International 1924\",6378388,29"
  "7,AUTHORITY[\"EPSG\",\"7022\"]],TOWGS84[103.25,-100.4,-307.19,0,0,0,0],AUTHORITY[\"EPSG\",\"6316\"]],PRIMEM[\"Greenwi"
  "ch\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\","
  "NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4316\"]],PROJECTION[\"Oblique Stereographic\",AUTHORITY[\"EPSG\",\"980"
  "9\"]],PARAMETER[\"Latitude of natural origin\",45.9111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude o"
  "f natural origin\",25.3924658888889,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996"
  "667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",500000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS["
  "\"Y\",NORTH],AUTHORITY[\"EPSG\",\"31600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31838, 1, 'NGN / UTM zone 38N', 'EPSG', 31838, 'PROJCS[\"NGN / UTM zone 38N\",GEOGCS[\"NGN\",D"
  "ATUM[\"National Geodetic Network\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-3"
  ".2,-5.7,2.8,0,0,0,0],AUTHORITY[\"EPSG\",\"6318\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"431"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31838\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31839, 1, 'NGN / UTM zone 39N', 'EPSG', 31839, 'PROJCS[\"NGN / UTM zone 39N\",GEOGCS[\"NGN\",D"
  "ATUM[\"National Geodetic Network\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],TOWGS84[-3"
  ".2,-5.7,2.8,0,0,0,0],AUTHORITY[\"EPSG\",\"6318\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree"
  "\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"431"
  "8\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUT"
  "HORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Sca"
  "le factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPS"
  "G\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"900"
  "1\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31839\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31901, 1, 'KUDAMS / KTM', 'EPSG', 31901, 'PROJCS[\"KUDAMS / KTM\",GEOGCS[\"KUDAMS\",DATUM[\"Ku"
  "wait Utility\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[226.702,-193.337,-35"
  ".371,-2.229,-4.391,9.238,0.9798],AUTHORITY[\"EPSG\",\"6319\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4319\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",48,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",1,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31901\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31965, 1, 'SIRGAS 2000 / UTM zone 11N', 'EPSG', 31965, 'PROJCS[\"SIRGAS 2000 / UTM zone 11N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31965"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31966, 1, 'SIRGAS 2000 / UTM zone 12N', 'EPSG', 31966, 'PROJCS[\"SIRGAS 2000 / UTM zone 12N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31966"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31967, 1, 'SIRGAS 2000 / UTM zone 13N', 'EPSG', 31967, 'PROJCS[\"SIRGAS 2000 / UTM zone 13N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]]"
  ",PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31967"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31968, 1, 'SIRGAS 2000 / UTM zone 14N', 'EPSG', 31968, 'PROJCS[\"SIRGAS 2000 / UTM zone 14N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31968"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31969, 1, 'SIRGAS 2000 / UTM zone 15N', 'EPSG', 31969, 'PROJCS[\"SIRGAS 2000 / UTM zone 15N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31969"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31970, 1, 'SIRGAS 2000 / UTM zone 16N', 'EPSG', 31970, 'PROJCS[\"SIRGAS 2000 / UTM zone 16N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31970"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31971, 1, 'SIRGAS 2000 / UTM zone 17N', 'EPSG', 31971, 'PROJCS[\"SIRGAS 2000 / UTM zone 17N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31971"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31972, 1, 'SIRGAS 2000 / UTM zone 18N', 'EPSG', 31972, 'PROJCS[\"SIRGAS 2000 / UTM zone 18N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31972"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31973, 1, 'SIRGAS 2000 / UTM zone 19N', 'EPSG', 31973, 'PROJCS[\"SIRGAS 2000 / UTM zone 19N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31973"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31974, 1, 'SIRGAS 2000 / UTM zone 20N', 'EPSG', 31974, 'PROJCS[\"SIRGAS 2000 / UTM zone 20N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31974"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31975, 1, 'SIRGAS 2000 / UTM zone 21N', 'EPSG', 31975, 'PROJCS[\"SIRGAS 2000 / UTM zone 21N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31975"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31976, 1, 'SIRGAS 2000 / UTM zone 22N', 'EPSG', 31976, 'PROJCS[\"SIRGAS 2000 / UTM zone 22N\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\""
  "8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"31976"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31977, 1, 'SIRGAS 2000 / UTM zone 17S', 'EPSG', 31977, 'PROJCS[\"SIRGAS 2000 / UTM zone 17S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31977\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31978, 1, 'SIRGAS 2000 / UTM zone 18S', 'EPSG', 31978, 'PROJCS[\"SIRGAS 2000 / UTM zone 18S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31978\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31979, 1, 'SIRGAS 2000 / UTM zone 19S', 'EPSG', 31979, 'PROJCS[\"SIRGAS 2000 / UTM zone 19S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31979\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31980, 1, 'SIRGAS 2000 / UTM zone 20S', 'EPSG', 31980, 'PROJCS[\"SIRGAS 2000 / UTM zone 20S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31980\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31981, 1, 'SIRGAS 2000 / UTM zone 21S', 'EPSG', 31981, 'PROJCS[\"SIRGAS 2000 / UTM zone 21S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31981\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31982, 1, 'SIRGAS 2000 / UTM zone 22S', 'EPSG', 31982, 'PROJCS[\"SIRGAS 2000 / UTM zone 22S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31982\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31983, 1, 'SIRGAS 2000 / UTM zone 23S', 'EPSG', 31983, 'PROJCS[\"SIRGAS 2000 / UTM zone 23S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31983\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31984, 1, 'SIRGAS 2000 / UTM zone 24S', 'EPSG', 31984, 'PROJCS[\"SIRGAS 2000 / UTM zone 24S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31984\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31985, 1, 'SIRGAS 2000 / UTM zone 25S', 'EPSG', 31985, 'PROJCS[\"SIRGAS 2000 / UTM zone 25S\","
  "GEOGCS[\"SIRGAS 2000\",DATUM[\"Sistema de Referencia Geocentrico para las AmericaS 2000\",SPHEROID[\"GRS 1980\",63781"
  "37,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6674\"]],PRIMEM[\"Greenwic"
  "h\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",N"
  "ORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4674\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\""
  "]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\""
  ",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],"
  "PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EP"
  "SG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\","
  "\"31985\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31986, 1, 'SIRGAS 1995 / UTM zone 17N', 'EPSG', 31986, 'PROJCS[\"SIRGAS 1995 / UTM zone 17N\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"319"
  "86\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31987, 1, 'SIRGAS 1995 / UTM zone 18N', 'EPSG', 31987, 'PROJCS[\"SIRGAS 1995 / UTM zone 18N\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"319"
  "87\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31988, 1, 'SIRGAS 1995 / UTM zone 19N', 'EPSG', 31988, 'PROJCS[\"SIRGAS 1995 / UTM zone 19N\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"319"
  "88\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31989, 1, 'SIRGAS 1995 / UTM zone 20N', 'EPSG', 31989, 'PROJCS[\"SIRGAS 1995 / UTM zone 20N\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"319"
  "89\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31990, 1, 'SIRGAS 1995 / UTM zone 21N', 'EPSG', 31990, 'PROJCS[\"SIRGAS 1995 / UTM zone 21N\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"319"
  "90\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31991, 1, 'SIRGAS 1995 / UTM zone 22N', 'EPSG', 31991, 'PROJCS[\"SIRGAS 1995 / UTM zone 22N\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\","
  "\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"319"
  "91\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31992, 1, 'SIRGAS 1995 / UTM zone 17S', 'EPSG', 31992, 'PROJCS[\"SIRGAS 1995 / UTM zone 17S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31992\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31993, 1, 'SIRGAS 1995 / UTM zone 18S', 'EPSG', 31993, 'PROJCS[\"SIRGAS 1995 / UTM zone 18S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31993\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31994, 1, 'SIRGAS 1995 / UTM zone 19S', 'EPSG', 31994, 'PROJCS[\"SIRGAS 1995 / UTM zone 19S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31994\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31995, 1, 'SIRGAS 1995 / UTM zone 20S', 'EPSG', 31995, 'PROJCS[\"SIRGAS 1995 / UTM zone 20S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31995\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31996, 1, 'SIRGAS 1995 / UTM zone 21S', 'EPSG', 31996, 'PROJCS[\"SIRGAS 1995 / UTM zone 21S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31996\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31997, 1, 'SIRGAS 1995 / UTM zone 22S', 'EPSG', 31997, 'PROJCS[\"SIRGAS 1995 / UTM zone 22S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31997\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31998, 1, 'SIRGAS 1995 / UTM zone 23S', 'EPSG', 31998, 'PROJCS[\"SIRGAS 1995 / UTM zone 23S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31998\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (31999, 1, 'SIRGAS 1995 / UTM zone 24S', 'EPSG', 31999, 'PROJCS[\"SIRGAS 1995 / UTM zone 24S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"31999\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32000, 1, 'SIRGAS 1995 / UTM zone 25S', 'EPSG', 32000, 'PROJCS[\"SIRGAS 1995 / UTM zone 25S\","
  "GEOGCS[\"SIRGAS 1995\",DATUM[\"Sistema de Referencia Geocentrico para America del Sur 1995\",SPHEROID[\"GRS 1980\",63"
  "78137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[0,0,0,0,0,0,0],AUTHORITY[\"EPSG\",\"6170\"]],PRIMEM[\"Green"
  "wich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\""
  ",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4170\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807"
  "\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin"
  "\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]"
  "],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\""
  "EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\""
  ",\"32000\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32001, 1, 'NAD27 / Montana North', 'EPSG', 32001, 'PROJCS[\"NAD27 / Montana North\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-109.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",48.7277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",47.861111"
  "1111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32001\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32002, 1, 'NAD27 / Montana Central', 'EPSG', 32002, 'PROJCS[\"NAD27 / Montana Central\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.8333333333333,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-109.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitu"
  "de of 1st standard parallel\",47.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard par"
  "allel\",46.45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924"
  ",AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32002\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32003, 1, 'NAD27 / Montana South', 'EPSG', 32003, 'PROJCS[\"NAD27 / Montana South\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-109.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",46.4,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.8666666666667,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32003\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32005, 1, 'NAD27 / Nebraska North', 'EPSG', 32005, 'PROJCS[\"NAD27 / Nebraska North\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standa"
  "rd parallel\",41.8611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",42.81666"
  "66666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32005\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32006, 1, 'NAD27 / Nebraska South', 'EPSG', 32006, 'PROJCS[\"NAD27 / Nebraska South\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-99.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",40.2833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",41.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32006\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32007, 1, 'NAD27 / Nevada East', 'EPSG', 32007, 'PROJCS[\"NAD27 / Nevada East\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-115.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32007\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32008, 1, 'NAD27 / Nevada Central', 'EPSG', 32008, 'PROJCS[\"NAD27 / Nevada Central\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-116.666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER"
  "[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32008\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32009, 1, 'NAD27 / Nevada West', 'EPSG', 32009, 'PROJCS[\"NAD27 / Nevada West\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHO"
  "RITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-118.594444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32009\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32010, 1, 'NAD27 / New Hampshire', 'EPSG', 32010, 'PROJCS[\"NAD27 / New Hampshire\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],"
  "PARAMETER[\"Longitude of natural origin\",-71.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at"
  " natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8"
  "806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32010\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32011, 1, 'NAD27 / New Jersey', 'EPSG', 32011, 'PROJCS[\"NAD27 / New Jersey\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAME"
  "TER[\"Longitude of natural origin\",-74.6666666666667,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.999975,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2000000,AUTHORITY[\"EPSG\",\"8806\"]],"
  "PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32011\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32012, 1, 'NAD27 / New Mexico East', 'EPSG', 32012, 'PROJCS[\"NAD27 / New Mexico East\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-104.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32012\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32013, 1, 'NAD27 / New Mexico Central', 'EPSG', 32013, 'PROJCS[\"NAD27 / New Mexico Central\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mer"
  "cator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAM"
  "ETER[\"Longitude of natural origin\",-106.25,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin"
  "\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\""
  "False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"900"
  "3\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32013\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32014, 1, 'NAD27 / New Mexico West', 'EPSG', 32014, 'PROJCS[\"NAD27 / New Mexico West\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\""
  ",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-107.833333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999916667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32014\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32015, 1, 'NAD27 / New York East', 'EPSG', 32015, 'PROJCS[\"NAD27 / New York East\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-74.3333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32015\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32016, 1, 'NAD27 / New York Central', 'EPSG', 32016, 'PROJCS[\"NAD27 / New York Central\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercat"
  "or\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETE"
  "R[\"Longitude of natural origin\",-76.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PA"
  "RAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EP"
  "SG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32016\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32017, 1, 'NAD27 / New York West', 'EPSG', 32017, 'PROJCS[\"NAD27 / New York West\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lo"
  "ngitude of natural origin\",-78.5944444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origi"
  "n\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETE"
  "R[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32017\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32019, 1, 'NAD27 / North Carolina', 'EPSG', 32019, 'PROJCS[\"NAD27 / North Carolina\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PAR"
  "AMETER[\"Longitude of false origin\",-79,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",34.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.1666666666667,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32019\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32020, 1, 'NAD27 / North Dakota North', 'EPSG', 32020, 'PROJCS[\"NAD27 / North Dakota North\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",47.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "48.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121"
  "924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32020\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32021, 1, 'NAD27 / North Dakota South', 'EPSG', 32021, 'PROJCS[\"NAD27 / North Dakota South\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.6777777777778,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",46.1944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",47.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0."
  "30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32021\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32022, 1, 'NAD27 / Ohio North', 'EPSG', 32022, 'PROJCS[\"NAD27 / Ohio North\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st"
  " standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",4"
  "1.7,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32022\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32023, 1, 'NAD27 / Ohio South', 'EPSG', 32023, 'PROJCS[\"NAD27 / Ohio South\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\""
  "Longitude of false origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard para"
  "llel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",40.033333333333"
  "3,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER"
  "[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32023\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32024, 1, 'NAD27 / Oklahoma North', 'EPSG', 32024, 'PROJCS[\"NAD27 / Oklahoma North\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",3"
  "5.5666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHOR"
  "ITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"North"
  "ing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\","
  "\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32024\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32025, 1, 'NAD27 / Oklahoma South', 'EPSG', 32025, 'PROJCS[\"NAD27 / Oklahoma South\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",33.9333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",35.233333"
  "3333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32025\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32026, 1, 'NAD27 / Oregon North', 'EPSG', 32026, 'PROJCS[\"NAD27 / Oregon North\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32026\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32027, 1, 'NAD27 / Oregon South', 'EPSG', 32027, 'PROJCS[\"NAD27 / Oregon South\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",44,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMET"
  "ER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32027\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32028, 1, 'NAD27 / Pennsylvania North', 'EPSG', 32028, 'PROJCS[\"NAD27 / Pennsylvania North\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.1666666666667,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77.75,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of"
  " 1st standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel"
  "\",41.95,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PA"
  "RAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTH"
  "ORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32028\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32030, 1, 'NAD27 / Rhode Island', 'EPSG', 32030, 'PROJCS[\"NAD27 / Rhode Island\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.0944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-71.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.9999938,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\""
  "]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY"
  "[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32030\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32031, 1, 'NAD27 / South Carolina North', 'EPSG', 32031, 'PROJCS[\"NAD27 / South Carolina Nort"
  "h\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORI"
  "TY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORI"
  "TY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert C"
  "onic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",33,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",33.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",34.977777"
  "7777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PAR"
  "AMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHO"
  "RITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32031\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32033, 1, 'NAD27 / South Carolina South', 'EPSG', 32033, 'PROJCS[\"NAD27 / South Carolina Sout"
  "h\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORI"
  "TY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORI"
  "TY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert C"
  "onic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.8333333333333,AUTHORIT"
  "Y[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",32.3333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",33.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32033\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32034, 1, 'NAD27 / South Dakota North', 'EPSG', 32034, 'PROJCS[\"NAD27 / South Dakota North\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1"
  "st standard parallel\",44.4166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",45.6944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012"
  "1924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32034\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32035, 1, 'NAD27 / South Dakota South', 'EPSG', 32035, 'PROJCS[\"NAD27 / South Dakota South\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42.3333333333333,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",42.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",44.4,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8"
  "826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601"
  "21924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32035\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32037, 1, 'NAD27 / Texas North', 'EPSG', 32037, 'PROJCS[\"NAD27 / Texas North\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER["
  "\"Longitude of false origin\",-101.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard pa"
  "rallel\",34.65,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",36.1944444444444,AUTHORI"
  "TY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northi"
  "ng at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\""
  "9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32037\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32038, 1, 'NAD27 / Texas North Central', 'EPSG', 32038, 'PROJCS[\"NAD27 / Texas North Central\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",31.6777777777778,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-97.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",32.1444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",33.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0."
  "30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32038\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32039, 1, 'NAD27 / Texas Central', 'EPSG', 32039, 'PROJCS[\"NAD27 / Texas Central\",GEOGCS[\"N"
  "AD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-100.333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",30.1166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",31.8833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32039\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32040, 1, 'NAD27 / Texas South Central', 'EPSG', 32040, 'PROJCS[\"NAD27 / Texas South Central\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",27.8333333333333,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-99,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1"
  "st standard parallel\",28.3833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",30.2833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.3048006096012"
  "1924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32040\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32041, 1, 'NAD27 / Texas South', 'EPSG', 32041, 'PROJCS[\"NAD27 / Texas South\",GEOGCS[\"NAD27"
  "\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008"
  "\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267"
  "\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9"
  "122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2S"
  "P)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",25.6777777777778,AUTHORITY[\"EPSG\",\"8821\""
  "]],PARAMETER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1s"
  "t standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "27.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121"
  "924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32041\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32042, 1, 'NAD27 / Utah North', 'EPSG', 32042, 'PROJCS[\"NAD27 / Utah North\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st"
  " standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",4"
  "1.7944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601219"
  "24,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32042\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32043, 1, 'NAD27 / Utah Central', 'EPSG', 32043, 'PROJCS[\"NAD27 / Utah Central\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2"
  "SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of "
  "1st standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\""
  ",40.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARA"
  "METER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHOR"
  "ITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32043\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32044, 1, 'NAD27 / Utah South', 'EPSG', 32044, 'PROJCS[\"NAD27 / Utah South\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal (2SP)"
  "\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]"
  "],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st"
  " standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",3"
  "8.35,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32044\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32045, 1, 'NAD27 / Vermont', 'EPSG', 32045, 'PROJCS[\"NAD27 / Vermont\",GEOGCS[\"NAD27\",DATUM"
  "[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\"]],TOW"
  "GS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\"]],PRI"
  "MEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],"
  "AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"E"
  "PSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\""
  "Longitude of natural origin\",-72.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural ori"
  "gin\",0.999964286,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG"
  "\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32045\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32046, 1, 'NAD27 / Virginia North', 'EPSG', 32046, 'PROJCS[\"NAD27 / Virginia North\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",38.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",39.2111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32046\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32047, 1, 'NAD27 / Virginia South', 'EPSG', 32047, 'PROJCS[\"NAD27 / Virginia South\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude "
  "of 1st standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parall"
  "el\",37.9777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\","
  "\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609"
  "60121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32047\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32048, 1, 'NAD27 / Washington North', 'EPSG', 32048, 'PROJCS[\"NAD27 / Washington North\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Con"
  "formal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],"
  "PARAMETER[\"Longitude of false origin\",-120.833333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st s"
  "tandard parallel\",47.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",48."
  "7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924"
  ",AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32048\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32049, 1, 'NAD27 / Washington South', 'EPSG', 32049, 'PROJCS[\"NAD27 / Washington South\",GEOG"
  "CS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPS"
  "G\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPS"
  "G\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\""
  "EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Con"
  "formal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.3333333333333,AUTHORITY[\"EPSG"
  "\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Lat"
  "itude of 1st standard parallel\",45.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard "
  "parallel\",47.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"E"
  "PSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304"
  "80060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32049\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32050, 1, 'NAD27 / West Virginia North', 'EPSG', 32050, 'PROJCS[\"NAD27 / West Virginia North\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",38.5111111111111,AUTHORITY[\""
  "EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",39,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\","
  "40.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"882"
  "6\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121"
  "924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32050\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32051, 1, 'NAD27 / West Virginia South', 'EPSG', 32051, 'PROJCS[\"NAD27 / West Virginia South\""
  ",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY["
  "\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY["
  "\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORI"
  "TY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Coni"
  "c Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",37,AUTHORITY[\"EPSG\",\"8821"
  "\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard p"
  "arallel\",37.4944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",38.883333333"
  "3333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAME"
  "TER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORIT"
  "Y[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32051\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32052, 1, 'NAD27 / Wisconsin North', 'EPSG', 32052, 'PROJCS[\"NAD27 / Wisconsin North\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",45.1666666666667,AUTHORITY[\"EPSG\","
  "\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st stan"
  "dard parallel\",45.5666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.766"
  "6666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],"
  "PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AU"
  "THORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32052\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32053, 1, 'NAD27 / Wisconsin Central', 'EPSG', 32053, 'PROJCS[\"NAD27 / Wisconsin Central\",GE"
  "OGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"E"
  "PSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"E"
  "PSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY["
  "\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic C"
  "onformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",43.8333333333333,AUTHORITY[\"EP"
  "SG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st"
  " standard parallel\",44.2611111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",4"
  "5.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826"
  "\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.304800609601219"
  "24,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32053\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32054, 1, 'NAD27 / Wisconsin South', 'EPSG', 32054, 'PROJCS[\"NAD27 / Wisconsin South\",GEOGCS"
  "[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\""
  ",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\""
  ",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPS"
  "G\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Confor"
  "mal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PAR"
  "AMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\""
  ",42.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",44.0777777777778,AUTH"
  "ORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Nor"
  "thing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\""
  ",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32054\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32055, 1, 'NAD27 / Wyoming East', 'EPSG', 32055, 'PROJCS[\"NAD27 / Wyoming East\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-105.177777777778,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32055\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32056, 1, 'NAD27 / Wyoming East Central', 'EPSG', 32056, 'PROJCS[\"NAD27 / Wyoming East Centra"
  "l\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORI"
  "TY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORI"
  "TY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.6777777777778,AUTHORITY[\"EPSG"
  "\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.333333333333,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"S"
  "cale factor at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32056\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32057, 1, 'NAD27 / Wyoming West Central', 'EPSG', 32057, 'PROJCS[\"NAD27 / Wyoming West Centra"
  "l\",GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORI"
  "TY[\"EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORI"
  "TY[\"EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTH"
  "ORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transvers"
  "e Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.6777777777778,AUTHORITY[\"EPSG"
  "\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-108.75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale facto"
  "r at natural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,"
  "AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32057\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32058, 1, 'NAD27 / Wyoming West', 'EPSG', 32058, 'PROJCS[\"NAD27 / Wyoming West\",GEOGCS[\"NAD"
  "27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"70"
  "08\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"62"
  "67\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\""
  "9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTH"
  "ORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.6777777777778,AUTHORITY[\"EPSG\",\"8801\"]],PAR"
  "AMETER[\"Longitude of natural origin\",-110.094444444444,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at na"
  "tural origin\",0.999941177,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806"
  "\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORI"
  "TY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32058\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32064, 1, 'NAD27 / BLM 14N (ftUS)', 'EPSG', 32064, 'PROJCS[\"NAD27 / BLM 14N (ftUS)\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32064\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32065, 1, 'NAD27 / BLM 15N (ftUS)', 'EPSG', 32065, 'PROJCS[\"NAD27 / BLM 15N (ftUS)\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32065\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32066, 1, 'NAD27 / BLM 16N (ftUS)', 'EPSG', 32066, 'PROJCS[\"NAD27 / BLM 16N (ftUS)\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32066\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32067, 1, 'NAD27 / BLM 17N (ftUS)', 'EPSG', 32067, 'PROJCS[\"NAD27 / BLM 17N (ftUS)\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",A"
  "UTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Lon"
  "gitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTH"
  "ORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False nort"
  "hing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS"
  "[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32067\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32081, 1, 'NAD27 / MTM zone 1', 'EPSG', 32081, 'PROJCS[\"NAD27 / MTM zone 1\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-53,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"32081\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32082, 1, 'NAD27 / MTM zone 2', 'EPSG', 32082, 'PROJCS[\"NAD27 / MTM zone 2\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-56,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY["
  "\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,A"
  "UTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHOR"
  "ITY[\"EPSG\",\"32082\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32083, 1, 'NAD27 / MTM zone 3', 'EPSG', 32083, 'PROJCS[\"NAD27 / MTM zone 3\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-58.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"32083\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32084, 1, 'NAD27 / MTM zone 4', 'EPSG', 32084, 'PROJCS[\"NAD27 / MTM zone 4\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-61.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"32084\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32085, 1, 'NAD27 / MTM zone 5', 'EPSG', 32085, 'PROJCS[\"NAD27 / MTM zone 5\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-64.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"32085\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32086, 1, 'NAD27 / MTM zone 6', 'EPSG', 32086, 'PROJCS[\"NAD27 / MTM zone 6\",GEOGCS[\"NAD27\""
  ",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\"7008\""
  "]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\"6267\""
  "]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"912"
  "2\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Transverse Mercator\",AUTHORI"
  "TY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude"
  " of natural origin\",-67.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9"
  "999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False "
  "northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\""
  ",NORTH],AUTHORITY[\"EPSG\",\"32086\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32098, 1, 'NAD27 / Quebec Lambert', 'EPSG', 32098, 'PROJCS[\"NAD27 / Quebec Lambert\",GEOGCS[\""
  "NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\"EPSG\",\""
  "7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\"EPSG\",\""
  "6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\""
  ",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic Conformal"
  " (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44,AUTHORITY[\"EPSG\",\"8821\"]],PARAME"
  "TER[\"Longitude of false origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standar"
  "d parallel\",60,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46,AUTHORITY[\"EPSG\",\""
  "8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\""
  ",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AU"
  "THORITY[\"EPSG\",\"32098\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32099, 1, 'NAD27 / Louisiana Offshore', 'EPSG', 32099, 'PROJCS[\"NAD27 / Louisiana Offshore\","
  "GEOGCS[\"NAD27\",DATUM[\"North American Datum 1927\",SPHEROID[\"Clarke 1866\",6378206.4,294.9786982138982,AUTHORITY[\""
  "EPSG\",\"7008\"]],TOWGS84[-32.3841359,180.4090461,120.8442577,-2.1545854,-0.1498782,0.5742915,8.1049164],AUTHORITY[\""
  "EPSG\",\"6267\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY"
  "[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4267\"]],PROJECTION[\"Lambert Conic "
  "Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",25.6777777777778,AUTHORITY[\"E"
  "PSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3333333333333,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\""
  "Latitude of 1st standard parallel\",27.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standa"
  "rd parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",2000000,AUTHORITY["
  "\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"US survey foot\",0."
  "30480060960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32099\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32100, 1, 'NAD83 / Montana', 'EPSG', 32100, 'PROJCS[\"NAD83 / Montana\",GEOGCS[\"NAD83\",DATUM"
  "[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,"
  "-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PRO"
  "JECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",44.261"
  "1111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-109.511111111111,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",49,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of "
  "2nd standard parallel\",45,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUTHORITY[\"EPS"
  "G\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32100\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32104, 1, 'NAD83 / Nebraska', 'EPSG', 32104, 'PROJCS[\"NAD83 / Nebraska\",GEOGCS[\"NAD83\",DAT"
  "UM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,"
  "1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01"
  "7453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],P"
  "ROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",39.8"
  "333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY[\"EPSG\",\"8822\"]]"
  ",PARAMETER[\"Latitude of 1st standard parallel\",43,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standar"
  "d parallel\",40,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\""
  "]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32104\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32107, 1, 'NAD83 / Nevada East', 'EPSG', 32107, 'PROJCS[\"NAD83 / Nevada East\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-115.594444444444,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",200000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",8000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32107\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32108, 1, 'NAD83 / Nevada Central', 'EPSG', 32108, 'PROJCS[\"NAD83 / Nevada Central\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34"
  ".75,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-116.666666666667,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\","
  "500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",6000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32108\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32109, 1, 'NAD83 / Nevada West', 'EPSG', 32109, 'PROJCS[\"NAD83 / Nevada West\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",34.75,AU"
  "THORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-118.594444444444,AUTHORITY[\"EPSG\",\"8802\"]]"
  ",PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",800000"
  ",AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",4000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32109\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32110, 1, 'NAD83 / New Hampshire', 'EPSG', 32110, 'PROJCS[\"NAD83 / New Hampshire\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42"
  ".5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.6666666666667,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.999966667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",300000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32110\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32111, 1, 'NAD83 / New Jersey', 'EPSG', 32111, 'PROJCS[\"NAD83 / New Jersey\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38.83333"
  "33333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111111111111,AUTHORITY[\"EPSG\""
  ",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32111\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32112, 1, 'NAD83 / New Mexico East', 'EPSG', 32112, 'PROJCS[\"NAD83 / New Mexico East\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-104.333333333333,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.999909091,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",165000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32112\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32113, 1, 'NAD83 / New Mexico Central', 'EPSG', 32113, 'PROJCS[\"NAD83 / New Mexico Central\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural or"
  "igin\",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-106.25,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32113\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32114, 1, 'NAD83 / New Mexico West', 'EPSG', 32114, 'PROJCS[\"NAD83 / New Mexico West\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\""
  ",31,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.833333333333,AUTHORITY[\"EPSG\",\"88"
  "02\"]],PARAMETER[\"Scale factor at natural origin\",0.999916667,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",830000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32114\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32115, 1, 'NAD83 / New York East', 'EPSG', 32115, 'PROJCS[\"NAD83 / New York East\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",38"
  ".8333333333333,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-74.5111111111111,AUTHORITY[\""
  "EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False"
  " easting\",150000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32115\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32116, 1, 'NAD83 / New York Central', 'EPSG', 32116, 'PROJCS[\"NAD83 / New York Central\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origi"
  "n\",40,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5944444444444,AUTHORITY[\"EPSG\",\""
  "8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easti"
  "ng\",250000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32116\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32117, 1, 'NAD83 / New York West', 'EPSG', 32117, 'PROJCS[\"NAD83 / New York West\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40"
  ",AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-78.5944444444444,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",3"
  "50000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32117\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32118, 1, 'NAD83 / New York Long Island', 'EPSG', 32118, 'PROJCS[\"NAD83 / New York Long Islan"
  "d\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude"
  " of false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-74,AUTHORI"
  "TY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]]"
  ",PARAMETER[\"Latitude of 2nd standard parallel\",40.6777777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting a"
  "t false origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8"
  "827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32118\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32119, 1, 'NAD83 / North Carolina', 'EPSG', 32119, 'PROJCS[\"NAD83 / North Carolina\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",33.75,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79,AUTHORITY[\"EPSG\",\"8822\"]],P"
  "ARAMETER[\"Latitude of 1st standard parallel\",36.1666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of"
  " 2nd standard parallel\",34.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",609601."
  "22,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32119\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32120, 1, 'NAD83 / North Dakota North', 'EPSG', 32120, 'PROJCS[\"NAD83 / North Dakota North\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.511111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",47.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32120\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32121, 1, 'NAD83 / North Dakota South', 'EPSG', 32121, 'PROJCS[\"NAD83 / North Dakota South\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",45.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.5111111111"
  "11,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.4944444444444,AUTHORITY[\"EPSG\","
  "\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",46.1944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\""
  "Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"E"
  "PSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\""
  ",\"32121\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32122, 1, 'NAD83 / Ohio North', 'EPSG', 32122, 'PROJCS[\"NAD83 / Ohio North\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "39.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lati"
  "tude of 2nd standard parallel\",40.4444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\","
  "600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"me"
  "tre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32122\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32123, 1, 'NAD83 / Ohio South', 'EPSG', 32123, 'PROJCS[\"NAD83 / Ohio South\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "38,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",40.0333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",38.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600"
  "000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre"
  "\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32123\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32124, 1, 'NAD83 / Oklahoma North', 'EPSG', 32124, 'PROJCS[\"NAD83 / Oklahoma North\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",35,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\"8822\"]],PARA"
  "METER[\"Latitude of 1st standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2n"
  "d standard parallel\",35.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,AUT"
  "HORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AU"
  "THORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32124\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32125, 1, 'NAD83 / Oklahoma South', 'EPSG', 32125, 'PROJCS[\"NAD83 / Oklahoma South\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",33.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",35.2333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",33.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32125\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32126, 1, 'NAD83 / Oregon North', 'EPSG', 32126, 'PROJCS[\"NAD83 / Oregon North\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",43.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",44.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",2500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32126\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32127, 1, 'NAD83 / Oregon South', 'EPSG', 32127, 'PROJCS[\"NAD83 / Oregon South\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",41.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"La"
  "titude of 2nd standard parallel\",42.3333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",1500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32127\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32128, 1, 'NAD83 / Pennsylvania North', 'EPSG', 32128, 'PROJCS[\"NAD83 / Pennsylvania North\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",40.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77.75,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.95,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\""
  "Latitude of 2nd standard parallel\",40.8833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origi"
  "n\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT["
  "\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32128\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32129, 1, 'NAD83 / Pennsylvania South', 'EPSG', 32129, 'PROJCS[\"NAD83 / Pennsylvania South\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",39.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-77.75,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]],"
  "PARAMETER[\"Latitude of 2nd standard parallel\",39.9333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at"
  " false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32129\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32130, 1, 'NAD83 / Rhode Island', 'EPSG', 32130, 'PROJCS[\"NAD83 / Rhode Island\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",41.0"
  "944444444444,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-71.5111111111111,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.99999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fal"
  "se easting\",100000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32130\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32133, 1, 'NAD83 / South Carolina', 'EPSG', 32133, 'PROJCS[\"NAD83 / South Carolina\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",31.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",34.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"L"
  "atitude of 2nd standard parallel\",32.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin"
  "\",609600,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32133\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32134, 1, 'NAD83 / South Dakota North', 'EPSG', 32134, 'PROJCS[\"NAD83 / South Dakota North\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100,AUTHORITY["
  "\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.6944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PA"
  "RAMETER[\"Latitude of 2nd standard parallel\",44.4166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32134\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32135, 1, 'NAD83 / South Dakota South', 'EPSG', 32135, 'PROJCS[\"NAD83 / South Dakota South\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",42.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.3333333333"
  "33,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",44.4,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",42.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"882"
  "7\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32135\"]"
  "]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32136, 1, 'NAD83 / Tennessee', 'EPSG', 32136, 'PROJCS[\"NAD83 / Tennessee\",GEOGCS[\"NAD83\",D"
  "ATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84["
  "1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0."
  "017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]]"
  ",PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\",34"
  ".3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-86,AUTHORITY[\"EPSG\",\"8822\"]"
  "],PARAMETER[\"Latitude of 1st standard parallel\",36.4166666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude"
  " of 2nd standard parallel\",35.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",6000"
  "00,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32136\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32137, 1, 'NAD83 / Texas North', 'EPSG', 32137, 'PROJCS[\"NAD83 / Texas North\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "34,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-101.511111111111,AUTHORITY[\"EPSG\",\"8822\""
  "]],PARAMETER[\"Latitude of 1st standard parallel\",36.1944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitud"
  "e of 2nd standard parallel\",34.65,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",200000,AUTHORI"
  "TY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32137\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32138, 1, 'NAD83 / Texas North Central', 'EPSG', 32138, 'PROJCS[\"NAD83 / Texas North Central\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",31.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.511111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",33.9777777777778,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",32.1444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER["
  "\"Easting at false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2000000,AUTHO"
  "RITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY["
  "\"EPSG\",\"32138\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32139, 1, 'NAD83 / Texas Central', 'EPSG', 32139, 'PROJCS[\"NAD83 / Texas Central\",GEOGCS[\"N"
  "AD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],"
  "TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",29.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-100.333333333333,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",31.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",30.1166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",700000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "32139\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32140, 1, 'NAD83 / Texas South Central', 'EPSG', 32140, 'PROJCS[\"NAD83 / Texas South Central\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",27.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-99,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",30.2833333333333,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",28.3833333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",4000000,AUTHORITY[\"EPSG\""
  ",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32"
  "140\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32141, 1, 'NAD83 / Texas South', 'EPSG', 32141, 'PROJCS[\"NAD83 / Texas South\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "25.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-98.5111111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",300000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",5000000,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32141"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32142, 1, 'NAD83 / Utah North', 'EPSG', 32142, 'PROJCS[\"NAD83 / Utah North\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "40.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",41.7944444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",40.7277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EPSG\",\""
  "8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32142"
  "\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32143, 1, 'NAD83 / Utah Central', 'EPSG', 32143, 'PROJCS[\"NAD83 / Utah Central\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origi"
  "n\",38.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHORIT"
  "Y[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.65,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\""
  "Latitude of 2nd standard parallel\",39.0277777777778,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origi"
  "n\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2000000,AUTHORITY[\"EPSG\",\"8827\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32143\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32144, 1, 'NAD83 / Utah South', 'EPSG', 32144, 'PROJCS[\"NAD83 / Utah South\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false origin\","
  "36.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-111.511111111111,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.35,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lat"
  "itude of 2nd standard parallel\",37.2166666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",3000000,AUTHORITY[\"EPSG\",\"8827\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32144\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32145, 1, 'NAD83 / Vermont', 'EPSG', 32145, 'PROJCS[\"NAD83 / Vermont\",GEOGCS[\"NAD83\",DATUM"
  "[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS84[1,1,"
  "-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174"
  "53292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\"]],PRO"
  "JECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",42.51111111111"
  "11,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-72.5111111111111,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.999964286,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False eastin"
  "g\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\","
  "1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32145\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32146, 1, 'NAD83 / Virginia North', 'EPSG', 32146, 'PROJCS[\"NAD83 / Virginia North\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",37.6777777777778,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",39.2111111111111,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",38.0333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",2000000,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"32146\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32147, 1, 'NAD83 / Virginia South', 'EPSG', 32147, 'PROJCS[\"NAD83 / Virginia South\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",36.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-78.5111111111111,AUTHOR"
  "ITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",37.9777777777778,AUTHORITY[\"EPSG\",\"8823\"]"
  "],PARAMETER[\"Latitude of 2nd standard parallel\",36.7666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting "
  "at false origin\",3500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",1000000,AUTHORITY[\"EP"
  "SG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\","
  "\"32147\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32148, 1, 'NAD83 / Washington North', 'EPSG', 32148, 'PROJCS[\"NAD83 / Washington North\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",47,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.833333333333,AUTHORITY[\"EP"
  "SG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",48.7444444444444,AUTHORITY[\"EPSG\",\"8823\"]],PARAME"
  "TER[\"Latitude of 2nd standard parallel\",47.5111111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false"
  " origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]]"
  ",UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32148\"]]', N"
  "ULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32149, 1, 'NAD83 / Washington South', 'EPSG', 32149, 'PROJCS[\"NAD83 / Washington South\",GEOG"
  "CS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"701"
  "9\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT"
  "[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPS"
  "G\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of fal"
  "se origin\",45.3333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-120.511111111111,"
  "AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",47.3333333333333,AUTHORITY[\"EPSG\",\"8"
  "823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",45.8333333333333,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Ea"
  "sting at false origin\",500000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPS"
  "G\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\""
  "32149\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32150, 1, 'NAD83 / West Virginia North', 'EPSG', 32150, 'PROJCS[\"NAD83 / West Virginia North\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",38.5111111111111,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-79.511111111"
  "1111,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",40.2611111111111,AUTHORITY[\"EPSG\""
  ",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",39,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at f"
  "alse origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827"
  "\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32150\"]]"
  "', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32151, 1, 'NAD83 / West Virginia South', 'EPSG', 32151, 'PROJCS[\"NAD83 / West Virginia South\""
  ",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\","
  "\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]]"
  ",UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY["
  "\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude o"
  "f false origin\",37,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-81,AUTHORITY[\"EPSG\",\"88"
  "22\"]],PARAMETER[\"Latitude of 1st standard parallel\",38.8833333333333,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Lat"
  "itude of 2nd standard parallel\",37.4944444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\""
  ",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"m"
  "etre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32151\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32152, 1, 'NAD83 / Wisconsin North', 'EPSG', 32152, 'PROJCS[\"NAD83 / Wisconsin North\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",45.1666666666667,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\""
  ",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",46.7666666666667,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER["
  "\"Latitude of 2nd standard parallel\",45.5666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false ori"
  "gin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32152\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32153, 1, 'NAD83 / Wisconsin Central', 'EPSG', 32153, 'PROJCS[\"NAD83 / Wisconsin Central\",GE"
  "OGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7"
  "019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UN"
  "IT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"E"
  "PSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of f"
  "alse origin\",43.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\""
  "EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",45.5111111111111,AUTHORITY[\"EPSG\",\"8823\"]],PARA"
  "METER[\"Latitude of 2nd standard parallel\",44.2611111111111,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at fal"
  "se origin\",600000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\""
  "]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32153\"]]',"
  " NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32154, 1, 'NAD83 / Wisconsin South', 'EPSG', 32154, 'PROJCS[\"NAD83 / Wisconsin South\",GEOGCS"
  "[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\""
  "]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\""
  "degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\""
  ",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false "
  "origin\",42,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-90,AUTHORITY[\"EPSG\",\"8822\"]],P"
  "ARAMETER[\"Latitude of 1st standard parallel\",44.0777777777778,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of"
  " 2nd standard parallel\",42.7444444444444,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",600000,"
  "AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32154\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32155, 1, 'NAD83 / Wyoming East', 'EPSG', 32155, 'PROJCS[\"NAD83 / Wyoming East\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105.177777777778,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",200000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\""
  "metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32155\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32156, 1, 'NAD83 / Wyoming East Central', 'EPSG', 32156, 'PROJCS[\"NAD83 / Wyoming East Centra"
  "l\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-107.333333333333"
  ",AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],P"
  "ARAMETER[\"False easting\",400000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY[\"EPSG\""
  ",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32"
  "156\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32157, 1, 'NAD83 / Wyoming West Central', 'EPSG', 32157, 'PROJCS[\"NAD83 / Wyoming West Centra"
  "l\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG"
  "\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\""
  "]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORIT"
  "Y[\"EPSG\",\"4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natura"
  "l origin\",40.5111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-108.75,AUTHORITY"
  "[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\""
  "False easting\",600000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNI"
  "T[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32157\"]]', NULL)"
  ";\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32158, 1, 'NAD83 / Wyoming West', 'EPSG', 32158, 'PROJCS[\"NAD83 / Wyoming West\",GEOGCS[\"NAD"
  "83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TO"
  "WGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degre"
  "e\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"42"
  "69\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",40.5"
  "111111111111,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-110.094444444444,AUTHORITY[\"EP"
  "SG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\",0.9999375,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"Fals"
  "e easting\",800000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",100000,AUTHORITY[\"EPSG\",\"8807\"]],UN"
  "IT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32158\"]]', NULL"
  ");\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32161, 1, 'NAD83 / Puerto Rico & Virgin Is.', 'EPSG', 32161, 'PROJCS[\"NAD83 / Puerto Rico & V"
  "irgin Is.\",GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORIT"
  "Y[\"EPSG\",\"7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\""
  ",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST]"
  ",AUTHORITY[\"EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\""
  "Latitude of false origin\",17.8333333333333,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-66"
  ".4333333333334,AUTHORITY[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",18.4333333333333,AUTHORI"
  "TY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd standard parallel\",18.0444444444444,AUTHORITY[\"EPSG\",\"8824\"]]"
  ",PARAMETER[\"Easting at false origin\",200000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",20"
  "0000,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],"
  "AUTHORITY[\"EPSG\",\"32161\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32164, 1, 'NAD83 / BLM 14N (ftUS)', 'EPSG', 32164, 'PROJCS[\"NAD83 / BLM 14N (ftUS)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32164\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32165, 1, 'NAD83 / BLM 15N (ftUS)', 'EPSG', 32165, 'PROJCS[\"NAD83 / BLM 15N (ftUS)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32165\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32166, 1, 'NAD83 / BLM 16N (ftUS)', 'EPSG', 32166, 'PROJCS[\"NAD83 / BLM 16N (ftUS)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32166\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32167, 1, 'NAD83 / BLM 17N (ftUS)', 'EPSG', 32167, 'PROJCS[\"NAD83 / BLM 17N (ftUS)\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060"
  "960121924,AUTHORITY[\"EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32167\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32181, 1, 'NAD83 / MTM zone 1', 'EPSG', 32181, 'PROJCS[\"NAD83 / MTM zone 1\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-53,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32181\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32182, 1, 'NAD83 / MTM zone 2', 'EPSG', 32182, 'PROJCS[\"NAD83 / MTM zone 2\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-56,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32182\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32183, 1, 'NAD83 / MTM zone 3', 'EPSG', 32183, 'PROJCS[\"NAD83 / MTM zone 3\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-58.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32183\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32184, 1, 'NAD83 / MTM zone 4', 'EPSG', 32184, 'PROJCS[\"NAD83 / MTM zone 4\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-61.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32184\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32185, 1, 'NAD83 / MTM zone 5', 'EPSG', 32185, 'PROJCS[\"NAD83 / MTM zone 5\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-64.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32185\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32186, 1, 'NAD83 / MTM zone 6', 'EPSG', 32186, 'PROJCS[\"NAD83 / MTM zone 6\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-67.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32186\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32187, 1, 'NAD83 / MTM zone 7', 'EPSG', 32187, 'PROJCS[\"NAD83 / MTM zone 7\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-70.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32187\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32188, 1, 'NAD83 / MTM zone 8', 'EPSG', 32188, 'PROJCS[\"NAD83 / MTM zone 8\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-73.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32188\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32189, 1, 'NAD83 / MTM zone 9', 'EPSG', 32189, 'PROJCS[\"NAD83 / MTM zone 9\",GEOGCS[\"NAD83\""
  ",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWGS8"
  "4[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\","
  "0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-76.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32189\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32190, 1, 'NAD83 / MTM zone 10', 'EPSG', 32190, 'PROJCS[\"NAD83 / MTM zone 10\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-79.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E(X)\",EAST],AXIS[\"N(Y)\",NORTH],AUTHORITY[\"EPSG\",\"32190\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32191, 1, 'NAD83 / MTM zone 11', 'EPSG', 32191, 'PROJCS[\"NAD83 / MTM zone 11\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-82.5111111111111,AUTHORITY[\"EPSG\",\"8802\"]],PAR"
  "AMETER[\"Scale factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUT"
  "HORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32191\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32192, 1, 'NAD83 / MTM zone 12', 'EPSG', 32192, 'PROJCS[\"NAD83 / MTM zone 12\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32192\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32193, 1, 'NAD83 / MTM zone 13', 'EPSG', 32193, 'PROJCS[\"NAD83 / MTM zone 13\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-84,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32193\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32194, 1, 'NAD83 / MTM zone 14', 'EPSG', 32194, 'PROJCS[\"NAD83 / MTM zone 14\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32194\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32195, 1, 'NAD83 / MTM zone 15', 'EPSG', 32195, 'PROJCS[\"NAD83 / MTM zone 15\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-90,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32195\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32196, 1, 'NAD83 / MTM zone 16', 'EPSG', 32196, 'PROJCS[\"NAD83 / MTM zone 16\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32196\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32197, 1, 'NAD83 / MTM zone 17', 'EPSG', 32197, 'PROJCS[\"NAD83 / MTM zone 17\",GEOGCS[\"NAD83"
  "\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]],TOWG"
  "S84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\""
  ",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4269\""
  "]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHOR"
  "ITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-96,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.9999,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",304800,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\""
  "]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32197\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32198, 1, 'NAD83 / Quebec Lambert', 'EPSG', 32198, 'PROJCS[\"NAD83 / Quebec Lambert\",GEOGCS[\""
  "NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\"7019\"]]"
  ",TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"de"
  "gree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of false ori"
  "gin\",44,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-68.5111111111111,AUTHORITY[\"EPSG\",\""
  "8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",60,AUTHORITY[\"EPSG\",\"8823\"]],PARAMETER[\"Latitude of 2nd"
  " standard parallel\",46,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "26\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"8827\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9"
  "001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32198\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32199, 1, 'NAD83 / Louisiana Offshore', 'EPSG', 32199, 'PROJCS[\"NAD83 / Louisiana Offshore\","
  "GEOGCS[\"NAD83\",DATUM[\"North American Datum 1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101,AUTHORITY[\"EPSG\",\""
  "7019\"]],TOWGS84[1,1,-1,0,0,0,0],AUTHORITY[\"EPSG\",\"6269\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],U"
  "NIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\""
  "EPSG\",\"4269\"]],PROJECTION[\"Lambert Conic Conformal (2SP)\",AUTHORITY[\"EPSG\",\"9802\"]],PARAMETER[\"Latitude of "
  "false origin\",25.5,AUTHORITY[\"EPSG\",\"8821\"]],PARAMETER[\"Longitude of false origin\",-91.3333333333333,AUTHORITY"
  "[\"EPSG\",\"8822\"]],PARAMETER[\"Latitude of 1st standard parallel\",27.8333333333333,AUTHORITY[\"EPSG\",\"8823\"]],P"
  "ARAMETER[\"Latitude of 2nd standard parallel\",26.1666666666667,AUTHORITY[\"EPSG\",\"8824\"]],PARAMETER[\"Easting at "
  "false origin\",1000000,AUTHORITY[\"EPSG\",\"8826\"]],PARAMETER[\"Northing at false origin\",0,AUTHORITY[\"EPSG\",\"88"
  "27\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32199\""
  "]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32201, 1, 'WGS 72 / UTM zone 1N', 'EPSG', 32201, 'PROJCS[\"WGS 72 / UTM zone 1N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32201\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32202, 1, 'WGS 72 / UTM zone 2N', 'EPSG', 32202, 'PROJCS[\"WGS 72 / UTM zone 2N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32202\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32203, 1, 'WGS 72 / UTM zone 3N', 'EPSG', 32203, 'PROJCS[\"WGS 72 / UTM zone 3N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32203\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32204, 1, 'WGS 72 / UTM zone 4N', 'EPSG', 32204, 'PROJCS[\"WGS 72 / UTM zone 4N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32204\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32205, 1, 'WGS 72 / UTM zone 5N', 'EPSG', 32205, 'PROJCS[\"WGS 72 / UTM zone 5N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32205\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32206, 1, 'WGS 72 / UTM zone 6N', 'EPSG', 32206, 'PROJCS[\"WGS 72 / UTM zone 6N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32206\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32207, 1, 'WGS 72 / UTM zone 7N', 'EPSG', 32207, 'PROJCS[\"WGS 72 / UTM zone 7N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32207\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32208, 1, 'WGS 72 / UTM zone 8N', 'EPSG', 32208, 'PROJCS[\"WGS 72 / UTM zone 8N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32208\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32209, 1, 'WGS 72 / UTM zone 9N', 'EPSG', 32209, 'PROJCS[\"WGS 72 / UTM zone 9N\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32209\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32210, 1, 'WGS 72 / UTM zone 10N', 'EPSG', 32210, 'PROJCS[\"WGS 72 / UTM zone 10N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32210\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32211, 1, 'WGS 72 / UTM zone 11N', 'EPSG', 32211, 'PROJCS[\"WGS 72 / UTM zone 11N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32211\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32212, 1, 'WGS 72 / UTM zone 12N', 'EPSG', 32212, 'PROJCS[\"WGS 72 / UTM zone 12N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32212\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32213, 1, 'WGS 72 / UTM zone 13N', 'EPSG', 32213, 'PROJCS[\"WGS 72 / UTM zone 13N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32213\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32214, 1, 'WGS 72 / UTM zone 14N', 'EPSG', 32214, 'PROJCS[\"WGS 72 / UTM zone 14N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32214\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32215, 1, 'WGS 72 / UTM zone 15N', 'EPSG', 32215, 'PROJCS[\"WGS 72 / UTM zone 15N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32215\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32216, 1, 'WGS 72 / UTM zone 16N', 'EPSG', 32216, 'PROJCS[\"WGS 72 / UTM zone 16N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32216\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32217, 1, 'WGS 72 / UTM zone 17N', 'EPSG', 32217, 'PROJCS[\"WGS 72 / UTM zone 17N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32217\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32218, 1, 'WGS 72 / UTM zone 18N', 'EPSG', 32218, 'PROJCS[\"WGS 72 / UTM zone 18N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32218\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32219, 1, 'WGS 72 / UTM zone 19N', 'EPSG', 32219, 'PROJCS[\"WGS 72 / UTM zone 19N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32219\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32220, 1, 'WGS 72 / UTM zone 20N', 'EPSG', 32220, 'PROJCS[\"WGS 72 / UTM zone 20N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32220\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32221, 1, 'WGS 72 / UTM zone 21N', 'EPSG', 32221, 'PROJCS[\"WGS 72 / UTM zone 21N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32221\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32222, 1, 'WGS 72 / UTM zone 22N', 'EPSG', 32222, 'PROJCS[\"WGS 72 / UTM zone 22N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32222\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32223, 1, 'WGS 72 / UTM zone 23N', 'EPSG', 32223, 'PROJCS[\"WGS 72 / UTM zone 23N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32223\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32224, 1, 'WGS 72 / UTM zone 24N', 'EPSG', 32224, 'PROJCS[\"WGS 72 / UTM zone 24N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32224\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32225, 1, 'WGS 72 / UTM zone 25N', 'EPSG', 32225, 'PROJCS[\"WGS 72 / UTM zone 25N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32225\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32226, 1, 'WGS 72 / UTM zone 26N', 'EPSG', 32226, 'PROJCS[\"WGS 72 / UTM zone 26N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32226\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32227, 1, 'WGS 72 / UTM zone 27N', 'EPSG', 32227, 'PROJCS[\"WGS 72 / UTM zone 27N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32227\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32228, 1, 'WGS 72 / UTM zone 28N', 'EPSG', 32228, 'PROJCS[\"WGS 72 / UTM zone 28N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32228\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32229, 1, 'WGS 72 / UTM zone 29N', 'EPSG', 32229, 'PROJCS[\"WGS 72 / UTM zone 29N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32229\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32230, 1, 'WGS 72 / UTM zone 30N', 'EPSG', 32230, 'PROJCS[\"WGS 72 / UTM zone 30N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32230\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32231, 1, 'WGS 72 / UTM zone 31N', 'EPSG', 32231, 'PROJCS[\"WGS 72 / UTM zone 31N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32231\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32232, 1, 'WGS 72 / UTM zone 32N', 'EPSG', 32232, 'PROJCS[\"WGS 72 / UTM zone 32N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32232\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32233, 1, 'WGS 72 / UTM zone 33N', 'EPSG', 32233, 'PROJCS[\"WGS 72 / UTM zone 33N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32233\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32234, 1, 'WGS 72 / UTM zone 34N', 'EPSG', 32234, 'PROJCS[\"WGS 72 / UTM zone 34N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32234\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32235, 1, 'WGS 72 / UTM zone 35N', 'EPSG', 32235, 'PROJCS[\"WGS 72 / UTM zone 35N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32235\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32236, 1, 'WGS 72 / UTM zone 36N', 'EPSG', 32236, 'PROJCS[\"WGS 72 / UTM zone 36N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32236\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32237, 1, 'WGS 72 / UTM zone 37N', 'EPSG', 32237, 'PROJCS[\"WGS 72 / UTM zone 37N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32237\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32238, 1, 'WGS 72 / UTM zone 38N', 'EPSG', 32238, 'PROJCS[\"WGS 72 / UTM zone 38N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32238\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32239, 1, 'WGS 72 / UTM zone 39N', 'EPSG', 32239, 'PROJCS[\"WGS 72 / UTM zone 39N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32239\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32240, 1, 'WGS 72 / UTM zone 40N', 'EPSG', 32240, 'PROJCS[\"WGS 72 / UTM zone 40N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32240\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32241, 1, 'WGS 72 / UTM zone 41N', 'EPSG', 32241, 'PROJCS[\"WGS 72 / UTM zone 41N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32241\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32242, 1, 'WGS 72 / UTM zone 42N', 'EPSG', 32242, 'PROJCS[\"WGS 72 / UTM zone 42N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32242\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32243, 1, 'WGS 72 / UTM zone 43N', 'EPSG', 32243, 'PROJCS[\"WGS 72 / UTM zone 43N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32243\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32244, 1, 'WGS 72 / UTM zone 44N', 'EPSG', 32244, 'PROJCS[\"WGS 72 / UTM zone 44N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32244\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32245, 1, 'WGS 72 / UTM zone 45N', 'EPSG', 32245, 'PROJCS[\"WGS 72 / UTM zone 45N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32245\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32246, 1, 'WGS 72 / UTM zone 46N', 'EPSG', 32246, 'PROJCS[\"WGS 72 / UTM zone 46N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32246\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32247, 1, 'WGS 72 / UTM zone 47N', 'EPSG', 32247, 'PROJCS[\"WGS 72 / UTM zone 47N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\","
  "\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32247\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32248, 1, 'WGS 72 / UTM zone 48N', 'EPSG', 32248, 'PROJCS[\"WGS 72 / UTM zone 48N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32248\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32249, 1, 'WGS 72 / UTM zone 49N', 'EPSG', 32249, 'PROJCS[\"WGS 72 / UTM zone 49N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32249\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32250, 1, 'WGS 72 / UTM zone 50N', 'EPSG', 32250, 'PROJCS[\"WGS 72 / UTM zone 50N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32250\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32251, 1, 'WGS 72 / UTM zone 51N', 'EPSG', 32251, 'PROJCS[\"WGS 72 / UTM zone 51N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32251\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32252, 1, 'WGS 72 / UTM zone 52N', 'EPSG', 32252, 'PROJCS[\"WGS 72 / UTM zone 52N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32252\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32253, 1, 'WGS 72 / UTM zone 53N', 'EPSG', 32253, 'PROJCS[\"WGS 72 / UTM zone 53N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32253\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32254, 1, 'WGS 72 / UTM zone 54N', 'EPSG', 32254, 'PROJCS[\"WGS 72 / UTM zone 54N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32254\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32255, 1, 'WGS 72 / UTM zone 55N', 'EPSG', 32255, 'PROJCS[\"WGS 72 / UTM zone 55N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32255\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32256, 1, 'WGS 72 / UTM zone 56N', 'EPSG', 32256, 'PROJCS[\"WGS 72 / UTM zone 56N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32256\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32257, 1, 'WGS 72 / UTM zone 57N', 'EPSG', 32257, 'PROJCS[\"WGS 72 / UTM zone 57N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32257\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32258, 1, 'WGS 72 / UTM zone 58N', 'EPSG', 32258, 'PROJCS[\"WGS 72 / UTM zone 58N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32258\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32259, 1, 'WGS 72 / UTM zone 59N', 'EPSG', 32259, 'PROJCS[\"WGS 72 / UTM zone 59N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32259\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32260, 1, 'WGS 72 / UTM zone 60N', 'EPSG', 32260, 'PROJCS[\"WGS 72 / UTM zone 60N\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\""
  ",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32260\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32301, 1, 'WGS 72 / UTM zone 1S', 'EPSG', 32301, 'PROJCS[\"WGS 72 / UTM zone 1S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32301\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32302, 1, 'WGS 72 / UTM zone 2S', 'EPSG', 32302, 'PROJCS[\"WGS 72 / UTM zone 2S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32302\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32303, 1, 'WGS 72 / UTM zone 3S', 'EPSG', 32303, 'PROJCS[\"WGS 72 / UTM zone 3S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32303\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32304, 1, 'WGS 72 / UTM zone 4S', 'EPSG', 32304, 'PROJCS[\"WGS 72 / UTM zone 4S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32304\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32305, 1, 'WGS 72 / UTM zone 5S', 'EPSG', 32305, 'PROJCS[\"WGS 72 / UTM zone 5S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32305\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32306, 1, 'WGS 72 / UTM zone 6S', 'EPSG', 32306, 'PROJCS[\"WGS 72 / UTM zone 6S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32306\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32307, 1, 'WGS 72 / UTM zone 7S', 'EPSG', 32307, 'PROJCS[\"WGS 72 / UTM zone 7S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32307\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32308, 1, 'WGS 72 / UTM zone 8S', 'EPSG', 32308, 'PROJCS[\"WGS 72 / UTM zone 8S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32308\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32309, 1, 'WGS 72 / UTM zone 9S', 'EPSG', 32309, 'PROJCS[\"WGS 72 / UTM zone 9S\",GEOGCS[\"WGS"
  " 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84[0"
  ",0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"deg"
  "ree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\""
  "4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,"
  "AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32309\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32310, 1, 'WGS 72 / UTM zone 10S', 'EPSG', 32310, 'PROJCS[\"WGS 72 / UTM zone 10S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32310\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32311, 1, 'WGS 72 / UTM zone 11S', 'EPSG', 32311, 'PROJCS[\"WGS 72 / UTM zone 11S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32311\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32312, 1, 'WGS 72 / UTM zone 12S', 'EPSG', 32312, 'PROJCS[\"WGS 72 / UTM zone 12S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32312\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32313, 1, 'WGS 72 / UTM zone 13S', 'EPSG', 32313, 'PROJCS[\"WGS 72 / UTM zone 13S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETE"
  "R[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORIT"
  "Y[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY["
  "\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32313\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32314, 1, 'WGS 72 / UTM zone 14S', 'EPSG', 32314, 'PROJCS[\"WGS 72 / UTM zone 14S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32314\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32315, 1, 'WGS 72 / UTM zone 15S', 'EPSG', 32315, 'PROJCS[\"WGS 72 / UTM zone 15S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32315\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32316, 1, 'WGS 72 / UTM zone 16S', 'EPSG', 32316, 'PROJCS[\"WGS 72 / UTM zone 16S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32316\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32317, 1, 'WGS 72 / UTM zone 17S', 'EPSG', 32317, 'PROJCS[\"WGS 72 / UTM zone 17S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32317\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32318, 1, 'WGS 72 / UTM zone 18S', 'EPSG', 32318, 'PROJCS[\"WGS 72 / UTM zone 18S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32318\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32319, 1, 'WGS 72 / UTM zone 19S', 'EPSG', 32319, 'PROJCS[\"WGS 72 / UTM zone 19S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32319\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32320, 1, 'WGS 72 / UTM zone 20S', 'EPSG', 32320, 'PROJCS[\"WGS 72 / UTM zone 20S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32320\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32321, 1, 'WGS 72 / UTM zone 21S', 'EPSG', 32321, 'PROJCS[\"WGS 72 / UTM zone 21S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32321\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32322, 1, 'WGS 72 / UTM zone 22S', 'EPSG', 32322, 'PROJCS[\"WGS 72 / UTM zone 22S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32322\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32323, 1, 'WGS 72 / UTM zone 23S', 'EPSG', 32323, 'PROJCS[\"WGS 72 / UTM zone 23S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32323\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32324, 1, 'WGS 72 / UTM zone 24S', 'EPSG', 32324, 'PROJCS[\"WGS 72 / UTM zone 24S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32324\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32325, 1, 'WGS 72 / UTM zone 25S', 'EPSG', 32325, 'PROJCS[\"WGS 72 / UTM zone 25S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32325\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32326, 1, 'WGS 72 / UTM zone 26S', 'EPSG', 32326, 'PROJCS[\"WGS 72 / UTM zone 26S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32326\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32327, 1, 'WGS 72 / UTM zone 27S', 'EPSG', 32327, 'PROJCS[\"WGS 72 / UTM zone 27S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32327\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32328, 1, 'WGS 72 / UTM zone 28S', 'EPSG', 32328, 'PROJCS[\"WGS 72 / UTM zone 28S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32328\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32329, 1, 'WGS 72 / UTM zone 29S', 'EPSG', 32329, 'PROJCS[\"WGS 72 / UTM zone 29S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32329\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32330, 1, 'WGS 72 / UTM zone 30S', 'EPSG', 32330, 'PROJCS[\"WGS 72 / UTM zone 30S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32330\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32331, 1, 'WGS 72 / UTM zone 31S', 'EPSG', 32331, 'PROJCS[\"WGS 72 / UTM zone 31S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32331\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32332, 1, 'WGS 72 / UTM zone 32S', 'EPSG', 32332, 'PROJCS[\"WGS 72 / UTM zone 32S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\""
  "Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\""
  "EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EP"
  "SG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32332\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32333, 1, 'WGS 72 / UTM zone 33S', 'EPSG', 32333, 'PROJCS[\"WGS 72 / UTM zone 33S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32333\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32334, 1, 'WGS 72 / UTM zone 34S', 'EPSG', 32334, 'PROJCS[\"WGS 72 / UTM zone 34S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32334\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32335, 1, 'WGS 72 / UTM zone 35S', 'EPSG', 32335, 'PROJCS[\"WGS 72 / UTM zone 35S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32335\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32336, 1, 'WGS 72 / UTM zone 36S', 'EPSG', 32336, 'PROJCS[\"WGS 72 / UTM zone 36S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32336\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32337, 1, 'WGS 72 / UTM zone 37S', 'EPSG', 32337, 'PROJCS[\"WGS 72 / UTM zone 37S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32337\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32338, 1, 'WGS 72 / UTM zone 38S', 'EPSG', 32338, 'PROJCS[\"WGS 72 / UTM zone 38S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32338\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32339, 1, 'WGS 72 / UTM zone 39S', 'EPSG', 32339, 'PROJCS[\"WGS 72 / UTM zone 39S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32339\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32340, 1, 'WGS 72 / UTM zone 40S', 'EPSG', 32340, 'PROJCS[\"WGS 72 / UTM zone 40S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32340\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32341, 1, 'WGS 72 / UTM zone 41S', 'EPSG', 32341, 'PROJCS[\"WGS 72 / UTM zone 41S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32341\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32342, 1, 'WGS 72 / UTM zone 42S', 'EPSG', 32342, 'PROJCS[\"WGS 72 / UTM zone 42S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32342\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32343, 1, 'WGS 72 / UTM zone 43S', 'EPSG', 32343, 'PROJCS[\"WGS 72 / UTM zone 43S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32343\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32344, 1, 'WGS 72 / UTM zone 44S', 'EPSG', 32344, 'PROJCS[\"WGS 72 / UTM zone 44S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32344\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32345, 1, 'WGS 72 / UTM zone 45S', 'EPSG', 32345, 'PROJCS[\"WGS 72 / UTM zone 45S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32345\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32346, 1, 'WGS 72 / UTM zone 46S', 'EPSG', 32346, 'PROJCS[\"WGS 72 / UTM zone 46S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32346\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32347, 1, 'WGS 72 / UTM zone 47S', 'EPSG', 32347, 'PROJCS[\"WGS 72 / UTM zone 47S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER["
  "\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY["
  "\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32347\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32348, 1, 'WGS 72 / UTM zone 48S', 'EPSG', 32348, 'PROJCS[\"WGS 72 / UTM zone 48S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32348\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32349, 1, 'WGS 72 / UTM zone 49S', 'EPSG', 32349, 'PROJCS[\"WGS 72 / UTM zone 49S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32349\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32350, 1, 'WGS 72 / UTM zone 50S', 'EPSG', 32350, 'PROJCS[\"WGS 72 / UTM zone 50S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32350\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32351, 1, 'WGS 72 / UTM zone 51S', 'EPSG', 32351, 'PROJCS[\"WGS 72 / UTM zone 51S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32351\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32352, 1, 'WGS 72 / UTM zone 52S', 'EPSG', 32352, 'PROJCS[\"WGS 72 / UTM zone 52S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32352\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32353, 1, 'WGS 72 / UTM zone 53S', 'EPSG', 32353, 'PROJCS[\"WGS 72 / UTM zone 53S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32353\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32354, 1, 'WGS 72 / UTM zone 54S', 'EPSG', 32354, 'PROJCS[\"WGS 72 / UTM zone 54S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32354\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32355, 1, 'WGS 72 / UTM zone 55S', 'EPSG', 32355, 'PROJCS[\"WGS 72 / UTM zone 55S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32355\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32356, 1, 'WGS 72 / UTM zone 56S', 'EPSG', 32356, 'PROJCS[\"WGS 72 / UTM zone 56S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32356\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32357, 1, 'WGS 72 / UTM zone 57S', 'EPSG', 32357, 'PROJCS[\"WGS 72 / UTM zone 57S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32357\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32358, 1, 'WGS 72 / UTM zone 58S', 'EPSG', 32358, 'PROJCS[\"WGS 72 / UTM zone 58S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32358\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32359, 1, 'WGS 72 / UTM zone 59S', 'EPSG', 32359, 'PROJCS[\"WGS 72 / UTM zone 59S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32359\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32360, 1, 'WGS 72 / UTM zone 60S', 'EPSG', 32360, 'PROJCS[\"WGS 72 / UTM zone 60S\",GEOGCS[\"W"
  "GS 72\",DATUM[\"World Geodetic System 1972\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\"]],TOWGS84"
  "[0,0,4.5,0,0,0.554,0.219],AUTHORITY[\"EPSG\",\"6322\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"d"
  "egree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\","
  "\"4322\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\","
  "0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER"
  "[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY"
  "[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\""
  "EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32360\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32401, 1, 'WGS 72BE / UTM zone 1N', 'EPSG', 32401, 'PROJCS[\"WGS 72BE / UTM zone 1N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32401\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32402, 1, 'WGS 72BE / UTM zone 2N', 'EPSG', 32402, 'PROJCS[\"WGS 72BE / UTM zone 2N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32402\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32403, 1, 'WGS 72BE / UTM zone 3N', 'EPSG', 32403, 'PROJCS[\"WGS 72BE / UTM zone 3N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32403\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32404, 1, 'WGS 72BE / UTM zone 4N', 'EPSG', 32404, 'PROJCS[\"WGS 72BE / UTM zone 4N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32404\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32405, 1, 'WGS 72BE / UTM zone 5N', 'EPSG', 32405, 'PROJCS[\"WGS 72BE / UTM zone 5N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32405\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32406, 1, 'WGS 72BE / UTM zone 6N', 'EPSG', 32406, 'PROJCS[\"WGS 72BE / UTM zone 6N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32406\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32407, 1, 'WGS 72BE / UTM zone 7N', 'EPSG', 32407, 'PROJCS[\"WGS 72BE / UTM zone 7N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32407\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32408, 1, 'WGS 72BE / UTM zone 8N', 'EPSG', 32408, 'PROJCS[\"WGS 72BE / UTM zone 8N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32408\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32409, 1, 'WGS 72BE / UTM zone 9N', 'EPSG', 32409, 'PROJCS[\"WGS 72BE / UTM zone 9N\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORI"
  "TY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32409\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32410, 1, 'WGS 72BE / UTM zone 10N', 'EPSG', 32410, 'PROJCS[\"WGS 72BE / UTM zone 10N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32410\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32411, 1, 'WGS 72BE / UTM zone 11N', 'EPSG', 32411, 'PROJCS[\"WGS 72BE / UTM zone 11N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32411\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32412, 1, 'WGS 72BE / UTM zone 12N', 'EPSG', 32412, 'PROJCS[\"WGS 72BE / UTM zone 12N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32412\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32413, 1, 'WGS 72BE / UTM zone 13N', 'EPSG', 32413, 'PROJCS[\"WGS 72BE / UTM zone 13N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUT"
  "HORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32413\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32414, 1, 'WGS 72BE / UTM zone 14N', 'EPSG', 32414, 'PROJCS[\"WGS 72BE / UTM zone 14N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32414\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32415, 1, 'WGS 72BE / UTM zone 15N', 'EPSG', 32415, 'PROJCS[\"WGS 72BE / UTM zone 15N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32415\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32416, 1, 'WGS 72BE / UTM zone 16N', 'EPSG', 32416, 'PROJCS[\"WGS 72BE / UTM zone 16N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32416\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32417, 1, 'WGS 72BE / UTM zone 17N', 'EPSG', 32417, 'PROJCS[\"WGS 72BE / UTM zone 17N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32417\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32418, 1, 'WGS 72BE / UTM zone 18N', 'EPSG', 32418, 'PROJCS[\"WGS 72BE / UTM zone 18N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32418\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32419, 1, 'WGS 72BE / UTM zone 19N', 'EPSG', 32419, 'PROJCS[\"WGS 72BE / UTM zone 19N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32419\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32420, 1, 'WGS 72BE / UTM zone 20N', 'EPSG', 32420, 'PROJCS[\"WGS 72BE / UTM zone 20N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32420\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32421, 1, 'WGS 72BE / UTM zone 21N', 'EPSG', 32421, 'PROJCS[\"WGS 72BE / UTM zone 21N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32421\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32422, 1, 'WGS 72BE / UTM zone 22N', 'EPSG', 32422, 'PROJCS[\"WGS 72BE / UTM zone 22N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32422\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32423, 1, 'WGS 72BE / UTM zone 23N', 'EPSG', 32423, 'PROJCS[\"WGS 72BE / UTM zone 23N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32423\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32424, 1, 'WGS 72BE / UTM zone 24N', 'EPSG', 32424, 'PROJCS[\"WGS 72BE / UTM zone 24N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32424\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32425, 1, 'WGS 72BE / UTM zone 25N', 'EPSG', 32425, 'PROJCS[\"WGS 72BE / UTM zone 25N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32425\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32426, 1, 'WGS 72BE / UTM zone 26N', 'EPSG', 32426, 'PROJCS[\"WGS 72BE / UTM zone 26N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32426\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32427, 1, 'WGS 72BE / UTM zone 27N', 'EPSG', 32427, 'PROJCS[\"WGS 72BE / UTM zone 27N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32427\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32428, 1, 'WGS 72BE / UTM zone 28N', 'EPSG', 32428, 'PROJCS[\"WGS 72BE / UTM zone 28N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32428\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32429, 1, 'WGS 72BE / UTM zone 29N', 'EPSG', 32429, 'PROJCS[\"WGS 72BE / UTM zone 29N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32429\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32430, 1, 'WGS 72BE / UTM zone 30N', 'EPSG', 32430, 'PROJCS[\"WGS 72BE / UTM zone 30N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32430\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32431, 1, 'WGS 72BE / UTM zone 31N', 'EPSG', 32431, 'PROJCS[\"WGS 72BE / UTM zone 31N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32431\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32432, 1, 'WGS 72BE / UTM zone 32N', 'EPSG', 32432, 'PROJCS[\"WGS 72BE / UTM zone 32N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32432\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32433, 1, 'WGS 72BE / UTM zone 33N', 'EPSG', 32433, 'PROJCS[\"WGS 72BE / UTM zone 33N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32433\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32434, 1, 'WGS 72BE / UTM zone 34N', 'EPSG', 32434, 'PROJCS[\"WGS 72BE / UTM zone 34N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32434\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32435, 1, 'WGS 72BE / UTM zone 35N', 'EPSG', 32435, 'PROJCS[\"WGS 72BE / UTM zone 35N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32435\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32436, 1, 'WGS 72BE / UTM zone 36N', 'EPSG', 32436, 'PROJCS[\"WGS 72BE / UTM zone 36N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32436\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32437, 1, 'WGS 72BE / UTM zone 37N', 'EPSG', 32437, 'PROJCS[\"WGS 72BE / UTM zone 37N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32437\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32438, 1, 'WGS 72BE / UTM zone 38N', 'EPSG', 32438, 'PROJCS[\"WGS 72BE / UTM zone 38N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32438\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32439, 1, 'WGS 72BE / UTM zone 39N', 'EPSG', 32439, 'PROJCS[\"WGS 72BE / UTM zone 39N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32439\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32440, 1, 'WGS 72BE / UTM zone 40N', 'EPSG', 32440, 'PROJCS[\"WGS 72BE / UTM zone 40N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32440\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32441, 1, 'WGS 72BE / UTM zone 41N', 'EPSG', 32441, 'PROJCS[\"WGS 72BE / UTM zone 41N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32441\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32442, 1, 'WGS 72BE / UTM zone 42N', 'EPSG', 32442, 'PROJCS[\"WGS 72BE / UTM zone 42N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32442\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32443, 1, 'WGS 72BE / UTM zone 43N', 'EPSG', 32443, 'PROJCS[\"WGS 72BE / UTM zone 43N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32443\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32444, 1, 'WGS 72BE / UTM zone 44N', 'EPSG', 32444, 'PROJCS[\"WGS 72BE / UTM zone 44N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32444\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32445, 1, 'WGS 72BE / UTM zone 45N', 'EPSG', 32445, 'PROJCS[\"WGS 72BE / UTM zone 45N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32445\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32446, 1, 'WGS 72BE / UTM zone 46N', 'EPSG', 32446, 'PROJCS[\"WGS 72BE / UTM zone 46N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32446\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32447, 1, 'WGS 72BE / UTM zone 47N', 'EPSG', 32447, 'PROJCS[\"WGS 72BE / UTM zone 47N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHOR"
  "ITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32447\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32448, 1, 'WGS 72BE / UTM zone 48N', 'EPSG', 32448, 'PROJCS[\"WGS 72BE / UTM zone 48N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32448\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32449, 1, 'WGS 72BE / UTM zone 49N', 'EPSG', 32449, 'PROJCS[\"WGS 72BE / UTM zone 49N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32449\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32450, 1, 'WGS 72BE / UTM zone 50N', 'EPSG', 32450, 'PROJCS[\"WGS 72BE / UTM zone 50N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32450\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32451, 1, 'WGS 72BE / UTM zone 51N', 'EPSG', 32451, 'PROJCS[\"WGS 72BE / UTM zone 51N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32451\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32452, 1, 'WGS 72BE / UTM zone 52N', 'EPSG', 32452, 'PROJCS[\"WGS 72BE / UTM zone 52N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32452\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32453, 1, 'WGS 72BE / UTM zone 53N', 'EPSG', 32453, 'PROJCS[\"WGS 72BE / UTM zone 53N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32453\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32454, 1, 'WGS 72BE / UTM zone 54N', 'EPSG', 32454, 'PROJCS[\"WGS 72BE / UTM zone 54N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32454\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32455, 1, 'WGS 72BE / UTM zone 55N', 'EPSG', 32455, 'PROJCS[\"WGS 72BE / UTM zone 55N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32455\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32456, 1, 'WGS 72BE / UTM zone 56N', 'EPSG', 32456, 'PROJCS[\"WGS 72BE / UTM zone 56N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32456\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32457, 1, 'WGS 72BE / UTM zone 57N', 'EPSG', 32457, 'PROJCS[\"WGS 72BE / UTM zone 57N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32457\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32458, 1, 'WGS 72BE / UTM zone 58N', 'EPSG', 32458, 'PROJCS[\"WGS 72BE / UTM zone 58N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32458\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32459, 1, 'WGS 72BE / UTM zone 59N', 'EPSG', 32459, 'PROJCS[\"WGS 72BE / UTM zone 59N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32459\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32460, 1, 'WGS 72BE / UTM zone 60N', 'EPSG', 32460, 'PROJCS[\"WGS 72BE / UTM zone 60N\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTH"
  "ORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32460\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32501, 1, 'WGS 72BE / UTM zone 1S', 'EPSG', 32501, 'PROJCS[\"WGS 72BE / UTM zone 1S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32501\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32502, 1, 'WGS 72BE / UTM zone 2S', 'EPSG', 32502, 'PROJCS[\"WGS 72BE / UTM zone 2S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32502\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32503, 1, 'WGS 72BE / UTM zone 3S', 'EPSG', 32503, 'PROJCS[\"WGS 72BE / UTM zone 3S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32503\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32504, 1, 'WGS 72BE / UTM zone 4S', 'EPSG', 32504, 'PROJCS[\"WGS 72BE / UTM zone 4S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32504\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32505, 1, 'WGS 72BE / UTM zone 5S', 'EPSG', 32505, 'PROJCS[\"WGS 72BE / UTM zone 5S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32505\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32506, 1, 'WGS 72BE / UTM zone 6S', 'EPSG', 32506, 'PROJCS[\"WGS 72BE / UTM zone 6S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32506\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32507, 1, 'WGS 72BE / UTM zone 7S', 'EPSG', 32507, 'PROJCS[\"WGS 72BE / UTM zone 7S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32507\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32508, 1, 'WGS 72BE / UTM zone 8S', 'EPSG', 32508, 'PROJCS[\"WGS 72BE / UTM zone 8S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32508\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32509, 1, 'WGS 72BE / UTM zone 9S', 'EPSG', 32509, 'PROJCS[\"WGS 72BE / UTM zone 9S\",GEOGCS[\""
  "WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"7043\""
  "]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]"
  "],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY"
  "[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural"
  " origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]"
  "],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50000"
  "0,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,"
  "AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32509\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32510, 1, 'WGS 72BE / UTM zone 10S', 'EPSG', 32510, 'PROJCS[\"WGS 72BE / UTM zone 10S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32510\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32511, 1, 'WGS 72BE / UTM zone 11S', 'EPSG', 32511, 'PROJCS[\"WGS 72BE / UTM zone 11S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32511\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32512, 1, 'WGS 72BE / UTM zone 12S', 'EPSG', 32512, 'PROJCS[\"WGS 72BE / UTM zone 12S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32512\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32513, 1, 'WGS 72BE / UTM zone 13S', 'EPSG', 32513, 'PROJCS[\"WGS 72BE / UTM zone 13S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"880"
  "2\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5"
  "00000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32513\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32514, 1, 'WGS 72BE / UTM zone 14S', 'EPSG', 32514, 'PROJCS[\"WGS 72BE / UTM zone 14S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32514\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32515, 1, 'WGS 72BE / UTM zone 15S', 'EPSG', 32515, 'PROJCS[\"WGS 72BE / UTM zone 15S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32515\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32516, 1, 'WGS 72BE / UTM zone 16S', 'EPSG', 32516, 'PROJCS[\"WGS 72BE / UTM zone 16S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32516\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32517, 1, 'WGS 72BE / UTM zone 17S', 'EPSG', 32517, 'PROJCS[\"WGS 72BE / UTM zone 17S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32517\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32518, 1, 'WGS 72BE / UTM zone 18S', 'EPSG', 32518, 'PROJCS[\"WGS 72BE / UTM zone 18S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32518\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32519, 1, 'WGS 72BE / UTM zone 19S', 'EPSG', 32519, 'PROJCS[\"WGS 72BE / UTM zone 19S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32519\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32520, 1, 'WGS 72BE / UTM zone 20S', 'EPSG', 32520, 'PROJCS[\"WGS 72BE / UTM zone 20S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32520\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32521, 1, 'WGS 72BE / UTM zone 21S', 'EPSG', 32521, 'PROJCS[\"WGS 72BE / UTM zone 21S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32521\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32522, 1, 'WGS 72BE / UTM zone 22S', 'EPSG', 32522, 'PROJCS[\"WGS 72BE / UTM zone 22S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32522\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32523, 1, 'WGS 72BE / UTM zone 23S', 'EPSG', 32523, 'PROJCS[\"WGS 72BE / UTM zone 23S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32523\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32524, 1, 'WGS 72BE / UTM zone 24S', 'EPSG', 32524, 'PROJCS[\"WGS 72BE / UTM zone 24S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32524\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32525, 1, 'WGS 72BE / UTM zone 25S', 'EPSG', 32525, 'PROJCS[\"WGS 72BE / UTM zone 25S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32525\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32526, 1, 'WGS 72BE / UTM zone 26S', 'EPSG', 32526, 'PROJCS[\"WGS 72BE / UTM zone 26S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32526\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32527, 1, 'WGS 72BE / UTM zone 27S', 'EPSG', 32527, 'PROJCS[\"WGS 72BE / UTM zone 27S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32527\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32528, 1, 'WGS 72BE / UTM zone 28S', 'EPSG', 32528, 'PROJCS[\"WGS 72BE / UTM zone 28S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32528\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32529, 1, 'WGS 72BE / UTM zone 29S', 'EPSG', 32529, 'PROJCS[\"WGS 72BE / UTM zone 29S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32529\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32530, 1, 'WGS 72BE / UTM zone 30S', 'EPSG', 32530, 'PROJCS[\"WGS 72BE / UTM zone 30S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32530\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32531, 1, 'WGS 72BE / UTM zone 31S', 'EPSG', 32531, 'PROJCS[\"WGS 72BE / UTM zone 31S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32531\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32532, 1, 'WGS 72BE / UTM zone 32S', 'EPSG', 32532, 'PROJCS[\"WGS 72BE / UTM zone 32S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32532\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32533, 1, 'WGS 72BE / UTM zone 33S', 'EPSG', 32533, 'PROJCS[\"WGS 72BE / UTM zone 33S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32533\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32534, 1, 'WGS 72BE / UTM zone 34S', 'EPSG', 32534, 'PROJCS[\"WGS 72BE / UTM zone 34S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32534\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32535, 1, 'WGS 72BE / UTM zone 35S', 'EPSG', 32535, 'PROJCS[\"WGS 72BE / UTM zone 35S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32535\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32536, 1, 'WGS 72BE / UTM zone 36S', 'EPSG', 32536, 'PROJCS[\"WGS 72BE / UTM zone 36S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32536\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32537, 1, 'WGS 72BE / UTM zone 37S', 'EPSG', 32537, 'PROJCS[\"WGS 72BE / UTM zone 37S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32537\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32538, 1, 'WGS 72BE / UTM zone 38S', 'EPSG', 32538, 'PROJCS[\"WGS 72BE / UTM zone 38S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32538\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32539, 1, 'WGS 72BE / UTM zone 39S', 'EPSG', 32539, 'PROJCS[\"WGS 72BE / UTM zone 39S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32539\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32540, 1, 'WGS 72BE / UTM zone 40S', 'EPSG', 32540, 'PROJCS[\"WGS 72BE / UTM zone 40S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32540\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32541, 1, 'WGS 72BE / UTM zone 41S', 'EPSG', 32541, 'PROJCS[\"WGS 72BE / UTM zone 41S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32541\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32542, 1, 'WGS 72BE / UTM zone 42S', 'EPSG', 32542, 'PROJCS[\"WGS 72BE / UTM zone 42S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32542\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32543, 1, 'WGS 72BE / UTM zone 43S', 'EPSG', 32543, 'PROJCS[\"WGS 72BE / UTM zone 43S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32543\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32544, 1, 'WGS 72BE / UTM zone 44S', 'EPSG', 32544, 'PROJCS[\"WGS 72BE / UTM zone 44S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32544\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32545, 1, 'WGS 72BE / UTM zone 45S', 'EPSG', 32545, 'PROJCS[\"WGS 72BE / UTM zone 45S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32545\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32546, 1, 'WGS 72BE / UTM zone 46S', 'EPSG', 32546, 'PROJCS[\"WGS 72BE / UTM zone 46S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32546\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32547, 1, 'WGS 72BE / UTM zone 47S', 'EPSG', 32547, 'PROJCS[\"WGS 72BE / UTM zone 47S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\""
  "]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",5000"
  "00,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1"
  ",AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32547\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32548, 1, 'WGS 72BE / UTM zone 48S', 'EPSG', 32548, 'PROJCS[\"WGS 72BE / UTM zone 48S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32548\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32549, 1, 'WGS 72BE / UTM zone 49S', 'EPSG', 32549, 'PROJCS[\"WGS 72BE / UTM zone 49S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32549\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32550, 1, 'WGS 72BE / UTM zone 50S', 'EPSG', 32550, 'PROJCS[\"WGS 72BE / UTM zone 50S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32550\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32551, 1, 'WGS 72BE / UTM zone 51S', 'EPSG', 32551, 'PROJCS[\"WGS 72BE / UTM zone 51S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32551\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32552, 1, 'WGS 72BE / UTM zone 52S', 'EPSG', 32552, 'PROJCS[\"WGS 72BE / UTM zone 52S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32552\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32553, 1, 'WGS 72BE / UTM zone 53S', 'EPSG', 32553, 'PROJCS[\"WGS 72BE / UTM zone 53S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32553\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32554, 1, 'WGS 72BE / UTM zone 54S', 'EPSG', 32554, 'PROJCS[\"WGS 72BE / UTM zone 54S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32554\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32555, 1, 'WGS 72BE / UTM zone 55S', 'EPSG', 32555, 'PROJCS[\"WGS 72BE / UTM zone 55S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32555\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32556, 1, 'WGS 72BE / UTM zone 56S', 'EPSG', 32556, 'PROJCS[\"WGS 72BE / UTM zone 56S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32556\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32557, 1, 'WGS 72BE / UTM zone 57S', 'EPSG', 32557, 'PROJCS[\"WGS 72BE / UTM zone 57S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32557\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32558, 1, 'WGS 72BE / UTM zone 58S', 'EPSG', 32558, 'PROJCS[\"WGS 72BE / UTM zone 58S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32558\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32559, 1, 'WGS 72BE / UTM zone 59S', 'EPSG', 32559, 'PROJCS[\"WGS 72BE / UTM zone 59S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32559\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32560, 1, 'WGS 72BE / UTM zone 60S', 'EPSG', 32560, 'PROJCS[\"WGS 72BE / UTM zone 60S\",GEOGCS"
  "[\"WGS 72BE\",DATUM[\"WGS 72 Transit Broadcast Ephemeris\",SPHEROID[\"WGS 72\",6378135,298.26,AUTHORITY[\"EPSG\",\"70"
  "43\"]],TOWGS84[0,0,1.9,0,0,0.814,-0.38],AUTHORITY[\"EPSG\",\"6324\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"890"
  "1\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHO"
  "RITY[\"EPSG\",\"4324\"]],PROJECTION[\"Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of nat"
  "ural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802"
  "\"]],PARAMETER[\"Scale factor at natural origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",50"
  "0000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\""
  ",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32560\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32600, 1, 'WGS 84 / UTM grid system (northern hemisphere)', 'EPSG', 32600, 'PROJCS[\"WGS 84 / "
  "UTM grid system (northern hemisphere)\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",63"
  "78137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercator Zoned Grid System\",AUTHORITY[\"EPSG\",\"9824\"]]"
  ",PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Initial longitude\",-180,AUTHOR"
  "ITY[\"EPSG\",\"8830\"]],PARAMETER[\"Zone width\",6,AUTHORITY[\"EPSG\",\"8831\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32600\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32601, 1, 'WGS 84 / UTM zone 1N', 'EPSG', 32601, 'PROJCS[\"WGS 84 / UTM zone 1N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32601\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32602, 1, 'WGS 84 / UTM zone 2N', 'EPSG', 32602, 'PROJCS[\"WGS 84 / UTM zone 2N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32602\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32603, 1, 'WGS 84 / UTM zone 3N', 'EPSG', 32603, 'PROJCS[\"WGS 84 / UTM zone 3N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32603\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32604, 1, 'WGS 84 / UTM zone 4N', 'EPSG', 32604, 'PROJCS[\"WGS 84 / UTM zone 4N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32604\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32605, 1, 'WGS 84 / UTM zone 5N', 'EPSG', 32605, 'PROJCS[\"WGS 84 / UTM zone 5N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32605\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32606, 1, 'WGS 84 / UTM zone 6N', 'EPSG', 32606, 'PROJCS[\"WGS 84 / UTM zone 6N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32606\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32607, 1, 'WGS 84 / UTM zone 7N', 'EPSG', 32607, 'PROJCS[\"WGS 84 / UTM zone 7N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32607\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32608, 1, 'WGS 84 / UTM zone 8N', 'EPSG', 32608, 'PROJCS[\"WGS 84 / UTM zone 8N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32608\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32609, 1, 'WGS 84 / UTM zone 9N', 'EPSG', 32609, 'PROJCS[\"WGS 84 / UTM zone 9N\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32609\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32610, 1, 'WGS 84 / UTM zone 10N', 'EPSG', 32610, 'PROJCS[\"WGS 84 / UTM zone 10N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32610\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32611, 1, 'WGS 84 / UTM zone 11N', 'EPSG', 32611, 'PROJCS[\"WGS 84 / UTM zone 11N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32611\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32612, 1, 'WGS 84 / UTM zone 12N', 'EPSG', 32612, 'PROJCS[\"WGS 84 / UTM zone 12N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32612\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32613, 1, 'WGS 84 / UTM zone 13N', 'EPSG', 32613, 'PROJCS[\"WGS 84 / UTM zone 13N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EA"
  "ST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32613\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32614, 1, 'WGS 84 / UTM zone 14N', 'EPSG', 32614, 'PROJCS[\"WGS 84 / UTM zone 14N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32614\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32615, 1, 'WGS 84 / UTM zone 15N', 'EPSG', 32615, 'PROJCS[\"WGS 84 / UTM zone 15N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32615\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32616, 1, 'WGS 84 / UTM zone 16N', 'EPSG', 32616, 'PROJCS[\"WGS 84 / UTM zone 16N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32616\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32617, 1, 'WGS 84 / UTM zone 17N', 'EPSG', 32617, 'PROJCS[\"WGS 84 / UTM zone 17N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32617\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32618, 1, 'WGS 84 / UTM zone 18N', 'EPSG', 32618, 'PROJCS[\"WGS 84 / UTM zone 18N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32618\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32619, 1, 'WGS 84 / UTM zone 19N', 'EPSG', 32619, 'PROJCS[\"WGS 84 / UTM zone 19N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32619\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32620, 1, 'WGS 84 / UTM zone 20N', 'EPSG', 32620, 'PROJCS[\"WGS 84 / UTM zone 20N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32620\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32621, 1, 'WGS 84 / UTM zone 21N', 'EPSG', 32621, 'PROJCS[\"WGS 84 / UTM zone 21N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32621\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32622, 1, 'WGS 84 / UTM zone 22N', 'EPSG', 32622, 'PROJCS[\"WGS 84 / UTM zone 22N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32622\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32623, 1, 'WGS 84 / UTM zone 23N', 'EPSG', 32623, 'PROJCS[\"WGS 84 / UTM zone 23N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32623\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32624, 1, 'WGS 84 / UTM zone 24N', 'EPSG', 32624, 'PROJCS[\"WGS 84 / UTM zone 24N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32624\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32625, 1, 'WGS 84 / UTM zone 25N', 'EPSG', 32625, 'PROJCS[\"WGS 84 / UTM zone 25N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32625\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32626, 1, 'WGS 84 / UTM zone 26N', 'EPSG', 32626, 'PROJCS[\"WGS 84 / UTM zone 26N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32626\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32627, 1, 'WGS 84 / UTM zone 27N', 'EPSG', 32627, 'PROJCS[\"WGS 84 / UTM zone 27N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32627\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32628, 1, 'WGS 84 / UTM zone 28N', 'EPSG', 32628, 'PROJCS[\"WGS 84 / UTM zone 28N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32628\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32629, 1, 'WGS 84 / UTM zone 29N', 'EPSG', 32629, 'PROJCS[\"WGS 84 / UTM zone 29N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32629\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32630, 1, 'WGS 84 / UTM zone 30N', 'EPSG', 32630, 'PROJCS[\"WGS 84 / UTM zone 30N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32630\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32631, 1, 'WGS 84 / UTM zone 31N', 'EPSG', 32631, 'PROJCS[\"WGS 84 / UTM zone 31N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32631\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32632, 1, 'WGS 84 / UTM zone 32N', 'EPSG', 32632, 'PROJCS[\"WGS 84 / UTM zone 32N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32632\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32633, 1, 'WGS 84 / UTM zone 33N', 'EPSG', 32633, 'PROJCS[\"WGS 84 / UTM zone 33N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32633\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32634, 1, 'WGS 84 / UTM zone 34N', 'EPSG', 32634, 'PROJCS[\"WGS 84 / UTM zone 34N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32634\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32635, 1, 'WGS 84 / UTM zone 35N', 'EPSG', 32635, 'PROJCS[\"WGS 84 / UTM zone 35N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32635\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32636, 1, 'WGS 84 / UTM zone 36N', 'EPSG', 32636, 'PROJCS[\"WGS 84 / UTM zone 36N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32636\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32637, 1, 'WGS 84 / UTM zone 37N', 'EPSG', 32637, 'PROJCS[\"WGS 84 / UTM zone 37N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32637\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32638, 1, 'WGS 84 / UTM zone 38N', 'EPSG', 32638, 'PROJCS[\"WGS 84 / UTM zone 38N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32638\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32639, 1, 'WGS 84 / UTM zone 39N', 'EPSG', 32639, 'PROJCS[\"WGS 84 / UTM zone 39N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32639\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32640, 1, 'WGS 84 / UTM zone 40N', 'EPSG', 32640, 'PROJCS[\"WGS 84 / UTM zone 40N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32640\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32641, 1, 'WGS 84 / UTM zone 41N', 'EPSG', 32641, 'PROJCS[\"WGS 84 / UTM zone 41N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32641\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32642, 1, 'WGS 84 / UTM zone 42N', 'EPSG', 32642, 'PROJCS[\"WGS 84 / UTM zone 42N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32642\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32643, 1, 'WGS 84 / UTM zone 43N', 'EPSG', 32643, 'PROJCS[\"WGS 84 / UTM zone 43N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32643\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32644, 1, 'WGS 84 / UTM zone 44N', 'EPSG', 32644, 'PROJCS[\"WGS 84 / UTM zone 44N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32644\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32645, 1, 'WGS 84 / UTM zone 45N', 'EPSG', 32645, 'PROJCS[\"WGS 84 / UTM zone 45N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32645\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32646, 1, 'WGS 84 / UTM zone 46N', 'EPSG', 32646, 'PROJCS[\"WGS 84 / UTM zone 46N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32646\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32647, 1, 'WGS 84 / UTM zone 47N', 'EPSG', 32647, 'PROJCS[\"WGS 84 / UTM zone 47N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST"
  "],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32647\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32648, 1, 'WGS 84 / UTM zone 48N', 'EPSG', 32648, 'PROJCS[\"WGS 84 / UTM zone 48N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32648\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32649, 1, 'WGS 84 / UTM zone 49N', 'EPSG', 32649, 'PROJCS[\"WGS 84 / UTM zone 49N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32649\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32650, 1, 'WGS 84 / UTM zone 50N', 'EPSG', 32650, 'PROJCS[\"WGS 84 / UTM zone 50N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32650\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32651, 1, 'WGS 84 / UTM zone 51N', 'EPSG', 32651, 'PROJCS[\"WGS 84 / UTM zone 51N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32651\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32652, 1, 'WGS 84 / UTM zone 52N', 'EPSG', 32652, 'PROJCS[\"WGS 84 / UTM zone 52N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32652\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32653, 1, 'WGS 84 / UTM zone 53N', 'EPSG', 32653, 'PROJCS[\"WGS 84 / UTM zone 53N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32653\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32654, 1, 'WGS 84 / UTM zone 54N', 'EPSG', 32654, 'PROJCS[\"WGS 84 / UTM zone 54N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32654\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32655, 1, 'WGS 84 / UTM zone 55N', 'EPSG', 32655, 'PROJCS[\"WGS 84 / UTM zone 55N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32655\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32656, 1, 'WGS 84 / UTM zone 56N', 'EPSG', 32656, 'PROJCS[\"WGS 84 / UTM zone 56N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32656\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32657, 1, 'WGS 84 / UTM zone 57N', 'EPSG', 32657, 'PROJCS[\"WGS 84 / UTM zone 57N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32657\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32658, 1, 'WGS 84 / UTM zone 58N', 'EPSG', 32658, 'PROJCS[\"WGS 84 / UTM zone 58N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32658\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32659, 1, 'WGS 84 / UTM zone 59N', 'EPSG', 32659, 'PROJCS[\"WGS 84 / UTM zone 59N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32659\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32660, 1, 'WGS 84 / UTM zone 60N', 'EPSG', 32660, 'PROJCS[\"WGS 84 / UTM zone 60N\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAS"
  "T],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32660\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32661, 1, 'WGS 84 / UPS North (N,E)', 'EPSG', 32661, 'PROJCS[\"WGS 84 / UPS North (N,E)\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitude of natural origin\",90,AUTHO"
  "RITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale "
  "factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2000000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"N\",SOUTH],AXIS[\"E\",SOUTH],AUTHORITY[\"EPSG\",\"32661\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32664, 1, 'WGS 84 / BLM 14N (ftUS)', 'EPSG', 32664, 'PROJCS[\"WGS 84 / BLM 14N (ftUS)\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32664\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32665, 1, 'WGS 84 / BLM 15N (ftUS)', 'EPSG', 32665, 'PROJCS[\"WGS 84 / BLM 15N (ftUS)\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32665\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32666, 1, 'WGS 84 / BLM 16N (ftUS)', 'EPSG', 32666, 'PROJCS[\"WGS 84 / BLM 16N (ftUS)\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32666\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32667, 1, 'WGS 84 / BLM 17N (ftUS)', 'EPSG', 32667, 'PROJCS[\"WGS 84 / BLM 17N (ftUS)\",GEOGCS"
  "[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\""
  "]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199"
  "43278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\""
  "Transverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8"
  "801\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natur"
  "al origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",1640416.67,AUTHORITY[\"EPSG\",\"8806\"]]"
  ",PARAMETER[\"False northing\",0,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"US survey foot\",0.30480060960121924,AUTHORITY[\""
  "EPSG\",\"9003\"]],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH],AUTHORITY[\"EPSG\",\"32667\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32700, 1, 'WGS 84 / UTM grid system (southern hemisphere)', 'EPSG', 32700, 'PROJCS[\"WGS 84 / "
  "UTM grid system (southern hemisphere)\",GEOGCS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",63"
  "78137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EP"
  "SG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",E"
  "AST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse Mercator Zoned Grid System\",AUTHORITY[\"EPSG\",\"9824\"]]"
  ",PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Initial longitude\",-180,AUTHOR"
  "ITY[\"EPSG\",\"8830\"]],PARAMETER[\"Zone width\",6,AUTHORITY[\"EPSG\",\"8831\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32700\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32701, 1, 'WGS 84 / UTM zone 1S', 'EPSG', 32701, 'PROJCS[\"WGS 84 / UTM zone 1S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32701\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32702, 1, 'WGS 84 / UTM zone 2S', 'EPSG', 32702, 'PROJCS[\"WGS 84 / UTM zone 2S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32702\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32703, 1, 'WGS 84 / UTM zone 3S', 'EPSG', 32703, 'PROJCS[\"WGS 84 / UTM zone 3S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32703\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32704, 1, 'WGS 84 / UTM zone 4S', 'EPSG', 32704, 'PROJCS[\"WGS 84 / UTM zone 4S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32704\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32705, 1, 'WGS 84 / UTM zone 5S', 'EPSG', 32705, 'PROJCS[\"WGS 84 / UTM zone 5S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32705\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32706, 1, 'WGS 84 / UTM zone 6S', 'EPSG', 32706, 'PROJCS[\"WGS 84 / UTM zone 6S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32706\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32707, 1, 'WGS 84 / UTM zone 7S', 'EPSG', 32707, 'PROJCS[\"WGS 84 / UTM zone 7S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32707\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32708, 1, 'WGS 84 / UTM zone 8S', 'EPSG', 32708, 'PROJCS[\"WGS 84 / UTM zone 8S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32708\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32709, 1, 'WGS 84 / UTM zone 9S', 'EPSG', 32709, 'PROJCS[\"WGS 84 / UTM zone 9S\",GEOGCS[\"WGS"
  " 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AU"
  "THORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278"
  ",AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tran"
  "sverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\""
  "]],PARAMETER[\"Longitude of natural origin\",-129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32709\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32710, 1, 'WGS 84 / UTM zone 10S', 'EPSG', 32710, 'PROJCS[\"WGS 84 / UTM zone 10S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32710\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32711, 1, 'WGS 84 / UTM zone 11S', 'EPSG', 32711, 'PROJCS[\"WGS 84 / UTM zone 11S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32711\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32712, 1, 'WGS 84 / UTM zone 12S', 'EPSG', 32712, 'PROJCS[\"WGS 84 / UTM zone 12S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32712\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32713, 1, 'WGS 84 / UTM zone 13S', 'EPSG', 32713, 'PROJCS[\"WGS 84 / UTM zone 13S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natura"
  "l origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARA"
  "METER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32713\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32714, 1, 'WGS 84 / UTM zone 14S', 'EPSG', 32714, 'PROJCS[\"WGS 84 / UTM zone 14S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32714\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32715, 1, 'WGS 84 / UTM zone 15S', 'EPSG', 32715, 'PROJCS[\"WGS 84 / UTM zone 15S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32715\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32716, 1, 'WGS 84 / UTM zone 16S', 'EPSG', 32716, 'PROJCS[\"WGS 84 / UTM zone 16S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32716\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32717, 1, 'WGS 84 / UTM zone 17S', 'EPSG', 32717, 'PROJCS[\"WGS 84 / UTM zone 17S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32717\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32718, 1, 'WGS 84 / UTM zone 18S', 'EPSG', 32718, 'PROJCS[\"WGS 84 / UTM zone 18S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32718\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32719, 1, 'WGS 84 / UTM zone 19S', 'EPSG', 32719, 'PROJCS[\"WGS 84 / UTM zone 19S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32719\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32720, 1, 'WGS 84 / UTM zone 20S', 'EPSG', 32720, 'PROJCS[\"WGS 84 / UTM zone 20S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32720\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32721, 1, 'WGS 84 / UTM zone 21S', 'EPSG', 32721, 'PROJCS[\"WGS 84 / UTM zone 21S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32721\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32722, 1, 'WGS 84 / UTM zone 22S', 'EPSG', 32722, 'PROJCS[\"WGS 84 / UTM zone 22S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32722\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32723, 1, 'WGS 84 / UTM zone 23S', 'EPSG', 32723, 'PROJCS[\"WGS 84 / UTM zone 23S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32723\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32724, 1, 'WGS 84 / UTM zone 24S', 'EPSG', 32724, 'PROJCS[\"WGS 84 / UTM zone 24S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32724\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32725, 1, 'WGS 84 / UTM zone 25S', 'EPSG', 32725, 'PROJCS[\"WGS 84 / UTM zone 25S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32725\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32726, 1, 'WGS 84 / UTM zone 26S', 'EPSG', 32726, 'PROJCS[\"WGS 84 / UTM zone 26S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32726\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32727, 1, 'WGS 84 / UTM zone 27S', 'EPSG', 32727, 'PROJCS[\"WGS 84 / UTM zone 27S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32727\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32728, 1, 'WGS 84 / UTM zone 28S', 'EPSG', 32728, 'PROJCS[\"WGS 84 / UTM zone 28S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32728\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32729, 1, 'WGS 84 / UTM zone 29S', 'EPSG', 32729, 'PROJCS[\"WGS 84 / UTM zone 29S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32729\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32730, 1, 'WGS 84 / UTM zone 30S', 'EPSG', 32730, 'PROJCS[\"WGS 84 / UTM zone 30S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",-3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32730\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32731, 1, 'WGS 84 / UTM zone 31S', 'EPSG', 32731, 'PROJCS[\"WGS 84 / UTM zone 31S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",3,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32731\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32732, 1, 'WGS 84 / UTM zone 32S', 'EPSG', 32732, 'PROJCS[\"WGS 84 / UTM zone 32S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",9,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural o"
  "rigin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMET"
  "ER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\""
  ",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32732\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32733, 1, 'WGS 84 / UTM zone 33S', 'EPSG', 32733, 'PROJCS[\"WGS 84 / UTM zone 33S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",15,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32733\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32734, 1, 'WGS 84 / UTM zone 34S', 'EPSG', 32734, 'PROJCS[\"WGS 84 / UTM zone 34S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",21,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32734\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32735, 1, 'WGS 84 / UTM zone 35S', 'EPSG', 32735, 'PROJCS[\"WGS 84 / UTM zone 35S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",27,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32735\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32736, 1, 'WGS 84 / UTM zone 36S', 'EPSG', 32736, 'PROJCS[\"WGS 84 / UTM zone 36S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",33,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32736\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32737, 1, 'WGS 84 / UTM zone 37S', 'EPSG', 32737, 'PROJCS[\"WGS 84 / UTM zone 37S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",39,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32737\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32738, 1, 'WGS 84 / UTM zone 38S', 'EPSG', 32738, 'PROJCS[\"WGS 84 / UTM zone 38S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",45,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32738\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32739, 1, 'WGS 84 / UTM zone 39S', 'EPSG', 32739, 'PROJCS[\"WGS 84 / UTM zone 39S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",51,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32739\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32740, 1, 'WGS 84 / UTM zone 40S', 'EPSG', 32740, 'PROJCS[\"WGS 84 / UTM zone 40S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",57,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32740\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32741, 1, 'WGS 84 / UTM zone 41S', 'EPSG', 32741, 'PROJCS[\"WGS 84 / UTM zone 41S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",63,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32741\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32742, 1, 'WGS 84 / UTM zone 42S', 'EPSG', 32742, 'PROJCS[\"WGS 84 / UTM zone 42S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",69,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32742\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32743, 1, 'WGS 84 / UTM zone 43S', 'EPSG', 32743, 'PROJCS[\"WGS 84 / UTM zone 43S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",75,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32743\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32744, 1, 'WGS 84 / UTM zone 44S', 'EPSG', 32744, 'PROJCS[\"WGS 84 / UTM zone 44S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",81,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32744\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32745, 1, 'WGS 84 / UTM zone 45S', 'EPSG', 32745, 'PROJCS[\"WGS 84 / UTM zone 45S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",87,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32745\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32746, 1, 'WGS 84 / UTM zone 46S', 'EPSG', 32746, 'PROJCS[\"WGS 84 / UTM zone 46S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",93,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32746\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32747, 1, 'WGS 84 / UTM zone 47S', 'EPSG', 32747, 'PROJCS[\"WGS 84 / UTM zone 47S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",99,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural "
  "origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAME"
  "TER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E"
  "\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32747\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32748, 1, 'WGS 84 / UTM zone 48S', 'EPSG', 32748, 'PROJCS[\"WGS 84 / UTM zone 48S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",105,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32748\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32749, 1, 'WGS 84 / UTM zone 49S', 'EPSG', 32749, 'PROJCS[\"WGS 84 / UTM zone 49S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",111,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32749\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32750, 1, 'WGS 84 / UTM zone 50S', 'EPSG', 32750, 'PROJCS[\"WGS 84 / UTM zone 50S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",117,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32750\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32751, 1, 'WGS 84 / UTM zone 51S', 'EPSG', 32751, 'PROJCS[\"WGS 84 / UTM zone 51S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",123,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32751\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32752, 1, 'WGS 84 / UTM zone 52S', 'EPSG', 32752, 'PROJCS[\"WGS 84 / UTM zone 52S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",129,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32752\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32753, 1, 'WGS 84 / UTM zone 53S', 'EPSG', 32753, 'PROJCS[\"WGS 84 / UTM zone 53S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",135,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32753\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32754, 1, 'WGS 84 / UTM zone 54S', 'EPSG', 32754, 'PROJCS[\"WGS 84 / UTM zone 54S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",141,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32754\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32755, 1, 'WGS 84 / UTM zone 55S', 'EPSG', 32755, 'PROJCS[\"WGS 84 / UTM zone 55S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",147,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32755\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32756, 1, 'WGS 84 / UTM zone 56S', 'EPSG', 32756, 'PROJCS[\"WGS 84 / UTM zone 56S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",153,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32756\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32757, 1, 'WGS 84 / UTM zone 57S', 'EPSG', 32757, 'PROJCS[\"WGS 84 / UTM zone 57S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",159,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32757\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32758, 1, 'WGS 84 / UTM zone 58S', 'EPSG', 32758, 'PROJCS[\"WGS 84 / UTM zone 58S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",165,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32758\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32759, 1, 'WGS 84 / UTM zone 59S', 'EPSG', 32759, 'PROJCS[\"WGS 84 / UTM zone 59S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",171,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32759\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32760, 1, 'WGS 84 / UTM zone 60S', 'EPSG', 32760, 'PROJCS[\"WGS 84 / UTM zone 60S\",GEOGCS[\"W"
  "GS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],"
  "AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925199432"
  "78,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Tr"
  "ansverse Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"880"
  "1\"]],PARAMETER[\"Longitude of natural origin\",177,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural"
  " origin\",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAM"
  "ETER[\"False northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\""
  "E\",EAST],AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32760\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32761, 1, 'WGS 84 / UPS South (N,E)', 'EPSG', 32761, 'PROJCS[\"WGS 84 / UPS South (N,E)\",GEOG"
  "CS[\"WGS 84\",DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"703"
  "0\"]],AUTHORITY[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.0174532925"
  "19943278,AUTHORITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTIO"
  "N[\"Polar Stereographic (variant A)\",AUTHORITY[\"EPSG\",\"9810\"]],PARAMETER[\"Latitude of natural origin\",-90,AUTH"
  "ORITY[\"EPSG\",\"8801\"]],PARAMETER[\"Longitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale"
  " factor at natural origin\",0.994,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",2000000,AUTHORITY[\"EPSG\""
  ",\"8806\"]],PARAMETER[\"False northing\",2000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\""
  "9001\"]],AXIS[\"N\",NORTH],AXIS[\"E\",NORTH],AUTHORITY[\"EPSG\",\"32761\"]]', NULL);\n",
"REPLACE INTO mysql.st_spatial_reference_systems(id, catalog_id, name, organization, organization_coordsys_id, definition"
  ", description) VALUES (32766, 1, 'WGS 84 / TM 36 SE', 'EPSG', 32766, 'PROJCS[\"WGS 84 / TM 36 SE\",GEOGCS[\"WGS 84\","
  "DATUM[\"World Geodetic System 1984\",SPHEROID[\"WGS 84\",6378137,298.257223563,AUTHORITY[\"EPSG\",\"7030\"]],AUTHORIT"
  "Y[\"EPSG\",\"6326\"]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.017453292519943278,AUTHO"
  "RITY[\"EPSG\",\"9122\"]],AXIS[\"Lat\",NORTH],AXIS[\"Lon\",EAST],AUTHORITY[\"EPSG\",\"4326\"]],PROJECTION[\"Transverse"
  " Mercator\",AUTHORITY[\"EPSG\",\"9807\"]],PARAMETER[\"Latitude of natural origin\",0,AUTHORITY[\"EPSG\",\"8801\"]],PA"
  "RAMETER[\"Longitude of natural origin\",36,AUTHORITY[\"EPSG\",\"8802\"]],PARAMETER[\"Scale factor at natural origin\""
  ",0.9996,AUTHORITY[\"EPSG\",\"8805\"]],PARAMETER[\"False easting\",500000,AUTHORITY[\"EPSG\",\"8806\"]],PARAMETER[\"Fa"
  "lse northing\",10000000,AUTHORITY[\"EPSG\",\"8807\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],AXIS[\"E\",EAST]"
  ",AXIS[\"N\",NORTH],AUTHORITY[\"EPSG\",\"32766\"]]', NULL);\n",
"COMMIT;\n",
"START TRANSACTION;\n",
"DELETE FROM mysql.plugin WHERE name='mysqlx';\n",
"COMMIT;\n",
"START TRANSACTION;\n",
"INSERT INTO mysql.replication_group_member_actions (name, event, enabled, type, priority, error_handling) VALUES (\"mysq"
  "l_disable_super_read_only_if_primary\", \"AFTER_PRIMARY_ELECTION\", 1, \"INTERNAL\", 1, \"IGNORE\");\n",
"INSERT INTO mysql.replication_group_member_actions (name, event, enabled, type, priority, error_handling) VALUES (\"mysq"
  "l_start_failover_channels_if_primary\", \"AFTER_PRIMARY_ELECTION\", 1, \"INTERNAL\", 10, \"CRITICAL\");\n",
"INSERT INTO mysql.replication_group_configuration_version (name, version) VALUES (\"replication_group_member_actions\", "
  "1);\n",
"COMMIT;\n",
nullptr
};
